/* MAINTAINER: Bernard */

// #define WANT_TRACE

#include "SDL.h"
#include "SDL_ttf.h"

#include "application_client.h"

#include "gfx.h"
#include "sfx.h"
#include "game_client.h"
#include "commander_client.h"
#include "networkmanager_client.h"
#include "inputmanager.h"
#include "keybindings.h"
#include "console.h"
#include "player.h"

#include "log.h"
#include "settings.h"

#include "translator.h"
#include "platform.h"

#include "gui_constants.h"
#include "thread.h"

#include "menubasic.h"
#include "menumain.h"

#include "iortransfer_client.h"
#include "timemanager.h"

ApplicationClient::ApplicationClient(int argc, char* argv[])
:Application(argc, argv)
{
	TIN

	mInputManager = new InputManager;
	mKeyBindings = new KeyBindings;

	mGFX = new GFX;
	mSFX = new SFX;
	
	mGameClient           = new GameClient;
	mConsole              = new Console;
	mCommanderClient      = new CommanderClient;
	mNetworkManagerClient = new NetworkManagerClient;

	if (!mInputManager || !mKeyBindings || !mGFX || !mSFX ||
	    !mGameClient || !mConsole || !mCommanderClient || !mNetworkManagerClient)
	{
		DIE("new failed");
	}
}


ApplicationClient::~ApplicationClient()
{
	TIN
	delete mInputManager;
	delete mKeyBindings;
	delete mGFX;
	delete mSFX;
	delete mGameClient;
	delete mConsole;
	delete mCommanderClient;
	delete mNetworkManagerClient;
}

ApplicationClient::ApplicationClient(ApplicationClient &copy)
:Application(0, 0)
{
	TIN
	
	LOGE << "It is forbidden to make copy of ApplicationClient." << endl;
	DIE();
}

void ApplicationClient::init()
{
	TIN
	
	// load user settings
	loadUserConfigFile(SETTING_CLIENT_CONFIG_FILE, SETTING_CLIENT_CONFIG_FILE_TEMPLATE);

	g_settings.parseCommandLine(mArgc, mArgv); // reparse the command line


	// Log - open logfile and set log level
	g_log.init(g_settings.getString(SETTING_LOGFILE_CLIENT).c_str(),
			   g_settings.isTrue(SETTING_TOSTDERR_CLIENT));
	g_log.set_loglevel(g_settings.getInt(SETTING_LOGLEVEL));


	LOGI << "Client setting and logging initialized." << endl;
	logBasicInfo();

	// Load gui config
	gc.load();

	// Load translations
	g_translator.loadTranslationsList( locateFile(g_settings.getString(SETTING_TRANSLATIONS_LIST)).c_str() );	
	g_translator.loadFile( locateFile(g_settings.getString(SETTING_TRANSLATION_FILE)).c_str() );
	


	// gfx
	mGFX->init();
	
	// sfx
	mSFX->init();

	// init for font rendering - SDL_ttf
	if (!TTF_WasInit() && TTF_Init() == -1)
	{
		LOGE << "Unable to init TTF: " << TTF_GetError() << endl;
		DIE();
	}
	
	//
	mConsole->init();

	
	// Enable key repeat
	if (SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL) < 0)
	{
		LOGE << "SDL_EnableKeyRepeat failed :-(" << endl;
	}

	// prepare tasks
	mTaskManager->add(FontManager::getRef());
	// probably task for music..

	mNetworkManagerClient->init();

	// Add key bindings & probably load some default
	mKeyBindings->init();
	mInputManager->addInputAcceptor(*mKeyBindings);

	//TODO: load some binding default
	mCommanderClient->addCommand("exec defaultkeybind.txt");
	mCommanderClient->processAllCommands();
	mCommanderClient->processAllCommands();
}

void ApplicationClient::run()
{
	TIN

	//runGame();	
	runMenu();	
}

void ApplicationClient::clean()
{
	TIN

	mNetworkManagerClient->clean();

	mConsole->clean();
		
		
	// sfx
	mSFX->clean();
	
	// gfx
	mGFX->clean();
	
	// save settings
	saveUserConfigFile(SETTING_CLIENT_CONFIG_FILE, SETTING_CLIENT_CONFIG_FILE_TEMPLATE);

	// stop SDL_TTF if we were using it - maybe this could be call to FontManager...
	if (TTF_WasInit()) TTF_Quit();
}


GFX& ApplicationClient::getGFX()
{
	return *mGFX;
}

SFX& ApplicationClient::getSFX()
{
	return *mSFX;
}

GameClient& ApplicationClient::getGame()
{
	return *mGameClient;
}
	
Game& ApplicationClient::getGameCommon()
{
	return *mGameClient;
}

InputManager& ApplicationClient::getInputManager()
{ 
	TIN

	return *mInputManager; 
}

Console& ApplicationClient::getConsole()
{
	TIN

	return *mConsole;
}


Commander& ApplicationClient::getCommander()
{
	return *mCommanderClient;
}

NetworkManagerClient& ApplicationClient::getNetworkManager()
{
	return *mNetworkManagerClient;
}

KeyBindings& ApplicationClient::getKeyBindings()
{
	return *mKeyBindings;
}


void ApplicationClient::loop(Frame &frame, TaskManager &tm)
{
	// Push new frame to stack and deactivate the old one
	if (!mFrameStack.empty())
	{
		assertL(*mFrameStack.begin() != 0)
		(*mFrameStack.begin())->deactivated();
	}
	mFrameStack.push_front(&frame);

	// Do the "main" loop
	while (frame.wantRun())
	{
		tm.update();

		// this may be in a different thread
		getInputManager().checkSDLEvents();
		getInputManager().processSDLEvents();

		getCommander().processAllCommands(); 

		// This is here for better multithread performance
		Thread::reshedule();
	}

	// Pop the old frame from stack
	mFrameStack.pop_front();
	if (!mFrameStack.empty())
	{
		assertL(*mFrameStack.begin() != 0)
		(*mFrameStack.begin())->activated();
	}

}

void ApplicationClient::runMenu()
{

	MenuMain mm;
	mm.init();
	mm.run();
	mm.clean();
}

void ApplicationClient::runGame(const ServerInfo  &sinfo)
{		
	mGFX->showLoadingScreen();	

	AsyncConnect::resultCPtr acResult = 
		mNetworkManagerClient->connect( sinfo.IOR, g_settings.getString(SETTING_PLAYER_PREFERED_NAME));
	
	// Passive waiting for connect
	acResult->waitForResult();
	
	// check if we are connected or not
	if (!mNetworkManagerClient->connected()) 
	{
		LOGW << "Unable to connect to server. Maybe netwok error or server refused to connect us" << endl;
		getConsole().errorMessage(_("Connection to the server failed."));
		return;
	}

	// We are connected so let's go and play :)
	mGameClient->init();

	mReconnect = true;
	while (mReconnect)
	{		
		mReconnect = false;
		
		mGFX->showLoadingScreen();

		// get game info - blocks and waits for result
		// gameInfo must be deleted		
		AsyncGameInfo::resultCPtr agiResult = mNetworkManagerClient->getGameInfo();
		GameInfo * gameInfo = agiResult->getResult();

		mGameClient->loadMap(gameInfo->mapName);

		// gameInfo must be released
		delete gameInfo;

		// Let the fun begin
		mGameClient->run();		
	}
	
	mGameClient->clean();					
	mNetworkManagerClient->disconnect();
	// Clean old events
	mGameClient->getEventQueue().cleanUp();

}

void ApplicationClient::quit()
{
	LOGT << "quit wanted" << endl;
	FrameStack::iterator it = mFrameStack.begin();
	FrameStack::iterator end = mFrameStack.end();
	for (; it != end; ++it)
	{
		assertL(*it);
		(*it)->quitWanted();
	}
	mReconnect = false;
}

void ApplicationClient::reconnect()
{
	gTime.pause(false);
	mReconnect = true;
	mGameClient->endFun();
}
