/* MAINTAINER: Bernard */

#ifndef _APPLICATION_CLIENT_H
#define _APPLICATION_CLIENT_H

#include <string>
#include <list>
#include "application.h"
#include "globals_client.h"

class GFX;
class SFX;
class GameClient;
class Game;
class Commander;
class CommanderClient;
class NetworkManagerClient;
class InputManager;
class KeyBindings;
class Console;
class Frame;
struct ServerInfo;

/**
 * \brief  Main client class for holding all together.
 * 
 * The ApplicationClient holds many client's specific
 * subsystems and handles starting menu and starting game
 * incuding connecting and map reload.
 */
class ApplicationClient : public Application
{
public:

	/**
	 * \brief Allocate all member pointer variables.
	 * 
	 * \param argc Used for inicialization of Application.
	 * \param argv Used for inicialization of Application.
	 */
	ApplicationClient(int argc, char* argv[]);
	
	/**
	 * \brief Free all member pointer vairables.
	 */
	~ApplicationClient();

	/**
	 * \brief Call init on nearly every possible subsystem :)
	 */
	virtual void init();

	/**
	 * \brief Currently starts the menu.
	 */
	virtual void run();

	/**
	 * \brief Clean subsystems.
	 */
	virtual void clean();

	/**
	 * \brief Tell application to quit.
	 * 
	 * Note that the quit does not have to be immedietly but it may 
	 * take some cycles in loop.
	 */
	virtual void quit();

	/**
	 * \brief Provides access to basic graphics.
	 * 
	 * \return mGFX
	 */
	GFX& getGFX();
	
	/**
	 * \brief Provides access to samples and music.
	 * 
	 * \return mSFX
	 */
	SFX& getSFX();

	/**
	 * \brief Provides access to the game.
	 * 
	 * \return mGameClient
	 */
	GameClient& getGame();
	
	/**
	 * \brief Provides access to basics of game.
	 * 
	 * Sources on client should use getGame(). This function is here becaue
	 * of virtual interface of Application and access to the game from
	 * common sources.
	 * \return mGameClient
	 */
	virtual Game& getGameCommon();

	/**
	 * \brief Provides access to commander.
	 * 
	 * \return mCommanderClient
	 */
	Commander& getCommander();

	/**
	 * \brief Provides access to network manager.
	 * 
	 * \return mNetworkManagerClient
	 */
	NetworkManagerClient& getNetworkManager();
	
	/**
	 * \brief Provides access to InputManager.
	 *
	 * \return mInputManager
	 */
	InputManager& getInputManager();

	/**
	 * \brief Provides access to key binding.
	 * 
	 * \return mKeyBindings
	 */
	KeyBindings& getKeyBindings();
	
	/**
	 * \brief Provides access to Console.
	 *
	 * \return mConsole
	 */
	Console& getConsole();


	/**
	 * \brief Jump into loop.
	 * 
	 * The loop is processing input and calls tm. The loop
	 * quits when frame has set that it wants to end (function wantRun()).
	 * \param frame Frame for deciding when end the loop.
	 * \param tm TaskManager used in the loop.
	 */
	void loop(Frame &frame, TaskManager &tm);

	/**
	 * \brief Start the menu.
	 * 
	 * Quit after menu decided to quit.
	 */
	void runMenu();

	/**
	 * \brief Connect to the server and start playing game.
	 * 
	 * Return after disconnected from server.
	 * \param sinfo Server we want to connect to.
	 */
	void runGame(const ServerInfo  &sinfo);

	/**
	 * \brief The name is wrong :) Only map change is here.
	 * 
	 * Originally changing map was done by reccontecting all clients
	 * but later we decided to make better code and now clients don't
	 * need to reconnect to be able to reload the map.
	 * This causes map reload.
	 */
	void reconnect();

private:
	/**
	 * \brief Program killer - trap
	 * 
	 * From security reasons we don't allow anyone to create
	 * copy of Application since it is probably a mistake.
	 * Anyway calling this constructor will kill program.
	 * \param copy not used
	 */
	ApplicationClient(ApplicationClient &copy);
	
protected:

	// All these are allocated in constructor
	InputManager *mInputManager;
	KeyBindings  *mKeyBindings;

	GFX  *mGFX; /// \brief Basic graphics system.
	SFX  *mSFX; /// \brief Basic sound system.
	
	GameClient      *mGameClient;
	Console         *mConsole; /// \brief Dropdown console.
	CommanderClient *mCommanderClient;

	NetworkManagerClient *mNetworkManagerClient;

	typedef std::list<Frame*> FrameStack;
	FrameStack mFrameStack;
	
	bool   mReconnect; /// \brief Determine if the app should reload map or quit.
};

#endif // _APPLICATION_CLIENT_H
