/* MAINATINER: Martin */

#ifndef _CAMERA_H
#define _CAMERA_H

#include "SDL_opengl.h"
#include "GLVector.h"
#include "frustum.h"

class WorldEntity;


/**
*	\brief Basic camera class.
*
*	Camera is represented by its position, reference point, up vector (all these vectors 
*	are in world space coordinate system) and information about perspective projection.
*	Perspective projection is described by the horizontal(!) field of view, height-to-width 
*	aspect ratio and near and far clipping planes. Viewing frustum can be calculated from
*	every camera instance.
*/
class Camera
{

public:
	
	static const GLVector3d 
		DEFAULT_POSITION, 			///< \brief Default camera position \code ( 0, 0, 0 ) \endcode
		DEFAULT_REFERENCE_POINT,	///< \brief Default camera reference point \code ( 0, 0, -1 ) \endcode
		DEFAULT_UP;					///< \brief Default camera up vector \code ( 0, 1, 0 ) \endcode

	/**
	*	\brief Describes perspective projection.
	*/
	struct Perspective {
		GLdouble 
			fovx,	///< \brief field of view in X axis ( horizontal )
			aspect,	///< \brief height / width (?) ratio
			zNear,	///< \brief near distance
			zFar;	///< \brief far distance
	};

	/**
	*	\brief Ctor - camera values are set to default.
	*	\note This will set only position, reference point, and up vector. Perspective, frustum and GFX stamp
	*	have to be set separately.
	*/
	Camera(void);
	
	/**
	*	\brief Ctor - camera values are set from arguments.
	*	\note This will set only position, reference point, and up vector. Perspective, frustum and GFX stamp
	*	have to be set separately.
	*/
	Camera(const GLVector3d & pos, const GLVector3d & ref, const GLVector3d & up);
	
	/**
	*	\brief Dtor - does nothing, only exists.
	*/
	virtual ~Camera(void);
	
	/**
	*	\brief Get camera position.
	*	\param pos where to store camera position
	*	\return pos argument
	*/
	inline GLVector3d& getPos(GLVector3d &pos) const;
	
	/**
	*	\brief Get camera direction vector.
	*
	*	Direction vector is computed as folows: \code camera_reference_point - camera_position \endcode
	*
	*	\param dir where to store camera direction vector
	*	\return dir argument
	*/
	inline GLVector3d& getDir(GLVector3d &dir) const;
	
	/**
	*	\brief Get camera reference point.
	*	\param ref where to store camera reference point
	*	\return ref argument
	*/
	inline GLVector3d& getRef(GLVector3d &ref) const;
	
	/**
	*	\brief Get camera up vector.
	*
	*	Up vector doesn't have to be perpendicular to direction vector.
	*
	*	\param up where to store camera up vector
	*	\return up argument
	*/
	inline GLVector3d& getUp (GLVector3d &up ) const;
	
	
	/**
	*	\brief Get camera position - short version.
	*	\return camera position.
	*/
	inline GLVector3d  getPos() const;
	
	/**
	*	\brief Get camera direction vector - short version.
	*	\return camera direction vector.
	*/
	inline GLVector3d  getDir() const;
	
	/**
	*	\brief Get camera referene point - short version.
	*	\return camera reference point.
	*/
	inline GLVector3d  getRef() const;
	
	/**
	*	\brief Get camera up vector - short version.
	*	\return camera up vector.
	*/
	inline GLVector3d  getUp () const;
	
	
	/**
	*	\brief Set new camera position.
	*	\param pos camera position to set.
	*/
	inline void setPos(const GLVector3d &pos);
	
	/**
	*	\brief Set new camera direction vector.
	*
	*	New reference point is computed from direction vector and camera position as 
	*	\code position + direction \endcode and this reference point is stored. 
	*	Direction vector is always calculated from camera position and reference point.
	*
	*	\param pos camera position to set.
	*/
	inline void setDir(const GLVector3d &dir);
	
	/**
	*	\brief Set new reference point.
	*
	*	\note Direction vector is computed from reference point.
	*	\note Up vector is not affected by setting new reference point. Just make You sure that You do not
	*		set reference point that makes direction vector parallel to up vector!
	*
	*	\param ref reference point
	*/
	inline void setRef(const GLVector3d &ref);
	
	
	/**
	*	\brief Set camera up vector.
	*
	*	Up vector doesn't have to be perpendicular to direction vector, but should not
	*	be parallel to direction vector.
	*
	*	\param up new up vector
	*/
	inline void setUp (const GLVector3d &up);

	/**
	*	\brief Set new perspective projection.
	*
	*	\param fovx field of view in X axis ( horizontal )
	*	\param aspect height / width (?) ratio of viewing window
	*	\param zNear distance to near clipping plane in world space coordinates
	*	\param zFar distance to far clipping plane in world space coordinates
	*/
	void setPerspective(const GLdouble fovx, const GLdouble aspect, const GLdouble zNear, const GLdouble zFar);

	/**
	 *	\brief Set perspective projection info from GFX.
	 */
	void setGFXPerspective();
	
	/**
	*	\brief world space to camera space
	*
	*	Calulate transformation matrix that transforms scene from world space to camera space.
	*
	*	\param m resultant transformation matrix
	*/
	void computeLookAt(GLMatrix4d &m);
	
	/**
	*	\brief camera space to clip space
	*
	*	Calculates transformation matrix that transforms scene from camera space to clip space using perspective projection.
	*
	*	\param resultant transformation matrix
	*/
	void computePerspective(GLMatrix4d &m);
	
	
	/**
	*	\brief frustum from camera position, orientation and perspective
	*	
	*	Six clipping planes of frustum are calulated using both camera transformation matrices (look-at and perspective).
	*/
	void computeFrustum();
	
	/**
	*	\brief get camera frustum
	*	\return last calculated frustum
	*	\note Make You sure You called computeFrustum() before calling this method.
	*/
	Frustum getFrustum() const { 
		return mFrustum; 
	}

	/**
	*	\brief postmultiplies current OpenGL matrix by look-at matrix ( simulates ::gluLookAt() ).
	*
	*	Look-at matrix is calculated, then current OpenGL matrix is postmultiplied by this transformation matrix.
	*
	*	\note Make You sure that GL_MODELVIEW_MATRIX is set.
	*/
	void gluLookAt();
	
	/**
	*	\brief postmultiplies current OpenGL matrix by perspective matrix ( simulates ::gluPerspective() ).
	*
	*	Projection matrix is calculated, then current OpenGL matrix is postmultiplied by this transformation matrix.
	*
	*	\note Make You sure that GL_PROJECTION_MATRIX is set.
	*/
	void gluPerspective();
	
	/**
	*	\brief set camera view to OpenGL
	*
	*	GFX perspective is obtained, GL_PROJECTION_MATRIX is set to camera perspective transformation, then
	*	GL_MODELVIEW_MATRIX is set to camera look-at transformation. Every next multipliation of GL_MODELVIEW_MATRIX
	*	will probably mean some transformation of world objects in world space.
	*/
	void setView();

	/**
	 * \brief Return monitored entity if any.
	 */
	virtual WorldEntity* monitoredEntity() { return NULL;  }

	/**
	 * \brief Clean current context of camera.
	 */
	virtual void clean() { }

	/**
	 * \brief Move the view to some default position.
	 * 
	 * In future the position may be loaded from map etc..
	 */
	 virtual void resetView();
	 
protected:
	
	GLVector3d	mPosition;			///< \brief Camera position (eye point).
	GLVector3d	mReferencePoint;	///< \brief Reference point (the point we are looking at).
	GLVector3d	mUpVector;			///< \brief Where is "UP", not necesary perpendicular to mPosition nor mReferencePoint.
	Frustum		mFrustum;			///< \brief View frustum.

private:

	Perspective mPerspective;			///< \brief Description of perspetive projection.
	bool		mPerspectiveChanged;	///< \brief perspective change flag
	int         mGFXStamp;				///< \brief some unknown "magic"
};

GLVector3d& Camera::getPos(GLVector3d &pos) const { return pos = mPosition; }
GLVector3d& Camera::getDir(GLVector3d &dir) const { return dir = mReferencePoint - mPosition; }
GLVector3d& Camera::getRef(GLVector3d &ref) const { return ref = mReferencePoint; }
GLVector3d& Camera::getUp (GLVector3d &up)  const { return up = mUpVector; }

GLVector3d  Camera::getPos() const { GLVector3d tmp; getPos(tmp); return tmp; }
GLVector3d  Camera::getDir() const { GLVector3d tmp; getDir(tmp); return tmp; }
GLVector3d  Camera::getRef() const { GLVector3d tmp; getRef(tmp); return tmp; }
GLVector3d  Camera::getUp () const { GLVector3d tmp; getUp(tmp); return tmp; }

void Camera::setPos(const GLVector3d &pos) { mPosition = pos; }
void Camera::setDir(const GLVector3d &dir) { mReferencePoint = mPosition + dir; }
void Camera::setRef(const GLVector3d &ref) { mReferencePoint = ref; }
void Camera::setUp (const GLVector3d &up)  { mUpVector = up; }

#endif // _CAMERA_H
