/* MAINTAINER: Bernard */

#ifndef _COMMANDER_CLIENT_H
#define _COMMANDER_CLIENT_H

#include "commander.h"
#include "hash_hack.h"
#include "stlextdefs.h"
#include <string>
#include <sstream>


typedef STL_EXT_NS::hash_map<std::string, std::string> aliasMapType;

/**
 * \brief Processor of client commands.
 * 
 * Accepts users commands from console and send them to the network
 * or process them directly.
 */
class CommanderClient : public Commander
{
public:
	CommanderClient();
	virtual ~CommanderClient(){ };


protected:


		
	static aliasMapType   mAliasMap;  /// \brief Map of command aliases
	
	/**
	 * \brief Process one command.
	 * 
	 * This function will be called in main task.
	 * \param command string with command
	 * \param playerID ID of player who created the command. It is ignored on client.
	 */
	virtual void processCommand(const std::string &command, int playerID);


	/**
	 * \brief Send command to the server.
	 *
	 * If first character is / the command go to server.
	 */
	void command_to_server(const std::string &command);

	/**
	 * \brief Does nothing
	 * \param args not used
	 */
	static void command_empty(std::stringstream &args, int playerID);

	/**
	 * \brief Quit the application
	 * \param args not used
	 */
	static void command_quit(std::stringstream &args, int playerID);

	/**
	 * \brief show some help
	 * \param args not used
	 */
	static void command_help(std::stringstream &args, int playerID);

	/**
	 * \brief Get setting variable
	 * \param args variable_name
	 */
	static void command_getvar(std::stringstream &args, int playerID);

	/**
	 * \brief Set setting variable to some value
	 * \param args variable_name value
	 */
	static void command_setvar(std::stringstream &args, int playerID);

	/**
	 * \brief Show all players in game
	 * \param args not used
	 */
	static void command_players(std::stringstream &args, int playerID);

	/**
	 * \brief Enable samples
	 * \param args on|off
	 */
	static void command_sound(std::stringstream &args, int playerID);
	
	/**
	 * \brief Set samples volume
	 * \params args new volume
	 */
	static void command_soundvolume(std::stringstream &args, int playerID);
	
	/**
	 * \brief Enable music
	 * \param args on|off
	 */
	static void command_music(std::stringstream &args, int playerID);

	/**
	 * \brief Sem music volume
	 * \param args new volume
	 */
	static void command_musicvolume(std::stringstream &args, int playerID);

	/**
	 * \brief Set sound theme
	 * \param args themefile
	 */
	static void command_soundtheme(std::stringstream &args, int playerID);

	/**
	 * \brief Write message to chat window
	 * \param args message
	 */
	static void command_chatmessage(std::stringstream &args, int playerID);
	
	/**
	 * \brief Write message to console window
	 * \param args message
	 */
	static void command_consolemessage(std::stringstream &args, int playerID);

	/**
	 * \brief Enable log in console
	 * \param args on|off
	 */
	static void command_log_in_console(std::stringstream &args, int playerID);
	
	/**
	 * \brief Change resolution of main window
	 * \param args resolution
	 */
	static void command_resolution(std::stringstream &args, int playerID);

	/**
	 * \brief CHange field of view in X axis
	 * \param args float value of angle
	 */
	static void command_fovx(std::stringstream &args, int playerID);

	/**
	 * \brief fullscreen/windowed mode
	 * \param args on|off
	 */
	static void command_fs(std::stringstream &args, int playerID);
	
	/**
	 * \brief clear console windows
	 * \param args not used
	 */
	static void command_clear(std::stringstream &args, int playerID);

	/**
	 * \brief grab mouse and keyboard input
	 * \param args on|off
	 */
	static void command_grab(std::stringstream &args, int playerID);

	/**
	 * \brief Unbind all key bindings.
	 * \param args not used
	 */
	static void command_unbind_all(std::stringstream &args, int playerID);

	/**
	 * \brief Bind key to command
	 * \param args keyname hi_priority command
	 */
	static void command_bind(std::stringstream &args, int playerID);

	/**
	 * \brief Same effect as pressing/releasing some button.
	 * \param args keyname up/down
	 */
	static void command_key(std::stringstream &args, int playerID);

	/**
	 * \brief Read file and add every line of it as a command to commander.
	 * \param args filename
	 */
	static void command_exec(std::stringstream &args, int playerID);

	/**
	 * \brief Show/hide console
	 * \param args not used
	 */
	static void command_console(std::stringstream &args, int playerID);

	/**
	 * \brief Sleep execution for some time.
	 * \param args time
	 */
	static void command_sleep(std::stringstream &args, int playerID);
		
	/**
	 * \brief Create new command alias.
	 * \param args alias_name aliased_thing
	 */
	static void command_alias(std::stringstream &args, int playerID);

	/**
	 * \brief Delete all aliases
	 * \param args not used
	 */
	static void command_unalias(std::stringstream &args, int playerID);

	/**
	 * \brief Provide feedback for picked bonus from server
	 * \param args id_of_bonus_type
	 */
	static void command_picked_bonus(std::stringstream &args, int playerID);

	/**
	 * \brief Provide feedback for expired bonus from server
	 * \param args id_of_bonus_type
	 */
	static void command_expired_bonus(std::stringstream &args, int playerID);

	/**
	 * \brief Change pause command from server.
	 * \param args pause_state
	 */
	static void command_pause(std::stringstream &args, int playerID);

	/**
	 * \brief Change pause command from server.
	 * \param args pause_state
	 */
	static void command_activate_chat(std::stringstream &args, int playerID);

	/**
	 * \brief Print coordinates of player's tank
	 * \param args not used
	 */
	static void command_coords(std::stringstream &args, int playerID);

	/**
	 * \brief Set targeting and tracing mode
	 * \param args [floating_yaw]
	 */
	static void command_targeting(std::stringstream &args, int playerID);

	/**
	 * \brief Set sensitivity for targeting
	 * \param args yaw(x)_sens pitch(y)_sens
	 */
	static void command_mousesensitivity(std::stringstream &args, int playerID);

	/**
	 * \brief Moves cannon to initial center position
	 * \param args not used
	 */
	static void command_resettargeting(std::stringstream &args, int playerID);

	/**
	 * \brief Moves cannon to initial center position
	 * \param args visible(bool)
	 */
	static void command_showstats(std::stringstream &args, int playerID);

	/**
	 * \brief Command for connecting to the server
	 * \param args serverAddress serverPort
	 */
	static void command_connect(std::stringstream &args, int playerID);

	/**
	 * \brief Command ending the game and return to the menu
	 * \param args not used
	 */
	static void command_disconnect(std::stringstream &args, int playerID);

	/**
	 * \brief For setting up camera (tracer)
	 * \param args pos_z distance ref_pos_z ref_distance
	 */
	static void command_camera(std::stringstream &args, int playerID);

	/**
	 * \brief Change observer mode
	 * \param args not used
	 */
	static void command_observer(std::stringstream &args, int playerID);

	/**
	 * \brief Change observer speed
	 * \param args new speed
	 */
	static void command_observerspeed(std::stringstream &args, int playerID);

	/**
	 * \brief Change activity of ingame menu
	 * \param args not used
	 */
	static void command_ingamemenu(std::stringstream &args, int playerID);
		
	/**
	 * \brief Make client load new map - this is command from server
	 * \param args not used
	 */
	static void command_reconnect(std::stringstream &args, int playerID);
	
	/**
	 * \brief Call server command for creating new tank with SETTING_PLAYER_PREFERED_TANK
	 *        as parameter.
	 * \param args not used
	 */
	static void command_newtank(std::stringstream &args, int playerID);

	/**
	 * \brief Call server command for setting name to SETTING_PLAYER_PREFERED_NAME
	 * \param args not used
	 */
	static void command_setpreferedname(std::stringstream &args, int playerID);
	
};

#endif
