/* MAINTAINER: Bernard */

// #define WANT_TRACE


#include "SDL.h"
#include "SDL_opengl.h"

#include <sstream>
#include <string>
#include <iostream>
#include <vector>

#include "globals_client.h"
#include "log.h"
#include "console.h"
#include "platform.h"
#include "settings.h"
#include "application_client.h"
#include "gfx.h"
#include "sfx.h"
#include "inputmanager.h"
#include "commander.h"
#include "conversions.h"
#include "translator.h"

#include "font_manager.h"
#include "opengl2d.h"
#include "gui_constants.h"
#include "timemanager.h"

using namespace std;

Console::Console()
{
	mPriority = IA_PRIORITY_CONSOLE;
	mActive = false;
	
	mFont = 0;
	mLineWrap = true;

	mTitle = 0; // some default value
	
	mInHistory = g_settings.getInt(SETTING_CONSOLE_LINES);

	mLogForwarding = g_settings.isTrue(SETTING_LOG_IN_CONSOLE);

	mHistoryIgnore = false;
	mTimestampOfErrorMessage = 0;
}

Console::~Console()
{
	TIN;
}


void Console::init()
{
	TIN;

	// registration to input manager
	getAppClient().getInputManager().addInputAcceptor(*this);

		
	// Loading font
 	mFont = FontManager::getRef().get(
					gc.con_font.c_str(),
					gc.con_font_size);	

	mErrorFont = FontManager::getRef().get(
					gc.con_errmsg_font.c_str(),
					gc.con_errmsg_font_size);	

	
	
	if (!mFont)
	{
		LOGE << "Loading font failed!" << endl;
		DIE();
	}
	
	// register as log acceptor
	mLogForwarding = g_settings.isTrue(SETTING_LOG_IN_CONSOLE);
	registerLog();

	// This is for testing :-) 
	mTitle = LNG_CACHE_("-- Console --");

	
	mSlideOff = (float)((mInHistory+2)*mFont->Ascent() + 2*(gc.con_box_dy) + gc.con_font_hy);
	mSlide = mSlideOff;
}

void Console::clean()
{
	TIN;

	// unregister from input
	getAppClient().getInputManager().removeInputAcceptor(*this);

	// Clean font
	mFont = 0; 
	mErrorFont = 0;

	// clean registration to log system
	g_log.setAcceptor(0); 


	LNG_UNCACHE(mTitle);
}


bool Console::wantKey(const SDL_keysym &keysym)
{
	return mActive;
}

bool  Console::keyDown(const SDL_keysym &keysym)
{
	TIN; 

	if (!mActive) return false;

	LOGT << "Processing new key " << keysym.sym << " :-)\n";
	
	switch (keysym.sym) 
	{

	// process command & erase commmand line
	case SDLK_RETURN:   
	case SDLK_KP_ENTER:	
	
		processCommand( mInputLine.getLineU() );	

		mInputLine.clear();
		
		break;
	
	// something with history of commands - not yet
	case SDLK_UP:
		//LOGI << "History in console not implemented yet." << endl;
		break;
	case SDLK_DOWN:
		//LOGI << "History in console not implemented yet." << endl;
		break;	
	case SDLK_TAB:
		//LOGI << "Completion in console is not implemented." << endl;
		break;	

	// adding one char to commandline
	default:
	
		mInputLine.processKey(keysym);
		
		break;
	}
	
	return true;
}

void Console::logLine(const std::string &line)
{
	if ( mLogForwarding )
		addMessage(line);
}


void Console::registerLog()
{
	if ( mLogForwarding )
	{
		g_log.setAcceptor(this);
	}
	else
	{
		g_log.setAcceptor(0);
	}
}


void Console::render()
{
	TIN;
	
	
	assertL(mFont.isValid());

	int w = getAppClient().getGFX().getWidth();
	int h = getAppClient().getGFX().getHeight();

	// OpenGL init
	getAppClient().getGFX().set2DOpenGL();
    glLoadIdentity();


	// Render FPS
	if (g_settings.isTrue(SETTING_SHOW_FPS))
	{
		mFont->RenderUTF8WithShadow(
			xtos(getAppClient().getGFX().getFps()),
			w - gc.fps_wx, h - gc.fps_hy,
			gc.fps_r, gc.fps_g, gc.fps_b
		);
	}
	
	
	// Finally render the error message
	renderErrorMessage();	
		
	// if we don't want "slidedown/slideup" :-)
	if (!mActive && mSlide >= mSlideOff) return;

	if (mActive)
	{
		if ((mSlide -= 25) < 0) mSlide = 0;
	}
	else
	{
		mSlide += 25;
	}

glPushMatrix();


	glTranslatef(0, mSlide, 0);

	// Decoration box
	gGL2D.drawRect(gc.con_box_x,  
	               h - gc.con_box_dy - gc.con_font_hy - (mInHistory+2)*mFont->Ascent(), 
	               w-2*gc.con_box_x, 
				   (mInHistory+2)*mFont->Ascent() + 2*(gc.con_box_dy),
				   gc.con_box_r, gc.con_box_g, gc.con_box_b, gc.con_box_a);


	glColor3f( 1.0, 1.0, 1.0);

	// Decorations
	mFont->RenderUTF8(
		LNG_GET(mTitle), 
		gc.con_font_x, 
		h - gc.con_font_hy - mFont->Ascent()
	);

	
	// History (messages) - must be locked
	mHistoryMutex.lock();
	mHistoryIgnore = true;

	list<string>::iterator it = mMessageHistory.begin();
	int i = mInHistory+1;
	for(; it != mMessageHistory.end(); ++it, --i)
	{
		mFont->RenderUTF8(
			*it,
			gc.con_font_x,
			h - (gc.con_font_hy + i*mFont->Ascent())
		);
	}	

	mHistoryIgnore = false;
	mHistoryMutex.unlock();
	
	// commandline
	mFont->RenderUTF8( 
		mInputLine.getLineURender(),
		gc.con_font_x,
		h - (gc.con_font_hy + (mInHistory+2)*mFont->Ascent())
	);

glPopMatrix();
}

void Console::addMessage(const std::string &msg)
{
	// DON'T DO ANY LOGGING HERE SICE THIS MAY BE CALLED FROM
	// LOG AND THUS IT WOULD DO INFINITE LOOP
	
	mHistoryMutex.lock();
	if (!mHistoryIgnore) 
	{
		// Add message to history
		// Line wrapping - the text if first translated to wide chars - we
		// want proper multibyte chars wrapping. Then one by one chars are
		// taken from src (original message) to temporary "line". 
		if (mFont && mLineWrap)
		{
			// maximal width of text in console
			int maxW = getAppClient().getGFX().getWidth()-2*gc.con_font_x-mFont->getWwidth();
		
			wstring src, dst;
			utf8tows(msg.c_str(), src);
			
			while (!src.empty())
			{
				wchar_t w;
				
				w = src[0];
				src.erase(0,1);
				dst += w;
				
				string dst_utf;
				wstoutf8(dst, dst_utf);
				
				if (mFont->getTextWidth(dst_utf) >= maxW)
				{
					mMessageHistory.push_front(dst_utf);
					dst.erase(0);
				}
				
			}
			
			if (!dst.empty())
			{
				string dst_utf;
				wstoutf8(dst, dst_utf);
				mMessageHistory.push_front(dst_utf);
			}
			
		}
		else
		{
			mMessageHistory.push_front(msg);
		}
	
		
		// Remove old messages from history
		while ((signed int)mMessageHistory.size() > mInHistory)
		{
			mMessageHistory.pop_back();
		}
	}

	mHistoryMutex.unlock();
}


void Console::processCommand(const std::string &cmd)
{
	TIN;
	
	LOGT << "Console command: " << cmd << endl;

	addMessage(cmd);
	
	getAppClient().getCommander().addCommand(cmd);	
}


void Console::setLogInConsole(bool enable)
{
	mLogForwarding = enable;
	g_settings.set(SETTING_LOG_IN_CONSOLE, mLogForwarding);
	registerLog();
}


void Console::clear()
{
	mMessageHistory.clear();
}


void Console::changeActivity()
{
	mActive = !mActive;
	getAppClient().getInputManager().wantAllInput(mActive);
	LOGT << " Console is " << mActive << endl;

	// Amuse user with super sound :-)
	if (mActive)
	{
		// getAppClient().getSFX().play("console.wav");
	}
}

void Console::errorMessage(const std::string &msg)
{
	mErrorMessage = msg;	
	mTimestampOfErrorMessage = gTime.getTicks();
}

void Console::renderErrorMessage()
{
	if (mErrorMessage.empty()) return;	

	int showed_time = gTime.getTicks() - mTimestampOfErrorMessage;

	if (showed_time > gc.con_errmsg_max_time) return;

	int h2 = getAppClient().getGFX().getHeight()/2;
	int w2 = getAppClient().getGFX().getWidth()/2;

	mErrorFont->RenderUTF8WithShadow(mErrorMessage, w2, h2, gc.con_errmsg_r, gc.con_errmsg_g, gc.con_errmsg_b);
/*
	// set alpha
	float alpha = 1.0;

	if (mTimestampOfErrorMessage > gc.con_errmsg_fade_time)
	{
		alpha = 1.0f - (float)(mTimestampOfErrorMessage-gc.con_errmsg_fade_time)/
			        (float)(gc.con_errmsg_max_time-gc.con_errmsg_fade_time);
	}
			
	// Shadow under message
	glColor4f(0, 0, 0, alpha);
	mErrorFont->RenderUTF8( mErrorMessage, w2+2, h2-2, Font::CENTER);
	mErrorFont->RenderUTF8( mErrorMessage, w2-2, h2+2, Font::CENTER);

			
	// The message itself
	glColor4f(gc.con_errmsg_r, gc.con_errmsg_g, gc.con_errmsg_b, alpha);		
	mErrorFont->RenderUTF8( mErrorMessage, w2, h2, Font::CENTER);
*/	
	
}
