/* MAINTAINER: Bernard */

#ifndef _CONSOLE_H
#define _CONSOLE_H

#include <string>
#include <list>

#include "inputacceptor.h"
#include "mutex.h"
#include "log.h"
#include "inputstring.h"

#include "font_manager.h"
#include "layer.h"

/**
 * \brief Simple dropdown console.
 * 
 * Process input from keyboard and send commands to Commander.
 * Doesn't process any command itself. It server also as log acceptor
 * and it is able to render itself.
 * TODO:
 *   commands history
 */
class Console : public InputAcceptor, public LogAcceptor, public Layer
{
public:
	Console();
	~Console();
	
	// from InputAcceptor
	/**
	 * Console takes every key when active
	 */
	virtual bool wantKey(const SDL_keysym &keysym);
	virtual bool keyDown(const SDL_keysym &keysym);

	// from LogAcceptor
	
	/**
	 * \brief Processing one line from log.
	 *
	 * The line is simply added as a message line.
	 * \param line one line
	 */
	 virtual void logLine(const std::string &line);
	

	/**
	 * \brief Render the console.
	 *
	 * Renders nothig if console is not active.
	 * Also render FPS - but it should be removed.
	 */
	virtual void render();

	/**
	 * \brief Inicialize the console and plug it into the system.
	 *
	 * Menas loading font, textures and adding to InputManager.
	 */
	void init();
	
	/**
	 * \brief Clean resources allocated by console
	 *
	 */
	void clean();

	/**
	 * \brief Set logging to console
	 * 
	 * \param enable if true the logging go to console
	 */
	void setLogInConsole(bool enable);

	/**
	 * \brief Add new message to console.
	 *
	 * Is thread safe.
	 * \param msg newly added message
	 */
	void addMessage(const std::string &msg);

	/**
	 * \brief Clear all messages in console windows
	 */
	void clear();


	/**
	 * \brief Change activity of console.
	 *
	 * If the console is active, it is deactivated, if is deactivated,
	 * it activates.
	 */
	void changeActivity();


	/**
	 * \brief Show some error message.
	 * 
	 * The message will dissapear afer while.
	 * \param msg The message to show.
	 */
	void errorMessage(const std::string &msg);

private:	
	
	/**
	 * \brief Handles the command
	 *
	 * Processing is simple forward to Commander.
	 * \param cmd string with command
	 */
	void processCommand(const std::string &cmd);

	/**
	 * \brief Update console state as log listener.
	 *
	 * Register/unregister to log as acceptor. Depending on
	 * mLogForwarding
	 */
	void registerLog();

	int mInHistory; ///< \brief Number of lines in console history.
	
	std::list<std::string> mMessageHistory;  ///< \brief Lines of history in console.
	Mutex                  mHistoryMutex;    ///< \brief Lock for mMessageHistory.

	/**
	 * \brief Render the error message. The last thing in render.
	 */
	void renderErrorMessage();

	/**
	 * \brief New messages are ignored if this is true.
	 *
	 * This exists becasue when you are rendering a history
	 * it may happen that new messages will appear and it would
	 * corrupt iterators.
	 */
	bool mHistoryIgnore;

	bool mLineWrap;  ///< \brief Console is able to wrap long lines if needed.
	
	bool mActive; ///< \brief True if console is active.

	InputString mInputLine; ///< \brief Currently edited command.
	
	FontRef  mFont;      ///< \brief Font for console.
	float    mSlide;     ///< \brief Console slides down and up - the current position.
	float    mSlideOff;  ///< \brief Console slides down and up - the off position.

	bool	mLogForwarding; ///< \brief Log messages are printed to console if true.

	int     mTitle;  ///< \brief Translation index for title of console.
	
	// commands history?
	
	// Error message
	std::string mErrorMessage;  ///< \brief Error message for the user.
	int         mTimestampOfErrorMessage; ///< \brief Time when the error message was created
	FontRef     mErrorFont;      ///< \brief Font for error message console.

};

#endif // _CONSOLE_H

