
//#define WANT_TRACE

#include "GL/glew.h"
#include "explosion_manager.h"
#include "timemanager.h"
#include "log.h"
#include "params.h"
#include "application_client.h"
#include "game_client.h"
#include "settings.h"
#include "tasksound.h"

#include "globals_client.h"
#include "gfx.h"

#include "platform.h"

using namespace std;

/// value [0..1)
inline float randf()
{
	return float(rand())/RAND_MAX;
}

/** \brief random point inside unit circle. */
inline void rand_circle( float &xresult, float &yresult )
{
	static const float radius2( 1 / ( sqrtf(2) - 1) );
	
	bool b0, b1, b2, b3, b4;
	
	float x, y;
	
	// 8 percent miss instead of 27 percent
	do {
		x = randf();
		y = randf();
		const float tmp  ( 2 + x*x + y*y );
		const float sum  ( 2 * ( x - y ) );
		const float diff ( 2 * ( x + y ) );

		b0 =  tmp		  > 1;
		b1 = (tmp + diff) > 1;
		b2 = (tmp - sum ) > 1;
		b3 = (tmp + sum ) > 1;
		b4 = (tmp - diff) > 1;

	} while ( !(b0 | b1 | b2 | b3 | b4) );
	
	xresult = b0 ? x : ( radius2 * ( (b1 | b3) ? -1 - x : 1 - x) );
	yresult = b0 ? y : ( radius2 * ( (b3 | b4) ? -1 - y : 1 - y) );
}

float randomFromTo( float from, float to )
{
    return from + (to - from) * randf();
}

/*
float ffast_cos( const float x )
{
    // assert:  0 <= fT <= PI/2
    // maximum absolute error = 1.1880e-03
    // speedup = 2.14

    const float xsqr = x * x;
    const float res( float(3.705e-02) );
    res *= xsqr;
    res -= float(4.967e-01);
    res *= xsqr;
    res += 1;
    return res;
}

*/
/*
template <typename Real>
inline void rand_circle( Real &x, Real &y)
{
	x = 2 * randr<Real>() - 1;
	y = 2 * randr<Real>() - 1;

	while( x*x + y*y > 1 )
	{
		x = 2 * randr<Real>() - 1;
		y = 2 * randr<Real>() - 1;
	}
}
*/

/**
*	\brief constrained random vector
*	Calculates a random vector deviated from given vector maximaly by devianceRAD.
*/
void random_vector( const GLVector3f vec, const float devianceRAD, GLVector3f & result )
{
	const float alpha( randomFromTo(-devianceRAD, devianceRAD) );
	const float beta( randomFromTo( 0, float(2 * M_PI) ) );
	const GLVector3f v1( vec.getCross( GLVector3f(1, 0, 0) ) );
	const GLVector3f v2( v1.getCross( vec ) );
	
	const float 
		cos_alpha( cos(alpha) ), 
		cos_beta( cos( beta ) ), 
		sin_alpha( sin( alpha ) ), 
		sin_beta( sin( beta ) );

	result = vec * cos_alpha + v1 * (sin_alpha * sin_beta) + v2 * (sin_alpha * cos_beta);
}

/** \brief returns a random unit vector */
GLVector3f random_vector( void )
{
	GLVector3f v;

	// Pick a random Z between -1.0f and 1.0f.
	v.z = randomFromTo( -1.0f, 1.0f );

	// Pick a random point on a circle.
	float x, y, tmp;
	do {
		rand_circle( x, y );
		tmp = x * x + y * y;
	} while ( tmp == 0 );

	// Get radius of this circle
	float sqr_radius = 1 - v.z * v.z;
	tmp = sqrt( sqr_radius / tmp );

	v.x = tmp * x ;
	v.y = tmp * y ;
	
	return v;
}

/* ******************************* Explosion manager ********************************** */

ExplosionManager::ExplosionManager()
{
	mDeltaGameTime  = 0;
	mLastUpdateTime = 0;
	mMinDelay    = g_settings.getInt(SETTING_EXPLOSIONS_MIN_DELAY);
	mSkipUpdates = g_settings.isTrue(SETTING_EXPLOSIONS_SKIP_ON_SLOW_CPU);
}

ExplosionManager::~ExplosionManager()
{

}

TextureRef ExplosionManager::getParticleTexture( int type )
{
	if ( type < 0 )
		return 0;

	if ( static_cast<int>( mTextureCache.size() ) <= type )
		mTextureCache.resize( type + 1, 0 );
	
	TextureRef &tr( mTextureCache[type] );
	if ( tr.isValid() )
		return tr;
	else
	{
		// --------------------------------------
		// --- sprite image and alpha channel ---
		// --------------------------------------
		const Params::sExplosionParams & params( gParams.explosion(type) );
		const std::string & explicitAlpha(params.particle.sprite.explicitAlpha), &image(params.particle.sprite.image );
		if ( !explicitAlpha.empty() ) 
		{
			SurfExternalAlpha sea(locateFile(image), locateFile(explicitAlpha) );	
			tr = TextureManager::getInstance()->add( &sea, GL_RGBA, Texture::BUILD_MIPMAPS );
		} 
		else 
		{
			tr = TextureManager::getInstance()->add( image, GL_RGBA, Texture::BUILD_MIPMAPS );
		}
		return tr;
	}
}

void ExplosionManager::update(int atime)
{
	int ticks = gTime.getGameTicks();
	mDeltaGameTime += ticks - mLastUpdateTime;
	mLastUpdateTime = ticks;

	while (mDeltaGameTime > mMinDelay)
	{
		if (mSkipUpdates) mDeltaGameTime = 0;
		else mDeltaGameTime -= 20;
		
		updateExplosions();
	}
}

void ExplosionManager::addExplosion(int type, const GLVector3f &postion, const GLVector3f &direction)
{
	TIN;

	Explosion tmp_expl( type, this );
	mExplosions.push_front(tmp_expl);

	Explosion &explosion = *mExplosions.begin();

	explosion.init(postion, direction);
}

void ExplosionManager::addExplosion(int type, const EntityPublicInfoPtr & pubInfo )
{
	TIN;

	Explosion tmp_expl( type, this );
	mExplosions.push_front(tmp_expl);

	Explosion &explosion = *mExplosions.begin();

	explosion.init( pubInfo );
}

void ExplosionManager::init()
{
	mLastUpdateTime = gTime.getGameTicks();
	mDeltaGameTime = 0;
	
	// Fill texture cache -just try to load texture for every particle
	for (int explID = 0; explID < static_cast<int>(gParams.mExplosions.size()); explID++)
	{
		if (gParams.mExplosions[explID].valid)
		{
			getParticleTexture( explID );
		}
	}
}

void ExplosionManager::clean()
{
	mExplosions.clear();
	mTextureCache.clear();
}

void ExplosionManager::updateExplosions()
{
	TIN;

	// update explosions and remove expired
	LOGD << "Update " << mExplosions.size() << endl;


	ExplosionList::iterator explosion_it = mExplosions.begin();

	while(explosion_it != mExplosions.end())
	{
		if (explosion_it->update(mLastUpdateTime))
		{
			++explosion_it;
		}
		else
		{
			explosion_it = mExplosions.erase(explosion_it);
		}
	}
}

void ExplosionManager::render()
{
	TIN;

	ExplosionList::iterator explosion_it = mExplosions.begin();

	for(;explosion_it != mExplosions.end(); ++explosion_it)
	{
		explosion_it->render();
	}
}

/* ******************************* Explosions ************************************ */

void Explosion::initParticle(int id, int timeStamp)
{
	int n(0);

	typedef Particles::types types;
	const Params::sExplosionParams::Particle & pparams( mParams.particle );
	//const Params::sExplosionParams::Emitter & eparams( mParams.emitter );

	// ----------------
	// --- position ---
	// ----------------
	GLVector3f v = random_vector();
	v.x *= pparams.position.radius;
	v.y *= pparams.position.radius;
	v.z *= randomFromTo( 0, pparams.position.height.to - pparams.position.height.from ) ;
	v.z += pparams.position.height.from;
	mParticles.actual_position[id] = mParticles.position[id] = types::position_type( getPosition() + v );

	// ----------------
	// --- movement ---
	// ----------------
	{
		// direction
		const int n( std::max( static_cast<int>( GLVector3f::D ), static_cast<int>( pparams.direction.vec.size() ) ) );
		GLVector3f tmp_vec;
		for (int i = 0; i < n; ++i)
			tmp_vec.val[ i ] = pparams.direction.vec[ i ];
		random_vector( tmp_vec, pparams.direction.deviance * float(M_PI / 180), mParticles.direction[id] );
		
		//mParticles.direction[id] = types::direction_type(0, 0, 1);
	}
	mParticles.speed[id] = types::speed_type( randomFromTo( pparams.speed.from, pparams.speed.to ) );
	{
		// global_acceleration
		const int n( std::max( static_cast<int>( GLVector3f::D ), static_cast<int>( pparams.globalAcceleration.vec.size() ) ) );
		GLVector3f tmp_vec;
		for (int i = 0; i < n; ++i)
			tmp_vec.val[ i ] = pparams.globalAcceleration.vec[ i ];
		random_vector( tmp_vec, pparams.globalAcceleration.deviance * float(M_PI / 180), mParticles.global_acceleration[id] );

		//mParticles.global_acceleration[id] = types::global_acceleration_type(0, 0, 0);
	}
	mParticles.speed_acceleration[id] = types::speed_type( randomFromTo( pparams.acceleration.from, pparams.acceleration.to ) );
	
	// ------------
	// --- time ---
	// ------------
	mParticles.birth[id] = types::birth_type( (timeStamp - mTimeStamp) / 1000.0f + randomFromTo( pparams.birth.from, pparams.birth.to ) );	
	mParticles.spanlife[id] = Particles::types::spanlife_type( randomFromTo( pparams.spanlife.from, pparams.spanlife.to ) );
	mParticles.age[id] = Particles::types::age_type(0);
	
	// --------------
	// --- visual ---
	// --------------
	n = static_cast<int>( pparams.color.size() );
	mParticles.color[id] = types::color_type( 
		n > 0 ? pparams.color[0] : 0, 
		n > 1 ? pparams.color[1] : 0, 
		n > 2 ? pparams.color[2] : 0, 
		0 );	// particle is invisible, it will show up at its first update ( particles are born in the future )

	mParticles.size[id] = types::size_type( pparams.size );
}


Explosion::Explosion(int type, ExplosionManager * emanager) 
: mExplosionType(type), mParams(gParams.explosion(type))
{
	setExplosionManager( emanager );

	mTimeStamp = gTime.getGameTicks();	
	mNextExplosionCreated = false;
	mValidParticles = 0;
}

/*
cp is a 4 element array where:
cp[0] is the starting point, or P0
cp[1] is the first control point, or P1
cp[2] is the second control point, or P2
cp[3] is the end point, or P3
t is the parameter value, 0 <= t <= 1
*/
void pointOnCubicBezier( const GLVector2f * cp, float t, GLVector2f & result )
{
    float   ax, bx, cx;
    float   ay, by, cy;
    float   tSquared, tCubed;

    /* calculate the polynomial coefficients */

    cx = 3.0f * (cp[1].x - cp[0].x);
    bx = 3.0f * (cp[2].x - cp[1].x) - cx;
    ax = cp[3].x - cp[0].x - cx - bx;
        
    cy = 3.0f * (cp[1].y - cp[0].y);
    by = 3.0f * (cp[2].y - cp[1].y) - cy;
    ay = cp[3].y - cp[0].y - cy - by;
        
    /* calculate the curve point at parameter value t */
        
    tSquared = t * t;
    tCubed = tSquared * t;
        
    result.x = (ax * tCubed) + (bx * tSquared) + (cx * t) + cp[0].x;
    result.y = (ay * tCubed) + (by * tSquared) + (cy * t) + cp[0].y;
}

// righ top quarter
/*
void pointOnCircleApprox( float t, GLVector2f & result )
{
	const float K( 0.5522847498307933984022516322796f );
    const float	cy(3.0f * K);
    const float tSquared(t * t), tCubed(tSquared * t);
        
    result.x = (4.0f - cy) * tCubed + (cy - 3.0f) * tSquared + 1;
    result.y = (cy - 2.0f) * tCubed + (3.0f - 2 * cy) * tSquared + cy * t;
}*/

bool Explosion::update(int currenttime)
{
	EntityPublicInfoPtr epi( getEntityPublicInfo() );
	if ( epi.get() )
	{
		setPosition( epi->getPosition() );
	}

	// in seconds
	const float timeFromStart( (currenttime - mTimeStamp) / 1000.0f );
	
	// expiration test
	if ( timeFromStart > mParams.duration )
		return false;

	// Update current particles & move expired to the end.
	// The expired may be recycled to avoid allocatios.
	
	int maxID = mValidParticles;
	int part_ID = 0;
	while (part_ID < maxID)
	{
		// calculate particle time that has elapsed from its birth time (in seconds, floats)
		const float parttime( timeFromStart - mParticles.birth[ part_ID ] );
		const float partspanlife( mParticles.spanlife[ part_ID ] );

		if ( parttime < 0 )
		// particle will be born in the future, but not right now
		{
			part_ID++;
			continue;
		}

		if ( parttime <= partspanlife )
		// update the particle at part_ID
		{
			// update age
			mParticles.age[ part_ID ] = partspanlife ? ( parttime / partspanlife ) : 0;

			// update position
			for (int i = 0; i < 3; ++i )
			{
				const float pos( mParticles.position[ part_ID ][ i ] );
				const float dir( mParticles.direction[ part_ID][ i ] );
				const float vel( mParticles.speed[ part_ID] );
				const float a1( mParticles.speed_acceleration[ part_ID ] );
				const float a2( mParticles.global_acceleration[ part_ID ][ i ] );

				mParticles.actual_position[ part_ID ][ i ] = 
					pos + ( dir * vel + 0.5f * (a2 + dir * a1) * parttime ) * parttime;

				// alpha component of color has to be updated (it has been initialized to zero)				
				{
					static const GLVector2f cp[] = 
					{ 
						GLVector2f(0.0f,0.0f), 
						GLVector2f(0.17f, 2.25f), 
						GLVector2f(0.25f, 0.0f), 
						GLVector2f(1.0f, 0.0f) 
					};
					GLVector2f tmp(0, 0);
					pointOnCubicBezier( cp, mParticles.age[ part_ID ], tmp );
					mParticles.color[part_ID].w = tmp.y * ( (mParams.particle.color.size() > 3) ? mParams.particle.color[3] : 1.0f );
				}
			}
		
			part_ID++;
		}
		else
		// the particle is gone :(
		{
			maxID--;
			mParticles.copyParticleBtoA( part_ID, maxID );
		}
	}

	const Params::sExplosionParams::Emitter &emitter( mParams.emitter );
		
	// Allocate new particles if wanted
	if ( timeFromStart < emitter.stopEmit )
	{
		int wantedParticles( static_cast<int>( emitter.initialParticlesCount + emitter.emitSpeed * timeFromStart + 0.5 ) );
		wantedParticles = std::max<int>( 0, std::min<int>( wantedParticles, emitter.maxParticlesCount ) );
		
		mParticles.resize( wantedParticles );
		
		for ( int particle_ID = maxID; particle_ID < wantedParticles; ++particle_ID )
		{
			initParticle( particle_ID, currenttime );
		}
		
		mValidParticles = wantedParticles;
	}
	else
	{
		mValidParticles = maxID;
		
		// only remove old paricles
		mParticles.resize( mValidParticles );
	}
	
	// Try to create next explosion
	if ( !mNextExplosionCreated && mParams.next_id >= 0 && timeFromStart > mParams.time_of_next )
	{
		EntityPublicInfoPtr epi( getEntityPublicInfo() );
		if ( !epi.get() )
			getAppClient().getGame().getExplosionManager().addExplosion(mParams.next_id, getPosition(), getDirection() ); 
		else
			getAppClient().getGame().getExplosionManager().addExplosion(mParams.next_id, epi );

		mNextExplosionCreated = true;
	}

	return true;
}

void Explosion::Particles::copyParticleBtoA(int a, int b)
{
	actual_position[a] = actual_position[b];
	position[a]		= position[b];
	color[a]		= color[b];
	size[a]			= size[b];
	birth[a]		= birth[b];
	spanlife[a]		= spanlife[b];
	age[a]			= age[b];
	direction[a]	= direction[b];
	speed[a]		= speed[b];
	speed_acceleration[a]	= speed_acceleration[b];
	global_acceleration[a]	= global_acceleration[b];
}

void Explosion::Particles::resize(int newsize)
{
	actual_position.resize( newsize );
	position.resize( newsize );
	color.resize( newsize );
	size.resize( newsize );
	birth.resize( newsize );
	spanlife.resize( newsize );
	age.resize( newsize );
	direction.resize( newsize );
	speed.resize( newsize );
	speed_acceleration.resize( newsize );
	global_acceleration.resize( newsize );
}


void Explosion::init(const GLVector3f &position, const GLVector3f &direction)
{
	ExplosionManager *eman( getExplosionManager() );
	assertL( eman && "The pointer to ExplosionManager can't be null!" );
	this->setParticleTexture( eman->getParticleTexture( mExplosionType ));
	
	this->setPosition( position );
	this->setDirection( direction );

	const int initialParticlesCount( std::max( 0, mParams.emitter.initialParticlesCount ) );
	mParticles.resize( initialParticlesCount );

	for (int particle_ID = 0; particle_ID < initialParticlesCount; ++particle_ID)
	{
		initParticle(particle_ID, mTimeStamp);
	}
	
	mValidParticles = initialParticlesCount;
	
	// Add the explosion to 3d sound
	TaskSound *ts = TaskSound::getInstance();
	if (ts)
	{
		ts->newExplosion(mExplosionType, GLVector3d(position.x,  position.y,  position.z));
	}	
}

void Explosion::init(const EntityPublicInfoPtr & pubInfo )
{
	if ( pubInfo.get() ) 
	{
		setEntityPublicInfo( pubInfo );
		init( pubInfo->getPosition(), GLVector3f(0, 0, 1) );
	}
}

int Explosion::Particles::count()
{
	// TODO: check all particle arrays for equal sizes!
	return static_cast<int>( position.size() );
}

void Explosion::calculatePointAttenuation( GLfloat &a, GLfloat &b, GLfloat &c )
{
	const GLfloat dMeters = mParams.particle.size;

	GLint viewport[ 4 ];
	GLfloat mat [ 16 ];
	glGetFloatv( GL_PROJECTION_MATRIX, mat );
	glGetIntegerv( GL_VIEWPORT, viewport );

	GLfloat H = (GLfloat) viewport[ 3 ];
	GLfloat tmp = ( 1 / mat[0] );
	GLfloat k = 1.0f / ( 1.0f + 2 * tmp * tmp );
	tmp = k / (H * dMeters * H * dMeters);
	
	a = GLfloat(0);
	b = GLfloat(0);
	c = tmp;
}

void Explosion::render()
{
	//GFX & gfx = getAppClient().getGFX();

	typedef Particles::types::position_type position_type;
	typedef Particles::types::color_type color_type;

	GLint count = mValidParticles; //mParticles.count();
	if ( count <= 0 )
		return;

	glVertexPointer( static_cast<GLint>( position_type::D ), GLE< position_type >::type, 0, &std::vector<position_type>::const_reference( mParticles.actual_position.front() ) );
	glColorPointer( static_cast<GLint>( color_type::D ), GLE<color_type>::type, 0, &std::vector<position_type>::const_reference( mParticles.color.front() ) );

	// ---------------------
	// --- enable arrays ---
	// ---------------------
	glEnableClientState( GL_VERTEX_ARRAY );
	glEnableClientState( GL_COLOR_ARRAY  );

	{
		// -------------------------
		// --- point parameters  ---
		// -------------------------
		{
			glPointSize( 1 );

			const GLfloat 
				minPointSize( 1 ), 
				maxPointSize( 1000 );

			GLfloat a[3] = { 0, 0, 0 };
			calculatePointAttenuation( a[0], a[1], a[2] );

			if ( GLEW_VERSION_1_4 )
			{
				glPointParameterf( GL_POINT_SIZE_MIN, minPointSize );
				glPointParameterf( GL_POINT_SIZE_MAX, maxPointSize );
				glPointParameterfv( GL_POINT_DISTANCE_ATTENUATION, a );				
			} 
			else if ( GLEW_ARB_point_parameters )
			{
				glPointParameterfARB( GL_POINT_SIZE_MIN_ARB, minPointSize );
				glPointParameterfARB( GL_POINT_SIZE_MAX_ARB, maxPointSize );
				glPointParameterfvARB( GL_POINT_DISTANCE_ATTENUATION_ARB, a );
			} 
			else if ( GLEW_EXT_point_parameters )
			{
				glPointParameterfEXT( GL_POINT_SIZE_MIN_EXT, minPointSize );
				glPointParameterfEXT( GL_POINT_SIZE_MAX_EXT, maxPointSize );
				glPointParameterfvEXT( GL_DISTANCE_ATTENUATION_EXT, a );
			}
			else 
			{
				// TODO: point parameters not supported...
			}
		}

		// ----------------
		// --- lighting ---
		// ----------------
		glDisable(GL_LIGHTING);
		
		// ---------------------
		// --- blending mode ---
		// ---------------------
		Params::sExplosionParams::BlendingType blendType = mParams.particle.blendingType;
		bool blending( true ), alpha(false);
		switch ( blendType )
		{
		case Params::sExplosionParams::BLENDING_ADDITIVE:
			glBlendFunc( GL_SRC_ALPHA, GL_ONE );
			glEnable( GL_BLEND );
			glDepthMask( GL_FALSE );
			break;
		case Params::sExplosionParams::BLENDING_INTERPOLATION:
			glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
			glEnable( GL_BLEND );
			glDepthMask( GL_FALSE );
			break;
		case Params::sExplosionParams::BLENDING_NONE:
		default:
			glDisable( GL_BLEND );
			glDepthMask( GL_TRUE );
			glAlphaFunc( GL_GEQUAL, 0.5 );
			glEnable( GL_ALPHA_TEST );
			blending = false;
			alpha = true;
		}

		// -------------
		// --- color ---
		// -------------
		// ... ?

		// ---------------
		// --- texture ---
		// ---------------
		TextureRef tr = getParticleTexture();
		if ( tr.isValid() )
		{
			glActiveTexture( GL_TEXTURE0 );
			tr->bind();
			
			glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE,	GL_COMBINE );
			
			glTexEnvf(GL_TEXTURE_ENV, GL_COMBINE_RGB,		GL_MODULATE );
			glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_RGB,		GL_PRIMARY_COLOR );
			glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_RGB,		GL_SRC_COLOR);
			glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE1_RGB,		GL_TEXTURE0 );
			glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND1_RGB,		GL_SRC_COLOR);

			glTexEnvf(GL_TEXTURE_ENV, GL_COMBINE_ALPHA,		GL_MODULATE );
			glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_ALPHA,		GL_PRIMARY_COLOR );
			glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_ALPHA,	GL_SRC_COLOR);
			glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE1_ALPHA,		GL_TEXTURE0 );
			glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND1_ALPHA,	GL_SRC_COLOR);

			glEnable( GL_TEXTURE_2D );
		} else
			glDisable( GL_TEXTURE_2D );

		// ---------------------
		// --- point sprites ---
		// ---------------------
		glEnable( GL_POINT_SPRITE );
		glTexEnvi( GL_POINT_SPRITE, GL_COORD_REPLACE, GL_TRUE );

		// -------------------
		// --- draw arrays ---
		// -------------------
		glDrawArrays( GL_POINTS, 0, count );

		// ----------------------
		// --- disable states ---
		// ----------------------
		if ( blending )
		{
			glDisable( GL_BLEND );
			glDepthMask( GL_TRUE );
		}
		
		if ( alpha )
			glDisable( GL_ALPHA_TEST );

		glTexEnvi( GL_POINT_SPRITE, GL_COORD_REPLACE, GL_FALSE );
		glDisable( GL_TEXTURE_2D );
		glDisable( GL_POINT_SPRITE );
	}

	// ----------------------
	// --- disable arrays ---
	// ----------------------
	glDisableClientState( GL_COLOR_ARRAY  );
	glDisableClientState( GL_VERTEX_ARRAY );

	// ---------------------------------
	// --- restore point parameters  ---
	// ---------------------------------
	{
		glPointSize( 1 );


		const GLfloat minPointSize(1), maxPointSize(8);
		GLfloat a[3] = { 1, 0, 0 };

		if ( GLEW_VERSION_1_4 )
		{
			glPointParameterf( GL_POINT_SIZE_MIN, minPointSize );
			glPointParameterf( GL_POINT_SIZE_MAX, maxPointSize );
			glPointParameterfv( GL_POINT_DISTANCE_ATTENUATION, a );				
		} 
		else if ( GLEW_ARB_point_parameters )
		{
			glPointParameterfARB( GL_POINT_SIZE_MIN_ARB, minPointSize );
			glPointParameterfARB( GL_POINT_SIZE_MAX_ARB, maxPointSize );
			glPointParameterfvARB( GL_POINT_DISTANCE_ATTENUATION_ARB, a );
		} 
		else if ( GLEW_EXT_point_parameters )
		{
			glPointParameterfEXT( GL_POINT_SIZE_MIN_EXT, minPointSize );
			glPointParameterfEXT( GL_POINT_SIZE_MAX_EXT, maxPointSize );
			glPointParameterfvEXT( GL_DISTANCE_ATTENUATION_EXT, a );
		}
		else 
		{
			// TODO: point parameters not supported...
		}
	}
}
