#ifndef _EXPLOSION_MANAGER_H
#define _EXPLOSION_MANAGER_H

#include "task.h"
#include <vector>
#include <cstdlib>
#include <list>
#include "GLVector.h"
#include "layer.h"

#include "params.h"
#include "texturemanager.h"
#include "property.h"
#include "entity_public_info.h"

class ExplosionManager;

class Explosion
{
public:
	/**
	 * \brief Load params reference
	 */
	Explosion( int type, ExplosionManager * manager );

	/**
	 * \create particles etc...
	 */
	void init(const GLVector3f &postion, const GLVector3f &direction);
	void init(const EntityPublicInfoPtr & pubInfo );

	void initParticle(int id, int timeStamp);

	
	/**
	 * \brief Update all particles in explosion.
	 *
	 * \return true if the explosion is gone (should be deleted)
	 */
	bool update(int currenttime);

	/**
	 * \brief Render all particles in explosion.
	 */
	void render();

private:
	void calculatePointAttenuation( GLfloat &a, GLfloat &b, GLfloat &c );

protected:

friend class ExplosionManager;

	/**
	 * \brief Time of explosion creation.
	 *
	 * It is used to determine that explosion should disapear.
	 */
	int mTimeStamp;  ///< \brief Time of creation;

	const int mExplosionType;
	const Params::sExplosionParams& mParams;
	
	
	bool mNextExplosionCreated; // you may create explosion in sequences
	
	struct Particles
	{
		struct types
		{
			typedef GLVector3f	position_type;
			typedef GLVector4f	color_type;
			typedef float		size_type;
			typedef float		birth_type;
			typedef float		spanlife_type;
			typedef float		age_type;
			typedef GLVector3f	direction_type;
			typedef float		speed_type;
			typedef GLVector3f	global_acceleration_type;
		};
		
		// data for all particles
		std::vector< types::position_type >		actual_position, position;
		std::vector< types::color_type >		color;
		std::vector< types::size_type >			size;
		std::vector< types::birth_type >		birth;
		std::vector< types::spanlife_type >		spanlife;
		std::vector< types::age_type >			age;
		std::vector< types::direction_type >	direction;
		std::vector< types::speed_type >		speed;
		std::vector< types::speed_type >		speed_acceleration;
		std::vector< types::global_acceleration_type > global_acceleration;
		

		int count();
		void copyParticleBtoA(int a, int b);
		void resize(int newsize); // change size of arrays
	};
	Particles mParticles;
	
	int mValidParticles;
	

	public: _property(TextureRef, ParticleTexture);
	public: _property(GLVector3f, Position);
	public: _property(GLVector3f, Direction);
	private: _vproperty(ExplosionManager *, ExplosionManager );
	private: _property(EntityPublicInfoPtr, EntityPublicInfo );
};

/**
 * \brief Class for holding all particle sources.
 * 
 * Originally the class was managing only explosions but somehow it was 
 * extended into particle subsystem. The ExplosionManager holds list
 * of all currently working explosions. Working means emitin particles or
 * having still some particles. It makes possible to create a new particle
 * emmiter (explosion) based on position and type or entity and type.
 * In case of entity creation the source of particle emmiter is moved
 * as the entity is moving. This is usefull eg for smohink tank etc.
 * 
 * The manager is task andit is happy if you treat it as any other task
 * and only call when you want to add an explosion.
 */
class ExplosionManager : public Task, public Layer
{
public:
	ExplosionManager();
	~ExplosionManager();

	/**
	 * \brief Add an explosion to the system based on absolute position.
	 * 
	 * \param type Explosion ID. This serves as lookup ID to params.
	 * \param postion Position where the explosion places particles.
	 * \param direction Required but not used at all.
	 */
	void addExplosion(int type, const GLVector3f &postion, const GLVector3f &direction);
	
	/**
	 * \brief Add an explosion connected to entity.
	 * 
	 * The explosion is following the entity. It may used for moving
	 * entity which is emiting some particles.
	 * \param type Explosion ID. This serves as lookup ID to params.
	 * \param pubInfo Info from entity.
	 */
	void addExplosion(int type, const EntityPublicInfoPtr & pubInfo );

	/**
	 * \brief Does some time inicialization and fill texutre cache.
	 */
	virtual void init();
	
	/**
	 * \brief Clean all explosions ald clean texture cache.
	 */
	virtual void clean();
	
	/**
	 * \brief Render all explosions in list.
	 */
	virtual void render();

	/**
	 * \brief Clean old explosions and update all other explosions.
	 * 
	 * The function takes care of time and doesn't do the update
	 * too often - depending on SETTING_EXPLOSIONS_MIN_DELAY. If there
	 * is not enough time - like the coputer is slow, the update is slowed
	 * down if SETTING_EXPLOSIONS_SKIP_ON_SLOW_CPU. The variable atime isn't
	 * used for time. Game time is used instead - when game is slowed down,
	 * the explosions are also slowed down. Also during pause the explosions
	 * are paused.
	 * \param atime Not used.
	 */
	virtual void update(int atime);

	/**
	 * \brief Return texture for specific particle type.
	 * 
	 * This is faster that loading textures each time explosion is created since
	 * these are loaded from cache. This is here to be used by Explosion. Probably
	 * nobody else has a good usage for these textures :-) 
	 * \param type ID of explosion. Same as ID used for explosion ID in params.
	 * \return The texture.
	 */
	TextureRef getParticleTexture( int type );

protected:
	
	/**
	 * \brief Dealing with explosion update.
	 * 
	 * Like deleting old and callung update to still valid explosions.
	 */
	void updateExplosions();

	typedef std::list<Explosion> ExplosionList;
	typedef std::vector<TextureRef> TextureCache;

	ExplosionList mExplosions; ///\brief Current explosions.
	
	TextureCache mTextureCache; ///\brief All textures used by particle system - here they are preloaded
	
	int   mDeltaGameTime; ///\brief Time left from last update call
	
	int   mLastUpdateTime; ///\brief Time of last explosion update
	
	int   mMinDelay; ///\brief Minimal delay between explosions update call
	
	bool  mSkipUpdates; ///\brief If true the time isn;t accumulated and max one update per frame.
};

#endif
