/* MAINTAINER: Bernard */

#ifndef FONT_H
#define FONT_H

#include <string>

/**
 * \brief Base class for all Fonts.
 *
 * The class itself doesn't do anything usefull. It is
 * more like interface. Anyone who need to access font
 * returned from FontManager will have access to
 * this interface.
 * Fonts aren't thread safe and shloud be used only from
 * one thread.
 */
class Font
{
public:

	enum Position {
		LEFT_BOTTOM,
		CENTER_CENTER,
		CENTER,
		CENTER_BOTTOM,
		RIGHT_BOTTOM,
	};

	
	Font();          ///< \brief Empty constructor
	virtual ~Font(); ///< \brief Empty destructor

	/**
	 * \brief Render UTF-8 encoded string.
	 *
	 * Render text to specified position. First call may be
	 * slower if implementation use some cache for rendered
	 * fonts. If you want to set a color to text use glColorX(....)
	 * and the font will be rendered wit current color.
	 * \note [0,0] is left bottom corner of screen.
	 * \param utf8str String that to be rendered.
	 * \param x X coodrinate if text.
	 * \param y Y coodrinate if text.
	 */
	virtual void RenderUTF8(const std::string &utf8str, int x, int y) = 0;

	/**
	 * \brief Rendering text with positioning.
	 * 
	 * The implementation of this function uses RenderUTF8 and derived class
	 * doesn't have to implement this directly.
	 * \param utf8str String that to be rendered.
	 * \param x X coodrinate if text.
	 * \param y Y coodrinate if text.
	 * \param pos Position of the text relative to x, y. Possible values ot
	 *        position are from enum Position. 
	 */
	virtual void RenderUTF8(const std::string &utf8str, int x, int y, Position pos);
	
	/**
	 * \brief Font rendering with shadow unter the font.
	 * 
	 * It may be hard to read white text on white screen. To preserve readablity
	 * of text we draw a shadow under the text. Shadow is the same test but
	 * translated a little. If you write something on the screen and it may
	 * happen that it may be rendered os nomething with nearly the same color,
	 * it is a good idea to render it with shadow. You may also specify what color you
	 * want to have since first the black color for shadow and then your
	 * color for final text.
	 * Implementations of Font doesn't have to implement this function since it uses
	 * RenderUTF8.
	 * \param utf8str String that to be rendered.
	 * \param x X coodrinate if text.
	 * \param y Y coodrinate if text.
	 * \param r Red color of wanted text.
	 * \param g Green color of wanted text.
	 * \param b Blue color of wanted text.
	 * \param pos Position of the text relative to x, y.
	 */
	virtual void RenderUTF8WithShadow(const std::string &utf8str, int x, int y,
					        float r, float g, float b, Font::Position pos = Font::CENTER);

	
	/**
	 * \brief Function for rendering 3D text turned to camera.
	 * 
	 * The text is in 3D space and it is oriented like going up. The orientation along
	 * Z axis is computed such that the text is turned to camera. The text is automatic
	 * centered. Implementation doesn't have to implement this function because it is
	 * currently implemented using RenderUTF83D.
	 * \param utf8str String to render.
	 * \param x X coordinate of the text.
	 * \param y Y coordinate ot the text.
	 * \param z Z coordinate of the text.
	 * \param scale Some scale of the text - it is not easy to guess the right size by pt for game so scale is avaliable.
	 */
	
	virtual void RenderUTF83D(const std::string &utf8str, float x, float y, float z, float scale);


	/**
	 * \brief Function for rendering 3D text.
	 * 
	 * The text is in 3D space and it is oriented like going up. It is possible
	 * to modify orientation along Z axis. The text is rendered without blending but
	 * also it doesn't spoil the Z buffer on places where it isn't readable - man where
	 * the text has empty spaces.
	 * \param utf8str String to render.
	 * \param x X coordinate of the text.
	 * \param y Y coordinate ot the text.
	 * \param z Z coordinate of the text.
	 * \param scale Some scale of the text - it is not easy to guess the right size by pt for game so scale is avaliable.
	 * \param dx Orientation of text in x coordinate.
	 * \param dy Orientation of the text in y coordinate.
	 * \param center True if the text should be centered.
	 * \param doublesided True if the text should be rendered as double sied.
	 */
	virtual void RenderUTF83D(const std::string &utf8str, float x, float y, float z, float scale,
							  float dx, float dy, bool center, bool doublesided) = 0;
	
	
	/**
	 * \brief Get With of text in pixels.
	 *
	 * Get width of utf-8 encoded text.
	 * \return Width of text.
	 */
	virtual int getTextWidth(const std::string &utf8str) = 0;
	
	
	/**
	 * \brief Get line height.
	 *
	 * The value is usefull if you want to render multiline
	 * text and need to know height of line.
	 * \return Height if line in font.
	 */
	virtual int	Ascent() = 0;
	
	/**
	 * \brief Clean the cache.
	 *
	 * Check the whole cache and delete members that weren't
	 * used for a long time.
	 * The exact behaviour is implementation dependent and 
	 * cache even doesn't have to be implemented.
	 */
	virtual void cleanCache();

	/**
	 * \brief Get width of W.
	 *
	 * W is usuallu the most width char in font so you may use
	 * this as aproximation.
	 * \return Width of W.
	 */
	virtual int getWwidth() = 0;
};

#endif
