/* MAINTAINER: Bernard */
//#define WANT_TRACE

#include "SDL.h"
#include "SDL_ttf.h"
#include "SDL_opengl.h"

#include "log.h"
#include "font_ttf.h"
#include "gfx.h"
#include "application_client.h"

FontTTF::FontTTF(const char* file, int ptsize)
{
	// Inicializing SDL_ttf
	if (!TTF_WasInit() && TTF_Init() == -1)
	{
		LOGE << "Unable to init TTF: " << TTF_GetError() << endl;
		DIE();
	}
	
	// Loading Font
	mFont = TTF_OpenFont(file, ptsize);
	if (!mFont)
	{
		LOGE << "TTF_OpenFont: " << TTF_GetError() << endl;
		DIE();
	}

	mWwidth = -1;
}

FontTTF::~FontTTF()
{
	if (mFont)
	{
		TTF_CloseFont(mFont);
		mFont = 0;
	}

	// clearing the cache - we need 2 calls to really clear the cache
	cleanCache();
	cleanCache();
}

void FontTTF::RenderUTF8(const std::string &utf8str, int x, int y)
{
	TIN;
	
	SCacheMember *scm = getSCM(utf8str);
	if (scm == 0) return;

	// Draw the text  -- Use one 2D Rectangle to draw it
	getAppClient().getGFX().set2DOpenGL();
	
	scm->texture.bind();

	float maxs = scm->texture.getMaxS();
	float maxt = scm->texture.getMaxT();
	
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    
	glBegin(GL_QUADS);
	glTexCoord2f(0, 0);
	glVertex3i(x, y, 0);

	glTexCoord2f(maxs, 0);
	glVertex3i(x + scm->w, y, 0);

	glTexCoord2f(maxs, maxt);
	glVertex3i(x + scm->w, y + scm->h, 0);

	glTexCoord2f(0, maxt);
	glVertex3i(x, y + scm->h, 0);
	glEnd();

	glDisable(GL_BLEND);

	glDisable(GL_TEXTURE_2D);
}

					  
void FontTTF::RenderUTF83D(const std::string &utf8str, float x, float y, float z, float scale,
					  float dx, float dy, bool center, bool doublesided)
{

	SCacheMember *scm = getSCM(utf8str);
	if (scm == 0) return;

	getAppClient().getGFX().set3DOpenGL();
	
	scm->texture.bind();

	float maxs = scm->texture.getMaxS();
	float maxt = scm->texture.getMaxT();

	float wx = scm->w*scale*dx;
	float wy = scm->w*scale*dy;
	float h = scm->h*scale;

//	if (center)
	{
		x -= wx/2;
		y -= wy/2;
		z -= h/2;
	}

	glEnable(GL_TEXTURE_2D);
	//glDisable(GL_BLEND);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	
 	glDepthMask( GL_TRUE );
	glAlphaFunc( GL_GEQUAL, static_cast<GLclampf>( 0.05 ) );
	glEnable( GL_ALPHA_TEST );
	   
	glBegin(GL_QUADS);
	glTexCoord2f(0,    0);    glVertex3f(x,           y, z);
	glTexCoord2f(maxs, 0);    glVertex3f(x + wx, y + wy, z);
	glTexCoord2f(maxs, maxt); glVertex3f(x + wx, y + wy, z + h);
	glTexCoord2f(0,    maxt); glVertex3f(x,      y,      z + h);

	if (doublesided)
	{
		glTexCoord2f(maxs, maxt); glVertex3f(x,      y,      z + h);
		glTexCoord2f(0,    maxt); glVertex3f(x + wx, y + wy, z + h);
		glTexCoord2f(0,    0);    glVertex3f(x + wx, y + wy, z);
		glTexCoord2f(maxs, 0);    glVertex3f(x,      y,      z);
	}
	glEnd();

	glDisable(GL_BLEND);
	glDisable( GL_ALPHA_TEST );

	glDisable(GL_TEXTURE_2D);

}

void FontTTF::cleanCache()
{
	LOGT << "Cleaning TTF chache. Size before: " << mRenderedCache.size() << endl;

	TStringHash::iterator it  = mRenderedCache.begin();
	
	while(it != mRenderedCache.end())
	{
		assertL(it->second);
		if (it->second->usages == 0)
		{
			// Remove from hash table
			TStringHash::iterator it2 = it;
			++it2;
			
			delete it->second;	
//			it = mRenderedCache.erase(it);				
			mRenderedCache.erase(it);				
			it = it2;
		}
		else
		{
			// Just reset usages to zero
			it->second->usages = 0;
			++it;
		}
	}
	
	LOGT << "Cleaning TTF chache. Size after: " << mRenderedCache.size() << endl;
}


int	FontTTF::Ascent()
{
	assertL(mFont);
	
	return TTF_FontAscent(mFont);
}

int FontTTF::getWwidth()
{
	assertL(mFont);

	if (mWwidth == -1)
	{
		TTF_SizeText(mFont, "W", &mWwidth, 0);
	}

	return mWwidth;
}
	
int FontTTF::getTextWidth(const std::string &utf8str)
{
	assertL(mFont);

	if (utf8str.empty()) return 0;

	int w;
	TTF_SizeUTF8(mFont, utf8str.c_str(), &w, 0);

	return w;
}


FontTTF::SCacheMember * FontTTF::getSCM(const std::string &utf8str)
{
	// searching in cache
	TStringHash::iterator it = mRenderedCache.find(utf8str);
	
	// Create new string not foutn on cache
	if (it != mRenderedCache.end()) 
	{	
		it->second->usages++;
		return it->second;
	}

	LOGT << "creating new text" << endl;
	
	// TTF_Render has problem with empty strings
	if (utf8str.empty()) return 0;
	
	// creating new record if not found
	SDL_Color col = {255,255,255};
	SDL_Surface *surf = TTF_RenderUTF8_Blended(mFont, utf8str.c_str(), col);
//	SDL_Surface *surf = TTF_RenderUTF8_Solid(mFont, utf8str.c_str(), col);
//	SDL_Surface *surf = TTF_RenderUTF8_Solid(mFont, utf8str.c_str(), col);
	if (!surf)
	{
	//		LOGE << "Unable to render '" << utf8str << "': " << TTF_GetError() << endl;
		return 0;
	//		DIE();
		}

	// Upgrading the surface :-)

	// convert to texture
	SCacheMember *scm = new SCacheMember;
	if (!scm)
	{
		DIE("new failed");
	}
	scm->usages  = 0;
	scm->w       = surf->w;
	scm->h       = surf->h;
	scm->surface = 0;
	
	SurfDefault sd( surf ); 
	scm->texture.create( &sd );                                                 
	
	SDL_FreeSurface(surf);

	mRenderedCache[utf8str] = scm;
	scm->usages++;
	return scm;
}


