/* MAINTAINER: Bernard */

#ifndef FONT_TTF_H
#define FONT_TTF_H

#include <string>
#include "SDL.h"
#include "SDL_ttf.h"

#include "texturemanager.h"
#include "hash_hack.h"
#include "stlextdefs.h"

#include "font.h"
#include "font_texture.h"

/**
 * \brief Implementation of font using SDL_ttf
 *
 * The font use caching of recently rendered strings.
 * So if you render some string often it should be fast.
 * But if you render different strings every frame the
 * performance won't be good.
 * 
 * TODO: Somenoe should call TTF_Quit
 */
class FontTTF : public Font
{
public:

	/**
	 * \brief Load font from file.
	 * 
	 * \param file File with font.
	 * \param ptsize Size of loaded font.
	 */
	FontTTF(const char* file, int ptsize);
	virtual ~FontTTF();

	/**
	 * \brief Render UFT8 string to some position.
	 * 
	 * \param utf8str The string to render.
	 * \param x X coordinate for render.
	 * \param x Y coordinate for render.
	 */
	virtual void RenderUTF8(const std::string &utf8str, int x, int y);

	/**
	 * \brief Function for rendering 3D text.
	 * 
	 * The text is in 3D space and it is oriented like going up. It is possible
	 * to modify orientation along Z axis. The text is rendered without blending but
	 * also it doesn't spoil the Z buffer on places where it isn't readable - man where
	 * the text has empty spaces.
	 * \param utf8str String to render.
	 * \param x X coordinate of the text.
	 * \param y Y coordinate ot the text.
	 * \param z Z coordinate of the text.
	 * \param scale Some scale of the text - it is not easy to guess the right size by pt for game so scale is avaliable.
	 * \param dx Orientation of text in x coordinate.
	 * \param dy Orientation of the text in y coordinate.
	 * \param center True if the text should be centered.
	 * \param doublesided True if the text should be rendered as double sied.
	 */
	virtual void RenderUTF83D(const std::string &utf8str, float x, float y, float z, float scale,
							  float dx, float dy, bool center, bool doublesided);
	/**
	 * \brief Function for getting text height.
	 */
	virtual int	Ascent();
	
	/**
	 * \brief Clear the cache.
	 *
	 * Check the whole cache and delete members that weren't
	 * used for a long time.
	 */
	virtual void cleanCache();
	
	/**
	 * \brief Get width of letter 'W'.
	 * 
	 * Letter W is suposed to be the widest letter ever so if you
	 * need to know what is the wides W is a good upper bound.
	 * \return Width of W.
	 */
	virtual int getWwidth();
	
	/**
	 * \brief Get With of text in pixels.
	 *
	 * Get width of utf-8 encoded text.
	 * \return Width of text.
	 */
	virtual int getTextWidth(const std::string &utf8str);
private:

	/**
	 * \brief Pointer to the font instance.
	 */
	TTF_Font *mFont;

	typedef struct SCacheMember
	{
		FontTexture texture; /// \brief OpenGL texture with the text.
		int usages; /// \brief Number of usages in last few seconds.
		int w;      /// \brief Width of text.
		int h;      /// \brief Height of text.
		SDL_Surface *surface; /// \brief Surface with rendered font.
		
	} SCacheMember;
	
	/**
	 * \brief Cache filling.
	 * 
	 * Check the cache for string utf8str. If the string is present,
	 * add some usages to the structure holdig texture for text.
	 * If utf8str is not in cache new structure SCacheMember is created,
	 * filled with texture of rendered text and added to cache.
	 * \param  utf8str String you want to have texture for.
	 * \return Member of chache with textures.
	 */
	SCacheMember * getSCM(const std::string &utf8str);

	typedef STL_EXT_NS::hash_map<std::string, SCacheMember*> TStringHash;

	TStringHash mRenderedCache; ///< \brief Allready rendered textures cache.

	int mWwidth; ///< \brief -1 means invalid / not measured yet
};

#endif
