/* MAINTAINER: Martin */

#ifndef _FRUSTUM_H
#define _FRUSTUM_H

#include "GLMatrix.h"
#include "GLPlane.h"
#include "log.h"
#include <vector>

class Frustum
{
public:
	typedef GLPlaned TPlane;
	
	enum TPlaneID {
		PL_LEFT = 0, 
		PL_RIGHT, 
		PL_TOP,
		PL_BOTTOM,
		PL_NEAR, 
		PL_FAR
	};

	enum ClipState {
		OUTSIDE,
		INSIDE,
		INTERSECT
	};

	Frustum();
	Frustum(const Frustum & f);

	Frustum(const GLMatrix4d &m):mPlanes(6){ set(m, true); }

	/**
	*	\brief Sets the frustrum clipping planes from matrix \c m.
	*
	*	Initializes 6 frustum clipping planes \c p[0] ... \c p[5]
	*
	*	\param m an OpenGL matrix the frustum extract from
	*	\param normalize if set to true, planes will be normalized (this is necessary for further tests against those planes)
	*	\note \c m should be <code>"OpenGL_Projection_Matrix" * "OpenGL Model View Matrix"</code>
	*/
	void set(const GLMatrix4d &m, bool normalize = true);

	/**
	*	\brief Transforms frustum planes to test objects in object space.
	*
	*	In many cases, the frustum is defined by projection a transformation matrix of the camera.
	*	But there can be other transformations that transforms objects from object space to world space.
	*	If we would like to test those objects to frustum intersection, we should apply the OS->WS 
	*	transformations to objects and test these transformed objects. Instead, for each object we transform 
	*	the frustum with the same OS->WS matrix and test the object against the frustum in OS.
	*/
	void transform(const GLMatrix4d &m, bool normalize = true);
	
	
	/**
	*	\brief Gets one frustum plane.
	*	\param planeID unique plane ID
	*	\return apropriate frustum plane
	*/
	const TPlane& getPlane(TPlaneID planeID){ return mPlanes[planeID]; }

	/**
	*	\brief Normalizes given plane.
	*	
	*	Divides the normal vector and D by normal vector length.
	*
	*	\param planeID which plane (0..6 or TPlaneID)
	*/
	void normalizePlane(TPlaneID planeID);

	/**
	*	\brief Normalizes all frustum planes.
	*/
	void normalizeAllPlanes(void);

	/**
	*	\brief test if point is inside frustum
	*	\param p point in world space coordsys
	*	\return true if point is inside frustum, false if outside frustum
	*/
	bool testPoint(const GLVector3d &p);

	enum Halfspace{
		NEGATIVE = -1,
		POSITIVE = +1,
		ON_PLANE = 0
	};

	static Halfspace classifyPoint(const TPlane &plane, const GLVector3d &point, const double eps = 0.0f);

	ClipState testSphere(const GLVector3d &center, const double r) const;
	ClipState testBox(const GLVector3d &center, const GLVector3d &extent) const;
	
private:
	/**
	*	\brief "matrix index" - helper method
	*
	*	computes index to GLMatrix4d from \c col, \c row coordinates
	*	
	*	\param row which row (1..4)
	*	\param col which column (1..4)
	*	\return resulting index (0 .. 16)
	*/
	int mi(int row, int col) 
	{ 
		assertL(row > 0 && row <= 4 && col > 0 && col <= 4 && "Bad row/column for computation of matrix index of view frustum."); 
		return ((col-1) * 4) + row-1; 
	}
	
	typedef std::vector<TPlane> TPlaneContainer;
	TPlaneContainer mPlanes; ///< \brief six clipping planes of frustum
};


#endif
