/* MAINATINER: Martin */

// #define WANT_TRACE

#include "GL/glew.h"
#include "SDL.h"
#include <CEGUI.h>
#include <RendererModules/OpenGLGUIRenderer/openglrenderer.h>

#include "log.h"
#include "gfx.h"
#include "settings.h"
#include "texturemanager.h"
#include "timemanager.h"
#include "opengl2d.h"
#include "platform.h"
#include "menubasic.h"
#include "translator.h"
#include "gui_constants.h"

#include <string>
#include <climits>
#include <sstream>
#include <ios>

GFX::GFX()
{
	mConfigStamp = 1;
	mScreen = 0;
	mCEGUIRenderer = 0;
	mMaxPointSize = GLfloat(1);
}

GFX::~GFX()
{
	if (mCEGUIRenderer) delete mCEGUIRenderer; 
}

void GFX::loadVideoSettings(void)
{
	TIN;

	std::istringstream iss(g_settings.getString(SETTING_RESOLUTION));
	char c;
	iss >> skipws >> mVS.width >> c >> mVS.height;

	mVS.bpp			= g_settings.getInt(SETTING_BPP);
	mVS.SDL_flags	= SDL_HWSURFACE | SDL_OPENGL | SDL_ANYFORMAT;
	if (g_settings.isTrue(SETTING_FULLSCREEN))
		mVS.SDL_flags |= SDL_FULLSCREEN;
	
	// TODO: settings: get attribute values from some setting file
	VideoSettings::GLAttribute attrs[] = 
	{
	/*	{ SDL_GL_RED_SIZE,		 5 },
		{ SDL_GL_GREEN_SIZE,	 5 },
		{ SDL_GL_BLUE_SIZE,		 5 },
		{ SDL_GL_DEPTH_SIZE,	16 },*/
		{ SDL_GL_DOUBLEBUFFER,	 g_settings.isTrue(SETTING_DOUBLE_BUFFERING) ? 1 : 0 },
	};

	for (int i = 0; i < (int)(sizeof(attrs)/sizeof(VideoSettings::GLAttribute)); ++i)
		mVS.attrs.push_back(attrs[i]);
	
	mConfigStamp++;
}

void GFX::createWindow()
{
	// Atributes for OpenGL
	{
		std::vector<VideoSettings::GLAttribute>::iterator it, ite;
		for (it = mVS.attrs.begin(), ite = mVS.attrs.end(); it != ite; ++it)
			SDL_GL_SetAttribute(it->attr,it->value);
	}
	
	// creating window
	if ( (mScreen = SDL_SetVideoMode(mVS.width, mVS.height, mVS.bpp, mVS.SDL_flags)) == NULL ) 
	{
		// TODO: design: DIE(message)
		LOGE << "SDL_SetVideoMode failed: " << SDL_GetError() << endl;
		exit(2);
	}
	SDL_WM_SetCaption(_("Battle Just Started"), NULL);
	
	// Disable the SDL cursor since cegui has it's own
	SDL_ShowCursor(SDL_DISABLE);
}

void GFX::init()
{
	TIN;
	resetFps();

	loadVideoSettings();
	
	// init SDL_VIDEO
	if (!SDL_WasInit(SDL_INIT_VIDEO))
	{
		if ( SDL_InitSubSystem(SDL_INIT_VIDEO) == -1 ) 
		{
			LOGE << "SDL_InitSubSystem(SDL_INIT_VIDEO) failed: " << SDL_GetError() << endl;
			exit(2);
		}
	}
	
	// create the SDL/OpenGL window
	createWindow();
	
	LOGI << "Window created" << endl;

	initOpenGL();
	
	// switch to 3D OpenGL 
	mGLMode3D = false;
	set3DOpenGL();
	
	// CEGUI wil take long time to load so put some loading screen there us
	showLoadingScreen();
	
	// init CEGUI
	initCEGUI();
}

void GFX::clean()
{
	TIN;
	
	// This could be our last chance to remove textures from OpenGL (erase them from TextureManager).
	TextureManager::getPtr()->erase();
	mFont = 0;

	// send CEGUI to hell :-)
	cleanCEGUI();
	
	// shit down VIDEO subsystem
	SDL_QuitSubSystem(SDL_INIT_VIDEO);
}

void glActiveTexture_dummy(GLenum unit) { }
void glBlendColor_dummy(GLclampf r, GLclampf g, GLclampf b, GLclampf a) {}
void glMultiTexCoord2f_dummy(GLenum unit, GLfloat u, GLfloat v)
{
	glTexCoord2f(u, v);
}

void GFX::initOpenGL()
{
	TIN;

	// initialize GLEW (OpenGL extensions library)
	if (glewInit() != GLEW_OK)
	{
		DIE("Glew initialization failed!");
	}
	
	// TODO: security: check existence of all needed extensions here
	if ( !GLEW_VERSION_1_4 )
	{
#ifdef _DEBUG
		LOGE << "OpenGL at least 1.4 is required!. This is a debug version so run at your own risk." << endl;
#else
		DIE( "OpenGL version at last 1.4 is required!" );
#endif
	}

	if (glActiveTexture   == 0) { glActiveTexture   = (PFNGLACTIVETEXTUREPROC)glActiveTexture_dummy;   }
	if (glBlendColor      == 0) { glBlendColor      = (PFNGLBLENDCOLORPROC)glBlendColor_dummy;      }
	if (glMultiTexCoord2f == 0) { glMultiTexCoord2f = (PFNGLMULTITEXCOORD2FPROC)glMultiTexCoord2f_dummy; }

	glGetFloatv( GL_POINT_SIZE_MAX, &mMaxPointSize );
	glGetFloatv( GL_POINT_SIZE_MIN, &mMinPointSize );
	

    // Make the viewport cover the whole window
	glViewport(0, 0, mVS.width, mVS.height);

	// Set the clear color to black 
	glClearColor(0.0, 0.0 ,0.0, 0);
							    
	// We want z-buffer tests enable
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);
	
	// Don't draw back-facing polygons
	glEnable(GL_CULL_FACE);

	glHint( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );

	gGL2D.init();
}

SDL_Surface* GFX::getScreen()
{
	return mScreen;
}

	
void GFX::set3DOpenGL()
{
	TIN
	
	if (mGLMode3D) return;

	// Set the camera projection matrix:
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	
	computePerspective();
	
	// Switch back to modelview
	glMatrixMode(GL_MODELVIEW);

	// OpenGL Enables - + probably turn on the lights..
	glEnable(GL_DEPTH_TEST);

	
	mGLMode3D = true;
}

void GFX::set2DOpenGL()
{
	TIN

	if (!mGLMode3D) return;

	// Set the camera projection matrix:
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	gluOrtho2D(0, mVS.width, 0, mVS.height);
	
	// Switch back to modelview
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	// OpenGL Enables - probably turn off the light
	glDisable(GL_DEPTH_TEST);

	// Disable texturing on other units
	disableMultitexturing();

	// Other "specialites"
//	glDisable( GL_ALPHA_TEST );
//	glEnable( GL_TEXTURE_2D );
	glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
//	glDisable(GL_LIGHTING);	

	mGLMode3D = false;
}

int GFX::getWidth()
{
	// some people wanted here to have got -1 since they'd like to divide without testing it to zero ;-))
	return mVS.width ? mVS.width : -1;
	
}
	
int GFX::getHeight()
{
	// some people wanted here to have got -1 since they'd like to divide without testing it to zero ;-))
	return mVS.height ? mVS.height : -1;
	
}

void GFX::swap_buffers(void)
{
	SDL_GL_SwapBuffers();
	
	// update frames-rendered count.
	fps.onFrameRendered(); // this is placed here because of menu (menu isn't placed in TaskRender)
	fps.update(gTime.getTicks() % (unsigned int)INT_MAX);
}

void GFX::resetFps(void)
{
	fps.init(gTime.getTicks() % (unsigned int)INT_MAX);
}


void GFX::reloadGFXSetting()
{
	// Don;t do this on platforms that don;t support it
	if (!canReloadGFXSetting) return;
	
	resetFps();

	loadVideoSettings();
	
	// recreate the SDL/OpenGL window
	createWindow();
	initOpenGL();
	computePerspective();
	
	// switch to 3D OpenGL 
	if (mGLMode3D)
	{
		mGLMode3D = false;
		set3DOpenGL();
	}
	else
	{
		mGLMode3D = true;
		set2DOpenGL();
	}
	
	// update size of window to CEGUI
	if (mCEGUIRenderer) 
	{
		mCEGUIRenderer->setDisplaySize(CEGUI::Size((float)getWidth(), (float)getHeight()));
	}

	mConfigStamp++;
}


void GFX::computePerspective()
{
	
	mPerspective.fovx = g_settings.getDouble(SETTING_FOVX);
	if (mVS.height != 0)
		mPerspective.aspect = (float)mVS.width / (float)mVS.height;
	else
		mPerspective.aspect = 0.5;
		
	// TODO: settings: get near and far lipping plane from settings
	mPerspective.zNear  = 1;
	mPerspective.zFar    = 2024;	// far distance

	mConfigStamp++;
}

void GFX::disableMultitexturing()
{
  	// turn OFF other texture units
	GLint maxtus(0);
	glGetIntegerv( GL_MAX_TEXTURE_UNITS, &maxtus );
	for ( GLint tu = 1; tu < maxtus; ++tu ) 
	{
		glActiveTexture( GL_TEXTURE0 + tu ); 
		glDisable( GL_TEXTURE_2D );
	}

	// And let only unit 0
	glActiveTexture(GL_TEXTURE0);

	// set normal TU0 mode
	glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
}


class BJSResourceProvider : public CEGUI::ResourceProvider
{
	public:
		virtual void loadRawDataContainer (const CEGUI::String &filename, 
                                   CEGUI::RawDataContainer &output, 
								   const CEGUI::String &resourceGroup)
		{
			ifstream file(locateFile(filename.c_str()).c_str(), ios::in|ios::binary);
			if (!file)
			{
				LOGE << "Can't open file "  << filename << endl;
				DIE("Resource locating");
			}

			file.seekg(0, ios::end);
			int size = file.tellg();
			file.seekg(0, ios::beg);

			char *data = new char[size];

			file.read(data, size);
			
			if (!file)
			{
				LOGE << "Some file read error!" << endl;
				DIE();
			}

			output.setSize(size);
			output.setData((CEGUI::uint8*)data);
		}
};


// When initializin cegui we want no logging from it. So let's do it our way :)
class EmptyLogger : public CEGUI::Logger
{
	virtual void logEvent (const CEGUI::String &message, CEGUI::LoggingLevel level)
	{
//		LOGD << message.c_str() << endl;
	}
	
 	virtual  void 	setLogFilename (const CEGUI::String &filename, bool append)
 	{ 		
 	}
};


void GFX::initCEGUI()
{
	using namespace CEGUI;   

	try{

		EmptyLogger *ceguiLOG = new EmptyLogger();
		ceguiLOG = 0; // CEGUI will free the logger later....

		int w = getWidth();
		int h = getHeight();

		BJSResourceProvider *resourcer = new BJSResourceProvider();
		mCEGUIRenderer = new OpenGLRenderer(0,w,h);
		new System(mCEGUIRenderer, resourcer);

//		SchemeManager::getSingleton().loadScheme("TaharezLook.scheme");
		SchemeManager::getSingleton().loadScheme("sleekspace.scheme");
		SchemeManager::getSingleton().loadScheme("TaharezLook.scheme");
		CEGUI::FontManager::getSingleton().createFont("DejaVuSans-10.font");
		System::getSingleton().setDefaultMouseCursor("SleekSpace", "MouseArrow");
/*
		// Load loading screen window
		WindowManager &wm = WindowManager::getSingleton();
		mLoadingScreen = wm.loadWindowLayout("loading.layout");
		MenuBasic::translateWindowTextRecursive(mLoadingScreen);
*/
		// Show the loading screen as soon as possible :-)
		//showLoadingScreen();

	} catch ( CEGUI::Exception &e)
	{
		LOGE << "CEGUI exception: " << e.getMessage() << endl;
		DIE("CEGUI ERROR");

	}
}


void GFX::cleanCEGUI()
{
	using namespace CEGUI;

//	WindowManager::getSingleton().destroyAllWindows();
//	SchemeManager::getSingleton().unloadAllSchemes();
//	CEGUI::FontManager::getSingleton().destroyAllFonts();

	CEGUI::System *cegui = CEGUI::System::getSingletonPtr();
	if (cegui)
	{
		// This is BJSResourceProvider
		ResourceProvider *resourcer = cegui->getResourceProvider();
		
		delete cegui;
		cegui = 0;
		
		if (mCEGUIRenderer) delete mCEGUIRenderer; 
		mCEGUIRenderer = 0;
		
		if (resourcer) delete resourcer;
	}
}

void GFX::showLoadingScreen()
{
//	set3DOpenGL();

	if (!mFont)
	{
		// Loading font
		mFont = FontManager::getRef().get(gc.loading_font.c_str(), gc.loading_font_size);
		if (!mFont)
		{
			LOGE << "Loading font failed!" << endl;
			DIE();
		}
	}

	// some Open GL settings
	glClearColor(0.3f, 0.3f, 0.3f, 1);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_MODELVIEW_MATRIX);
	glLoadIdentity();
/*
	if (mLoadingScreen)
	{
		CEGUI::Window *old = CEGUI::System::getSingleton().setGUISheet(mLoadingScreen);
		CEGUI::System::getSingleton().signalRedraw();
		CEGUI::System::getSingleton().renderGUI();	
		CEGUI::System::getSingleton().setGUISheet(old);
	}
	*/
   	mFont->RenderUTF8WithShadow( _("Loading..."), getWidth()/2, getHeight()/2, 1.0, 1.0, 1.0, Font::CENTER);

	swap_buffers();
}

