/* MAINTAINER: Martin */
#ifndef _GFX_H
#define _GFX_H

#include <vector>
#include "SDL_opengl.h"
#include "SDL.h"

#include "stlextdefs.h"
#include "hash_hack.h"
#include "fps.h"
#include "camera.h"
#include "font_manager.h" 

struct SDL_Surface;

namespace CEGUI
{
	class OpenGLRenderer;
	class Window;
};

/**
*	\brief OpenGL Enum Helper - it maps gl types to gl enums.
*/
template <class T> struct dummy_s {};
template < class T = void, template <class T> class C = dummy_s >
struct GLE { static const GLenum type = GL_INVALID_ENUM; };
template<> struct GLE<GLbyte> { static const GLenum type = GL_BYTE; };
template<> struct GLE<GLubyte> { static const GLenum type = GL_UNSIGNED_BYTE; };
template<> struct GLE<GLshort> { static const GLenum type = GL_SHORT; };
template<> struct GLE<GLushort> { static const GLenum type = GL_UNSIGNED_SHORT; };
template<> struct GLE<GLint> { static const GLenum type = GL_INT; };
template<> struct GLE<GLuint> { static const GLenum type = GL_UNSIGNED_INT; };
template<> struct GLE<GLfloat> { static const GLenum type = GL_FLOAT; };
template<> struct GLE<GLdouble> { static const GLenum type = GL_DOUBLE; };
// this one works with vector<T>, or GLVector3<T>
template< typename T, template <class T> class C > struct GLE< C<T> > { static const GLenum type = GLE<T>::type; };

class GFX
{
public:
	GFX();
	~GFX();

	/**
	*	\brief initialize SDL_VIDEO, OpenGL, fps counter and create window
	*/
	void init();
	
	/**
	*	\brief code at GFX shutdown
	*
	*	Releases textures from TextureManager. If there still exist referenced textures, those becomes
	*	zombie. \see TextureManager::erase()
	*/
	void clean();

	/**
	*	\return Surface that represents screen. May be NULL.
	*/
	SDL_Surface* getScreen();
	
	/**
	*	\brief Switch OpenGL rendering to 3D mode. Is used when rendering
	*		game world.
	*/
	void set3DOpenGL();
	
	/**
	*	\brief Switch OpenGL rendering to 2D mode. It is usefull for menu
	*		and ingame GUI.
	*/
	void set2DOpenGL();

	/**
	 * \brief Disable texturing units all but one.
	 */
	void disableMultitexturing();

	/**
	*	\brief Get width of screen or -1 if unknown.
	*/
	int getWidth();
	
	/**
	*	\brief Get height of screen or -1 if unknown.
	*/
	int getHeight();

	
	/**
	*	\brief swaps OpenGL buffers.
	*	
	*	This should be called after rendering, when we want to switch front and back buffers.
	*	You would preffer calling GFX::swap_buffers rather than SDL_GL_SwapBuffers and vice versa, 
	*	because this counts frames and updates fps counter. Fps counter could be updated somewhere
	*	else, but this is sufficient.
	*
	*/
	void swap_buffers(void);

	/**
	*	\brief actual fps ratio
	*	\return fps ratio
	*/
	int getFps(void) 
	{ return fps.get(); }

	/**
	*	\brief resets the fps counter
	*/
	void resetFps(void);

	
	/**
	 * \brief Load setting from g_setting and resize/refullcreen/.. window
	 */
	void reloadGFXSetting();

	/**
	 * \brief get "default" perspective.
	 */
	const Camera::Perspective &getPerspective()
	{ return mPerspective; }

	/**
	 * \brief Compute default perspective
	 */
	void computePerspective();	


	/**
	 * \brief Return some stamp which is incremented in every change to GFX.
	 *
	 * For example camera use the stamp to keep perspecive up to date..
	 * Allways >= 0;
	 */
	int  getConfigStamp()
	{
		return mConfigStamp;
	}
	
	/**
	 * \brief Only draw the Loading... screen
	 */
	void showLoadingScreen();

	GLfloat getMaxPointSize() const { return mMaxPointSize; }
	GLfloat getMinPointSize() const { return mMinPointSize; }
	
private:

	/**
	 * \brief Init CEGUI gui system
	 */
	void initCEGUI();

	/**
	 * \brief Try to clean as much from CEGUI as possible
	 */
	void cleanCEGUI();

	
	/**
	 * \brief Create the window
	 */
	void createWindow();
	
	/**
	*	\brief OpenGL initialization code
	*/
	void initOpenGL();


	/**
	*	\brief Loads settings from settings-file (or uses default values if not found).
	*/
	void loadVideoSettings(void);

	
	Camera::Perspective mPerspective;
	
	SDL_Surface *mScreen;
	
	/**
	 * Determines in what state the OpenGL is. 2D or 3D.
	 * It is used in set2DOpenGL() and set3DOpenGL() to
	 * not waist OpenGL commands.
	 */
	bool mGLMode3D;

	struct VideoSettings {
		struct GLAttribute{
			SDL_GLattr attr;
			int value;
		};

		int 
			width,		///< Height of the screen (window)
			height,		///< Width of the screen (window)
			bpp;
		Uint32
			SDL_flags;
		std::vector<GLAttribute> attrs;
	};

	struct VideoSettings mVS;			///< \brief current video settings
	Fps fps;							///< \brief frames-per-second counter

	int mConfigStamp;
	
	CEGUI::OpenGLRenderer *mCEGUIRenderer;
	
	//CEGUI::Window*   mLoadingScreen;
	FontRef    mFont;  ///< font for writing "Loading...."
	
	GLfloat mMaxPointSize;
	GLfloat mMinPointSize;
};

#endif
