/* MAINTAINER: Bernard */

//#define WANT_TRACE

#include "log.h"
#include "sfx.h"

#include "sound_sdl_mixer.h"
#include "sound_empty.h"
#include "platform.h"
#include "settings.h"
#include "application_client.h"
#include "taskmanager.h"
#include "conversions.h"
#include "tinyxml/tinyxml.h"

#ifndef NO_OPENAL

	#if defined(_MSC_VER)
	#include <al.h>
	#include <alc.h>
	#elif defined(__APPLE__)
	#include <OpenAL/al.h>
	#include <OpenAL/alc.h>
	#else
	#include <AL/al.h>
	#include <AL/alc.h>
	#endif	
	
	#ifdef ALUT_API_MAJOR_VERSION
	#define ALUT_ERROR_CHECK \
		{ \
			int err; \
			if ((err = alutGetError()) != ALC_NO_ERROR) \
			{ \
				LOGE << "OpenAL: " << alutGetErrorString(err) << endl; \
	/*	DIE(); \  */ \
			}\
		}
	#else
		#define ALUT_ERROR_CHECK
	#endif
	
	#include "alut.h"
	
#endif

SFX::SFX()
{
	mSoundSystem = 0;
	mSamplesEnabled = true;
	mMusicEnabled = true;
}

SFX::~SFX()
{
	// delete some singleton
	if (SoundTheme::getInstance())
	{
		SoundTheme::getInstance()->deleteMe();
	}
}

void SFX::init()
{
	TIN;
	
	assertL(mSoundSystem == 0);
	
	if (g_settings.isTrue(SETTING_ENABLE_SOUND))
	{
		mSoundSystem      = new SoundSystem_SDL();
	}
	else
	{
		mSoundSystem = new SoundSystem_Empty();
	}
	mSoundSystemEmpty = new SoundSystem_Empty();

	
	if (mSoundSystem == 0 || mSoundSystemEmpty == 0)
	{
		DIE("Sound class creation failed!");
	}

	
	// Prepare music manager
	mMusicManager.init(&mSoundSystem->getMusic());
	
	// enable/disable as setting
	enableSamples(g_settings.isTrue(SETTING_ENABLE_SAMPLES));
	enableMusic(g_settings.isTrue(SETTING_ENABLE_MUSIC));

	// load theme from setting
	loadThemesList();
	loadSettingTheme();
	
	// register music :-)
	// So the update will work
	getAppClient().getTaskManager().add(mMusicManager);	

	// apply volume
	setSamplesVolume(g_settings.getInt(SETTING_SAMPLES_VOLUME));
	setMusicVolume(g_settings.getInt(SETTING_MUSIC_VOLUME));
	
	//TODO: do something with this this is here only for testing
//	mMusicManager.loadList("music.xml"); -- allready loaded :)
	mMusicManager.play();
	
	// Init ALUT
#ifndef NO_OPENAL
		// Inicialize ALUT
	#ifdef ALUT_API_MAJOR_VERSION
		alutInitWithoutContext(0,0);
	#else
		// backward compatibility alut 0.x
		alutInit(0,0);
	#endif
		ALUT_ERROR_CHECK;	
#endif
}
	
void SFX::clean()
{
	TIN;
	
#ifndef NO_OPENAL
	alutExit();
	ALUT_ERROR_CHECK;
#endif

	// Unregister music
	getAppClient().getTaskManager().remove(mMusicManager);
		
	assertL(mSoundSystem != 0);
	delete mSoundSystem;
	mSoundSystem = 0;

	assertL(mSoundSystemEmpty != 0);
	delete mSoundSystemEmpty;
	mSoundSystemEmpty = 0;
}

ISoundSamplesManager& SFX::getSamplesManager()
{
	assertL(mSoundSystem != 0);
	assertL(mSoundSystemEmpty != 0);

	if (mSamplesEnabled)
		return mSoundSystem->getSamplesManager();
	else
		return mSoundSystemEmpty->getSamplesManager();
}

void SFX::play(const char *file)
{
//	LOGT << "PLaying sound: " << file << endl;
	
	ISoundSample *sample = getSamplesManager().newSample(file);
	sample->selfDeleteAfterStop();
	sample->play();
}


void SFX::loadTheme(const std::string &file)
{
	SoundTheme::getRef().load(locateFile(file));
	mMusicManager.reloadList();

	g_settings.set(SETTING_SOUNDTHEME, file);
}


void SFX::enableSamples(bool enable)
{
	if (mSamplesEnabled == enable) return;
	
	if (enable)
	{
		mSamplesEnabled = enable;
		g_settings.set(SETTING_ENABLE_SAMPLES, "true");
	}
	else
	{
		mSoundSystem->getSamplesManager().stopAll();
		mSamplesEnabled = enable;
		g_settings.set(SETTING_ENABLE_SAMPLES, "false");
	}
}

void SFX::enableMusic(bool enable)
{
	assertL(mSoundSystem);
	
	if (enable && !mMusicEnabled)
	{
		mMusicManager.setISM(&mSoundSystem->getMusic());
		mMusicManager.update(0);
		g_settings.set(SETTING_ENABLE_MUSIC, "true");
	}

	if (!enable && mMusicEnabled)
	{
		mSoundSystem->getMusic().stop(); // don't call mMusicManager.stop()
		mMusicManager.setISM(&mSoundSystemEmpty->getMusic());
		g_settings.set(SETTING_ENABLE_MUSIC, "false");
	}
	mMusicEnabled = enable;
}

void SFX::setSamplesVolume(int volume)
{
	// store volume to settings
	g_settings.set(SETTING_SAMPLES_VOLUME, itos(volume));
	
	// apply volume
	assertL(mSoundSystem);
	mSoundSystem->getSamplesManager().setVolume(volume);
}

void SFX::setMusicVolume(int volume)
{
	// store volume to settings
	g_settings.set(SETTING_MUSIC_VOLUME, itos(volume));
	
	// apply volume
	assertL(mSoundSystem);
	mSoundSystem->getMusic().setVolume(volume);
}

ISoundSystem& SFX::getSoundSystem()
{
	assertL(mSoundSystem);
	return *mSoundSystem;
}

void SFX::loadSettingTheme()
{
	loadTheme(g_settings.getString(SETTING_SOUNDTHEME));	
}
	
void SFX::loadThemesList()
{
	
	TiXmlDocument doc( locateFile(g_settings.getString(SETTING_SOUNDTHEMES_LIST)).c_str() );
	TiXmlHandle   handle(&doc);


	if (!doc.LoadFile()) 
	{
		LOGE << "Unable to load sound themes list. Probably file not exist or isn't a valid XML file" << endl;
		return;
	}	
	
	// Forget previous translatins list
	mThemes.clear();

	// Search whole documents for translation pairs and add them to 
	// translation table
	TiXmlElement* child = handle.FirstChild( "themes" ).FirstChild( "theme" ).Element();
	for (; child; child=child->NextSiblingElement()) 
	{
		const char *langfile   = child->Attribute("file");
		if (langfile != 0) 
		{
			mThemes.push_back(langfile);
		}
	}	
}

#ifdef NO_OPENAL
	bool SFX::getOpenALDevices(std::vector<std::string>& deviceV)
	{
		deviceV.clear();
		return false;
	}
#else
	bool SFX::getOpenALDevices(std::vector<std::string>& deviceV)
	{
		deviceV.clear();

		// Enumerate OpenAL devices
#ifdef AL_VERSION_1_1		
		if (alcIsExtensionPresent (NULL, (const ALCchar *) "ALC_ENUMERATION_EXT") == AL_TRUE)
#else
		if (alcIsExtensionPresent (NULL, (const ALCubyte *) "ALC_ENUMERATION_EXT") == AL_TRUE) // For OpenAL 1.0
#endif
		{
			const char *s = (const char *) alcGetString(NULL, ALC_DEVICE_SPECIFIER);
			while (*s != '\0')
			{
				deviceV.push_back(s);
				LOGD << "OpenAL avaliable device: " << s << endl;
				while (*s++ != '\0');
			}
			return true;
		}
		else
		{
			LOGD << "OpenAL device enumeration isn't available." << endl;
			return false;
		}
	}
#endif	

