/* MAINTAINER: Bernard */

// #define WANT_TRACE

#include "GL/glew.h"

#include "log.h"

#include "tasksound.h"
#include "taskrender.h"
#include "tasksimulation.h"
#include "taskapplyserverchanges.h"
#include "tasksendyawpitch.h"

#include "game_client.h"

#include "application_client.h"
#include "inputmanager.h"
#include "keybindings.h"
#include "networkmanager_client.h"
#include "commander.h"

#include "font_manager.h"
#include "thread.h"

#include "player.h"
#include "tank.h"
#include "world.h"

#include "inputtonetwork.h"
#include "mousemotion.h"

#include "settings.h"

#include "layer_game.h"
#include "console.h"

#include "visual_terrain.h"
#include "platform.h"


#include "tracer.h"
#include "observer.h"
#include "mousemotion.h"
#include "gamegui.h"
#include "game_state_copy.h"

#include "explosion_manager.h"

#include "timemanager.h"

GameClient::GameClient(void):
	mActualCamera(0)
{
	TIN

	mTankTracer  = new Tracer;
	mObserver    = new Observer;
	mMouseMotion = new MouseMotion;
	mGUI         = new GameGUI;
	mGameStateCopy = new GameStateCopyClient;
	mExplosionManager = new ExplosionManager;
	mStateOfGame = WaitForPlayers;
	mLastGSCommitTime = 0;

	if (!mTankTracer || !mObserver || !mMouseMotion || !mGUI || 
	    !mGameStateCopy || !mExplosionManager)
	{
		DIE("new error");
	}
}
 
GameClient::~GameClient(void)
{
	TIN
	assertL(mTankTracer && mObserver && mMouseMotion && mGUI && mGameStateCopy && mExplosionManager);

	delete mTankTracer;
	delete mObserver;
	delete mMouseMotion;
	delete mGUI;
	delete mGameStateCopy;
	delete mExplosionManager;
}

GameClient::GameClient(GameClient &copy)
{
	TIN
	
	LOGE << "It is forbidden to make copy of GameClient." << endl;
	DIE();
}


void GameClient::run()
{
	TIN;
	
	// Everybody wants to have some fun.
	mEndFun = false;
	
	// This is the main loop for client engine. How do you like it :-)
	InputManager &input = getAppClient().getInputManager();

	// Create nework keys acceptor and plug it in input
	InputToNetwork& itn = getAppClient().getKeyBindings().getInputToNetwork();
	input.setNetAcceptor(&itn);


	// Prepare tasks
	TaskSound               sound;
	TaskRender              render;
	TaskSimulation          simulation;
	TaskApplyServerChanges  applyServerChanges;
	TaskSendYawPitch        sendYawPitch;


	LayerGame layerGame;
	render.pushLayer(layerGame);
	assertL(mMouseMotion && mExplosionManager && mGUI);
	render.pushLayer(*mMouseMotion);
	render.pushLayer(*mExplosionManager);
	render.pushLayer(*mGUI);
	render.pushLayer(dynamic_cast<Layer&>(getAppClient().getConsole()));
//	mTaskManager.add(render); -will be last


	mTaskManager.add(applyServerChanges);
	mTaskManager.add(simulation);
	mTaskManager.add(sendYawPitch);
	mTaskManager.add(*mExplosionManager);
		
	mTaskManager.initTasks();
	render.init();

	// We also add font manager because font need clearing cache
	mTaskManager.add(getAppClient().getTaskManager());
	mTaskManager.add(*mGUI);

	TPOINTS("Tasks Inicialized");

	// Prepare Mouse Motion
	// NOTE that mMouseMotion is added to task manager BEFORE mTankTracer
	assertL(mMouseMotion);
	mMouseMotion->init();
	input.addInputAcceptor(*mMouseMotion);
	mTaskManager.add(*mMouseMotion);

	// Prepare tank tracer
	assertL(mTankTracer && mObserver);
	mTankTracer->resetView();
	mObserver->resetView();
	mObserver->deactivate();
	mTaskManager.add(*mTankTracer);
	mTaskManager.add(*mObserver);
	
	// Rendering is last because first frame would look strange :-)
	// if rendering would be first
	mTaskManager.add(render);

	// nono, it was not true, the sound goes last since it also has some impact on client
	if (g_settings.isTrue(SETTING_3DSOUND_ENABLE) && g_settings.isTrue(SETTING_ENABLE_SOUND))
	{ 
		sound.init();
		mTaskManager.add(sound);
	}
	

	mGUI->gameStarted();

	// Adding world entites from map - they are done by server
	
	// Enter the task loop
	gTime.enablePause(true);
	getAppClient().loop(*this, mTaskManager);
	gTime.enablePause(false);
	
	mGUI->gameEnded();
	
//	mWorld->cleanUp();

	mTaskManager.remove(getAppClient().getTaskManager());
	mTaskManager.remove(*mGUI);
	
	mTaskManager.cleanTasks();
	mTaskManager.removeAll();
}

void GameClient::loadMap(const std::string &mapName)
{
	Game::loadMap(mapName);
	
	if ( !getWorld().getLandscape().getTerrain().initClient() )
	{
		DIE("Can't initialize terrain!");
	}

	loadLandscapeEntities();
}

void GameClient::unloadMap()
{
	mTankTracer->clean();
}


void GameClient::init()
{
	Game::init();
	
	mGUI->init();

	mStateOfGame = WaitForPlayers;

	loadGFX();
}

void GameClient::clean()
{
	mGUI->clean();
	getEventQueue().cleanUp();
	
	Game::clean();
}

Player * GameClient::getClientPlayer()
{
	int id = getAppClient().getNetworkManager().getPlayerID();
	if ( isPlayerIDValid(id) )
		return &getPlayer( id );
	else
		return 0;
}

Camera*	GameClient::getCamera()
{
	if (mObserver->isActive()) return mObserver;
	return mTankTracer;
}

void GameClient::loadGFX()
{

	GLfloat 
		LightAmbient[] = {0.1f, 0.1f, 0.1f, 1.0f},
		LightDiffuse[] = {0.9f, 0.9f, 0.9f, 1.0f},
		LightSpecular[] = {1.0f, 1.0f, 1.0f, 1.0f};
	//	LightPosition[] = {100, 100, 40, 1 };

	// smooth shading
	glShadeModel( GL_SMOOTH );
	
	// -----------------------------
	// --- global lighting model ---
	// -----------------------------
	// 0.2 is the default value
	GLfloat global_ambient[] = { 0.2f, 0.2f, 0.2f, 1.0f };
	glLightModelfv( GL_LIGHT_MODEL_AMBIENT, global_ambient );						
	// apply specular lighting term AFTER texturing
	glLightModeli( GL_LIGHT_MODEL_COLOR_CONTROL, GL_SEPARATE_SPECULAR_COLOR );
	// the eye is NOT infinitely remote
	glLightModeli( GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE );

	// light parameters
	glLightfv(GL_LIGHT0, GL_AMBIENT, LightAmbient);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, LightDiffuse);
	glLightfv(GL_LIGHT0, GL_SPECULAR, LightSpecular);
	// light position (or direction) has to be set after view transformation in order to
	// place it correctly with respect to BJS world coordinate system
	//glLightfv(GL_LIGHT0, GL_POSITION, LightPosition);

	// enable lighting and color material
	// Only the LIGHT0 has preset default values! Other light have zeros as defaults.
	glEnable(GL_LIGHT0);
	glEnable(GL_COLOR_MATERIAL);
	
	// development purposes
	/*
	if (g_settings.isTrue(SETTING_MARTIN_TMP_RENDER_WIREFRAME))
		glPolygonMode(GL_FRONT, GL_LINE);
	*/
}
