/* MAINTAINER: Bernard */

#ifndef _GAME_CLIENT_H
#define _GAME_CLIENT_H

#include "game.h"
#include "frame.h"
#include "log.h"


class Tracer;
class Observer;
class MouseMotion;
class GameGUI;
class GameStateCopyClient;
class Player;
class Camera;
class ExplosionManager;

class GameClient : public Game, public Frame
{

public:
	GameClient(void);
	virtual ~GameClient(void);

	
	/**
	 * Start the game.
	 */
	virtual void run();

	virtual bool wantRun() { return !mEndFun; }
	virtual void quitWanted() { endFun(); }
	
	/*
	*	\return Player's camera (actual camera on client, exactly).
	*	\note This isn't necessary tank's tracer. The tank doesn't have to 
	*		exist and this can be the "freelook camera" or whatever else.
	*/
	Camera*	getCamera();
	
	virtual void init();

	virtual void clean();

	virtual void loadMap(const std::string &mapName);
	
	virtual void unloadMap();
	

	/**
	 * \brief Return player controled by this client.
	 *
	 * \return player for this client.
	 */
	Player *getClientPlayer();

	/**	\brief Gets current state of game - wether we are waiting for players to get ready, playing, or en of map
	 *
	 *	\return Current state of game
	 */
	eStateOfGame getStateOfGame() const
	{
		return mStateOfGame;
	}

	/**	\brief Sets new state of game - wether we are waiting for players to get ready, playing, or en of map
	 *
	 *	\param state  New state of game
	 */
	void setStateOfGame( eStateOfGame state )
	{
		mStateOfGame = state;
	}

	/**	\brief Sets time of last gamestate commit
	 *
	 *	\param time commit time
	 */
	void setCommitTime( int time )
	{
		mLastGSCommitTime = time;
	}

	/**	\brief Gets time of last gamestate commit
	 *
	 *	\return Last commit time
	 */
	int getCommitTime() const
	{
		return mLastGSCommitTime;
	}

	/**
	 * \brief Return class that hold some variables from game state.
	 *
	 * The returned class is important because access to it is thread safe.
	 * So you may access these variables as you wisch. The copy is old at most
	 * one frame of game. So it doesn't have to be the newest but new.
	 */
	GameStateCopyClient &getGameStateCopy() { assertL(mGameStateCopy); return *mGameStateCopy; }
	ExplosionManager    &getExplosionManager() {assertL(mExplosionManager); return *mExplosionManager; }
	Tracer&      getTankTracer()  { assertL(mTankTracer);  return *mTankTracer; }
	MouseMotion& getMouseMotion() { assertL(mMouseMotion); return *mMouseMotion; }
	Observer&    getObserver()    { assertL(mObserver);    return *mObserver; }
	GameGUI&     getGUI()         { assertL(mGUI);         return *mGUI; }
private:

	void loadGFX();

	/**
	 * \brief Program killer - trap
	 * 
	 * From security reasons we don't allow anyone to create
	 * copy of GameClient since it is probably a mistake.
	 * Anyway calling this constructor will kill program.
	 * \param copy not used
	 */
	GameClient(GameClient &copy);

	
	
	Camera * mActualCamera; ///< actual view, it can be one of tracers
	
	//TODO: Do some tracer-managing. We could probably want to have got a tracer
	//	for the player's tank, for some team-tanks, for rockets, etc.
	Tracer      *mTankTracer;
	Observer    *mObserver;
	MouseMotion *mMouseMotion;
	GameGUI     *mGUI;
	GameStateCopyClient  *mGameStateCopy;
	ExplosionManager *mExplosionManager;

	eStateOfGame mStateOfGame;		///< \brief Client here remembers state of game
	int			mLastGSCommitTime;	///< \brief time of last commit of gamestate
};

#endif // _GAME_CLIENT_H
