/* MAINATINER: Martin */

/* CLIENT */

#include "platform.h"
#include "terrain.h"
#include "terrain_base.h"
#include "visual_terrain.h"

#include "weightfun.h"
#include "noise.h"

#include "terrain_geomipmap.h"
typedef class TerrainGeoMipMap CurrentTerrainImplementation;

#include <typeinfo>

#include "terrain_map_attributes.h"

// <map_settings_access>
#include "application_client.h"
#include "game_client.h"
#include "globals_client.h"
#include "params.h"
// </map_settings_access>

bool Terrain::initClient()
{
	//const int GLOBAL_TEXTURE_EXTENT = 64;
	//const int MIXMAP_TEXTURE_EXTENT = 64;

	const int MIXMAP_TEXTURES_NUM = 1;

	// shortcuts for some types
	typedef TerrainSurface::SurfaceLayer::Tile Tile;
	typedef TerrainSurface::SurfaceLayer::WeightFunctionPtr wfptr;
	typedef TerrainWeightFunction::Rectangle rect;
	typedef counted_ptr<TerrainWeightFunction::NoiseFunction> nfptr;
	typedef ::TerrainAttributes::SurfaceAttributes::layers_container_type LayersContainer;

	// get access to map parameters, especially to terrain attributes
	Params::sMapParams &map = getAppClient().getGame().getMapParams();
	const TerrainAttributes &ta = map.terrain;

	// create new terrain, ensure valid terrain base pointer
	mTerrainBase.reset( new CurrentTerrainImplementation );
	TerrainBase * tb = getTerrainBase();
	if ( tb == 0 )
	{
		LOGD << "Failed to create new terrain."  << std::endl;
		return false;
	}
	
	// common init ( load a heightmap, set virtual extents )
	if ( !initCommon() )
	{
		LOGD << "Common terrain initialization failed!" << std::endl;
		return false;
	}

	// create terrain surface layers using map attributes
	for ( LayersContainer::const_iterator it( ta.surface.surfaceLayers.begin() ), ite( ta.surface.surfaceLayers.end() ); it != ite; ++it )
	{
		// one layer description
		LayersContainer::const_reference ref = (*it);

		// weight function ( layer coverage ) - create
		TerrainWeightFunction * twf;
		wfptr wf( twf = new TerrainWeightFunction( tb ) );
		if ( twf )
		{
			// parametric description
			twf->setParameters(

				rect(
					ref.coverage.domain.height1,
					ref.coverage.domain.height2,
					ref.coverage.domain.slope1,
					ref.coverage.domain.slope2 ), 

				rect(
					ref.coverage.full.height1,
					ref.coverage.full.height2,
					ref.coverage.full.slope1,
					ref.coverage.full.slope2 ), 

				ref.noiseAmount

			);
		
			// noise - assign to weight function
			if ( !ref.noise.fileName.empty() )
			{
				nfptr n( new ExternalNoise( locateFile( ref.noise.fileName ) ) );
				if ( n.get() != 0 )
					twf->setNoise( n );
			}
		}

		// create new terrain surface layer p
		TerrainSurface::LayersContainer::value_type p = tb->createLayer();
		if ( p.get() )
		{
		
			// weight function - assign to p
			p->setWeightFunction( wf );
			
			// layer color - assign to p
			p->setColorFunction( 
				TerrainSurface::SurfaceLayer::ColorFunctionPtr(
					new TerrainSurface::SurfaceLayer::TerrainConstantColorFunction(
						ref.color.constantColor.r / 255.0, 
						ref.color.constantColor.g / 255.0, 
						ref.color.constantColor.b / 255.0 ) ) );

			// detail tile
			if ( !ref.detail.fileName.empty() )
			{
				p->setDetailTile(
					Tile::createTile(
						locateFile( ref.detail.fileName ), 
						ref.detail.extentsInPixels.width, 
						ref.detail.extentsInPixels.height, 
						ref.detail.extentsInMeters.width, 
						ref.detail.extentsInMeters.height));
			}
		} 
		else 
		{
			LOGD << "Failed to create new terrain surface layer!" << std::endl;
		}
	}

	// --- TEXTURES ---

	// try to get visual terrain subclass
	VisualTerrain * vt(0);
	try 
	{
		vt = dynamic_cast<VisualTerrain*>( tb );
	} 
	catch ( std::bad_cast ) 
	{
		LOGD << "Can't cast TerrainBase to VisualTerrain." << std::endl;
		return false;
	}


	// lightmap
	vt->loadLightmap( locateFile( map.terrain.surface.shadowTextureFileName ) );

	// create terrain global color texture
	vt->createGlobalMapTexture( ta.surface.globalTextureExtents.width, ta.surface.globalTextureExtents.height );
	
	// select surface layers for detail map and mix map creation
	VisualTerrain::LayerSelection laysel;
	vt->selectLayers(laysel);

	// create mix and detail textures
	vt->createMixTextures( laysel, ta.surface.mixmapExtents.width, ta.surface.mixmapExtents.height, MIXMAP_TEXTURES_NUM);
	vt->createDetailTextures( laysel, MIXMAP_TEXTURES_NUM );

	vt->init();
	
	return true;
}

//------------------------------------
// WRAPPING
//------------------------------------

void Terrain::render(void)
{
	if ( this->getTerrainBase() != 0 )
		dynamic_cast<VisualTerrain*>(this->getTerrainBase())->render();
}

void Terrain::update(const class Camera * camera)
{
	if ( this->getTerrainBase() != 0 )
		dynamic_cast<VisualTerrain*>(this->getTerrainBase())->update(camera);
}

void Terrain::loadLightmap(std::string path)
{
	if ( this->getTerrainBase() != 0 )
		dynamic_cast<VisualTerrain*>(this->getTerrainBase())->loadLightmap(path);
}

void Terrain::loadGlobalTexture(std::string path)
{
	if ( this->getTerrainBase() != 0 )
		dynamic_cast<VisualTerrain*>(this->getTerrainBase())->loadGlobalTexture(path);
}
