/* MAINTAINER: Martin */

#ifndef _TERRAIN_GEOMIPMAP_H
#define _TERRAIN_GEOMIPMAP_H

#include "visual_terrain.h"
#include "frustum.h"
#include "camera.h"
#include "hierarchy_tree.h"

/**
*	\brief GeoMipMap LOD terrain representation
*	
*	Divides heightmap into terrain blocks. Introduces full binary HierarchyTree that is builded over terrain blocks.
*	Nodes of this tree contain 3D bounding-box information and are used for frustum culling.
*/
class TerrainGeoMipMap: public VisualTerrain
{
public:

	/**
	*	\brief Node of hierarchy tree
	*	
	*	Contains 3D bounding box and frustum-culling state.
	*/
	// TODO: more powerful visibility testing using occluders
	class Node: public HierarchyTreeBase::Node {
	public:
		typedef Frustum::ClipState ClipState;

		/**
		*	\brief gets previously set frustum state - used for frustum culling
		*	\return Frustum::ClipState (in/out/intersect)
		*/
		ClipState getFrustumClipState(void) { return mFrustumClipState; }

		/**
		*	\brief sets frustum state - used for frustum culling
		*	\param Frustum::ClipState (in/out/intersect)
		*/
		void setFrustumClipState(ClipState state) { mFrustumClipState = state; }
		
		/**
		*	\brief sets user-defined 3D bounding box
		*	\param center bounding box's center point
		*	\param extent bounding box's extent vector (extent for each axis of this box)
		*/
		void setBoundingBox(const float center[3], const float extent[3])
		{ 
			for (int i = 0; i < 3; ++i) {
				this->center[i] = center[i]; 
				this->extent[i] = extent[i];
			}
		}

		/**
		*	\brief gets user-defined 3D bounding box stored in this node
		*	\param center bounding box's center point
		*	\param extent bounding box's extent vector (extent for each axis of this box)
		*/
		void getBoundingBox(float center[3], float extent[3]) const
		{ 
			for (int i = 0; i < 3; ++i){
				center[i] = this->center[i]; 
				extent[i] = this->extent[i];
			}
		}

		const float * getBoundingBoxCenter(void) const
		{ return this->center; }

		const float * getBoundingBoxExtent(void) const
		{ return this->extent; }

	private:
		ClipState mFrustumClipState; ///< \brief node outside frustum, inside frustum, or intersect (don't know)
		
		struct {
			float center[3]; ///< \brief The center point of user's bounding box.
			float extent[3]; ///< \brief The extent of user's bounding box.
		};
	};

	/** \brief tree over terrain blocks, used for frustum culling */
	typedef HierarchyTree<Node> TerrainHierarchyTree;

private:
	/**
	*	\brief small block of terrain 
	*
	*	It is able to render itself.
	*	In future, there could be some data for optimization, i.e. texture info, geo-mip-map-level, etc.
	*/
	class TerrainBlock
	{
	public:
		struct BlockDescriptor
		{
			union {
				struct {
					int 
						x, ///< \brief starting x-coord of block - in heightmap units
						y; ///< \brief starting y-coord of block - in heightmap units
				};
				int coord[2]; ///< \brief vector access, x is coord[0], y is coord[1]
			};
			union {
				struct{
					int
						w, ///< \brief block width - in heightmap units
						h; ///< \brief block width - in heightmap units
				};
				int extent[2]; ///< \brief vector access, w is extent[0], h is extent[1]
			};

			TerrainGeoMipMap *terr; ///< \brief reference to terrain base.
		};
		
		enum NEIGHBOUR_ID { N_LEFT = 0, N_RIGHT = 1, N_TOP = 2, N_BOTTOM = 3 };

		TerrainBlock(): mActualLevel(0), mLevels(0) {}

		void setNeighbours(TerrainBlock * neighbours[4])
		{ 
			for (int i = 0; i < 4; ++i)
				mNeighbours[i] = neighbours[i];
		}

		void initLevels(const BlockDescriptor &bd);
		
		void selectLevel(const TerrainGeoMipMap::TerrainBlock::BlockDescriptor &bd, double blockSqrDistance, double pixelTolerance);

		/**
		*	\brief how many vertices to skip horizontaly/verticaly when the block is being rendered
		*	\return LOD step
		*/
		int getStep(void) { return 1 << mActualLevel; }

		/**
		*	\brief current LOD level of this block
		*	\return non-negative LOD level number, 0 is full detail
		*/
		int getLevel(void) { return mActualLevel; }

		/**
		*	\brief renders one terrain block
		*	
		*	Storing block descriptor inside each of blocks would be possible, but redundant.
		*	Instead of doing this, block descriptor is provided from caller.
		*
		*	\param bd block descriptor
		*/
		void render(const BlockDescriptor &bd);

	private:
		
		void render(const BlockDescriptor &bd, int x1, int y1, int x2, int y2, int step);
				
		typedef void (TerrainGeoMipMap::TerrainBlock::*PRenderTriFanMethod)(TerrainBase *, int points[4][2], int step, int axis);

		void renderTriFanA(TerrainBase * tb, int points[4][2], int step, int axis);
		void renderTriFanB(TerrainBase * tb, int points[4][2], int step, int axis);
		
		void blend(const BlockDescriptor &bd, int step2, bool b1, bool b2, int axis, int blendStripCoord1, int blendStripCoord2, PRenderTriFanMethod renderTriFan);
		void blendLeft(const BlockDescriptor &bd, int step2, bool b1, bool b2);
		void blendTop(const BlockDescriptor &bd, int step2, bool b1, bool b2);
		void blendRight(const BlockDescriptor &bd, int step2, bool b1, bool b2);
		void blendBottom(const BlockDescriptor &bd, int step2, bool b1, bool b2);
		
		/**
		*	\brief computes non-scaled signed difference value at the line [[x1, y1], [x2, y2]]
		*
		*	Computes and returns the difference between interpolated height value in the middle of line 
		*	[[x1, y1], [x2, y2]] and the real height sample in the middle.
		*
		*	\return signed difference value between [0.0, 255.0]
		*/
		float getDelta(const BlockDescriptor &bd, int x1, int y1, int x2, int y2)
		{			
			return bd.terr->getTrueHeight((x1 + x2) / 2, (y1 + y2) / 2) - 0.5f * (bd.terr->getTrueHeight(x1, y1) + bd.terr->getTrueHeight(x2, y2));			
		}

		float getFovX(const BlockDescriptor &bd);
		int getViewHRes(const BlockDescriptor &bd);

		/**
		*	\brief minimum distance of level with given delta
		*/
		float getMinDistance(const BlockDescriptor &bd, float delta)
		{			
			return fabs( getViewHRes(bd) * delta / ( 2.0f * tan( (float)(getFovX(bd) * M_PI / 360.0) ) ) );
		}

		int getLevelsCount(const BlockDescriptor &bd) 
		{ return bd.terr->getBlockShift() + 1; }

		TerrainBlock * getNeighbour(const NEIGHBOUR_ID id);

		class GeoMipMapLevel 
		{
		public:
			float getSqrDist(void) const { return mSqrDist; }
			void setSqrDist(float sqrDist) { mSqrDist = sqrDist; }
		private:
			float mSqrDist;	///< \brief minimum distance at which this level can be used
		};
		
		// actual geomipmap level
		int mActualLevel;
		
		// geomipmap levels
		GeoMipMapLevel *mLevels;
		
		TerrainBlock * mNeighbours[4];
	};

	/**
	*	\brief Node delegate for hierarchy tree
	*
	*	Computes 3D bounding boxes when building the hierarchy tree.
	*/
	class NodeDelegate : public TerrainHierarchyTree::NodeDelegate
	{
	public:
		NodeDelegate(): mTerrain(0){}
		virtual ~NodeDelegate(){}

		virtual void nodeCreation(TerrainHierarchyTree::NodeNavigator &navig);

		void setTerrain(TerrainGeoMipMap * terr) {
			mTerrain = terr; 
		}

	private:
		TerrainGeoMipMap * getTerrain(void) { return mTerrain; }
		TerrainGeoMipMap * mTerrain;
	};

public:

	TerrainGeoMipMap(int blockSize = 65);
	virtual ~TerrainGeoMipMap(){}

	/**
	*	\brief Performs rendering - setups texturing environment and draws geometry. [CLIENT]
	*/
	virtual void render(void);

	/**
	*	\brief renders terrain geometry, no texturing
	*	If frustum culling is enabled, culls terrain blocks against frustum.
	*/
	void drawGeometry();

	/**
	*	\brief Data update (LOD-related things). [CLIENT]
	*
	*	Frustum culling, GeoMipMap level selection for each visible block.
	*/
	virtual void update(const class Camera * camera);

	/**
	*	\brief preprocessing step - offline calculations for LOD
	*	
	*	Creates blocks over the heightmap, builds quadtree over blocks, 
	*	calculates error values.
	*/
	virtual void init(void);

	/**
	*	\brief gets binary shift of terrain's block
	*	\return Shift value of block size ( size of block is (1 << shift) + 1 )
	*/
	int getBlockShift(void) { return mBlockShift; }
	
	/**
	*	\brief gets block size
	*	\return block size computed from block shift value
	*/
	int getBlockSize(void) { return (1 << getBlockShift()) + 1; }

	/**
	*	\brief enables/disables frustum culling
	*	\param enabled true enables frustum, false disables
	*/
	void setFrustumState(bool enabled) { mFrustumEnabled = enabled; }
	
	//void getFrustumState(void) { return mFrustumEnabled; }

	//const Frustum & getFrustum(void) { return mFrustum; }

	float getFovX(void){ return 60; }

private:

	/**
	*	\brief Creates block structure over heightmap.
	*	\note Uses block size specified in ctor.
	*/
	void createBlocks();

	void initBlocks();

	/**
	*	\brief builds the hierarchy tree over blocks
	*/
	void createHierarchyTree();

	/**
	*	\brief performs frustum culling
	*	\param navigator starting node
	*/
	void frustumCulling(TerrainHierarchyTree::NodeNavigator & navigator);

	void processVisibleBlock(TerrainBlock::BlockDescriptor &bd, TerrainBlock & block);	

	/**
	*	\brief rendering using frustum culling
	*
	*	It recursively traverses the tree starting at node \c navigator, skips invisible nodes.
	*
	*	\param navigator starting node
	*/
	void hierarchicRender(const TerrainHierarchyTree::NodeNavigator & navigator);

	/**
	*	\brief render terrain boundary at given block if there exists any
	*	\param x block x coordinate ( 0 .. mBlocksW )
	*	\param y block y coordinate ( 0 .. mBlocksH )
	*	\param bd terrain block descriptor
	*/
	void renderBlockBoundaryIfAny( int x, int y, const TerrainBlock::BlockDescriptor &bd );

	/**
	*	\brief creates block descriptor from given block's coordinates \c bx, \c by (in block matrix).
	*	\param bx x-coordinate of block in block-matrix, zero-based (i.e. second block will have bx == 1)
	*	\param by y-coordinate of block in block-matrix, zero-based
	*	\param d  descriptor to fill
	*	\return reference to \c d
	*/
	TerrainBlock::BlockDescriptor& setDescriptor(int bx, int by, TerrainBlock::BlockDescriptor & d);

	/**
	*	\brief Array of blocks covering the whole terrain.
	*/
	TerrainBlock* mBlocks;
	int 
		mBlocksW, ///< \brief the number of blocks in x-axis
		mBlocksH; ///< \brief the number of blocks in y-axis
	
	/**
	*	\brief block's binary shift value
	*
	*	Determines the block size. Block size will be ((1 << mBlockShift) + 1) - this means that 
	*	all blocks are squares with common power-of-two-plus-one size. 
	*	The reason for "POT size" is, that each GeoMipMap level has steps of 2^n size.
	*	The reason for "POT+1 size" is, that terrain blocks shares their edges with other blocks.
	*/
	int mBlockShift;
	
	NodeDelegate mNodeDelegate;				///< \brief node delegate that is provided to hierarchy tree
	TerrainHierarchyTree mHierarchyTree;	///< \brief hierarrchy tree over terrain blocks, used for frustum culling 

	GLVector3d mLocation;	///< \brief actual location of camera
	Frustum mFrustum;		///< \brief actual frustum, got from camera

	bool mFrustumEnabled;
	bool mFrustumNeedsUpdate;

};

#endif // _TERRAIN_GEOMIPMAP_H
