/* MAINATINER: Martin */
#include "terrain_roetger.h"
#include "camera.h"

TerrainRoetger::TerrainRoetger():
	m_minimum_global_res(20),
	m_desired_global_res(12),
	m_faces_rendered(0),
	m_roughness(0)
{
}

TerrainRoetger::~TerrainRoetger(void)
{
	delete[] m_roughness; m_roughness = 0;
}

void TerrainRoetger::init(void)
{
	calculate_roughness();
}

bool TerrainRoetger::allocateHeightmap(int sizeX, int sizeY)
{
	int size(std::max(sizeX, sizeY));
	m_roughness = new float [size * size];
	m_qt.Alloc(size);
	return TerrainBase::allocateHeightmap(size, size);
}

void TerrainRoetger::update(const class Camera * camera)
{
	GLVector3d pos, inv;
	getRealScaleInv(inv);
	camera->getPos(pos);
	pos = pos * inv;
	m_location = pos;
	
	mFrustum = camera->getFrustum();
	
	GLMatrix4d m;
	m = m.identity();
	GLVector3d scale(getRealScale());
	m.applyScale(scale);
	mFrustum.transform(m, true);

	build_quadtree();
}

void TerrainRoetger::render(void)
{	
	glDisable(GL_LIGHTING);
	//glActiveTexture(0);
	glEnable(GL_TEXTURE_2D);

	int sizeX, sizeY;
	getExtents(sizeX, sizeY);
	// TODO: preserving/restoring modelview matrix should be avoided!

	// Preserve modelview matrix.
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	
	// Set scale.
	getRealScale().glScale();
	
	// s, t vectors for generating global texture coordinates
	GLVector4f tv[2] = {
		GLVector4f(1.0f/sizeX, 0.0f, 0.f, 0.0f),
		GLVector4f(0.0f, 1.0f/sizeY, 0.f, 0.0f)
	};

	// generate texture coordinates
	glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
	glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
	glTexGenfv(GL_S, GL_OBJECT_PLANE, tv[0].val);
	glTexGenfv(GL_T, GL_OBJECT_PLANE, tv[1].val);
	glEnable(GL_TEXTURE_GEN_S);
	glEnable(GL_TEXTURE_GEN_T);

	mGlobalTexture->bind();
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

	/* <rendering> */

	render_quadtree();

	/* <\rendering> */


	glDisable(GL_TEXTURE_GEN_S);
	glDisable(GL_TEXTURE_GEN_T);

	glLoadIdentity();

	glDisable(GL_TEXTURE_2D);

	// Restore modelview matrix.
	glPopMatrix();
}

void TerrainRoetger::set_minimum_quality(float minimum_global_res)
{
  if (minimum_global_res <= 2.0f)
    minimum_global_res = 2.0f;
  if (minimum_global_res == m_minimum_global_res)
    return;
  m_minimum_global_res = minimum_global_res;
  calculate_roughness();
  build_quadtree();
}

void TerrainRoetger::set_desired_quality(float desired_global_res)
{
  m_desired_global_res = (desired_global_res <= 1.0f) ? 1.0f : desired_global_res;
}

void TerrainRoetger::add_height_error(int x1, int y1, int x2, int y2, float * max_error)
{
  double h1, h2, hmid, herr;

  h1 = getScaledHeight(x1, y1);
  h2 = getScaledHeight(x2, y2);
  hmid = getScaledHeight((x1+x2)>>1, (y1+y2)>>1);
  herr = fabs(hmid - 0.5f * (h1 + h2));
  if (herr > *max_error)
    *max_error = static_cast<float>(herr);
}

void TerrainRoetger::calculate_roughness(void)
{
	int size, h;
	getExtents(size, h);

	int x, y, step, w2 = size >> 1, i;
  float d2, t, max_rough = 0.0f;
  const float K = m_minimum_global_res / (2.0f*(m_minimum_global_res - 1.0f));
  
  for (x = 0; x < size; ++x)
    m_roughness[x] = 0.0f;

  for (step = 2; step < size; step <<= 1){
    w2 = step >> 1;
    for (y = w2; y < size - w2; y += step){
      for (x = w2; x < size - w2; x += step){
        // maximalize d2 along middles of four edges and two diagonals
        d2 = 0;
        add_height_error(x-w2, y+w2, x+w2, y+w2, &d2);
        add_height_error(x-w2, y-w2, x+w2, y-w2, &d2);
        add_height_error(x-w2, y+w2, x-w2, y-w2, &d2);
        add_height_error(x+w2, y+w2, x+w2, y-w2, &d2);
        add_height_error(x-w2, y+w2, x+w2, y-w2, &d2);
        add_height_error(x+w2, y+w2, x-w2, y-w2, &d2);

        // calculate d2 value
        d2 /= step;

        if (step > 2){
          // Opravime d2 abysme dostali absolutni rozdil urovni sousednich uzlu
          // roven 1, pokud budeme provadet deleni podle kriteria s d2.
          if ((t=K*m_roughness[index(x, y+w2)]) > d2) d2 = t;
          if ((t=K*m_roughness[index(x, y-w2)]) > d2) d2 = t;
          if ((t=K*m_roughness[index(x+w2, y)]) > d2) d2 = t;
          if ((t=K*m_roughness[index(x-w2, y)]) > d2) d2 = t;
        }          
        
        // distribute error to four corners of its square, 
        // maximalize with d2
        if (m_roughness[i=index(x-w2, y+w2)] < d2) m_roughness[i] = d2;
        if (m_roughness[i=index(x+w2, y+w2)] < d2) m_roughness[i] = d2;
        if (m_roughness[i=index(x+w2, y-w2)] < d2) m_roughness[i] = d2;
        if (m_roughness[i=index(x-w2, y-w2)] < d2) m_roughness[i] = d2;

        if (max_rough < d2) max_rough = d2;
        m_roughness[index(x, y)] = d2;        
      }    
    }
  }  
  if (max_rough){
    float d2_scale = 1.0f/max_rough;
    for (step = 2; step < size; step <<= 1){
      w2 = step >> 1;
      for (y = w2; y < size - w2; y += step)
        for (x = w2; x < size - w2; x += step)
          m_roughness[index(x, y)] *= d2_scale;
    }
  }
}

void TerrainRoetger::build_quadtree(const Quadtree::QuadtreeNode &node)
{
	
	if (node.width >= 16){
	
		for (int pl = 0; pl < 6; ++pl){
			if (pl == Frustum::PL_BOTTOM || pl == Frustum::PL_TOP)
				continue;
		int outside = 0;		
		// 4 points of the quad
		for (int i = -1; i <= 1; i+=2)
			for (int j = -1; j <= 1; j+=2)
			{
				// test against one plane
				Frustum::Halfspace h;
				h = mFrustum.classifyPoint(mFrustum.getPlane((Frustum::TPlaneID)pl), GLVector3d(node.x+i*node.width/2.0f, node.y+j*node.width/2.0f, 0.0f));
				if (h == Frustum::NEGATIVE){
					outside++;
				}
			}
		// all points are outside one plane
		if (outside == 4){
			m_qt.set_leaf(node);
			return;
		}
		}
	}
	
  if (subdivision_criterium(node) == true)
  {
    Quadtree::QuadtreeNode subnodes[4];
    
    // Subdivide this node:
    // Mark this node as a subdivided node, obtain it's subnodes and 
    // recursively continue building quadtree.
    m_qt.set_node(node);
    if ((node.width > 3) && (m_qt.get_subnodes(node, subnodes) == true))
      for (int i = 0; i < 4; ++i)
        build_quadtree(subnodes[i]);
  }else{
    // Do not subdivide, set this node as a leaf.
    m_qt.set_leaf(node);
  }
}

void TerrainRoetger::emit_vertices(const GLVector3d &v1, const GLVector3d & v2, const GLVector3d & v3)
{
	v1.glVertex();
	v2.glVertex();
	v3.glVertex();
	++m_faces_rendered;
}

void TerrainRoetger::render_quadtree(const Quadtree::QuadtreeNode &node)
{
  //  Quadtree::get_subnodes(node, subnodes) - subnodes[] order
  // + | 0 1
  // y | 3 2
  // - L----
  //   - x +
  //

  int w2 = node.width >> 1, x = node.x, y = node.y;    
  if (w2 >= 1){
    int xp, yp;
    GLVector3d v1, v2, v3;
    Quadtree::QuadtreeNode subnodes[4], neighbor;
    // we are not at finest level, render possible leaves and
    // render other nodes recursively
    if (!m_qt.get_subnodes(node, subnodes) && (w2 != 1))
      return;

    bool b0, b1, b2, b3;
    if (w2 != 1){
      b0 = m_qt.is_subdivided(subnodes[0]);
      b1 = m_qt.is_subdivided(subnodes[1]);
      b2 = m_qt.is_subdivided(subnodes[2]);
      b3 = m_qt.is_subdivided(subnodes[3]);

      if (b0) render_quadtree(subnodes[0]);
      if (b1) render_quadtree(subnodes[1]);
      if (b2) render_quadtree(subnodes[2]);
      if (b3) render_quadtree(subnodes[3]);
    }else
      b0 = b1 = b2 = b3 = false;

    glBegin(GL_TRIANGLES);

    xp = x; yp = y; v1 = GLVector3d(static_cast<float>(xp), static_cast<float>(yp), static_cast<float>(getTrueHeight(xp, yp)));
    if (!(b0 || b1) && m_qt.get_neighbor(Quadtree::n_up, node, neighbor) && !m_qt.is_subdivided(neighbor)){
      xp=x+w2; yp=y+w2; v2=GLVector3d(static_cast<float>(xp), static_cast<float>(yp), static_cast<float>(getTrueHeight(xp, yp)));
      xp=x-w2; yp=y+w2; v3=GLVector3d(static_cast<float>(xp), static_cast<float>(yp), static_cast<float>(getTrueHeight(xp, yp)));
      emit_vertices(v1, v2, v3);
    }else{
      if (!b0){
        xp=x; yp=y+w2; v2=GLVector3d(static_cast<float>(xp), static_cast<float>(yp), static_cast<float>(getTrueHeight(xp, yp))); 
        xp=x-w2; yp=y+w2; v3=GLVector3d(static_cast<float>(xp), static_cast<float>(yp), static_cast<float>(getTrueHeight(xp, yp))); 
        emit_vertices(v1, v2, v3);
      }
      if (!b1){
        xp=x+w2; yp=y+w2; v2=GLVector3d(static_cast<float>(xp), static_cast<float>(yp), static_cast<float>(getTrueHeight(xp, yp))); 
        xp=x; yp=y+w2; v3=GLVector3d(static_cast<float>(xp), static_cast<float>(yp), static_cast<float>(getTrueHeight(xp, yp))); 
        emit_vertices(v1, v2, v3);
      }
    }
    
    if (!(b0 || b3) && m_qt.get_neighbor(Quadtree::n_left, node, neighbor) && !m_qt.is_subdivided(neighbor)){      
      xp=x-w2; yp=y+w2; v2=GLVector3d(static_cast<float>(xp), static_cast<float>(yp), static_cast<float>(getTrueHeight(xp, yp))); 
      xp=x-w2; yp=y-w2; v3=GLVector3d(static_cast<float>(xp), static_cast<float>(yp), static_cast<float>(getTrueHeight(xp, yp)));
      emit_vertices(v1, v2, v3);

    }else{
      if (!b0){
        xp=x-w2; yp=y+w2; v2=GLVector3d(static_cast<float>(xp), static_cast<float>(yp), static_cast<float>(getTrueHeight(xp, yp))); 
        xp=x-w2; yp=y; v3=GLVector3d(static_cast<float>(xp), static_cast<float>(yp), static_cast<float>(getTrueHeight(xp, yp))); 
        emit_vertices(v1, v2, v3);
      }
      if (!b3){
        xp=x-w2; yp=y; v2=GLVector3d(static_cast<float>(xp), static_cast<float>(yp), static_cast<float>(getTrueHeight(xp, yp))); 
        xp=x-w2; yp=y-w2; v3=GLVector3d(static_cast<float>(xp), static_cast<float>(yp), static_cast<float>(getTrueHeight(xp, yp))); 
        emit_vertices(v1, v2, v3);
      }
    }

    if (!(b1 || b2) && m_qt.get_neighbor(Quadtree::n_right, node, neighbor) && !m_qt.is_subdivided(neighbor)){      
      xp=x+w2; yp=y-w2; v2=GLVector3d(static_cast<float>(xp), static_cast<float>(yp), static_cast<float>(getTrueHeight(xp, yp))); 
      xp=x+w2; yp=y+w2; v3=GLVector3d(static_cast<float>(xp), static_cast<float>(yp), static_cast<float>(getTrueHeight(xp, yp)));
      emit_vertices(v1, v2, v3);

    }else{
      if (!b1){
        xp=x+w2; yp=y; v2=GLVector3d(static_cast<float>(xp), static_cast<float>(yp), static_cast<float>(getTrueHeight(xp, yp))); 
        xp=x+w2; yp=y+w2; v3=GLVector3d(static_cast<float>(xp), static_cast<float>(yp), static_cast<float>(getTrueHeight(xp, yp))); 
        emit_vertices(v1, v2, v3);
      }
      if (!b2){
        xp=x+w2; yp=y-w2; v2=GLVector3d(static_cast<float>(xp), static_cast<float>(yp), static_cast<float>(getTrueHeight(xp, yp))); 
        xp=x+w2; yp=y; v3=GLVector3d(static_cast<float>(xp), static_cast<float>(yp), static_cast<float>(getTrueHeight(xp, yp))); 
        emit_vertices(v1, v2, v3);
      }
    }
    
    if (!(b3 || b2) && m_qt.get_neighbor(Quadtree::n_down, node, neighbor) && !m_qt.is_subdivided(neighbor)){
      xp=x-w2; yp=y-w2; v2=GLVector3d(static_cast<float>(xp), static_cast<float>(yp), static_cast<float>(getTrueHeight(xp, yp))); 
      xp=x+w2; yp=y-w2; v3=GLVector3d(static_cast<float>(xp), static_cast<float>(yp), static_cast<float>(getTrueHeight(xp, yp)));
      emit_vertices(v1, v2, v3);

    }else{
      if (!b3){
        xp=x-w2; yp=y-w2; v2=GLVector3d(static_cast<float>(xp), static_cast<float>(yp), static_cast<float>(getTrueHeight(xp, yp))); 
        xp=x; yp=y-w2; v3=GLVector3d(static_cast<float>(xp), static_cast<float>(yp), static_cast<float>(getTrueHeight(xp, yp))); 
        emit_vertices(v1, v2, v3);
      }
      if (!b2){
        xp=x; yp=y-w2; v2=GLVector3d(static_cast<float>(xp), static_cast<float>(yp), static_cast<float>(getTrueHeight(xp, yp))); 
        xp=x+w2; yp=y-w2; v3=GLVector3d(static_cast<float>(xp), static_cast<float>(yp), static_cast<float>(getTrueHeight(xp, yp))); 
        emit_vertices(v1, v2, v3);
      }
    }
    glEnd();
  }
}
