/* MAINTAINER: Martin */

#ifndef _TERRAIN_ROETGER_H
#define _TERRAIN_ROETGER_H

#include "visual_terrain.h"
#include "quadtree.h"
#include "frustum.h"

/**
*	\brief Simple LOD terrain implementation. Level Of Detail - Roetger.
*
*
*	\note This is the specific terrain implementation (where some LOD could be implemented).
*		Implements VisualTerrain::render() and possibly VisualTerrain::update() methods.
*/
class TerrainRoetger: public VisualTerrain
{
public:
	TerrainRoetger();
	~TerrainRoetger();
	
	virtual void render(void);
	virtual void update(const class Camera * camera);
	virtual void init(void);

	void set_minimum_quality(float minimum_global_res);
	void set_desired_quality(float desired_global_res);
	float get_minimum_quality(void) { return m_minimum_global_res; }
	float get_desired_quality(void) { return m_desired_global_res; }
	void calculate_roughness(void);

	/** Recursively builds the quadtree of this terrain.
	*  Works top-down, uses <c>subdivision_criterium(<a>node</a>)</c> to determine whether to
	*  subdivide quadtree node or not. Possibly stops building the tree-branch at node of width 1.
	*  @param node Starting node, root at startup.
	*/
	void build_quadtree(const Quadtree::QuadtreeNode &node);

	/** Recursively builds the quadtree of this terrain.
	*  @see build_quadtree(node)
	*/
	inline void build_quadtree(void)
	{ Quadtree::QuadtreeNode root; m_qt.get_root(root); build_quadtree(root); }

	void render_quadtree(const Quadtree::QuadtreeNode &node);
	inline void render_quadtree(void)
	{ Quadtree::QuadtreeNode root; m_qt.get_root(root); render_quadtree(root); }

	/** Computes subdivision criterium of the given <a>node</a>.
	*  @return True for subdivide, otherwise false.
	*/
	inline bool subdivision_criterium(const Quadtree::QuadtreeNode & node)
	{
		double l = metricsL1(GLVector3d((float)node.x, (float)node.y, 0.0f), m_location);
		return l < (node.width-1) * m_minimum_global_res * (1.0 + (m_desired_global_res - 1.0) * m_roughness[index(node.x, node.y)]);
		// m_desired_global_res has to be >= 1.0
	}
	
	void init_faces_counter(void) {m_faces_rendered = 0;}
	int faces_counter(void) {return m_faces_rendered; }

	void set_location(GLVector3d loc) {m_location = loc; }

protected:
	virtual bool allocateHeightmap(int sizeX, int sizeY);
	
private:
	
	inline double metricsL1(GLVector3d a, GLVector3d b)
	{ return fabs(b.x - a.x) + fabs(b.y - a.y); }
	
	inline int index(int x, int y)
	{ int w, h; getExtents(w, h); return y * w + x; }

	void add_height_error(int x1, int y1, int x2, int y2, float * max_error);
	
	// temporary solution
	void emit_vertices(const GLVector3d &v1, const GLVector3d & v2, const GLVector3d & v3);
	
	Quadtree m_qt;      //!< Quad tree over this terrain, root node size is m_size

	float  m_minimum_global_res;
	float  m_desired_global_res;
	GLVector3d  m_location;
	
	int m_faces_rendered; //!< the number of faces rendered
	float* m_roughness;

	Frustum mFrustum;
};

#endif //_TERRAIN_SIMPLE_H
