/* MAINTAINER: Martin */

#include <string>
#include <algorithm>

#include "visual_terrain.h"
#include "SDL.h"
#include "grutils.h"
#include "texturemanager.h"
#include "platform.h"
#include "conversions.h"

#include "SDL_rotozoom.h"

VisualTerrain::VisualTerrain(void)
{
}

void VisualTerrain::loadLightmap(std::string path)
{
	SurfDefault sf(path);
	mLightmapTexture = TextureManager::getInstance()->add(&sf);
}

void VisualTerrain::loadGlobalTexture(std::string path)
{
	SurfDefault sf(path);
	mGlobalTexture = TextureManager::getInstance()->add(&sf);
}

// layer type shortcut
typedef VisualTerrain::LayerSelection::value_type layer_type;

	// sort laysel by periods, then by image pixel extents
	struct comparer : public binary_function<layer_type, layer_type, bool> {
		bool operator()(first_argument_type x, second_argument_type y) 
		{ 
			counted_ptr< TerrainSurface::SurfaceLayer::Tile > 
				tile1( x->getDetailTile() ), tile2( y->getDetailTile() );

			// criterias:
			// 1. period x
			// 2. period y
			// 3. absolute width
			// 4. absolute height
			bool result =  
				tile1->getPeriodX() < tile2->getPeriodX() || 
				(
					tile1->getPeriodX() == tile2->getPeriodX() && 
					(
						tile1->getPeriodY() < tile2->getPeriodY() || 
						( 
							tile1->getPeriodY() == tile2->getPeriodY() &&
							(
								tile1->getAbsoluteWidth() < tile2->getAbsoluteWidth() ||
								(
									tile1->getAbsoluteWidth() == tile2->getAbsoluteWidth() &&
									tile1->getAbsoluteHeight() < tile2->getAbsoluteHeight()
								)
							)
						)
					)
				);

			return result;
		}
    };



void VisualTerrain::selectLayers(LayerSelection & laysel)
{
	// select only those layers that have weight function and detail tile defined
	laysel.resize(0);
	TerrainSurface::LayersContainer::size_type nlayers, ilayer;
	nlayers = mSurface.getLayersCount();
	laysel.reserve(nlayers);
	for (ilayer = 0; ilayer < nlayers; ++ilayer)
	{
		// layer at index ilayer
		TerrainSurface::LayersContainer::value_type 
			layer = mSurface.getLayer(ilayer);
		
		// detail tile of this layer
		counted_ptr< TerrainSurface::SurfaceLayer::Tile > 
			detail( layer->getDetailTile() );
		
		// weight function of this layer
		TerrainSurface::SurfaceLayer::WeightFunctionPtr 
			weight( layer->getWeightFunction() );

		// need both layers to exist
		if ( detail.get() && weight.get() )
		{
			laysel.push_back( layer );
		}
	}
	
	std::stable_sort( laysel.begin(), laysel.end(), comparer());
}

bool VisualTerrain::createTexturesFromSurfaces(std::vector<SDL_Surface *> & surfaces, std::vector<TextureRef> & textures, const std::string & textureNameHashPrefix)
{
	textures.resize( 0 );
	textures.reserve( surfaces.size() );

	for (std::vector<SDL_Surface *>::size_type isurf = 0, nsurfaces = surfaces.size(); isurf < nsurfaces; ++isurf)
	{
		grutils::SDL_Surface_Accessor sac( surfaces[isurf] );
		
		sac.lock();
	
		SurfDefault sf(sac.get());
		// TODO: store name as class member constant
		sf.pref( textureNameHashPrefix + xtos(isurf) );
		sf.setULC(true);

		textures.push_back( TextureManager::getInstance()->add(&sf, GL_RGBA, Texture::BUILD_MIPMAPS) );
		if (!textures.back().isValid())
			return false;

		sac.unlock();
	}
	return true;
}

void VisualTerrain::freeSurfaces(std::vector<SDL_Surface *> & surfaces)
{
	for ( std::vector<SDL_Surface *>::iterator it = surfaces.begin(), ite = surfaces.end(); it != ite; ++it )
	{
		if ( *it )
		{
			SDL_FreeSurface(*it);
			*it = 0;
		}
	}
	surfaces.clear();
}

void VisualTerrain::createGlobalMapTexture(int extentX, int extentY)
{	
	// A. create an empty SDL surface of desired size and pass it to a surface accessor
	grutils::SDL_Surface_Accessor surf(
		SDL_CreateRGBSurface(
			SDL_SWSURFACE,
			extentX, extentY, 
			32,
			0x000000ff, 
			0x0000ff00, 
			0x00ff0000, 
			0xff000000));
	
	if (!surf.isValid())
	// invalid surface
	{
		LOGD << "Terrain global texture was not created (invalid SDL surface)!" << std::endl;
		return;
	}

	// B. calculate global texture colors
	{
		int numOfLayers = static_cast<int>(mSurface.getLayersCount());

		double dx = 1.0 / (extentX - 1), dy = 1.0 / (extentY - 1);

		typedef TerrainSurface::SurfaceLayer::WeightFunction::value_type weight_value_type;
		typedef TerrainSurface::SurfaceLayer::ColorFunction::value_type color_value_type;
		
		std::vector<weight_value_type> 
			weights(extentX), 
			sumWeights(extentX, 0);

		std::vector<color_value_type> 
			colors(extentX),
			finalColors(extentX, color_value_type(0, 0, 0));
		
		for (double y = 0; y <= 1.0; y += dy)
		{
			for (int ilayer = 0; ilayer < numOfLayers; ++ilayer)
			{
				TerrainSurface::LayersContainer::value_type layer = mSurface.getLayer( ilayer );
				if (layer.get() == 0 || layer->getWeightFunction().get() == 0 || layer->getColorFunction().get() == 0)
					continue;

				// get weights and colors on row y
				layer->getWeightFunction()->sampleValues(0.0, 1.0, y, dx, dy, weights.begin());
				layer->getColorFunction()->sampleValues(0.0, 1.0, y, dx, dy, colors.begin());

				// add weighted color to final color and calculate next part of weight sum
				for (int x = 0; x < extentX; ++x)
				{
					weight_value_type w = weights[x];
					finalColors[x].r += float( colors[x].r * w );
					finalColors[x].g += float( colors[x].g * w );
					finalColors[x].b += float( colors[x].b * w );
					sumWeights[x] += w;
				}
			}

			// normalize final color by weight sum (for cases when weight sum doesn't equal to 1.0)
			int iy = static_cast<int>(y * (extentY - 1) + 0.5);
			for (int ix = 0; ix < extentX; ++ix)
			{
				weight_value_type w = sumWeights[ix];
				if ( w ) {
					finalColors[ix].r /= float(w);
					finalColors[ix].g /= float(w);
					finalColors[ix].b /= float(w);
				}else {
					finalColors[ix].r = 0;
					finalColors[ix].g = 0;
					finalColors[ix].b = 0;
				}
				Uint8 
					r(static_cast<Uint8>(255 * max(0.0, min(1.0, (double)finalColors[ix].r)))),
					g(static_cast<Uint8>(255 * max(0.0, min(1.0, (double)finalColors[ix].g)))),
					b(static_cast<Uint8>(255 * max(0.0, min(1.0, (double)finalColors[ix].b))));

				surf.setColorRGBA(ix, iy, r, g, b, SDL_ALPHA_OPAQUE);
				finalColors[ix] = color_value_type(0, 0, 0);
				sumWeights[ix] = 0;
			}
		}
	}

	// C. create the global texture from surf		
	surf.lock();
	
	SurfDefault sf(surf.get());
	// TODO: store name as class member constant
	sf.pref("terrain_global_texture");
	sf.setULC(true);

	TextureRef texture = TextureManager::getInstance()->add(&sf, GL_RGB, Texture::BUILD_MIPMAPS);

	surf.unlock();
	surf.free();
	
	// set the new global texture
	if ( texture.isValid() )
	{
		this->mGlobalTexture = texture;
		LOGD << "Terrain global texture successfully created." << std::endl;
	} else {
		LOGD << "Terrain global texture was not created (invalid texture)!" << std::endl;
	}
}

void VisualTerrain::createMixTextures(const LayerSelection & laysel, int extentX, int extentY, int maxTexturesCount)
{
	if (extentY < 2 || extentX < 2)
		return;

	// A. Create vector with layers that will go to mix textures. This can exclude some layers
	//	or  change their order. The main aim of doing this is to have layers only with weight and detail function.
	LayerSelection::size_type nlayers = laysel.size(), ilayer, maxlayers, layersdelta;
	
	// calculate maximum layers count from maximum allowed textures count
	maxlayers = (LayerSelection::size_type)max( min( maxTexturesCount * 4, int(nlayers) ), 0); // 4 channels per one texture
	layersdelta = nlayers - maxlayers;

	if ( maxlayers == 0 )
		return;
			
	// B. create empty SDL surfaces of desired size
	const int nsurfaces((int)((maxlayers + 3) >> 2));
	std::vector< SDL_Surface* > surf;
	surf.reserve( nsurfaces );

	for (int isurf = 0; isurf < nsurfaces; ++isurf)
	{
		SDL_Surface * s = 
			SDL_CreateRGBSurface(
				SDL_SWSURFACE,
				extentX, extentY, 
				32,
				0x000000ff, 
				0x0000ff00, 
				0x00ff0000, 
				0xff000000
			);
		
		if (!s)
		// invalid surface
		{
			freeSurfaces( surf );
			// todo: handle error when creating surface
			LOGD << "Terrain mix textures were not been created (invalid SDL surface)!" << std::endl;
			return;
		}
		surf.push_back( s );
	}

	// C. loop through rows - initialization
	typedef TerrainSurface::SurfaceLayer::WeightFunction::value_type weight_value_type;
	typedef TerrainSurface::LayersContainer::value_type layer_value_type;

	// create row-buffer for each layer
	std::vector< std::vector<weight_value_type> > weights(nlayers);
	for (ilayer = 0; ilayer < nlayers; ++ilayer)
		weights[ilayer].resize(extentX);
	
	// row-buffer for weight sum in one row
	std::vector<weight_value_type> weightSum(extentX, 0), weightResiduum(extentX, 0);

	double dx(1.0 / (extentX - 1)), dy(1.0 / (extentY - 1)), yf(0.0);
	
	// D. loop through rows - loop
	for ( int y = 0; y < extentY; ++y, yf += dy )
	{
		// initialize weightSum to zeros
		weightSum.resize(0);
		weightSum.resize(extentX, 0);

		// D.1 loop through layers in row y and calculate the sum of weights in this row.
		for ( ilayer = 0; ilayer < nlayers; ++ilayer )
		{
			// sample weights
			layer_value_type layer( laysel[ilayer] );
			layer->getWeightFunction()->sampleValues(0.0, 1.0, yf, dx, dy, weights[ilayer].begin());
			
			// calculate the sum
			for (int x = 0; x < extentX; ++x)
				weightSum[x] += weights[ilayer][x];
		}

		// calculate the weight residuum
		if ( layersdelta )
		{
			// clear residuum
			weightResiduum.resize(0);
			weightResiduum.resize(extentX, 0);
			
			// calculate residual sum for layers maxlayers to nlayers - 1
			for ( ilayer = maxlayers; ilayer < nlayers; ++ilayer )
				for (int x = 0; x < extentX; ++x)
					weightResiduum[x] += weights[ilayer][x];
			
			// spread to maxlayers
			weight_value_type f( weight_value_type(1) / maxlayers );
			for (int x = 0; x < extentX; ++x)
				weightResiduum[x] *= f;
		}
		
		// D.2 loop again and normalize weights (divide them by the sum) in the row. Convert these
		//	weights to [0,255] and output to apropriate channel of apropriate surface.
		//	At this point, maxTexturesCount is reflected by adding the residuum of weights from layers 
		//	that will not go to textures, uniformly, to weights of layers that will go to textures.
		for ( ilayer = 0; ilayer < maxlayers; ilayer += 4 )
		{
			int surfIndex( (int)(ilayer >> 2) );
			grutils::SDL_Surface_Accessor sac( surf[ surfIndex ] );
			for (int x = 0; x < extentX; ++x)
			{
				
				weight_value_type residuum( layersdelta ? weightResiduum[x] : 0 );

				Uint8 ch[4] = {255, 255, 255, 255 };
				
				// normalization factor - normalizes weights to [0, 255]
				weight_value_type f( (weight_value_type)((weightSum[x] ? 1.0 / weightSum[x] : 0) * 255) );

				// fill in channel buffer ch, add residuum
				// TODO: aby soucet ch[i] dal 1. Jedno zaokrouhleni na vsechny slozky - obecne nestaci. Chceme napriklad
				// vysledek 127 + 128 a ne 127+127 nebo 128 + 128.
				for (int j = 0; j < 4; ++j)
				{
					ch[j] = 
						((ilayer + j) < nlayers) 
						? grutils::clamp<int>(0, 255, static_cast<int>( (weights[ ilayer + j ][x] + residuum) * f + 0.1)) 
						: 0;
				}

				// output channel buffer to surface
				sac.setColorRGBA(x, y, ((int)ch[0] + 255)/2, ((int)ch[1] + 255)/2, ((int)ch[2] + 255)/2, ch[3]);
			}
		}
	}

	// E. create textures from surfaces
	std::vector<TextureRef> texrefs;
	bool texturesCreated = createTexturesFromSurfaces( surf, texrefs, "terrain_mixmap_texture_");
	freeSurfaces( surf );

	if ( !texturesCreated )
	{
		LOGD << "Terrain mixmap textures were not created (invalid texture)!" << std::endl;
		return;
	}

	this->mMixMapTextures.assign(texrefs.begin(), texrefs.end());
	LOGD << "Terrain mixmap textures were successfully created." << std::endl;
}

bool VisualTerrain::createDetailTextures(const LayerSelection &laysel, int maxTexturesCount)
{
	LayerSelection::size_type nlayers = laysel.size(), ilayer, maxlayers;
	std::vector<GLVector2d> periods;
	
	// calculate maximum layers count from maximum allowed textures count
	maxlayers = (LayerSelection::size_type)max( min( maxTexturesCount * 4, int(nlayers) ), 0); // 4 channels per one texture
	nlayers = maxlayers;

	if ( maxlayers == 0 )
		return false;
			
	// B. create empty SDL surfaces of desired size
	const int nsurfaces((int)((nlayers + 3) >> 2));
	std::vector< SDL_Surface* > surf;
	surf.reserve( nsurfaces );
	periods.reserve(nsurfaces);

	typedef TerrainSurface::SurfaceLayer::WeightFunction::value_type weight_value_type;
	typedef TerrainSurface::LayersContainer::value_type layer_value_type;	
	typedef TerrainSurface::SurfaceLayer::Tile Tile;

	// C. Output each tile to a separate channel
	for ( ilayer = 0; ilayer < nlayers; ilayer += 4 )
	{
		counted_ptr<Tile> tile[4] = 
		{
			(ilayer + 0 < nlayers && laysel[ilayer + 0].get()) ? laysel[ ilayer + 0]->getDetailTile() : counted_ptr<Tile>(0),
			(ilayer + 1 < nlayers && laysel[ilayer + 1].get()) ? laysel[ ilayer + 1]->getDetailTile() : counted_ptr<Tile>(0),
			(ilayer + 2 < nlayers && laysel[ilayer + 2].get()) ? laysel[ ilayer + 2]->getDetailTile() : counted_ptr<Tile>(0),
			(ilayer + 3 < nlayers && laysel[ilayer + 3].get()) ? laysel[ ilayer + 3]->getDetailTile() : counted_ptr<Tile>(0)
		};

		// check tiles for equal periods and calculate their common extent
		bool equalPeriods(true);
		int requestedWidth(0), requestedHeight(0);
		double periodX(0), periodY(0);
		for (int i = 0; i < 4; ++i) 
		{
			if (tile[i].get()) 
			{
				requestedWidth = max(tile[i]->getAbsoluteWidth(), requestedWidth);
				requestedHeight = max(tile[i]->getAbsoluteHeight(), requestedHeight);

				periodX = tile[i]->getPeriodX();
				periodY = tile[i]->getPeriodY();

				for (int j = i + 1; j < 4; ++j) 
				{
					counted_ptr<Tile> p(tile[j]);
					if (p.get() ) 
					{
						if ( p->getPeriodX() != periodX || p->getPeriodY() != periodY )
						{
							equalPeriods = false;
							break;
						}
						requestedWidth = max(p->getAbsoluteWidth(), requestedWidth);
						requestedHeight = max(p->getAbsoluteHeight(), requestedHeight);
					}
				}
				break;
			}
		}
		
		if (!equalPeriods)
		{
			LOGE << "Wrong terrain tileset - try to make periods of all tiles equal." << std::endl;
			freeSurfaces(surf);
			return false;
		}

		//int surfIndex( (int)(ilayer >> 2) );
		//assertL(surfIndex < nsurfaces && "Surfaces index out of bounds!");

		grutils::SDL_Surface_Accessor dst(
			SDL_CreateRGBSurface(
				SDL_SWSURFACE,
				requestedWidth, requestedHeight, 
				32,
				0x000000ff, 
				0x0000ff00, 
				0x00ff0000, 
				0xff000000
			)
		);
		
		if ( !dst.isValid() )
		// invalid surface
		{
			freeSurfaces(surf);
			LOGD << "Terrain detail textures were not been created (invalid SDL surface)!" << std::endl;
			return false;
		}
		surf.push_back( dst.get() );
		
		periods.push_back(GLVector2d(periodX, periodY));

		grutils::SDL_Surface_Accessor tileSAC[4] =
		{
			grutils::SDL_Surface_Accessor( (tile[0].get()) ? tile[0]->getSurface() : 0 ),
			grutils::SDL_Surface_Accessor( (tile[1].get()) ? tile[1]->getSurface() : 0 ), 
			grutils::SDL_Surface_Accessor( (tile[2].get()) ? tile[2]->getSurface() : 0 ),
			grutils::SDL_Surface_Accessor( (tile[3].get()) ? tile[3]->getSurface() : 0 )
		};

		vector<SDL_Surface *> scaled;
		scaled.reserve(4);

		// rescale small tiles if necessary
		for (int i = 0; i < 4; ++i) 
		{
			grutils::SDL_Surface_Accessor & sa = tileSAC[i];
			if ( sa.isValid() ) 
			{
				if (sa.getWidth() != requestedWidth || sa.getHeight() != requestedHeight)
				{
					double 
						zoomx( sa.getWidth() ? double( requestedWidth ) / sa.getWidth() : 0 ), 
						zoomy( sa.getHeight() ? double( requestedHeight ) / sa.getHeight() : 0 );
					SDL_Surface * scaledSurf = zoomSurface(sa.get(), zoomx, zoomy, 0);
					if ( !scaledSurf )
					{
						LOGD << "Can not scale surface to size " << requestedWidth << "x" << requestedHeight << "!" << std::endl;

						freeSurfaces( surf );
						freeSurfaces( scaled );
						return false;
					}
					scaled.push_back( scaledSurf );
					sa.setSurface( scaledSurf );
				}
			}
		}

		// lock surfaces
		for (int i = 0; i < 4; ++i) 
			if (tileSAC[i].isValid())
				tileSAC[i].lock();

		// Every pixel of dst surface consists of 4 channels (rgba). Channel number i will be
		// calculated from surface of tileSAC[i].
		dst.lock();
		for ( int y = 0; y < dst.getHeight(); ++y ) 
		{
			for ( int x = 0; x < dst.getWidth(); ++x )
			{
				Uint8 ch[4] = {255,255,255,255}, r, g, b, a;

				for (int channelNum = 0; channelNum < 4; ++channelNum)
				{
					grutils::SDL_Surface_Accessor & sac = tileSAC[ channelNum ];

					if (!sac.isValid())
						continue;
					sac.getColorRGBA( x, y, r, g, b, a );
					ch[ channelNum ] = grutils::calculateIntensity( r, g, b );
				}
				dst.setColorRGBA( x, y, ((int)ch[0] + 255)/2, ((int)ch[1] + 255)/2, ((int)ch[2] + 255)/2, ch[3] );
			}
		}
		dst.unlock();
		
		for (int i = 0; i < 4; ++i)
			if (tileSAC[i].isValid())
				tileSAC[i].unlock();
		
		freeSurfaces( scaled );

	}

	// E. create textures from surfaces
	std::vector<TextureRef> texrefs;
	
	bool texturesCreated = createTexturesFromSurfaces( surf, texrefs, "terrain_detail_texture_");
	freeSurfaces( surf );

	if ( !texturesCreated )
	{
		LOGD << "Terrain detail textures were not created (invalid texture)!" << std::endl;
		return false;
	}

	// assign new set of detail textures
	this->mDetailTextures.assign(texrefs.begin(), texrefs.end());	
	
	// assign periods
	this->mDetailPeriods.assign(periods.begin(), periods.end());

	LOGD << "Terrain detail textures were successfully created." << std::endl;

	return true;
}

/*
TerrainRenderer::TerrainRenderer()
{
	setTerrain( 0 );
}

TerrainRenderer::TerrainRenderer(VisualTerrain * vt)
{
	setTerrain( vt );
}

TerrainRenderer::~TerrainRenderer()
{
}

void TerrainRenderer::render()
{
	VisualTerrain * t(getTerrain());
	if ( t == 0 )
		return;

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();

	setScale();

	glColor3f( 1, 1, 1 );
	t->render();

	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
}

void TerrainRenderer::setScale()
{
	VisualTerrain *tmp = getTerrain();
	assertL(tmp && "Accessing null terrain pointer!");
	tmp->getRealScale().glScale();
}

TextureRef TerrainRenderer::getLightmapTexture()
{
	VisualTerrain *tmp = getTerrain();
	assertL(tmp && "Accessing null terrain pointer!");
	return tmp->mLightmapTexture;
}

TextureRef TerrainRenderer::getGlobalTexture()
{
	VisualTerrain *tmp = getTerrain();
	assertL(tmp && "Accessing null terrain pointer!");
	return tmp->mGlobalTexture;
}

void TerrainRenderer::getMixMapTextures(std::vector<TextureRef> & values)
{
	VisualTerrain *tmp = getTerrain();
	assertL(tmp && "Accessing null terrain pointer!");
	values.assign(tmp->mMixMapTextures.begin(), tmp->mMixMapTextures.end());
}

void TerrainRenderer::getDetailMapTextures(std::vector<TextureRef> & values)
{
	VisualTerrain *tmp = getTerrain();
	assertL(tmp && "Accessing null terrain pointer!");
	values.assign(tmp->mDetailTextures.begin(), tmp->mDetailTextures.end());
}

void TerrainRenderer::getDetailMapPeriods(std::vector<GLVector2d> & values)
{
	VisualTerrain *tmp = getTerrain();
	assertL(tmp && "Accessing null terrain pointer!");
	values.assign(tmp->mDetailPeriods.begin(), tmp->mDetailPeriods.end());
}
*/
