/* MAINTAINER: Martin */

#ifndef _VISUAL_TERRAIN_H_
#define _VISUAL_TERRAIN_H_

#include "terrain_base.h"
#include "texturemanager.h"
#include "GLVector.h"

/** \addtogroup game_data */ /*@{ */

/**
*	\brief specific terrain behavior on client
*
*	This is an ABSTRACT class! Rendering (and possibly updating as well) should be implemented in
*	some derived class.
*	All terrain textures are stored on this level of terrain hierarchy.
*	
*/

class VisualTerrain : public TerrainBase
{
	// TODO: remove this friend!
	friend class TaskRender;
	friend class TerrainRenderer;

public:
	VisualTerrain(void);

	/**
	*	\brief Performs rendering. [CLIENT]
	*
	*	Possibly "only" sends data to OpenGL API.
	*/
	virtual void render(void) = 0;

	/**
	*	\brief Data update (LOD-related things). [CLIENT]
	*
	*	This is usually a *very* time consuming task. However, it should be 
	*	performed once a frame, because it takes care about LOD.
	*	In good implementation, it should be time-constrained, for example by
	*	the maximal number of used triangles.
	*/
	virtual void update(const class Camera * camera){};

	/**
	*	\brief preprocessing step - offline calculations for LOD
	*/
	virtual void init(void){}

	/**
	*	\brief loads lightmap texture
	*
	*	Initializes lightmap texture using TextureManager.
	*
	*	\param path texture path
	*/
	void loadLightmap(std::string path);

	/**
	*	\brief loads global texture
	*
	*	Initializes global texture using TextureManager.
	*
	*	\param path texture path
	*/
	void loadGlobalTexture(std::string path);

private:
	bool createTexturesFromSurfaces(std::vector<SDL_Surface *> & surfaces, std::vector<TextureRef> & textures, const std::string & textureNameHashPrefix);
	void freeSurfaces(std::vector<SDL_Surface *> & surfaces);
	
public:
	
	typedef std::vector< LayerPtr > LayerSelection;

	/**
	*	\brief select layers for building mixmap and detail textures
	*
	*	Only layers with detail tile and weight function are selected and then sorted by criterias of
	*	detail tile:
	*	-# period in X axis
	*	-# period in Y axis
	*	-# absolute width in pixels
	*	-# absolute height in pixels
	*/
	void selectLayers(LayerSelection & laysel);

	void createMixTextures(const LayerSelection &laysel, int extentX, int extentY, int maxTexturesCount);
	bool createDetailTextures(const LayerSelection &laysel, int maxTexturesCount);

	/**
	*	\param extentX global texture extent in X axis
	*	\param extentY global texture extent in Y axis
	*/
	void createGlobalMapTexture(int extentX, int extentY);	// using mix map and average tile colors

protected:

	TextureRef mGlobalTexture;		///< \brief Global, unique texture of the terrain.
	TextureRef mLightmapTexture;	///< \brief Lightmap texture of the terrain.
	std::vector<TextureRef> 
		mMixMapTextures;			///< \brief Mixmap, mix weights are stored in rgba channels.
	std::vector<TextureRef> 
		mDetailTextures;			///< \brief Detail tiles of equal size and repetition stored in rgba channels.
	std::vector<GLVector2d> 
		mDetailPeriods;
};

/*@}*/

#endif //_VISUAL_TERRAIN_H_

