/* MAINTAINER: Bernard */

#ifndef _GAME_GUI_H
#define _GAME_GUI_H

#include <string>
#include <vector>
#include "task.h"
#include "game.h"
#include "layer.h"

// forward declarations to save includes and speed up the compilation
class Player;
class StatsWindow;
class InfoLayer;
class Messages;
class MenuInGame;

/**
 * \brief Class for holding GUI when playing.
 * 
 * The class itself doesn't render anything but only
 * provide container for other inner classes and try
 * to make gui look like one big class from engine point of view.
 * 
 * The class directly act as Task so it may receive init/clean/update messages
 * as well as Layer and may be placed into TaskRender.
 */
class GameGUI : public Task, public Layer
{
	public:
		GameGUI();

		/**
		 * \brief Delete all allocated variables.
		 */
		virtual ~GameGUI();

		/**
		 * \brief Does basic inicialization. 
		 * 
		 * All inner pointer variables are allocated it they
		 * weren't allocated before and init is called on them.
		 * The init is called once when connected to the server.
		 */
		virtual void init();
		
		/**
		 * \brief Call clean on all variables and delete mMenuInGame.
		 * 
		 * The clean() is called once when disconnecting the server.
		 */
		virtual void clean();
		
		/**
		 * \brief Render the whole GUI.
		 * 
		 * The GUI is rendered by calling render to all
		 * GUI parts.
		 */
		virtual void render();

		/**
		 * \brief Pass update to parts that require it.
		 */
		virtual void update(int dtime, int atime);
		
		/**
		 * \brief Callback from engine that game will start soon.
		 * 
		 * The callback is done by engine after every load of map.
		 * mMenuInGame is allocated if it wasn't before.
		 */
		void gameStarted();
		
		/**
		 * \brief Callback from engine that game ended.
		 * 
		 * It doesn't imply disconnect from server. It may be caused
		 * also by reloading the map. mMenuInGame if freed here.
		 */		
		void gameEnded();
		
		/**
		 * Distribute current player to other GUI components
		 */
		void setPlayer(Player &player);
		
		/**
		 * \brief Add a message to messages window.
		 * 
		 * \param msg Message to add.
		 */
		void addMessage(const std::string &msg);

		/**
		 * \brief Callback from server that player picked some bonus.
		 * 
		 * The bonus will appear as an icon on the screen if the bonus
		 * is with some time limit. Forwarded to mInfoLayer.
		 * \param id ID of picked bonus from params.
		 */
		void pickedBonus(int id);

		/**
		 * \brief Callback from server that some bonus expired.
		 * 
		 * One icon of bonus of this type will disapear from the screen.
		 * \param id ID of picked bonus from params.
		 */
		void expiredBonus(int id);

		/**
		 * \brief Make chat window active.
		 * 
		 * The chat window deactivate automaticly when chat
		 * message is sended.
		 */
		void activateChatWindow();

		/**
		 * \brief Show or hide stats window.
		 * \param visibility State to which we want to chage it.
		 */
		void showStatsWindow(bool visibility);


		/**
		 * \brief Makes mMenuInGame active.
		 */
		void activateMenu();
		
		/**
		 * \brief Deactivate mMenuInGame.
		 */
		void deactivateMenu();
		
		/**
		 * \brief Check if mMenuInGameis active.
		 * 
		 * \return True if mMenuInGameis is active.
		 */
		bool isMenuActive();
		
		/**
		 * \brief Current state of game conditions.
		 * 
		 * The conditions are forwarded to mInfoLayer.
		 * \param conditions The current state.
		 */
		void setEndConditionStatuses( const std::list<ConditionStatus> &conditions );
			
	private:
		
		// our components
		StatsWindow  *mStatsWindow;  ///< \brief Window wit list of all players and their score
		InfoLayer    *mInfoLayer;    ///< \brief Targeting, icons and info messages.
		Messages     *mMessages;     ///< \brief Chat messages and other server messages.
		MenuInGame   *mMenuInGame;   ///< \brief Small ingame menu.

		bool          mInicialized;  ///< \brief Check if class was inicialized.
};


#endif // _GAME_GUI_H
