/* MAINTAINER: Bernard */

#ifndef _MESSAGES_H
#define _MESSAGES_H

#include <list>
#include <string>

#include "inputacceptor.h"
#include "inputstring.h"
#include "font_manager.h"
#include "timemanager.h"

/**
 * \brief Class that holds messages in game.
 * 
 * The messages come from server. They may be either chat messages
 * or some other info messages like who kill who or whatever.
 * The message from server is delivered using command_chatmessage.
 * The messages are put into queue with limited size and the messages also
 * automaticly disapear if they are shown for too long.
 * The Messages also enable user to input string for a chat message.
 * Variable mInputLine is holdind the line with message and if actvivated
 * the input is forwarded to it.
 * 
 * Chat notes. Chat works such that client write something to mInputLine, the
 * input is taken and sended to the server and server distribute the message
 * to every client. On every client the message goes back to Messages.
 * Commands from Commander is used for transfering messages.
 */
class Messages : public InputAcceptor
{
public:	
	Messages();
	~Messages();

	/**
	 * \brief Add one message to messages.
	 * 
	 * If the queue is allreadu full last message is removed.
	 * \param msg A message to add.
	 */
	void addMessage(const std::string &msg);

	/**
	 * \brief Render messages and currently edited line.
	 * 
	 * The currently edited line of chat message is rendered
	 * only if activated. All messages are rendered allways.
	 */
	void render();
	
	/**
	 * \brief Distinguish between activated state and then want all and
	 *        not active state and then it wants nothing.
	 * 
	 * \param keysym Ignored.
	 * \return True if console is active.
	 */
	virtual bool wantKey(const SDL_keysym &keysym); // from InputAcceptor
	
	/**
	 * \brief Adding key to editline with current chat line.
	 * 
	 * If the chat line is active the key is passed to the chat line. 
	 * Else the key is ignored and false is returned.
	 * Return serves as a button for accepting the input and sending
	 * '/chat' command to the server. The chat line will be automaticly
	 * disabled then and player will continue playing.
	 */
	virtual bool keyDown(const SDL_keysym &keysym); // from InputAcceptor


	/**
	 * \brief Inicialize the messages window and plug it into the system.
	 *
	 * Menas loading font, textures and adding to InputManager.
	 */
	void init();
	
	/**
	 * \brief Clean resources allocated by Messages window
	 *
	 */
	void clean();


	/**
	 * \brief change active state of messages window
	 */
	void activate(bool state);

private:

	/**
	 * \brief Structure containing one message.
	 * 
	 * When the message is created it automaticly read current time
	 * so it is easy to chceck when the message should be removed because
	 * of time.
	 */
	struct Message
	{
		Message(std::string message)
		{
			msg = message;
			timeStamp = gTime.getTicks();
		}

		std::string msg;        ///< \brief Message string.
		int         timeStamp;  ///< \brief Time when the message was created.
	};
	typedef std::list<Message> MessageList;

	MessageList mMessages;  ///< \brief List of messages.

	InputString mInputLine; ///< \brief Currently edited command.
	
	FontRef mFont; ///< \brief Font for messages.
	
	bool mActive;  ///< \brief True if chat message editing is active.
};

#endif // _MESSAGES_H
