/* MAINTAINER: Martin */

#ifndef _HIERARCHY_TREE
#define _HIERARCHY_TREE

#include "log.h"
#include "frustum.h"

/**
*	\brief Base class for HierarchyTree	(KD tree, K = HierarchyTree::D = 2)
*
*	Defines dimension od node-box, Node base-class, extent of node-box of root node.
*/
class HierarchyTreeBase
{
public:
	/**
	*	\brief dimension of node's bounding box 
	*	\note don't be confused with user-defined floating point bounding box
	*/
	enum { D = 2 };
	
	/**
	*	\brief node of hierarchy tree
	*	
	*	Contains information used for tree (inner bounding box, leaf-flag) as well as user's defined 
	*	information (user's defined 3D bounding box, frustum culling state).
	*/
	class Node
	{
	friend class HierarchyTreeBase;
	public:

		/**
		*	\brief determines if this node is a leaf or not
		*	\return true on leaf-node
		*/
		bool isLeaf(void) { return mLeaf; }

		const int* getNodeBoxMin(void) const
		{ return mMinPoint; }

		const int* getNodeBoxMax(void) const
		{ return mMaxPoint; }

	private:

		/**
		*	\brief sets bounding box of this node
		*	\param boxMin minimal point of the box
		*	\param boxMax maximal point of the box
		*	\note used only in tree-building by HierarchyTree class
		*/
		void setNodeBox(const int boxMin[D], const int boxMax[D])
		{ 
			for (int i = 0; i < D; ++i) {
				assertL(boxMin[i] <= boxMax[i] && "Invariant is broken: minimal point has to be <= maximal point!");
				mMinPoint[i] = boxMin[i]; 
				mMaxPoint[i] = boxMax[i];
			}
		}

		/**
		*	\brief sets the leaf property of this node
		*	\param value true for leaf, false for inner node
		*/
		void setLeaf(bool value) 
		{ mLeaf = value; }

		bool mLeaf; ///< \brief true for leaf node, false for subdivided node
		int 
			mMinPoint[D], ///< \brief minimal node point
			mMaxPoint[D]; ///< \brief maximal node point
	};

	HierarchyTreeBase() 
	{	// TODO: ensure this works
		const int zero[D] = {0, }; setExtent(zero, zero); 
	}
	
	~HierarchyTreeBase(){}
	
	/**
	*	\brief sets the root's min-/max-point
	*	\param minPoint minimal point of root's node-bounding-box
	*	\param maxPoint maximal point of root's node-bounding-box
	*/
	void setExtent(const int minPoint[D], const int maxPoint[D]) { 
		for (int i = 0; i < D; ++i) {
			assertL(minPoint[i] <= maxPoint[i] && "Invariant is broken: minimal point has to be <= maximal point!");
			mRootMin[i] = minPoint[i];
			mRootMax[i] = maxPoint[i];
		}
	}
	
protected:
	/**
	*	\brief sets the node-bounding-box of given \c node
	*	\param node which node
	*	\param boxMin minimal point of the box
	*	\param boxMax maximal point of the box
	*/
	void setNodeBox(Node & node, const int boxMin[D], const int boxMax[D])
	{ node.setNodeBox(boxMin, boxMax); }

	/**
	*	\brief sets the leaf property of given \c node
	*	\param node the node which leaf-flag we want to set
	*	\param value true for leaf, false for inner node
	*/
	void setNodeLeaf(Node & node, bool value) 
	{ node.setLeaf(value); }

	int	
		mRootMin[D],	///< \brief bounding box's min point of root-node
		mRootMax[D];	///< \brief bounding box's max point of root-node
};

/*
*	\brief Implementation of KDTree (k = 2) that defines hierarchy over terrain blocks.
*
*	As a template parameter is class NodeClass that has to be a subclass of 
*	HierarchyTreeBase::Node, which defines common properties of tree node. 
*	HierarchyTree class defines NodeNavigator class for easy tree access and 
*	NodeDelegate class that will be used for filling the properties of NodeClass when
*	building the tree.
*/
template <class NodeClass>
class HierarchyTree: public HierarchyTreeBase {
public:
	enum { ChildrenCount = 2 }; ///< \brief the count of children of each node (2 as default)

	/**
	*	\brief provides easy access to the tree hierarchy
	*/
	class NodeNavigator
	{
	friend class HierarchyTree<NodeClass>;
	public:
		
		/**
		*	\brief constructs empty navigator
		*/
		NodeNavigator(void):
		  mTree(0), mNodeLevel(0), mNodeIndex(0)
		{}
		
		/**
		*	\brief copy ctor
		*/
		NodeNavigator(const NodeNavigator & navig)
			:mTree(navig.mTree), mNodeLevel(navig.mNodeLevel), mNodeIndex(navig.mNodeIndex)
		{}

		/**
		*	\brief reference access operator for node-data access
		*/
		NodeClass & operator *() const
		{  
			return getTree()->getNode(mNodeLevel, mNodeIndex); 
		}

		/**
		*	\brief member access operator for node-data access
		*/
		NodeClass * operator ->() const
		{ return &(this->operator *()); }
		
		/*operator NodeClass*(void) const
		{ return &(this->operator *()); }*/

		/**
		*	\brief children access
		*	\param child_id number of the child, zero based
		*	\param child node navigator that will be filled with data
		*	\return reference to \c child that was provided as argument
		*/
		NodeNavigator& getChild(const int child_id, NodeNavigator &child) const
		{
			child.mTree = getTree();
			child.mNodeLevel = mNodeLevel + 1;
			child.mNodeIndex = mNodeIndex * HierarchyTree<NodeClass>::ChildrenCount + child_id;
			assertL(child.isValid() && "Got invalid child from node!");
			return child;
		}

		/**
		*	\brief checks for validity of encapsulated node
		*	\return true, if node exists in the tree, false otherwise
		*	\note checks only if there is memory allocated in the tree for this node
		*/
		bool isValid(void) const 
		{ return getTree()->isNodeValid(mNodeLevel, mNodeIndex); }

	private:
		/**
		*	\brief access the tree referenced by this navigator
		*	\return tree pointer
		*/
		HierarchyTree * getTree() const {
			assertL(mTree && "Using zero pointer in NodeNavigator class!");
			return mTree;
		}
		
		/**
		*	\brief creates a navigator with explicit tree and node position
		*	\param t tree pointer
		*	\param level node's level, zero based (root is on level 0)
		*	\param index node's index (root has index 0, its left child 0, its right child 1 )
		*/
		NodeNavigator(HierarchyTree * t, int level, int index)
			: mTree(t), mNodeLevel(level), mNodeIndex(index) 
		{}

		HierarchyTree * mTree;	///< \brief pointer to tree
		int 
			mNodeLevel,			///< \brief position of node in tree (depth)
			mNodeIndex;			///< \brief position of node in tree (left offset) 
	};

	/**
	*	\brief interface for NodeClass definition at building phase
	*	\note used when building the tree structure
	*/
	class NodeDelegate
	{
	public:
		NodeDelegate(){}
		virtual ~NodeDelegate(){}
		
		/**
		*	\brief NodeClass definition at building phase
		*	\param navig current node being constructed (children are already finished)
		*/
		virtual void nodeCreation(NodeNavigator &navig) = 0;
	};

	HierarchyTree(): mNodeDelegate(0) {}
	~HierarchyTree();

	/**
	*	\brief access to root
	*	\return node navigator initialized to root-node position
	*/
	NodeNavigator getRoot(void) { return NodeNavigator(this, 0, 0); }
	
	/**
	*	\brief common node access
	*	\param level node's depth in tree
	*	\param index node's offset in tree
	*	\return user's node class reference
	*/
	NodeClass& getNode(const int level, const int index) { return (*mLevels[level])[index]; }
	
	/**
	*	\brief checks node validity (wheter it is out-of-tree-bounds or not)
	*	\return true if valid
	*/
	bool isNodeValid(const int level, const int index) 
	{ return level >= 0 && level < (int)mLevels.size() && index >= 0 && index < (int)mLevels[level]->size(); }
	
	/**
	*	\brief constructs full tree based on previous definition of extent of root node
	*	\note extent of the root node is defined by HierarchyTreeBase::SetExtent()
	*/
	void build(void)
	{
		build(mRootMin, mRootMax, 0, 0);
	}

	/**
	*	\brief sets the node delegate
	*	\param del pointer to delegate
	*	\note provides the ability of building user data in hierachy of this tree
	*/
	void setNodeDelegate(NodeDelegate * del) { mNodeDelegate = del; }
	
	// TODO: consider this...	
	void render(void);
	void render(const NodeNavigator &navigator);
	
private:
	typedef std::vector< NodeClass > TOneLevel;	///< \brief container for nodes - one tree level
	typedef std::vector< TOneLevel * > TLevels;	///< \brief container for tree levels
	
	/**
	*	\brief allocates next level
	*
	*	The size is ChildrenCount-times greater than the size of previous level.
	*/
	void addLevel(void);
	
	/**
	*	\brief constructs full tree based on extent definition
	*
	*	Recursively constructs the tree.
	*
	*	\param minPoint defines extent
	*	\param maxPoint defines extent
	*	\param level current node - level
	*	\param index current node - offset
	*	\note extent of the root node is defined by HierarchyTreeBase::SetExtent()
	*/
	void build(
		const int minPoint[D], const int maxPoint[D], 
		typename TLevels::size_type level, typename TOneLevel::size_type index);
	
	NodeDelegate * mNodeDelegate;	///< \brief delegate pointer
	TLevels	mLevels;				///< \brief nodes
};

template <class NodeClass>
HierarchyTree<NodeClass>::~HierarchyTree()
{
	typename TLevels::iterator it, ite;
	for (it = mLevels.begin(), ite = mLevels.end(); it != ite; ++it) {
		delete *it;
		*it = 0;
	}
	mLevels.clear();
}

template <class NodeClass>
void HierarchyTree<NodeClass>::addLevel(void)
{
	if (mLevels.size() != 0) {
		typename TLevels::iterator lev, prelev;
		TOneLevel * one_level = new TOneLevel();
		lev = mLevels.insert(mLevels.end(), one_level);
		prelev = lev;
		(*lev)->resize((*--prelev)->size() * ChildrenCount );
	} else {
		TOneLevel * one_level = new TOneLevel(1);
		mLevels.push_back(one_level);
	}
}

template <class NodeClass>
void HierarchyTree<NodeClass>::build(
						  const int minPoint[HierarchyTreeBase::D], 
						  const int maxPoint[HierarchyTreeBase::D], 
						  typename HierarchyTree<NodeClass>::TLevels::size_type level, 
						  typename HierarchyTree<NodeClass>::TOneLevel::size_type index)
{
	// while we are out-of-levels, build new level (this should iterate only once)
	while (level >= mLevels.size())
		addLevel();

	assertL(index < (*mLevels[level]).size() && "Node index out of bounds when building the hierarchy tree!");
	
	int d[D], j = 0;
	
	// Compute box dimension and index j of dimension with maximal box's edge size.
	for (int i = 0; i < D; ++i)
	{
		d[i] = maxPoint[i] - minPoint[i];
		if (d[i] > d[j])
			j = i;
	}
	int delta = d[j];

	assertL(delta >= 1 && "Node extent was less than 1 when building the hierarchy tree!");
	
	// Stop if maximal edge size equals 1.
	bool leaf_flag = (delta == 1);
	
	NodeClass & node = (*mLevels[level])[index];
	setNodeLeaf(node, leaf_flag);
	
	if (!leaf_flag) {

		// Otherwise, divide and build recursively.
		assertL(ChildrenCount == 2 && "Ehm, only trees with children count = 2 are supported, sorry...");

		// left subtree (from minPoint to half of the box (divided in j-th axis))
		for (int i = 0; i < D; ++i)
			d[i] = maxPoint[i];	
		d[j] -= delta / 2 + (delta & 1); // subtracting upper part of delta/2
		build(minPoint, d, level + 1, ChildrenCount * index);

		// right subtree (from half of the box (divided in j-th axis) to maxPoint)
		for (int i = 0; i < D; ++i)
			d[i] = minPoint[i];	
		d[j] += delta / 2; // adding lower part of delta/2
		build(d, maxPoint, level + 1, ChildrenCount * index + 1);
	}
	
	// Fill the bounding box information in to the Node data structure.
	setNodeBox(node, minPoint, maxPoint);
	NodeNavigator navig(this, (int)level, (int)index);
	mNodeDelegate->nodeCreation(navig);
}

template <class NodeClass>
void HierarchyTree<NodeClass>::render(void)
{
	glColor3f(1, 0, 0);
	GLint a[2];
	glGetIntegerv(GL_POLYGON_MODE, a);
	glPolygonMode(GL_FRONT, GL_LINE);
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_DEPTH_TEST);
	
	render(getRoot());

	glEnable(GL_TEXTURE_2D);
	glEnable(GL_DEPTH_TEST);
	glPolygonMode(GL_FRONT, a[0]);
}

#include "GLMatrix.h"
#include "grutils.h"
using namespace grutils;

template <class NodeClass>
void HierarchyTree<NodeClass>::render(const NodeNavigator & navigator)
{
	bool bRender = true;
	switch(navigator->getFrustumClipState()) {
		case Frustum::INSIDE:
			glColor3f(1, 0, 0);
			break;
		case Frustum::OUTSIDE:
			glColor3f(0, 1, 0);
			break;
		default:
			if (navigator->isLeaf()){
				glColor3f(1, 1, 0);
				break;
			}
			bRender = false;
			break;
	}
	if (bRender) {
		GLMatrix4d mat;
		float center[3], extent[3];
		navigator->getBoundingBox(center, extent);
		mat.glGet(GL_MODELVIEW_MATRIX);
		glTranslatef(center[0], center[1], center[2]);
		glScalef(extent[0], extent[1], extent[2]);
		glDrawCube();
		mat.glLoadMatrix();
	} else {
		NodeNavigator child;
		for (int i = 0; i < ChildrenCount; ++i)
			render(navigator.getChild(i, child));
	}
}



#endif
