/* MAINTAINER: Bernard */

#ifndef _INPUT_MANAGER_H
#define _INPUT_MANAGER_H


//#include <vector>
#include <map>
#include <queue>
#include "SDL.h"

#include "mutex.h"

class InputAcceptor;

/**
 * InputManager is class for handling all input. It reads input from
 * SDL and find acceptor who wants the input. There may be more acceptors
 * for same input and then priority of accepter are used.
 * The class is designed such that it would be possible that the key receiving 
 * could run in a separate thread. In fact it runs in main thread.
 * TODO: Mouse motion doesn't go to network - probably this isn't needed
 */
class InputManager : public Mutex
{
public:
	InputManager();
	~InputManager();

	/**
	 * \brief Add input acceptor to input acceptors list mIAVector.
	 * 
	 * Note that InputAcceptor doesn't need to be registered in list
	 * unless it needs to get all input at some point.
	 * \param ia input acceptor that should be added. Priority of acceptor
	 *           will be determined by calling ia.getNice() The priority
	 *           MUST be unique.
	 */
	void addInputAcceptor(InputAcceptor &ia);
	
	/**
	 * \brief Return input acceptor from input acceptors list.
	 *
	 * \param ia input acceptor that should be removed.
	 */
	void removeInputAcceptor(InputAcceptor &ia);

	/**
	 * \brief Read input from SDL. 
	 *
	 * This may be called from separate thread. When engine is
	 * running it will be in thread togehter with network.
	 * Check against mWantedKeys_netThread and if position with key is != 0,
	 * the key is added to queue mKeys with keys waiting for processing.
	 * The key is put into queue also when mWantedAllInput is true.
	 */
	void checkSDLEvents();

	/**
	 * \brief Process events from SDL stored in queue.
	 *
	 * This should be called from engine. Check against mWantedKeys_mainThread
	 * and calls input acceptors.
	 */
	void processSDLEvents();
	
	/**
	 * \brief Proces one key.
	 *
	 * If you want to process SDL events on you own and only use this for
	 * keys handling simply call this when you get new key.
	 * \param keysym key that was pressed
	 * \param down true if the key is down now
	 */
	void keyChange(SDL_keysym keysym, bool down);

	/**
	 * \brief Process mouse button change state.
	 *
	 * The mouse buttons are checked in the second thread
	 * and aren't stored for later processing like key does.
	 * \param button index of button
	 * \param pressed current state of button
	 */
	void mouseButton(int button, bool pressed);

	/**
	 * \brief Check if someone wants this keyevent in main thread.
	 *
	 * If the key is wanted in main thread it is placed into queue for
	 * later processing in main thread. If not is is passed to mNetAcceptor.
	 * \param event Event that produced this key.
	 * \param key The key from event.
	 */
	void checkKeyEvent(SDL_Event &event, SDL_keysym keysym);
	
	
	/**
	 * \brief Processing of mouseMotion event in main thread.
	 *
	 * \param dx Change of x mouse cursor coordinate from event.
	 * \param dy Change of y mouse cursor coordinate from event.
	 */
	void mouseMotion(int dx, int dy);
		
	/**
	 * \brief Send all input to queue for processing and none to netwok.
	 *
	 * Component that wants all input may set this to true and then all input
	 * will go to queue. It also must be in mIAVector if it wants to get the input.
	 * mAllInputWanters will automaticly set to 0 when some key will pass
	 * all components in mIAVector and noone will want it.
	 * \param wanted You may set to false if you want to be polite and 
	 *               decrease mAllInputWanters. Note that you may lost one key
	 *               for network network if you will not be polite.
	 */
	void wantAllInput(bool wanted = true);

	/**
	 * \brief Set forwaring of one key to input acceptor.
	 *
	 * Note that the acceptor doesn't need to be rigistered by add InputAcceptor.
	 * This will cause forwarding one key to input acceptor. InputAcceptor
	 * may call this function more times to receive more keys.
	 * 
	 * \param key key to forward. Use keys from SDL (like SDLK_e,..)
	 * \param receiver input acceptor that want to receive this key.
	 *                 by setting to NULL you may cancel forwading
	 *                 keys.
	 */
	void setKeyReceiver(int key, InputAcceptor *receiver);

	/**
	 * \brief Clear all entry from receiver
	 */
	void cleanKeyReceiver(InputAcceptor *receiver);

	/**
	 * \brief Set acceptor that receive input from keyDown that nobody
	 *        wants.
	 * Note that mNetAcceptor isn't locked so be sure that only one game thread
	 * is running when calling this or call this from network thread.
	 * \param netAcceptor network acceptor. May be 0.
	 */
	void setNetAcceptor(InputAcceptor *netAcceptor);
	
private:

	/**
	 * \brief Disable copying input manager.
	 *
	 * The InputManager shoudn't be copied. Anyway
	 * if you call this private method, the application
	 * fails.
	 */
	InputManager(const InputManager &im);
	
	std::map<int,InputAcceptor*> mIAVector; ///< list of input wanters
	
	/**
	 * List of events that are waiting for processing in main thread.
	 */
	std::queue<SDL_Event> mEvents;

	
	/**
	 * \brief Keys that are wanted in main thread.
	 */
	InputAcceptor* mWantedKeys_mainThread[SDLK_LAST]; 

	/**
	 * \brief Number of components that would like to take full input.
	 */
	int mAllInputWanters;

	/**
	 * \brief Object that accepts input that nobody else wants.
	 *
	 * It should be threadsafe since it runs (or may run) in different thread.
	 */
	InputAcceptor* mNetAcceptor;
};

#endif // _INPUT_MANAGER_H
