/* MAINTAINER: Bernard */

//#define WANT_TRACE

#include "keybindings.h"
#include "application_client.h"
#include "commander.h"
#include "log.h"
#include "conversions.h"
#include "tinyxml/tinyxml.h"
#include "inputmanager.h"

using namespace std;

const char undefinedButton[] = "----";

KeyBindings::KeyBindings()
{
	mPriority = IA_PRIORITY_KEYBINDINGS;
	mHiPriority.setNice(IA_PRIORITY_KEYBINDINGS_HI_PRIORITY);
}

KeyBindings::~KeyBindings()
{
}


void KeyBindings::init()
{
	// Create key names
	for (int i = 0; i < SDLK_LAST; i++)
	{
		mKeyNames[i] = SDL_GetKeyName((SDLKey)i);
		for (int j = 0; j < (signed int)mKeyNames[i].length(); j++)
		{
			if (mKeyNames[i][j] == ' ') mKeyNames[i][j] = '_';
		}
	}

#ifdef WANT_TRACE
	for (int i = 0; i < SDLK_LAST; i++)
	{
		LOGT << i << "  is:'" << mKeyNames[i] << "'" << endl;
	}
#endif

	// Create mouse buttons names
	for (int i = 0; i < MAX_SDL_MOUSE_BUTTONS; i++)
	{
		mMouseButtonNames[i].append("mouse_").append(itos(i));
	}

	mHiPriority.setOwner(this);
	getAppClient().getInputManager().addInputAcceptor(mHiPriority);

	
	// Loading basic command map
	loadBasicCommandsMap();

}

bool KeyBindings::keyDown(const SDL_keysym &keysym)
{
	TIN;

	assertL((keysym.sym < SDLK_LAST && keysym.sym >= 0));

	lock();
	if (!mKeyBindingsDown[keysym.sym].empty())
	{
		LOGT << "Know keysym " << mKeyNames[keysym.sym];
		getAppClient().getCommander().addCommand( mKeyBindingsDown[keysym.sym] );
		unlock();
		return true;
	}
	unlock();

	TPOINTS("Key not binded");

	return false;
}

bool KeyBindings::keyUp(const SDL_keysym &keysym)
{
	TIN;

	assertL((keysym.sym < SDLK_LAST && keysym.sym >= 0));

	lock();
	if (!mKeyBindingsUp[keysym.sym].empty())
	{
		getAppClient().getCommander().addCommand( mKeyBindingsUp[keysym.sym] );
		unlock();
		return true;
	}
	unlock();

	return false;
}

bool KeyBindings::mouseButton(int button, bool pressed)
{
	assertL(button < MAX_SDL_MOUSE_BUTTONS && button >= 0);
	lock();
		if (pressed)
		{
			if (!mMouseBindingsDown[button].empty())
			{		
				getAppClient().getCommander().addCommand( mMouseBindingsDown[button] );
				unlock();
				return true;
			}
		}
		else
		{
			if (!mMouseBindingsUp[button].empty())
			{		
				getAppClient().getCommander().addCommand( mMouseBindingsUp[button] );
				unlock();
				return true;
			}
		}
	unlock();
	return false;
}

bool KeyBindings::wantKey(const SDL_keysym &keysym)
{
	TIN;

	lock();
		bool rv = !mKeyBindingsDown[keysym.sym].empty() || !mKeyBindingsUp[keysym.sym].empty() ;
	unlock();

	return rv;
}

void KeyBindings::unbindAll()
{
	lock();
	for ( int i = 0; i < SDLK_LAST; i++ )
	{
		mKeyBindingsDown[i].clear();
		mKeyBindingsUp[i].clear();
		mKeyBindingsBasic[i].clear();
	}

	for ( int i = 0; i < MAX_SDL_MOUSE_BUTTONS; i++ )
	{
		mMouseBindingsDown[i].clear();
		mMouseBindingsUp[i].clear();
		mMouseBindingsBasic[i].clear();		
	}

	getAppClient().getInputManager().cleanKeyReceiver(this);
	mInputToNetwork.clearActionTables();
	mHiPriority.clearForwardTable();
	unlock();
}


void KeyBindings::bindKey( SDLKey key, bool buttondown, const std::string &command, bool highpriority)
{
	TIN;

	assertL((key < SDLK_LAST && key >= 0));

	// Try to map to basic command
	basicCommandMapType::iterator it = mBasicCommandMap.find(command);
	if (it != mBasicCommandMap.end())
	{
		mKeyBindingsBasic[key] = command;
		mInputToNetwork.setKeyAction(key, it->second);
		return;
	}


	// Map as standart command
	lock();
	if (buttondown) mKeyBindingsDown[key] = command;
	else            mKeyBindingsUp[key]   = command;

	if (highpriority)  
	{
		LOGT << "High priority key " << mKeyNames[key] << endl;
		mHiPriority.forwardKey(key);
	}
	else 
	{	
		LOGT << "Low priority key " << mKeyNames[key] << endl;
		getAppClient().getInputManager().setKeyReceiver(key, this);
	}
	unlock();

}

void KeyBindings::bindMouse( int butt_num, bool buttondown, const std::string &command)
{
	TIN;

	// Try to map to basic command
	basicCommandMapType::iterator it = mBasicCommandMap.find(command);
	if (it != mBasicCommandMap.end())
	{
		mMouseBindingsBasic[butt_num] = command;
		mInputToNetwork.setMouseAction(butt_num, it->second);
		return;
	}

	// Map is as standart command
	assertL(butt_num < MAX_SDL_MOUSE_BUTTONS && butt_num >= 0);
	lock();
	if (buttondown) mMouseBindingsDown[butt_num] = command;
	else            mMouseBindingsUp[butt_num] = command;
	unlock();
}


bool KeyBindings::bind( const std::string &key, bool buttondown, const std::string &command, bool highpriority)
{
	// The key is directly ine letter
	if (key.length() == 1)
	{
		bindKey((SDLKey)key[0], buttondown, command, highpriority);
		return true;
	}

	// Try to recognize key name
	for (int i = 0; i < SDLK_LAST; i++)
	{
		if (key == mKeyNames[i])
		{
			bindKey((SDLKey)i, buttondown, command, highpriority);
			return true;
		}
	}
	
	// Try to match key to mouse buttons
	for (int i = 0; i < MAX_SDL_MOUSE_BUTTONS; i++)
	{
		if (key == mMouseButtonNames[i])
		{
			bindMouse(i, buttondown, command);
			return true;
		}
	}
	
	// ignore undefined button
	if (key == undefinedButton) return true;
	
	// Keyname doesn't match anything
	LOGW << "Unknown key name: " << key << endl;
	return false;
}

void  KeyBindings::loadBasicCommandsMap()
{
	// Maybe it could be loaded from some file one day, but we dont bother :)
	mBasicCommandMap["left"]  =  0;	
	mBasicCommandMap["right"] =  1;	
	mBasicCommandMap["up"]    =  2;	
	mBasicCommandMap["down"]  =  3;	
	mBasicCommandMap["fire"]  =  4;	
}

InputToNetwork& KeyBindings::getInputToNetwork()
{
	return mInputToNetwork;
}


bool KeyBindings::setButton(const std::string &button, bool value)
{
	long action;
	
	basicCommandMapType::iterator it = mBasicCommandMap.find(button);
	if (it != mBasicCommandMap.end())
	{
		action = it->second;
	}
	else
	{
		if (!try_get_long(button, action) || action < 0) return false;
	}
	
	mInputToNetwork.setButton(action, value);

	return true;
}

std::string KeyBindings::keyToString(SDLKey key)
{
	if (key >= 0 && key < SDLK_LAST)
		return mKeyNames[key];
	return undefinedButton;
}
	
std::string KeyBindings::mouseToString(int butt_num)
{
	if (butt_num >= 0 && butt_num < MAX_SDL_MOUSE_BUTTONS)
		return mMouseButtonNames[butt_num];
	return undefinedButton;
}
		

std::string KeyBindings::whatKeyForCommand(const std::string &command)
{
	// Search for binding in keys
	for (int key = 0; key < SDLK_LAST; key++)
	{
		if (mKeyBindingsDown[key]  == command) return keyToString((SDLKey)key);
		if (mKeyBindingsBasic[key] == command) return keyToString((SDLKey)key);
	}
	
	// Search for binding in mouse buttons
	for (int mouse_button = 0; mouse_button < MAX_SDL_MOUSE_BUTTONS; mouse_button++)
	{
		if (mMouseBindingsDown[mouse_button]  == command) return mouseToString(mouse_button); 
		if (mMouseBindingsBasic[mouse_button] == command) return mouseToString(mouse_button); 
	}

	return undefinedButton;		
}

void KeyBindings::unbindButton(const std::string &keyName)
{
	// The key is directly ine letter
	SDLKey key = SDLK_LAST;
	if (keyName.length() == 1)
	{
		key = (SDLKey)keyName[0];
	}
	else
	{
		for (int i = 0; i < SDLK_LAST; i++)
		{
			if (keyName == mKeyNames[i])
			{
				key = (SDLKey)i;
				break;
			}
		}
	}
	
	if (key != SDLK_LAST)
	{
		mKeyBindingsDown[key] = "";
		mKeyBindingsUp[key] = "";
		mKeyBindingsBasic[key] = "";
		getAppClient().getInputManager().setKeyReceiver(key, 0);
		mHiPriority.clearForwardKey(key);
		mInputToNetwork.setKeyAction(key, -1);
		
		return;
	}
		
	// Try to match key to mouse buttons
	for (int butt_num = 0; butt_num < MAX_SDL_MOUSE_BUTTONS; butt_num++)
	{
		if (keyName == mMouseButtonNames[butt_num])
		{
			
			mMouseBindingsDown[butt_num] = "";
			mMouseBindingsUp[butt_num] = "";
			mMouseBindingsBasic[butt_num] = "";
			mInputToNetwork.setMouseAction(butt_num, -1);
			return;
		}
	}
	
	
}

