/* MAINTAINER: Bernard */

#ifndef _KEYBINDINGS_H
#define _KEYBINDINGS_H

#include "SDL.h"
#include "mutex.h"
#include "inputtonetwork.h"
#include "inputacceptor.h"
#include "log.h"
#include <string>
#include "hash_hack.h"
#include "stlextdefs.h"

typedef STL_EXT_NS::hash_map<std::string, int> basicCommandMapType;

/**
 * \brief Class that makes possible to map keys into commands.
 *
 * Goal of the class is provide easy way to mak keys into commands.
 * It is possible to create mapping of key to command and key to tank
 * constrol in this class.
 * Holds tables for button names and for bind buttons to commands.
 * The table for buttons to commands like left, right which directly
 * controls tank are stored in mInputToNetwork.
 * It is possible to bind three different things. First are commands
 * that directly controls tank. They are not commands at all. They are
 * translated into tank commands and stored in mInputToNetwork.
 * Second type are high priority keys/commands. The hight priority
 * thing are the first things that are processed. It means that it
 * is possible to use them for openenin console and also for closing console.
 * They will be available everytime. Even in console.
 * Last type are regular commands. They work only of noone else takes focus.
 * It means that for example console will overtake these keys.
 * It is possible to bind keys anytime in the game.
 */
class KeyBindings : public InputAcceptor, public Mutex
{
public:
	KeyBindings();
	virtual ~KeyBindings();

	/**
	 * \brief Message from InputManager.
	 *
	 * Check if some command is mapped on the key and if so,
	 * add the command to commander.
	 * \param keysym pressed key
	 * \return true if key was processed
	 */
	virtual bool keyDown(const SDL_keysym &keysym);

	/**
	 * \brief Message from InputManager.
	 *
	 * Does nothing. Only check if we want this kay and then return true.
	 * \param keysym released key
	 * \return true if key was processed
	 */
	virtual bool keyUp(const SDL_keysym &keysym);

	/**
	 * \brief Message from InputManager.
	 *
	 * Check if the button is associated with some command and do the
	 * command if so. The commands are emited only if pressed is true.
	 * If pressed is false no command is called.
	 * \param button mouse button
	 * \param pressed state of the button
	 * \return true if the button had mapping
	 */
 	virtual bool mouseButton(int button, bool pressed);

	/**
	 * \brief Check if the key is associated as normal 
	 *        command as send true if so.
	 *
	 * \param keysym offer for a key
	 * \return true if key is wanted
	 */
	virtual bool wantKey(const SDL_keysym &keysym);


	/**
	 * \brief Associate keynum with command for commander or with logical button
	 *
	 * \param key number of key to associate
	 * \param command command for commander or name of logical button
	 * \param highpriority specifi if this is high priority bind
	 * \param buttondown If the response is for key down or key up
	 */
	void bindKey( SDLKey key, bool buttondown, const std::string &command, bool highpriority);

	/**
	 * \brief Associate mouse button with command.
	 *
	 * \param butt_num mouse button id
	 * \param command Command for commander or into table for logic buttons.
	 * \param buttondown If the response is for key down or key up
	 */
	void bindMouse( int butt_num, bool buttondown, const std::string &command);

	/**
	 * \brief Associate key with command.
	 *
	 * \param key may be a single character directly saying what key it is
	 *            or may be entry in mKeyNames or mMouseButtonNames
	 * \param command basic command from table mBasicCommandMap for basic command
	 *            or command isn;t basic it is used as string for commanders command
	 * \param highpriority High priority has sense only for classic commands. If true,
	 *            the key will be processed before any other input
	 * \param buttondown If the response is for key down or key up
	 * \return false if the key is not recognized
	 */
	bool bind( const std::string &key, bool buttondown, const std::string &command, bool highpriority);

	/**
	 * \brief Clear all bindings.
	 */
	void unbindAll();

	/**
	 * \brief Does some inicialization of structures.
	 *
	 * For example mKeyNames must be inicialized here. It isn't possible
	 * to do it in contructor since at that tome SDL isn;t inicialized
	 * and so knows no keys.
	 */
	void init();

	/**
	 * \brief Provide access to mInputToNetwork
	 * \return mInputToNetwork
	 */
	InputToNetwork& getInputToNetwork();

	/**
	 * \brief Simulates button state for tank control.
	 *
	 * It may be entry from mBasicCommandMap or directly
	 * number of button. The button name is resolved and
	 * then mInputToNetwork.setButton is called
	 */
	bool setButton(const std::string &button, bool value);
	

	/**
	 * \brief Get string representation of key
	 */	
	std::string keyToString(SDLKey key);
	
	std::string mouseToString(int butt_num);
	
	/**
	 * \brief Get what key is used for this command
	 */
	std::string whatKeyForCommand(const std::string &command);
	
	/**
	 * \brief Unbind one button.
	 * 
	 * Clear it's binding to all command binds except newtwork layer.
	 */
	void unbindButton(const std::string &keyName);
	
private:

	void loadBasicCommandsMap(); /// \brief Inicialization for mBasicCommandMap

	std::string mKeyBindingsDown[SDLK_LAST]; /// \brief Binding of key presses to commands
	std::string mKeyBindingsUp[SDLK_LAST]; /// \brief Binding of key releases to commands
	std::string mMouseBindingsDown[MAX_SDL_MOUSE_BUTTONS]; /// \brief Binding of mouse buttons presses
	std::string mMouseBindingsUp[MAX_SDL_MOUSE_BUTTONS]; /// \brief Binding of mouse buttons releases

	std::string mKeyBindingsBasic[SDLK_LAST]; /// \brief keys for basic commands - just to remember keybinding
	std::string mMouseBindingsBasic[SDLK_LAST]; /// \brief keys for basic commands - just to remember keybinding

	std::string mKeyNames[SDLK_LAST]; /// \brief Names of keys on keyboard
	std::string mMouseButtonNames[MAX_SDL_MOUSE_BUTTONS]; /// \brief Names of mouse buttons


	basicCommandMapType mBasicCommandMap; /// \brief Maps basic command strings to key numbers. Like "left = 0".

	InputToNetwork mInputToNetwork; /// \brief Class for input to network - it is here because of key binding 

	InputForward mHiPriority;  /// \brief Class for receiving high priority key events

	/**
	 * \brief Anti copy contructor
	 * \param kb not used
	 */
	KeyBindings( const KeyBindings & kb )
	{
		DIE( "Attempt to copy bindings" );
	}


};

#endif // _KEYBINDINGS_H
