/* MAINTAINER: Martin */

#include "GL/glew.h"
#include "layer_game.h"
#include "log.h"
#include "camera.h"
#include "GLMatrix.h"

#include "globals_client.h"
#include "application_client.h"
#include "world.h"
#include "game_client.h"
#include "gfx.h"
#include "frustum.h"
#include "worldentity.h"

// TODO: remove this testing code...
#ifdef TESTING_ODE
#include "networkmanager_client.h" // tmp contacts & object lines
#include "globals.h"
#endif

#include "render_visitor.h"
#include "terrain.h"
#include "model.h"
#include "platform.h"

// terrain attributes should be in map params
#include "terrain_map_attributes.h"

#include "queue_inserter.h"

LayerGame::LayerGame()
:	mRV(0), 
	mOuterFogRenderer( 0 )
{
}

LayerGame::~LayerGame() 
{
	if (mRV) {
		delete mRV;
		mRV = 0;
	}
}

RenderVisitor * LayerGame::getRenderVisitor()
{
	if (mRV != 0)
		return mRV;
	
	mRV = new RenderVisitor();
	return mRV;
}

counted_ptr< SkyboxRenderer > LayerGame::getSkyboxRenderer()
{
	if ( mSkyboxRenderer.get() == 0 )
	{
		counted_ptr< SkyboxRenderer > tmp( new SkyboxRenderer() );
		if ( tmp.get() == 0 )
			DIE("Can't create skybox renderer instance!");

		// terrain attributes from map params
		const TerrainAttributes &ta = getAppClient().getGame().getMapParams().terrain;
		
		// get skybox description ( names of 6 textures )
		#define _side(side, path) { true, SkyboxRenderer::side, locateFile( path ), },
		struct 
		{
			bool valid;
			SkyboxRenderer::BoxSide sideType;
			std::string sideName;
		} sides[ SkyboxRenderer::SIDES_COUNT ] = 
		{
			_side(LEFT,		ta.skybox.leftTextureFileName	)
			_side(FRONT,	ta.skybox.frontTextureFileName	)
			_side(RIGHT,	ta.skybox.rightTextureFileName	)
			_side(BACK,		ta.skybox.backTextureFileName	)
			_side(TOP,		ta.skybox.topTextureFileName	)
			_side(BOTTOM,	ta.skybox.bottomTextureFileName	)
		};
			
		
		// initialize skybox renderer
		for (int i = 0; i < SkyboxRenderer::SIDES_COUNT; ++i)
			if ( sides[i].valid ) 
			{
				SurfDefault sf(sides[i].sideName);
				sf.setULC(true);
				TextureRef tref = TextureManager::getInstance()->add( &sf, GL_RGB, 0 );
				tref->bind();
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
				tmp->assignTexture( sides[i].sideType, tref );

			}

		mSkyboxRenderer = tmp;
	}
	return mSkyboxRenderer;
}

counted_ptr< OuterFogRenderer > LayerGame::getOuterFogRenderer()
{
	if ( mOuterFogRenderer.get() == 0 )
		// outer fog renderer doesn't exist, create a new one
	{
		const References * refs = getRefs();
		TerrainBase * tb ( refs->terrain ? refs->terrain->getTerrainBase() : 0 );
		if ( tb != 0 )
		{
			counted_ptr< OuterFogRenderer > tmp( new OuterFogRenderer( tb ) );
			if ( tmp.get() == 0 )
				DIE( "Can't create OuterFogRenderer instance!" );
			
			mOuterFogRenderer = tmp;
		}
	}
	return mOuterFogRenderer;
}

void LayerGame::initRefs(void)
{
	// get world
	mRefs.world		= &getAppClient().getGame().getWorld();
	assertL(mRefs.world && "GameLayer - Cannot obtain world!");
	
	// get landscape
	mRefs.landscape	= &mRefs.world->getLandscape();
	assertL(mRefs.landscape && "GameLayer - Cannot obtain landscape!");

	// get terrain
	mRefs.terrain	= &mRefs.landscape->getTerrain();
	assertL(mRefs.terrain && "GameLayer - Cannot obtain terrain!");

	// get camera
	mRefs.camera = getAppClient().getGame().getCamera();
	assertL(mRefs.camera && "GameLayer - Cannot obtain camera!");
}

void LayerGame::renderWorldEntities()
{
	// TODO: grouping entities using some quadtree
	
	// get frustum
	Frustum f;
	this->getFrustumCopy( f );
	
	// Get iterators for world entities.
	worldent_iterator 
		it = getRefs()->world->getWorldEntitiesBegin(),
		ite = getRefs()->world->getWorldEntitiesEnd();

	RenderVisitor * rv = getRenderVisitor();
	
	while (it != ite)
	{
		CommonEntity * e = *it;
		if ( !isEntityOutsideFrustum( e, &f ) & ( e != 0 ) ) 
		{
			Model * m = e->getModel();

			if ( m ) 
			{
				HierarchyNode * root = m->getRoot();
				if ( root ) 
				{
					m->updateAnimation(0);

					glColor3f(1, 1, 1);
					glPushMatrix();
					e->getPosition().glTranslate();
					e->getOrientation().matrix().glMultMatrix();

					root->accept( rv );
					glPopMatrix();
				}
			}
		}
		
		++it;
	}
}

bool LayerGame::getFrustumCopy( class Frustum & f )
{
	initRefs();
	if ( getRefs() && getRefs()->camera ) 
	{
		f = getRefs()->camera->getFrustum();
		return true;
	}
	return false;
}


bool LayerGame::isEntityOutsideFrustum( class CommonEntity * e, const class Frustum * f ) const
{
	if ( e == 0 || f == 0 )
		return true;

	// 1. get bounding sphere in model space ( described by its center and radius )
	GLVector3d center(0, 0, 0), pos( e->getPosition() );
	double radius(0);
	
	Model * m = e->getModel();
	if ( m == 0 )
		return true;

	m->getBoundingSphere( center, radius );

	// 2. transform bounding sphere to world space by adjusting its position
	center += pos;

	// 3. test sphere against frustum
	Frustum::ClipState res;
	res = f->testSphere( center, radius );
	if ( res == Frustum::OUTSIDE )
		return true;
	/*
	else if ( res == Frustum::INTERSECT )
	{
		// HINT: maybe, bounding box could be used for occlusion culling test only. This will be performed
		// in hardware with terrain (plus landscape entities?) as ocludders.

		// try to test with bounding box
		GLVector3d extents( 0, 0, 0 );
		center.set(0, 0, 0);
		// TODO:
		//m->getAABB( center, extents );

		center += pos;
		Frustum f2( *f );
		f2.transform( e->getOrientation().matrix() );
		if ( f->testBox( center, extents ) == Frustum::OUTSIDE )
			return true;	
	}
	*/
	return false;
}

counted_ptr< RenderQueue > LayerGame::getRenderQueue()
{
	if ( !mRenderQueue.get() )
	{
		mRenderQueue = counted_ptr< RenderQueue >( new RenderQueue() );
		mQueueInserter = counted_ptr< QueueInserter >(0);
	}
	return mRenderQueue;
}

counted_ptr< QueueInserter > LayerGame::getQueueInserter()
{
	if ( !mQueueInserter.get() && getRenderQueue().get() )
	{
		mQueueInserter = counted_ptr< QueueInserter > ( new QueueInserter( getRenderQueue() ) );
	}
	return mQueueInserter;
}


template <class ent_iterator>
void LayerGame::updateEntities( ent_iterator begin, ent_iterator end )
{
	Frustum f; 
	getFrustumCopy( f );
	counted_ptr< QueueInserter > qins = getQueueInserter();
	
	while ( begin != end )
	{
		CommonEntity * e = *begin;
		if ( !isEntityOutsideFrustum( e, &f ) & ( e != 0 ) ) 
		// entity passed the visibility test
		{
			Model * m = e->getModel();
			if ( m )
			{
				m->updateAnimation(0);
				qins->insertModel( m, e->getPosition(), e->getOrientation() );
			}
		}
		
		++begin;
	}
}

void LayerGame::renderLandscapeEntities()
{
	// TODO: grouping entities using some quadtree
	// TODO: frustum culling enabled/disabled
	// get copy of actual frustum
	Frustum f;
	this->getFrustumCopy( f );
	
	// Get iterators for world entities.
	landent_iterator 
		it = getRefs()->landscape->getLandEntitiesBegin(),
		ite = getRefs()->landscape->getLandEntitiesEnd();

	RenderVisitor * rv = getRenderVisitor();
	
	while (it != ite)
	{		
		CommonEntity * e = *it;
		if ( !isEntityOutsideFrustum( e, &f ) & ( e != 0 ) ) 
		{
			Model * m = e->getModel();
			if ( m ) 
			{
				HierarchyNode * root = m->getRoot();
				if ( root ) 
				{
					// landscape entities could be animated as well as world entities
					m->updateAnimation(0);

					glColor3f(1, 1, 1);
					glPushMatrix();
					e->getPosition().glTranslate();
					e->getOrientation().matrix().glMultMatrix();

					root->accept( rv );

					glPopMatrix();
				}
			}
		}
		
		++it;
	}
}

void LayerGame::renderLandscape(void)
{
	Terrain * terrain = &getRefs()->landscape->getTerrain();
	if ( terrain )
		terrain->render();	
}

void LayerGame::render()
{
	initRefs();

	setView();

	GLfloat LightPosition[] = {100, 0, 100, 0 };
	glLightfv(GL_LIGHT0, GL_POSITION, LightPosition);
	glEnable( GL_LIGHT0 );


	// ---------------------
	// --- RENDER SKYBOX ---
	// ---------------------
	counted_ptr< SkyboxRenderer > skyboxRenderer = this->getSkyboxRenderer();
	if ( skyboxRenderer.get() ) 
		skyboxRenderer->render();
	
	// -----------------
	// --- SETUP FOG ---
	// -----------------
	// TODO: design: less hardwired and more configurable
	GLfloat density = 0.0009f;
	GLfloat fogColor[4] = {0.5, 0.5, 0.5, 1.0};
	glEnable( GL_DEPTH_TEST );
	glEnable( GL_FOG );
	glFogi( GL_FOG_MODE, GL_EXP2 );
	glFogfv (GL_FOG_COLOR, fogColor);
	glFogf (GL_FOG_DENSITY, density);
	glHint (GL_FOG_HINT, GL_DONT_CARE);
	glFogf(GL_FOG_START, 150.0f);
	glFogf(GL_FOG_END, 1500.0f);

	// ------------------------
	// --- RENDER LANDSCAPE ---
	// ------------------------
	glDisable( GL_LIGHTING );
	glDisable( GL_BLEND );
	glDisable( GL_ALPHA_TEST );

	renderLandscape();
	
	// -----------------------------
	// --- RENDER QUEUED OBJECTS ---
	// -----------------------------
	counted_ptr< RenderQueue > q = getRenderQueue();
	if ( q.get() ) 
	{
		glEnable( GL_LIGHTING );
		q->render();
		glDisable( GL_LIGHTING );
	}
	q->clear();
	/*
	glEnable( GL_LIGHTING );

	// ---------------------------------
	// --- RENDER LANDSCAPE ENTITIES ---
	// ---------------------------------
	renderLandscapeEntities();

	// -----------------------------
	// --- RENDER WORLD ENTITIES ---
	// -----------------------------
	renderWorldEntities();

	glDisable( GL_LIGHTING );
	*/
#ifdef TESTING_ODE
	/* DEBUG: (zobrazuje kontakt pointy)
	*/
	extern std::list<GLVector3d> contacts_dbg;
	glPointSize( 10.0 );
	glBegin( GL_POINTS );
	glColor3d( 0.0, 1.0, 0 );
	std::list< GLVector3d >::iterator it;
	std::list< GLVector3d >::iterator ite = contacts_dbg.end();
	for ( it = contacts_dbg.begin(); it != ite; ++it )
	{
		glVertex3d( (*it).x, (*it).y, (*it).z );
	}
	glEnd();

	extern std::list< struct _info > objs_dbg;
	glBegin( GL_LINES );
	glColor3d( 0.0, 1.0, 0 );
	std::list< struct _info >::iterator it2;
	std::list< struct _info >::iterator ite2 = objs_dbg.end();
	for ( it2 = objs_dbg.begin(); it2 != ite2; ++it2 )
	{
		GLVector3d dir = (*it2).ori.rotateVector( ZERO_ROTATION_VECTOR );
		dir.unit();
		dir *= 10;
		dir+= (*it2).pos;
		dir.glVertex();
		(*it2).pos.glVertex();
	}
	glEnd();

	/**/
#endif

	// ------------------------
	// --- RENDER OUTER FOG ---
	// ------------------------
	counted_ptr< OuterFogRenderer > outerFogRenderer = getOuterFogRenderer();
	if ( outerFogRenderer.get() )
	{
		glColor3f( 0.8f, 0.8f, 0.8f );
		outerFogRenderer->render();
	}

	glDisable( GL_FOG );
}

void LayerGame::update()
{
	initRefs();

	// test name of current map for changes, call onMapChange if map name has changed
	Params::sMapParams &map = getAppClient().getGame().getMapParams();
	if ( currentMapName != map.name )
	{
		this->onMapChanged();
		currentMapName = map.name;
	}

	getRefs()->camera->computeFrustum();
	getRefs()->terrain->update( getRefs()->camera );

	// -------------------------------------------
	// --- UPDATE WORLD AND LANDSCAPE ENTITIES ---
	// -------------------------------------------
	// - determine visibility
	// - for each visible entity
	//		- update animation
	//		- insert renderable parts into render queue 
	updateEntities( getRefs()->world->getWorldEntitiesBegin(), getRefs()->world->getWorldEntitiesEnd() );
	updateEntities( getRefs()->landscape->getLandEntitiesBegin(), getRefs()->landscape->getLandEntitiesEnd() );

	// --- Tady se provede nejake to castecne usporadani fronty
	// 		- sort by opaque/masked/translucent
	//		- group by material ( or texture id )
	getRenderQueue()->sort();
}

void LayerGame::onMapChanged()
{
	// if the map has changed, release skybox renderer and let she ;) reinitialize on next access
	mSkyboxRenderer = counted_ptr<SkyboxRenderer>( 0 );
}

void LayerGame::setView()
{
	// TODO: but keep this code... and remove this comment
	getRefs()->camera->setView();
}

/* DEBUG: (zobrazuje kontakt pointy

extern std::list<GLVector3d> contacts_dbg;
glPointSize( 10.0 );
glBegin( GL_POINTS );
glColor3d( 1.0, 0, 0 );
std::list< GLVector3d >::iterator it;
std::list< GLVector3d >::iterator ite = contacts_dbg.end();
for ( it = contacts_dbg.begin(); it != ite; ++it )
{
	glVertex3d( (*it).x, (*it).y, (*it).z );
}
glEnd();
*/
// Debug - bounding sphere rendering
/*
glPushMatrix();
glPushAttrib(GL_POLYGON_BIT | GL_ENABLE_BIT );
	glDepthFunc( GL_ALWAYS );
	glDisable(GL_TEXTURE_2D);
	glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );
	GLVector3d c;
	double r;
	m->getBoundingSphere(c, r);
	e->getPosition().glTranslate();
	c.glTranslate();
	e->getOrientation().matrix().glMultMatrix();
	grutils::glDrawSphere(r, 12, 12);
	glDepthFunc( GL_LEQUAL );
glPopAttrib();
glPopMatrix();
*/
