/* MAINTAINER: Martin */

#ifndef _LAYER_GAME_H
#define _LAYER_GAME_H

#include "layer.h"
#include "log.h"

#include "counted_ptr.h"
#include "skybox_renderer.h"
#include "outer_fog_renderer.h"
#include "render_queue.h"

class RenderVisitor;
class QueueInserter;

/**
*	\brief world rendering, terrain updating
*/
class LayerGame : public Layer
{
public:
	LayerGame();
	virtual ~LayerGame();

	/** \brief game-rendering code */
	virtual void render();
	
	/** \brief frustum calculation, LOD-update code */
	virtual void update();
	
private:

	struct References
	{
		class World		* world;
		class Landscape	* landscape;
		class Terrain	* terrain;
		class Camera	* camera;

	} mRefs; ///< \brief some useful references

	/** \brief obtains references mRefs to world, landcape, etc. for easy access */
	void initRefs(void);
	const References * getRefs(void) { return &mRefs; }

	/**	\brief sets the projection and model-view matrix for concrete layer */
	virtual void setView();

	bool getFrustumCopy( class Frustum & f );
	bool isEntityOutsideFrustum( class CommonEntity * e, const class Frustum * f ) const;
	
	/** \brief rendering code for world entities */
	void renderWorldEntities(void);
	void renderLandscapeEntities(void);

	/**
	*	\brief prepare entities for rendering
	*
	*	For each entity:
	*	- check visibility
	*	- update animation of visible entity
	*	- insert renderable parts of visible entity into render queue
	*/
	template <typename ent_iterator>
	void updateEntities( ent_iterator begin, ent_iterator end );

	/** 
	*	\brief get render queue 
	*	
	*	If render queue doesn't exist, it will be created.
	*	
	*	\return render queue
	*/
	counted_ptr< RenderQueue > getRenderQueue();
	counted_ptr< QueueInserter > getQueueInserter();

	RenderVisitor * getRenderVisitor();
	counted_ptr< SkyboxRenderer > getSkyboxRenderer();
	counted_ptr< OuterFogRenderer > getOuterFogRenderer();
	
	/** \brief rendering code for landscape */
	void renderLandscape(void);

	RenderVisitor * mRV;
	counted_ptr< SkyboxRenderer > mSkyboxRenderer;
	counted_ptr< OuterFogRenderer > mOuterFogRenderer;
	counted_ptr< RenderQueue > mRenderQueue;
	counted_ptr< QueueInserter > mQueueInserter;

	void onMapChanged();

	std::string currentMapName;
	
};

/* TODO:
Split renderLandscape() to terrain rendering and landscape entities rendering or implement 
World::render() method that will merge landscape rendering and world entities rendering.
*/

#endif //_LAYER_GAME_H
