/* MAINTAINER: Bernard */
//#define WANT_TRACE

#include <CEGUI.h>
#include <algorithm>
#include <sstream>
#include <vector>
#include <string>
#include "menumain.h"
#include "log.h"
#include "application_client.h"
#include "commander.h"
#include "globals.h"
#include "iortransfer_client.h"
#include "settings.h"
#include "networkmanager_client.h"
#include "sfx.h"
#include "gfx.h"
#include "inputmanager.h"
#include "keybindings.h"
#include "platform.h"
#include "iortransfer_client.h"
#include "thread.h"
#include "translator.h"
#include "conversions.h"
#include "globals.h"
#include "tinyxml/tinyxml.h"
#include "timemanager.h"
#include "gui_constants.h"
#include "console.h"

using namespace CEGUI;
using namespace std;

/**
 * \brief Class for accepting input from InputManager with high priorioty.
 * 
 * It is used for changing keybinding in MenuMain.
 */
class OneKeyAcceptor: public InputAcceptor
{
public:	
	OneKeyAcceptor(MenuMain *mmain)
	{
		assertL(mmain);
		
		mMMain = mmain;
		mPriority = IA_PRIORITY_ONEKEY;
	}
	
	/**
	 * \brief Retrurn always true since it accepts any input.
	 */
	virtual bool wantKey(const SDL_keysym &keysym)
	{
		return true;
	}	
	
	/**
	 * \brief Forward key to MenuMain.
	 * 
	 * Hide warning dialog about activated acceptor.
	 * \param keysym The pressed key for forwarding.
	 * \return Always true since the input is accepted.
	 */
	virtual bool keyDown(const SDL_keysym &keysym)
	{
		TIN;
		
		if (mMMain && keysym.sym != SDLK_ESCAPE)
		{
			KeyBindings& keybind = getAppClient().getKeyBindings();
			mMMain->oneKey(keybind.keyToString(keysym.sym), mCommandName);
		}
		
		getAppClient().getInputManager().removeInputAcceptor(dynamic_cast<InputAcceptor&>(*this));
		mMMain->mMOptions->getChild("options.message")->setVisible(false);	
		
		return true;	
	}
	
	/**
	 * \brief Forward mouse press to MainMenu.
	 * 
	 * The button is forwarded only if pressed.
	 * Hide warning dialog about activated acceptor.
	 * \param button The button to forward.
	 * \param pressed Not pressed button isn't passed to the MainMenu
	 *        and we will wait for next one.
	 * \return Always true.
	 */
 	virtual bool mouseButton(int button, bool pressed)
 	{
 		TIN;
 		
 		if (!pressed) return true;
 		
		if (mMMain) 
		{
	 		KeyBindings& keybind = getAppClient().getKeyBindings();
			mMMain->oneKey(keybind.mouseToString(button), mCommandName);
		}
		
		getAppClient().getInputManager().removeInputAcceptor(dynamic_cast<InputAcceptor&>(*this));
		mMMain->mMOptions->getChild("options.message")->setVisible(false);	
		
		return true;	
 	}
	
	
	/**
	 * \brief Activatiing the acceptor.
	 * 
	 * The acceptor will became active.
	 * Show warning dialog about activated acceptor.
	 * \param command_name Command for which the keybind is changed.
	 *        The command name is later returned back to MenuMain.
	 *        The value is stored in mCommandName.
	 */
	void getKey(const std::string& command_name)
	{
		TIN;
		
		assertL(mMMain);
		
		mCommandName = command_name;
		getAppClient().getInputManager().addInputAcceptor(dynamic_cast<InputAcceptor&>(*this));
		getAppClient().getInputManager().wantAllInput(true);
		mMMain->mMOptions->getChild("options.message")->setVisible(true);			
	}

private:	
	MenuMain    *mMMain; ///< \brief Pointer to our owner. Must be valid.
	std::string  mCommandName; ///< \brief Last command name.
};


// Sample sub-class for ListboxTextItem that auto-sets the selection brush
// image.  This saves doing it manually every time in the code.
/**
 * \brief Class for associating ServerInfo to ListboxTextItem and
 *        for setting CEGUI selecting brush 
 */
class SInfoListItem: public CEGUI::ListboxTextItem
{
public:
    SInfoListItem(const CEGUI::String& text, const ServerInfo &sinfo) : ListboxTextItem(text)
    {
        setSelectionBrushImage("TaharezLook", "MultiListSelectionBrush");
		mSInfo = sinfo;
    }

	ServerInfo mSInfo; ///< \brief The associated info.
};

/**
 * \brief ListboxTextItem with added selection brush.
 */
class ResolutionListItem: public CEGUI::ListboxTextItem
{
public:
    ResolutionListItem(const CEGUI::String& text) : ListboxTextItem(text)
    {
        setSelectionBrushImage("TaharezLook", "ComboboxSelectionBrush");
//        setSelectionBrushImage("SleekSpace", "ComboboxSelectionBrush");
    }
};

/**
 * \brief Class for associating aditional string to lisbox item.
 */
class ListItemAndString: public CEGUI::ListboxTextItem
{
public:
    ListItemAndString(const CEGUI::String& text, std::string &datastring) : ListboxTextItem(text)
    {
        setSelectionBrushImage("TaharezLook", "MultiListSelectionBrush");
		mString = datastring;
    }
    
    std::string mString; ///< \brief The associated string.
};


/** \brief Thread for receiving broadcast info from network
 */
class GetNetworkGames : public Thread
{
public:
	GetNetworkGames(MenuMain *mmain)
	{
		assertL(mmain);
		mMenuMain = mmain;
	}
	
	/**
	 * \brief Wait for broadcast info and pass new info to menu.
	 * 
	 * The list is sorted using IOR since servers may be returned in different order
	 * in each call and it would look ugly if the server would be mixing in menu
	 * in every refresh.
	 */	
	virtual int run()
	{
		assertL(mMenuMain);
		std::vector<ServerInfo> sviv;
		sviv = getLANServers( g_settings.getInt(SETTING_TMP_NETWORK_CONNECT_PORT) );
		sort(sviv.begin(), sviv.end(), compareServerInfo);

		mMenuMain->setServerInfos(sviv);
		
		return 0;
	}

	/**
	 * \brief Compare two IOR info.
	 */
	static bool compareServerInfo(const ServerInfo &sviv_a, const ServerInfo &sviv_b)
	{
		return sviv_a.IOR < sviv_b.IOR;
	}
	
protected:
	MenuMain* mMenuMain; ///< \brief Menu which spawned this thread.
};

MenuMain::MenuMain()
{

	mTickDelta = g_settings.getInt(SETTING_UPDATE_SERVER_LIST_DELAY);
	mOneKeyAcceptor = new OneKeyAcceptor(this);
	mGetNetworkGamesThread = new GetNetworkGames(this);
	if (!mOneKeyAcceptor || !mGetNetworkGamesThread) DIE("new failed.");

	// Load background
	mBackGround  = TextureManager::getInstance()->add( locateFile( gc.mmain_background ) );
	if ( ! mBackGround.isValid() )
    {
		DIE( "Cannot load texture for menu background" );
	}
	
	// Load and prepare menu screens
	try
	{
		WindowManager &wm = WindowManager::getSingleton();

		// Main menu
		mMMain = wm.loadWindowLayout("menu.main.layout");
		translateWindowTextRecursive(mMMain);

		Window *w;
		w = mMMain->getChild("main.quit");
		w->subscribeEvent(PushButton::EventClicked, CEGUI::Event::Subscriber(&MenuMain::handleQuit, this));

		w = mMMain->getChild("main.credits");
		w->subscribeEvent(PushButton::EventClicked, CEGUI::Event::Subscriber(&MenuMain::handleToCredists, this));

		w = mMMain->getChild("main.single");
		w->subscribeEvent(PushButton::EventClicked, CEGUI::Event::Subscriber(&MenuMain::handleToSingle, this));

		w = mMMain->getChild("main.multi");
		w->subscribeEvent(PushButton::EventClicked, CEGUI::Event::Subscriber(&MenuMain::handleToMulti, this));
		
		w = mMMain->getChild("main.options");
		w->subscribeEvent(PushButton::EventClicked, CEGUI::Event::Subscriber(&MenuMain::handleToOptions, this));

		w = mMMain->getChild("main.version");
		w->setText(BJS_VERSION_STRING);
#ifndef _DEBUG
		w->setVisible(false);
#endif			

		// Credits menu
		mMCredits = wm.loadWindowLayout("menu.credits.layout");
		translateWindowTextRecursive(mMCredits);

		w = mMCredits->getChild("credits.back");
		w->subscribeEvent(PushButton::EventClicked, CEGUI::Event::Subscriber(&MenuMain::handleToMain, this));

		// Set main menu as default
		mWindow = mMMain;

		// Multi player
		mMMulti =  wm.loadWindowLayout("menu.multi.layout");
		translateWindowTextRecursive(mMMulti);

		w = mMMulti->getChild("multi.back");
		w->subscribeEvent(PushButton::EventClicked, CEGUI::Event::Subscriber(&MenuMain::handleToMain, this));

		w = mMMulti->getChild("multi.connect"); 
		w->subscribeEvent(PushButton::EventClicked, CEGUI::Event::Subscriber(&MenuMain::handleConnect, this));

		w = mMMulti->getChild("multi.create"); 
		w->subscribeEvent(PushButton::EventClicked, CEGUI::Event::Subscriber(&MenuMain::handleToMultiCreate, this));
		
		w = mMMulti->getChild("multi.edit"); 
		w->subscribeEvent(Editbox::EventTextAccepted,     CEGUI::Event::Subscriber(&MenuMain::handleConnect, this));

		MultiColumnList* mclbox = static_cast<MultiColumnList*>(mMMulti->getChild("multi.gameslist"));
		mclbox->subscribeEvent(Editbox::EventMouseDoubleClick, CEGUI::Event::Subscriber(&MenuMain::handleConnectFromList, this));
		mclbox->setSelectionMode(MultiColumnList::RowSingle);
	 	mclbox->addColumn( (const CEGUI::utf8*)_("Welcome msg"),  0, UDim(0.25f, 0));
	 	mclbox->addColumn( (const CEGUI::utf8*)_("Map"),          1, UDim(0.28f,0));
	 	mclbox->addColumn( (const CEGUI::utf8*)_("Players"),      2, UDim(0.20f,0));
	 	mclbox->addColumn( (const CEGUI::utf8*)_("IP/DNS"),       3, UDim(0.20f,0));

		// Options
		mMOptions = wm.loadWindowLayout("menu.options.layout");
		translateWindowTextRecursive(mMOptions);
		
		w = mMOptions->getChild("options.cancel");
		w->subscribeEvent(PushButton::EventClicked, CEGUI::Event::Subscriber(&MenuMain::handleToMain, this));
		
		w = mMOptions->getChild("options.ok");
		w->subscribeEvent(PushButton::EventClicked, CEGUI::Event::Subscriber(&MenuMain::handleApplyOptions, this));
		
		w = mMOptions->getChild("options.pane.audio");
		w->subscribeEvent(PushButton::EventClicked, CEGUI::Event::Subscriber(&MenuMain::handleOptionsAudio, this));
		w = mMOptions->getChild("options.pane.video");
		w->subscribeEvent(PushButton::EventClicked, CEGUI::Event::Subscriber(&MenuMain::handleOptionsVideo, this));
		w = mMOptions->getChild("options.pane.keybind");
		w->subscribeEvent(PushButton::EventClicked, CEGUI::Event::Subscriber(&MenuMain::handleOptionsKeybind, this));
		w = mMOptions->getChild("options.pane.other");
		w->subscribeEvent(PushButton::EventClicked, CEGUI::Event::Subscriber(&MenuMain::handleOptionsOther, this));
		
		FrameWindow *frame = static_cast<FrameWindow*>(mMOptions->getChild("options.frame.other"));
		frame->setRollupEnabled(false);
		frame = static_cast<FrameWindow*>(mMOptions->getChild("options.frame.video"));
		frame->setRollupEnabled(false);
		frame = static_cast<FrameWindow*>(mMOptions->getChild("options.frame.audio"));
		frame->setRollupEnabled(false);
		frame = static_cast<FrameWindow*>(mMOptions->getChild("options.frame.keybind"));
		frame->setRollupEnabled(false);

		// keybind
		w = mMOptions->getChild("options.frame.keybind");
		mclbox = static_cast<MultiColumnList*>(w->getChild("options.keybind.list"));
		mclbox->subscribeEvent(Editbox::EventMouseDoubleClick, CEGUI::Event::Subscriber(&MenuMain::handleChangeKeybind, this));
		mclbox->setSelectionMode(MultiColumnList::RowSingle);
	 	mclbox->addColumn( (const CEGUI::utf8*)_("Action"),  0, UDim(0.65f, 0));
	 	mclbox->addColumn( (const CEGUI::utf8*)_("Key"),     1, UDim(0.28f,0));


		// Single player
		mMSingle = wm.loadWindowLayout("menu.single.layout");
		translateWindowTextRecursive(mMSingle);

		w = mMSingle->getChild("single.back");
		w->subscribeEvent(PushButton::EventClicked, CEGUI::Event::Subscriber(&MenuMain::handleToMain, this));	
		w = mMSingle->getChild("single.play");
		w->subscribeEvent(PushButton::EventClicked, CEGUI::Event::Subscriber(&MenuMain::handleSinglePlay, this));
		w = mMSingle->getChild("single.aiplayers.plus");
		w->subscribeEvent(PushButton::EventClicked, CEGUI::Event::Subscriber(&MenuMain::handleSingleAIPlus, this));
		w = mMSingle->getChild("single.aiplayers.minus");
		w->subscribeEvent(PushButton::EventClicked, CEGUI::Event::Subscriber(&MenuMain::handleSingleAIMinus, this));
		w = mMSingle->getChild("single.timelimit.plus");
		w->subscribeEvent(PushButton::EventClicked, CEGUI::Event::Subscriber(&MenuMain::handleSingleTimeLimitPlus, this));
		w = mMSingle->getChild("single.timelimit.minus");
		w->subscribeEvent(PushButton::EventClicked, CEGUI::Event::Subscriber(&MenuMain::handleSingleTimeLimitMinus, this));
		w = mMSingle->getChild("single.fraglimit.plus");
		w->subscribeEvent(PushButton::EventClicked, CEGUI::Event::Subscriber(&MenuMain::handleSingleFragLimitPlus, this));
		w = mMSingle->getChild("single.fraglimit.minus");
		w->subscribeEvent(PushButton::EventClicked, CEGUI::Event::Subscriber(&MenuMain::handleSingleFragLimitMinus, this));

		MultiColumnList* maplist = static_cast<MultiColumnList*>(mMSingle->getChild("single.maplist"));
		maplist->setSelectionMode(MultiColumnList::RowSingle);
	 	maplist->addColumn( (const CEGUI::utf8*)_("Name"),    0, UDim(0.40f, 0));
	 	maplist->addColumn( (const CEGUI::utf8*)_("File"),    1, UDim(0.30f,0));
	 	maplist->addColumn( (const CEGUI::utf8*)_("Players"), 2, UDim(0.24f,0));

		// Create server
		mMMultiCreate = wm.loadWindowLayout("menu.multicreate.layout");
		translateWindowTextRecursive(mMMultiCreate);
		
		w = mMMultiCreate->getChild("multicreate.back");
		w->subscribeEvent(PushButton::EventClicked, CEGUI::Event::Subscriber(&MenuMain::handleToMulti, this));			
		w = mMMultiCreate->getChild("multicreate.create");
		w->subscribeEvent(PushButton::EventClicked, CEGUI::Event::Subscriber(&MenuMain::handleCreateServer, this));	
		w = mMMultiCreate->getChild("multicreate.aiplayers.plus");
		w->subscribeEvent(PushButton::EventClicked, CEGUI::Event::Subscriber(&MenuMain::handleMultiCreateAIPlus, this));	
		w = mMMultiCreate->getChild("multicreate.aiplayers.minus");
		w->subscribeEvent(PushButton::EventClicked, CEGUI::Event::Subscriber(&MenuMain::handleMultiCreateAIMinus, this));	
		w = mMMultiCreate->getChild("multicreate.fraglimit.plus");
		w->subscribeEvent(PushButton::EventClicked, CEGUI::Event::Subscriber(&MenuMain::handleMultiCreateFragLimitPlus,  this));	
		w = mMMultiCreate->getChild("multicreate.fraglimit.minus");
		w->subscribeEvent(PushButton::EventClicked, CEGUI::Event::Subscriber(&MenuMain::handleMultiCreateFragLimitMinus, this));	
		w = mMMultiCreate->getChild("multicreate.timelimit.plus");
		w->subscribeEvent(PushButton::EventClicked, CEGUI::Event::Subscriber(&MenuMain::handleMultiCreateTimeLimitPlus,  this));	
		w = mMMultiCreate->getChild("multicreate.timelimit.minus");
		w->subscribeEvent(PushButton::EventClicked, CEGUI::Event::Subscriber(&MenuMain::handleMultiCreateTimeLimitMinus, this));	
		w = mMMultiCreate->getChild("multicreate.maxplayers.plus");
		w->subscribeEvent(PushButton::EventClicked, CEGUI::Event::Subscriber(&MenuMain::handleMultiCreateMaxPlayersPlus,  this));	
		w = mMMultiCreate->getChild("multicreate.maxplayers.minus");
		w->subscribeEvent(PushButton::EventClicked, CEGUI::Event::Subscriber(&MenuMain::handleMultiCreateMaxPlayersMinus, this));	
		
		maplist = static_cast<MultiColumnList*>(mMMultiCreate->getChild("multicreate.maplist"));
		maplist->setSelectionMode(MultiColumnList::RowSingle);
	 	maplist->addColumn( (const CEGUI::utf8*)_("Name"),    0, UDim(0.40f, 0));
	 	maplist->addColumn( (const CEGUI::utf8*)_("File"),    1, UDim(0.30f,0));
	 	maplist->addColumn( (const CEGUI::utf8*)_("Players"), 2, UDim(0.24f,0));

		

		LOGD << "Menu constructed" << endl;
	} catch ( CEGUI::Exception &e)
	{
		LOGE << "CEGUI exception: " << e.getMessage() << endl;
		DIE("CEGUI ERROR");

	}

}

MenuMain::~MenuMain()
{
    WindowManager &wm = WindowManager::getSingleton();
	wm.destroyWindow(mMMain);
	wm.destroyWindow(mMCredits);
	wm.destroyWindow(mMMulti);
	wm.destroyWindow(mMSingle);
	wm.destroyWindow(mMOptions);

	if (mGetNetworkGamesThread->isRunning()) mGetNetworkGamesThread->wait();
	delete mGetNetworkGamesThread;
	
	delete  mOneKeyAcceptor;
}

void MenuMain::update(int dtime, int atime)
{
	MenuBasic::update(dtime, atime);

	if (mWindow == mMMulti)
	{
		refreshNetGames();
	}
	if (mWindow == mMMultiCreate)
	{
		Window *w = mMMultiCreate->getChild("multicreate.password");
		if (w->getText() == "")
		{
			mMMultiCreate->getChild("multicreate.create")->setVisible(false);
			mMMultiCreate->getChild("multicreate.password.warning")->setVisible(true);
		}
		else
		{
			mMMultiCreate->getChild("multicreate.password.warning")->setVisible(false);
			mMMultiCreate->getChild("multicreate.create")->setVisible(true);			
		}
	}
	
}
void MenuMain::tick()
{
	mServerInfosMutex.lock();
	mNeedNetRefresh = true;
	mServerInfosMutex.unlock();
}

bool MenuMain::handleQuit(const CEGUI::EventArgs& e)
{
	LOGD << "Exiting Program!!" << endl;
	mWantRun = false;
	return true;
}

bool MenuMain::handleToMain(const CEGUI::EventArgs& e)
{
	System::getSingleton().setGUISheet(mMMain);
	mWindow = mMMain;
	return true;
}

bool MenuMain::handleToCredists(const CEGUI::EventArgs& e)
{
	System::getSingleton().setGUISheet(mMCredits);
	mWindow = mMCredits;
		
	// Secret order which will be broken by shuffle :-)	
	vector<string> names;
	names.push_back("BERNARD LIDICKY\n");
 	names.push_back("MARTIN MANAK\n");
 	names.push_back("PETER LIBIC\n");
 	names.push_back("LUKAS MAREK\n");
 	
 	random_shuffle(names.begin(), names.end());
 	
 	string result;
 	for (unsigned int i = 0; i < names.size(); i++)
 	{
 		result.append(names[i]);
 	}
 
 	mMCredits->getChild("credits.names")->setText(result);
 	
	return true;
}

bool MenuMain::handleToMulti(const CEGUI::EventArgs& e)
{
	System::getSingleton().setGUISheet(mMMulti);
	mWindow = mMMulti;
	mNeedNetRefresh = true; // Refresh game list 
	return true;
}


/* *************************************** Multiplayer ********************************* */

void MenuMain::setServerInfos(const std::vector<ServerInfo> &svinfo)
{
	mServerInfosMutex.lock();
	mNeedNetRefresh = false;
	mServerInfos = svinfo;
	mServerInfosMutex.unlock();
}

void MenuMain::refreshNetGames()
{
	mServerInfosMutex.lock();
	if (mNeedNetRefresh && !mGetNetworkGamesThread->isRunning())
	{
		mGetNetworkGamesThread->start();
	}	
	mServerInfosMutex.unlock();
	
	try
	{
		mServerInfosMutex.lock();
		

		MultiColumnList* mclbox = static_cast<MultiColumnList*>(mMMulti->getChild("multi.gameslist"));

		ServerInfo selectedInfo;
		SInfoListItem* p_selected = static_cast<SInfoListItem*>(mclbox->getFirstSelectedItem());
		if (p_selected != 0)
		{
			selectedInfo = p_selected->mSInfo;
		}


		mclbox->resetList();

		for(int row = 0; (unsigned int)row < mServerInfos.size(); ++row)
		{
			ServerInfo sinfo = mServerInfos[row];
			mclbox->addRow();
			mclbox->setItem(new SInfoListItem(sinfo.HelloMessage,      sinfo), 0, row);
			mclbox->setItem(new SInfoListItem(sinfo.MapDescription,    sinfo), 1, row);
			mclbox->setItem(new SInfoListItem(itos(sinfo.PlayerCount), sinfo), 2, row);
			mclbox->setItem(new SInfoListItem(sinfo.IP, sinfo),                3, row);

			if (p_selected && selectedInfo.IOR == sinfo.IOR)
			{
				mclbox->setItemSelectState(MCLGridRef(row, 0), true);
				mclbox->setItemSelectState(MCLGridRef(row, 1), true);
				mclbox->setItemSelectState(MCLGridRef(row, 2), true);
				mclbox->setItemSelectState(MCLGridRef(row, 3), true);
			}
		}
		mServerInfosMutex.unlock();
	} catch ( CEGUI::Exception &e)
	{
		LOGE << "CEGUI exception: " << e.getMessage() << endl;
		DIE("CEGUI ERROR");

	}

}

bool MenuMain::handleConnect(const CEGUI::EventArgs& e)
{
	LOGD << "Connecting to some server" << endl;

	// For sure clear the old commands if any
	g_settings.set(SETTING_GAME_INIT_COMMANDS, "");

	// First check if we are disconnected
	assertL(!getAppClient().getNetworkManager().connected());
	if (getAppClient().getNetworkManager().connected()) 
	{
		return true;
	}

    // get the text entry editbox
    Editbox* editbox = static_cast<Editbox*>(mMMulti->getChild("multi.edit"));
    // get text out of the editbox
    String edit_text(editbox->getText());

	// Use string from 
    if (!edit_text.empty())
    {
		string address = edit_text.c_str();

		getAppClient().getCommander().addCommand("connect " + address);
	}
	else // selection
	{
		handleConnectFromList(e);
	}

	return true;
}

bool MenuMain::handleConnectFromList(const CEGUI::EventArgs& e)
{
	// First check if we are disconnected
	assertL(!getAppClient().getNetworkManager().connected());
	if (getAppClient().getNetworkManager().connected()) 
	{
		return true;
	}

	// read data for connectiong from list	
	MultiColumnList* mclbox = static_cast<MultiColumnList*>(mMMulti->getChild("multi.gameslist"));
	SInfoListItem* selected = static_cast<SInfoListItem*>(mclbox->getFirstSelectedItem());
	if (selected == 0) return true;
	
	// For sure clear the old commands if any
	g_settings.set(SETTING_GAME_INIT_COMMANDS, "");

	getAppClient().runGame(selected->mSInfo);
	
	return true;
}

/* *************************************** OPtions ********************************* */

// Clear ".xml"
std::string fileToName(const std::string &filename)
{
	std::string tmp = filename;
	tmp.resize(tmp.size()-4);
	return tmp;  
}

// Add ".xml"
std::string nameToFile(const std::string &langname)
{
	return 	langname + ".xml";
}

void addKeybindRow(MultiColumnList* keylist, const CEGUI::String& text, std::string command)
{
	KeyBindings& keybind = getAppClient().getKeyBindings();
	std::string key = keybind.whatKeyForCommand(command);
	int row = keylist->addRow();
	keylist->setItem(new ListItemAndString(text,  command), 0, row);
	keylist->setItem(new ListItemAndString(key,   command), 1, row);				
}

bool MenuMain::handleToOptions(const CEGUI::EventArgs& e)
{
	try
	{
		System::getSingleton().setGUISheet(mMOptions);
		mWindow = mMOptions;
	
/*	// Do actualization of variables in options
	TabControl *tabs = static_cast<TabControl*>(mMOptions->getChild("options.tab"));
	LOGD << "Tabs: " << tabs->getTabCount() << endl;
	tabs->addTab(mMOptions->getChild("options.audio"));
	tabs->makeTabVisible("options.audio");
	*/
	// AUDIO
		Window *frame = mMOptions->getChild("options.frame.audio");
		SFX &sfx = getAppClient().getSFX();
		ISoundSystem& sound = sfx.getSoundSystem();
		LOGD << sfx.isSamplesEnabled() << " " << sfx.isMusicEnabled() << endl;
	
		(static_cast<Checkbox*>(frame->getChild("options.audio.music")))->setSelected(
			sfx.isMusicEnabled()
		);
	
		(static_cast<Checkbox*>(frame->getChild("options.audio.samples")))->setSelected(
			sfx.isSamplesEnabled()
		);
	
		(static_cast<Checkbox*>(frame->getChild("options.audio.3Dsamples")))->setSelected(
			g_settings.isTrue(SETTING_3DSOUND_ENABLE)
		);
		
		Slider *slider = static_cast<Slider*>(frame->getChild("options.audio.samplesvolume"));
		slider->setMaxValue((float)sound.getSamplesManager().getMaxVolume());
		slider->setCurrentValue((float)sound.getSamplesManager().getVolume());
	
		slider = static_cast<Slider*>(frame->getChild("options.audio.musicvolume"));
		slider->setMaxValue((float)sound.getMusic().getMaxVolume());
		slider->setCurrentValue((float)sound.getMusic().getVolume());
		
		slider = static_cast<Slider*>(frame->getChild("options.audio.globalgain"));
		slider->setMaxValue((float)1.0f);
		slider->setCurrentValue((float)g_settings.getDouble(SETTING_3DSOUND_GLOBAL_GAIN));
		
		std::vector<std::string> openALDevicesV;
		Combobox *openALdevice = static_cast<Combobox*>(frame->getChild("options.audio.openALdevice"));
		if (sfx.getOpenALDevices(openALDevicesV))
		{
			openALdevice->resetList();
			openALdevice->getEditbox()->setText(g_settings.getString(SETTING_3DSOUND_DEVICE));
			for (int i = 0; i < (int)openALDevicesV.size(); i++)
			{
				openALdevice->addItem(new ListboxTextItem(openALDevicesV[i]));
			}
		}
		else
		{
			openALdevice->setVisible(false);
			frame->getChild("options.audio.openALdevice.label")->setVisible(false);
		}
		
			// themes list
		Combobox *themes = static_cast<Combobox*>(frame->getChild("options.audio.theme"));
		themes->resetList();
		themes->getEditbox()->setText(fileToName(g_settings.getString(SETTING_SOUNDTHEME)));
		
		const std::vector<std::string>& themesVector = sfx.getThemes();
		
		for (unsigned int themeID = 0; themeID < themesVector.size(); themeID++)
		{
			themes->addItem(new ListboxTextItem(fileToName(themesVector[themeID])));
		}	
	
	
		// Graphics
		frame = mMOptions->getChild("options.frame.video");
		Combobox *videomode = static_cast<Combobox*>(frame->getChild("options.graphics.videomodes"));
		videomode->resetList();
		videomode->getEditbox()->setText(g_settings.getString(SETTING_RESOLUTION));
		videomode->addItem(new ListboxTextItem("640x480"));
		videomode->addItem(new ListboxTextItem("800x600"));
		videomode->addItem(new ListboxTextItem("1024x768"));	
		videomode->addItem(new ListboxTextItem("1280x800"));
		videomode->addItem(new ListboxTextItem("1280x900"));
		videomode->addItem(new ListboxTextItem("1280x1024"));
		videomode->addItem(new ListboxTextItem("1680x1050"));
		videomode->addItem(new ListboxTextItem("1600x1200"));


		(static_cast<Checkbox*>(frame->getChild("options.graphics.fullscreen")))->setSelected(
			g_settings.isTrue(SETTING_FULLSCREEN)
		);

		(static_cast<Checkbox*>(frame->getChild("options.graphics.showfps")))->setSelected(
			g_settings.isTrue(SETTING_SHOW_FPS)
		);

	
		// show warning if needed
		frame->getChild("options.video.restartwarning")->setVisible(!canReloadGFXSetting);
	
		// KLeybinding
	
		mMOptions->getChild("options.message")->setVisible(false);
		
		frame = mMOptions->getChild("options.frame.keybind");
		MultiColumnList* keylist = static_cast<MultiColumnList*>(frame->getChild("options.keybind.list"));
		keylist->resetList();
		addKeybindRow(keylist, (const CEGUI::utf8*)_("Fire"),            "fire");
		addKeybindRow(keylist, (const CEGUI::utf8*)_("Forward"),         "up");
		addKeybindRow(keylist, (const CEGUI::utf8*)_("Backward"),        "down");
		addKeybindRow(keylist, (const CEGUI::utf8*)_("Left"),            "left");
		addKeybindRow(keylist, (const CEGUI::utf8*)_("Right"),           "right");
		addKeybindRow(keylist, (const CEGUI::utf8*)_("Next weapon"),     "/nextweapon");
		addKeybindRow(keylist, (const CEGUI::utf8*)_("Previous weapon"), "/previousweapon");
		addKeybindRow(keylist, (const CEGUI::utf8*)_("Weapon 1"),        "/weapon 0");
		addKeybindRow(keylist, (const CEGUI::utf8*)_("Weapon 2"),        "/weapon 1");
		addKeybindRow(keylist, (const CEGUI::utf8*)_("Weapon 3"),        "/weapon 2");
		addKeybindRow(keylist, (const CEGUI::utf8*)_("Weapon 4"),        "/weapon 3");
		addKeybindRow(keylist, (const CEGUI::utf8*)_("Weapon 5"),        "/weapon 4");
		addKeybindRow(keylist, (const CEGUI::utf8*)_("Free targeting"),  "targeting");
		addKeybindRow(keylist, (const CEGUI::utf8*)_("New tank"),        "newtank");
		addKeybindRow(keylist, (const CEGUI::utf8*)_("Suicide"),         "/suicide");
		addKeybindRow(keylist, (const CEGUI::utf8*)_("Ready"),         "/ready");
		
		// Others - language
		frame = mMOptions->getChild("options.frame.other");
		
		(static_cast<Checkbox*>(frame->getChild("options.other.invertmouse")))->setSelected(
			g_settings.isTrue(SETTING_INVERT_MOUSE)
		);
		
		(static_cast<Checkbox*>(frame->getChild("options.other.autoshowingame")))->setSelected(
			g_settings.isTrue(SETTING_AUTO_INGAME_MENU)
		);
				
		Combobox *translation = static_cast<Combobox*>(frame->getChild("options.other.language"));
		translation->resetList();
		translation->getEditbox()->setText(fileToName(g_settings.getString(SETTING_TRANSLATION_FILE)));
		
		std::vector<std::string>& tVector = g_translator.getTranslations();
		
		for (unsigned int transID = 0; transID < tVector.size(); transID++)
		{
			translation->addItem(new ListboxTextItem(fileToName(tVector[transID])));
		}	
		
		
	} catch ( CEGUI::Exception &e)
	{
		LOGE << "CEGUI exception: " << e.getMessage() << endl;
		DIE("CEGUI ERROR");

	}
	
	
	return true;
}


void MenuMain::oneKey(const std::string &key_name, const std::string &command)
{
	try
	{
		Window *frame = mMOptions->getChild("options.frame.keybind");
		MultiColumnList* keylist = static_cast<MultiColumnList*>(frame->getChild("options.keybind.list"));
	
		for(int row = 0; (unsigned int)row < keylist->getRowCount(); ++row)
		{
			
			ListItemAndString* keyitem = static_cast<ListItemAndString*>(keylist->getItemAtGridReference(MCLGridRef(row, 1)));
			if (keyitem == 0) break;

			if (keyitem->mString == command)
			{
				keyitem->setText(key_name);
				break;
			}
		}
		keylist->clearAllSelections();
	} catch ( CEGUI::Exception &e)
	{
		LOGE << "CEGUI exception: " << e.getMessage() << endl;
	}		
}


bool MenuMain::handleApplyOptions(const CEGUI::EventArgs& e)
{
	assertL(mMOptions);
	// Apply all options from options window
	// AUDIO
	SFX &sfx = getAppClient().getSFX();
	
	Window *frame = mMOptions->getChild("options.frame.audio");
	sfx.enableMusic(
		(static_cast<Checkbox*>(frame->getChild("options.audio.music")))->isSelected()
	);

	sfx.enableSamples(
		(static_cast<Checkbox*>(frame->getChild("options.audio.samples")))->isSelected()
	);
	
	sfx.setMusicVolume((int)
		(static_cast<Slider*>(frame->getChild("options.audio.musicvolume")))->getCurrentValue()
	);
	
	sfx.setSamplesVolume((int)
		(static_cast<Slider*>(frame->getChild("options.audio.samplesvolume")))->getCurrentValue()
	);
	
	g_settings.set(SETTING_3DSOUND_GLOBAL_GAIN, xtos(
		(float)((static_cast<Slider*>(frame->getChild("options.audio.globalgain")))->getCurrentValue())
	));

	g_settings.set(SETTING_3DSOUND_ENABLE, xtos(
		(static_cast<Checkbox*>(frame->getChild("options.audio.3Dsamples")))->isSelected()
	));

	g_settings.set(SETTING_SOUNDTHEME, nameToFile(xtos(
		(static_cast<Combobox*>(frame->getChild("options.audio.theme")))->getEditbox()->getText()
	)));
	sfx.loadSettingTheme();
	

	Combobox *openALdevice = static_cast<Combobox*>(frame->getChild("options.audio.openALdevice"));
	if (openALdevice->isVisible())
	{
		g_settings.set(SETTING_3DSOUND_DEVICE, xtos(
			openALdevice->getEditbox()->getText()
		));	
	}
	
	
	// GRAPHICS
	frame = mMOptions->getChild("options.frame.video");
	g_settings.set(SETTING_FULLSCREEN, xtos(
		(static_cast<Checkbox*>(frame->getChild("options.graphics.fullscreen")))->isSelected()
	));

	g_settings.set(SETTING_RESOLUTION, xtos(
		(static_cast<Combobox*>(frame->getChild("options.graphics.videomodes")))->getEditbox()->getText()
	));
	
	g_settings.set(SETTING_SHOW_FPS, xtos(
		(static_cast<Checkbox*>(frame->getChild("options.graphics.showfps")))->isSelected()
	));

	// Keybinding
	frame = mMOptions->getChild("options.frame.keybind");
	KeyBindings& keybind = getAppClient().getKeyBindings();	
	MultiColumnList* keylist = static_cast<MultiColumnList*>(frame->getChild("options.keybind.list"));
	for(int row = 0; (unsigned int)row < keylist->getRowCount(); ++row)
	{
		ListItemAndString* keyitem = static_cast<ListItemAndString*>(keylist->getItemAtGridReference(MCLGridRef(row, 1)));
		if (keyitem != 0)
		{
			// unbind old value for button and bind new
			keybind.unbindButton(keyitem->getText().c_str()); 
			keybind.bind(keyitem->getText().c_str(),  true, keyitem->mString, false);
		}
	}
			
	// Others
	frame = mMOptions->getChild("options.frame.other");	
	g_settings.set(SETTING_TRANSLATION_FILE, nameToFile(xtos(
		(static_cast<Combobox*>(frame->getChild("options.other.language")))->getEditbox()->getText()
	)));
	
	g_settings.set(SETTING_INVERT_MOUSE, xtos(
		(static_cast<Checkbox*>(frame->getChild("options.other.invertmouse")))->isSelected()
	));
	
	g_settings.set(SETTING_AUTO_INGAME_MENU, xtos(
		(static_cast<Checkbox*>(frame->getChild("options.other.autoshowingame")))->isSelected()
	));
	
	getAppClient().getGFX().reloadGFXSetting();
	
	// And let's go back to main menu
	handleToMain(e);
	return true;	
}

bool MenuMain::handleChangeKeybind(const CEGUI::EventArgs& e)
{			
	
	// read data for keybind from list
	Window *frame = mMOptions->getChild("options.frame.keybind");
	MultiColumnList* keylist = static_cast<MultiColumnList*>(frame->getChild("options.keybind.list"));
	ListItemAndString* selected = static_cast<ListItemAndString*>(keylist->getFirstSelectedItem());
	if (selected == 0) return true;
	

	assertL(mOneKeyAcceptor);	
	mOneKeyAcceptor->getKey(selected->mString);
	
	return true;
}	

bool MenuMain::handleOptionsAudio(const CEGUI::EventArgs& e)
{	
	mMOptions->getChild("options.frame.audio")->setVisible(true);
	mMOptions->getChild("options.frame.video")->setVisible(false);
	mMOptions->getChild("options.frame.keybind")->setVisible(false);
	mMOptions->getChild("options.frame.other")->setVisible(false);	
	return true;
}

bool MenuMain::handleOptionsVideo(const CEGUI::EventArgs& e)
{
	mMOptions->getChild("options.frame.audio")->setVisible(false);
	mMOptions->getChild("options.frame.video")->setVisible(true);
	mMOptions->getChild("options.frame.keybind")->setVisible(false);
	mMOptions->getChild("options.frame.other")->setVisible(false);		
	return true;	
}

bool MenuMain::handleOptionsKeybind(const CEGUI::EventArgs& e)
{
	mMOptions->getChild("options.frame.audio")->setVisible(false);
	mMOptions->getChild("options.frame.video")->setVisible(false);
	mMOptions->getChild("options.frame.keybind")->setVisible(true);
	mMOptions->getChild("options.frame.other")->setVisible(false);		
	return true;
}

bool MenuMain::handleOptionsOther(const CEGUI::EventArgs& e)
{
	mMOptions->getChild("options.frame.video")->setVisible(false);
	mMOptions->getChild("options.frame.audio")->setVisible(false);
	mMOptions->getChild("options.frame.keybind")->setVisible(false);
	mMOptions->getChild("options.frame.other")->setVisible(true);		
	return true;
}


bool MenuMain::handleToSingle(const CEGUI::EventArgs& e)
{
	System::getSingleton().setGUISheet(mMSingle);
	mWindow = mMSingle;
	
	// Load maps:
	vector<string> mapV;
	listDirectory(g_settings.getString(SETTING_MAP_LIST_DIRECTORY), mapV);
	MultiColumnList* maplist = static_cast<MultiColumnList*>(mMSingle->getChild("single.maplist"));
	maplist->resetList();
	for (int mapID = 0; mapID < (int)mapV.size(); mapID++)
	{
		if (Params::isProbablyMapFile(locateFile(mapV[mapID])))
		{
			Params::sMap mapInfo;
			gParams.loadMapBasicInfo(locateFile(mapV[mapID]), mapInfo);
		
			int row = maplist->addRow();
			maplist->setItem(new ListItemAndString((const CEGUI::utf8*)_(mapInfo.name.c_str()),  mapV[mapID]), 0, row);
			maplist->setItem(new ListItemAndString(mapV[mapID],   mapV[mapID]), 1, row);	
			maplist->setItem(new ListItemAndString(xtos(mapInfo.max_players), mapV[mapID]), 2, row);	
		}
	}
		
	return true;
}

bool MenuMain::handleSingleAIPlus(const CEGUI::EventArgs& e)
{
	long value;
	if (try_get_long(mMSingle->getChild("single.aiplayers")->getText().c_str(), value))
	{
		value++;
	}
	else
	{
		value = 0;
	}
	mMSingle->getChild("single.aiplayers")->setText(xtos(value));
	
	return true;
}

bool MenuMain::handleSingleAIMinus(const CEGUI::EventArgs& e)
{
	long value;
	if (try_get_long(mMSingle->getChild("single.aiplayers")->getText().c_str(), value))
	{
		if (--value < 0) value = 0;
	}
	else
	{
		value = 0;
	}
	mMSingle->getChild("single.aiplayers")->setText(xtos(value));
	
	return true;	
}

bool MenuMain::handleSingleTimeLimitPlus(const CEGUI::EventArgs& e)
{
	long value;
	if (try_get_long(mMSingle->getChild("single.timelimit")->getText().c_str(), value))
	{
		value++;
	}
	else
	{
		value = 0;
	}

	mMSingle->getChild("single.timelimit")->setText(xtos(value));
	
	return true;
}

bool MenuMain::handleSingleTimeLimitMinus(const CEGUI::EventArgs& e)
{
	long value;
	if (try_get_long(mMSingle->getChild("single.timelimit")->getText().c_str(), value))
	{
		if (--value < 0) value = 0;
	}
	else
	{
		value = 0;
	}

	mMSingle->getChild("single.timelimit")->setText(xtos(value));
	
	return true;
}

bool MenuMain::handleSingleFragLimitPlus(const CEGUI::EventArgs& e)
{
	long value;
	if (try_get_long(mMSingle->getChild("single.fraglimit")->getText().c_str(), value))
	{
		value++;
	}
	else
	{
		value = 0;
	}

	mMSingle->getChild("single.fraglimit")->setText(xtos(value));
	
	return true;
}

bool MenuMain::handleSingleFragLimitMinus(const CEGUI::EventArgs& e)
{
	long value;
	if (try_get_long(mMSingle->getChild("single.fraglimit")->getText().c_str(), value))
	{
		if (--value < 0) value = 0;
	}
	else
	{
		value = 0;
	}

	mMSingle->getChild("single.fraglimit")->setText(xtos(value));
	
	return true;	
}


bool MenuMain::handleSinglePlay(const CEGUI::EventArgs& e)
{
// Get commands for server	
	stringstream server_cmd;
		
	long value;
	if (try_get_long(mMSingle->getChild("single.aiplayers")->getText().c_str(), value) && value > 0)
	{
		server_cmd << "/addai " << value << endl;
	}
	
	if (try_get_long(mMSingle->getChild("single.fraglimit")->getText().c_str(), value))
	{
		server_cmd << "/fraglimit " << value << endl;
	}

	if (try_get_long(mMSingle->getChild("single.timelimit")->getText().c_str(), value))
	{
		server_cmd << "/timelimit " << value << endl;
	}
	
	server_cmd << "/maxplayers 1" << endl; // Allow only one player :-)
	// add automatic shutdown limit
	server_cmd << "/maxtimenohuman " <<  g_settings.getInt(SETTING_SINGLE_TIME_WITHOUT_HUMAN) << endl;
	
	
	g_settings.set(SETTING_GAME_INIT_COMMANDS, server_cmd.str());
	
	
// Get the map	
	MultiColumnList* maplist = static_cast<MultiColumnList*>(mMSingle->getChild("single.maplist"));
	ListItemAndString* selected = static_cast<ListItemAndString*>(maplist->getFirstSelectedItem());
	if (selected == 0)
	{
		g_settings.set(SETTING_SERVER_MAP, "");		
	}
	else
	{
		g_settings.set(SETTING_SERVER_MAP, selected->mString);				
	}
	
// create passwod
	string password = xtos(rand()%1000);
	g_settings.set(SETTING_SERVER_PASSWORD, password);				
	
	
// Start server	
	g_settings.set(SETTING_SERVER_SINGLE_PLAYER, "true");
	if (!startServer()) 
	{
		getAppClient().getConsole().errorMessage(_("Starting server failed."));		
		return true;
	}
	
// Prepare commands for client
	stringstream client_cmd;
	client_cmd << "/wannabeadmin " << password << endl;
	g_settings.set(SETTING_GAME_INIT_COMMANDS, client_cmd.str());

// Connect to localhost
	getAppClient().getCommander().addCommand("connect 127.0.0.1");	
	
	return true;		
}

/* ****************************** Multi player - create server ************************** */

bool MenuMain::handleToMultiCreate(const CEGUI::EventArgs& e)
{
	System::getSingleton().setGUISheet(mMMultiCreate);
	mWindow = mMMultiCreate;
	
	// Load maps:
	vector<string> mapV;
	listDirectory(g_settings.getString(SETTING_MAP_LIST_DIRECTORY), mapV);
	MultiColumnList* maplist = static_cast<MultiColumnList*>(mMMultiCreate->getChild("multicreate.maplist"));
	maplist->resetList();
	for (int mapID = 0; mapID < (int)mapV.size(); mapID++)
	{
		if (Params::isProbablyMapFile(locateFile(mapV[mapID])))
		{
			Params::sMap mapInfo;
			gParams.loadMapBasicInfo(locateFile(mapV[mapID]), mapInfo);
		
			int row = maplist->addRow();
			maplist->setItem(new ListItemAndString((const CEGUI::utf8*)_(mapInfo.name.c_str()),  mapV[mapID]), 0, row);
			maplist->setItem(new ListItemAndString(mapV[mapID],   mapV[mapID]), 1, row);	
			maplist->setItem(new ListItemAndString(xtos(mapInfo.max_players), mapV[mapID]), 2, row);	
		}
	}
		
	
	return true;
}

bool MenuMain::handleMultiCreateAIPlus(const CEGUI::EventArgs& e)
{
	long value;
	if (try_get_long(mMMultiCreate->getChild("multicreate.aiplayers")->getText().c_str(), value))
	{
		value++;
	}
	else
	{
		value = 0;
	}
	mMMultiCreate->getChild("multicreate.aiplayers")->setText(xtos(value));
	
	return true;
}

bool MenuMain::handleMultiCreateAIMinus(const CEGUI::EventArgs& e)
{
	long value;
	if (try_get_long(mMMultiCreate->getChild("multicreate.aiplayers")->getText().c_str(), value))
	{
		if (--value < 0) value = 0;
	}
	else
	{
		value = 0;
	}
	mMMultiCreate->getChild("multicreate.aiplayers")->setText(xtos(value));
	
	return true;	
}

bool MenuMain::handleMultiCreateFragLimitPlus(const CEGUI::EventArgs& e)
{
	long value;
	if (try_get_long(mMMultiCreate->getChild("multicreate.fraglimit")->getText().c_str(), value))
	{
		value++;
	}
	else
	{
		value = 0;
	}
	mMMultiCreate->getChild("multicreate.fraglimit")->setText(xtos(value));
	
	return true;
}

bool MenuMain::handleMultiCreateFragLimitMinus(const CEGUI::EventArgs& e)
{
	long value;
	if (try_get_long(mMMultiCreate->getChild("multicreate.fraglimit")->getText().c_str(), value))
	{
		if (--value < 0) value = 0;
	}
	else
	{
		value = 0;
	}
	mMMultiCreate->getChild("multicreate.fraglimit")->setText(xtos(value));
	
	return true;	
}

bool MenuMain::handleMultiCreateTimeLimitPlus(const CEGUI::EventArgs& e)
{
	long value;
	if (try_get_long(mMMultiCreate->getChild("multicreate.timelimit")->getText().c_str(), value))
	{
		value++;
	}
	else
	{
		value = 0;
	}
	mMMultiCreate->getChild("multicreate.timelimit")->setText(xtos(value));
	
	return true;
}

bool MenuMain::handleMultiCreateTimeLimitMinus(const CEGUI::EventArgs& e)
{
	long value;
	if (try_get_long(mMMultiCreate->getChild("multicreate.timelimit")->getText().c_str(), value))
	{
		if (--value < 0) value = 0;
	}
	else
	{
		value = 0;
	}
	mMMultiCreate->getChild("multicreate.timelimit")->setText(xtos(value));
	
	return true;	
}

bool MenuMain::handleMultiCreateMaxPlayersPlus(const CEGUI::EventArgs& e)
{
	long value;
	if (try_get_long(mMMultiCreate->getChild("multicreate.maxplayers")->getText().c_str(), value))
	{
		value++;
	}
	else
	{
		value = 0;
	}
	mMMultiCreate->getChild("multicreate.maxplayers")->setText(xtos(value));
	
	return true;
}

bool MenuMain::handleMultiCreateMaxPlayersMinus(const CEGUI::EventArgs& e)
{
	long value;
	if (try_get_long(mMMultiCreate->getChild("multicreate.maxplayers")->getText().c_str(), value))
	{
		if (--value < 0) value = 0;
	}
	else
	{
		value = 0;
	}
	mMMultiCreate->getChild("multicreate.maxplayers")->setText(xtos(value));
	
	return true;	
}

	
bool MenuMain::handleCreateServer(const CEGUI::EventArgs& e)
{

// check for password
	string password = mMMultiCreate->getChild("multicreate.password")->getText().c_str();
	if (password.empty())
	{
		return true;	
	}
	g_settings.set(SETTING_SERVER_PASSWORD, password);				
	
	
// Get commands for server	
	stringstream server_cmd;
		
	long value;
	if (try_get_long(mMMultiCreate->getChild("multicreate.aiplayers")->getText().c_str(), value) && value > 0)
	{
		server_cmd << "/addai " << value << endl;
	}
	
	if (try_get_long(mMMultiCreate->getChild("multicreate.fraglimit")->getText().c_str(), value))
	{
		server_cmd << "/fraglimit " << value << endl;
	}

	if (try_get_long(mMMultiCreate->getChild("multicreate.timelimit")->getText().c_str(), value))
	{
		server_cmd << "/timelimit " << value << endl;
	}
	
	if (try_get_long(mMMultiCreate->getChild("multicreate.maxplayers")->getText().c_str(), value))
	{
		server_cmd << "/maxplayers " << value << endl;
	}
	
	// add automatic shutdown limit
	server_cmd << "/maxtimenohuman " <<  g_settings.getInt(SETTING_MULTI_TIME_WITHOUT_HUMAN) << endl;
	
	g_settings.set(SETTING_GAME_INIT_COMMANDS, server_cmd.str());
	
	
// Get the map	
	MultiColumnList* maplist = static_cast<MultiColumnList*>(mMMultiCreate->getChild("multicreate.maplist"));
	ListItemAndString* selected = static_cast<ListItemAndString*>(maplist->getFirstSelectedItem());
	if (selected == 0)
	{
		g_settings.set(SETTING_SERVER_MAP, "");		
	}
	else
	{
		g_settings.set(SETTING_SERVER_MAP, selected->mString);				
	}
		
	
// Start server	
	g_settings.set(SETTING_SERVER_SINGLE_PLAYER, "false");
	if (!startServer()) 
	{
		getAppClient().getConsole().errorMessage(_("Starting server failed."));
		return true;
	}

// Prepare commands for client
	stringstream client_cmd;
	client_cmd << "/wannabeadmin " << password << endl;
	g_settings.set(SETTING_GAME_INIT_COMMANDS, client_cmd.str());


// Connect to localhost
	getAppClient().getCommander().addCommand("connect 127.0.0.1");	
	
	
	return true;		
}


