/* MAINTAINER: Bernard */
#ifndef _MENU_MAIN_H
#define _MENU_MAIN_H

#include "menubasic.h"
#include "mutex.h"
#include "iortransfer_client.h"

namespace CEGUI
{
	class EventArgs;
};

class OneKeyAcceptor;
class GetNetworkGames;

/**
 * \brief Class holding whole main menu.
 * 
 * It takes care for all main menu window sheets and also for the functionality of menu.
 * It means response to clicking a button.
 */
class MenuMain : public MenuBasic
{
public:

	MenuMain();
	~MenuMain();

	/**
	 * \brief Do update of some menu items.
	 * 
	 * If menu is in network game selection the update takes care of
	 * updating server list and in mutiplayer server ceration window
	 * the password field is checked and if not filled, the button
	 * for starting server is disabled.
	 * \params dtime Only forwarded to ancestor.
	 * \params atime Only forwarded to ancestor.
	 */
	virtual void update(int dtime, int atime);
	
	/**
	 * \brief Every tick  mNeedNetRefresh is set to false.
	 * 
	 * It causes that next call to refreshNetGames() will
	 * refresh the net games.
	 */
	virtual void tick();

// callbacks
	bool handleQuit(const CEGUI::EventArgs& e);  ///< \brief Handle for Quit button quits the application.

	bool handleToMain(const CEGUI::EventArgs& e);     ///< \brief Change current sheet to first (main) sheet.
	bool handleToCredists(const CEGUI::EventArgs& e); ///< \brief Change current sheet to credits sheet.
	
	/**
	 * \brief Change current sheet to menu with selecting server to join.
	 * 
	 * On entering the menu, network game list is forced to refresh.
	 */
	bool handleToMulti(const CEGUI::EventArgs& e);
	
	/**
	 * \brief Callback for Connect button in sheet for connecting to server.
	 * 
	 * First the editbox is checked if there is something written in there
	 * and if so game try to connect to that ip/hostname. If editbox is empty,
	 * the game try to connect to server selected in list of servers.
	 * This callback is used also when Enter is pressed in ip/hostname editbox.
	 */
	bool handleConnect(const CEGUI::EventArgs& e);
	
	/**
	 * \brief Callback for doubleclick on list of servers.
	 * 
	 * The game try to connect to currently selected server in listbox.
	 */
	bool handleConnectFromList(const CEGUI::EventArgs& e);
	
	/**
	 * \brief Change current sheet to options menu.
	 * 
	 * Also suck all current settings from game and put them into
	 * menu windows.
	 */
	bool handleToOptions(const CEGUI::EventArgs& e);
	
	/**
	 * \brief Suck oprions from menu and put them to game.
	 */
	bool handleApplyOptions(const CEGUI::EventArgs& e);
	
	/**
	 * \brief Change key bind for currently selected line in list of
	 *        keybinds.
	 * 
	 * The keybind is changed using high priority input acceptor OneKeyAcceptor
	 * which takes one button from input and pass it back to function oneKey
	 */
	bool handleChangeKeybind(const CEGUI::EventArgs& e);
	
	bool handleOptionsAudio(const CEGUI::EventArgs& e);    ///< \brief Show tab with audio options.
	bool handleOptionsVideo(const CEGUI::EventArgs& e);    ///< \brief Show tab with video options.
	bool handleOptionsKeybind(const CEGUI::EventArgs& e);  ///< \brief Show tab with keybind options.
	bool handleOptionsOther(const CEGUI::EventArgs& e);    ///< \brief Show tab with other game options.

	bool handleToSingle(const CEGUI::EventArgs& e);  ///< \brief Change current sheet to single player menu.
	bool handleSingleAIPlus(const CEGUI::EventArgs& e);   ///< \brief Add one to AI count in single player menu.
	bool handleSingleAIMinus(const CEGUI::EventArgs& e);  ///< \brief Substract one from AI count in single player menu.
	bool handleSingleTimeLimitPlus(const CEGUI::EventArgs& e);  ///< \brief Add one to TimeLimit in single player menu.
	bool handleSingleTimeLimitMinus(const CEGUI::EventArgs& e); ///< \brief Substract one from TimeLimit in single player menu.
	bool handleSingleFragLimitPlus(const CEGUI::EventArgs& e);  ///< \brief Add one to FragLimit in single player menu.
	bool handleSingleFragLimitMinus(const CEGUI::EventArgs& e); ///< \brief Substract one from FragLimit in single player menu.
	
	/**
	 * \brief Start single player.
	 * 
	 * The function read input in menu and pack it to server's command line
	 * such that server is able to decode it. It also generate random password
	 * for server admin. 
	 * Then switch from menu to game and send to server command for this player
	 * to became admin - it may since it knows the password. This is a little 
	 * protection from stealing server.
	 */
	bool handleSinglePlay(const CEGUI::EventArgs& e);


	bool handleToMultiCreate(const CEGUI::EventArgs& e);      ///< \brief Change current sheet to multi player server starter.
	bool handleMultiCreateAIPlus(const CEGUI::EventArgs& e);  ///< \brief Add one to AI count in multi player server starter.
	bool handleMultiCreateAIMinus(const CEGUI::EventArgs& e); ///< \brief Substract one from AI count in multi player server starter.
	bool handleMultiCreateTimeLimitPlus(const CEGUI::EventArgs& e);   ///< \brief Add one to TimeLimit in multi player server starter.
	bool handleMultiCreateTimeLimitMinus(const CEGUI::EventArgs& e);  ///< \brief Substract one from TimeLimit in multi player server starter.
	bool handleMultiCreateFragLimitPlus(const CEGUI::EventArgs& e);   ///< \brief Add one to FragLimit in multi player server starter.
	bool handleMultiCreateFragLimitMinus(const CEGUI::EventArgs& e);  ///< \brief Substract one from FragLimit in multi player server starter.
	bool handleMultiCreateMaxPlayersPlus(const CEGUI::EventArgs& e);  ///< \brief Add one to max players limit in multi player server starter.
	bool handleMultiCreateMaxPlayersMinus(const CEGUI::EventArgs& e); ///< \brief Substract one from max players limit in multi player server starter.	
	
	/**
	 * \brief Start multiplayer server and connect to it.
	 * 
	 * Suck players wishes about server from menu and start server according to
	 * them. Then switch from menu to game mode and as server to grand admin 
	 * rights to our player. The player would probably ask for them anyway so let's
	 * save him some time since we know the server password.
	 */
	bool handleCreateServer(const CEGUI::EventArgs& e);

	
	/**
	 * \brief Set current info about located servers.
	 * 
	 * This function is thread safe and may be called from thread listing
	 * the servers on network.
	 * \param svinfo New vector of servers.
	 */
	void setServerInfos(const std::vector<ServerInfo> &svinfo);

protected:

	friend class GetNetworkGames;
	
	std::vector<ServerInfo> mServerInfos; ///< \brief List of discovered multiplayer servers.
	bool mNeedNetRefresh;	 ///< \brief True if it has been a long time since last server list refresh.
	Mutex mServerInfosMutex; ///< \brief Mutex for mServerInfos and mNeedNetRefresh.

	friend class OneKeyAcceptor;
	
	/**
	 * \brief Callback from mOneKeyAcceptor.
	 * 
	 * If you try to change keybinding the class mOneKeyAcceptor is activated and after
	 * receiving one button, the button is forwared back to this class and keybind
	 * is changed. Well changed is only an entry in keybind list. The keybind is really
	 * changed after clicking OK button/
	 * \param key_name Pressed key.
	 * \param command  Command for which mOneKeyAcceptor was called.
	 */
	void oneKey(const std::string &key_name, const std::string &command);

	CEGUI::Window*   mMMain;           ///< \brief Window with first menu screen.
	CEGUI::Window*   mMCredits;        ///< \brief Window with credits screen.
	CEGUI::Window*   mMMulti;          ///< \brief Window for connecting to multiplayer server.
	CEGUI::Window*   mMMultiCreate;    ///< \brief Window for creating multiplayer server.
	CEGUI::Window*   mMSingle;         ///< \brief Window for starting singleplayer game.
	CEGUI::Window*   mMOptions;        ///< \brief Window with smaller setting subwindows.
	
	OneKeyAcceptor*  mOneKeyAcceptor;  ///< \brief High priority input acceptor for changing keybinding.
	GetNetworkGames* mGetNetworkGamesThread; ///< \brief Thread for broadcast listing network servers on local network.
	
	/**
	 * \brief Check mNeedNetRefresh and create thread for listing servers if needed.
	 */
	void refreshNetGames();
	
};

#endif
