/* MAINTAINER: Bernard */
#ifndef _MENU_BASIC_H
#define _MENU_BASIC_H

#include "layer.h"
#include "task.h"
#include "frame.h"
#include "inputacceptor.h"
#include "opengl2d.h"

namespace CEGUI
{
	class Window;
};

/**
 * \brief Base class for menu.
 * 
 * Class providing all necessary support for running CEGUI menu.
 * In may be placed as a Layer into rendering task, As Task to task maneger
 * for regullar update calls, as Frame to application and run in ow loop
 * and can accept input from InputManager.
 * 
 * Derived classes basicly only need to load window layout and callback functions
 * for menu events. Maybe also do some update if they want to change something
 * regullary.
 */
class MenuBasic : public Layer, public Task, public Frame, public InputAcceptor
{
public:	

	MenuBasic();  ///< \brief Just some basic inicialization.
	virtual ~MenuBasic() = 0;

	/**
	 * \brief Full message loop showing the menu.
	 * 
	 * If you want to run menu like only the menu in a message loop,
	 * this is the loop. Does all task manager inicializations to make
	 * menu running. The function return after menu quits.
	 */
	virtual void run();

	/**
	 * \brief Render the menu if menu is activated.
	 */
	virtual void render();

	// Input aceptor callback functions
	virtual bool keyDown(const SDL_keysym &keysym);      ///< \brief Pass input to CEGUI.
	virtual bool keyUp(const SDL_keysym &keysym);        ///< \brief Pass input to CEGUI.
 	virtual bool mouseButton(int button, bool pressed);  ///< \brief Pass input to CEGUI.
	virtual bool mouseMotion(int dx, int dy);            ///< \brief Pass input to CEGUI.
	virtual bool wantKey(const SDL_keysym &keysym);  ///< \brief Input is wanted if menu is activated.

	/**
	 * \brief Send time notes to CEGUI.
	 * 
	 * CEGUI needs to get time events and this takes care of
	 * these time events.
	 */
	virtual void update(int dtime, int atime);

	/**
	 * \brief Activate the menu.
	 * 
	 * Means the meny will render and the menu starts to get input
	 * by registering to InputManager.
	 */
	virtual void activated();

	/**
	 * \brief Deactivate the menu.
	 * 
	 * The menu will no longer render and removes this from
	 * Input Manager so it also get no input.
	 */
	virtual void deactivated();

	/**
	 * \brief Get menu state.
	 * 
	 * \return True if menu is crrently activated.
	 */
	bool isActivated() { return mActivated; }

	
protected:

	/**
	 * \brief Try to translate every static text in window.
	 * 
	 * Recursivelly traverse all child windows and try to translate 
	 * text in every window. This should be called right after 
	 * loading window from file.
	 * \param window Window to translate.
	 */
	static void translateWindowTextRecursive(CEGUI::Window*  window);

	bool mActivated; ///< \brief Activity state of this menu
	
    CEGUI::Window*  mWindow; ///< \brief Currently active window in menu.

	/**
	 * \brief Texture for screen background.
	 * 
	 * If you want backgroun to menu just load this texture. It the
	 * mBackGround is 0 no background is drawn nor cleared. It is usefull
	 * for ingame menu to leave the texture 0. By tedaulf it is 0.
	 */
	TextureRef  mBackGround;
};


#endif
