#include "ac_transformation.h"
#include "model.h"
#include "hna_transform.h"

TransformationController::TransformationController( anim_source_type * animationDataSource, const std::string & nodeName )
	: mAnimationDataSource( animationDataSource ), 
	mNodeName( nodeName ), 
	mNode(0),
	mOriginalTransformationSaved( false ), 
	mOriginalNodeTM( GLMatrix4d::identity() )
{}

HierarchyNode * TransformationController::getNodeByName(class Model * model, const std::string & nodeName )
{
	class NodeFinder: public HierarchyNodeVisitor
	{
	public:
		NodeFinder( std::string targetNodeName )
			: mTargetNodeName( targetNodeName ), mFoundNode(0)
		{}

		virtual void visitHierarchyNode(class HierarchyNode * node) 
		{
			if ( node->getName() == mTargetNodeName )
			{
				mFoundNode = node;
				return;
			}
			for (HierarchyNode::ChildIterator it = node->getChildrenBegin(), ite = node->getChildrenEnd(); mFoundNode == 0 && it != ite; ++it)
				(*it)->accept(this);
		}

		HierarchyNode * getResult() { return mFoundNode; }

	private:
		std::string mTargetNodeName;
		class HierarchyNode * mFoundNode;
	} nf ( mNodeName );

	HierarchyNode * root = model->getRoot();
	if ( root )
	{
		root->accept( &nf );
		return nf.getResult();
	}

	return 0;
}

TransformationController::transformation_ptr_type TransformationController::getTransformation( class Model * model )
{
	// TODO: safety: access mNode by some private method rather than directly
	if ( !mNode )
	{
		mNode = getNodeByName( model, mNodeName );
		if (!mNode )
			return 0;
	}

	// TODO: speedup: avoid using transformation obtainer
	class TransformationObtainer: public hna::AttrVisitor
	{
	public:
		TransformationObtainer()
			: mTransformation( 0 )
		{}

		virtual void visitTransformation(class hna::Transformation * t)
		{ mTransformation = t; }
		hna::Transformation * getResult() { return mTransformation; }
	private:
		hna::Transformation * mTransformation;
	} to;

	mNode->acceptAttributeVisitor( &to );
	return to.getResult();
}

void TransformationController::saveOriginalTransformation( class Model * model )
{
	transformation_ptr_type pt = getTransformation( model );
	
	if ( !pt )
		// TODO: handle error
		return;

	mOriginalNodeTM = pt->getNodeTM();
	mOriginalTransformationSaved = true;
}

void TransformationController::update( class Model * model, unsigned int time_ms )
// modifies transformation attribute by multiplying it with value from animationDataSource
{
	if ( !isOriginalTransformationSaved() )
		saveOriginalTransformation( model );
	
	transformation_ptr_type pt = getTransformation( model );

	if ( pt )
	{
		// set new node transform matrix
		GLMatrix4d new_tm( mOriginalNodeTM );
		anim_source_type::value_type transformation_value_in_time;
		mAnimationDataSource->getValue( transformation_value_in_time, time_ms );
		new_tm *= transformation_value_in_time;
		pt->setNodeTM( new_tm );
	}
	// TODO: log: if pt was not found (is null), log a warning message 
}
