#include "GL/glew.h"
#include "color_texture_material.h"

using namespace hna;

const ColorTextureMaterial::InfoExt::Coordinate ColorTextureMaterial::InfoExt::DEFAULT_COORDINATE_VALUE = 
{
	ColorTextureMaterial::InfoExt::TILE, 
	1.0, 
	0.0,
};

const ColorTextureMaterial::InfoExt ColorTextureMaterial::InfoExt::DEFAULT_INFO_EXT_VALUE =
{
	ColorTextureMaterial::InfoExt::DEFAULT_COORDINATE_VALUE,	// s
	ColorTextureMaterial::InfoExt::DEFAULT_COORDINATE_VALUE,	// t
	true,	// mipmaps
	false,	// translucent
};

// ----
// ctor
ColorTextureMaterial::ColorTextureMaterial( std::string materialName, std::string textureFileName, int explicitMapChannel, const InfoExt &ie )
:	Material( materialName ),
	mExplicitMapChannel( explicitMapChannel ),
	mInfoExt( ie )
{
	SurfDefault surfFactoryInstance( textureFileName );
	SurfaceFactory * psf = &surfFactoryInstance;

	// -------------------
	// --- repeat mode ---
	// -------------------
	GLint repmode[2] = { GL_REPEAT, GL_REPEAT };
	const InfoExt::Coordinate *coord[2] = { &ie.s, &ie.t };
	for (int i = 0; i < 2; ++i)
	{
		switch ( coord[i]->repeatMode )
		{
		case InfoExt::CLAMP:
			psf->suff("_c");
			repmode[i] = GL_CLAMP_TO_EDGE;
			break;
		case InfoExt::MIRROR:
			if ( GLEW_VERSION_1_4 || GLEW_ARB_texture_mirrored_repeat ) 
			{
				psf->suff("_m");
				repmode[i] = GLEW_VERSION_1_4 ? GL_MIRRORED_REPEAT : GL_MIRRORED_REPEAT_ARB;
			}
			break;
		case InfoExt::TILE:
		default:
			// tiled repeat is default, without additional suffix
			break;
		};
	}

	// ----------------------
	// --- create texture ---
	// ----------------------
	// TODO: texture internal format could be specified from outside
	mTexture = TextureManager::getRef().add( psf, GL_RGBA, ie.generateMipmaps ? Texture::BUILD_MIPMAPS : 0 );

	// --------------------------------------
	// --- set texture repeat parameters ---- 
	// --------------------------------------
	// (mTexture is stil bound)
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, repmode[0] );
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, repmode[1] );
}

// ---------
// copy ctor
ColorTextureMaterial::ColorTextureMaterial( const ColorTextureMaterial & mat ):
	Material( mat ),
	mTexture( mat.mTexture ),
	mExplicitMapChannel( mat.mExplicitMapChannel ),
	mInfoExt( mat.mInfoExt )
{
}

// ----
// dtor
ColorTextureMaterial::~ColorTextureMaterial(void)
{
}

bool ColorTextureMaterial::isTranslucent()
{
	return mInfoExt.translucent;
}

void ColorTextureMaterial::InfoExt::setRepeat( CoordID c, RepeatMode mode, double tiling )
{
	switch (c)
	{
	case COORD_S:
		s.repeatMode	= mode;
		s.tiling		= tiling;
		break;
	case COORD_T:
		t.repeatMode	= mode;
		t.tiling		= tiling;
		break;
	default:
		break;
	};
}

void ColorTextureMaterial::InfoExt::setOffset( CoordID c, double offset )
{
	switch (c)
	{
	case COORD_S:
		s.offset = offset;
		break;
	case COORD_T:
		t.offset = offset;
		break;
	default:
		break;
	};
}

void ColorTextureMaterial::applyEffect()
{
	// multitexturing environmnent?
	const int textureUnitID(1);
	
	// clear binding
	clearMapChannelInfo();
	
	glColor3f( 1.0f, 1.0f, 1.0f );
	glActiveTexture( GL_TEXTURE0 + textureUnitID );
	mTexture->bind();
	glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
	
	/*
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
	glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB, GL_ADD_SIGNED);
	glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_RGB, GL_TEXTURE0 + textureUnitID );
	glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_RGB, GL_SRC_COLOR );
	glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE1_RGB, GL_PRIMARY_COLOR );
	glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND1_RGB, GL_SRC_COLOR );
	*/

	// change texture matrix
	glMatrixMode( GL_TEXTURE );
	glLoadIdentity();

	// coordinates shift and tiling transformation
	glTranslated(mInfoExt.s.offset, mInfoExt.t.offset, 0 );
	glScaled( mInfoExt.s.tiling, mInfoExt.t.tiling, 1.0 );

	// back to modelview
	glMatrixMode( GL_MODELVIEW );

	// enable texturing
	glEnable( GL_TEXTURE_2D );

	// if the material is translucent, set up the blending and alpha testing
	// alpha testing discards (almost) fully transparent fragments
	if ( isTranslucent() )
	{
		glAlphaFunc( GL_GEQUAL, 0.01f );
		glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
		glEnable( GL_BLEND );
		glEnable( GL_ALPHA_TEST );
	}

	// bind mapping channel to texture unit textureUnitID
	this->assignMapChannelInfo( mExplicitMapChannel, textureUnitID );
}

void ColorTextureMaterial::removeEffect()
{
	if ( isTranslucent() )
	{
		glDisable( GL_BLEND );
		glDisable( GL_ALPHA_TEST );
	}

	ChannelToTextureUnitIterator it, ite;
	getMapChannelInfo( it, ite );
	for ( int channelID, textureUnitID ; it != ite; ++it )
	{
		translateMapChannelInfo( it, channelID, textureUnitID );
		glActiveTexture( textureUnitID );
		
		// restore identity texture matrix
		glMatrixMode( GL_TEXTURE );
		glLoadIdentity();
		glMatrixMode( GL_MODELVIEW );

		glDisable( GL_TEXTURE_2D );
	}
}

