#include "grep_dl.h"
#include "model_loader.h"
#include "tinyxml/tinyxml.h"
#include "color_texture_material.h"
#include "platform.h"
#include "gobj_all.h"

void ModelLoader::onMaterialsElement(const TiXmlHandle &handle)
// materials are loaded only on client
{
	TiXmlElement *el = handle.FirstChildElement("material").Element();

	for ( ; el != 0; el = el->NextSiblingElement("material") )
	{
		int materialID;
		std::string materialName;
			
		// add new material
		if ( el->QueryIntAttribute( "id", &materialID ) != TIXML_SUCCESS )
		{
			// TODO: missing id attribute
			continue;
		}

		TiXmlHandle h(el);
		TiXmlText * text = h.FirstChildElement("name").FirstChild().Text();

		if ( text )
		{
			// get material name
			materialName = text->Value();
		}

		// ------------------------------
		// --- color texture material ---
		// ------------------------------
		TiXmlElement *specificMat = h.FirstChildElement("color_texture").Element();
		if ( specificMat )
		// !color_texture material!
		{
			using namespace hna;
			ColorTextureMaterial::InfoExt infoExt = ColorTextureMaterial::InfoExt::DEFAULT_INFO_EXT_VALUE;

			// -----------------
			// --- file name ---
			// -----------------
			std::string textureFileName;
			{
				const char * fname = specificMat->Attribute("file_name");
				if (fname)
					textureFileName = fname;
			}
			
			// -------------------------------
			// --- explicit map channel id ---
			// -------------------------------
			int explicitMapChannel;
			if ( specificMat->QueryIntAttribute("explicit_map_channel", &explicitMapChannel) != TIXML_SUCCESS)
			{
				// TODO: missing explicit_map_channel attribute
				continue;
			}

			// -------------------
			// --- coordinates ---
			// -------------------
			TiXmlElement * coordinates( specificMat->FirstChildElement( "coordinates" ) );
			if ( coordinates )
			{
				// this will help against copy-paste-like code ( u, v coordinates - two similar codes )
				const int HELPER_SIZE( 2 );
				struct 
				{
					const std::string 
						offsetStr, 
						repeatStr;
					hna::ColorTextureMaterial::InfoExt::RepeatMode repeatMode;
					double offsetAmount, repeatAmount;
				} helper[ HELPER_SIZE ] = 
				{
					{ "u_offset", "u_repeat", hna::ColorTextureMaterial::InfoExt::TILE, },
					{ "v_offset", "v_repeat", hna::ColorTextureMaterial::InfoExt::TILE, },
				};
				
				//std::stringstream sstream; // string-to-number conversions
				TiXmlElement * tmp;

				// --------------
				// --- offset ---
				// --------------
				for ( int i = 0; i < HELPER_SIZE; ++i )
					if ( ( tmp = coordinates->FirstChildElement( helper[i].offsetStr ) ) != 0 ) 
					{
						const char * text ( tmp->GetText() );
						if ( text ) 
						{
							std::stringstream sstream( text );
							sstream >> helper[i].offsetAmount;
						}
					}
				
				// -------------------
				// --- repetition ----
				// -------------------
				for ( int i = 0; i < HELPER_SIZE; ++i )
					if ( ( tmp = coordinates->FirstChildElement( helper[i].repeatStr ) ) != 0 ) 
					{
						// -----------------------
						// --- repetition type ---
						// -----------------------
						const char * ch ( tmp->Attribute( "type" ) );
						const std::string a( ch != 0 ? ch : "" );
						if ( a == "tile" )
							// tiled repeat
							{
								helper[i].repeatMode = hna::ColorTextureMaterial::InfoExt::TILE;
							} 
						else if ( a == "mirror" )
							// mirrored repeat
							{
								helper[i].repeatMode = hna::ColorTextureMaterial::InfoExt::MIRROR;
							}
						else if ( a == "none" )
							// clamp
							{
								helper[i].repeatMode = hna::ColorTextureMaterial::InfoExt::CLAMP;
							} 
						else 
							// not supported
							{
								// TODO: repetition mode not supported
							}
						
						// -------------------------
						// --- repetition amount ---
						// -------------------------
						
						{
							const char * text ( tmp->GetText() );
							if ( text )
							{
								std::stringstream sstream( text );
								sstream >> helper[i].repeatAmount;
							}
						}
					}
				
				infoExt.setOffset( ColorTextureMaterial::InfoExt::COORD_S, helper[0].offsetAmount );
				infoExt.setOffset( ColorTextureMaterial::InfoExt::COORD_T, helper[1].offsetAmount );
				infoExt.setRepeat( ColorTextureMaterial::InfoExt::COORD_S, helper[0].repeatMode, helper[0].repeatAmount );
				infoExt.setRepeat( ColorTextureMaterial::InfoExt::COORD_T, helper[1].repeatMode, helper[1].repeatAmount );
			}

			TiXmlElement * transparency( specificMat->FirstChildElement( "transparency" ) );
			Material::TransparencyType transpatype(Material::TYPE_OPAQUE);
			if ( transparency ) 
			{
				const char * text ( transparency->GetText() );
				if ( text )
				{
					const std::string s( text );
					transpatype = ( s == "masked" ? Material::TYPE_MASKED : ( s == "translucent" ? Material::TYPE_TRANSLUCENT : Material::TYPE_OPAQUE ) );
				}
			}
			
			// create ColorTextureMaterial using materialName, textureFileName, explicitMapChannel
			ColorTextureMaterial *mat(0);
			counted_ptr< Attribute > 
				materialRef
				(
					mat = new ColorTextureMaterial
					(
						materialName,
						locateFile(textureFileName),
						explicitMapChannel,
						infoExt
					)
				);
			
			if ( mat )
			{
				mat->setTransparencyType( transpatype );
			}

			ModelLoader::insertMaterialIntoMap(materialID, materialRef);
		}
	}
}

void ModelLoader::onTrimeshElement(const class TiXmlHandle &handle)
// trimeshes are loaded only on client
{	
	gobj::TriMesh * mesh = new gobj::TriMesh;
	insertGeometryObject( mData.gobjId, mesh );

	TiXmlElement * el = handle.FirstChildElement("channels").FirstChildElement("channel").Element();	
	for ( ; el != 0; el = el->NextSiblingElement("channel") )
	{	
		TiXmlHandle h(el);

		const char * idStr(el->Attribute("id"));
		if ( !idStr )
			return;
		
		const std::string s( idStr );
		
		gobj::TriMesh::Channel * channel;
		
		if ( s == "mesh" )
		// channel with vertices
		{
			channel = mesh->getMeshChannel();
		}
		else if ( s == "normals" )
		// channel with normals
		{
			channel = mesh->getNormalsChannel();
		}
		else
		// texture mapping channel number
		{
			int id;
			std::stringstream(s) >> id;
			channel = mesh->createTexMappingChannel( id );	
		}
		
		if ( !channel )
			return;

		// load channel data

		// vertices
		TiXmlText * text( h.FirstChildElement( "vertices" ).FirstChild().Text() );
		if ( text )
		{
			std::stringstream data( text->Value() );
			gobj::TriMesh::Channel::TCoord c;
			unsigned int i = 0;
			while ( !data.eof() )
			{
				data >> c[0] >> c[1] >> c[2];	
				channel->setCoordinate( i, c );
				++i;
			}
		}
		
		// indices
		text = h.FirstChildElement( "indices" ).FirstChild().Text();
		if ( text )
		{
			std::stringstream data( text->Value() );
			gobj::TriMesh::Channel::TFace f;
			unsigned int i = 0;
			while ( !data.eof() )
			{
				data >> f[0] >> f[1] >> f[2];
				--f[0]; --f[1]; --f[2];
				channel->setFace( i, f );
				++i;
			}
		}
	}
}

#include "graphical_representation_creator.h"

bool ModelLoader::createGraphicalRepresentation( )
{
	
	Model * m ( mModel.get() );
	if ( m )
	{
		// calculate bounding sphere in advance
		GLVector3d tmp1;
		double tmp2;
		m->getBoundingSphere(tmp1, tmp2);

		GraRepCreator c;
		return c.createGraphicalRepresentation( m );
	}
	return false;
}

