/* MAINTAINER: Petko */

#include "iortransfer_client.h"
#include "globals.h"
#include "log.h"
#include "tcpiptools.h"
#include "tinyxml/tinyxml.h"

bool parseIORResponse( const char * data, ServerInfo & sinfo )
{
	TiXmlDocument response;
	TiXmlHandle   handle(&response);
	response.Parse( data );
	if ( response.Error() )
	{
		LOGE << "Parsing IOR response failed:" << response.ErrorDesc() <<  endl;
		return false;
	}

	TiXmlElement* child = handle.FirstChild( "BJS_IOR_RESPONSE" ).FirstChild( "IOR" ).Element();
	if ( child == 0 )
	{
		child = handle.FirstChild( "ERROR" ).Element();
		if ( child == 0 )
		{
			LOGE << "Invalid server response for IOR request" << endl;
			return false;
		}
		else
		{
			TiXmlText * text = 0;

			for ( TiXmlNode * node = child->FirstChild(); node != 0; node = node->NextSibling()) 
			{
				if ( node->Type() == TiXmlNode::TEXT )
				{
					text = node->ToText();
					break;
				}
			}

			if ( text == 0 )
			{
				LOGE << "Invalid server response for IOR request" << endl;
				return false;
			}

			LOGE << "Server response on IOR request: Error: " << text->ValueStr() << endl;
			return false;
		}
	}

	TiXmlText * text = 0;

	for ( TiXmlNode * node = child->FirstChild(); node != 0; node = node->NextSibling()) 
	{
		if ( node->Type() == TiXmlNode::TEXT )
		{
			text = node->ToText();
			break;
		}
	}

	if ( text == 0 )
	{
		LOGE << "IOR tag is empty" << endl;
		return false;
	}

	if ( ( sinfo.IOR = text->ValueStr()) == "" )
	{
		LOGE << "Server returned empty IOR" << endl;
		return false;
	}

	child = handle.FirstChild( "BJS_IOR_RESPONSE" ).FirstChild( "WELCOME" ).Element();
	if ( child )
	{
		if ( child->GetText() )
			sinfo.HelloMessage = child->GetText();
	}

	child = handle.FirstChild( "BJS_IOR_RESPONSE" ).FirstChild( "MAPDESCRIPTION" ).Element();
	if ( child )
	{
		if ( child->GetText() )
			sinfo.MapDescription = child->GetText();
	}

	child = handle.FirstChild( "BJS_IOR_RESPONSE" ).FirstChild( "PLAYERCOUNT" ).Element();
	if ( child )
	{
		if ( child->GetText() )
			sinfo.PlayerCount = atoi( child->GetText() );
	}

	return true;
}


bool getServerIORByAddress(const char * address, unsigned short port, ServerInfo & sinfo )
{
	sinfo.HelloMessage = "";
	sinfo.IOR = "";
	sinfo.MapDescription = "";
	sinfo.PlayerCount = -1;

	TcpIpTools::BJS_SOCKET_TYPE sock = TcpIpTools::socket( SOCK_DGRAM );
	if ( sock == TcpIpTools::BAD_BJS_SOCKET )
	{
		LOGE << "Cannot create socket!" << endl;
		return false;
	}

	if ( TcpIpTools::selectWrite( sock, 100 ) != 1 )
	{
		LOGW << "Socket is not ready for writing for too long" << endl;
		TcpIpTools::closesocket( sock );
		return false;
	}

	std::string request = std::string("<?xml version=\"1.0\" ?><BJS_IOR_REQUEST><VERSION>") + 
							BJS_VERSION_STRING + "</VERSION></BJS_IOR_REQUEST>";

	char * dsize = new char[IOR_TRANSFER_HEADER_SIZE + 1];
	memset( dsize, ' ', IOR_TRANSFER_HEADER_SIZE );
	dsize[IOR_TRANSFER_HEADER_SIZE] = 0;

#ifdef _MSC_VER
#define snprintf _snprintf
#endif

	snprintf( dsize, IOR_TRANSFER_HEADER_SIZE, "%d", (int)request.length() );
	dsize[ strlen(dsize) ]=' ';

	request = std::string(dsize) + request;

	delete[] dsize;

	if ( TcpIpTools::sendto( sock, request.c_str(), (int)request.length(), address, port ) != (int)request.length() )
	{
		LOGE << "Did not sent complete data - probably too long" << endl;
		TcpIpTools::closesocket( sock );
		return false;
	}

	if ( TcpIpTools::selectRead( sock, 1000) != 1 )		//server 'has' 1 second for reply
	{
		LOGE << "Socket not ready for reading for too long. Maybe server didn' responded or firewall blocked packet?" << endl;
		TcpIpTools::closesocket( sock );
		return false;
	}

	char * recv_buf = new char[IOR_TRANSFER_HEADER_SIZE + IOR_TRANSFER_DATA_MAX_SIZE + 1];
	
	memset( recv_buf, 0, IOR_TRANSFER_HEADER_SIZE + IOR_TRANSFER_DATA_MAX_SIZE + 1 );
	struct sockaddr_in sa;

	int really_read = TcpIpTools::recvfrom( sock, recv_buf, IOR_TRANSFER_HEADER_SIZE + IOR_TRANSFER_DATA_MAX_SIZE, &sa );
	if ( really_read <= IOR_TRANSFER_HEADER_SIZE )
	{
		LOGW << "Bad data size - received only " << really_read << " bytes" << endl;
		TcpIpTools::closesocket( sock );
		delete[] recv_buf;
		return false;
	}

	LOGD << "received:" <<endl;
	LOGD << recv_buf << endl;

	TcpIpTools::closesocket( sock );

	recv_buf[ IOR_TRANSFER_HEADER_SIZE - 1] = 0;

	int data_size = atoi( recv_buf );
	if ( data_size <= 0 || data_size > IOR_TRANSFER_DATA_MAX_SIZE )
	{
		LOGW << "Bad header size - received " << recv_buf 
			<< " value (interpreted as: " << data_size << " )" << endl;
		delete[] recv_buf;
		return false;
	}

	if ( really_read - IOR_TRANSFER_HEADER_SIZE != data_size )
	{
		LOGW << "Invalid request - bad header data size. Header reports " 
			<< data_size << " bytes, but I've received " 
			<< (really_read - IOR_TRANSFER_HEADER_SIZE) << " bytes" << endl;
		delete[] recv_buf;
		return false;
	}

	char * data = recv_buf + IOR_TRANSFER_HEADER_SIZE;

	bool res = parseIORResponse( data, sinfo );
	sinfo.IP = address;

	delete[] recv_buf;

	return res;
}

std::vector< ServerInfo > getLANServers( unsigned short port )
{
	std::vector< ServerInfo > result;
	std::vector< ServerInfo > empty;

	TcpIpTools::BJS_SOCKET_TYPE sock = TcpIpTools::socket( SOCK_DGRAM );
	if ( sock == TcpIpTools::BAD_BJS_SOCKET )
	{
		LOGE << "Cannot create socket!" << endl;
		return empty;
	}

	//TODO: here continue!!!
	if ( TcpIpTools::setSockBroadcast( sock ) != 0 )
	{
		LOGW << "Socket could not be set to send broadcast data" << endl;
		TcpIpTools::closesocket( sock );
		return empty;
	}

	if ( TcpIpTools::selectWrite( sock, 100 ) != 1 )
	{
		LOGW << "Socket is not ready for writing for too long" << endl;
		TcpIpTools::closesocket( sock );
		return empty;
	}

	std::string request = std::string("<?xml version=\"1.0\" ?><BJS_IOR_REQUEST><VERSION>") + 
		BJS_VERSION_STRING + "</VERSION></BJS_IOR_REQUEST>";

	char * dsize = new char[IOR_TRANSFER_HEADER_SIZE + 1];
	memset( dsize, ' ', IOR_TRANSFER_HEADER_SIZE );
	dsize[IOR_TRANSFER_HEADER_SIZE] = 0;

#ifdef _MSC_VER
#define snprintf _snprintf
#endif

	snprintf( dsize, IOR_TRANSFER_HEADER_SIZE, "%d", (int)request.length() );
	dsize[ strlen(dsize) ]=' ';

	request = std::string(dsize) + request;

	delete[] dsize;

	if ( TcpIpTools::sendto( sock, request.c_str(), (int)request.length(), INADDR_BROADCAST, port ) != (int)request.length() )
	{
		LOGE << "Did not sent complete data - probably too long" << endl;
		TcpIpTools::closesocket( sock );
		return empty;
	}

	while ( TcpIpTools::selectRead( sock, 500) == 1 )		//servers 'have' 0.5 seconds for reply
	{
		ServerInfo sinfo;
		sinfo.PlayerCount = -1;

		char * recv_buf = new char[IOR_TRANSFER_HEADER_SIZE + IOR_TRANSFER_DATA_MAX_SIZE + 1];

		memset( recv_buf, 0, IOR_TRANSFER_HEADER_SIZE + IOR_TRANSFER_DATA_MAX_SIZE + 1 );
		struct sockaddr_in sa;

		int really_read = TcpIpTools::recvfrom( sock, recv_buf, IOR_TRANSFER_HEADER_SIZE + IOR_TRANSFER_DATA_MAX_SIZE, &sa );
		if ( really_read <= IOR_TRANSFER_HEADER_SIZE )
		{
			LOGW << "Bad data size - received only " << really_read << " bytes" << endl;
			delete[] recv_buf;
			continue;
		}

		LOGD << "received:" <<endl;
		LOGD << recv_buf << endl;

		recv_buf[ IOR_TRANSFER_HEADER_SIZE - 1] = 0;

		int data_size = atoi( recv_buf );
		if ( data_size <= 0 || data_size > IOR_TRANSFER_DATA_MAX_SIZE )
		{
			LOGW << "Bad header size - received " << recv_buf 
				<< " value (interpreted as: " << data_size << " )" << endl;
			delete[] recv_buf;
			continue;
		}

		if ( really_read - IOR_TRANSFER_HEADER_SIZE != data_size )
		{
			LOGW << "Invalid request - bad header data size. Header reports " 
				<< data_size << " bytes, but I have received " 
				<< (really_read - IOR_TRANSFER_HEADER_SIZE) << " bytes" << endl;
			delete[] recv_buf;
			continue;
		}

		char * data = recv_buf + IOR_TRANSFER_HEADER_SIZE;

		bool res = parseIORResponse( data, sinfo );

		hostent * remoteHost = gethostbyaddr((char *) &(sa.sin_addr), 4, AF_INET);
		if ( remoteHost )
		{
			sinfo.IP = remoteHost->h_name;
		}
		else
		{	//resolution failed -> get IP in std format
			unsigned ip = ntohl( sa.sin_addr.s_addr );
//			unsigned ip = ntohl( sa.sin_addr.s_addr ); // this is for linux
			
			ip&= 0xFFFFFFFF;		//for case int has more than 32 bits

			unsigned int b1 = ip >> 24;
			unsigned int b2 = (ip & 0x00FFFFFF )  >> 16;
			unsigned int b3 = (ip & 0x0000FFFF )  >> 8;
			unsigned int b4 = ip & 0x000000FF;

			sinfo.IP = itos( b1 ) + "." + itos( b2 ) + "." + itos( b3 ) + "." + itos( b4 );
		}

		delete[] recv_buf;

		if ( res )
			result.push_back( sinfo );
	}

	TcpIpTools::closesocket( sock );

	return result;
}
