/* MAINTAINER: Lukas */

#include "networkmanager_client.h"
#include "netobjects_client.h"
#include "log.h"
#include "application_client.h"
#include "game_client.h"
#include "corbaexceptionmacros.h"
#include "settings.h"

void NetworkManagerClient::init()
{
	TIN

	mUnpackerManager.init();
	
	mNetworkReceiver.init(this);

	// !!! ORB starts last - all initialized !!!

	try {
		const int configLength = 7;
		
		int count = configLength;
		char * config[configLength];
		
		config[0] = "nic.exe";
		// BiDirectional settings
		config[1] = "-ORBofferBiDirectionalGIOP";
		config[2] = "1";
		config[3] = "-ORBclientTransportRule";
		config[4] = "* unix,tcp,bidir";

		// trace settings
		config[5] = "-ORBtraceLevel";
		char traceBuffer[100];
		strcpy(traceBuffer, itos(g_settings.getInt(SETTING_NETWORK_CORBA_TRACE_LEVEL)).c_str());
		config[6] = traceBuffer;

		// Initialise the ORB.
		mOrb = CORBA::ORB_init(count, config); // orb reference

		assertL(! CORBA::is_nil(mOrb));

		// initialise the POA
		CORBA::Object_var tmp = mOrb->resolve_initial_references("RootPOA");
		assertL(! CORBA::is_nil(tmp));
		PortableServer::POA_var rootpoa = PortableServer::POA::_narrow(tmp);
		assertL(! CORBA::is_nil(rootpoa));
		PortableServer::POAManager_var pman = rootpoa->the_POAManager();
		assertL(! CORBA::is_nil(pman));
		pman->activate();

		// Create a POA with the Bidirectional policy
		CORBA::PolicyList pl;
		pl.length(1);
		CORBA::Any a;
		a <<= BiDirPolicy::BOTH;
		pl[0] = mOrb->create_policy(BiDirPolicy::BIDIRECTIONAL_POLICY_TYPE, a);

		mBidirPoa = rootpoa->create_POA("bidir", pman, pl);
		assertL(! CORBA::is_nil(mBidirPoa));
	}
	catch(CORBA::SystemException& e)
	{
		SYSEXCEPTION_FATALERROR("Corba start failed")
	}
	catch(CORBA::Exception& e)
	{
		EXCEPTION_FATALERROR("Corba start failed")
	}
	catch(omniORB::fatalException& e)
	{
		OMNIORB_FATAL_EXCEPTION("Corba start failed")
	}
	catch(...)
	{
		UNKNOWN_ERROR("Corba start failed")
	}
}

void NetworkManagerClient::clean()
{
	// !!! ORB stops first - no more requests !!!

	TIN

	if(connected())
		disconnect();

	if(g_settings.getBool(SETTING_NETWORK_CORBA_SHUTDOWN))
	{
		try
		{
			// TODO LUKAS BUG - dotestovat
			// kdyz se tady odmitne odchozi spojeni tak zustane viset...
			// a pri ukoncovani to trva pekne dlouho - co s tim nevim
			// a obcas jako premie to padne :(
			mOrb->shutdown(false);
			mOrb->destroy();
		}
		catch(CORBA::SystemException& e)
		{
			SYSEXCEPTION_FATALERROR("Corba shutdown failed")
		}
		catch(CORBA::Exception& e)
		{
			EXCEPTION_FATALERROR("Corba shutdown failed")
		}
		catch(omniORB::fatalException& e)
		{
			OMNIORB_FATAL_EXCEPTION("Corba shutdown failed")	
		}
		catch(...)
		{
			UNKNOWN_ERROR("Corba shutdown failed")
		}
	}

	mNetworkReceiver.clean();

	mUnpackerManager.clean();
}
AsyncConnect::resultCPtr NetworkManagerClient::connect(const std::string serverIOR, const string playerName)
{
	TIN

	if(connected())
		disconnect();

	mSharedVars.lock();

	mPlayerID = INVALID_PLAYER_ID;

	mSharedVars.unlock();

	Connector_var connector;
	Client_var clientRef;

	// call prepare
	try
	{
		// create our 'ClientSide' object
		Client_i * client = new Client_i();

		// Register a 'ClientSide' object in this process.
		PortableServer::ObjectId_var oid = mBidirPoa->activate_object(client);

		// my object reference
		clientRef = client->_this();

		assertL(! CORBA::is_nil(clientRef));

		mNetworkReceiver.setNetworkObjects(client, mBidirPoa);

		CORBA::Object_var obj = mOrb->string_to_object(serverIOR.c_str());
		
		// connector reference
		connector = Connector::_narrow(obj);
		
	}
	catch(CORBA::SystemException& e)
	{
		SYSEXCEPTION_ERROR("Can't prepare connect")
	}
	catch(CORBA::Exception& e)
	{
		EXCEPTION_ERROR("Can't prepare connect")
	}
	catch(omniORB::fatalException& e)
	{
		OMNIORB_FATAL_EXCEPTION("Can't prepare connect")
	}
	catch(...)
	{
		UNKNOWN_ERROR("Can't prepare connect")
	}

	// call
	return mAsyncConnect.invoke(connector, playerName, clientRef);
}

void NetworkManagerClient::disconnect()
{
	TIN

	mSharedVars.lock();

	if(! CORBA::is_nil(mServerRef))
	{

		try
		{
			mServerRef->disconnect();
		}
		catch(omniORB::fatalException& e)
		{
			OMNIORB_FATAL_EXCEPTION("Can't disconnect")
		}
		catch(...)
		{
			// nothing to report
		}
	}

	// no more inputs from network
	mNetworkReceiver.deactivate();

	mServerRef = Server::_nil();

	mPlayerID = INVALID_PLAYER_ID;

	mSharedVars.unlock();

	// TODO LUKAS konstanta klient-side disconnect
	// TODO LUKAS disconnect handler
	disconnectedCallback(10, "Client calls disconnect from server.");
}

void NetworkManagerClient::sendCommand(string command)
{
	TIN

	mSharedVars.lock();

	if(! CORBA::is_nil(mServerRef))
	{
		bool connectionProblem = false;

		try
		{
			mServerRef->sendCommand(command.c_str());
		}
		catch(CORBA::SystemException& e)
		{
			SYSEXCEPTION_ERROR("Can't send command")

			connectionProblem = true;
		}
		catch(CORBA::Exception& e)
		{
			EXCEPTION_ERROR("Can't send command")

			connectionProblem = true;
		}
		catch(omniORB::fatalException& e)
		{
			OMNIORB_FATAL_EXCEPTION("Can't send command")
		}
		catch(...)
		{
			UNKNOWN_ERROR("Can't send command")
		}

		if(connectionProblem)
		{
			LOGI << "Connection problem." << std::endl;

			disconnect();
		}
	}

	mSharedVars.unlock();
}

void NetworkManagerClient::sendInputButton(int button, bool value, double yaw, double pitch)
{
	TIN

	mSharedVars.lock();

	if(! CORBA::is_nil(mServerRef))
	{
		bool connectionProblem = false;

		try
		{
			mServerRef->sendInputButton(button, value, yaw, pitch);
		}
		catch(CORBA::SystemException& e)
		{
			SYSEXCEPTION_ERROR("Can't send input")

			connectionProblem = true;
		}
		catch(CORBA::Exception& e)
		{
			EXCEPTION_ERROR("Can't send input")

			connectionProblem = true;
		}
		catch(omniORB::fatalException& e)
		{
			OMNIORB_FATAL_EXCEPTION("Can't send input")
		}
		catch(...)
		{
			UNKNOWN_ERROR("Can't send input")
		}

		if(connectionProblem)
		{
			LOGI << "Connection problem." << std::endl;

			disconnect();
		}
	}

	mSharedVars.unlock();
}

void NetworkManagerClient::sendYawPitch(double yaw, double pitch)
{
	TIN

	mSharedVars.lock();

	if(! CORBA::is_nil(mServerRef))
	{
		bool connectionProblem = false;

		try
		{
			mServerRef->sendYawPitch(yaw, pitch);
		}
		catch(CORBA::SystemException& e)
		{
			SYSEXCEPTION_ERROR("Can't send yaw/pitch")

			connectionProblem = true;
		}
		catch(CORBA::Exception& e)
		{
			EXCEPTION_ERROR("Can't send yaw/pitch")

			connectionProblem = true;
		}
		catch(omniORB::fatalException& e)
		{
			OMNIORB_FATAL_EXCEPTION("Can't send yaw/pitch")
		}
		catch(...)
		{
			UNKNOWN_ERROR("Can't send yaw/pitch")
		}

		if(connectionProblem)
		{
			LOGI << "Connection problem." << std::endl;

			disconnect();
		}

	}

	mSharedVars.unlock();
}

void NetworkManagerClient::connectedCallback(string msg)
{
	// !!! Called from corba !!!

	TIN

	if(connected())
	{
		LOGI << "Client was connected to server." << std::endl;
	}
	else
	{
		LOGI << "Client wasn't connected to server with message: " << msg << std::endl;
	}

	// TODO LUKAS predelat na handler
}


void NetworkManagerClient::disconnectedCallback(int reason, std::string message)
{
	// !!! Called from corba !!!

	TIN

	LOGI << "Client was disconnected from server." << std::endl;

	// TODO LUKAS predelat na handler
}

AsyncGameInfo::resultCPtr NetworkManagerClient::getGameInfo()
{
	return mAsyncGameInfo.invoke(mServerRef);
}
