/* MAINTAINER: Lukas */

#ifndef _NETWORKMANAGER_CLIENT_H
#define _NETWORKMANAGER_CLIENT_H

#include <string>
#include <list>
#include "netobjects_client.h"
#include "mutex.h"
#include "unpackermanager.h"
#include "networkreceiver_client.h"
#include "globals.h"
#include "game.h"
#include "asyncconnect.h"
#include "asyncgameinfo.h"

/**
 * Manages all network related stuff
 */
class NetworkManagerClient
{

// TODO LUKAS odstranit az budou joby
friend class AsyncConnect;

friend class NetworkReceiverClient;
friend class Client_i; // network corba object - network receiver access

public:
	NetworkManagerClient() { mPlayerID = INVALID_PLAYER_ID; };
	virtual ~NetworkManagerClient() { };

	/**
	 * Initializes ORB and UnpackerManager
	 */
	void init();

	/**
	 * Cleanups ORB and UnpackerManager
	 */
	void clean();
	
	// TODO LUKAS callback bud zavola corbacke vlakno s vysledkem nebo zavola job(send vlakno) s chybou
	/**
	 * Connects to given server
	 *
	 * \param serverIOR server IOR
	 * \param playerName player's name(nick)
	 *
	 * \return AsyncResult class
	 */
	AsyncConnect::resultCPtr connect(const std::string serverIOR, const string playerName);

	// TODO LUKAS ohlidat disconnect - pokud se zavola tak nam uz nedojdou odpovedi
	// v disconnectu by se meli vsechny pozadavky vyridit errorem
	/**
	 * Disconnects player from server
	 */
	void disconnect();

	/**
	 * Return player's ID
	 *
	 * \return valid player's ID otherwise INVALID_PLAYER_ID
	 */
	int getPlayerID()
	{
		mSharedVars.lock();

		int id = mPlayerID;

		mSharedVars.unlock();

		return id;
	}

	/**
	 * Determines whether is client connected to server
	 *
	 * \return true if client is connected to server otherwise false
	 */
	bool connected()
	{
		mSharedVars.lock();

		int id = mPlayerID;

		mSharedVars.unlock();

		return id != INVALID_PLAYER_ID;
	}

	/**
	 * Sends command to all players
	 *
	 * \param command string that contains command
	 */
	void sendCommand(string command);

	/**
	 * Send keyboard input + cannon position
	 *
	 * \param button affected button
	 * \param value button value
	 * \param yaw cannon yaw
	 * \param pitch cannon pitch
	 */
	void sendInputButton(int button, bool value, double yaw, double pitch);
	
	/**
	 * Send cannon position
	 *
	 * \param yaw cannon yaw
	 * \param pitch cannon pitch
	 */
	void sendYawPitch(double yaw, double pitch);

	/**
	 * Get game info from server. Asynchronous call.
	 *
	 * \return AsyncResult class
	 */
	AsyncGameInfo::resultCPtr getGameInfo();

	/**
	 * Returns UnpackerManager
	 *
	 * \return UnpackerManager
	 */
	UnpackerManager & getUnpackerManager() { return mUnpackerManager; };

protected:
	// TODO LUKAS predelat na handler(+ pridat duvod(vyctovy typ) a message)
	/**
	 * function is called if client was disconnected from server
	 */
	virtual void disconnectedCallback(int reason, std::string message);
	
	// TODO LUKAS predelat na handler
	/**
	 * function is called if client was connected to server
	 *
	 * \param msg additional message for client
	 */
	virtual void connectedCallback(std::string msg);

	NetworkReceiverClient & getNetworkReceiver() { return mNetworkReceiver; };

private:
	// no copy
	NetworkManagerClient(NetworkManagerClient &) { };

	// --- network receiver interface ---

	/**
	 * Lock shared vars
	 */
	void sharedVarsLock()
	{
		mSharedVars.lock();
	}

	/**
	 * Unlock shared vars
	 */
	void sharedVarsUnlock()
	{
		mSharedVars.unlock();
	}

	/**
	 * Set player id
	 *
	 * This function isn't synchronized use sharedVarsLock() before working with shared vars and
	 * sharedVarsUnlock() when you are done
	 */
	void setPlayerID(int plID)
	{
		mPlayerID = plID;
	}

	/**
	 * Set server reference
	 *
	 * This function isn't synchronized use sharedVarsLock() before working with shared vars and
	 * sharedVarsUnlock() when you are done
	 */
	void setServerRef(Server_var serverRef)
	{
		mServerRef = serverRef;
	}

	// --- network receiver interface ---

	/**
	 * UnpackerManager
	 */
	UnpackerManager mUnpackerManager;

	/**
	 * Pointer to ORB that handles communication
	 */
	CORBA::ORB_var mOrb;

	// --- this vars can set corba thread we must lock them ---

	Mutex mSharedVars;

	/**
	 * Player's ID (Shared var)
	 */
	int mPlayerID;

	/**
	 * Pointer to object that represents server (Shared var)
	 */
	Server_var mServerRef;

	/**
	 * Bidir Poa
	 */
	PortableServer::POA_var mBidirPoa;

	/**
	 * Network reciever
	 */
	NetworkReceiverClient mNetworkReceiver;

	/**
	 * Async function
	 */
	AsyncConnect mAsyncConnect;
	
	/**
	 * Async function
	 */
	AsyncGameInfo mAsyncGameInfo;
};

#endif // _NETWORKMANAGER_CLIENT_H
