/* MAINTAINER: Lukas */

#ifndef _UNPACKERS_H
#define _UNPACKERS_H

#include "unpackerworker.h"
#include "events_client.h"

/**
 * Helper template
 *
 * Contains unpack method which simply creates event from network data
 */
template<class TEvent>
class HelperUnpackerWorker : public UnpackerWorker
{
protected:
	HelperUnpackerWorker() { };
	HelperUnpackerWorker(int priority) : UnpackerWorker(priority) { };

	/**
	 * Simply creates event from network data
	 *
	 * \param data network data
	 */
	template <class TData>
	void helperUnpack(TData & data);
};

/**
 * Unpacker for basic HistoryPoint
 */
class BasicHistoryUnpacker : public HelperUnpackerWorker<EvBasicWorldEntityChanged>
{
public:
	BasicHistoryUnpacker() { };
	virtual ~BasicHistoryUnpacker() { };

	/**
	 * Unpack history point that is stored in gamestate
	 *
	 * Event is created for each packed HistoryPoint
	 *
	 * \param gameState net data that will be unpacked
	 */
	virtual void unpack(const NETGameState & gameState);
};

/**
 * Unpacker for tank HistoryPoint
 */
class TankHistoryUnpacker : public HelperUnpackerWorker<EvTankWorldEntityChanged>
{
public:
	TankHistoryUnpacker() { };
	virtual ~TankHistoryUnpacker() { };

	/**
	 * Unpack history point that is stored in gamestate
	 *
	 * Event is created for each packed tank
	 *
	 * \param gameState net data that will be unpacked
	 */
	virtual void unpack(const NETGameState & gameState);
};

/**
 * Unpacker for commands
 */
class CommandUnpacker : public UnpackerWorker
{
public:
	CommandUnpacker() { };
	virtual ~CommandUnpacker() { };

	/**
	 * Unpack command that is stored in gamestate
	 *
	 * Event is created for each packed command
	 *
	 * \param gameState net data that will be unpacked
	 */
	virtual void unpack(const NETGameState & gameState);
};

/**
 * Unpacker for collisions
 */
class CollisionUnpacker : public HelperUnpackerWorker<EvNewCollision>
{
public:
	CollisionUnpacker() { };
	virtual ~CollisionUnpacker() { };

	/**
	 * Unpack collision that is stored in gamestate
	 *
	 * Event is created for each packed collision
	 *
	 * \param gameState net data that will be unpacked
	 */
	virtual void unpack(const NETGameState & gameState);
};

/**
 * Unpacker for tank creation data
 */
class NewTankUnpacker : public HelperUnpackerWorker<EvNewTank>
{
public:
	// set priority to 90 - this unpacker runs sooner then default(second new unpack)
	NewTankUnpacker():HelperUnpackerWorker<EvNewTank>(90) { };
	virtual ~NewTankUnpacker() { };

	/**
	 * Unpack tank creation data that is stored in gamestate
	 *
	 * Event is created for each packed tank creation data
	 *
	 * \param gameState net data that will be unpacked
	 */
	virtual void unpack(const NETGameState & gameState);
};

/**
 * Unpacker for projectile creation data
 */
class NewProjectileUnpacker : public HelperUnpackerWorker<EvNewProjectile>
{
public:
	// set priority to 100 - this unpacker runs sooner then default
	NewProjectileUnpacker():HelperUnpackerWorker<EvNewProjectile>(100) { };
	virtual ~NewProjectileUnpacker() { };

	/**
	 * Unpack projectile creation data that is stored in gamestate
	 *
	 * Event is created for each packed projectile creation data
	 *
	 * \param gameState net data that will be unpacked
	 */
	virtual void unpack(const NETGameState & gameState);
};

/**
 * Unpacker for bonus box creation data
 */
class NewBonusBoxUnpacker : public HelperUnpackerWorker<EvNewBonusBox>
{
public:
	// set priority to 100 - this unpacker runs sooner then default
	NewBonusBoxUnpacker():HelperUnpackerWorker<EvNewBonusBox>(100) { };
	virtual ~NewBonusBoxUnpacker() { };

	/**
	 * Unpack bonus box creation data that is stored in gamestate
	 *
	 * Event is created for each packed bonus box creation data
	 *
	 * \param gameState net data that will be unpacked
	 */
	virtual void unpack(const NETGameState & gameState);
};

/**
 * Unpacker for deleted objects
 */
class DeletedObjUnpacker : public HelperUnpackerWorker<EvDeletedObj>
{
public:
	// set priority to 290 - this unpacker runs before last
	DeletedObjUnpacker():HelperUnpackerWorker<EvDeletedObj>(290) { };
	virtual ~DeletedObjUnpacker() { };

	/**
	 * Unpack deleted objects that is stored in gamestate
	 *
	 * Event is created for each packed deleted object
	 *
	 * \param gameState net data that will be unpacked
	 */
	virtual void unpack(const NETGameState & gameState);
};

/**
 * Unpacker for player
 */
class PlayerUnpacker : public HelperUnpackerWorker<EvPlayerChanged>
{
public:
	PlayerUnpacker() { };
	virtual ~PlayerUnpacker() { };

	/**
	 * Unpack player that is stored in gamestate
	 *
	 * Event is created for each packed player
	 *
	 * \param gameState net data that will be unpacked
	 */
	virtual void unpack(const NETGameState & gameState);
};

/**
 * Unpacker for player creation data
 */
class NewPlayerUnpacker : public HelperUnpackerWorker<EvNewPlayer>
{
public:
	// set priority to 80 - this unpacker runs sooner then default(first new unpack)
	NewPlayerUnpacker():HelperUnpackerWorker<EvNewPlayer>(80) { };
	virtual ~NewPlayerUnpacker() { };

	/**
	 * Unpack player creation data that is stored in gamestate
	 *
	 * Event is created for each packed player creation data
	 *
	 * \param gameState net data that will be unpacked
	 */
	virtual void unpack(const NETGameState & gameState);
};

/**
 * Unpacker for deleted players
 */
class DeletedPlayerUnpacker : public HelperUnpackerWorker<EvDeletedPlayer>
{
public:
	// set priority to 300 - this unpacker runs later then default
	DeletedPlayerUnpacker():HelperUnpackerWorker<EvDeletedPlayer>(300) { };
	virtual ~DeletedPlayerUnpacker() { };

	/**
	 * Unpack deleted objects that is stored in gamestate
	 *
	 * Event is created for each packed deleted object
	 *
	 * \param gameState net data that will be unpacked
	 */
	virtual void unpack(const NETGameState & gameState);
};

/**
 * Unpacker for state of game
 */
class StateOfGameUnpacker : public UnpackerWorker
{
public:
	StateOfGameUnpacker() { };
	virtual ~StateOfGameUnpacker() { };

	/**
	 * Unpack state of game that is stored in gamestate
	 *
	 * Event is created for each state of game that is different then NETNoChange
	 *
	 * \param gameState net data that will be unpacked
	 */
	virtual void unpack(const NETGameState & gameState);
};

/**
 * Unpacker for condition statuses
 */
class ConditionStatusUnpacker : public UnpackerWorker
{
public:
	ConditionStatusUnpacker() { };
	virtual ~ConditionStatusUnpacker() { };

	/**
	 * Unpack condition statuses that is stored in gamestate
	 *
	 * Event is created for each packed condition status
	 *
	 * \param gameState net data that will be unpacked
	 */
	virtual void unpack(const NETGameState & gameState);
};

// HERE PLACE YOU UNPACKER CLASS

#endif // _UNPACKERS_H
