/* MAINTAINER: Bernard */

//#define WANT_TRACE
#include "log.h"
#include "observer.h"
#include "application_client.h"
#include "game_client.h"
#include "terrain.h"
#include "world.h"
#include "gfx.h"
#include "inputmanager.h"
#include <algorithm>
#include <cmath>
#include "platform.h"
#include "settings.h"
#include "gamegui.h"

using namespace std;

Observer::Observer()
{
	mPriority = IA_PRIORITY_OBSERVER;
	mActive = false;

	mTickDelta = 20;
	mSpeed = 0;

	mUpVector = GLVector3d(0,0,1);
	mPosition = GLVector3d(0, 0, 60);
	mReferencePoint = GLVector3d(128, 128, 60);
}


bool Observer::mouseMotion(int dx, int dy)
{
	if (!mActive) return false;	

	TIN;

	int x, y;
	SDL_GetMouseState(&x, &y);

	int width  = getAppClient().getGFX().getWidth();
	int height = getAppClient().getGFX().getHeight();
	int w2     = width/2;
	int h2     = height/2;
	dx = x - w2;
	dy = y - h2;
	

	double sensitivity = 0.02;

	mYaw   -= dx * sensitivity;
	if (g_settings.isTrue(SETTING_INVERT_MOUSE))
	{
		mPitch -= dy * sensitivity * 2;
	}
	else
	{
		mPitch += dy * sensitivity;
	}
	
	if (mPitch  < -89) mPitch = -89;
	if (mPitch  >  89) mPitch =  89;

	
	if (x != w2 || y != h2) SDL_WarpMouse(w2, h2);
	SDL_GetMouseState(&x, &y);
	if (x != w2 || y != h2) SDL_WarpMouse(w2, h2);
	
	return true;
}


void Observer::activate()
{
	if (mActive) return;

	TIN;

	Camera *cam = getAppClient().getGame().getCamera();
	if (cam)
	{
		cam->getPos(mPosition);
		
		// And compute yaw and pitch
		GLVector3d refPoint;
		cam->getRef(refPoint);
		
		refPoint -= mPosition;
		refPoint.normalize();
		
		mYaw   = atan2(refPoint.y, refPoint.x)*180.0/M_PI;
		mPitch = asin(-refPoint.z) * 180.0/M_PI;
		
		if (isnan(mYaw) || isnan(mPitch))
		{
			mYaw   = 0;
			mPitch = 0;
			LOGW << "isnan in camera encountered" << endl;
		}
	}
	else
	{
		mPosition.x = 100;
		mPosition.y = 100;
		mPosition.z = 250;
		
		mYaw   = 0;
		mPitch = 0;
	}

	mActive = true;

	mSpeed = 0;

	// move the mouse only if the ingame menu is not active
	if (!getAppClient().getGame().getGUI().isMenuActive())
	{
		int w2 = getAppClient().getGFX().getWidth()/2;
		int h2 = getAppClient().getGFX().getHeight()/2;
		SDL_WarpMouse(w2, h2);
		SDL_WarpMouse(w2, h2);
	}

	getAppClient().getInputManager().addInputAcceptor(*this);

	tick();
	setGFXPerspective();
}

void Observer::deactivate()
{
	if (!mActive) return;

	TIN;

	mActive = false;
	getAppClient().getInputManager().removeInputAcceptor(*this);
}

bool Observer::isActive()
{
	TIN; 

	return mActive;
}

inline double getZ(double x, double y, Terrain &t)
{
	if (x < 0) x = 0;
	if (y < 0) y = 0;
	if (x > t.getScale().x) x = t.getScale().x;
	if (y > t.getScale().y) y = t.getScale().y;

	return t.getHeight(x,y);
}

void Observer::tick()
{
	// Don;t allow observer mode if player has tank
	if (getAppClient().getGame().getClientPlayer() &&
	    getAppClient().getGame().getClientPlayer()->getTank())
	{
		deactivate();
	}
	
	if (!mActive) return;

	TIN;

	GLQuaterniond yaw(GLVector3d(0, 0, 1), mYaw);
	GLQuaterniond pitch(GLVector3d(0, 1, 0), mPitch);
	GLQuaterniond rot = yaw * pitch;	

	GLVector3d direction = rot.rotateVector(GLVector3d(1,0,0));
	if (direction.lengthSqr() == 0) return;
	direction.normalize();

	mPosition += direction*mSpeed;

	// Do some constraints - but only in release mode
#ifndef _DEBUG	
	// Test if mPosition is above terrain
	Terrain &t = getAppClient().getGame().getWorld().getLandscape().getTerrain();

	// Round mPosition to the map - else we have problems with height map
	t.cropToValidXY(mPosition);

	// Compute Z
	GLVector3d tmp = mPosition + direction*4;
	double x = mPosition.x;
	double y = mPosition.y;
	double h = max(max(getZ(x, y, t), getZ(tmp.x, tmp.y, t)),
				   max(max(getZ(x+3, y, t), getZ(x-3, y, t)),
				       max(getZ(x, y+3, t), getZ(x, y-3, t))));
	if (mPosition.z < h + 2) mPosition.z = h + 2;

#endif

	mReferencePoint = mPosition + direction*10;
	mUpVector = rot.rotateVector(GLVector3d(0,0,1));

}

void Observer::setSpeed(double speed)
{
	mSpeed = speed;
}
