#include "render_queue.h"
#include "graphical_rep.h"

#include <algorithm>

RenderQueue::RenderQueue( size_type initialSize )
{
	setPacketSorting( true );
	assertL( INITIAL_SIZE >= 1 && "Initial size has to be at least one!" );
	mOpaque.reserve( INITIAL_SIZE );
	mMasked.reserve( INITIAL_SIZE );
	mTranslucent.reserve( INITIAL_SIZE );
}

RenderQueue::TransparencyHint RenderQueue::getTransparencyHint( const RenderPacket & p )
{
	TransparencyHint hint(HINT_OPAQUE);
	// TODO: transparency detection
	if ( p.mat )
	{
		switch( p.mat->getTransparencyType() )
		{
		case hna::Material::TYPE_MASKED:
			hint = HINT_MASKED;
			break;
		case hna::Material::TYPE_TRANSLUCENT:
			hint = HINT_TRANSLUCENT;
			break;
		default:
			break;
		}
	}
	return hint;
}

RenderQueue::container_type & RenderQueue::getBufferByHint( const TransparencyHint hint )
{
	switch (hint)
	{
	case HINT_TRANSLUCENT:
		return mTranslucent;
	case HINT_MASKED:
		return mMasked;
	case HINT_OPAQUE:
	default:
		return mOpaque;
	};
}

bool RenderQueue::insertPacket( const RenderPacket & p )
{
	getBufferByHint( getTransparencyHint( p ) ).push_back( p );
	return true;
}

void RenderQueue::renderOpaque()
{
	if ( !mOpaque.empty() )
	{		
		glColor3f(1, 1, 1);
		hna::Material * currentMaterial( 0 ); // start as null
		GLenum currentFrontFace( GL_CCW );

		for ( container_type::iterator it( mOpaque.begin() ), ite( mOpaque.end() ); it < ite; ++it )
		{
			RenderPacket & p = *it;
			gobj::GeometryObjectRef gobj = p.gobj;
			const gobj::GraphicalRepresentation * grep = gobj.isValid() ? gobj->getGraphicalRepresentation() : 0;
			if ( grep && grep->isValid() )
			{				
				if ( p.mat != currentMaterial )
				// material changed
				{
					if ( currentMaterial )
						currentMaterial->removeEffect();

					currentMaterial = p.mat;

					if ( currentMaterial )
						// note: this should be always true when strictly sorted
						currentMaterial->applyEffect();
				}

				if ( currentFrontFace != p.frontFaceOrientation )
				{
					currentFrontFace = p.frontFaceOrientation;
					glFrontFace( currentFrontFace );
				}

				glPushMatrix();
					p.toWorld.glMultMatrix();
					grep->render();
				glPopMatrix();
			}
		}
		if ( currentMaterial )
			currentMaterial->removeEffect();
		if ( currentFrontFace != GL_CCW )
			glFrontFace( GL_CCW );
	}
}

void RenderQueue::renderMasked()
{
	if ( !mOpaque.empty() )
	{
		hna::Material * currentMaterial( 0 ); // start as null
		GLenum currentFrontFace( GL_CCW );

		glDisable(GL_CULL_FACE);
		glAlphaFunc( GL_GREATER, 0.5 );
		glEnable( GL_ALPHA_TEST );

		glColor3f(1, 1, 1);

		for ( container_type::iterator it( mMasked.begin() ), ite( mMasked.end() ); it < ite; ++it )
		{
			RenderPacket & p = *it;
			gobj::GeometryObjectRef gobj = p.gobj;
			const gobj::GraphicalRepresentation * grep = gobj.isValid() ? gobj->getGraphicalRepresentation() : 0;
			if ( grep && grep->isValid() )
			{
				if ( p.mat != currentMaterial )
				// material changed
				{
					if ( currentMaterial )
						currentMaterial->removeEffect();

					currentMaterial = p.mat;

					if ( currentMaterial )
						// note: this should be always true when strictly sorted
						currentMaterial->applyEffect();
				}

				if ( currentFrontFace != p.frontFaceOrientation )
				{
					currentFrontFace = p.frontFaceOrientation;
					glFrontFace( currentFrontFace );
				}

				glFrontFace( p.frontFaceOrientation );
				glPushMatrix();
					p.toWorld.glMultMatrix();
					grep->render();
				glPopMatrix();
			}
		}

		if ( currentMaterial )
			currentMaterial->removeEffect();
		if ( currentFrontFace != GL_CCW )
			glFrontFace( GL_CCW );

		glDisable( GL_ALPHA_TEST );
		glEnable(GL_CULL_FACE);
	}
}

void RenderQueue::renderTranslucent()
{
	// TODO: render-pass
	if ( mTranslucent.size() > 0 )
	{
		DIE("should never ocure! Something is wrong");
	}
}

void RenderQueue::render()
{
	glEnable( GL_NORMALIZE );
	glFrontFace( GL_CCW );

	renderOpaque();
	renderMasked();
	renderTranslucent();
	glDisable( GL_NORMALIZE );

	glFrontFace( GL_CCW );
}

bool RenderQueue::isGreater( const RenderPacket & left, const RenderPacket & right )
{
	return (left.mat > right.mat);
}

void RenderQueue::sort()
{
	if ( isPackedSortingEnabled() )
	{
		// TODO: this is potentially slow...
		std::sort( mOpaque.begin(), mOpaque.end(), isGreater );
		std::sort( mMasked.begin(), mMasked.end(), isGreater );
		std::sort( mTranslucent.begin(), mTranslucent.end(), isGreater );
	}
}

void RenderQueue::clear()
{
	mOpaque.clear();
	mMasked.clear();
	mTranslucent.clear();
}
