/* MAINTAINER: Martin */
#include "GL/glew.h"
#include "skybox_renderer.h"
#include "grutils.h"
#include "globals_client.h"
#include "application_client.h"
#include "gfx.h"

SkyboxRenderer::SkyboxRenderer()
: mSides( SIDES_COUNT , TextureRef(0) )
{}

void SkyboxRenderer::assignTexture( const BoxSide side, TextureRef texture )
{
	mSides[ getSideIndex( side ) ] = texture;
}

int SkyboxRenderer::getSideIndex( const BoxSide side ) const
{
	assertL( (int)side >= 0 && (int)side <= (int)mSides.size() && "Wrong side!");
	return (int)side;
}

TextureRef SkyboxRenderer::getTexture( const BoxSide side )
{
	return mSides[ getSideIndex( side ) ];
}

void SkyboxRenderer::getCameraTransformation( GLMatrix4d & m )
{
	// TODO: get matrix directly from camera, not some OpenGL matrix.
	m.glGet( GL_MODELVIEW_MATRIX );
}

void SkyboxRenderer::render()
{
	// TODO: low priority: do better skybox rendering than immediate mode
	static const double halfSide = 10;
	static const GLVector3d front(0, halfSide, 0), top(0, 0, halfSide), right(halfSide, 0, 0);
	static const GLVector3d left(-right), bottom(-top), back(-front);

	// push attributes
	glPushAttrib( 
		GL_COLOR_BUFFER_BIT | 
		GL_TEXTURE_BIT | 
		GL_TRANSFORM_BIT | 
		GL_DEPTH_BUFFER_BIT | 
		GL_POLYGON_BIT );

	// get a copy of camera matrix ( assumed orthogonal ) tm
	GLMatrix4d tm( GLMatrix4d::identity() );
	this->getCameraTransformation( tm );

	// zero translation part (this moves camera to the skybox center)
	tm.m12 = 0;
	tm.m13 = 0;
	tm.m14 = 0;

	// texture matrix - preserve and then set to identity 
	glMatrixMode( GL_TEXTURE );
	glPushMatrix();
	glLoadIdentity();

	// modelview matrix - preserve and then set to tm
	glMatrixMode( GL_MODELVIEW );
	glPushMatrix();
	tm.glLoadMatrix();
	
	// disable multitexturing and enable texture unit 0
	getAppClient().getGFX().disableMultitexturing();

	glActiveTexture( GL_TEXTURE0 );
	glEnable( GL_TEXTURE_2D );
	
	// set texture environment to GL_REPLACE
	glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE );

	// disable automatic texgen
	glDisable( GL_TEXTURE_GEN_S );
	glDisable( GL_TEXTURE_GEN_T );
	glDisable( GL_TEXTURE_GEN_R );
	glDisable( GL_TEXTURE_GEN_Q );

	// set polygon mode
	glPolygonMode( GL_FRONT, GL_FILL );
	
	// set backface culling
	glCullFace( GL_BACK );

	// disable blending
	glDisable( GL_BLEND );

	// disable alpha testing
	glDisable( GL_ALPHA_TEST );

	// disable depth testing and set the depth mask to false	
	glDisable( GL_DEPTH_TEST );
	glDepthMask( GL_FALSE );

	glDisable( GL_LIGHTING );
	
	// for each side:
	//  - bind texture
	//  - draw side
	
	TextureRef tex;
	tex = getTexture( SkyboxRenderer::LEFT );
	if ( tex.isValid() ) 
	{
		tex->bind();
		glBegin(GL_QUADS);
			glTexCoord2f(0, 0);
			(left + back + top).glVertex();

			glTexCoord2f(0, tex->getMaxT());
			(left + back + bottom).glVertex();

			glTexCoord2f(tex->getMaxS(), tex->getMaxT());
			(left + front + bottom).glVertex();

			glTexCoord2f(tex->getMaxS(), 0);
			(left + front + top).glVertex();
		glEnd();
	}
	
	tex = getTexture( SkyboxRenderer::FRONT );
	if ( tex.isValid() ) 
	{
		tex->bind();
		glBegin(GL_QUADS);
			glTexCoord2f(0, 0);
			(front + left + top).glVertex();

			glTexCoord2f(0, tex->getMaxT());
			(front + left + bottom).glVertex();

			glTexCoord2f(tex->getMaxS(), tex->getMaxT());
			(front + right + bottom).glVertex();

			glTexCoord2f(tex->getMaxS(), 0);
			(front + right + top).glVertex();
		glEnd();
	}

	tex = getTexture( SkyboxRenderer::RIGHT );
	if ( tex.isValid() ) 
	{
		tex->bind();
		glBegin(GL_QUADS);
			glTexCoord2f(0, 0);
			(right + front + top).glVertex();

			glTexCoord2f(0, tex->getMaxT());
			(right + front + bottom).glVertex();

			glTexCoord2f(tex->getMaxS(), tex->getMaxT());
			(right + back + bottom).glVertex();

			glTexCoord2f(tex->getMaxS(), 0);
			(right + back + top).glVertex();
		glEnd();
	}

	tex = getTexture( SkyboxRenderer::BACK );
	if ( tex.isValid() ) 
	{
		tex->bind();
		glBegin(GL_QUADS);
			glTexCoord2f(0, 0);
			(back + right + top).glVertex();

			glTexCoord2f(0, tex->getMaxT());
			(back + right + bottom).glVertex();

			glTexCoord2f(tex->getMaxS(), tex->getMaxT());
			(back + left + bottom).glVertex();

			glTexCoord2f(tex->getMaxS(), 0);
			(back + left + top).glVertex();
		glEnd();
	}

	tex = getTexture( SkyboxRenderer::TOP );
	if ( tex.isValid() ) 
	{
		tex->bind();
		glBegin(GL_QUADS);
			glTexCoord2f(0, 0);	
			(top + left + back).glVertex();
		
			glTexCoord2f(0, tex->getMaxT());
			(top + left + front).glVertex();

			glTexCoord2f(tex->getMaxS(), tex->getMaxT());
			(top + right + front).glVertex();

			glTexCoord2f(tex->getMaxS(), 0);
			(top + right + back).glVertex();
		glEnd();	
	}

	tex = getTexture( SkyboxRenderer::BOTTOM );
	if ( tex.isValid() ) 
	{
		tex->bind();
		glBegin(GL_QUADS);
			
			glTexCoord2f(0, 0);
			(bottom + left + front).glVertex();

			glTexCoord2f(0, tex->getMaxT());
			(bottom + left + back).glVertex();

			glTexCoord2f(tex->getMaxS(), tex->getMaxT());
			(bottom + right + back).glVertex();

			glTexCoord2f(tex->getMaxS(), 0);
			(bottom + right + front).glVertex();
		glEnd();
	}

	// restore modelview matrix
	glPopMatrix();

	// restore texture matrix
	glMatrixMode( GL_TEXTURE );
	glPopMatrix();

	// restore attributes
	glPopAttrib();
}
