/* MAINTAINER: Bernard */


/**
 * \file isound.h
 *
 * Interfaces for sound. The rest of program (except sfx) sould
 * use only these interfaces.
 */

#ifndef _I_SOUND_H
#define _I_SOUND_H


#include <string>
#include "log.h"
#include "task.h"

class ISoundSystem;
class ISoundMusic;
class ISoundSample;
class ISoundSamplesManager;

/**
 * \brief Basic class that holds all sound.
 *
 * Handles the inicialization etc.
 *
 * Note that children should (must) redefine getMusic() or create
 * mMusic in constructor & same with getSamplesManager;
 * 
 */
class ISoundSystem
{
public:	
	/**
	 * \brief Inicialization of sound subsytem
	 */
	ISoundSystem() {}

	/**
	 * \brief Clean sound subsystem.
	 */
	virtual ~ISoundSystem() {}

	/**
	 * \brief Return object that handle the music.
	 *
	 * Note that only one object if type ISoundMusic should exist.
	 */
	virtual ISoundMusic& getMusic()
	{
		assertL(mMusic != 0);

		return *mMusic;
	}

	/**
	 * \brief Return object for handling samples.
	 */
	virtual ISoundSamplesManager& getSamplesManager()
	{
		assertL(mSamplesManager != 0);
		return *mSamplesManager;
	}
	
protected:

	/**
	 * \brief Delete mMusic and mSamplesManager
	 */
	void deleteObjects();
	
	ISoundMusic* mMusic;
	ISoundSamplesManager* mSamplesManager;
	
};

/**
 * \brief Virtual class for holding sound samples.
 *
 *  Nowhere implemented yet.
 *  For example of usage see ISoundSamplesManager;
 */
class ISoundSample
{
public:
	ISoundSample() { mSuicide = false; };
	virtual ~ISoundSample() {};

	/**
	 * \brief Delete this object after stop playing of the sound.
	 *
	 * This is good if you want to create some sound and don't want
	 * to hold the object with the sound. Like sound of explosion.
	 * Example of usage:
	 *   ISoundSample *mysound = SoundSamplesManager.newSample("ahoj.vaw");
	 *   mysound->selfDeleteAfterStop();
	 *   mysound->play();
	 *   // now don't use mysound pointer any more since you reinicialize it.
	 *   // the sound is playing and no(?) memory leaks :-)
	 * Note the order selfDeleteAfterStop is called BEFORE play command. The
	 * playcommand itself may delete the object (if was invalid - no sound loaded)
	 * so carefully :-)
	 */
	virtual void selfDeleteAfterStop()
	{
		mSuicide = true;
	}

	virtual void play(bool loop = false) = 0;

	
protected:
	/**
	 * \brief The object try to sucide.
	 *
	 * \return true if the object did self delte.
	 */
//	virtual bool suicideOpportunity() = 0;
	virtual bool suicideOpportunity();

	bool  mSuicide; /// <\brief true if selfdelete is wanted :-)
};

/**
 * \brief Handle creation and management of samples.
 *
 * You soudn't create the manager yourself but ask for it from
 * ISoundSystem. Note that this class should be added to
 * TaskManager as task, because it needs to handle with position
 * sound and need to change sound volumes during the game.
 * 
 * Usage:
 * ISoundSystem* ss = getSomehowSomewhere();
 * ISoundSamplesManager* sm = ss->getSamplesManager();
 * 
 * ISoundSample* sample = sm->newSample("MegaFun.wav");
 * ..
 * sample->play();
 * ..
 * sm->deleteSample(sample);
 * // don't use sample any more
 * 
 * 
 */
class ISoundSamplesManager : public Task
{
public:

	ISoundSamplesManager() {};
	virtual ~ISoundSamplesManager() {};
	
	/**
	 * \brief Create new object with sample from file.
	 *
	 * Allways return valid value. The object is allocated with new.
	 * You may delete it yourself  or may call deleteSample, which
	 * is prefered, becasue in some future reimplementation...
	 * \param file File with sample to load. Usually wav. The file
	 *        will be themed and locateFile will be perofrmed on it.
	 * \return new sample
	 */
	virtual ISoundSample* newSample(const char* file) = 0;

	/**
	 * \brief Delete sample created by newSample if not needed any more.
	 * 
	 * This call must be allways same, because it may happen, that sfx return
	 * different ISoundSamplesManager during the program so it would be bad
	 * if the samples would come to wrong method of deletion.
	 * \param sample Sample to delete
	 */
	void deleteSample(ISoundSample* sample)
	{
		if (sample) delete sample;
	}
	
	/**
	 * \brief Stop all currently playing samples.
	 *
	 * Call probably at the end of game - when jumping from game
	 * to menu or vice versa.
	 */
	virtual void stopAll() = 0;
	
	/**
	 * \brief Get current volume of samples.
	 * 
	 * \return volume is in range 0..getMaxVolume()
	 */
	virtual int getVolume() = 0;
	
	/**
	 * \brief Get maximum possible volume.
	 * 
	 * \return maximum possible volume.
	 */
	virtual int getMaxVolume() = 0;
	
	/**
	 * \brief Set samples volume to some variable.
	 * 
	 * \param volume new volume
	 */
	virtual void setVolume(int volume) = 0;
};



/**
 * \brief Handle the music.
 *
 * There may be omly one object for handling music and only one music
 * playing at time. This is only a interface and is implemented by SDL.
 * Usage:
 * ISoundSystem* ss = getSomehowSomewhere();
 *
 * ISoundMusic* music = ss->getMusic(); // don't delete this pointer when not used
 * music->play("my_song.pm3", true);
 * ..
 * music->stop(); 
 * music = 0;
 *
 * Note that you don't need to hold pointer to music while playing...
 * 
 */
class ISoundMusic
{
public:	
	
	ISoundMusic() {}
	virtual ~ISoundMusic() {}

	
	/**
	 * \brief Start playing music file.
	 *
	 * The file to play isn't searched in file system.
	 * If there was some music allready playing - it is stopped.
	 * \param file name of file to play. locateFile isn't called.
	 * \param loop true if the music should loop forever
	 */
	virtual void play(const char* file, bool loop) = 0;
	
	/**
	 * \brief Start playing music file.
	 *
	 * Equal to call play(file.c_str(), loop)
	 */
	virtual void play(const std::string & file, bool loop)
	{
		play(file.c_str(), loop);
	}

	/**
	 * \brief Stop currently playing music.
	 *
	 * If there is no music, nothing happens.
	 */
	virtual void stop() = 0;
	
	/**
	 * \brief Get music status.
	 *
	 * \return true if the music is playing.
	 */
	virtual bool isPlaying() = 0;

	// May be added
	// virtual void pause() = 0;
	// virtual void resume() = 0;

	/**
	 * \brief Get current volume of music.
	 * 
	 * \return volume is in range 0..getMaxVolume()
	 */
	virtual int getVolume() = 0;
	
	/**
	 * \brief Get maximum possible volume.
	 * 
	 * \return maximum possible volume.
	 */
	virtual int getMaxVolume() = 0;
	
	/**
	 * \brief Set music volume to some variable.
	 * 
	 * \param volume new volume
	 */
	virtual void setVolume(int volume) = 0;	
};

#endif
