/* MAINTAINER: Bernard */

/**
 * \file Provide empty implementation of sound classes.
 */

#ifndef _I_SOUND_EMPTY_H
#define _I_SOUND_EMPTY_H


#include "isound.h"
#include "log.h"

/**
 * \brief Empty implementation of ISoundMusic.
 * 
 * The empty implementation is usefull if you want to turn off 
 * the sound subsystem completly.
 */
class SoundMusic_Empty : public ISoundMusic
{
public:	

	SoundMusic_Empty()
	{
		mPlaying = false;	
	}

	~SoundMusic_Empty(){};
	
	/**
	 *\brief Pretend like the sound started to play.
	 * 
	 * Just set mPlaying to true.
	 */
	virtual void play(const char* file, bool loop)
	{
		mPlaying = true;
	}

	/**
	 * \brief Pretend like stopping the sound.
	 * 
	 * Just set mPlaying to false.
	 */
	virtual void stop()
	{
		mPlaying = false;
	}
	
	/**
	 * \brief Get current status of music.
	 * 
	 * \return mPlaying
	 */
	virtual bool isPlaying()
	{
		return mPlaying;
	}
	
	/**
	 * \brief Return some fake min volume.
	 * 
	 * \return Always 0.
	 */
	virtual int getVolume() { return 0; }
	
	/**
	 * \brief Return some fake max volume.
	 * 
	 * \return Always 128.
	 */	
	virtual int getMaxVolume() { return 128; }
	
	/**
	 * \brief Does nothing.
	 * 
	 * \param volume Not used.
	 */		
	virtual void setVolume(int volume) { }
	
protected:

	bool       mPlaying; /// \brief Just for remembering last state.
};	

/**
 * \brief Empty implementation of ISoundSample.
 * 
 * The implementation try to suicide in every play() call.
 */
class SoundSample_Empty : public ISoundSample
{
public:
	SoundSample_Empty() {};
	virtual ~SoundSample_Empty() {};

	/**
	 * \brief Try to sucide.
	 * 
	 * \param loop Not used.
	 */
	virtual void play(bool loop = false)
	{
		suicideOpportunity();
	}
};

/**
 * \brief Empty implementation of ISoundSamplesManager.
 */
class SoundSamplesManager_Empty : public ISoundSamplesManager
{
public:

	SoundSamplesManager_Empty() {};
	virtual ~SoundSamplesManager_Empty() {};
	
	/**
	 * \brief Create new SoundSample_Empty
	 *
	 * Allways return valid value. The object is allocated with new.
	 * You may delete it yourself  or may call deleteSample, which
	 * is prefered, becasue in some future reimplementation...
	 * \param file ignored
	 * \return new sample
	 */
	virtual ISoundSample* newSample(const char* file)
	{
		SoundSample_Empty *empty = new SoundSample_Empty();
		if (empty == 0)
		{
			DIE("Failed allocation of SoundSample_Empty");
		}

		return empty;
	}

	/**
	 * \brief Does nothing.
	 */
	virtual void stopAll()
	{
		
	}
	
	/**
	 * \brief Return some fake volume.
	 * 
	 * \return always 0.
	 */
	int getVolume() { return 0; }
	
	/**
	 * \brief Return some fake max volume.
	 * 
	 * \return always 128.
	 */	
	int getMaxVolume() { return 128; }
	
	/**
	 * \brief Does nothing.
	 * 
	 * \param volume Not used.
	 */
	void setVolume(int volume) {}
	
};


class SoundSystem_Empty : public ISoundSystem
{
public:	
	SoundSystem_Empty() 
	{
		mMusic = new SoundMusic_Empty();
		if (mMusic == 0)
		{
			DIE("Unable to allcoate object SoundMusic_Empty");
		}

		mSamplesManager = new SoundSamplesManager_Empty();
		if (mSamplesManager == 0)
		{
			DIE("Unable to allcoate object SoundSamplesManager_Empty");
		}
	}

	virtual ~SoundSystem_Empty() 
	{
		deleteObjects();
	}
};

#endif
