/* MAINTAINER: Bernard */

#ifndef NO_OPENAL

#define WANT_TRACE 

#include "tinyxml/tinyxml.h"
#include "tasksound.h"
#include "params.h"
#include "projectile.h"
#include "application_client.h"
#include "params.h"
#include "platform.h"
#include "settings.h"
#include <cstring>
#include "log.h"
#include "game_client.h"
#include "camera.h"

using namespace std;

TaskSound* TaskSound::mThis = 0;


#define OPEN_AL_ERROR_CHECK \
	{ \
		int err; \
		if ((err = alGetError()) != AL_NO_ERROR) \
		{ \
/*			LOGE << "OpenAL: " << alGetString(err) << endl;*/ \
/*	DIE(); \  */ \
		}\
	}

#define OPEN_ALC_ERROR_CHECK \
	{ \
		int err; \
		if ((err = alcGetError(mDevice)) != ALC_NO_ERROR) \
		{ \
/*			LOGE << "OpenAL: " << alcGetString(mDevice, err) << endl;*/ \
/*	DIE(); \  */ \
		}\
	}
#ifdef ALUT_API_MAJOR_VERSION
	#define ALUT_ERROR_CHECK \
		{ \
			int err; \
			if ((err = alutGetError()) != ALC_NO_ERROR) \
			{ \
/*				LOGE << "OpenAL: " << alutGetErrorString(err) << endl;*/ \
	/*	DIE(); \  */ \
			}\
		}
#else
	#define ALUT_ERROR_CHECK
#endif

#define OPEN_AL_ERROR_CHECK_DEBUG OPEN_AL_ERROR_CHECK

TaskSound::TaskSound()
{
	mInited  = false;
	mDevice  = 0;
	mContext = 0;
	
	mGlobalGain = (ALfloat)g_settings.getDouble(SETTING_3DSOUND_GLOBAL_GAIN);
	
	mInicialized = false;
	mOpenALInicilized = false;
	
	assertL(mThis == 0);
	mThis = this;
}

TaskSound::~TaskSound()
{
	assertL(mThis);
	mThis = 0;
}

void TaskSound::initOpenAL()
{	
	// test if allready inicialized
	if (mOpenALInicilized) return;
	
	// Open sound device
	string device = g_settings.getString(SETTING_3DSOUND_DEVICE);
	if (device.empty())
	{
    	mDevice = alcOpenDevice( NULL );
	}
	else
	{
#ifdef ALUT_API_MAJOR_VERSION	
    	mDevice = alcOpenDevice(device.c_str());
#else
		// backward compatibility alut 0.x
    	mDevice = alcOpenDevice((ALCubyte*)device.c_str());
#endif
	}
	OPEN_ALC_ERROR_CHECK;
    
	if (mDevice == NULL) return;

	// Create context
	mContext = alcCreateContext( mDevice, NULL );
	if( mContext == NULL ) 
	{
		alcCloseDevice( mDevice );
		OPEN_ALC_ERROR_CHECK;
		return;
	}   
	OPEN_ALC_ERROR_CHECK;
	
	alcMakeContextCurrent(mContext);
	OPEN_ALC_ERROR_CHECK;

	//alDistanceModel(AL_LINEAR_DISTANCE_CLAMPED); // this doesn't exist in OpenAL 1.0
	alDistanceModel(AL_INVERSE_DISTANCE);
	//alDistanceModel(AL_INVERSE_DISTANCE_CLAMPED);
	//alDistanceModel(AL_NONE);

	OPEN_AL_ERROR_CHECK;

	// just some info about OpenAL
	printOpenALInfo();
	
	mOpenALInicilized = true;
}

ALuint TaskSound::loadBuffer(const std::string &file)
{
	if (!mOpenALInicilized) return AL_NONE;
	
	ALuint buffer = AL_NONE;

#ifdef ALUT_API_MAJOR_VERSION

	buffer = alutCreateBufferFromFile(locateFile(file).c_str());
	if (buffer == AL_NONE)
	{
		LOGE << "Unable to load file " << file 
		     << " because " << alutGetErrorString(alutGetError ()) << endl;
	}

#else // old version of ALUT
    // property of wav
    ALenum  format;
    ALsizei   size;
    ALvoid*   data;
    ALsizei   freq;

    alGenBuffers(1, &buffer);
    if (alGetError() != AL_NO_ERROR) return AL_NONE;

	// there is different function version on MacOS and Linux
#if defined(__APPLE__)
    alutLoadWAVFile((ALbyte*)locateFile(file).c_str(), &format, &data, &size, &freq);
#else
    alutLoadWAVFile((ALbyte*)locateFile(file).c_str(), &format, &data, &size, &freq, 0);
#endif
	
    alBufferData(buffer, format, data, size, freq);
    alutUnloadWAV(format, data, size, freq);
#endif

	ALUT_ERROR_CHECK;

	return buffer;
}


void TaskSound::playBufferForPosition(const GLVector3d &pos, const sound3Ddescription &dsc)
{
	if (dsc.buffer == AL_NONE) return;
	if (!mInicialized) return;
	
	// Don't plat distatnt things
	if ( abs(mListener.x - pos.x) + abs(mListener.y - pos.y) > dsc.distance_silence ) return;

	ALuint source;	
	alGenSources(1, &source);
	
	OPEN_AL_ERROR_CHECK;
	
	alSourcei (source, AL_BUFFER,   dsc.buffer);                      OPEN_AL_ERROR_CHECK_DEBUG;
	alSourcef (source, AL_PITCH,    dsc.pitch);                       OPEN_AL_ERROR_CHECK_DEBUG;
	alSourcef (source, AL_GAIN,     dsc.gain*mGlobalGain);            OPEN_AL_ERROR_CHECK_DEBUG;
	alSourcei (source, AL_LOOPING,  AL_FALSE     );                   OPEN_AL_ERROR_CHECK_DEBUG;
	alSourcef (source, AL_REFERENCE_DISTANCE,  dsc.reference_distance); OPEN_AL_ERROR_CHECK_DEBUG;
	alSourcef (source, AL_ROLLOFF_FACTOR,      dsc.rolloff_factor);      OPEN_AL_ERROR_CHECK_DEBUG;

	alSource3f(source, AL_POSITION, 
		(ALfloat)(pos.x), 
		(ALfloat)(pos.y),
		(ALfloat)(pos.z)); OPEN_AL_ERROR_CHECK_DEBUG;
	alSourcePlay(source); OPEN_AL_ERROR_CHECK_DEBUG;
	
	mOnlySources.push_back(source);

	OPEN_AL_ERROR_CHECK;
}

void TaskSound::playBufferForEntity(WorldEntity* we, const sound3Ddescription &dsc)
{
	if (dsc.buffer == AL_NONE) return;
	if (!mInicialized) return;
	
	list3Dmember m;
	m.we = we;

	alGenSources(1, &m.source);
	OPEN_AL_ERROR_CHECK;
	
	alSourcei (m.source, AL_BUFFER,   dsc.buffer  );                     OPEN_AL_ERROR_CHECK_DEBUG;
	alSourcef (m.source, AL_PITCH,    dsc.pitch   );                     OPEN_AL_ERROR_CHECK_DEBUG;
	alSourcef (m.source, AL_GAIN,     dsc.gain*mGlobalGain);                     OPEN_AL_ERROR_CHECK_DEBUG;
	alSourcei (m.source, AL_LOOPING,  AL_TRUE     );                     OPEN_AL_ERROR_CHECK_DEBUG;
	alSourcef (m.source, AL_REFERENCE_DISTANCE, dsc.reference_distance); OPEN_AL_ERROR_CHECK_DEBUG;
	alSourcef (m.source, AL_ROLLOFF_FACTOR,     dsc.rolloff_factor    ); OPEN_AL_ERROR_CHECK_DEBUG;

	m.playing = false;
	m.distance_silence = dsc.distance_silence;
	m.distance_sound   = dsc.distance_sound;
	updateSource(m);

	mEntites.push_back(m);
	
	OPEN_AL_ERROR_CHECK;
}

void TaskSound::addedWorldObject(WorldEntity* we)
{
	assertL(we != NULL);
	if (!mInicialized) return;
	
	// one shot from projectile
	switch (we->getClassType())
	{
	case cProjectile :
		playBufferForPosition(we->getPosition(), mProjectileBuffers[we->getSubType()]);
		break;

	case cBonusBox :
//		playBufferForEntity(we, mBuffer);
		break;

	case cTank :
		playBufferForEntity(we, mTankEngineBuffers[we->getSubType()]);
		break;
	default:
		;
	}
}

void TaskSound::deletedWorldObject(WorldEntity* we)
{
	if (!mInicialized) return;
	
	// Find if it is playing some sound and if so, delete it all
	list3DOpenAL::iterator it = mEntites.begin();
	for (;it != mEntites.end(); ++it)
	{
		if (it->we == we) 
		{
			alSourceStop(it->source);
			alDeleteSources(1, &(it->source));
			mEntites.erase(it);	
			break;
		}
	}
	
	OPEN_AL_ERROR_CHECK;
}

void TaskSound::init()
{
	if (mInicialized) return;
	
	// game_client.cpp allready does this check but for sure
	if (!g_settings.isTrue(SETTING_3DSOUND_ENABLE)) return;
	
	initOpenAL();


	//mBuffer = loadBuffer("sample.wav");

	// Load all projectile buffers
	mProjectileBuffers.resize(gParams.mProjectiles.size());
	for (int i = 0; (unsigned int)i < gParams.mProjectiles.size(); i++)
	{
		const string &file = gParams.projectile(i).shoot_sound;
		if (!file.empty())
		{
			mProjectileBuffers[i] = loadDescription(file);
		}
	}

	// Load all tank engine buffers
	mTankEngineBuffers.resize(gParams.mTanks.size());
	for (int i = 0; (unsigned int)i < gParams.mTanks.size(); i++)
	{
		const string &file = gParams.tank(i).engine_sound;
		if (!file.empty())
		{
			mTankEngineBuffers[i] = loadDescription(file);
		}
	}

	// Load explosiosn buffers
	mExplosionBuffers.resize(gParams.mExplosions.size());
	for (int i = 0; (unsigned int)i < gParams.mExplosions.size(); i++)
	{
		const string &file = gParams.explosion(i).sound;
		if (!file.empty())
		{
			mExplosionBuffers[i] = loadDescription(file);
		}
	}

	OPEN_AL_ERROR_CHECK;
	
	mInicialized = true;
}

void TaskSound::clean()
{
	if (!mInicialized) return;
	
	// clear entites sources
	{
		list3DOpenAL::iterator it = mEntites.begin();
		for (;it != mEntites.end(); ++it)
		{
			alSourceStop(it->source);          OPEN_AL_ERROR_CHECK_DEBUG;
			alDeleteSources(1, &(it->source)); OPEN_AL_ERROR_CHECK_DEBUG;
		}
		mEntites.clear();
	}	

	// clear other sources
	{
		list3DSources::iterator it = mOnlySources.begin();
		for (;it != mOnlySources.end(); ++it)
		{
			alSourceStop(*it);          OPEN_AL_ERROR_CHECK_DEBUG;
			alDeleteSources(1, &(*it)); OPEN_AL_ERROR_CHECK_DEBUG;
		}
		mOnlySources.clear();
	}

	// clear buffers
	list3DBuffers::iterator it = mProjectileBuffers.begin();
	for (;it != mProjectileBuffers.end(); ++it)
	{
		if (it->buffer != AL_NONE)
		{
			alDeleteBuffers(1, &(it->buffer)); OPEN_AL_ERROR_CHECK_DEBUG;
		}
	}
	mProjectileBuffers.clear();

	it = mTankEngineBuffers.begin();
	for (;it != mTankEngineBuffers.end(); ++it)
	{
		if (it->buffer != AL_NONE)
		{
			alDeleteBuffers(1, &(it->buffer)); OPEN_AL_ERROR_CHECK_DEBUG;
		}
	}
	mTankEngineBuffers.clear();

	it = mExplosionBuffers.begin();
	for (;it != mExplosionBuffers.end(); ++it)
	{
		if (it->buffer != AL_NONE)
		{
			alDeleteBuffers(1, &(it->buffer)); OPEN_AL_ERROR_CHECK_DEBUG;
		}
	}
	mExplosionBuffers.clear();


	OPEN_AL_ERROR_CHECK;

	cleanOpenAL();
	
	mInicialized = false;
}

void TaskSound::cleanOpenAL()
{
	if (!mOpenALInicilized) return;
	
	// Clean device and context
	alcMakeContextCurrent(NULL);
	alcDestroyContext(mContext);
	alcCloseDevice(mDevice);
	mContext = NULL;
	mDevice = NULL;

	
	mOpenALInicilized = false;
}

void TaskSound::update(int atime)
{
	OPEN_AL_ERROR_CHECK;

	// Update listener
	Camera *cam = getAppClient().getGame().getCamera();
	if (cam)
	{
		WorldEntity *we;
		if ( (we = cam->monitoredEntity()))
		{
			cam->getPos(mListener);
			mListener = we->getPosition()*0.9 + mListener*0.1;
		}
		else
		{
			cam->getPos(mListener);
		}
		alListener3f(AL_POSITION, (ALfloat)mListener.x, 
		             (ALfloat)mListener.y, (ALfloat)mListener.z);  OPEN_AL_ERROR_CHECK_DEBUG;

		GLVector3d dir, up;
		cam->getDir(dir);
		cam->getDir(up);
		ALfloat orient[] = {(ALfloat)dir.x, (ALfloat)dir.y, (ALfloat)dir.z, 
		                    (ALfloat)up.x,  (ALfloat)up.y,   (ALfloat)up.z  };
		alListenerfv(AL_ORIENTATION, orient); OPEN_AL_ERROR_CHECK_DEBUG;

		//TODO: Velocity
		//alListenerfv(AL_VELOCITY,    lvelocity); 
	}

	OPEN_AL_ERROR_CHECK;

	// Update positions of other objects
	{
		list3DOpenAL::iterator it = mEntites.begin();
		for (;it != mEntites.end(); ++it)
		{
			updateSource(*it);
		}
	}	
	
	OPEN_AL_ERROR_CHECK;

	// Check if only sources are still playing
	{
		list3DSources::iterator it = mOnlySources.begin();
		while ( it!= mOnlySources.end() )
		{
			ALint state;
			alGetSourcei(*it, AL_SOURCE_STATE, &state); OPEN_AL_ERROR_CHECK_DEBUG;
			if (state == AL_STOPPED)
			{
				alSourceStop(*it);           OPEN_AL_ERROR_CHECK_DEBUG;
				alDeleteSources(1, &(*it));  OPEN_AL_ERROR_CHECK_DEBUG;
				it = mOnlySources.erase(it);	
		//		LOGD << "erasing" << endl;
			}
			else
			{
				++it;
			}
		}
	}	
	
	OPEN_AL_ERROR_CHECK;
}

void TaskSound::updateSource(list3Dmember &lm)
{
	const GLVector3d& pos = lm.we->getPosition();
	alSource3f(lm.source, AL_POSITION, 
		(ALfloat)(pos.x), (ALfloat)(pos.y), (ALfloat)(pos.z)); 
	OPEN_AL_ERROR_CHECK_DEBUG;

//	LOGT << "Source: " << pos.x << " " << pos.y << " " << pos.z << endl;

	const GLVector3d& speed = lm.we->getSpeed();
	alSource3f(lm.source, AL_VELOCITY,
		(ALfloat)(-speed.x/80),  //TODO: Those - are here because our tank has beter sound
		(ALfloat)(-speed.y/80),  // then. They should be removed and something like our tank
		(ALfloat)(-speed.z/80)); // recognization should be placed here.
	OPEN_AL_ERROR_CHECK_DEBUG;

	if ( lm.playing && (abs(mListener.x - pos.x) + abs(mListener.y - pos.y)) > lm.distance_silence )
	{
		alSourcePause(lm.source);  OPEN_AL_ERROR_CHECK_DEBUG;
		lm.playing = false;
//		LOGT << "STOP!!" << endl;
	}
	else if ( !lm.playing && abs(mListener.x - pos.x) + abs(mListener.y - pos.y) < lm.distance_sound )
	{
		alSourcePlay(lm.source);   OPEN_AL_ERROR_CHECK_DEBUG;
		lm.playing = true;
//		LOGT << "PLAY!!" << endl;
	}
}

void TaskSound::printOpenALInfo()
{

	// Write some interesting info about OpenAL
	if (alGetString(AL_VERSION))  
		LOGD << "OpenAL version: "    << alGetString(AL_VERSION)    << endl;
	if (alGetString(AL_RENDERER)) 
		LOGD << "OpenAL renderer: "   << alGetString(AL_RENDERER)   << endl;
	if (alGetString(AL_VENDOR)) 
		LOGD << "OpenAL vendor: "     << alGetString(AL_VENDOR)     << endl;
	if (alGetString(AL_EXTENSIONS)) 
		LOGD << "OpenAL extensions: " << alGetString(AL_EXTENSIONS) << endl;
	OPEN_AL_ERROR_CHECK;

#ifdef ALUT_API_MAJOR_VERSION
	// Enumerate OpenAL devices
	if (alcIsExtensionPresent (NULL, (const ALCchar *) "ALC_ENUMERATION_EXT") == AL_TRUE)
	{
		const char *s = (const char *) alcGetString(NULL, ALC_DEVICE_SPECIFIER);
		while (*s != '\0')
		{
			LOGD << "OpenAL avaliable device: " << s << endl;
			while (*s++ != '\0');
		}
	}
	else
	{
		LOGD << "OpenAL device enumeration isn't available." << endl;
	}
	OPEN_ALC_ERROR_CHECK;
#endif

	// Print default device name
	LOGD << "OpenAL default device: " 
	     << (const char *)alcGetString(NULL, ALC_DEFAULT_DEVICE_SPECIFIER) 
		 << endl;
	OPEN_ALC_ERROR_CHECK;

	// Print current device name
	if (mDevice)
	{
		LOGD << "OpenAL current device: " 
			 << (const char *)alcGetString(mDevice, ALC_DEVICE_SPECIFIER) 
			 << endl;
		OPEN_ALC_ERROR_CHECK;
	}
}


void TaskSound::newExplosion(int id, const GLVector3d &pos)
{
	if (id < 0 || id >= (int)mExplosionBuffers.size()) return;
		
	playBufferForPosition(pos, mExplosionBuffers[id]);	
}

#define LOAD_VALUE(element) \
	child = handle.FirstChild("buffer").FirstChild(element).Element(); \
	if (!child) \
	{ \
		LOGE << "There is no element " element  " in file " << filepath << endl; \
		return dsc; \
	} \
	value = child->GetText();

sound3Ddescription TaskSound::loadDescription(const std::string &file)
{
	sound3Ddescription dsc;
	std::string wavfile;
	
	TiXmlDocument doc;
	TiXmlHandle   handle(&doc);
	TiXmlElement* child;
	const char *value;
	
	string filepath = locateFile(file);
	if (!doc.LoadFile(filepath)) 
	{ 
		LOGE << "Unable to load " << filepath
		     << " Probably file not exist or isn't a valid XML file" << endl; 
		return dsc; // No need to die  - jus the sound is not available
	} 
	
	
	LOAD_VALUE("wav_file");
	wavfile = value;
	
	LOAD_VALUE("distance_silence");
	dsc.distance_silence  = (float)get_double(value);
	
	LOAD_VALUE("distance_sound");
	dsc.distance_sound  = (float)get_double(value);
	
	LOAD_VALUE("rolloff_factor");	
	dsc.rolloff_factor     = (ALfloat)get_double(value);
	
	LOAD_VALUE("reference_distance");	
	dsc.reference_distance = (ALfloat)get_double(value);
	
	LOAD_VALUE("gain");		
	dsc.gain  = (ALfloat)get_double(value);
	
	LOAD_VALUE("gain");			
	dsc.pitch = (ALfloat)get_double(value);
	
	// Fill the buffer with sound
	dsc.buffer = loadBuffer(wavfile);
	
	return dsc;
}

#endif // NO_OPENAL

