/* MAINTAINER: Martin */

#include "GL/glew.h" // OpenGL extensions

#include "texturemanager.h"
#include "SDL_image.h"
#include "application_client.h"
#include "globals_client.h"
#include "platform.h"
#include "conversions.h"
#include "gfx.h"

#include "grutils.h"
using namespace grutils;

#include <sstream>

const GLenum Texture::firstTextureUnitID( GL_TEXTURE0 );

Uint32 SurfaceFactory::toPOT(Uint32 arg)
{
	Uint32 pot = 1;
	while (pot < arg && pot < ((Uint32)1 << 31))
		pot <<= 1;
	return pot;
}

TextureManager::TextureManager()
{
	
	mNPOT = GLEW_ARB_texture_non_power_of_two != 0;
}

TextureManager::TUniversalManager::TRef TextureManager::add(SurfaceFactory * psf, GLint internalFormat, GLint level)
{
	// I. FIND
	ManagedItem * pmi;
	std::string hashString = psf->getHashString();
	if ((pmi = find(hashString)) != 0)
		return pmi;

    // II. INIT
	Texture * ptex= new Texture();
	ptex->upload(psf, internalFormat, level);

	// III. REGISTER
	return createAndRegister(hashString, ptex);
}

TextureManager::TUniversalManager::TRef TextureManager::add(std::string fileName, GLint internalFormat, GLint level)
{
	SurfDefault sd(fileName);
	return add(&sd, internalFormat, level);
}

SurfaceFactory::SurfaceFactory(SDL_Surface *src)
	: mSrcSurface(src), mSrcSurfaceOwned(false), mULC(false)
{}

std::string SurfColorKey::createHashSuffix()
{
	std::stringstream s;
	s << " keycolor(" << mKeyR << "," << mKeyG << ","  << mKeyB << ")"; 
	return s.str();
}

std::string SurfAlpha::createHashSuffix()
{
	std::stringstream s;
	s << " alpha(" << mAlpha << ")";
	return s.str();
}

SDL_Surface * SurfaceFactory::getSrcSurface()
{	TIN;
	if (!mSrcSurface){
		// Load the image file into a SDL_Surface.
		mSrcSurface = IMG_Load(mImagePath.c_str());
		if ( !mSrcSurface ) {
			LOGE << "Surface loading: Can not load '"<< mImagePath << "' (" << IMG_GetError() << ")" << std::endl;
			return 0;
		}
		mSrcSurfaceOwned  = true;
	}
	return mSrcSurface;
}

SurfaceFactory::~SurfaceFactory()
{
	if (mSrcSurfaceOwned && mSrcSurface){
		SDL_FreeSurface(mSrcSurface);
		mSrcSurface = 0;
		mSrcSurfaceOwned = false;
	}
}

void SurfaceFactory::setDestPixelFormat(SDL_PixelFormat *pf)
{
	memset(pf, 0, sizeof(SDL_PixelFormat));

	pf->BitsPerPixel = 32;
	pf->BytesPerPixel = 4;
	pf->alpha = SDL_ALPHA_OPAQUE;

	// SDL interprets each pixel as a 32-bit number, so our masks must depend
    // on the endianness (byte order) of the machine
	#if SDL_BYTEORDER == SDL_BIG_ENDIAN
	pf->Rmask = 0xff000000; pf->Rshift = 24;
	pf->Gmask = 0x00ff0000; pf->Gshift = 16;
	pf->Bmask = 0x0000ff00; pf->Bshift = 8;
	pf->Amask = 0x000000ff; pf->Ashift = 0;
	#else
	pf->Rmask = 0x000000ff; pf->Rshift = 0;
	pf->Gmask = 0x0000ff00; pf->Gshift = 8;
	pf->Bmask = 0x00ff0000; pf->Bshift = 16;
	pf->Amask = 0xff000000; pf->Ashift = 24;
	#endif
}

SurfaceFactory::FactoryItem& SurfaceFactory::create(SDL_Surface * s, FactoryItem &item )
{
	TIN;
	item.surface = 0;
	item.type = GL_UNSIGNED_BYTE;
	item.format = GL_RGBA;	// format of the new surface

	// Note: item.type from packed pixels donesn't work for gluBuildMipmaps(...)!
	if (!s) 
		return item;

	// temporary surface for conversion to common OpenGL texture format
	SDL_Surface *newSurf;

	// texture pixel format
	SDL_PixelFormat tpf = *s->format;
	setDestPixelFormat(&tpf);
	
	// upper-left-corner origin; power-of-two alignment
	bool bULC = isULC();
	Uint32 w, h;
	if (!TextureManager::getInstance()->getNPOT()){
		w = toPOT(s->w);
		h = toPOT(s->h);
	}else{
		w = s->w;
		h = s->h;
	}

	newSurf = SDL_CreateRGBSurface(
		SDL_SWSURFACE, 
		w, h, 
		tpf.BitsPerPixel, 
		tpf.Rmask, tpf.Gmask, tpf.Bmask, tpf.Amask);
	
	if (!newSurf) {
		LOGE << "Implicit surface creation error: Can not convert SDL-surface!" << std::endl;
		SDL_FreeSurface(newSurf);
		return item;
	}
	// TODO: copy data from s to newSurf using tpf PixelFormat and some custom color-key and surface-alpha.
	// ...
	SDL_LockSurface(newSurf);
	SDL_LockSurface(s);

	SDL_PixelFormat 
		*pf1(s->format), 
		*pf2(newSurf->format);
	Uint32 
		xstop(std::min(w, Uint32(s->w))), ystop(std::min(h, Uint32(s->h))),
		bypp1(s->format->BytesPerPixel), bypp2(newSurf->format->BytesPerPixel);
	Sint32
		pitch2( (bULC ? 1 : -1) * Sint32(newSurf->pitch));
	Uint8
		*xstartp1(reinterpret_cast<Uint8*>(s->pixels)), 
		*xstartp2(reinterpret_cast<Uint8*>(newSurf->pixels) + ((bULC) ? 0 : (ystop-1) * newSurf->pitch)),
		*p1, *p2;

	if (!(xstartp1 && xstartp2))
		DIE("null texture pointers!");

	// loop through pixels of s and newSurf
	// p1 is pointing to actual pixel of s, p2 is pointing to actual pixel of newSurf
	for (Uint32 y = 0; y < ystop; ++y, xstartp1 += s->pitch, xstartp2 += pitch2) {
		p1 = xstartp1; p2 = xstartp2;
		for (Uint32 x = 0; x < xstop; ++x, p1 += bypp1, p2 += bypp2) 
		{
			putpixel(p2, convert(p1, pf1, pf2), pf2);
		}
	}

	SDL_UnlockSurface(s);
	SDL_UnlockSurface(newSurf);
	
	item.origWidth = s->w;
	item.origHeight = s->h;
	item.surface = newSurf;
	return item;
}

SurfaceFactory::FactoryItem& SurfaceFactory::create(FactoryItem &item)
{ 
	return create(getSrcSurface(), item);
}

SurfaceFactory::FactoryItem& SurfDefault::create(FactoryItem &item)
{ 
	item.surface = 0;
	
	SDL_PixelFormat fmt;
	setDestPixelFormat(&fmt);

	// Converting surface to the OpenGL native texture data format is easier than
	// trying to tell OpenGL something about srcImage data format explicitly.
	// However, it needs one more temporary surface.
	SDL_Surface * surf = getSrcSurface();

	if (surf)
		surf = SDL_ConvertSurface(surf, &fmt, SDL_SWSURFACE);

	if (!surf)
		return item;
	
	// things like ULC/LLC or NPOT knows how to do the SurfaceFactory.
	SurfaceFactory::create(surf, item);

	SDL_FreeSurface(surf);
	return item;
}


Uint32 SurfDefault::convert(Uint8 *pix, SDL_PixelFormat *pf1, SDL_PixelFormat *pf2)
{
	Uint8 r, g, b, a;
	SDL_GetRGBA(getpixel(pix, pf1), pf1, &r, &g, &b, &a);
	return SDL_MapRGBA(pf2, r, g, b, a);
}

Uint32 SurfExternalAlpha::convert(Uint8 *pix, SDL_PixelFormat *pf1, SDL_PixelFormat *pf2)
{
	Uint8 r, g, b, a;
	SDL_GetRGBA(getpixel(pix, pf1), pf1, &r, &g, &b, &a);
	return SDL_MapRGBA(pf2, r, g, b, SDL_ALPHA_OPAQUE);
}

Uint32 SurfColorKey::convert(Uint8 *pix, SDL_PixelFormat *pf1, SDL_PixelFormat *pf2)
{
	Uint8 r, g, b, a;
	SDL_GetRGBA(getpixel(pix, pf1), pf1, &r, &g, &b, &a);
	return SDL_MapRGBA(pf2, r, g, b, ((r == mKeyR) && (g == mKeyG) && (b == mKeyB)) ? SDL_ALPHA_TRANSPARENT : SDL_ALPHA_OPAQUE);
}

Uint32 SurfAlpha::convert(Uint8 *pix, SDL_PixelFormat *pf1, SDL_PixelFormat *pf2)
{
	Uint8 r, g, b, a;
	SDL_GetRGBA(getpixel(pix, pf1), pf1, &r, &g, &b, &a);
	return SDL_MapRGBA(pf2, r, g, b, mAlpha);
}

SurfExternalAlpha::SurfExternalAlpha(std::string imagePath, std::string alphaImagePath)
: SurfaceFactory(imagePath), mExternalAlphaOwned(false)
{
	TIN;

	// Load the image file into a SDL_Surface.
	mExternalAlpha = IMG_Load(alphaImagePath.c_str());
	if ( !mExternalAlpha ) {
		LOGE << "External alpha-channel surface loading: Can not load '" << alphaImagePath << "' (" << IMG_GetError() << ")" << std::endl;
	    return;
	}
	mExternalAlphaOwned  = true;
	suff(" ExtAlpha(" + alphaImagePath + ")");
}

SurfExternalAlpha::SurfExternalAlpha(SDL_Surface *src, std::string alphaImagePath)
: SurfaceFactory(src), mExternalAlphaOwned(false)
{
	TIN;

	// Load the image file into a SDL_Surface.
	mExternalAlpha = IMG_Load(alphaImagePath.c_str());
	if ( !mExternalAlpha ) {
		LOGE << "External alpha-channel surface loading: Can not load '"<< alphaImagePath << "' (" << IMG_GetError() << ")" << std::endl;
	    return;
	}
	mExternalAlphaOwned  = true;
	suff(" ExtAlpha(" + alphaImagePath + ")");
}

SurfExternalAlpha::~SurfExternalAlpha()
{
	if (mExternalAlphaOwned && mExternalAlpha){
		SDL_FreeSurface(mExternalAlpha);
		mExternalAlpha = 0;
		mExternalAlphaOwned = false;
	}
}

SurfaceFactory::FactoryItem&  SurfExternalAlpha::create(FactoryItem &item)
{
	TIN;
	SurfaceFactory::create(getSrcSurface(), item);
	SDL_Surface * s = item.surface;
	if (!s) {
		return item;
	}
	bool bULC = isULC();

	if (!mExternalAlpha){
		LOGE << "External alpha map not found." << std::endl;
		return item;
	}
	
	// Fill the alpha channel of just created surface s (takes only the red component of mExternalAlpha)
	
	SDL_LockSurface(mExternalAlpha);
	if (mExternalAlpha != s)
		SDL_LockSurface(s);

	SDL_PixelFormat 
		*pf1(s->format), 
		*pf2(mExternalAlpha->format);
	Uint32 
		xstop(std::min(s->w, mExternalAlpha->w)), 
		ystop(std::min(s->h, mExternalAlpha->h)),
		bypp1(pf1->BytesPerPixel), bypp2(pf2->BytesPerPixel);
	Sint32
		pitch2( (bULC ? 1 : -1) * Sint32(mExternalAlpha->pitch));
	Uint8
		*xstartp1(reinterpret_cast<Uint8*>(s->pixels)), 
		*xstartp2(reinterpret_cast<Uint8*>(mExternalAlpha->pixels) + ((bULC) ? 0 : (ystop-1) * mExternalAlpha->pitch)),
		*p1, *p2,
		r, g, b, a;

	if (!(xstartp1 && xstartp2))
		DIE("null texture pointers!");

	// loop through pixels of s and mExternalAlpha
	// p1 is pointing to actual pixel of s, p2 is pointing to actual pixel of mExternalAlpha
	for (Uint32 y = 0; y < ystop; ++y, xstartp1 += s->pitch, xstartp2 += pitch2) {
		p1 = xstartp1; p2 = xstartp2;
		for (Uint32 x = 0; x < xstop; ++x, p1 += bypp1, p2 += bypp2) {
			SDL_GetRGB(getpixel(p2, pf2), pf2, &r, &g, &b);
			a = r;
			SDL_GetRGB(getpixel(p1, pf1), pf1, &r, &g, &b);
			putpixel(p1, SDL_MapRGBA(pf1, r, g, b, a), pf1);
		}
	}

	if (mExternalAlpha != s)
		SDL_UnlockSurface(s);
	SDL_UnlockSurface(mExternalAlpha);
	
	return item;
}

Texture::Texture()
	: mID(0), mMaxS(0), mMaxT(0)
{
}

Texture::~Texture()
{
	// remove texture from OpenGL
	if (mID)
	{
		glDeleteTextures(1, &mID);
		mID = 0;
	}
}

void Texture::upload(SurfaceFactory * psf, GLint internalFormat, GLint level)
{	TIN;
	
	if (!psf || glIsTexture(mID))
		return;

	GLint	min_filter, mag_filter;

	// Create OpenGL-friendly surface for texture using SurfaceFactory
	SurfaceFactory::FactoryItem fi;
	SDL_Surface * textureSurf = psf->create(fi).surface;
	if (!textureSurf)
		return;

	const GLenum format(fi.format); // pixel data format - can be any supported OpenGL texture format
	const GLenum type(fi.type); // type of pixel components
	
	// Compute texture coordinates maxima and save it to mMaxS and mMaxT
	
	if (textureSurf->w != fi.origWidth)
		mMaxS = (textureSurf->w) ? float(fi.origWidth) / textureSurf->w : 0;
	else
		mMaxS = 1.0f;
	
	if (textureSurf->h != fi.origHeight)
		mMaxT = (textureSurf->h) ? float(fi.origHeight) / textureSurf->h : 0;
	else
		mMaxT = 1.0f;
	
	// Generate texture ID.
	glGenTextures(1, &mID);

	// Set this texture as current.
	bind();
	
	glPushClientAttrib(GL_CLIENT_PIXEL_STORE_BIT);

	// Setup pixel store
	// - select the best-fit alignment
	{
		// Calculate the maximal possible alignment (up to 8) in the good hope that
		// driver can optimize aligned rows:
		GLint pitch( textureSurf->pitch );
		GLint alignment = 
			static_cast<GLint>( 1 ) +						// only multiple of 1		-> alignment = 1
			static_cast<GLint>( (pitch & 1) == 0 ) +		// only multiple of 1, 2	-> alignment = 2
			static_cast<GLint>( (pitch & 3) == 0 ) * 2 +	// only multiple of 1, 2, 4 -> alignment = 4
			static_cast<GLint>( (pitch & 7) == 0 ) * 4;		// multiple of 1, 2, 4, 8	-> alignment = 8
		glPixelStorei( GL_UNPACK_ALIGNMENT, alignment );
		
		// Check alignment validity:
		assertL( "Bad texture row alignment!" && pitch == ( ( textureSurf->w * textureSurf->format->BytesPerPixel + alignment - 1) / alignment ) * alignment );
	}
	// - set row length
	glPixelStorei( GL_UNPACK_ROW_LENGTH, textureSurf->w );
	// - set other values - ensure they are set to default
	glPixelStorei(GL_UNPACK_SWAP_BYTES, GL_FALSE);
	glPixelStorei(GL_UNPACK_LSB_FIRST, GL_FALSE);
	glPixelStorei(GL_UNPACK_SKIP_ROWS, 0);
	glPixelStorei(GL_UNPACK_SKIP_PIXELS, 0);

	// Copy image data from SDL_Surface to OpenGL.
	SDL_LockSurface(textureSurf);
	if (level >= 0)
	// custom mipmap level
	{
		glTexImage2D(
			GL_TEXTURE_2D,		// target texture - must be GL_TEXTURE_2D
			level,				// level-of-detail number
			internalFormat,		// internal format
			textureSurf->w,		// width (must be 2^m + 2(border))
			textureSurf->h,		// height (must be 2^n + 2(border))
			0,					// border size (0 or 1)
			format,				// pixel data format
			type,
			textureSurf->pixels	// image data pointer
			);
		min_filter = GL_NEAREST;
		mag_filter = GL_LINEAR;
	}
	else if (level == BUILD_MIPMAPS)
	// generate all mipmaps
	{
		if ( GLEW_VERSION_1_4 || GLEW_SGIS_generate_mipmap )
		// automatic mipmap generation is supported
		{
			glTexParameteri( GL_TEXTURE_2D, GLEW_VERSION_1_4 ? GL_GENERATE_MIPMAP : GL_GENERATE_MIPMAP_SGIS, GL_TRUE );

			glTexImage2D(
				GL_TEXTURE_2D,		// target texture - must be GL_TEXTURE_2D
				0,					// level-of-detail number
				internalFormat,		// internal format
				textureSurf->w,		// width (must be 2^m + 2(border))
				textureSurf->h,		// height (must be 2^n + 2(border))
				0,					// border size (0 or 1)
				format,				// pixel data format
				type,
				textureSurf->pixels	// image data pointer
			);
		} 
		else 
		// no support for automatic mipmaps, let GLU to generate mipmaps
		{
			// pitfall: gluBuild2DMipmaps nezna formaty z packed pixels extension!!
			// Takze napriklad GL_UNSIGNED_INT_8_8_8_8_REV z OpenGL 1.2 neprojde jako <type> parametr
			gluBuild2DMipmaps(
				GL_TEXTURE_2D,
				internalFormat,
				textureSurf->w,
				textureSurf->h,
				format,
				type,
				textureSurf->pixels
			);
		}
		min_filter = GL_LINEAR_MIPMAP_NEAREST;
		mag_filter = GL_LINEAR_MIPMAP_LINEAR;
	}
	else
	// unknown texture mipmap level
	{
		SDL_UnlockSurface( textureSurf );
		SDL_FreeSurface( textureSurf );
		textureSurf = 0;
		LOGE << "Uploading texture: Unknown mipmap level '" << level << "'!" << std::endl;
		glPopClientAttrib();
		return;
	}
	SDL_UnlockSurface(textureSurf);	
	SDL_FreeSurface(textureSurf);
	textureSurf = 0;
	
	glPopClientAttrib();

	// Set texture 'My Default' filtering methods.
	// (This is necessary for glTexImage2D(), because OpenGL uses mipmaps for MIN_FILTER as default!!)
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, min_filter);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, mag_filter);
}

void Texture::bind()
{
	// Clear previously set error flags.
	glGetError();

	// Bind the texture given by mID to the current texture unit.
	glBindTexture(GL_TEXTURE_2D, mID);
	
	// Pop possible error code and if error is set, log the error string.
	GLenum err = glGetError();
	if (err != GL_NO_ERROR)
		LOGE << "Texture binding error: " << gluErrorString(glGetError()) << std::endl;
}

void Texture::bindDirectly( GLenum textureUnitID )
{	
	glActiveTexture(textureUnitID);
	bind();
}
