/* MAINTAINER: Martin */

#ifndef _TEXTURE_MANAGER_H
#define _TEXTURE_MANAGER_H

#include <string>
#include "log.h"

#include "SDL.h"
#include "SDL_opengl.h"

#include "universal_manager.h"

class SurfaceFactory;

/**
*	\brief Object that represents 2D texture, managed by TextureManager.
*
*	This class controls one OpenGL texture. Texture manager is able to call upload() method, that gives it
*	the ability of creating OpenGL-friendly surface and uploading it to OpenGL. Texture is deleted from
*	OpenGL in destructor.
*
*	\note This class is not for direct using, use it rather in the form that the TextureManager provides to you.
*/
class Texture
{
friend class TextureManager; ///< \brief friendship allows the TextureManager to call protected upload() method

private:
	GLuint	mID;			///< \brief OpenGL texture object identifier (zero = invalid texture)
	
	/**
	*	\brief range of texture coordinates
	*
	*	If the texture manager doesn't support NPOT textures, each one of those should lie in interval (0, 1) 
	*	in the case non-power-of-two extents, 1.0 otherwise.
	*
	*	In all cases, the domain of this 2D texture is <tt>[0, mMaxS]x[0, mMaxT]</tt>.
	*/
	GLfloat	mMaxS, mMaxT;

public:

	/**
	*	\brief special \c level value of upload() method.
	*
	*	If this value is passed to argument \c level of upload() method, mipmap levels 
	*	will be automatically generated using gluBuild2DMipmaps().
	*/
	static const GLint BUILD_MIPMAPS = -1;

	/**
	*	\brief Ctor - simple initialization.
	*/
	Texture();

	/**
	*	\brief Dtor - will free the OpenGL texture if this one exists.
	*/
	~Texture();

	/**
	*	\brief gets the ID of this OpenGL texture.
	*	\return OpenGL texture ID or 0 if invalid.
	*/
	GLuint	getID() { return mID; }
	
	/**
	*	\brief gets the maximum S texture coordinate
	*	\return maximum S coordinate
	*	\see mMaxS, mMaxT, bind()
	*/
	GLfloat	getMaxS() { return mMaxS; }
	
	/**
	*	\brief gets the maximum T texture coordinate
	*	\return maximum T coordinate
	*	\see mMaxS, mMaxT, bind()
	*/
	GLfloat	getMaxT() { return mMaxT; }

	/**
	*	\brief binds this texture to current active texturing unit.
	*
	*	It uses the glBindTexture call to bind the texture object represented by mID.
	*
	*	\note If you want the NPOT texture map on some specific interval, yoy can modify the
	*		texture matrix (preserved for each texture unit) using getMaxS() and getMaxT().
	*		Binding texture destroys the previous texture unit context.
	*/
	void bind();

	/**
	*	\brief binds the texture directy to a specified texture unit
	*
	*	Bindiding a texture to other texture unit than GL_TEXTURE0_ARB requires GL_ARB_multitexture extension.
	*
	*	\param textureUnitID texture unit identificator
	*	\note textureUnitID should be in the form of GL_TEXTUREi_ARB, where 0 <= i < GL_MAX_TEXTURE_UNITS_ARB
	*/
	void bindDirectly( GLenum textureUnitID );
	
	/**
	*	\brief binds the texture to a specified texture unit using texture unit number
	*	\param textureUnit texture unit number, 0 <= textureUnit < max texture units count
	*/
	void bind( unsigned int textureUnit )
	{ bindDirectly( firstTextureUnitID + textureUnit ); }
	
protected:

	/**
	*	\brief Uploads the texture from external source to current texture unit.
	*
	*	A factory-mechanism ( SurfaceFactory ) is used for creating the OpenGL-friendly surface and passing source 
	*	parameters (file name or surface, transparency type, or you can implement our own). Texture coordinates bound
	*	(mMaxS and mMaxT) are computed. A texture is allocated in OpenGL, bound, and pixels are copied from the surface 
	*	created by \c psf and min-NEAREST mag-LINEAR filter is set by default (mipmapping is taken care of).
	*
	*	\param psf Pointer to some Texture::SurfaceFactory implementation (here you can fill up the file name or 
	*		source surface). This is used for creating an OpenGL-friendly surface from this source.
	*	\param internalFormat viz. glTexImage2D, the number of color components of the texture (1..4)
	*	\param level the texture mipmap level. 0 is the base image, n is the n-th mipmap reduction, or
	*		Texture::BUILD_MIPMAPS for automatic mipmaps generation.
	*/
	void upload(SurfaceFactory * psf, GLint internalFormat = GL_RGBA, GLint level = 0);

public:

	static const GLenum firstTextureUnitID;
};

/**
*	\brief manages all textures
*	
*	This is a singleton.
*	It provides several <code>add(...)</code> methods to the user. You describe texture
*	source in arguments of add() and you will get the reference TRef, with that you can 
*	access the Texture class. If there is not any texture managed matching given description, 
*	it will be created and uploaded to OpenGL. Otherwise only the reference to managed 
*	item found, is returned.
*
*	Example:
*	\code
*	// Loads the file "image.png" with an external alpha channel describet as red channel od "image_alpha.png"
*	// (i.e. some grayscale image), builds 2D mipmaps and returns texture reference to it:
*	TextureRef tr1 = TextureManager::getInstance()->add(
*						&SurfExternalAlpha("image.png", "image_alpha.png"), 
*						GL_RGBA, Texture::BUILD_MIPMAPS);
*
*	// Finds the texture in the registry and returns only a reference to it
*	// (GL_RGBA and BUILD_MIPMAPS is ignored when image found in registry):
*	TextureRef tr2 = TextureManager::getInstance()->add(
*						&SurfExternalAlpha("image.png", "image_alpha.png"), 
*						GL_RGBA, Texture::BUILD_MIPMAPS);
*
*	// Increments the reference count only..
*	TextureRef tr3 = tr1;
*
*	// Now, tr1, tr2, tr3 refers to only one texture.
*
*	// ...
*	// binds the texture to current texture unit
*	t2->bind();
*
*	// Loads image2.png with an explicit color key (0x00, 0x00, 0x00):
*	tr1 = TextureManager::getInstance()->add(
*						&SurfColorKey("image2.png", 0x00, 0x00, 0x00));
*
*	// SDL_Surface * customSurfacePtr = getSomeSurface();
*	// Loads the texture from custom surface with explicit color key and prefix its name with "surface1"
*	tr2 = TextureManager::getInstance()->add(
*						&SurfColorKey(customSurfacePtr, 0x00, 0x00, 0x00).pref("surface1"));
*
*	// release last reference to the first texture "image.png"
*	tr3 = 0;
*
*	// the first texture is now released...
*	\endcode
*
*	\note You should free all texture references before graphics subsystem exits to avoid 
*		warning logs.
*	\note OpenGL has to be initialized before registering any texture!
*/
class TextureManager: public UniversalManager<Texture, TextureManager>
{
friend class Singleton<TextureManager>;	///< due to singleton DP
protected:
	TextureManager();
	~TextureManager(){}

public:
	/**
	*	\brief Registers the texture using arbitrary SurfaceFactory.
	*	
	*	If this texture (described by \c psf) doesn't exist, it will be created from source described
	*	in psf, uploaded to current texture unit and a reference to it will be returned.
	*	If this texture already exists (is registered yet), only a reference will be returned.
	*
	*	\param psf pointer to surface factory holding texture description
	*	\param internalFormat see glTexImage2D, the number of color components of the texture (1..4)
	*	\param level the texture mipmap level. 0 is the base image, n is the n-th mipmap reduction, or
	*		Texture::BUILD_MIPMAPS for automatic mipmaps generation.
	*	\return reference to managed data providing access to Texture class
	*/
	TRef add(SurfaceFactory * psf, GLint internalFormat = GL_RGBA, GLint level = 0);
	
	/**
	*	\brief Registers the texture using default factory with the file name only
	*	\param fileName the name of external image-file
	*/
	TRef add(std::string fileName, GLint internalFormat = GL_RGBA, GLint level = 0);

	/**
	*	\brief Non-power-of-two textures extension presence
	*	\return true, if the GL_ARB_texture_non_power_of_two extension is present, false if not present.
	*/
	bool getNPOT(void) { return mNPOT; }

private:
	bool mNPOT;	///< \brief true for NPOT textures extension presence, false otherwise

};

/**
*	\brief base abstract factory class which creates SDL_Surface and hash string of texture for TextureManager
*
*	It is able to create a FactoryItem object containing OpenGL-friendly SDL surface, its original 
*	extents and OpenGL pixel format of texture. TextureManager class uses this for creation of surface that 
*	it uploads to OpenGL afterwards. SurfaceFactory knows how to convert a surface to power-of-two 
*	extents, convert it to lower-left-corner origin and provides some basic methods for derived classes.
*
*	As a source you can specify a SDL_Surface or path to a image file. Various derived classes can provide
*	more source parameters and more functionality to creation of the destination surface.
*
*	\see SurfaceFactory::FactoryItem::surface for description about "OpenGL-friendly"
*	
*/
class SurfaceFactory
{
public:
	/**
	*	\brief structure providing destination surface and its parameters
	*
	*	This is the "product" that the user of this surface factory will get.
	*/
	struct FactoryItem
	{
		/**
		*	\brief OpenGL-friendly surface
		*
		*	OpenGL-friendly means 32bpp, RGBA (in GL_RGBA order), lower-left-corner (LLC) origin, 
		*	power-of-two extents if necessary. Some of those properties can be explicitly changed.
		*/
		SDL_Surface * surface;
		int origWidth, origHeight;	///< \brief extents of the source image
		GLenum format;				///< \brief format of pixels for OpenGL texture (i.e. GL_RGBA)
		GLenum type;				///< \brief type of pixel component
	};

protected:
	
	/**
	*	\brief ctor using SDL_Surface source
	*
	*	You should explicitly specify the hash string using the suff() or pref() methods, it is useful for the
	*	texture manager.
	*
	*	\param source surface - factory will read its pixels and store them into a destination surface
	*/
	SurfaceFactory(SDL_Surface *src);
	
	/**
	*	\brief ctor using imagePath
	*
	*	If external file is specified, the factory will attempt to load it using SDL_Image library later.
	*	Hash string for the texture manager is set to \c imagePath.
	*
	*	\param imagePath path to file with an image readable by IMG_Load() - part of the SDL_Image library.
	*/
	SurfaceFactory(std::string imagePath)
		: mSrcSurface(0), mSrcSurfaceOwned(false), mULC(false), mImagePath(imagePath), mHashString(imagePath)
	{}

public:
	/**
	*	\brief dtor - releases intrinsic temporary surface, if allocated 
	*
	*	Temporary surface is allocated in the case of ctor using the \c imagePath as a source (external file).
	*	It is allocated in getSrcSurface() if source surface doesn't exist.
	*/
	virtual ~SurfaceFactory();

	/**
	*	\brief factory method, fills up the FactoryItem structure (produces the OpenGL friendly surface).
	*
	*	It is used by the user to actualy create the factory item.
	*
	*	At this class level, it simply calls the create(SDL_Surface*, FactoryItem&) method only to
	*	perform basic conversion (see this method). However, derived classes can provide more powerful
	*	code.
	*
	*	\param item a structure for holding the data that will be created
	*	\return the \c item passed into its argument for easy access
	*
	*	\note if this method fails, \c item.surface will be set to 0. Otherwise, \c item.surface will be
	*		set to destination surface and the user is responsible for releasing this surface by calling the 
	*		SDL_FreeSurface() method.
	*/
	virtual FactoryItem& create(FactoryItem &item);

	/**
	*	\brief provides hash string based on source data
	*	\return hash string for the TextureManager class
	*/
	virtual std::string getHashString() { return mHashString; }
	
	/**
	*	\brief appends a suffix to the internal hash string
	*
	*	Suffix or prefix should the user use when creating factory product using SDL_Surface as input, or
	*	any input with that hash string isn't supported.
	*
	*	\param str suffix string
	*	\return \c *this instance for chaining commands and compact code
	*/
	SurfaceFactory& suff(std::string str) { mHashString += str; return *this; }

	/**
	*	\brief inserts a prefix to the internal hash string
	*
	*	Suffix or prefix should the user use when creating factory product using SDL_Surface as input, or
	*	any input with that hash string isn't supported.
	*
	*	\param str suffix string
	*	\return \c *this instance for chaining commands and compact code
	*/
	SurfaceFactory& pref(std::string str) { mHashString = str + mHashString; return *this; }

	/**
	*	\brief enables or disables the Upper Left Corner origin
	*	
	*	After enabling ULC, (0,0)-origin is at upper left corner, x increases right, y increases down.
	*	By default, LLC (lower left corner) is set, so images that will be created by create(SDL_Surface*, FactoryItem&)
	*	will have (0,0)-origin at their LLC (src is mirrored along the x axis). This is for support of OpenGL's
	*	LLC origin.
	*
	*	\param ulc set true to enable ULC, false to disable ULC ( = enable LLC)
	*	\return *this instance for chaining commands and compact code
	*/
	SurfaceFactory& setULC(bool ulc) { mULC = ulc; return *this; }
	
	/**
	*	\brief ULC/LLC determination
	*	
	*	\return true, if ULC (upper-left-corner) origin is enabled, false otherwise.
	*/
	bool isULC(void){ return mULC; }
	
protected:
	/**
	*	\brief gets the source surface
	*
	*	The source surface is returned by this method. If the source surface doesn't exist yet (in the case
	*	of ctor using only the path to image file), the surface is loaded using IMG_Load and then returned.
	*	Owner of surface created this way is this instance.
	*
	*	\return the source image surface - input data for this factory
	*/
	SDL_Surface * getSrcSurface();	
	
	
	/**
	*	\brief creates the OpenGL-friendly surface that can be used to define a texture
	*
	*	Creates the destination surface using OpenGL friendly format (taken from setDestPixelFormat() ) and
	*	transfers source pixels to this destination surface regarding the ULC flag and POT support using
	*	the polymorphic convert() method.
	*
	*	\param s source surface
	*	\param item a structure for holding the data that will be created
	*	\return the \c item passed into its argument for easy access
	*
	*	\see create(FactoryItem &item);
	*/
	FactoryItem& create(SDL_Surface *s, FactoryItem& item);

	/**
	*	\brief converts source pixel to destination pixel using source and destination pixel formats
	*
	*	Behavior of this polymorphic method can be redefined in derived classes and it can provide for 
	*	example a color-key filter or set explicit alpha, etc.
	*
	*	\param pix adress of source pixel
	*	\param pf1 source pixel format
	*	\param pf2 destination pixel format
	*	\return pixel value in destination source format
	*/
	virtual Uint32 convert(Uint8 *pix, SDL_PixelFormat *pf1, SDL_PixelFormat *pf2) = 0;

	/**
	*	\brief fills up the pixel format of the destination surface
	*
	*	By default, bpp is set to 32, 4 bytes per pixel, per-surface alpha is set to SDL_ALPHA_OPAQUE and
	*	R-,G-,B-,A-mask are set according to endianness to 0xFF, shifts are set and other fields are set to zero.
	*
	*	\param pf destination pixel format
	*/
	void setDestPixelFormat(SDL_PixelFormat *pf);

	/**
	*	\brief converts the \c arg to the nearest upper power-of-two value
	*	\return the nearest upper power-of-two value
	*/
	static Uint32 toPOT(Uint32 arg);

private:

	SDL_Surface * mSrcSurface;	///< \brief source surface
	bool mSrcSurfaceOwned;		///< \brief flag determining ownership of this instance over the source surface
	bool mULC;					///< \brief flag for upper-left-corner origin enabled
	std::string mImagePath;		///< \brief path to file with image data, if specified in ctor
	std::string mHashString;	///< \brief hash string for the TextureManager
};

/**
*	\brief surface factory using default surface properties (surface alpha, colorkey, alpha channel, flags)
*/
class SurfDefault: public SurfaceFactory{
public:
	/**
	*	\brief ctor using SDL_Surface source
	*	\param source surface
	*/
	SurfDefault(SDL_Surface *src):
	  SurfaceFactory(src)
	{}
	
	/**
	*	\brief ctor using imagePath
	*	\param imagePath path to file with an image readable by IMG_Load()
	*/
	SurfDefault(std::string imagePath):
	  SurfaceFactory(imagePath)
	{}

	~SurfDefault(){}

	/**
	*	\brief factory method, fills up the FactoryItem structure (produces the OpenGL friendly surface).
	*
	*	Uses SDL's rules for creating the destination surface. More specificly, it uses the SDL_ConvertSurface()
	*	function to convert the source surface to the destination surface.
	*
	*	\param item a structure for holding the data that will be created
	*	\return the \c item passed into its argument for easy access
	*
	*	\note if this method fails, \c item.surface will be set to 0. Otherwise, \c item.surface will be
	*		set to destination surface and the user is responsible for releasing this surface by calling the 
	*		SDL_FreeSurface() method.
	*/
	FactoryItem& create(FactoryItem &item);

protected:
	/**
	*	\brief converts source pixel to destination pixel using source and destination pixel formats
	*
	*	Simply converts pixels between two pixel formats.
	*
	*	\param pix adress of source pixel
	*	\param pf1 source pixel format
	*	\param pf2 destination pixel format
	*	\return pixel value in destination source format
	*/
	Uint32 convert(Uint8 *pix, SDL_PixelFormat *pf1, SDL_PixelFormat *pf2);
};

/**
*	\brief surface factory using external file for defining the alpha channel of the destination surface
*/
class SurfExternalAlpha: public SurfaceFactory{
public:
	/**
	*	\brief ctor using source surface and second surface for defining the alpha channel
	*	\param src source surface
	*	\param externalAlpha external surface defining the alpha channel (in its red channel)
	*/
	SurfExternalAlpha(SDL_Surface *src, SDL_Surface * externalAlpha)
		: SurfaceFactory(src), mExternalAlpha(externalAlpha) {}

	/**
	*	\brief ctor using imagePath and external surface for defining the alpha channel
	*	\param imagePath path to the source file
	*	\param externalAlpha external surface defining the alpha channel (in its red channel)
	*/
	SurfExternalAlpha(std::string imagePath, SDL_Surface * externalAlpha)
		: SurfaceFactory(imagePath), mExternalAlpha(externalAlpha) {}

	/**
	*	\brief ctor using imagePath and external image file for defining the alpha channel
	*	\param imagePath path to the source file
	*	\param externalAlpha path to the image file defining the alpha channel (in its red channel)
	*/
	SurfExternalAlpha(std::string imagePath, std::string alphaImagePath);

	/**
	*	\brief ctor using source surface and external image file for defining the alpha channel
	*	\param src source surface
	*	\param externalAlpha path to the image file defining the alpha channel (in its red channel)
	*/
	SurfExternalAlpha(SDL_Surface *src, std::string alphaImagePath);
	
	~SurfExternalAlpha();

	/**
	*	\brief factory method, fills up the FactoryItem structure (produces the OpenGL friendly surface).
	*
	*	As a alpha channel of the destination surface uses the red channel of specified second surface
	*	(\c externalAlpha ).
	*
	*	\param item a structure for holding the data that will be created
	*	\return the \c item passed into its argument for easy access
	*
	*	\note if this method fails, \c item.surface will be set to 0. Otherwise, \c item.surface will be
	*		set to destination surface and the user is responsible for releasing this surface by calling the 
	*		SDL_FreeSurface() method.
	*/
	FactoryItem& create(FactoryItem &item);

protected:
	
	/**
	*	\brief converts source pixel to destination pixel using source and destination pixel formats and sets opaque alpha channel
	*
	*	This is used for the "middle-step" in the create() fucntion.
	*
	*	\param pix adress of source pixel
	*	\param pf1 source pixel format
	*	\param pf2 destination pixel format
	*	\return pixel value in destination source format
	*/
	Uint32 convert(Uint8 *pix, SDL_PixelFormat *pf1, SDL_PixelFormat *pf2);

private:

	SDL_Surface * mExternalAlpha;	///< \brief external surface defining the alpha in its red channel
	bool mExternalAlphaOwned;		///< \brief flag that says if the \c mExternalAlpha is owned (and should be freed) by this instance or not
};

/**
*	\brief surface factory using explicit specified color key value for definition the alpha channel of the destination surface
*/
class SurfColorKey: public SurfaceFactory{
public:
	
	/**
	*	\brief ctor using the source surface and explicit color key
	*
	*	\param src	source surface
	*	\param keyR	red component of the color key
	*	\param keyG green component of the color key
	*	\param keyB blue component of the color key
	*	\note It also creates the suffix of the hash string that includes the color key.
	*/
	SurfColorKey(SDL_Surface *src, Uint8 keyR, Uint8 keyG, Uint8 keyB)
		: SurfaceFactory(src), mKeyR(keyR), mKeyG(keyG), mKeyB(keyB) 
	{ suff(createHashSuffix()); }
	
	/**
	*	\brief ctor using the image path and explicit color key
	*
	*	\param imagePath	path to source image
	*	\param keyR	red		component of the color key
	*	\param keyG green	component of the color key
	*	\param keyB blue	component of the color key
	*	\note It also creates the suffix of the hash string that includes the color key.
	*/
	SurfColorKey(std::string imagePath, Uint8 keyR, Uint8 keyG, Uint8 keyB)
		: SurfaceFactory(imagePath), mKeyR(keyR), mKeyG(keyG), mKeyB(keyB) 
	{ suff(createHashSuffix()); }
	
	~SurfColorKey(){}

protected:
	
	/**
	*	\brief converts source pixel to destination pixel using source and destination pixel formats and sets the destination alpha with respect to colorkey
	*
	*	\param pix adress of source pixel
	*	\param pf1 source pixel format
	*	\param pf2 destination pixel format
	*	\return pixel value in destination source format
	*/
	Uint32 convert(Uint8 *pix, SDL_PixelFormat *pf1, SDL_PixelFormat *pf2);

private:
	Uint8 mKeyR, mKeyG, mKeyB;			///< \brief the colorkey
	
	/**
	*	\brief creates a suffix of the hash string that includes the colorkey information
	*	\return the suffix created
	*/
	std::string createHashSuffix();		
};


/**
*	\brief surface factory using explicit specified per-surface alpha value for definition of the alpha channel of the destination surface
*/
class SurfAlpha: public SurfaceFactory{
public:
	
	/**
	*	\brief ctor using source surface and specified per-surface alpha
	*	\param src source surface
	*	\param alpha explicit per-surface alpha
	*	\note It also creates the suffix of the hash string that includes the alpha.
	*/
	SurfAlpha(SDL_Surface *src, Uint8 alpha)
		: SurfaceFactory(src), mAlpha(alpha)
	{ suff(createHashSuffix()); }
	
	/**
	*	\brief ctor using source image and specified per-surface alpha
	*	\param imagePath source image path
	*	\param alpha explicit per-surface alpha
	*	\note It also creates the suffix of the hash string that includes the alpha.
	*/
	SurfAlpha(std::string imagePath, Uint8 alpha)
		: SurfaceFactory(imagePath), mAlpha(alpha)
	{ suff(createHashSuffix()); }
	
	~SurfAlpha(){}

protected:
	/**
	*	\brief converts source pixel to destination pixel using source and destination pixel formats and 
	*		sets the destination alpha with respect to explicitly specified alpha value
	*
	*	\param pix adress of source pixel
	*	\param pf1 source pixel format
	*	\param pf2 destination pixel format
	*	\return pixel value in destination source format
	*/
	Uint32 convert(Uint8 *pix, SDL_PixelFormat *pf1, SDL_PixelFormat *pf2);

private:
	Uint8 mAlpha;	///< \brief explicitly specified per-surface alpha
	
	/**
	*	\brief creates a suffix of the hash string that includes the explicitly 
	*		specified per-surface alpha information
	*	\return the suffix created
	*/
	std::string createHashSuffix();
};


typedef TextureManager::TUniversalManager::TRef TextureRef; ///< \brief texture reference that can be used instead of the TextureManager::TRef

// Used fot the explicit template instantiation in template_instantces_*.cpp file
#ifdef INSTANTIATE_TEMPLATE
	template class Singleton<TextureManager>;
#endif

#endif
