/* MAINTAINER: Bernard */

#ifndef _APPLICATION_H
#define _APPLICATION_H

/**	\defgroup common Common
 *
 *	\brief Code common for client and server
 *
 *	In this module is code that is used in both server and client. 
 *	Contains classes and interfaces that can, eventually, be derived
 *	into client/server -specific subclasses.
 *	Also contains \link game_data Game data \endlink , essential for 
 *	everything in game.
 */


class Commander;
class Game;
class TaskManager;

/**
 * \brief Base class for both server and client application class.
 *
 * It serves as some kind of interface. For example if we will need
 * application to provide something to common objects, we may use this.
 * And Martin wished to have this class :-)
 * This is abstract class that can't be created directly. It should only
 * be usede as base class for ApplicationServer and/or ApplicationClient.
 * This class hold is all together.
 */
class Application
{
public:
	
	/**
	 * \brief Calls initBasicSubsystems and allocate mTaskManager.
	 * 
	 * \param argc Variable argc from main.
	 * \param argv Variable argv from main.
	 */
	Application(int argc, char* argv[]);
	
	/**
	 * \brief Calls cleanBasicSubsystems anc free mTaskManager.
	 */
	virtual ~Application();

	/**
	 * \brief Inicialization of Application.
	 */
	virtual void init() = 0;

	/**
	 * \brief Running the Application.
	 */
	virtual void run() = 0;
	
	/**
	 * \brief Cleaning up the Application.
	 */
	virtual void clean() = 0;

	/**
	 * \brief Callback from engine that the Application should quit.
	 * 
	 * It may be generated by user from menu, from console, by clicnking
	 * on X button in window title bar... If you think that user wishes
	 * to quit the application, cal this.
	 */
	virtual void quit() = 0;

	/**
	 * \brief Get pointer to Game. 
	 * 
	 * If you are in client or server code use getGame() which return
	 * server or client version of Game so you have acces to more features :-)
	 * Please never retype this GameClient or GameServer.
	 * \return pointer common game
	 */
	virtual Game& getGameCommon() = 0;

	/**
	 *\brief Provide access to the commander.
	 *
	 * This may be used to push commands to commander anywhere from
	 * the code.
	 */
	virtual Commander& getCommander() = 0;

	/**
	 * \brief Provide access to task manager
	 *
	 * This task manager groups all task that needed to be called
	 * periodicaly. So if you have some loop you should call
	 * app.getTaskManager().update();
	 * Or if you have you own task manager you may (and should) add it as task
	 * to your manager.
	 */
	virtual TaskManager& getTaskManager();

	
protected:

private:
	/**
	 * \brief Does some basic inicialization.
	 *
	 * Loads settings, inicialize loging, inicialize SDL.
	 * clean for this inicialization is in cleanBasicSubsystems()
	 * Should be called early in application. Better it should be the first
	 * executed code - it is called in contructor of Application so
	 * it is quite early.
	 * It is important becase it sets up loging and loads configuration, co
	 * other objects may read variables in contructor from settings.
	 */
	void initBasicSubsystems();
	
	/**
	 * \brief Clean what initBasicSubsystems inicialized.
	 *
	 * quit SDL.
	 */
	void cleanBasicSubsystems();

protected:	
	/**
	 * \brief Load user config file.
	 *
	 * Create if doesn't exist.
	 * \param file is resolved using g_settings 
	 * \param file_template is resolved using g_settings 
	 */
	void loadUserConfigFile(const char *file, const char *file_template);

	/**
	 * \brief Write some basic info about program to log - Info pririty.
	 */
	void logBasicInfo();

	/**
	 * \brief Save user config files.
	 * 
	 * \param file File where to save.
	 */	
	void saveUserConfigFile(const char *file, const char *file_template);

	TaskManager *mTaskManager; ///< \brief global version of task manager. Allways != 0
	
	int    mArgc;  ///< argc from constructor
	char **mArgv; //< argv from contructor
};

#endif // _APPLICATION_H
