/* MAINTAINER: Bernard */

#ifndef _COMMANDER_H
#define _COMMANDER_H


#include <string>
#include <sstream>
#include <list>


#include "mutex.h"
#include "task.h"
#include "stlextdefs.h"
#include "hash_hack.h"
#include "globals.h"

/**
 * The commands and functions for commands are stored in hash map. Its typedef is here.
 */
typedef STL_EXT_NS::hash_map<std::string, void (*)(std::stringstream&, int)>  commads_hash;

/**
 * \brief Processor of server commands.
 * 
 * Accepts users commands from console and from network.
 * Inherits from Task so you may put it in task manager.
 * The commander needs regular processAllCommands(); call
 * or update call will do the same job.
 */
class Commander : public Task
{
public:
	Commander() { };
	virtual ~Commander() { };

	/**
	 * \brief Add command to command queues.
	 *
	 * The command will be processed later in main thread.
	 * This function is thread safe.
	 * Add one element to mCommQueue. Command will be processed
	 * in processAllCommands() call.
	 * \param command string containing command that contains
	 *                the command and it arguments (if any)
	 * \param playerID ID of player who created this command. You don't have to
	 *                 set it on client side - it will be set automaticly in receiver
	 *                 of commands on server. Value INVALID_PLAYER_ID is invalid value and indicates
	 *                 unknown source like internal command or command from server console.
	 *                 The ID is used when command provide some output so we need to send
	 *                 it back to player who created it.
	 */
	void addCommand(const std::string &command, int playerID = INVALID_PLAYER_ID);

	/**
	 * \brief Process all commands in queue
	 *
	 * This should be regullary called from main thread.
	 * Simplye emty mCommQueue and call processCommand on every
	 * member if the list.
	 */
	void processAllCommands();

	/**
	 * \brief Equal to processAllCommands().
	 *
	 * This fucntion is here if you need to put commander in
	 * task manager.
	 */
	virtual void update(int atime);
	
protected:

	/**
	 * \brief Process one command.
	 * 
	 * This function will be called in main task. Should process the command.
	 * \param command string with command
	 * \param playerID ID of player who created the command
	 */
	virtual void processCommand(const std::string &command, int playerID) = 0;

	Mutex mQueueMutex; /// Mutex for commands queue

	/** 
	 * \brief Holds info abou one message in queue.
	 */
	typedef struct sCommand
	{
		/// \brief Just simple inicialization
		sCommand(const std::string &cmd, int plID)
		{
			command  = cmd;
			playerID = plID;
		}
		std::string command;  ///< string with command
		int         playerID; ///< ID of player who created the command
	} sCommand;

	/**
	 * \brief Queue of commands that are waiting for processing.
	 *
	 * Use mQueueMutex when accessing the.
	 */
	std::list<sCommand> mCommQueue;

	
	/**
	 * \brief Hash table that holds all commands.
	 */
	commads_hash mCommandsHash;
};



#endif
