#ifndef _FUNCTION_2D_
#define _FUNCTION_2D_

#include <vector>

/**
*	\brief interface for function defined at 2D real support
*/
template <class T>
class Function2D
{
public:
	typedef T value_type;
	typedef std::vector< value_type > buffer_type;

	virtual ~Function2D() {}

	/**
	*	\brief sample values at given row and stores them into some preallocated vector
	*	\param x1 start position on row y
	*	\param x2 end position on row y
	*	\param y row
	*	\param sampleX one step in x axis (row)
	*	\param sampleY one step in y axis (column)
	*	\param first where to output the first result
	*	\note The preallocated vector (the one the first iterator is taken from) must have
	*		enough entries to store all samples.
	*/
	virtual void sampleValues(double x1, double x2, double y, double sampleX, double sampleY, typename buffer_type::iterator first) {}

	virtual void addSubFunction( Function2D<value_type> * fn, buffer_type * buffer ) {}
	
	virtual void getValue(double x, double y, value_type & value ) const {}
};


/**
*	\brief consists of more 2D functions and evaluates them all
*/
template <class T>
class CompositeFunction2D: public Function2D<T>
{
protected:
	
	typedef T value_type;
	typedef std::vector< value_type > buffer_type;

	/**
	*	\brief subfunction - function and its output buffer
	*/
	struct SubFunction 
	{
		SubFunction()
			: function(0), buffer(0)
		{}

		SubFunction(Function2D<value_type> * fn, buffer_type * b)
			: function(fn), buffer(b)
		{}

		Function2D<value_type> * function;
		buffer_type * buffer;
	};

	/** \brief container type definition for subfunctions */
	typedef std::vector<SubFunction> sub_functions_type;

	/** \brief subfunctions container */
	sub_functions_type subFunctions;

public:

	/**
	*	\brief insert a subfunction to container
	*	\param fn pointer to function
	*	\param buffer pointer to buffer
	*/
	virtual void addSubFunction( Function2D<value_type> * fn, buffer_type * buffer )
	{
		subFunctions.push_back( SubFunction(fn, buffer) );
	}

	void ensureBuffersSize(typename buffer_type::size_type newSize )
	{
		typename sub_functions_type::iterator it( subFunctions.begin() ), ite( subFunctions.end() );
		for (; it != ite; ++it)
		{
			if ( it->buffer && it->buffer->size() < newSize )
				it->buffer->resize( newSize );
		}
	}

	/**
	*	\brief sample all subfunctions to buffers
	*	\param first is ignored, buffers are filled instead of first
	*/
	virtual void sampleValues(double x1, double x2, double y, double sampleX, double sampleY, typename buffer_type::iterator first)
	{
		typename sub_functions_type::iterator it( subFunctions.begin() ), ite( subFunctions.end() );
		for (; it != ite; ++it)
		{
			if (it->function && it->buffer )
				it->function->sampleValues(x1, x2, y, sampleX, sampleY, it->buffer->begin() );
		}
	}
};

template <class T>
class PeriodicFunction2D: public Function2D<T>
{
public:
	virtual ~PeriodicFunction2D() {}
	virtual int getPeriod() = 0;
};

#endif //  _FUNCTION2D_

