/* MAINTAINER: Bernard */

#include "log.h"
#include "game.h"
#include "platform.h"
#include "conversions.h"
#include "world.h"
#include "tinyxml/tinyxml.h"
#include "settings.h"
#include <sstream>
#include "application.h"
#include "commander.h"
#include "timemanager.h"
#include "model_cache.h"

using namespace std;

Game::Game(void): 
mEndFun(false)
{
	mRunning = false;
	mWorld = 0;
}

Game::~Game(void)
{
	clean();	
}

void Game::cacheModels() const
{
	LOGI << "Caching models..." << std::endl;
	ModelCache * mc = ModelCache::getInstance();
	// just for sure, check mc for null
	if ( !mc )
	{
		LOGE << "Some serious problem with model cache has occured!" << std::endl;
		return;
	}
	
	std::vector<std::string>::iterator it( gParams.m3DModels.begin() ), ite( gParams.m3DModels.end() );
	for (; it != ite; ++it )
		mc->cacheIt( *it );

	int cnt( mc->count() );
	LOGI << "Caching models: " << cnt << ( cnt > 0 ? " successfully cached." : " cached." ) << std::endl;
}

void Game::clearModelCache() const
{
	LOGI << "Clearing model cache..." << std::endl;
	ModelCache * mc = ModelCache::getInstance();
	// just for sure, check mc for null
	if ( !mc )
	{
		LOGE << "Some serious problem with model cache has occured!" << std::endl;
		return;
	}

	int cnt( mc->count() );
	mc->clear();
	cnt -= mc->count();

	LOGI << "Clearing model cache: " << cnt << " entries released." << std::endl;
}

void Game::init()
{
	cacheModels();
	recreateWorld();
	
	// Add commands from SETTING_GAME_INIT_COMMANDS
	stringstream commands(g_settings.getString(SETTING_GAME_INIT_COMMANDS));
	string line;
	while ( getline(commands, line).good() )
	{
		getApp().getCommander().addCommand(line);
	}	
}


void Game::clean()
{
	clearModelCache();
	// remove "OLD" world data from world (but players are still here
	if (mWorld)
	{
		mWorld->cleanUp();
		delete mWorld;
		mWorld = 0;		
	}
	
	removeAllPlayers();	
	mEventQueue.cleanUp();
}

void Game::removeAllPlayers()
{
		// Remove all players
	for (int i = 0; (i < (int)mPlayers.size()); ++i)
	{
		if (mPlayers[i] != 0) delete mPlayers[i];
		mPlayers[i] = 0;
	}
	updateValidPlayersID();	
}


World& Game::getWorld()
{
	assertL(mWorld);
	
	return *mWorld;
}

void Game::endFun()
{ 
	TIN;
	gTime.enablePause(false);
	mEndFun = true; 
}

void Game::loadMap(const std::string &mapName)
{
	TIN;

	mMapPath = locateFile(mapName);

	if (!exist(mMapPath.c_str()))
	{
		LOGE << "Unable to find/open file wit map: " << mapName << endl;
		DIE("Map file doesn't exist");
	}


	Params::loadMapBasicInfo(mMapPath, mMapParams);

	getWorld().loadedMap();

	TiXmlDocument doc;
	TiXmlHandle   handle(&doc);

	if (!doc.LoadFile(mMapPath) || !loadMapParts(handle))
	{
		LOGE << "Unable to load " << mapName << endl;
		DIE("Loading parts of map failed");
	}

	LOGI << "Loaded map " << mMapParams.name << endl;
}

void Game::unloadMap()
{
	
}

bool Game::loadMapParts(TiXmlHandle &docHandle)
{

	//LOGD << " loadMapParts " << endl;

	//Loading weapons from MAP
	mDefaultWeapons.clear();
	mDefaultWeapons.assign(gParams.mWeapons.size(), false);
	
	TiXmlElement* child;
	child = docHandle.FirstChild("map").FirstChild("basic_equip").FirstChild("weapon").Element();
	for (; child; child=child->NextSiblingElement("weapon")) 
	{
		// enabled weapon id
		const char *idc = child->Attribute("id");
		long id;
		if (idc == NULL || !try_get_long(idc, id) || id < 0 || (unsigned int)id >= mDefaultWeapons.size())
		{
			LOGE << "Wrong weapon id attribute in map." << endl;
			return false;
		}

		mDefaultWeapons[id] = true;
		//LOGD << "Enabling weapon " << id << endl;

	}

	//Loading projectiles from MAP
	mDefaultProjectiles.clear();
	mDefaultProjectiles.assign(gParams.mProjectiles.size(), 0);

	child = docHandle.FirstChild("map").FirstChild("basic_equip").FirstChild("ammo").Element();
	for (; child; child=child->NextSiblingElement("ammo")) 
	{
		// ID of projectil
		const char *idc = child->Attribute("id");
		long id;
		if (idc == NULL || !try_get_long(idc, id) || id < 0 || (unsigned int)id >= mDefaultProjectiles.size())
		{
			LOGE << "Wrong ammo id attribute in map." << endl;
			return false;
		}

		// Number of projectiles
		const char *countc = child->Attribute("ammount");
		long count;
		if (countc == NULL || !try_get_long(countc, count) || count < 0)
		{
			LOGE << "Wrong weapon ammount attribute in map." << endl;
			return false;
		}

		mDefaultProjectiles[id] = count;

		//LOGD << "Adding " << count << " projectiles of type " << id << endl;
	}

	return true;
}


int Game::newPlayer(Player &player, int ID/* = INVALID_PLAYER_ID*/)
{
	LOGD << "new player with ID " << ID << endl;

	// search for new ID
	if (ID == INVALID_PLAYER_ID)
	{
		for (ID = 0; (ID < (int)mPlayers.size()) && (mPlayers[ID] != 0); ++ID);
	}

	// resize the array for sure
	if (ID >= (int)mPlayers.size()) 
	{
		// resize and init to zero
		int oldsz = (int)mPlayers.size();
		mPlayers.resize(ID+1);
		for (int i = oldsz; i < ID+1; ++i) mPlayers[ID] = 0;
	}
	
	if (mPlayers[ID] != 0)
	{
		LOGE << "Something is wrong. Wanted to add allready existing ID." << endl;
		DIE();
	}
		
	mPlayers[ID] = new Player(player);
	if (mPlayers[ID] == 0)
	{
		LOGE << "new failed. And this will probably fail too :0)" << endl;
		DIE();
	}
	
	// *mPlayers[ID] = player;
	mPlayers[ID]->setID(ID);

	updateValidPlayersID();

	LOGD << "final player ID " << ID << endl;


	return ID;
}


Player & Game::getPlayer(int ID)
{
	if (ID < 0 || ID >= (int)mPlayers.size() || mPlayers[ID] == 0)
	{
		LOGE << "Invalid player ID " << ID << endl;
		DIE();
	}

	return *mPlayers[ID];
}


void Game::removePlayer(int ID)
{
	if (ID < 0 || ID >= (int)mPlayers.size() || mPlayers[ID] == 0)
	{
		LOGE << "Invalid player ID " << ID << endl;
		DIE();
	}

	delete mPlayers[ID];
	mPlayers[ID] = 0;

	updateValidPlayersID();
}


const std::vector<int>& Game::getValidPlayersID()
{
	return mValidPlayersID; 
}
	
void Game::updateValidPlayersID()
{
	mValidPlayersID.clear();
	mValidPlayers.clear();
	
	for (int i = 0; i < (int)mPlayers.size(); ++i)
	{
		if (mPlayers[i] != 0)
		{
			mValidPlayersID.push_back(mPlayers[i]->getID());
			mValidPlayers.push_back(mPlayers[i]);
		}
	}
}

bool Game::isPlayerIDValid(int ID)
{
	if (ID == INVALID_PLAYER_ID) return false;
	if (ID < 0 || ID >= (int)mPlayers.size() || mPlayers[ID] == 0) return false;
	return true;
}
	

bool Game::isRunning()
{
	mRunningMutex.lock();
	bool rv = mRunning;
	mRunningMutex.unlock();

	return rv;
}


void Game::setRunning(bool state)
{
	mRunningMutex.lock();
	mRunning = state;
	mRunningMutex.unlock();
}

void Game::recreateWorld()
{
	if (mWorld != 0) delete mWorld;
	
	mWorld = new World();
	if (mWorld == 0) DIE("Creating new world failed!");
}


TaskManager& Game::getTaskManager()
{
	return mTaskManager;
}


Params::sMapParams & Game::getMapParams()
{
	assertL(mMapParams.valid);
	return mMapParams;
}

void Game::loadLandscapeEntities()
{
	assertL(mMapParams.valid);
	getWorld().getLandscape().loadEntities(mMapPath);
}

weaponVectorType& Game::getDefaultWeapons()
{
	return mDefaultWeapons;
}

projectilesNumberType&  Game::getDefaultProjectiles()
{
	return mDefaultProjectiles;
}
	
playerVectorType::iterator Game::getValidPlayersVectorBegin()
{
	return mValidPlayers.begin();	
}

playerVectorType::iterator Game::getValidPlayersVectorEnd()
{
	return mValidPlayers.end();	
}

bool comparePlayers(Player *a, Player *b)
{
	assertL(a && b);
	if (a->getScore() > b->getScore()) return true;
	if (a->getScore() < b->getScore()) return false;

	if (a->getKills() > b->getKills()) return true;
	if (a->getKills() < b->getKills()) return false;

	if (a->getDeaths() > b->getDeaths()) return true;
	if (a->getDeaths() < b->getDeaths()) return false;

	if (a->getID() < b->getID()) return true;

	return false;
}

void Game::sortValidPlayers()
{
	sort(mValidPlayers.begin(), mValidPlayers.end(), comparePlayers);
}

int Game::getPlayerCount()
{
	return (int)mValidPlayersID.size(); 
}

const std::string& Game::getMapPath()
{
	return mMapPath;
}

void Game::setReadyForAllPlayers(bool new_ready)
{
	for (unsigned int playerIDX = 0; playerIDX < mValidPlayers.size(); playerIDX++)
	{	
		assertL(mValidPlayers[playerIDX]);
		mValidPlayers[playerIDX]->ready(new_ready);
	}
}
