/* MAINTAINER: Bernard */

#ifndef _GAME_H
#define _GAME_H

#include <string>
#include <vector>
#include "taskmanager.h"
#include "player.h"
#include "mutex.h"
#include "eventqueue.h"
#include "params.h"
#include "globals.h"

class World;
class TiXmlHandle;

typedef std::vector<bool>     weaponVectorType;
typedef std::vector<int>      projectilesNumberType;
typedef std::vector<Player*>  playerVectorType;

/**	\brief States (phases) in which the game can be
 */
enum eStateOfGame
{
	// NETNoChange = -1, // dont use -1 - this is reserved for network
	WaitForPlayers,		///< \brief Waiting fo players to get ready
	Playing,			///< \brief Normal play with scoring
	EndOfGame			///< \brief Game stopped, chattime
};

/**	\brief Structure that holds information about current status of map ending conditions
 */
struct ConditionStatus
{
	std::string name;	///< \brief Name of condition
	int targetvalue;	///< \brief Some value that wants this condition to reach
	int currentvalue;	///< \brief Some value that is the current status of condition
};

/**
 * \brief Structire containing info about Game.
 * 
 * It is transfered from server to client when client get's conncetd.
 * (the client asks for it);
 */
struct GameInfo
{
	std::string mapName;  ///<\brief Map name 
};

/**
 * \brief Holds all content of game / provide access to it.
 *
 * Itself implements access to players in game. You can't create
 * instance of this class. 
 * This class implements common Game features for server and client.
 */
class Game 
{
public:
	Game(void);
	virtual ~Game(void);
	
	/**
	 * \brief Init data structures for game.
	 *
	 * This is for example loading some textures, models, registering
	 * to input manager,...
	 */
	virtual void init();
	
	/**
	 * \brief Start the game.
	 *
	 * Creates tasks and start main loop.
	 */
	virtual void run() = 0;
    
	/**
	 * \brief Clean data structures for game. 
	 *
	 * It mean for examlpe unloading textures and models...
	 * This function shouls clean players vector. So don't forget
	 * to call it in derived classes.
	 */
	virtual void clean();
	
	/**
	 * \brief Set flag that signalises engine to end.
	 */
	virtual void endFun();
        
	/**
	 * \brief Provide access to World
	 *
	 * \return mWorld
	 */
	World& getWorld();


	/**
	 * \brief Register new player to game
	 *
	 * Note that generated ID doesn't has to be unique. It may
	 * be some older ID that has been invalidated.
	 * \param player structure with new player to add. The copy of this
	 *        structure is saved.
	 * \param ID of new player. If INVALID_PLAYER_ID first free ID is taken.
	 *        The INVALID_PLAYER_ID call is for server and specific ID call will be used
	 *        in client. If you set ID and ID is allready in use, the program
	 *        will halt.
	 * \return id of player
	 */
	virtual int newPlayer(Player &player, int ID = INVALID_PLAYER_ID);

	
	/**
	 * \brief Return player associated with this ID.
	 *
	 * The program will be halted if you call this with invalid ID.
	 * \param ID id if player you want to receive.
	 * \return Reference to player that has ID ID
	 */
	Player & getPlayer(int ID);

	/**
	 * \brief Invalidate ID of player. 
	 *
	 * The ID must not be used then. But the ID may be
	 * recycled and later returned by newPlayer.
	 * TODO: there should be some fetare that check te vector
	 *       of players and resize it if number of players is <= 1/2
	 *       of allocated space for players. (But at least 8 ? Or something
	 *       like that)
	 * \param ID that will be invalidated.
	 */
	virtual void removePlayer(int ID);

	/**
	 * \brief Clear all player from Game.
	 */
	virtual void removeAllPlayers();

	/**
	 * \brief Create vector of all valid players ID.
	 *
	 * NOTE that this vector may change in next loop run
	 * since some players may leave/join the game. So be careful.
	 * Please don't modify returned vector.
	 * \return vector of ID's.
	 */
	const std::vector<int>& getValidPlayersID();


	playerVectorType::iterator getValidPlayersVectorBegin();

	playerVectorType::iterator getValidPlayersVectorEnd();

	/**
	 * \brief test of ID validity
	 *
	 * This function may be needed because client may kick some
	 * other player using its ID but the player already left so
	 * we may get in troubles when calling removePlayer directly.
	 * \param ID ID to test for validity
	 * \return true if ID valid.
	 */
	bool isPlayerIDValid(int ID);

	/**
	 * \brief Get number of player in game
	 */
	int getPlayerCount();

	/**
	 * \brief Sort players ID in mValidPlayersID using their score.
	 */
	void sortValidPlayers();

	/**
	 * \brief Set ready flag for all players
	 * 
	 * \param new_ready New value to set.
	 */
	void setReadyForAllPlayers(bool new_ready);

	/**
	 * \brief Thread safe access to state of game.
	 *
	 * This is good to test if it is a good idea to send/receive/process
	 * network messages like Up/down/tank movement/...
	 * Provides access to mRunning.
	 * \return true if whole world is loaded and the game is running.
	 */
	bool isRunning();

	/**
	 * \brief Get task manager for game
	 * 
	 * The task manager may contain App.TaskManager as a member
	 * \return TaskManager from game
	 */
	TaskManager& getTaskManager();

	/**	\brief Returns reference to event queue
	 *
	 *	\return Reference to event queue	
	 */
	EventQueue & getEventQueue()
	{
		return mEventQueue;
	}


	/** \brief Return reference to currently active map params.
	 *
	 * Say just a reference to mMapParams. If it is invalid, the
	 * assert is produced
	 */
	Params::sMapParams & getMapParams();

	
	weaponVectorType&  getDefaultWeapons();
	projectilesNumberType&  getDefaultProjectiles();


	const std::string& getMapPath();

	/**
	 * Load data for some map.
	 * 
	 * \param mapName name of file to load
	 * 
	 * Check if the file may be located and opened for reading.
	 * Also loads mMapParams. But other things like texture loading and
	 * so will be performed later. If the file doesn't exist, return false.
	 * If the file has some errors in it, the program is killed.
	 * \return true if loaded successfully.
	 */
	virtual void loadMap(const std::string &mapName);

	virtual void unloadMap();

protected:


	virtual bool loadMapParts(TiXmlHandle &docHandle);


	/**
	 * \brief Load basic map params.
	 *
	 * They are stored in mMapParams. You may get them by calling
	 * getMapParams(). The program is killed if fails.
	 */
	void loadMapParams(std::string path);

	/**
	 * \brief Load Entities from map
	 */
	void loadLandscapeEntities();

	/**	\brief Queue to handle events from network or other input - for update tasks
	 *
	 */
	EventQueue	mEventQueue;	


	/**
	 * \brief Create new world
	 *
	 * delete old if existed and create new. Be careful when calling this :-)
	 */
	void recreateWorld();
	
	/**
	 * \brief change running state.
	 *
	 * Thread safe. Provides access to mRunning.
	 * \param state new running state
	 */
	void setRunning(bool state);

	void cacheModels() const;
	void clearModelCache() const;

	bool  mRunning; ///< True if game is not loading or.. simply the game is running
	Mutex mRunningMutex; ///< For safe access to mRunning
	
	
	
	World		*mWorld;         ///< World structure holden by Game

	TaskManager	mTaskManager;	///< All tasks

	bool		mEndFun;		///< if true, engine will stop


	/** 
	 * \brief Vector of all players in game.
	 *
	 * Note that not all positions has to be valid.
	 */
	playerVectorType mPlayers;  ///< Vector from 0 to maxID with players
	std::vector<int> mValidPlayersID; ///< List of valid players ID - is generated using mPlayers
	playerVectorType mValidPlayers;  // only valid Players - is generated using mPlayers


	/**
	 * \brief Update mValidPlayersID to match mPlayers
	 *
	 * This should be called after every change of mPlayers
	 */
	void updateValidPlayersID();


	Params::sMapParams mMapParams;  ///< basic info about current map
	std::string        mMapPath; ///< Path to map config file


	weaponVectorType       mDefaultWeapons;     ///< \brief Which weapons the tank has
	projectilesNumberType  mDefaultProjectiles; ///< \brief Number of projectiles for use
};

#endif // _GAME_H
