/* MAINTAINER: Petko */
#ifndef _COMONENTITY_H
#define _COMONENTITY_H

#include "GLVector.h"
#include "GLQuaternion.h"
#include <string>
#include <memory>

class Model;
class ServerPhysicsData;

/** \addtogroup game_data */ /*@{ */

/**	\brief These values are used to identify the type of object
 *
 */
enum ClassType 
{
	cInvalid,
	cProjectile,
	cTank,
	cBonusBox,
	cStatic
};

/**	\brief Basic game entity representation
 *
 *	This class is abstract ancestor of all objects in game,
 *	as moving as steady ones. It has methods that draws the object,
 *	object initialization, collision detection helpers,...
 */
class CommonEntity
{
public:
	CommonEntity();
	virtual ~CommonEntity() = 0;

	void setID(int ID);
	int  getID();

	/**	\brief Makes basic initialzation of object - fills in ID, position and orientation
	 *
	 *	\param ID ID to set for this object.
	 *	\param position Position of new object in scene.
	 *	\param orientation Orientation of object in scene.
	 */
	void basicInit( int ID, GLVector3d position, GLQuaterniond orientation );

	/**	\brief Makes object-specific initialization
	 *
	 *	This method is intended to do stuff like initialization of 3D model...
	 *	If it returns false, the object won't be added to list of objects and
	 *	will be deleted. This method should have no paramters. Every inherited class
	 *	can contain method bool initSpecific(), where can be used specific parameters 
	 *	for initialization.
	 *
	 *	\return True if init is succesfull, otherwise the object id deleted
	 */
	virtual bool init();

	/**	\brief Makes object-specific actions when being removed from world.
	 *
	 *	This method is called from Wrld::delete* methods to allow objects to
	 *	do their specific cleanup operations, when removed from world. Now this
	 *	means also object destruction, but this may change in future, so use this 
	 *	method, NOT destructor!
	 */
	virtual void destroy();
		
	// Maybe this routine could do the rendering and then the others
	// doesn't have to override it. Except tank probably.
	virtual void render();

	/**	\brief Returns which descendant of CommonEntity this class is.
	 *	
	 *
	 *	\return Type of this object.
	 */
	ClassType getClassType();

	/**	\brief Return position of object.
	 *
	 *	\return Position of object.
	 */
	const GLVector3d& getPosition() const
	{
		return mPosition;
	}

	/**	\brief Sets the position of object. BE CAREFULL!
	 *
	 *	\param pos New position of object.
	 */
	void setPosition( const GLVector3d &pos )
	{
		mPosition = pos;
	}

	/**	\brief Return orientation of object.
	 *
	 *	\return Orientation of object.
	 */
	GLQuaterniond getOrientation() const
	{
		return mOrientation;
	}

	/**	\brief Sets the orientation of object. BE CAREFULL!
	 *
	 *	\param orient New orientation of object.
	 */
	void setOrientation( GLQuaterniond orient )
	{
		mOrientation = orient;
		mOrientation.normalize();
	}

	/**
	 * \brief Return angle of the entity when considering only x,y plane.
	 * 
	 * \return Angle of orientation of the entity.
	 */
	double getXYangle();
	

	/**	\brief Returns server physics data of this entity
	 *
	 *	\return Physics data  pointer
	 */
	ServerPhysicsData * getPhysicsData() const
	{
		return mPhysicsData;
	}

	/**	\brief Sets server physics data of this entity
	 *
	 *	\param Physics data  pointer
	 */
	void setPhysicsData(ServerPhysicsData * spd)
	{
		mPhysicsData = spd;
	}

	/**	\brief Returns subtype of object - index for params.
	 *
	 *	\return Subtype of object.
	 */
	int getSubType() const
	{
		return mSubType;
	}

	/**	\brief Returns pointer to model of this object
	 *
	 *	\return pointer to model of this object
	 */
	Model * getModel()
	{
		return mModel.get();
	}
		
protected:	
	/// \brief Position of an object in the world
	GLVector3d		mPosition;

	/// \brief Orientation of object in the world
	GLQuaterniond	mOrientation;

	/// \brief Unique indetificator of object in all world. Note that valid value is > 0.
	int				mID; 

	/// \brief Determines which class this is
	ClassType		mType;

	/** \brief Data for physics - ODE.
	 *
	 *	Data for physics - ODE. 
	 *	WARNING! This pointer is deleted in TaskSimulationServer::deletedWorldEntity() for WorldEntities
	 *	and in destructor of entity for LandscapeEntities
	 */
	ServerPhysicsData * mPhysicsData;

	/// \brief Identifies subtype of thi object - eg. which type of tank is this.
	int				mSubType;

	// Model for graphics or physics
	// TODO: client/server stuffs (do not load trimesh on server)
	std::auto_ptr< Model > mModel;			///< \brief Instance of tank's model

	void setModelName( const std::string & modelName )
	{
		mModelName = modelName;
	}

	const std::string & getModelName() 
	{
		return mModelName;
	}

private:
	/// \brief Private copy contstruktor - we don't want anybody to copy this
	CommonEntity( const CommonEntity & );
	
	std::string mModelName; ///< \brief 3D model name
};

/*@}*/ 
#endif // _COMONENTITY_H
