/* MAINTAINER: Petko  CURRENT_EDITOR: Bernard */
/* COMMON */

//#define WANT_TRACE

#include "params.h"
#include "landscape.h"
#include "log.h"
#include "game.h"
#include "world.h"
#include "application.h"
#include "tinyxml/tinyxml.h"
#include "conversions.h"
#include "globals.h"

using namespace std;
using namespace STL_EXT_NS;



Landscape::Landscape(void)
{
	mMinID = -1;
}

Landscape::~Landscape(void)
{
	clean();
}



LandscapeEntity* Landscape::createLandEntity( GLVector3d position, GLQuaterniond orientation,
                                              int param_id, int id)
{
	LandscapeEntity * object = new LandscapeEntity;
	assertL( object != NULL );

	object->initByID(param_id);

	object->basicInit( (id==INVALID_ENTITY_ID)?getNextID():id, position, orientation );

	if ( !object->init() )
	{
		LOGW << "Failed one of object initializations!" << endl;
		delete object;

		return 0;
	}

	addLandscapeEntity( object );

	LOGT << "created, id=" << param_id << endl;

	return object;
}


LandscapeEntity* Landscape::createLandEntity( GLVector3d position, GLQuaterniond orientation,
                                              std::string modelName, int id)
{
	LandscapeEntity * object = new LandscapeEntity;
	assertL( object != NULL );

	object->initByModelName(modelName);

	object->basicInit( (id==INVALID_ENTITY_ID)?getNextID():id, position, orientation );

	if ( !object->init() )
	{
		LOGW << "Failed one of object initializations!" << endl;
		delete object;

		return 0;
	}

	addLandscapeEntity( object );
	
	LOGT << "created, modelName=" << modelName << endl;
	
	return object;
}

bool Landscape::deleteLandEntity( int id )
{
	LandscapeEntity * object = getLandEntity( id );
	if ( object == NULL )
	{
		LOGW << "LandscapeEntity lookup NOT successfull!" << endl;
		return false;
	}

	

	delLandscapeEntity( object );
	return true;
}


landent_iterator Landscape::getLandEntitiesBegin()
{
	return mLandList.begin();
}

landent_iterator Landscape::getLandEntitiesEnd()
{
	return mLandList.end();
}

LandscapeEntity* Landscape::getLandEntity(int id)
{
	landent_hash::iterator it = mLandHash.find( id );

	if ( it == mLandHash.end() )
	{
		LOGD << "Entity with id = " << id << " not found." << endl;
		return NULL;
	}
	
	if ( it->second == NULL ) 
	{
		LOGE << "Critical internal error. NULL MUST NOT BE here.\n";
		DIE();
	}
	return it->second;
}

LandscapeEntity* Landscape::getLandEntity(landent_iterator it) const
{
	if ( it == mLandList.end() )
		return NULL;

	if (*it == NULL) 
	{
		LOGE << "Critical internal error. NULL MUST NOT BE here.\n";
		DIE();
	}
	return *it;
}


void Landscape::addLandscapeEntity( LandscapeEntity* le )
{
#ifdef _DEBUG
	if ( mLandHash.find( le->getID() ) != mLandHash.end() )
	{
		LOGE << "Trying to add an object with ID already used\n";
		DIE();
	}
	if ( le->getID() == INVALID_ENTITY_ID ) 
	{
		LOGE << "Try to add an object with invalid ID " << le->getID() << "\n";
		DIE();
	}
#endif

	//adding to list:
	mLandList.push_back( le );

	//adding to hashtable:
	mLandHash[ le->getID() ] = le;

	//notyfying tasks that new entity was created ? Is this needed? - probably not
	//gApp->getGameCommon()->getTaskManager().addedWorldObject( le );
}

void Landscape::delLandscapeEntity( LandscapeEntity* le )
{
#ifdef _DEBUG
	if ( mLandHash.find( le->getID() ) == mLandHash.end() )
	{
		LOGE << "Trying to delete object with invalid ID!\n";
		DIE();
	}
#endif

	// notyfying tasks that entity is going to be deleted - probably not needed
	// gApp->getGameCommon()->getTaskManager().deletedWorldObject( le );

	//delete from World containers:
	
	//delete from list:
	mLandList.remove( le );

	//delete from hashtable:
	mLandHash.erase( le->getID() );
	
	le->destroy();
	delete le;
}

int Landscape::getNextID()
{
//	return getApp().getGameCommon().getWorld().getNextID();
	return --mMinID;
}

Landscape::Landscape( const Landscape & )
{
	DIE( "Attempt to copy the instance of World!" );
}

#define GET_VALUE(x) \
	value = child->Attribute( #x ); \
	if ( value == 0 ) \
	{ \
		LOGE << "Missing attribute " << #x << " in file " << mapPath << endl; \
		DIE(); \
	} 

#define LOAD_DOUBLE(x) \
	GET_VALUE(x) \
	x = get_double(value);

void Landscape::loadEntities(std::string mapPath)
{	
	// Clean old entites
	clean();
	
	
	TiXmlDocument doc;
	TiXmlHandle   handle(&doc);
	TiXmlElement* child;

	const char *value, *id, *model_name;
	double x,y,z, rot_x, rot_y, rot_z, angle;



	if (!doc.LoadFile(mapPath))
	{ 
		LOGE << "Unable to load " << mapPath
		     << " Probably file not exist or isn't a valid XML file" << endl;
		DIE(); 
	}

	child = handle.FirstChild("map").FirstChild("static_objects").FirstChild("entity").Element();
	if (!child)
	{
		LOGI << "There are no static objects in file " << mapPath  << endl;
	}

	for (; child; child=child->NextSiblingElement()) 
	{
		LOAD_DOUBLE(x)
		LOAD_DOUBLE(y)
		LOAD_DOUBLE(rot_x)
		LOAD_DOUBLE(rot_y)
		LOAD_DOUBLE(rot_z)
		LOAD_DOUBLE(angle)

		// attribute z isn't required
		if ( child->Attribute("z") != 0)
		{
			LOAD_DOUBLE(z)
		}
		else
		{
			z = mTerrain.getHeight(x, y);
		}

		id         = child->Attribute("id");
		model_name = child->Attribute("model_name");
		if (id)
		{
			createLandEntity(GLVector3d(x, y, z), 
			                  GLQuaterniond(GLVector3d(rot_x, rot_y, rot_z),angle), get_long(id));

		} 
		else 
		{
			if (model_name)
			{
				createLandEntity(GLVector3d(x, y, z), 
			                  GLQuaterniond(GLVector3d(rot_x, rot_y, rot_z),angle), model_name);
			}
			else
			{
				LOGE << "In map file " << mapPath 
				     << " is missing id or model_name in some land etity" << endl;
				DIE();
			}
		}
	}
}

void Landscape::clean()
{
	// deleting all objects
	while (mLandList.begin() != mLandList.end())
	{
		delLandscapeEntity(	*mLandList.begin() );
	}	
}
