/* MANTAINER: Bernard */

#ifndef _PARAMS_H
#define _PARAMS_H

#include <string>
#include <vector>
#include <cstdlib>
#include "terrain_map_attributes.h"

/**
 * \brief class that holds all parameters for tanks etc.
 *
 * It is inicialized in Application
 */
class Params
{	
public:
	Params();
	~Params();

	/** 
	 * \brief Load parameters for everything :-) 
	 *
	 * Die on error.
	 */
	void load();

/************* Tank section **************/

	/** 
	 * \brief Tank parameters
	 */
	typedef struct sTankParams
	{
		sTankParams(bool val = false) { valid = val; }
		bool        valid;        ///< true if structure is valid
		
		int         id;           ///< must be unique
		std::string name;         ///< some nice fantasy name to display
		double      max_speed;    ///< in m/s
		int         engine_power; ///< in watt
		int         hp;           ///< 1 HP == HP of one human
		int         armour;        ///< in percents
		int         turning;      ///< ability to turn - in degrees per second (old model)
		double      turning_delta_speed; ///< delta speed of wheels when turning (new model)
		std::string turning_user_info;   ///< some house numero to inform user about turning :-)
		double      weight;       ///< int kg. Double because of weight of projectiles
		std::string model;        ///< (file) (collision model will be generated on the fly)
		std::string engine_sound; ///< (file) sound of tank's engine
		int         explosion_id; ///< id of explosion of tank
	} sTankParams;
	
	/**
	 * \brief find params for tank ID.
	 *
	 * \param id ID of tank that you want to have. 
	 *           Program is killed if ID doesn't exist (is invalid)
	 * \return Reference to wanted tank type.
	 */
	const sTankParams& tank(int id);
	
	typedef std::vector<sTankParams> TankParamsVector;
	
	TankParamsVector mTanks; ///< Parameters of all tanks types

/************* Projectile section **************/
	
	/** 
	 * \brief Projectile parameters
	 */
	typedef struct sProjectileParams
	{
		sProjectileParams(bool val = false) { valid = val; }
		bool        valid;    ///< true if structure is valid
		
		int         id;       ///< must be unique
		std::string name;     ///< some cool name
		std::string model;    ///< some cool model
		std::string shoot_sound;    ///< some cool model
		double      speed;    ///< in m/s
		double      weight;   ///< in kg
		int         damage_min;   ///< minimal damage in HP units
		int         damage_max;   ///< maximal damage in HP units
	 	bool        piercing; ///< ignore armour
	 	double      radius;   ///< int m,   may be 0
		int         carry_max; ///< maximum that player may have at once
		
		int         explosion_hit_ground_id; ///< explosion when hitting ground
		int         explosion_hit_entity_id; ///< explosion when hitting entity
		
		/**
		 * \brief Return damage done by the projectile
		 *
		 * Please not that this function may return different values
		 * in every call.
		 * \return Damage done by the projectile
		 */
		int getDamage() const
		{	
			if (damage_max == damage_min) return damage_max;
			return damage_min + std::rand()%(damage_max-damage_min);
		}
		
	} sProjectileParams;
		
	/**
	 * \brief find params for projectile ID.
	 *
	 * \param id ID of projectile type you want to have.
	 *           Program is killed if ID doesn't exist (is invalid)
	 */
	const sProjectileParams& projectile(int id);
	
	std::vector<sProjectileParams> mProjectiles; ///< Parameters of all projectiles types

/************* Weapon section **************/
	
	/** 
	 * \brief Weapon parameters
	 */
	typedef struct sWeaponParams
	{
		sWeaponParams(bool val = false) { valid = val; }
		bool        valid;    ///< true if structure is valid
		
		int         id;              ///< must be unique
		std::string name;            ///< some cool name
		std::string model;           ///< some cool model
		std::string icon;            ///< some cool icon for weapon
		double      reload_time;     ///< in ms
		int         projectile;      ///< id in mProjectiles
		int         change_in_time;  ///< in ms time that takes to show weapon and make it ready for shooting
		int         change_out_time; ///< in ms time that takes to hide the weapon
		double      pitch_min;       ///< minimal possible shooting angle
		double      pitch_max;       ///< maximal shooting angle
	} sWeaponParams;
		
	/**
	 * \brief find params for weapon ID.
	 *
	 * \param id ID of weapon type you want to have.
	 *           Program is killed if ID doesn't exist (is invalid)
	 */
	const sWeaponParams& weapon(int id);
	
	std::vector<sWeaponParams> mWeapons; ///< Parameters of all weapons types

/************* LandscapeEntity section **************/
	
	/** 
	 * \brief LandscapeEntity parameters
	 */
	typedef struct sLandscapeEntityParams
	{
		sLandscapeEntityParams(bool val = false) { valid = val; }
		bool        valid;    ///< true if structure is valid
		
		int         id;         ///< must be unique
		std::string name;       ///< some cool name
		std::string model;      ///< some cool model
	} sLandscapeEntityParams;
		
	/**
	 * \brief find params for landscape entity ID.
	 *
	 * \param id ID of landscape entity type you want to have.
	 *           Program is killed if ID doesn't exist (is invalid)
	 */
	const sLandscapeEntityParams& landscapeEntity(int id);
	
	std::vector<sLandscapeEntityParams> mLandscapeEntities; ///< Parameters of all landscape entities types


	
/************* Bonus section **************/

	/** 
	 * \brief Bonus parameters
	 */
	typedef struct sBonusParams
	{
		sBonusParams(bool val = false) { valid = val; }
		bool        valid;    ///< true if structure is valid
		
		int         id;           ///< must be unique
		std::string name;         ///< some cool name
		std::string model;        ///< some cool model
		std::string pick_message; ///< amuse user :-)
		std::string pick_sound;   ///< amuse user :-)
		std::string icon;         ///< has sense only if bonus has some duration
		/**
		 * \brief time in ms how long the bonus will work.
		 * If == 0 (limited) params will be ignored.
		 */
		int         duration;    
		bool        outlive_tank;   ///< if true the bonus will appear after tank destruction
		bool        drop_on_expire; ///< if true the bonus will appear after it expires
		
		/// -- types -- I don't like this solution :( But don't know better idea
		int    HP_change;       ///< Number of added HP   (permanent)
		int    ammo_addition;   ///< Add ammo (permanent)
		int    ammo_type;       ///< type of added ammo
		double max_speed;       ///< change max speed of tank in percents (limited)
		double reload_time;     ///< change time for reloading in percents (limited)
		double damage;          ///< change damage in percents (limited)
		int    armour;           ///< add some armor (don't know if percents od absolute)  (limited)
		bool   invulnerability; ///< (limited)
		bool   invisibility;    ///< (limited)		
		int    weapon;          ///< Makes available some weapon. -1 no weapon
	} sBonusParams;
		
	/**
	 * \brief find params for weapon ID.
	 *
	 * \param id ID of bonus type you want to have.
	 *           Program is killed if ID doesn't exist (is invalid)
	 * \return may return NULL if called with invalid ID
	 */
	const sBonusParams& bonus(int id);
	
	std::vector<sBonusParams> mBonus; ///< Parameters of all weapons types

/************** Map section **************/
	/** \brief Holds info about map.
	 *
	 * Not all info, just basic one. It means that for example static objects are ignored and so one.
	 */
	typedef struct sMapParams
	{
		sMapParams(bool val = false) { valid = val; }
		bool        valid;    ///< true if structure is valid
		
		std::string  name;
		std::string  loading_image;

		std::string  game_type; ///< like "deadmatch", "rabbit hunt",..
		std::string  victory_settings; ///< like "fraglimit", "timelimit",...
		
		int max_players;
		int max_bonus_boxes;

		double gravity_constant;
		std::string heightmap;
		double heightmap_scalex;
		double heightmap_scaley;
		double heightmap_scalez;

		double boundaries_min_x;
		double boundaries_min_y;
		double boundaries_min_z;
		double boundaries_max_x;
		double boundaries_max_y;
		double boundaries_max_z;

		TerrainAttributes terrain;

	} sMap;
	
	static void loadMapBasicInfo(const std::string path, sMap &map);

	static void loadMapTerrain(const std::string path, TerrainAttributes &terrain);

	/**
	 * \brief Function try to open the file and check if it is probably a map file.
	 * 
	 * File is considered to be a map file if it is a valid XML file and main tag
	 * is "map"
	 * \return true if it is a map file
	 */
	static bool isProbablyMapFile(const std::string path);
	
/************* Model preloading section **********/

	/**
	 * \brief Vector listing all models in game.
	 * 
	 * Usage of this is that it is possible to preload all models
	 * before the game is started.
	 */
	std::vector<std::string> m3DModels;
/************* Explosions section **************/
	
	/** 
	 * \brief LandscapeEntity parameters
	 */
	typedef struct sExplosionParams
	{
		enum BlendingType{
			BLENDING_NONE,
			BLENDING_INTERPOLATION,
			BLENDING_ADDITIVE,
		};

		typedef struct
		{
			float from, to;
		} RandomNumber;

		typedef struct 
		{
			float deviance;
			std::vector<float> vec;
		} RandomVector;

		typedef std::vector<float> Color;
		
		sExplosionParams(bool val = false) { valid = val; }
		bool        valid;      ///< true if structure is valid
		int         id;         ///< explosion id, it has to be unique
		std::string sound;      ///< sound of explosion - will be played once only I guess
		float		duration;

		struct Emitter
		{
			int		initialParticlesCount;
			int		maxParticlesCount;
			float	emitSpeed;
			float	stopEmit;
			bool	resurrectParticles;
		} emitter;

		struct Particle
		{
			struct {
				float radius;
				RandomNumber height;
			}position;

			// velocity
			RandomVector direction;
			RandomNumber speed;
			RandomNumber acceleration;
			RandomVector globalAcceleration;

			//time
			RandomNumber birth;
			RandomNumber spanlife;
			
			// visuals
			Color		 color;
			struct {
				std::string	 image;
				std::string  explicitAlpha;
			} sprite;
			BlendingType blendingType;
			float		 size;
		} particle;

		int         next_id; ///< if there should be created another particle source after some time == means -1 invalid
		float       time_of_next; ///< how long after us should be the next created
	} sExplosionParams;
		
	/**
	 * \brief find params for landscape entity ID.
	 *
	 * \param id ID of explosion type you want to have.
	 *           Program is killed if ID doesn't exist (is invalid)
	 */
	const sExplosionParams& explosion(int id);
	
	std::vector<sExplosionParams> mExplosions;



protected:

};


extern Params gParams; 

#endif
