/* MAINTAINER: Petko */
#ifndef _PLAYER_H
#define _PLAYER_H

#include <string>

class Tank;
class PlayerNetData;

/** \addtogroup game_data */ /*@{ */

/**	\brief Class with player's data.
 *
 */
class Player
{
public:
	Player()
	{
		dataInit();
	}
	
	/** \brief Creates new player with network data
	 *
	 * This is only valid constructor that can be called on server
	 *
	 * \praram pnd pointer to network data(must be PlayerNetDataServer class or descendant)
	 */
	Player(PlayerNetData * pnd)
	{
		dataInit();

		mNetData = pnd;
	}

	/**
	 * New copy constructor
	 *
	 * This copy constructor doesn't copy mNetData but steals it
	 */
	Player(Player & copyFrom);

	~Player();

	/**	\brief Returns the name of player.
	 *
	 *	\return Player's name
	 */
	const std::string &getName();
	void		setName(const std::string &name);

	/**	\brief (Re-)initializes player's statistics (score/kills/deaths/...) and ready status
	 */
	void		initStats();

	/**
	 * \brief Return players tank.
	 *
	 * Every player may have tank. But the player may  also be without it.
	 * So check return value for 0.
	 * \return Associated tank. May be 0
	 */
	Tank* getTank()
	{
		return mTank;
	} 
	
	int	getKills() const
	{ 
		return mKills;  
	}
	void setKills(int kills)
	{ 
		mKills = kills; 
	}
	void killsUP()
	{ 
		mKills++;       
	}
	void killsDown()
	{ 
		mKills--;       
	}


	int	getDeaths() const     
	{ 
		return mDeaths;   
	}
	void setDeaths(int deaths) 
	{ 
		mDeaths = deaths; 
	}
	void deathsUP()
	{ 
		mDeaths++;        
	}
	
	int getScore() const
	{ 
		return mScore;
	}
	void setScore(int score)
	{ 
		mScore = score;
	}

	/**	\brief Changes score by change (can be negative)
	 *
	 *	\param change Requested score change
	 */
	void addToScore( int change )
	{
		mScore+= change;
	}

	int getTeam() const
	{ 
		return mTeam;
	}
	void setTeam(int teamID)
	{ 
		mTeam = teamID; 
	}

	/**
	 * \brief Return id of player.
	 *
	 * The id corresponds to id in Game
	 * \return id of player. -1 (negative) id means invalid value.
	 */
	int getID() const;
	
	/**
	 * \brief Set new tank for this player.
	 *
	 * Also does some work around yaw and pitch. So if there will
	 * be client player this should be virtual.
	 * \param tank new tank. May be 0 if removing tank from player.
	 *              (tank was destroyed or whatever)
	 */
	void setTank(Tank *tank);

	/**
	 * \brief Set players ID
	 *
	 * This function should be called only from Game since the
	 * ID must correspond to ID in Game. So DON'T call this if 
	 * you aren't unsure.
	 * \param newID new players ID
	 */
	void setID(int newID);

	/**
	 * \brief returns players network data
	 * 
	 * \return reference to network data
	 */
	PlayerNetData & netData()
	{
		return *mNetData;
	}

	/**	\brief Returns true, if player has to be deleted after next step of simulation
	 *
	 *	When player is being removed from game, the flag mDeleteMe has to be set to true.
	 *	Then ALL PLAYER ENTITIES (projectiles/tanks) must be marked with same flag. The actual 
	 *	deletion of player will be called on server at the end of next simulation step, on clients
	 *	it should be OK to do actual deletion it directly after the appropriate event appears,
	 *	since on server all player's entities will be destroyed before player's destruction.
	 *
	 *	\note This means the flag CANNOT be set asynchronously! It must be set in main thread.
	 *
	 *	\return True, if player should be deleted
	 */
	bool getDeleteMe()
	{
		return mDeleteMe;
	}

	/**	\brief Sets mDeleteMe flag, so the player will be in next simulation deleted. It also set the flag for network to not send data to this player
	 *
	 *	\note This MUST be called from main thread - or at least cannot be called while TaskSimulationServer is being executed.
	 */
	void setDeleteMe()
	{
		mDeleteMe = true;
		mDontUpdateClient = true;
	}
	
	/**	\brief Return if client associated with this player should receive gamestates
	 *
	 *	\return True, if data should be sent, false otherwise
	 */
	bool canUpdateClient()
	{
		return !mDontUpdateClient;
	}

	/**	\brief Sets flag that should ensure that there will be not sent any more data to client associated with this player
	 */
	void dontUpdateClient()
	{
		mDontUpdateClient = true;
		mIsHuman          = false;
	}

	/**	\brief Sets player status to ready for live play
	 *
	 */
	void ready(bool value = true)
	{
		mReady = value;
	}

	/**	\brief Returns true, if player is ready for live play.
	 *
	 *	\return Wether player wants to play already
	 */
	bool isReady() const
	{
		return mReady;
	}

	/**
	 * \brief return true if player is big admin.
	 */
	bool isBigAdmin()
	{
		return mBigAdmin;
	}

	/**
	 * \brief Modifi Big Admin status of player.
	 * If you are a BIG admin, you are allowed to do admin
	 * server commands.
	 * 
	 * \param reallyAdmin Use true if the player is really a big admin.
	 */
	void setBigAdmin(bool reallyAdmin)
	{
		mBigAdmin = reallyAdmin;
	}

	/**
	 * \brief Check if the player is human player.
	 */
	bool isHumanPlayer()
	{
		return mIsHuman;
	}

private:
	
	/** \brief private data init method
	 *
	 */
	void dataInit();

	int         mID;        ///< \brief Players unique ID. -1 means invalid ID
	
	std::string	mName;		///< \brief Player's name
	int         mTeam;      ///< \brief ID of team where player is playing

	int			mKills;		///< \brief Number of killed enemies
	int			mDeaths;	///< \brief How much times the player was killed
	int         mScore;     ///< \brief Some score in game - doesn't have to be mKills

	bool		mReady;		///< \brief If true, player is ready for live play

	Tank		*mTank;		///< \brief The tank controlled by Player. May be NULL
	int			mPing;		///< \brief Player's network response time in milliseconds

	bool		mDeleteMe;	///< \brief True, if the player has to be deleted at the end of next simulation step

	bool		mDontUpdateClient;	///< \brief If true, the network subsystem will not send gamestate updates TO this player

	PlayerNetData * mNetData; ///< \brief Data that are used by network
	
	bool        mBigAdmin;	///< \brief true if player is bug admin
	bool        mIsHuman;   ///< \brief true if player is human
};

/*@}*/ 
#endif // _PLAYER_H
