/* MAINTAINER: Petko */
#ifndef _PROJECTILE_H
#define _PROJECTILE_H

#include "worldentity.h"

class Player;

/** \addtogroup game_data */ /*@{ */

/**	\brief Class represents "flying" projectile
 */
class Projectile : public WorldEntity
{
public:
	Projectile();
	~Projectile();

	/**
	 * Pack new created Projectile
	 *
	 * Packs data that is needed for creating object on client.
	 *
	 * \param data net game state structure where data will be packed
	 */
	virtual void packCreationData(NETGameState & data);

	/**	\brief Initializes Projectile according to its ID in params
	 *
	 *	\param params_ID ID from Params to initialize by.
	 */
	void initByID( int params_ID );

	/**	\brief Inits mOwner  field
	 *
	 *	\param shooting_player_id ID of player that shot this projectile
	 *	\return true,  if initialization successful, false otherwise
	 */
	bool initSpecific( int shooting_player_id );

	virtual void handleCollisionWEntityServer( int ID, const GLVector3d & coll_pos, const GLVector3d & coll_normal );

	virtual void handleCollisionWStaticServer( const GLVector3d & coll_pos, const GLVector3d & coll_normal );
	
	/**	\brief Returns true, if this object is ready to do damage
	 *
	 *	This object should return true at the moment, when it is ready to do damage.
	 *	Usually this is at first hit of another object, but can be redefined - for example for
	 *	grenade, that jumps 4 times before exploding...
	 *
	 *	\return True, if projectile is ready to do damage
	 */
	bool getHit() const
	{
		return mHit;
	}

	/**	\brief Sets projectile to state to be ready to do damage
	 *
	 *	Call this method (i.e. from collision handlers), when the projectile is 
	 *	ready to do damage. Examples: for normal shot - call this after first collision
	 *	for bomb with timer - call this method when timer expires
	 *
	 *	\param entity Entity that hit this projectile (or 0)
	 */
	void hit( int entity = 0 )
	{
		mHit = true;
		mEntity = entity;
	}

	/**	\brief Returns ID of entity that hit this projectile.
	 *
	 *	\return ID of hit entity ( 0 for static )
	 */
	int getEntity() const
	{
		return mEntity;
	}

	/**	\brief Returns player that fired this projectile
	 *
	 *	\return Author of the shot
	 */
	Player * getOwner() const
	{
		return mOwner;
	}

	virtual void render();

	/**
	 * \brief Get id of params for this projectile.
	 */
	int getParamsID()
	{
		return 	mParamsID;
	}

	/**
	 * \brief Return damage from projectile.
	 * 
	 * Not that it may be different from params damage since the projectile
	 * may be influenced by some bonuses from player.
	 */
	int getDamage()
	{
		return mDamage;	
	}

private:
	Player*	mOwner;		///< \brief Player who owns this projectile. We need to know who add a frag :-)
	int		mDamage;	///< \brief Damage that causes this projectile
	bool	mPiercing;	///< \brief Tells whether the projectile ignores armour
	double	mRadius;	///< \brief What is the damage radius in case of impact
	bool	mHit;		///< \brief False, until this projectile hits something
	int		mEntity;	///< \brief Entity that hit this projectile, or 0 if no entity or area damage
	int     mParamsID;  ///< \brief ID of params for creating this projectile
};

/*@}*/ 
#endif // _PROJECTILE_H
