/* MAINTAINER: Petko */
#ifndef _TANK_H
#define _TANK_H

#include "worldentity.h"
#include <vector>
#include <list>
#include "wep_obtainer.h"

typedef std::vector<bool> weaponVectorType;
typedef std::vector<int>  projectilesNumberType;

class Player;
class Bonus;
struct NETGameState;

typedef std::list<Bonus*> bonusListType;

/** \addtogroup game_data */ /*@{ */

/**	\brief Representation of tank.
 */
class Tank : public WorldEntity
{
public:
	Tank();

	//DEBUG:
	virtual ~Tank();

	/**
	 * Pack new created Tank
	 *
	 * Packs data that is needed for creating object on client.
	 *
	 * \param data net game state structure where data will be packed
	 */
	virtual void packCreationData(NETGameState & data);

	/**	\brief Overloaded creation of present history point from WorldEntity
	 *
	 *	\param time_stamp What is the time stamp of point created
	 */
	virtual void createPresentPoint( int time_stamp );

	virtual bool init();

	virtual void render();

	/**	\brief Initializes Tank according to its ID in params
	 *
	 *	\param params_ID ID from Params to initialize by.
	 */
	void initByID( int params_ID );

	/**	\brief Inits mPlayer field
	 *
	 *	\param player_id ID of player that will control this tank
	 *	\return true,  if initialisation successfull, false otherwise
	 */
	bool initSpecific( int player_id );

	/**	\brief return mYaw
	 */
	double getYaw() const
	{ 
		return mYaw; 
	}

	/**	\brief return mPitch
	 */
	double getPitch() const
	{ 
		return mPitch; 
	}

	/**
	 * \brief Returning Yaw relatively to tank.
	 * 
	 * It is not trivial computation.
	 */
	double getTankYaw() const;
	
	/**
	 * \brief Returning Pitch relatively to tank coordinates
	 * 
	 * It is not trivial computation.
	 */
	double getTankPitch() const;

	/**	\brief Set new values mPitch and mYaw.
	 *
	 *	\param pitch new value for mPitch
	 *	\param yaw new value for mYaw
	 */
	void setYawPitch(double yaw, double pitch);

	/**
	 * \brief Compute yaw and pitch for aiming the target.
	 * 
	 * \params target Coordinates of target.
	 * \params iterations Iterations for more accurate aiming.
	 *         human player should have at least 3 and AI player may have
	 *         for example only 1 to don't have so good aiming. The iteration
	 *         makes couting with tanks current speed and direction.
	 * \return True if aiming is good. False if target is out of range.
	 */
	bool setYawPitchFrom3DTarget(const GLVector3d &target, int iterations = 3);

	/**	\brief Structure that describes user's commands to the tank
	 *
	 */
	typedef struct TankCommands
	{
		TankCommands()
		{
			reset();
		};

		void reset()
		{
			forward = false; backward = false;
			left = false; right = false;
			fire = false;
		}

		bool forward;
		bool backward;
		bool left;
		bool right;
		bool fire;
	} TankCommands;

	/**	\brief Returns reference to commands structure
	 *
	 *	\return Non-constant reference to the commands structure
	 */
	TankCommands & getCmds()
	{
		return mCmds;
	};

	/**	\brief Makes computation of tank's movement. It computes speeds of tracks
	 *
	 *	\param ode_to_real Multiple between time flow in ODE and in simulation
	 */
	void tankControllServer( const double ode_to_real );

	/**	\brief Returns top speed of the tank
	 *
	 *	\return Tank's top speed
	 */
	double getMaxSpeed() const
	{
		return mMaxSpeed;
	}

	/**	\brief Returns weight of the tank in kgs
	 *
	 *	\return Tank's weight
	 */
	double getWeight() const
	{
		return mWeight;
	}

	/**	\brief Returns force of tank's engine
	 *
	 *	\return Force of tank's  engine
	 */
	double getEngineForce() const
	{
		return mEngineForce;
	}

	/**	\brief Tracks in oposite of turning directiom moves this faster when turning. Other track moves oposite 1/5th of this speed
	 *
	 *	\return Addition in speed for turning track
	 */
	double getTurningTrackSpeedAdd() const
	{
		return mTurnSpeedAdd;
	}

	/**	\brief Computes the launch point of shot and its direction
	 *
	 *	\param orientation Orientation of future shot
	 *	\param start Point from where the shot will be launched
	 */
	void getShotOriAndStart( GLQuaterniond & orientation, GLVector3d & start );

	/** 
	 * \brief Get roraion point of the weapon in world space coordinates.
	 */
	void getWeaponRotationPoint(GLVector3d & rot_point);
	
	/**	\brief Gets ID of player that controlls this tank
	 *
	 *	\return ID of player
	 */
	int	 getPlayerID() const;

	/**	\brief Tells if the player can fire now and returns projectile ID.
	 *
	 *	\param simulation_time Time that takes one simulation step in real, not in ODE
	 *	\param sim_timestamp timestamp of 'now' in simulation
	 *	\return ID of projectile that should be fired, or INVALID_PROJECTILE_ID if fireing is imposible
	 */
	int canFire( const double simulation_time, const int sim_timestamp );

	/**	\brief Sets weapon variables appropriate to new weapon
	 *
	 *	\param type_id Id of new weapon type
	 */
	void setWeaponServer( int type_id );

	virtual void handleCollisionWEntityServer( int ID, const GLVector3d &coll_pos, const GLVector3d &coll_normal );

	/**	\brief Set next available weapon.
	 */
	void setNextWeaponServer();

	/**
	 * \brief Return current weaponID
	 * INVALID_WEAPON_ID if no weapon
	 */
	int getWeapon() { return mWeapon;  }

	/**
	 * \brief Return current HP
	 */
	int getHP() { return mHP; }

	/**
	 * \brief Return maximum possible HP
	 */
	 int getMaxHP() { return mMaxHP; }

	/**
	 * \brief Return current armour
	 */
	int getArmour() { return mArmour; }

	/**
	 * \brief Return max armour
	 */
	int getMaxArmour() { return mMaxArmour; }


	/**
	 * \brief Return number of projectiles for current weapon.
	 *
	 * Zero if no weapon;
	 */
	int getProjectiles();

	/**
	 * \brief return true if tank is "working"
	 */
	 bool isAlive() { return mHP > 0; }

	/**
	 * \brief Set previous available weapon.
	 */
	void setPreviousWeaponServer();

	/**	\brief Makes weapon from tank available.
	 *
	 *	\param id Id of enabled weapon.
	 */
	void enableWeapon(int id);

	/**	\brief Add ammo.
	 *
	 *	\param id Ammo type.
	 *	\param number number of projectiles to add
	 */
	void addProjectiles(int id, int number);


	virtual void doDamage( int damage, bool piercing );

	virtual void checkDestroy( Player * player );

	virtual void updateServer( const int timestamp );


	/**
	 * \brief Implements only tank specific part of TanksHistoryPoint.
	 *
	 * But also calls orogonal.
	 */
	virtual void interpolateHistory( int timestamp );

	void setClientsMain() 
	{ 
		mClientsMain = true; 
	}

	/**	\brief Determines whether tank still lives or not
	 *
	 *	\return True, if tank is alive, false if it is killed
	 */
	bool lives()
	{
		return mHP > 0;
	}

	/**	\brief Called by simulation in case of fall from the map (leaving skybox)
	 */
	virtual void fallenServer();

	/**	\brief Commit suicide.
	 */
	void suicide()
	{
		mHP = -1000000;
		mSuicide = true;
		checkDestroy( 0 );
	}

	/**	\brief This method is called when tank has lost all of its HP.
	 *	\note Cannot be called on server.
	 */
	void killedCallbackClient();


	/**
	 * \brief The tank may modify damage if the projectiles. 
	 * 
	 * This is the mdification. Projectile will apply the damage multiplication itself.
	 * \retrun mDamageMuptiplication
	 */
	double getDamageMultiplication()
	{
		return mDamageMultiplication;	
	}

private:

	/**
	 * \brief Compute relative yaw and pitch to tanks position.
	 * 
	 * \param yaw   Yaw relatively to tank.
	 * \param pitch Pitch relatively to tank.
	 */
	void computeRelativeYawPitch(double &yaw, double &pitch) const;

	friend class Bonus;

	int		mHP;			///< \brief Current HPs of tank.
	int		mArmour;		///< \brief Current armour per cents of tank.

	int		mMaxHP;			///< \brief Maximum of HPs.
	int		mMaxArmour;		///< \brief Maximum of armour per cents (cannot be more than 100 :-))
	
	int		mWeight;		///< \brief Weight of the tank in kgs.
	
	double	mMaxSpeed;		///< \brief The top speed that can the tank achieve.
	int		mEnginePower;	///< \brief Power of engine in Watts.
	double	mEngineForce;	///< \brief Force of tank's engine. Computed from engine power and tank max speed

	double	mTurnSpeedAdd;	///< \brief Tracks in oposite of turning directiom moves this faster when turning

	/**
	 * \brief Addition to normal damage. 
	 * 
	 * The adition is percetages of normal damage. Projectile read the value automaticly on server
	 * when the projectile is creted. Default value is 1. Bonus may modify it.
	 */
	double  mDamageMultiplication; ///< \brief Tank has ability to have mora damage.
	     

	int     mExplosionID;   ///< \brief ID of explosion of this tank
	
	Player*	mPlayer;		///< \brief Player who controlls this tank.

	// This is usefull in client's decoding history points - because yaw and pitch
	// is determined by client and thus it should be ignired form client's tank
	// when it comes from server.
	bool    mClientsMain;   ///< \brief Only for client if the tank is client's main tank


	TankCommands mCmds;		///< \brief Commands that user applied to the tank

	//Some extra staff for weapon
	double	mYaw;			///< \brief	Yaw of the tank's weapon in degrees. Change only by setYawPitch. In World coordinates.
	double	mPitch;			///< \brief Pitch of the tank's weapon in degrees. Change only by setYawPitch. In World coordinates.
	int		mWeapon;		///< \brief ID of weapon currently armed
	double  mReloadTime;	///< \brief Time in milisecond to reload  currently armed weapon
	int     mLastFireTime;  ///< \brief Time of last attempt to fire in game-absolute time in miliseconds
	int     mProjectile;    ///< \brief Type of projectile that the weapon shoots

	bool	mSuicide;		///< \brief True, if player commited suicide in this tank

	int		mDestroyTime;	///< \brief Time in that the tank was destroyed
	int		mDestroyDelay;	///< \brief Time to wait before tank destruction after beeing killed

	weaponVectorType       mWeapons;     ///< \brief Which weapons the tank has
	projectilesNumberType  mProjectiles; ///< \brief Number of projectiles for use


	bonusListType mBonuses; ///< \brief List of bonuses associated with this tank

	WepObtainer mWepObtainer; ///< \brief obtains tank weapon point position and direction (in object space)
	
	float mNearNoAim; ///< \brief if target is nearer than this, the aiming is probably wrong. In meters.
};

/*@}*/ 
#endif

