/* MAINTAINER: Martin */

#ifndef _TERRAIN_H_
#define _TERRAIN_H_

#include <string>
#include <vector>
#include <memory>

#include "GLVector.h"
#include "terrain_base.h"

/** \addtogroup game_data */ /*@{ */

/**
*	\brief Terrain component for public usage.
*
*	It is a simple wrapper over some sophisticated terrain-class.
*
*	\see TerrainBase for details
*/
class Terrain
{
public:

	typedef std::vector<GLVector3d> THeights;	///< \brief height samples vector type for getHeightsInsideQuad() method

	Terrain();

	/**
	*	\brief initializes terrain on client 
	*	\note It dynamically creates an instance of CurrentTerrainImplementation.
	*/
	bool initClient();
	
	/** 
	*	\brief initializes terrain on server 
	*	\note It dynamically creates an instance of TerrainBase class only.
	*/
	bool initServer();

	/**
	*	\brief Load heightmap from file.
	*	\param fileName path to file with heightmap picture readable by SDL_Image
	*	\return true on success, otherwise false
	*	\note Width and height of the heightmap picture should be power of two plus one.
	*/
	bool loadHeightmap(const std::string fileName);
	
	// TODO: this method is not needed, lightmap can be loaded in initClient() by calling VisualTerrain method
	/// \brief Load lightmap file
	void loadLightmap(std::string path);

	// TODO: not needed, same as with lightmap
	/// \brief wrapped method
	void loadGlobalTexture(std::string path);
	
	/// \brief Render terrain.
	void render(void);
	
	/// \brief Update terrain LOD according to given camera
	void update(const class Camera * camera);
	
	/**
	*	\brief Get heightmap extens (in heightmap pixels)
	*	\param x number of heightmap pixels in X axis
	*	\param y number of heightmap pixels in Y axis
	*/
	void getExtents(int &x, int &y);
	
	/**
	*	\brief Get terrain height at point [x, y].
	*	
	*	\param x coordinate on X axis in virtual terrain coordinates
	*	\param y coordinate on Y axis in virtual terrain coordinates
	*	\return height at [x, y], scaled to virtual terrain height
	*	\note virtual terrain extents are set by setScale() method. These extens define logical terrain
	*	size and height ( eg. 1550.3 x 1500 x 800 meters ) 
	*/
	double getHeight(double x, double y);
	
	/**
	*	\brief Get heightmap value at point [x, y].
	*	
	*	\param x coordinate on X axis in heightmap pixels
	*	\param y coordinate on Y axis in heightmap pixels
	*	\return heightmap value at [x, y].
	*/
	unsigned char getTrueHeight(int x, int y);
	
	/**
	*	\brief Get heightmap value at point [x, y], but scaled to virtual terrain height.
	*	
	*	\param x coordinate on X axis in heightmap pixels
	*	\param y coordinate on Y axis in heightmap pixels
	*	\return heightmap value at [x, y] scaled to virtual terrain height.
	*	\see Terrain::getHeight(int &x, int &y)
	*/
	double getScaledHeight(int x, int y);
	
	/**
	*	\brief Get terrain maximal possible height in virtual coordinates
	*	\return maximal possible height
	*	\see Terrain::getHeight(int &x, int &y)
	*/
	double getScaledMaxHeight(void);
	
	/// \brief wrapped method
	GLVector3d& getScale(GLVector3d &scale);
	/// \brief wrapped method
	GLVector3d  getScale(void);
	/// \brief wrapped method
	void setScale(const GLVector3d &scale);

	/// \brief returns terrain base (class Terrain is a wrapper for class TerrainBase)
	TerrainBase * getTerrainBase(void) { return mTerrainBase.get(); }

	// \brief test if vector has valid x,y coordinates for testing terrain height
	bool isXYValid(const GLVector3d &vtotest);

	// \brief Crop x and y coordinate of vector such that x and y are valid
	void cropToValidXY(GLVector3d &vtocrop);

private:

	/** 
	*	\brief common initialization (after server or client is initialized)
	*/
	bool initCommon();

	/**
	*	\brief wrapped instance of some terrain implementation
	*	
	*	On client, TerrainBase if polymorphic class and some wrapped functionalities are obtained by
	*	<code>dynamic_cast<VisualTerrain>(mTerrainBase)</code>. VisualTerrain extends TerrainBase and it
	*	is an interface (abstract class) that is implemented by CurrentTerrainImplementation.
	*
	*	On server, on the other hand, PhysicalTerrain is the subclass that is created and used.
	*/
	std::auto_ptr< TerrainBase > mTerrainBase;
};

/*@}*/ 
#endif // _TERRAIN_H_
