/* MAINTAINER: Martin */

#include "terrain_base.h"

#include "SDL_image.h"
#include "SDL_rotozoom.h"

#include "log.h"
#include "grutils.h"

#include <cmath>

TerrainBase::TerrainBase():
	mSizeX(-1), mSizeY(-1),
	mScale(1.0f, 1.0f, 1.0f),
	mScaleInv(1.0f, 1.0f, 1.0f),
	mHeightMap(0)
{
}

TerrainBase::~TerrainBase()
{
	delete[] mHeightMap;
}

double TerrainBase::getHeight(double x, double y)
{
	double fx, fy, dx, dy, ffx, ffy, h = 0;
	int ix, iy;
	unsigned char h00=0, h01=0, h10=0, h11=0;

	// (fx, fy) is (x, y) scaled to heightmap size
	fx = x * mScaleInv[0];
	fy = y * mScaleInv[1];

	// real part of fx, fy (for interpolation)
	dx = fx - (ffx = floor(fx));
	dy = fy - (ffy = floor(fy));
	
	// integer part of fx, fy (for indexing purposes)
	ix = static_cast<int>(ffx);
	iy = static_cast<int>(ffy);

	h00 = getTrueHeight(ix, iy);
	h10 = (ix + 1 < mSizeX) ? getTrueHeight(ix + 1, iy) : h00;
	h11 = (ix + 1 < mSizeX && iy + 1 < mSizeY) ? getTrueHeight(ix + 1, iy +1) : h00;
	h01 = (iy + 1 < mSizeY) ? getTrueHeight(ix, iy + 1) : h00;
	h =
	(h00 * (1.0 - dx) + h10 * dx) * (1.0 - dy)
	+
	(h01 * (1.0 - dx) + h11 * dx) * dy;

	return h * mScale[2];

}

bool TerrainBase::allocateHeightmap(int sizeX, int sizeY)
{
	mSizeX = mSizeY = 0;
	if (mHeightMap != 0)
	{
		delete[] mHeightMap;
		mHeightMap = 0;
	}

	if (sizeX < 2 || sizeY < 2)
		return false;

	mHeightMap = new THeightMapValue [sizeX * sizeY];
	if (mHeightMap == 0)
		return false;
	mSizeX = sizeX;
	mSizeY = sizeY;
	return true;
}

bool TerrainBase::loadHeightmap(const std::string fileName)
{
	TIN;
	SDL_Surface *image;

	// Load the image file into a surface
	image = IMG_Load(fileName.c_str());
	if ( image == 0 ) {
		LOGE << "Heightmap loading: Can't load '"<< fileName << "' (" << IMG_GetError() << ")" << std::endl;
	    return false;
	}
	if (image->w <= 0 || image->h <= 0)
	{
		SDL_FreeSurface( image );
		LOGD << "Image of zero extents are not allowed!" << std::endl;
		return false;
	}

	// Check for 2^n + 1 extents. If not pass, then rescale (mostly down) to 2^n + 1.
	Uint32
		w( image->w ), h( image->h ), // given extents from the image
		w2( 1 << 31 ), h2( 1 << 31 ); // desired extents
	while ( ( w & w2 ) == 0 ) 
		w2 >>= 1;
	while ( ( h & h2 ) == 0 ) 
		h2 >>= 1;
	w2 = std::max<Uint32>( 3, w2 | 1 );
	h2 = std::max<Uint32>( 3, h2 | 1 );

	if ( w2 != w || h2 != h )
	// It needs to be rescaled.
	{
		SDL_Surface * zoomedSurf = zoomSurface( image, double( w2 ) / w, double( h2 ) / h, 0 );
		SDL_FreeSurface( image );
		image = zoomedSurf;
		if ( image == 0 )
		{
			LOGE << "En error occured when rescaling heightmap image." << std::endl;
			return false;
		}
	}

	// allocate heightmap
	if ( !allocateHeightmap( image->w, image->h ) ) 
	{
		LOGD << "Heightmap loading: Cant't allocate heightmap." << std::endl;
		SDL_FreeSurface( image );
	    return false;
	}
	
	grutils::SDL_Surface_Accessor sac(image);	

	const int
		width = std::min( sac.getWidth(), mSizeX ),	// X-boundary when reading pixels
		height  = std::min( sac.getHeight(), mSizeY ),	// Y-boundary when reading pixels
		hShift = mSizeX - width;						// how many bytes to skip at the end of each heightmap's line

	// flip surface about X axis and read pixels to heightmap array
	sac.lock();
	{
		THeightMapValue *h = mHeightMap;
		assertL( h != 0 && "Heightmap should be allocated!" );

		SDL_Color color;
		for ( int y = height - 1; y >= 0; --y, h += hShift )
			for ( int x = 0; x < width; ++x, ++h )
			{
				sac.getColorRGBA(x, y, color.r, color.g, color.b, color.unused);
				*h = createHeightValueFromColor(&color);
			}
	}
	sac.unlock();
	sac.free();
	return true;
}

GLVector3d & TerrainBase::normal(int x, int y, GLVector3d &n)
{
	// calculate normal
	int w, h;
	GLVector3d & scale = n;
	getExtents(w, h);
	getRealScale( scale );

	// center position
	bool up(y == 0), down(y == h - 1), left(x == 0), right(x == w - 1);
	
	// get heightmap values
	int hcenter = getTrueHeight(x, y);
	int hup = (!up) ? getTrueHeight(x, y - 1) : hcenter;
	int hleft = (!left) ? getTrueHeight(x - 1, y) : hcenter;
	int hright = (!right) ? getTrueHeight(x + 1, y) : hcenter;
	int hdown = (!down) ? getTrueHeight(x, y + 1) : hcenter;

	int dx = (hright - hleft);
	int dy = (hdown - hup);

	GLVector3d 
		vx((left | right) ? scale.x : 2.0 * scale.x, 0.0, dx * scale.z), 
		vy(0.0, (up | down) ? scale.y : 2.0 * scale.y, dy * scale.z);
	vx.cross(vy);
	
	double len = vx.length();
	if (len > 0.000001)
		vx *= 1.0 / len;
	else
		vx.set(0.0, 0.0, 1.0);

	n.set(vx.x, vx.y, vx.z);
	return n;
}

double TerrainBase::slope(int x, int y)
{
	GLVector3d n;
	const GLVector3d z(0.0, 0.0, 1.0);	
	double slope = fabs(z.dot(normal(x, y, n)));
	return 2 * acos(slope) / M_PI;
}

TerrainBase::LayerPtr TerrainBase::createLayer( int * layerIndex )
{
	LayerPtr layerPtr( mSurface.createLayer() );
	if ( layerPtr.get() != 0 && layerIndex )
		*layerIndex = static_cast<int>(mSurface.getLayersCount() - 1);
	return layerPtr;
}
