/* MAINTAINER: Martin */

#ifndef _TERRAIN_BASE_H_
#define _TERRAIN_BASE_H_

#include <string>
#include <vector>

#include "GLVector.h"
#include "SDL.h"

#include "terrain_surface.h"

/** \addtogroup game_data */ /*@{ */

/**
 *	\brief basic terrain class
 *
 *	Provides heightmap and extents manipulation routines.
 */
class TerrainBase
{
public:
	typedef Uint8 THeightMapValue;

	TerrainBase();
	virtual ~TerrainBase();

	/**
	*	\brief Loads heightmap.
	*
	*	Performs following steps:
	*	1) Loads picture given by FileName using SDL_Image, saves it temporary in SDL_Surface
	*	2) Converts to compact-form (1-byte-height-value)
	*	3) Frees SDL_Surface
	*
	*	\param fileName SDL_Image's supported picture format (*.BMP, *.PNG, ...)
	*	\return true on success, otherwise false
	*	\note Width and height of the heightmap picture should be power of two plus one.
	*/
	bool loadHeightmap(const std::string fileName);

	/**
	*	\brief Obtains terrain size (heightmap extents).
	*
	*	\param x heightmap extent in x axis
	*	\param y heightmap extent in y axis
	*/
	inline void getExtents(int &x, int &y);
		
	/**
	*	\brief Height at given point (interpolated, continuous case).
	*
	*	It scales x, y to Terrain's extent, determines height values of 
	*	closest neighbours (given by integer part of translated x, y) and 
	*	returns interpolated height from them (using non-integer part). 
	*	Heightmap values can be stored in the range of <0, 255>. 
	*	However, this function returns scaled height - height is
	*	scaled to <0.0, 1.0> and multiplied by previously set scale-factor.
	*
	*	\param x x-coordinate of point - should be in terrain's scaled extent
	*	\param y y-coordinate of point - should be in terrain's scaled extent
	*	\return interpolated height or some constant value if unknown
	*/
	double getHeight(double x, double y);

	/**
	*	\brief Height at given point (heightmap value, discrete case).
	*
	*	Performs heightmap lookup at map's x, y coordinates.
	*	If terrain paging is implemented, the page that contains x, y
	*	is determined first and the lookup is delegated to it.
	*
	*	\param x x-coord of point - should be in terrain's non-scaled extent
	*	\param y y-coord of point - should be in terrain's non-scaled extent
	*	\return heightmap value at x, y or some constant if unknown
	*/
	inline THeightMapValue getTrueHeight(int x, int y);

	/**
	*	\brief Height at given point (heightmap scaled value, discrete case).
	*
	*	Does the same as ::getTrueHeightAt(int x, int y) but scales the
	*	returned value.
	*
	*	\param x x-coord of point - should be in terrain's non-scaled extent
	*	\param y y-coord of point - should be in terrain's non-scaled extent
	*	\return  scaled heightmap value at x, y or some constant if unknown
	*	\see getTrueHeight
	*/
	inline double getScaledHeight(int x, int y);

	/**
	*	\brief Obtains the maximal height scale.
	*	\return maximal height scale
	*/
	inline double getScaledMaxHeight(void);
	
	/**
	*	\brief Obtains the current scale vector.
	*
	*	Current scale vector is a vector that describes the "public extents" of the terrain.
	*
	*	\param scale 3D scale vector (where to store result)
	*	\return current scale vector
	*	\note The scale vector describes, how much to scale the terrian's normalized extent
	*		<<0.0, 1.0>, <0.0, 1.0>, <0.0, 1.0>>. One should take care of aspect ratio when
	*		setting the scale.
	*/
	inline GLVector3d& getScale(GLVector3d &scale);
	
	/**
	*	\brief	Obtains a scale vector (vector, that describes the terrain extent).	
	*	\return current scale vector
	*	\see	getScale(GLVector3d &scale)
	*/
	inline GLVector3d  getScale(void);

	/** 
	*	\brief Sets a scale vector (vector, that describes the terrain extent).
	*
	*	\code
	*	// An example of setting "my own" scale in meters:
	*	// for example in meters (m)
	*	const double ToMyUnits = 5;	// "one heightmap pixel is a 5 m"
	*	double desiredTerrainHeightInMyUnits = 100;	// "maximal height will be 100 m"
	*	
	*	terrain.GetExtents(extX, extY); 
	*	terrain.setScale(GLVector3d(extX*ToMyUnits, extY*ToMyUnits, desiredTerrainHeightInMyUnits)) 
	*	\endcode
	*	
	*	Note that rendered output will be scaled by the scale vector.
	*
	*	\param scale 3D scale vector
	*	\note The scale vector describes, how much to scale the terrian's normalized extent
	*		<<0.0, 1.0>, <0.0, 1.0>, <0.0, 1.0>>. One should take care of aspect ratio when 
	*		setting the scale.
	*/
	inline void setScale(const GLVector3d &scale);

	GLVector3d & normal(int x, int y, GLVector3d &n);
	double slope(int x, int y);

	// ---------------------------- surface -----------------------------
	
	
	typedef TerrainSurface::LayersContainer::value_type LayerPtr;
	
	/**
	*	\brief creates a new layer
	*	\param layerIndex if not null, the index of the new layer
	*	\return layer pointer
	*/
	LayerPtr createLayer(int * layerIndex = 0);

	// ---------------------------- end surface -------------------------
	

private:
	
	/**
	*	\brief color-to-height value transform routine
	*	Calculates intensity (Uint8) from the RGB triplet and rounds it.
	*	\param col	RGB triplet.
	*	\return		8-bit unsigned height value
	*/
	inline THeightMapValue createHeightValueFromColor(SDL_Color * col);
	
	int mSizeX, mSizeY;	///< \brief terrain extents (physical size of the heightmap).

	GLVector3d mScale;		///< \brief height and extents scale.
	GLVector3d mScaleInv;	///< \brief inverted height and extents scale.

protected:
	THeightMapValue *	mHeightMap;	///< \brief heightmap values, mSizeX * mSizeY bytes
	
	/**
	*	\brief layered surface - set of layers
	*
	*	You can create your own set of layers. Each layer should be described by global weight 
	*	function (impact), it can have a global color map and periodic detail function. The first
	*	one could be used on server, the rest is useful for rendering on client.
	*
	*	For example, if you want to have a surface that consists of grass, rock and sand,
	*	you should create layer for grass, layer for rock, layer for sand.
	*/
	TerrainSurface		mSurface;
	
	/**
	*	\brief allocates memory and sets the heightmap pointer.
	*	\param sizeX the size of the heightmap in x axis
	*	\param sizeY the size of the heightmap in y axis
	*	\return false, if there is not enough memory ;-)
	*	\note If the heightmap was already allocated, it is deleted.
	*/
	bool allocateHeightmap(int sizeX, int sizeY);

public:

	/**
	*	\brief obtains the real scale-vector.
	*	
	*	This is mainly for rendering purposes. Others will use public getScale that
	*	is derived from normalized extent.
	*
	*	\return vector that scales terrain from its heightmap extents
	*		to virtual extents.
	*	\note This difers from public getScale which returns scale that is derived from
	*		normalized (1, 1, 1) extent.
	*/
	inline GLVector3d& getRealScale(GLVector3d &scale);

	/**
	*	\brief obtains the real scale-vector.
	*	\return vector that scales terrain from its heightmap extents
	*		to virtual extents.
	*	\see getRealScale(GLVector3d &scale)
	*/
	inline GLVector3d  getRealScale(void);

	/**
	*	\brief obtains the real scale-vector inverted.
	*	\return vector that scales terrain from its virtual extents to heightmap extents.
	*	\see getRealScale(GLVector3d &scale)
	*/
	inline GLVector3d& getRealScaleInv(GLVector3d &scaleInv);

	/**
	*	\brief obtains the real scale-vector inverted.
	*	\return vector that scales terrain from its virtual extents to heightmap extents.
	*	\see getRealScale(GLVector3d &scale)
	*/
	inline GLVector3d  getRealScaleInv(void);	

	// NOTE (author's personal):
	//	Don't attempt to even think about terrain paging! 
	//	It's NOT really needed in this project.
	//	"All-In-Mem heightmap will be suficient for everybody." ;-)
};
/*@}*/

// ------------- Inline Methods Definition ----------------

void TerrainBase::getExtents(int &x, int &y)
{
	x = mSizeX; y = mSizeY;
}
/*
void TerrainBase::getMixMapExtents(int &extentX, int &extentY)
{
	extentX = mMixMapExtentX; extentY = mMixMapExtentY;
}

const TerrainBase::MixMapValue & TerrainBase::getMixMapValue(int x, int y)
{
	assert(mMixMapExtentX >= 0);
	assert((x >= 0) && (x < mMixMapExtentX) & (y >= 0) && (y < mMixMapExtentY));
	
	// return get_height(x, y)
	return mMixMap[y * mMixMapExtentX + x];
}

void TerrainBase::setMixMapValue(int x, int y, const TerrainBase::MixMapValue & v)
{
	assert(mMixMapExtentX >= 0);
	assert((x >= 0) && (x < mMixMapExtentX) & (y >= 0) && (y < mMixMapExtentY));
	
	// return get_height(x, y)
	mMixMap[y * mMixMapExtentX + x] = v;
}
*/
TerrainBase::THeightMapValue TerrainBase::getTrueHeight(int x, int y)
{
	if (!((x >= 0) & (x < mSizeX) & (y >= 0) & (y < mSizeY)))
		return 0;
	
	// return get_height(x, y)
	return mHeightMap[y * mSizeX + x];
}

double TerrainBase::getScaledHeight(int x, int y)
{
	return getTrueHeight(x, y) * mScale[2];
}

double TerrainBase::getScaledMaxHeight(void)
{
	return mScale[2] * 255.0f;
}

GLVector3d& TerrainBase::getScale(GLVector3d &scale)
{
	scale[0] = mScale[0] * (mSizeX - 1);
	scale[1] = mScale[1] * (mSizeY - 1);
	scale[2] = mScale[2] * (255.0f);
	return scale;
}

GLVector3d TerrainBase::getScale()
{
	GLVector3d temp;
	return getScale(temp);
}

GLVector3d& TerrainBase::getRealScale(GLVector3d &scale)
{
	return scale = mScale;
}

GLVector3d TerrainBase::getRealScale()
{
	GLVector3d temp;
	return getRealScale(temp);
}

GLVector3d& TerrainBase::getRealScaleInv(GLVector3d &scaleInv)
{
	return scaleInv = mScaleInv;
}

GLVector3d TerrainBase::getRealScaleInv()
{
	GLVector3d temp;
	return getRealScaleInv(temp);
}

void TerrainBase::setScale(const GLVector3d &scale)
{
	mScale[0] = scale[0] * (1.0f/(mSizeX - 1));
	mScale[1] = scale[1] * (1.0f/(mSizeY - 1));
	mScale[2] = scale[2] * (1.0f/255.0f);
	mScaleInv[0] = 1.0f / mScale[0];
	mScaleInv[1] = 1.0f / mScale[1];
	mScaleInv[2] = 1.0f / mScale[2];
}

TerrainBase::THeightMapValue TerrainBase::createHeightValueFromColor(SDL_Color * col)
{ return static_cast<Uint8>(0.299f*col->r + 0.587f*col->g + 0.114f*col->b + 0.5f); }

#endif // _TERRAIN_BASE_H_
