/* MAINTAINER: Petko */
#ifndef _WORLD_H
#define _WORLD_H

#include "landscape.h"
#include "stlextdefs.h"
#include "hash_hack.h"

#include <list>

class WorldEntity;
class BonusBox;
class Projectile;
class Tank;
class Bonus;

/**	\defgroup game_data Game Data
 *	\brief In this module are stored all whole-game relevant data.
 *	\ingroup common
 *
 *	All data of the game are stored in classes included in this module.
 */


/** \addtogroup game_data */ /*@{ */


	/// \brief Defines type for sequential list of entities
typedef std::list< WorldEntity* >			worldent_list;	

	/// \brief Defines iterator for sequential pass of components
typedef std::list< WorldEntity* >::iterator	worldent_iterator;

	/// \brief Type for hash table of objects
typedef STL_EXT_NS::hash_map< int, WorldEntity* >	worldent_hash;
	

/**	\brief Contains all objects in game and manages them
 *
 *	Class that is owner of Landscape (with its Terrain, Skybox and LandscapeEntities) and 
 *	WorldEntities. It also manages WorldEntities and provides access to Landscape.
 */
class World
{
public:

	World(void);
	~World(void);

	/** \brief Gets the start of the list of objects.	*/
	worldent_iterator	getWorldEntitiesBegin();

	/** \brief Gets the end of the list of objects		*/
	worldent_iterator	getWorldEntitiesEnd();


		// Acces methods to the objects

	/**	\brief Finds ant returns entity with given ID
	 *
	 *	\param id ID of wanted object
	 *	\return Pointer to the found object, or NULL if doesn't exists object with that ID
	 */
	WorldEntity*		getWorldEntity(int id);

	/**	\brief Finds ant returns entity with given iterator
	 *
	 *	\param iterator Iterator to list of objects.
	 *	\return Pointer to the found object, or NULL if the iterator was equal to end().
	 */
	WorldEntity*		getWorldEntity(worldent_iterator iterator) const;
		
	/** \brief Provides access to Landscape				*/
	Landscape&			getLandscape();


		//Object construction methods - for every type of object one:

	/**	\brief Creates new bonus box on the map
	 *
	 *	\param position Position on the map where should new object appear.
	 *	\param orientation Orientation of the object on the map, in which it should appear.
	 *	\param bonus_function_id ID of bonus type - its effect
	 *	\param id Required ID of object - if zero, it's assigned some value automatically
	 *	\param create_bonus If true, Bonus object is created, not otherwise
	 *	\return Pointer to newly created object, or NULL, if not successful.
	 */
	BonusBox*	createBonusBox( GLVector3d position, GLQuaterniond orientation, int bonus_function_id, int id = 0, bool create_bonus = true );
	// bonus is deleted if creation / initialization of bonus box fails
	BonusBox*	createBonusBox( GLVector3d position, GLQuaterniond orientation, Bonus* bonus, int id = 0 );

	/**	\brief Creates new projectile in the world
	 *
	 *	\param position Position on the map where should new object appear.
	 *	\param orientation Orientation of the object on the map, in which it should appear.
	 *	\param type_id ID of projectile type
	 *	\param shooter ID of player who shot this projectile
	 *	\param id Required ID of object - if zero, it's assigned some value automatically
	 *	\return Pointer to newly created object, or NULL, if not successful.
	 */
	Projectile*	createProjectile( GLVector3d position, GLQuaterniond orientation, int type_id, int shooter, int id = 0 );

	/**	\brief Creates new tank in the world
	 *
	 *	Creates new tank in the world and associates it with player 'player_id'. After this is called addedWorldObject callback
	 *
	 *	\param position Position on the map where should new object appear.
	 *	\param orientation Orientation of the object on the map, in which it should appear.
	 *	\param type_id ID of tank type
	 *	\param player_id ID of player who controls this tank
	 *	\param id Required ID of object - if zero, it's assigned some value automatically
	 *	\return Pointer to newly created object, or NULL, if not successful.
	 */
	Tank*		createTank( GLVector3d position, GLQuaterniond orientation, int type_id, int player_id, int id = 0 );


		// Object destruction methods - for every type of object one

	/**	\brief Deletes a bonus box from the world.
	 *
	 *	\param id ID of object to delete.
	 *	\param delete_bonus Tells, whether the method has to delete bonus's effect class, too.
	 *	\return True if successful.
	 */
	bool		deleteBonusBox( int id, bool delete_bonus = false );

	/**	\brief Deletes a projectile from the world.
	 *
	 *	\param id ID of object to delete.
	 *	\return True if successful.
	 */
	bool		deleteProjectile( int id );

	/**	\brief Deletes a tank from the world.
	 *
	 *	\param id ID of object to delete.
	 *	\return True if successful.
	 */
	bool		deleteTank( int id );

	/**	\brief Deletes objects - it detects which objects it is and then calls appropriate delete function
	 *	\note It calls deleteBonusBox with second parameter true, so don't use it when somebody takes the bonus!
	 *
	 *	\param id ID of object to delete.
	 *	\return True if successful.
	 */
	bool		deleteObjectGeneric( int id );

	/*	TODO: Zmazat tohto frienda!!!! 
		Je to tu z testovacich dovodov!!!!
	*/
	friend class TaskSimulationServer;

	/**
	 * Return next unique ID in world.
	 * Always return positive number
	 */
	int getNextID();

	/**	\brief Marks all entities, that have some relation to player ID, for deletion
	 *
	 *	\param ID ID of player whose entities should be deleted
	 */
	void deletePlayersEntities( int ID );

	/**	\brief Something like destructor - destroys all entities in world
	 *
	 */
	void cleanUp();

	void loadedMap();

private:

	void addWorldEntity(WorldEntity* we);
	
	void delWorldEntity(WorldEntity* we);
	
	Landscape	*mLandscape;

	// container of world objects
	worldent_list	mWorldList;		///< \brief List of all movable objects
	worldent_hash	mWorldHash;		///< \brief Hash table of all movable objects
		
	int mMaxID;						///< \brief Maximal unique ID reached so far

	World( const World & );			///< \brief Copy constructor is private, because we don't want anybody to copy that
	
};

/*@}*/ 

#endif // _WORLD_H

