/* MAINTAINER: Petko */
#include "worldentity.h"
#include "log.h"
#include "history.h"
#include "application.h"
#include "game.h"

WorldEntity::WorldEntity()
{
	mSpeed = GLVector3d( 0, 0, 0 );
	mAngVelocity = GLVector3d( 0, 0, 0 );

	mPresentPoint = 0;
	mHistoryUpdated = false;

	for ( int i = 0; i < HISTORY_LENGTH; i++ )
		mHistoryPoints[i]= 0;

	mDeleteMe = false;
	mSpeedChange = true;
	mUnPredictable = true;
}

WorldEntity::~WorldEntity()
{
	for ( int i = 0; i < HISTORY_LENGTH; i++ )
	{
		if ( mHistoryPoints[i] )
			delete mHistoryPoints[i];
	}
	mPresentPoint = 0;
}

void WorldEntity::shiftHistory()
{
	if ( mHistoryPoints[ HISTORY_LENGTH - 1 ] )
		delete mHistoryPoints[ HISTORY_LENGTH - 1 ];

	for ( int i = HISTORY_LENGTH - 1; i > 0; --i )
		mHistoryPoints[i]= mHistoryPoints[ i-1 ];

	mHistoryPoints[0] = mPresentPoint;
}

void WorldEntity::createPresentPoint( int time_stamp )
{
	mPresentPoint = new HistoryPoint;

	mPresentPoint->mOrientation = mOrientation;
	mPresentPoint->mPosition	= mPosition;
	mPresentPoint->mSpeed		= mSpeed;
	mPresentPoint->mAngVelocity	= mAngVelocity;

	mPresentPoint->mTimeStamp	= time_stamp;
	mPresentPoint->mID			= mID;
}

HistoryPoint * WorldEntity::getPresentPoint()
{
	return mPresentPoint;
}

void WorldEntity::updateHistory( HistoryPoint * new_point )
{
	mPresentPoint= new_point;
	shiftHistory();
	mHistoryUpdated = true;
}

void WorldEntity::commitHistoryUpdate( int time )
{
	if ( !mHistoryUpdated && mPresentPoint )		//entity wasn't updated in the last update
	{
		if ( mHistoryPoints[0] && mHistoryPoints[1] )	//if there are at least two histry points, I can make decision:
		{
			if			// both previous states are same -> object doesn't move => copy present point
			( 
				vecEqual( mHistoryPoints[0]->mPosition, mHistoryPoints[1]->mPosition, 0.05 ) &&
				vecEqual( mHistoryPoints[0]->mOrientation.getV(), mHistoryPoints[1]->mOrientation.getV(), 0.05 ) &&
				dblEqual( mHistoryPoints[0]->mOrientation.getW(), mHistoryPoints[1]->mOrientation.getW(), 0.05 )
			)
			{
			}
			else		//there's change -> object's moving -> apply gravity and move it
			{
				double rtime = (double)time/1000.0;								//ms to s
				mPresentPoint->mPosition += mPresentPoint->mSpeed * rtime;		//add change from velocity
				mPresentPoint->mPosition.z -=									//add change from gravity (s=(a*t^2)/2)
					(getApp().getGameCommon().getMapParams().gravity_constant * rtime * rtime) / 2;
				mPresentPoint->mSpeed += getApp().getGameCommon().getMapParams().gravity_constant * time;

				double degrees = ( mPresentPoint->mAngVelocity.length() * M_PI / 180.0 ) * rtime;
				GLQuaterniond chng( mPresentPoint->mAngVelocity, degrees );
				chng.normalize();
				mPresentPoint->mOrientation *= chng;
			}
		}

		HistoryPoint * new_point = mPresentPoint->copy();
		updateHistory( new_point );
	}

	mHistoryUpdated = false;
}

void WorldEntity::interpolateHistory(int timestamp)
{
	if ( mPresentPoint == 0 )
	{
		//LOGW << "mPresentPoint is 0!" << endl;
		return;
	}
	mPosition	= mPresentPoint->mPosition;
	mOrientation= mPresentPoint->mOrientation;
	mSpeed		= mPresentPoint->mSpeed;
}
