/* MAINTAINER: Petko */
#ifndef _WORLD_ENTITY_H
#define _WORLD_ENTITY_H

#include "commonentity.h"
#include "entity_public_info.h"

/** \addtogroup game_data */ /*@{ */

const int HISTORY_LENGTH = 5;

class Player;
class HistoryPoint;
struct NETGameState;

/**	\brief Represents objects in game, that can change theirs state (move).
 *
 *
 *
 */
class WorldEntity : public CommonEntity
{
public:
	WorldEntity();
	virtual ~WorldEntity() = 0;
	
	/**	\brief Shifts interpolation history by one step
	 *
	 *	
	 */
	void shiftHistory();

	/**	\brief Creates from current available data new mPresentPoint
	 *
	 *	\param time_stamp Defines the timestamp of the point being created
	 */
	virtual void createPresentPoint( int time_stamp );

	/**	\brief Return history point that represents current state
	 *
	 *	\return history point
	 */
	HistoryPoint * getPresentPoint();

	/**	\brief Adds new point to the history and present point.
	 *
	 *	Changes the present history point to the new point and then shifts the
	 *	history. WARNING! It doesn't delete the present point. This is behaviour 
	 *	that is expected as long as mPresentPoint and mHistoryPoints[0] points to
	 *	the same objects (this is correct).
	 *
	 *	\param new_point Point that should be added as the beginning of history :P
	 */
	void updateHistory( HistoryPoint * new_point );

	/**	\brief If entity was updated, resets the 'history updated' flag, otherwise duplicates present history point or calculate new position, if allowed
	 *
	 *	\param time Time that should elapsed since last gamestate commit
	 */
	void commitHistoryUpdate( int time );

	/**
	 * Pack new created entity
	 *
	 * Packs data that is needed for creating object on client.
	 *
	 * \param data net game state structure where data will be packed
	 */
	virtual void packCreationData(NETGameState & data) = 0;

	/**	\Brief Sets the data of entity by data in history
	 *
	 *	\note Zatial len akoze!!!!
	 *
	 *	\param timestamp What is the current 'time'
	 */
	virtual void interpolateHistory( int timestamp );

	/**	\brief Returns object's weight.
	 *
	 *	\return Object's weight.
	 */
	double getWeight() const
	{
		return mWeight;
	}

	/**	\brief Returns object's speed.
	 *
	 *	\return Object's speed.
	 */
	const GLVector3d& getSpeed() const
	{
		return mSpeed;
	}

	/**	\brief Sets new speed vector that will be updatesd in object in next simulation step.
	 *
	 *	\param newspeed New speed vector.
	 */
	void setSpeed( const GLVector3d & newspeed )
	{
		mSpeed = newspeed;
		mSpeedChange = true;
	}

	/**	\brief Returns object's angular velocity.
	 *
	 *	\return Object's angular velocity.
	 */
	const GLVector3d& getAngVelocity() const
	{
		return mAngVelocity;
	}

	/**	\brief Sets new angular velocity vector.
	 *
	 *	\note This has no effect for object movement. For internal use only.
	 *	\param newspeed New angular velocity vector.
	 */
	void setAngVelocity( const GLVector3d & newavel )
	{
		mAngVelocity = newavel;
	}

	/**	\brief Sets new speed vector that will NOT be updated in object in next simulation step.
	 *	\note Use only when you now what you are doing!
	 *
	 *	\param newspeed New speed vector.
	 */
	void setSpeedSilent( const GLVector3d & newspeed )
	{
		mSpeed = newspeed;
	}

	/**	\brief Returns true if the speed was updated using setSpeed()
	 *
	 *	\return true if speed changed
	 */
	bool getSpeedChange() const
	{
		return mSpeedChange;
	}

	/**	\brief Resets mSpeedChange - that means speed change was handled
	 */
	void resetSpeedChange()
	{
		mSpeedChange = false;
	}

	

	/**	\brief Entity-type-specific handler of collision with landscape
	 *
	 *	This method is called every time, if this entity touches terrain or any landscapeentity.
	 *	You can perform some collision handling here - but NO PHYSICS.
	 *	Collisions between tanks' wheels and terrain DON'T call this method.
	 *	If there are more collisions with this object, the order of handlers is 
	 *	not defined - so do not assume any order!
	 *
	 *	\param coll_pos Position of the detected collision - in world coordinates - if there are more contacts, one is (undefined how) chosen
	 *	\param coll_normal Normal of the contact surface - it is undefined which normal this will be
	 */
	virtual void handleCollisionWStaticServer( const GLVector3d & coll_pos, const GLVector3d & coll_normal ){}

	/**	\brief Entity-type-specific handler of collision with projectiles
	 *
	 *	This method is called every time, if this entity is hit by projectile.
	 *	You can perform some collision handling here - but NO PHYSICS.
	 *	After this handler is called, normal handleCollisionWEntityServer() is also called,
	 *	so put into this method only shot-hit-it specific code.
	 *	There are NO collisions between projectiles.
	 *	If there are more collisions with this object, the order of handlers is 
	 *	not defined - so do not assume any order!
	 *
	 *	\param ID ID of second projectile that hit this object.
	 *	\param coll_pos Position of the detected collision - in world coordinates - if there are more contacts, one is (undefined how) chosen
	 *	\param coll_normal Normal of the contact surface - it is undefined which normal this will be
	 */
	virtual void handleCollisionWProjectileServer( int ID, const GLVector3d & coll_pos, const GLVector3d & coll_normal ){}

	/**	\brief Entity-type-specific handler of collision with world entities
	 *
	 *	This method is called every time, if this entity touches another - including projectiles.
	 *	You can perform some collision handling here - but NO PHYSICS.
	 *	If there are more collisions with this object, the order of handlers is 
	 *	not defined - so do not assume any order!
	 *
	 *	\param ID ID of object that collided with this object
	 *	\param coll_pos Position of the detected collision - in world coordinates - if there are more contacts, one is (undefined how) chosen
	 *	\param coll_normal Normal of the contact surface - it is undefined which normal this will be
	 */
	virtual void handleCollisionWEntityServer( int ID, const GLVector3d & coll_pos, const GLVector3d & coll_normal ){}

	/**	\brief Returns mDelete, it tells, if it is time to delete this object
	 *
	 *	\return true, if object has to be deleted in the end of this simulation step
	 */
	virtual bool getDeleteFlag() const
	{
		return mDeleteMe;
	}

	/**	\brief Sets the delete flag, so the entity will be deleted at next simulation pass (on server)
	 */
	virtual void setDeleteFlag()
	{
		mDeleteMe = true;
	}

	/**	\brief Changes state of entity appropriate to hit by missile 
	 *
	 *	\param damage Damage that makes missile
	 *	\param piercing True, if missile ignores armour
	 */
	virtual void doDamage( int damage, bool piercing ){}

	/**	\brief Has to check if the entity should still live after hit by projectile from player
	 *
	 *	\param player Player that fired shot that hit this object
	 */
	virtual void checkDestroy( Player * player ){}

	/**	\brief This method is called in every simulation step
	 *
	 *	\param timestep Timestamp of current simulation time
	 */
	virtual void updateServer( const int timestamp ) {}

	/**	\brief This method is called in every simulation step on client.
	 *
	 *	This method is called (on client) from simulation task. It should pass some timestamp
	 *	that defines current game time. This is not availanle yet!
	 *
	 *	\note Now, timestamp will be 0 allways!
	 *
	 *	\param timestep Timestamp of current simulation time
	 */
	virtual void updateClient( const int timestamp );

	/**	\brief Simulation calls this when object leaves the skybox
	 */
	virtual void fallenServer(){}

	/**	\brief Sets the flag that this entity cannot be predicted on the client
	 *	\note For simulation use only!
	 */
	void setUnpredictableFlag()
	{
		mUnPredictable = true;
	}

	/**	\brief Resets the flag that this entity could be predicted on the client
	 *	\note For netwrok use only!
	 */
	void resetUnpredictableFlag()
	{
		mUnPredictable = false;
	}

	/**	\brief Gets flag if entity can be predicted on the client
	 *
	 *	\return UnPredictable flag
	 */
	// TODO LUKAS dodelat ze UnpredictableFlag() bude pro tank testovat zda se mu
	// zmenili HP atd
	bool getUnpredictableFlag() const
	{
		return mUnPredictable;
	}

protected:
		//WorldEntity represents movable objects, so we need speed vector and weight...
	GLVector3d	mSpeed;			///< \brief Vector of current speed
	bool		mSpeedChange;	///< \brief true if speed has been changed using setSpeed()
	GLVector3d	mAngVelocity;	///< \brief Angular velocity of object. Got from ODE.
	double		mWeight;		///< \brief Weight of object in kg

		//Data for interpolation
	HistoryPoint * mPresentPoint;					///< \brief Point in history, that is now valid
	HistoryPoint * mHistoryPoints[HISTORY_LENGTH];	///< \brief Here are cached some points of history	
	bool		mHistoryUpdated;					///< \Whether the history was updated

	bool		mDeleteMe;		///< \brief Tells, if it is time to delete this object

	bool		mUnPredictable;	///< \brief The simulation sets this to true, if there are no foreces except gravity

	/** \brief get public info about this entity 
	 */
	EntityPublicInfoPtr getPublicInfo() const 
	{ 
		return mPublicInfo; 
	}

	/** \brief set the entity public info 
	 */
	void setPublicInfo( EntityPublicInfoPtr value ) 
	{ 
		mPublicInfo = value; 
	}

private:
	EntityPublicInfoPtr mPublicInfo; ///< \brief Position, etc. - useful for explosions, survives entity destruction

};
	
/*@}*/ 
#endif // _WORLD_ENTITY_H
