/* MAINTAINER: Bernard */

#ifndef _GLOBALS_H
#define _GLOBALS_H

#include "GLVector.h"

extern const char* BJS_VERSION_STRING;

class Application;

/**
 * \brief Pointer to global Application class.
 *
 * It may be 0 - when trying to use it in constructors that are
 * called when creating Application object. Dotn't cast this to
 * (ApplicationClient*) or (ApplicationServer*). Use gAppClient or
 * gAppServer instead.
 * Don't forget to inicialize this in main :)
 */
extern Application *gApp;

// global constants

/**
 * \brief List of subdirectories with data
 *
 * Last member is 0.
 */
extern const char* SUBDIRECTORIES[];

/**
 * \brief Vector with zero rotation.
 * 
 * If you need vector of some value - speed as value and rotate it
 * use speed * ZERO_ROTATION_VECTOR and then rotate and you get 
 * oriented speed.
 */
extern const GLVector3d ZERO_ROTATION_VECTOR;

/**	\brief Tolerrance in comparing doubles for equality
 */
extern const double DBL_TOLER;

/**
 * \brief The message will not be translated.
 *
 * If there is a message that shoudn't be translated from server
 * to client for example the first char of message should be
 * NOT_TRANSLATE_CHAR. Else there will by attempt to translate
 * the message. It is for for example user chat messages...
 */
extern const char NOT_TRANSLATE_CHAR;

/**
 * \brief ID which is never a valid player ID
 */
extern const int INVALID_PLAYER_ID;

/**
 * \brief ID which is never a valid entity (landscape or world) ID
 */
extern const int INVALID_ENTITY_ID;

/**
 * \brief ID which is never a valid weapon ID
 */
extern const int INVALID_WEAPON_ID;

/**
 * \brief ID which is never a valid projectile ID
 */
extern const int INVALID_PROJECTILE_ID;

/**
 * \brief ID which is never a valid explosion ID
 */
extern const int INVALID_EXPLOSION_ID;

/**
 * \brief first game state number that will be send over network
 */
extern const int GS_START_NUMBER;


/**
 * \brief Safer version of *gApp
 */
Application& getApp();

/**	\brief Checks for equality of doubles with some tolerance
 *
 *	\param a One of numbers
 *	\param b One of numbers
 *	\return True if approximately equal
 */
inline bool dblEqual( const double a, const double b )
{
	return (std::abs( a - b ) < DBL_TOLER );
}

/**	\brief Checks for equality of vectors with some tolerance
 *
 *	\param a One of vectors
 *	\param b One of vectors
 *	\return True if approximately equal
 */
inline bool vecEqual( const GLVector3d & a, const GLVector3d & b )
{
   return ( dblEqual( a.x, b.x ) && dblEqual( a.y, b.y ) && dblEqual( a.z, b.z ) );
}

/**	\brief Checks for equality of doubles with some tolerance
 *
 *	\param a One of numbers
 *	\param b One of numbers
 *	\param toler Tolerance
 *	\return True if approximately equal
 */
inline bool dblEqual( const double a, const double b, const double toler )
{
	return (std::abs( a - b ) < toler );
}

/**	\brief Checks for equality of vectors with some tolerance
 *
 *	\param a One of vectors
 *	\param b One of vectors
 *	\param toler Tolerance
 *	\return True if approximately equal
 */
inline bool vecEqual( const GLVector3d & a, const GLVector3d & b, const double toler )
{
   return ( dblEqual( a.x, b.x, toler ) && dblEqual( a.y, b.y, toler ) && dblEqual( a.z, b.z, toler ) );
}

/**	\brief Defines the size of header part of the packet (packets) that are used to exchange IOR and other server information
 *
 *	Defines the size of header part of the packet (packets) that are used to exchange IOR and other server information.
 *	It is in text and decimal. For example, the packet data could look like: 
 *	'78       Here begins data. It has some size and it is stored in the beginning of packet'. In header there's only
 *	size of data stored - size of header is ommited.
 *	Fill in the size of data in the front of header. The last byte of header is reserved - fill with space
 */
extern const int IOR_TRANSFER_HEADER_SIZE;

/**	\brief Maximum size of data that are transfered in IOR exchange as data (header not included in this)
 */
extern const int IOR_TRANSFER_DATA_MAX_SIZE;

/**	\brief Returns true, if this application is running as server
 *	
 *	\return True, if server, false otherwise
 */
bool isServer();


#endif
