#include "grutils.h"
#include "log.h"


namespace grutils
{

bool SDL_Surface_ReadOnlyAccessor::isValid()
{
	return mSurfaceWrapper.surface != 0 && mSurfaceWrapper.surface->format != 0;
}

void SDL_Surface_ReadOnlyAccessor::setSurface(const SDL_Surface * surf)
{
	SurfaceWrapper tmp(surf);
	mSurfaceWrapper = tmp;
}

SDL_Surface_ReadOnlyAccessor::SDL_Surface_ReadOnlyAccessor(const struct SDL_Surface * surface)
	: mSurfaceWrapper(surface) 
{}

SDL_Surface_ReadOnlyAccessor::SDL_Surface_ReadOnlyAccessor()
	: mSurfaceWrapper( 0 ) 
{}

int SDL_Surface_ReadOnlyAccessor::getWidth()
{
	assertL(mSurfaceWrapper.surface != 0 && "Attempt to access null SDL surface!");
	return mSurfaceWrapper.surface->w;
}

int SDL_Surface_ReadOnlyAccessor::getHeight()
{
	assertL(mSurfaceWrapper.surface != 0 && "Attempt to access null SDL surface!");
	return mSurfaceWrapper.surface->h;
}

void SDL_Surface_ReadOnlyAccessor::getColorRGBA(int x, int y, Uint8 &r, Uint8 &g, Uint8 &b, Uint8 &a) const
{
	const SDL_Surface *s(mSurfaceWrapper.surface);

	assertL(s != 0 && "Attempt to access null SDL surface!");
	assertL(x >= 0 && y >= 0 && x < s->w && y < s->h && "Surface coordinates out of bounds!");
	assertL(s->format != 0 && "Attempt to acces null pixel format of a SDL surface!");

	SDL_GetRGBA(
		getpixel(
			reinterpret_cast<Uint8*>(s->pixels) + y * s->pitch + x * s->format->BytesPerPixel,
			s->format ),
		s->format,
		&r, &g, &b, &a);
}

SDL_Surface_Accessor::SDL_Surface_Accessor(struct SDL_Surface * surface)
	: SDL_Surface_ReadOnlyAccessor(surface), mSurface(surface)
{}
	
void SDL_Surface_Accessor::lock()
{
	assertL(mSurface != 0 && "Attempt to lock null SDL surface!");
	SDL_LockSurface(mSurface);
}

void SDL_Surface_Accessor::unlock()
{
	assertL(mSurface != 0 && "Attempt to unlock null SDL surface!");
	SDL_UnlockSurface(mSurface);
}

void SDL_Surface_Accessor::setColorRGBA(int x, int y, Uint8 r, Uint8 g, Uint8 b, Uint8 a)
{
	assertL(mSurface != 0 && "Attempt to access null SDL surface!");
	assertL(x >= 0 && y >= 0 && x < mSurface->w && y < mSurface->h && "Surface coordinates out of bounds!");
	assertL(mSurface->format != 0 && "Attempt to acces null pixel format of a SDL surface!");

	putpixel(
		reinterpret_cast<Uint8*>(mSurface->pixels) + y * mSurface->pitch + x * mSurface->format->BytesPerPixel, 
		SDL_MapRGBA(this->mSurface->format, r, g, b, a),
		mSurface->format
	);
}

void SDL_Surface_Accessor::free()
{
	if ( isValid() )
	{
		SDL_FreeSurface( mSurface );
		setSurface( mSurface = 0 );
	}
}

Uint32 getpixel(const Uint8 * pixels, const SDL_PixelFormat * pf)
{
	switch (pf->BytesPerPixel) {
	case 1:
		return *pixels;
	case 2:
		return *(Uint16 *)pixels;
	case 3:
		#if SDL_BYTEORDER == SDL_BIG_ENDIAN
			return (Uint32)pixels[0] << 16 | (Uint32)pixels[1] << 8 | (Uint32)pixels[2];
		#else
			return (Uint32)pixels[0] | (Uint32)pixels[1] << 8 | (Uint32)pixels[2] << 16;
		#endif
	case 4:
		return *(Uint32 *)pixels;
	default:
		return 0;       /* shouldn't happen, but avoids warnings */
	}
}

void putpixel(Uint8 * pixels, const Uint32 pixel_value, const SDL_PixelFormat * pf)
{
	switch(pf->BytesPerPixel) {
	case 1:
		*pixels = pixel_value;
		break;
	case 2:
		*(Uint16 *)pixels = pixel_value;
		break;
	case 3:
		#if SDL_BYTEORDER == SDL_BIG_ENDIAN
			pixels[0] = (pixel_value >> 16) & 0xff;
			pixels[1] = (pixel_value >> 8) & 0xff;
			pixels[2] = pixel_value & 0xff;
		#else
			pixels[0] = pixel_value & 0xff;
			pixels[1] = (pixel_value >> 8) & 0xff;
			pixels[2] = (pixel_value >> 16) & 0xff;
		#endif
		break;
	case 4:
		*(Uint32 *)pixels = pixel_value;
		break;
	}
}

// copy-pasted from NeHe ;-)
GLvoid glDrawCube()          // Draw A Cube
{
	glBegin(GL_QUADS);      // Start Drawing Quads
	// Front Face
	glNormal3f( 0.0f, 0.0f, 1.0f);    // Normal Facing Forward
	glTexCoord2f(0.0f, 0.0f); glVertex3f(-1.0f, -1.0f,  1.0f);  // Bottom Left Of The Texture and Quad
	glTexCoord2f(1.0f, 0.0f); glVertex3f( 1.0f, -1.0f,  1.0f);  // Bottom Right Of The Texture and Quad
	glTexCoord2f(1.0f, 1.0f); glVertex3f( 1.0f,  1.0f,  1.0f);  // Top Right Of The Texture and Quad
	glTexCoord2f(0.0f, 1.0f); glVertex3f(-1.0f,  1.0f,  1.0f);  // Top Left Of The Texture and Quad
	// Back Face
	glNormal3f( 0.0f, 0.0f,-1.0f);    // Normal Facing Away
	glTexCoord2f(1.0f, 0.0f); glVertex3f(-1.0f, -1.0f, -1.0f);  // Bottom Right Of The Texture and Quad
	glTexCoord2f(1.0f, 1.0f); glVertex3f(-1.0f,  1.0f, -1.0f);  // Top Right Of The Texture and Quad
	glTexCoord2f(0.0f, 1.0f); glVertex3f( 1.0f,  1.0f, -1.0f);  // Top Left Of The Texture and Quad
	glTexCoord2f(0.0f, 0.0f); glVertex3f( 1.0f, -1.0f, -1.0f);  // Bottom Left Of The Texture and Quad
	// Top Face
	glNormal3f( 0.0f, 1.0f, 0.0f);    // Normal Facing Up
	glTexCoord2f(0.0f, 1.0f); glVertex3f(-1.0f,  1.0f, -1.0f);  // Top Left Of The Texture and Quad
	glTexCoord2f(0.0f, 0.0f); glVertex3f(-1.0f,  1.0f,  1.0f);  // Bottom Left Of The Texture and Quad
	glTexCoord2f(1.0f, 0.0f); glVertex3f( 1.0f,  1.0f,  1.0f);  // Bottom Right Of The Texture and Quad
	glTexCoord2f(1.0f, 1.0f); glVertex3f( 1.0f,  1.0f, -1.0f);  // Top Right Of The Texture and Quad
	// Bottom Face
	glNormal3f( 0.0f,-1.0f, 0.0f);    // Normal Facing Down
	glTexCoord2f(1.0f, 1.0f); glVertex3f(-1.0f, -1.0f, -1.0f);  // Top Right Of The Texture and Quad
	glTexCoord2f(0.0f, 1.0f); glVertex3f( 1.0f, -1.0f, -1.0f);  // Top Left Of The Texture and Quad
	glTexCoord2f(0.0f, 0.0f); glVertex3f( 1.0f, -1.0f,  1.0f);  // Bottom Left Of The Texture and Quad
	glTexCoord2f(1.0f, 0.0f); glVertex3f(-1.0f, -1.0f,  1.0f);  // Bottom Right Of The Texture and Quad
	// Right face
	glNormal3f( 1.0f, 0.0f, 0.0f);    // Normal Facing Right
	glTexCoord2f(1.0f, 0.0f); glVertex3f( 1.0f, -1.0f, -1.0f);  // Bottom Right Of The Texture and Quad
	glTexCoord2f(1.0f, 1.0f); glVertex3f( 1.0f,  1.0f, -1.0f);  // Top Right Of The Texture and Quad
	glTexCoord2f(0.0f, 1.0f); glVertex3f( 1.0f,  1.0f,  1.0f);  // Top Left Of The Texture and Quad
	glTexCoord2f(0.0f, 0.0f); glVertex3f( 1.0f, -1.0f,  1.0f);  // Bottom Left Of The Texture and Quad
	// Left Face
	glNormal3f(-1.0f, 0.0f, 0.0f);    // Normal Facing Left
	glTexCoord2f(0.0f, 0.0f); glVertex3f(-1.0f, -1.0f, -1.0f);  // Bottom Left Of The Texture and Quad
	glTexCoord2f(1.0f, 0.0f); glVertex3f(-1.0f, -1.0f,  1.0f);  // Bottom Right Of The Texture and Quad
	glTexCoord2f(1.0f, 1.0f); glVertex3f(-1.0f,  1.0f,  1.0f);  // Top Right Of The Texture and Quad
	glTexCoord2f(0.0f, 1.0f); glVertex3f(-1.0f,  1.0f, -1.0f);  // Top Left Of The Texture and Quad
  glEnd();          // Done Drawing Quads
}


GLvoid glDrawSphere(double radius, int slices, int stacks)          // Draw A Cube
{
	GLUquadric * pq = ::gluNewQuadric();
	::gluSphere(pq, radius, slices, stacks);
	::gluDeleteQuadric(pq);
}
} // namespace grutils
