/* MAINTAINER: Bernard */

#ifndef _LOG_H_
#define _LOG_H_

#include <string>
#include <map>
#include <ostream>
#include <fstream>
#include <iostream>
#include <sstream>
#include <cassert>
#include "conversions.h"

// will turn all trace
// You should write this to your own sources where you want
// to turn on logging. See doc for class Log
//#ifdef _DEBUG
//	#define WANT_TRACE
//#endif


/**
 * \brief Basic class for classes that want process log messages.
 */

class LogAcceptor
{
public:
	LogAcceptor() {};
	virtual ~LogAcceptor() {};

	/**
	 * \brief Forwarded input from Log
	 *
	 * DONT USE ANY LOGING IN THIS FUNCTION!
	 * \param t what was written ti log
	 */
	template<class T>  LogAcceptor& operator<< (const T& t)
	{
		mLine << t;
		// Here could be check for '\n' in mLine and process
		// the line
		return *this;
	}

	/**
	 * \brief Input for manipulators. 
	 * 
	 * Current status is that any call to this method
	 * will send current stream to logLine and empty the stream.
	 * DONT USE ANY LOGING IN THIS FUNCTION!
	 * \param op manipulator
	 */
	virtual LogAcceptor& operator<< (std::ostream& (*op)(std::ostream&));
	
	/**
	 * \brief Processing one line from log
	 *
	 * DONT USE ANY LOGING IN THIS FUNCTION!
	 * \param line one line
	 */
	 virtual void logLine(const std::string &line) = 0;
	
private:

	std::stringstream mLine;
};

// We need to explore what compiler support what

#ifndef  _MSC_VER
	#define __HAS_PRETTY_FUNCTION__
	#define __HAS_FUNCTION__
#else
	#define __HAS_FUNCTION__
#endif


using namespace std;

/// Loglevels
enum 
{ 
	LOG_LEVEL_DEVNULL = -1, LOG_LEVEL_TRACE = 0, LOG_LEVEL_DEBUG = 1, 
    LOG_LEVEL_INFO = 2, LOG_LEVEL_WARNING = 3, LOG_LEVEL_ERROR = 4
};


/**
 * \brief Class to handle loging
 *
 * Provides loging to file and loging to stderr.
 * If you want to use main log, use macros like LOGD, LOGE,...
 * You may use this class also to create special custom log.
 *
 * Using TRACE macros minihowto. First when you want to see trace
 * output in your log files you need to define WANT_TRACE. It is good
 * idea to define at the beginnning of each cpp file. It if then easy
 * to select what trace you want to see and you don't have to reedit
 * too many sources / you don't have too search too large logfile.
 * The macros: 
 *     TIN  you may write this macro at the beggining of function.
 *          It produce message that execution entered the function
 *			ald also produce message when execution leave the function
 *			(The leave part is automatic. You don't need to write
 *			anything else)
 *	   TVARIABLE(v)  print message to log that varibale v has value v.
 *			Both name and value are printed.
 *	   TPOINT Write mark to log that execution passed this point
 *     TPOINTS(s) Same as TPOINT but also add string message s
 *	   LOGT return stream with level set to TRACE. So you may use
 *          operator << and write to stream anything you want.
 *          When tracing is disabled LOGT return stream with log level
 *          set to devnull so no output will be produced.
 */
class Log
{

public:	

	Log();

	/**
	 * \brief Write message and kills program.
	 * 
	 * This call never return.
	 * \param s string to print
	 */
	void die(const char *s);

	/**
	 * \brief Kill the program.
	 *
	 * Write message die. This call never return.
	 */
	void die();

	/**
	 * Write message s and add explanation of current errno
	 * \param s message before explanation of errno
	 */
	void exit_failure(const char *s);
	
	/**
	 * Write message and explanation of errorcode.
	 * \param s          message before explanation of error code.
	 * \param errorcode  errorcode to explain
	 */
	void exit_failure(const char *s, int errorcode);
	
	/**
	 * Inicialize the logging system. Opens log file.
	 * Also connect to syslog.
	 * \param file name of file where to write. The program DIE if
	 *             the file can't be opened.
	 * \param tostderr if true, messages are send to stderr also
	 */
	void init(const char *file, bool tostderr);
	
	/**
	 * Write message with loglevel.
	 * '\n' will be added at the end of message.
	 * \param level loging level for this message
	 * \param s     text of message
	 */
	void message(int level, const char *s);

	/**
	 * Write message with loglevel. 
	 * '\n' will be added at the end of message.
	 * \param level loging level for this message
	 * \param s     text of message
	 */
	void message(int level, std::string &s);

	/**
	 * Write message with debug level.
	 * '\n' will be added at the end of message.
	 * \param s  text of message
	 */
	void debug(const char *s) { debug() << s << endl; }
	
	/**
	 * Write message with debug level.
	 * '\n' will be added at the end of message.
	 * \param s  text of message
	 */
	void debug(string &s) { debug() <<  s << endl; }
	
	
	//void debug(const char *s) { message(LOG_LEVEL_DEBUG, s); }
	//void debug(string &s) { message(LOG_LEVEL_DEBUG, s); }
	
	void set_loglevel(int level) { m_min_level = level; };

	/**
	 * Behaves same as perror function but the output
	 * is better stylised.
	 */
	void perror(const char *s);

	/**
	 * Operator for writing to log. You should use this operator
	 * if you need to write anything to logfile.
	 * \param t what you want to write
	 */
	template<class T> inline Log& operator<< (const T& t)
	{
		if (m_level < m_min_level) return *this;
		if (m_tostderr) cerr  << t;
		m_logfile << t;
#ifndef _DEBUG		
		if (m_level >= LOG_LEVEL_WARNING)
#endif			
			m_logfile.flush(); // this hurts performance but helps to write log
		
		if ( mLogAcceptor ) (*mLogAcceptor) << t;
		
		return *this;
	}

	/**
	 * Operator for manipulators - like flush, endl,..
	 * Also works with formating manipulators.
	 * \param op manipulator
	 */
	inline Log& operator<< (ostream& (*op)(ostream&))
	{
		if (m_level < m_min_level || m_level == LOG_LEVEL_DEVNULL) return *this;

		if (m_tostderr) op(cerr);
		op(m_logfile);
		
		if ( mLogAcceptor ) *mLogAcceptor << op;
		
		return *this;
	}
	
		
	Log& devnull() { m_level = LOG_LEVEL_DEVNULL;   return *this; }
	Log& trace()   { m_level = LOG_LEVEL_TRACE;     *this << "T "; return time(); }
	Log& debug()   { m_level = LOG_LEVEL_DEBUG;     *this << "D "; return time(); }
	Log& info()    { m_level = LOG_LEVEL_INFO;      *this << "I "; return time(); }
	Log& warning() { m_level = LOG_LEVEL_WARNING;   *this << "W "; return time(); }
	Log& error()   { m_level = LOG_LEVEL_ERROR;     *this << "E "; return time(); }

	/**
	 * \brief Set actual acceptor. May be 0 to set empty acceptor
	 *
	 * \param la new log acceptor
	 */
	void setAcceptor(LogAcceptor *la)
	{
		mLogAcceptor = la;
	}
	
private:	

	LogAcceptor       *mLogAcceptor; ///< may be NULL

	ofstream           m_logfile; ///< log file
	
	int m_level; ///< current log level

	int m_min_level; ///< minimal level to log

	bool m_tostderr; ///< if true, the loging goes to stderr also
	
	/**
	 * Write current time in the log file
	 */
	Log& time();

};


/**
 * Class for automatic TRACE_OUT
 * Remembers funcion name and print it in destructor
 */
class AutoTraceOut
{
public:	
	AutoTraceOut(Log &logClass, const char* functionName) :
		mLogClass(logClass)
	{
		mFunctionName = functionName;
	}
	~AutoTraceOut()
	{
		mLogClass.trace() << "TRACEOUT " << mFunctionName << endl;
	}
	
private:
	const char* mFunctionName;
	Log &mLogClass;
};


/**
 * Call this macro when you enter function that you want to trace.
 */
#ifdef __HAS_PRETTY_FUNCTION__

	#define TRACE_IN(logclass) \
		logclass.trace() << "TRACEIN " << __PRETTY_FUNCTION__ << endl;

	#define TRACE_OUT(logclass) \
		logclass.trace() << "TRACEOUT " << __PRETTY_FUNCTION__ << endl;

	#define TRACE_IN_AUTO_OUT(logclass) \
		TRACE_IN(logclass) \
		AutoTraceOut _temporary_trace_object(logclass, __PRETTY_FUNCTION__);

#elif defined __HAS_FUNCTION__

	#define TRACE_IN(logclass) \
		logclass.trace() << "TRACEIN " << __FUNCTION__ << endl;

	#define TRACE_OUT(logclass) \
		logclass.trace() << "TRACEOUT " << __FUNCTION__ << endl;

	#define TRACE_IN_AUTO_OUT(logclass) \
		TRACE_IN(logclass) \
		AutoTraceOut _temporary_trace_object(logclass, __FUNCTION__);


#else

	#define TRACE_IN(logclass) \
		logclass.trace() << "TRACEIN " << __FILE__ << ":" << __LINE__ << endl;

	#define TRACE_OUT(logclass) \
		logclass.trace() << "TRACEOUT " << __FILE__ << ":" << __LINE__ << endl;

	#define TRACE_IN_AUTO_OUT(logclass) \
		TRACE_IN(logclass) \
		string __temporary_trace_string s; \
		s.append(__FILE__).append(":").append(__LINE__); \
		AutoTraceOut _temporary_trace_object(logclass, s.c_str());

#endif


/**
 * Print "variable" = variable
 */
#define TRACE_VARIABLE1(logclass, variable) \
	logclass.trace() << "TRACEVARIABLE " << __FILE__ << ":" << __LINE__ \
                     << " " << #variable << " = " << variable << endl;

#define TRACE_VARIABLE2(logclass, v1, v2) \
	logclass.trace() << "TRACEVARIABLE " << __FILE__ << ":" << __LINE__ \
                     << " " << #v1 << "=" << v1 << " " << #v2 << "=" << v2 <<endl;

#define TRACE_VARIABLE3(logclass, v1, v2, v3) \
	logclass.trace() << "TRACEVARIABLE " << __FILE__ << ":" << __LINE__ \
                     << " " << #v1 << "=" << v1 << " " << #v2 << "=" << v2 \
					 << " " << #v3 << "=" << v3 << endl;

/**
 * Use this macro if you want to see if execution enter some place in code.
 */
#define TRACE_POINT(logclass) \
	logclass.trace() << "TRACEPOINT " << __FILE__ << ":" << __LINE__ << endl; 

#define TRACE_POINT_S(logclass, s) \
	logclass.trace() << "TRACEPOINT " << __FILE__ << ":" << __LINE__ << " " << s << endl; 


// Shortcuts for loging
extern Log g_log;         ///< MAIN LOG OBJECT
/*
extern Log g_log_bernard; ///< Logfile for developer 
extern Log g_log_martin;  ///< Logfile for developer 
extern Log g_log_michal;  ///< Logfile for developer 
extern Log g_log_petko;   ///< Logfile for developer 
extern Log g_log_lukas;   ///< Logfile for developer 
*/

/* ****************************** What you should use is HERE and down ********* */

#ifndef _DEBUG   // don't write any info from source if not in debug
		#define LOGN  g_log.devnull()
		#define LOGT_ g_log.trace()   
		#define LOGD  g_log.debug()   
		#define LOGI  g_log.info()    
		#define LOGW  g_log.warning() 
		#define LOGE  g_log.error()
#else
	#ifdef __HAS_FUNCTION__
		#define LOGN  g_log.devnull()
		#define LOGT_ g_log.trace()   << __FILE__ << ":" << __LINE__ << " " << __FUNCTION__ << " "
		#define LOGD  g_log.debug()   << __FILE__ << ":" << __LINE__ << " " << __FUNCTION__ << " "
		#define LOGI  g_log.info()    << __FILE__ << ":" << __LINE__ << " " << __FUNCTION__ << " "
		#define LOGW  g_log.warning() << __FILE__ << ":" << __LINE__ << " " << __FUNCTION__ << " "
		#define LOGE  g_log.error()   << __FILE__ << ":" << __LINE__ << " " << __FUNCTION__ << " "
	#else
		#define LOGN  g_log.devnull()
		#define LOGT_ g_log.trace()   << __FILE__ << ":" << __LINE__ << " " 
		#define LOGD  g_log.debug()   << __FILE__ << ":" << __LINE__ << " "
		#define LOGI  g_log.info()    << __FILE__ << ":" << __LINE__ << " "
		#define LOGW  g_log.warning() << __FILE__ << ":" << __LINE__ << " "
		#define LOGE  g_log.error()   << __FILE__ << ":" << __LINE__ << " "
	#endif
#endif

	// Trace
#ifdef WANT_TRACE

	#define TIN                  TRACE_IN_AUTO_OUT(g_log);
	#define TVARIABLE(v1)        TRACE_VARIABLE1(g_log, v1);
	#define TVARIABLE2(v1,v2)    TRACE_VARIABLE2(g_log, v1, v2);
	#define TVARIABLE3(v1,v2,v3) TRACE_VARIABLE3(g_log, v1, v2, v3);
	#define TPOINT               TRACE_POINT(g_log);
	#define TPOINTS(s)           TRACE_POINT_S(g_log, s);
	#define LOGT                 LOGT_

#else

	#define TIN             
	#define TVARIABLE(v1)  
	#define TVARIABLE2(v1, v2)  
	#define TVARIABLE3(v1, v2, v3)  
	#define TPOINT          
	#define TPOINTS(s)      
	#define LOGT            LOGN

#endif


/**
 * \brief Program killing function
 *
 * The program will be killed even if logging is disabled.
 * exit() function is called inside. So this call never return.
 * \param msg The last thing you want to say before you DIE :-)
 */

inline void DIE(const char *msg)
{
	g_log.die(msg);
}

/**
 * \brief Program killing function.
 *
 * The program will be killed. If you have logging some default
 * message will apear. exit() function is called inside. So this
 * call never return.
 */
inline void DIE()
{
	g_log.die();
}

/***************** assert ****************/
#ifdef _DEBUG
	#define assertL(condition) \
		if (!(condition)) \
		{ \
			LOGE << __FILE__ << ":" << __LINE__ << ": failed assertion " << #condition << endl; \
			assert(condition); /*MSCV show special dialog */ \
			DIE(); \
		}
#else
	#define assertL(condition) 
#endif

#endif // #ifndef _LOG_H_
