/* MAINTAINER: Martin */
#ifndef _GEOMETRY_OBJECT_H
#define _GEOMETRY_OBJECT_H

#include "refcounter.h"

namespace gobj
{
	class GraphicalRepresentation;

	/**
	*	\brief base class of geometry objects (TriMesh, Box, Sphere)
	*
	*	It is an ref-counted object that deletes itself when last reference
	*	that refers to it is destroyed.
	*/
	class GeometryObject: public RefCounter
	{
	protected:
		GeometryObject(): mBBComputed( false ), mGraRep( 0 ) {}
	public:
		
		struct BoundingBox {
			double center[3];
			double extents[3];
		};

		virtual ~GeometryObject(){}

		/**
		*	\brief accepts the incoming visitor
		*	\param vis visitor to accept
		*/
		virtual void accept( class Visitor * vis ) const {};
		
		virtual void computeBoundingBox( BoundingBox & bb ) const {};

		/** 
		*	\brief obtains object bounding box, computes it if necessary
		*	\param bb where to store object bounding box
		*/
		void getBoundingBox( BoundingBox &bb )
		{
			if ( !mBBComputed )
			{
				computeBoundingBox( mBB );
				mBBComputed = true;
			}			
			bb = mBB;
		}

		/**
		*	\brief get graphical representation of this geometry object
		*	\return pointer to graphical representation
		*/
		const GraphicalRepresentation * getGraphicalRepresentation() const
		{
			return mGraRep;
		}

		
		/**
		*	\brief assign a graphical representation to this geometry object
		*	This instance takes ownership over rep.
		*/
		void setGraphicalRepresentation( GraphicalRepresentation * rep )
		{
			mGraRep = rep;
		}

	protected:
		
		/**
		*	\brief occurs when reference count reaches zero
		*	
		*	On client, graphical representation is destroyed and then GeometryObject::OnZeroRefCount() is 
		*	called which deletes instance of this GeometryObject. On server, graphical representation is not 
		*	used.
		*/
		virtual void OnZeroRefCount();
	
	private:
		BoundingBox mBB;
		bool mBBComputed;
		
		/**
		*	\brief graphical representation of geometry object (on client only) - Display List
		*/
		GraphicalRepresentation * mGraRep;
	};
	
	/** \brief reference to geometry object */
	typedef RefCounterPtr<GeometryObject> GeometryObjectRef;

	//typedef ConstRefCounterPtr<GeometryObject> GeometryObjectConstRef;
}

#endif //_GEOMETRY_OBJECT_H
