/* MAINTAINER: Martin */
#ifndef _GROUP_NODE_VISITOR_H
#define _GROUP_NODE_VISITOR_H

#include "hierarchynodevisitor.h"

/**
	\brief visitor for group nodes

	By implementing visitGroupHead() and/or visitGroupMember in a derived class, 
	you will be able to traverse groups in the hierarchy tree.

	\code
	// ----------------------
	// -- A simple example --
	// ----------------------

	// Given a visibility flag, enables/disables rendering of group and its members
	// including nested groups and their members.
	class SetRenderable: public GroupNodeVisitor
	{
	public:
		SetRenderable( bool visibility ): mVisibility( visibility )
		{}

		void forEveryGroupMember( HierarchyNode * node )
		{
			node->setRenderable( mVisibility );
			visitDirectGroupMembers( node );
		}

		virtual void visitGroupHead( HierarchyNode * node ) { forEveryGroupMember( node ); }
		virtual void visitGroupMember( HierarchyNode * node ) { forEveryGroupMember( node ); }
	private:
		bool mVisibility;
	};

	// ...

	// Usage example:
	
	HierarchyNode *hn;
	// get hn from somewhere, somehow
	// ...

	// Create the visitor, that sets the renderable-flag of group and all group 
	// members to false and use it ( works only if hn is a group head ).
	SetRenderable v(false);
	hn->accept( &v );

	\endcode

	\code
	// --------------------------------------------------------
	// -- A "template", how to create your own group visitor --
	// --------------------------------------------------------

	class MyGroupVisitor: public GroupNodeVisitor
	{
	public:
		virtual void visitGroupHead( class HierarchyNode * node )
		{
			// Do some work on group head:
			// ...
			
			// Visit propagation:
			// A) Visit group children if you want to by calling visitDirectGroupMembers();
			// B) terminate visit
		}

		virtual void visitGroupMember( class HierarchyNode * node )
		{
			// Do some work on group member:
			// ...
			
			// Visit propagation:
			
			if ( !isNestedGroup(node) )
			// 1) ordinary group member node
			{
				// A) Visit group children if you want to:
				//		visitDirectGroupMembers();
				// B) terminate visit
			}
			else
			// 2) nested group node
			{
				// A) Visit nested group members if you want to:
				//		visitDirectGroupMembers();
				// B) create a new group visitor and visit group children
				//		GroupNodeVisitor gv2;
				//		node->accept(&gv2);
				// C) terminate visit
			}
		}
	}
	\endcode
	
	\code
	// -------------------------------
	// -- An example, how to use it --
	// -------------------------------

	HierarchyNode * hn;	
	// get 'hn' from somewhere, somehow
	// ...
	
	// create your group visitor
	MyGroupVisitor gv;

	// visit group node ( applies to hn only if it is a group head )
	hn->accept(&gv);
	
	\endcode
*/

class GroupNodeVisitor: public HierarchyNodeVisitor
{
private:

	void visitHierarchyNode(class HierarchyNode * node);

public:

	virtual void visitGroupHead( class HierarchyNode * node ){}
	virtual void visitGroupMember( class HierarchyNode * node ){}

	bool isNestedGroup( class HierarchyNode * node );

protected:

	void visitDirectGroupMembers( class HierarchyNode * node );

};

#endif //_GROUP_NODE_VISITOR_H
