/* MAINTAINER: Martin */

#include "hierarchy_node.h"
#include "groupnodevisitor.h"
#include "log.h"

HierarchyNode::HierarchyNode(): 
	mGroup(0), mParent(0), 
	mRenderableNode(true), mCanDelete(true), 
	mName(""), mGeometryObject(0), mGroupHead(false), mGroupMember(false)
{}

HierarchyNode::HierarchyNode(const HierarchyNode & node):
	mGroup(node.mGroup), mParent(node.mParent), 
	mRenderableNode(node.mRenderableNode), mCanDelete(false),
	mChildren(node.mChildren), mAttributes(node.mAttributes),
	mAttrTransformation( node.mAttrTransformation ), mAttrMaterial( node.mAttrMaterial ),
	mName(node.mName), mGeometryObject(node.mGeometryObject), mGroupHead( node.isGroupHead() ), mGroupMember( node.isGroupMember() )	
{}

void HierarchyNode::setTransformation( counted_ptr< hna::Attribute > transformation )
{
	mAttrTransformation = transformation;
}

void HierarchyNode::setMaterial( counted_ptr< hna::Attribute > material )
{
	mAttrMaterial = material;
}

HierarchyNode::~HierarchyNode()
{
	// if children and attributes are already cloned, delete them
	if (mCanDelete) 
	{
		// delete children
		for (ChildIterator it = getChildrenBegin(), ite = getChildrenEnd(); it != ite; ++it)
		{
			delete *it;
			*it = 0;
		}
		mChildren.clear();

		// delete attributes
		// - automatic when using counted_ptr
		/*
		for (AttributeIterator it = mAttributes.begin(), ite = mAttributes.end(); it != ite; ++it)
		{
			delete *it;
			*it = 0;
		}
		mAttributes.clear();
		*/
		mCanDelete = false;
	} else {
		// do not delete children or attributes if they are not yet cloned - log warning instead
		LOGW << "Attempting to delete children and attributes that doesn't belong to this instance of HierarchyNode!" << std::endl;
	}
	
}

void HierarchyNode::addChild( HierarchyNode * node )
{	
	node->setParent(this);
	mChildren.push_back(node);
}

void HierarchyNode::addAttribute( counted_ptr<hna::Attribute> attribute )
{
	mAttributes.push_back( attribute );
}

bool HierarchyNode::checkChildrenInvariant(void)
{
	for (ChildIterator it = getChildrenBegin(), ite = getChildrenEnd(); it != ite; ++it)
	{
		// neni member a mel by ukazatel na nejakou skupinu? No toto?!
		if (!(*it)->isGroupMember() && ((*it)->getGroup() != 0))
			return false;
		
		if (this->isGroupHead())
		// sme v uzlu definujicim skupinu. 
		{
			// a jeho deti, ktere jsou cleny, ze by na nej neukazovaly? No toto?!
			if ((*it)->isGroupMember() && (*it)->getGroup() != this )
			{
				return false;
			}
		}
		else if (this->isGroupMember())
		// jsme v uzlu, ktery neni skupina, ale je soucasti jine skupiny
		{
			// a jeho deti, ktere jsou cleny te same skupiny, ze by ukazovaly na jinou skupinu? No toto?!
			if ((*it)->isGroupMember() && (*it)->getGroup() != this->getGroup() )
			{
				return false;
			}		
		}else
		// jsme v uzlu, ktery neni v zadne skupine, ani hlavou zadne skupiny
		{
			// takze jeho deti nesmi byt ani cleny zadne skupiny, ani mit pointer na nejakou skupinu. To by jako neslo.
			if ( (*it)->isGroupMember() || (*it)->getGroup() != 0 )
				return false;		
		}
	}
	return true;
}

void HierarchyNode::assertChildrensInvariant(void)
{
	assertL(
		checkChildrenInvariant() &&
		"Children invariant is broken!"
	);
}

void HierarchyNode::setRenderable( bool enable )
{
	
	mRenderableNode = enable;
	
	if ( isGroupHead() )
	// TODO: affect group children
	{
		class SetRenderable: public GroupNodeVisitor
		{
		public:
			SetRenderable( bool visibility ): mVisibility( visibility )
			{}

			void forEveryGroupMember( HierarchyNode * node )
			{ 
				node->setRenderable( mVisibility );
				visitDirectGroupMembers( node );
			}

			virtual void visitGroupHead( HierarchyNode * node ) { visitDirectGroupMembers( node ); }
			virtual void visitGroupMember( HierarchyNode * node ) { forEveryGroupMember( node ); }
		private:
			bool mVisibility;
		} myVisitor( enable );
	
		accept(&myVisitor);
	}
}

void HierarchyNode::cloneChildren(HierarchyNode * group)
{ 	
	for (ChildIterator it = getChildrenBegin(), ite = getChildrenEnd(); it != ite; ++it)
	{
		// non-null hierarchy child's group pointer must be equal to group pointer of this node
		*it = (*it)->clone(((*it)->getGroup() == 0) ? 0 : group);
		(*it)->setParent(this);
	}
}

void HierarchyNode::cloneAttributes(void)
{ 	
	if ( mAttrTransformation.get() != 0 && !mAttrTransformation->isShared() )
	{
		mAttrTransformation = counted_ptr<hna::Attribute>( mAttrTransformation->clone() );
	}
	
	if ( mAttrMaterial.get() != 0 && !mAttrMaterial->isShared() )
		mAttrMaterial = counted_ptr<hna::Attribute>( mAttrMaterial->clone() );

	for (AttributeIterator it = getAttributesBegin(), ite = getAttributesEnd(); it != ite; ++it)
		if ( !(*it)->isShared() )
		{
			counted_ptr<hna::Attribute> clonedAttribute( (*it)->clone() );
			(*it) = clonedAttribute;
		}
}

void HierarchyNode::cloneCode(HierarchyNode * group)
{ 
	cloneChildren(group);
	cloneAttributes();
	mCanDelete = true;
}

HierarchyNode* HierarchyNode::clone(HierarchyNode * group)
{
	this->assertChildrensInvariant();	
	HierarchyNode * new_node = new HierarchyNode(*this); // TODO: check NULL
	new_node->cloneCode(group);
	return new_node;
}

void HierarchyNode::accept(HierarchyNodeVisitor * vis)
{
	// perform visitor's action on node
	vis->visitHierarchyNode(this);
}

void HierarchyNode::acceptAttributeVisitor(hna::AttrVisitor * vis)
{
	if (mAttrTransformation.get() != 0 )
		mAttrTransformation->accept( vis );

	if ( mAttrMaterial.get() != 0 )
		mAttrMaterial->accept( vis );

	for (AttributeIterator it = mAttributes.begin(), ite = mAttributes.end(); it != ite; ++it)
		(*it)->accept(vis);
}

