/* MAINTAINER: Martin */
#ifndef _HIERARCHY_NODE_H_
#define _HIERARCHY_NODE_H_

#include <vector>
#include <string>
#include <list>

#include "hna.h"
#include "hierarchynodevisitor.h"
#include "hna_visitor.h"
#include "gobj.h"
#include "counted_ptr.h"

/**
*	\brief node of model hierarchic tree - base class
*
*	Maintains children nodes, group pointers, attributes.
*	Provides methods for cloning, applying attributes and structure traversing (using visitors).
*/
class HierarchyNode
{
public:
	friend class ModelLoader;

	typedef std::vector<HierarchyNode*> ChildrenContainer;		///< \brief container type for children
	typedef std::list< counted_ptr<hna::Attribute> > AttributesContainer;	///< \brief container type for attributes
	typedef ChildrenContainer::iterator ChildIterator;			///< \brief iterator type for children
	typedef AttributesContainer::iterator AttributeIterator;	///< \brief iterator type for attributes

	/**
	*	\brief ctor
	*	Simple initialization - pointers to null, flags to true.
	*/
	HierarchyNode();

	/**
	*	\brief dtor
	*	
	*	Destroys all children of this node and all attributes of this node 
	*	(that means iff mCanDelete is set to true). 
	*	Children are destroyed recursively. 
	*/
	virtual ~HierarchyNode();

	// TODO: iterators - access rights (maybe this should be const-iterator...) - maybe not (considered RenderVisitor)
	
	/** \brief child iterator - begin */
	ChildIterator getChildrenBegin(void) { return mChildren.begin(); }
	
	/** \brief child iterator - end */
	ChildIterator getChildrenEnd(void)   { return mChildren.end(); }
	
	/** \brief attribute iterator - begin */
	AttributeIterator getAttributesBegin(void) { return mAttributes.begin(); }
	
	/** \brief attribute iterator - end */
	AttributeIterator getAttributesEnd(void)   { return mAttributes.end(); }

	/** \brief parent of this node */
	HierarchyNode * getParent(void) { return mParent; }
	
	/**	\brief group of this node */
	HierarchyNode * getGroup(void) { return mGroup; }

	/**
	*	\brief adds a child subtree
	*	\param node root of the subtree
	*/
	void addChild(HierarchyNode * node);
	
	/**
	*	\brief adds an attribute
	*	\param attribute the new attribute
	*	\note this node takes ownership over the attribute
	*/
	void addAttribute( counted_ptr<hna::Attribute> attribute );
	
	/**
	*	\brief clones this node as a root (its parent and group will be \tt null )
	*	\return cloned node with cloned subtrees
	*/
	HierarchyNode * clone(void) 
	{ 
		HierarchyNode *clonedNode;
		clonedNode = clone(0);
		clonedNode->setParent(0);
		clonedNode->mGroup = 0;
		return clonedNode;
	}

	/**
	*	\brief accepts the node visitor \tt v
	*
	*	The right method of node visitor will be called in subclass, that will override
	*	accept(). By implementing visitor interface you can provide your own visitor code.
	*	
	*	\param vis visitor to accept
	*/
	virtual void accept( HierarchyNodeVisitor * vis );

	/** \brief accepts attribute visitor on each of node attributes 
	*	\param vis visitor to accept
	*/
	void acceptAttributeVisitor( hna::AttrVisitor * vis );

	/**
	*	\brief gets the geometry object associated with this node
	*	\return geometry object of this model-node
	*/
	gobj::GeometryObjectRef getGeometryObject(void)
	{ return mGeometryObject; }

	/** \brief sets node name */
	void setName(const std::string & name) { mName = name; }

	/** \brief gets node name */
	const std::string & getName() { return mName; }

	/**
	*	\brief if this node should be rendered or not
	*/
	bool isRenderable(void) const { return mRenderableNode; }

	/**
	*	\brief enable or disable node rendering
	*
	*	Change performed to group node changes all children as well.
	*/
	void setRenderable( bool enable = true );

	
	bool isGroupHead() const { return mGroupHead; }
	bool isGroupMember() const { return mGroupMember; }

	void setTransformation( counted_ptr< hna::Attribute > transformation );
	void setMaterial( counted_ptr< hna::Attribute > material );

	counted_ptr< hna::Attribute > getTransformation() { return mAttrTransformation; }
	counted_ptr< hna::Attribute > getMaterial() { return mAttrMaterial; }

protected:
	/**
	*	\brief copy ctor, except the \tt mCanDelete flag setting to false
	*
	*	Copy ctor will create new node. Its children and attributes are shared with
	*	those of \tt node. This is not very consistent state until they are cloned.
	*	Setting \tt mCanDelete to \c false ensures that dtor will not destroy children
	*	and attributes that are not its own.
	*/
	HierarchyNode(const HierarchyNode & node);

	/**	\brief sets the parent */
	void setParent(HierarchyNode * parent) { mParent = parent; }
	
	/**
	*	\brief static method for setting the parent of a node
	*
	*	Protected non-static method of HierarchyNode can't be called in its
	*	subclasses directly on HierarchyNode pointer. 
	*	This static method solves that problem.
	*
	*	\param node concrete instance of \tt node
	*	\param paren its parent to set
	*/
	static void setParent(HierarchyNode * node, HierarchyNode * parent)
	{ node->setParent(parent); }

	/**
	*	\brief static method for cloning a node
	*
	*	Protected non-static method of HierarchyNode can't be called in its
	*	subclasses directly on HierarchyNode pointer. 
	*	This static method solves that problem.
	*
	*	\param node concrete instance of node
	*	\param group new group for cloned children (pointer to already cloned hierarchy node)
	*	\return cloned node with cloned subtrees
	*/
	static HierarchyNode * clone(HierarchyNode * node, HierarchyNode * group)
	{ return node->clone(group); }
	
	/**
	*	\brief clones this node - subclasses shall override it completely
	*	\param group new group for cloned children (pointer to already cloned hierarchy node)
	*	\return cloned node with cloned subtrees
	*/
	virtual HierarchyNode * clone(HierarchyNode * group);
	
	/**
	*	\brief clones children and attributes
	*	\param group new group for cloned children (pointer to already cloned hierarchy node)
	*/
	void cloneCode(HierarchyNode * group);
	
	/**
	*	\brief tests children invariant, write assert message and exits program when broken
	*/
	void assertChildrensInvariant(void);

	/**
	*	\brief associates a geometry object with this model-node
	*	\param geometryObject counted reference to the geometry object
	*/
	void setGeometryObject(gobj::GeometryObjectRef geometryObject)
	{
		mGeometryObject = geometryObject;
	}

private:
	/**
	*	\brief clones all node children - recursively
	*
	*	Recursively clones node children and sets their parent to \tt this.
	*	If some child belongs to a group, its group will be set to \tt group.
	*
	*	\param group new group for cloned children (pointer to already cloned hierarchy node)
	*/
	void cloneChildren(HierarchyNode * group);
	
	/** \brief clones all node attributes */
	void cloneAttributes(void);
	
	/**
	*	\brief test children group invariant and returns false iff broken
	*
	*
	*	\return false if the invariant is broken, otherwise true
	*/
	bool checkChildrenInvariant(void);

	/** \brief group node this node belongs to (or null if doesn't belong to any group) */
	HierarchyNode * mGroup;
	
	/** parent node (or null if this node has no parent) */
	HierarchyNode * mParent;
	
	// TODO: rendering was taken outside this class, so those flags shouldn't be here
	//	- minimally they should be renamed
	//  - couldn't they be some attributes?

	bool mRenderableNode;	///< \brief renderable node or not 
	
	/** 
	*	\brief whether the destructor is allowed to delete children and attributes or not (safety reasons) 
	*	
	*	Ensures that dtor will not delete children nor attributes that this instance doesn't own. 
	*	This flag is set to false only in copy ctor (before the cloning code is performed) and set to true 
	*	right after cloning complete.
	*/
	bool mCanDelete;
	
	ChildrenContainer mChildren;		///< \brief children nodes
	AttributesContainer mAttributes;	///< \brief attributes list

	counted_ptr< hna::Attribute > mAttrTransformation;
	counted_ptr< hna::Attribute > mAttrMaterial;
		
	// TODO: replace string with some ID when needed
	/** \brief node name */
	std::string mName;

	/**
	*	\brief counted reference to geometry object associated with this model-node
	*/
	gobj::GeometryObjectRef mGeometryObject; 

	bool mGroupHead;
	bool mGroupMember;
};


#endif //_HIERARCHY_NODE_H_
