/* MAINTAINER: Martin */
#ifndef _ATTR_TRANSFORM_H_
#define _ATTR_TRANSFORM_H_

#include "hna.h"
#include "GLMatrix.h"

namespace hna 
{
	/**
	*	\brief geometry transformation
	*
	*	Transformation attribute consists of two matrices - ObjectOffsetTM and NodeTM.
	*	NodeTM transforms parent node coordinate system to current node coordinate system and 
	*	propagates itself down the	hierarchy. ObjectOffsetTM is transformation in current node
	*	coordinate system and do not propagates itself down the hierarchy. By multiplying NodeTM
	*	matrices on path from root to current node, transformation matrix to object space of this
	*	node is obtained. Further, by multiplying this by ObjectOffsetTM, the transformation
	*	to world space is obtained.
	*/
	class Transformation: public Attribute
	{
	public:		

		/** \brief both matrices are set to identity */
		Transformation(void);
		
		/** \brief copy ctor */
		Transformation(const Transformation &t);

		virtual ~Transformation();
				
		
		// TODO: implement applyEffect, removeHierarchyEffect, removeTemporaryEffect in visitor

		/** \brief temporary - performs OpenGL transform on client, nothing on server */
		virtual void applyEffect(void);

		/** \brief temporary - removes OpenGL transform (NodeTM) on client, nothing on server */
		virtual void removeHierarchyEffect(void);		

		/** \brief temporary - remopves OpenGL transform (ObjectOffsetTM) on client, nothing on server */
		virtual void removeTemporaryEffect(void);

		/** \brief clones this attribute */
		virtual Attribute* clone(void);
		
		/**
		*	\brief sets node transform matrix
		*	\param tm node transform matrix
		*/
		void setNodeTM(const GLMatrix4d & tm)
		{ 
			mNodeTM = tm; 
			mMirrorNTM = isMirrorMatrix( mNodeTM ); 
		}
		
		/**
		*	\brief sets object offset transform matrix
		*	\param tm object offset transform matrix
		*/
		void setObjectOffsetTM(const GLMatrix4d & tm)
		{ 
			mObjectOffsetTM = tm;
			mMirrorOTM = isMirrorMatrix( mObjectOffsetTM );
		}

		
		/** \brief gets node transform matrix */
		const GLMatrix4d& getNodeTM(void) const
		{ return mNodeTM; }

		/** \brief gets object offset transform matrix */
		const GLMatrix4d& getObjectOffsetTM(void) const
		{ return mObjectOffsetTM; }

		bool isMirrorNodeTM() const
		{ return mMirrorNTM; }

		bool isMirrorObjectOffsetTM() const
		{ return mMirrorOTM; }

		/** 
		*	\brief accepts attribute visitor 
		*	\param vis visitor to accept
		*/
		virtual void accept(class AttrVisitor * vis);

	private:
		
		/**
		*	\brief test for matrix that changes orientation of faces
		*	\param m matrix to test
		*	\return true if matrix changes orientation, false if it doesn't
		*/
		static bool isMirrorMatrix( const GLMatrix4d & m );

		GLMatrix4d 
			mNodeTM,			///< \brief transformation from parent to this node, inherited by hierarchy children
			mObjectOffsetTM;	///< \brief relative transformation in this object space, not inherited by hierarchy children

		bool mMirrorNTM, mMirrorOTM;

		// parentTM = parent.nodeTM * parent.offsetTM
		// toWorldSpaceTM = parentTM * this.mNodeTM * this.mObjectOffsetTM
	};
	
	// Note: DO NOT try to estabilish links between transformation attributes 
	// (since it's hard to clone them and hard to maintain consistency)
}


#endif //_ATTR_TRANSFORM_H_
