/* MAINTAINER: Martin */
#include "model.h"

#include "hierarchy_node.h"
#include "hierarchynodevisitor.h"
#include "transform_obtainer.h"


Model::Model() 
	: mPrototype(0), mRoot(0)
{
	mBoundingSphere.center.set(0, 0, 0);
	mBoundingSphere.radius = 0;
	mBoundingSphere.valid = false;
}


Model::Model(const Model & model)
{
	// clone the hierarchy tree
	setRoot(model.mRoot->clone());
	
	mPrototype = model.mPrototype;
	mBoundingSphere = model.mBoundingSphere;
}

Model::~Model()
{
	// delete nodes in the hierarchy tree
	if (mRoot)
	{
		delete mRoot;
		mRoot = 0;
	}
	
	// delete all animation
	clearAnimation();
	mPrototype = 0;
}

void Model::getBoundingSphere( GLVector3d &center, double &radius )
{
	if ( !mBoundingSphere.valid ) 
	{
		
		// visitor that calculates bounding sphere of goven model
		class BoundingSphereBuilder: public HierarchyNodeVisitor
		{
		public:
			BoundingSphereBuilder()
			: mTM( GLMatrix4d::identity() ), mSphereCenter(0, 0, 0), mSphereSqrRadius(0)
			{}

			void visitHierarchyNode( HierarchyNode * node )
			{
				TransformObtainer to;
				node->acceptAttributeVisitor( &to );
				hna::Transformation * t = to.get();
				
				GLMatrix4d tmp, m;
				if ( t )
				{
					// preserve node transformation matrix
					tmp = mTM;
					
					// update node transformation matrix and calculate to-world transformation
					mTM *= t->getNodeTM();
					m = mTM * t->getObjectOffsetTM();
				}else{
					m = mTM;
				}

				// get geometry object
				gobj::GeometryObject * gobj = node->getGeometryObject();
				if ( gobj )
				{
					// get its bounding box
					gobj::GeometryObject::BoundingBox bbox;
					gobj->getBoundingBox( bbox );

					// calculate bounding box vertices
					GLVector3d vertex[8] = 
					{
						GLVector3d( bbox.center[0] + bbox.extents[0], bbox.center[1] + bbox.extents[1], bbox.center[2] + bbox.extents[2] ),
						GLVector3d( bbox.center[0] + bbox.extents[0], bbox.center[1] + bbox.extents[1], bbox.center[2] - bbox.extents[2] ),
						GLVector3d( bbox.center[0] + bbox.extents[0], bbox.center[1] - bbox.extents[1], bbox.center[2] + bbox.extents[2] ),
						GLVector3d( bbox.center[0] + bbox.extents[0], bbox.center[1] - bbox.extents[1], bbox.center[2] - bbox.extents[2] ),
						GLVector3d( bbox.center[0] - bbox.extents[0], bbox.center[1] + bbox.extents[1], bbox.center[2] + bbox.extents[2] ),
						GLVector3d( bbox.center[0] - bbox.extents[0], bbox.center[1] + bbox.extents[1], bbox.center[2] - bbox.extents[2] ),
						GLVector3d( bbox.center[0] - bbox.extents[0], bbox.center[1] - bbox.extents[1], bbox.center[2] + bbox.extents[2] ),
						GLVector3d( bbox.center[0] - bbox.extents[0], bbox.center[1] - bbox.extents[1], bbox.center[2] - bbox.extents[2] )
					};
					
					// transform bounding box vertices to world space and possibly enlarge bounding sphere 
					// to contain all these 8 vertices
					for ( int i = 0; i < 8; ++i ) 
					{
						GLVector3d tmp( m.dot3( vertex[i] ) - mSphereCenter );
						double d( tmp.lengthSqr() );
						if ( d > mSphereSqrRadius )
							mSphereSqrRadius = d;
					}
				}

				visitHierarchyChildren( node );
				
				if ( t )
				{
					// restore node transformation matrix
					mTM = tmp;
				}
			}

			void setSphere( const GLVector3d &center, double radius )
			{
				mSphereCenter = center;
				mSphereSqrRadius = radius * radius;
			}
			
			void getSphere( GLVector3d &center, double &radius ) const
			{
				center = mSphereCenter;
				radius = mSphereSqrRadius > 0 ? sqrt( mSphereSqrRadius ) : 0;
			}

		private:
			GLMatrix4d mTM;

			GLVector3d	mSphereCenter;
			double		mSphereSqrRadius;
		
		} sphereBuilder;

		HierarchyNode * root = getRoot();
		if ( root )
		{
			sphereBuilder.setSphere( GLVector3d(0, 0, 0), 0 );
			root->accept( &sphereBuilder );
			sphereBuilder.getSphere( mBoundingSphere.center, mBoundingSphere.radius );

			// set valid
			mBoundingSphere.valid = true;
		}
	}

	if ( mBoundingSphere.valid )
	{
		// return bounding sphere
		center = mBoundingSphere.center;
		radius = mBoundingSphere.radius;
	}
}
