/* MAINTAINER: Martin */
#ifndef _MODEL_H_
#define _MODEL_H_

#include "hierarchy_node.h"
#include <list>
#include <string>

class Animation;

/**
*	\brief public interface for accessing a 3D model
*
*	Model consists of
*	<ul>
*	<li>tree of nodes (model nodes with geometry objects or group nodes)</li>
*	<li>structure for quick access to nodes [not implemented yet]</li>
*	<li>container for animation controllers [not implemented yet]</li>
*	</ul>
*/
class Model 
{
	friend class ModelLoader;

public:
	Model();
	
	/**
	*	\brief dtor
	*	
	*	Deletes the hierarchy tree. 
	*	
	*	\note Inside this dtor it is possible that the hierarchic tree of 
	*		model prototype doesn't exist yet (due to destroyed counted 
	*		reference in ModelInstance subclass).
	*/
	virtual ~Model();

protected:
	
	/**
	*	\brief sets the model prototype pointer
	*
	*	\param prototype model protype
	*
	*	\note This method is used from ModelInstance subclass to set up the model prototype
	*	pointer.	
	*/
	void setPrototype(Model * prototype)
	{ mPrototype = prototype; }

	/**
	*	\brief copy ctor - clones hierarchy structure of the \c model.
	*	\note animation is not cloned nor copied yet.
	*/
	Model(const Model & model);
	
	/**
	*	\brief sets the root of hierarchy tree
	*	\param hn root node
	*	\brief this instance takes the ownership over the hierarchy tree
	*/
	void setRoot(HierarchyNode * hn) { mRoot = hn; }

public:
	//TODO: access rights (this should not be public)
	/**
	*	\brief access to hierarchic tree
	*	\return root node of hierarchic tree (or null if is not set)
	*/
	HierarchyNode * getRoot() { return mRoot; }

	/**
	*	\note returned animation is owned by model
	*/
	// TODO: design & safety: consider returning some kind of handler instead of " animation * "
	Animation * createAnimation( const std::string & animationName );

	void updateAnimation( unsigned int time_ms );

	/** 
	*	\brief clears animation owned by this instance of model
	*/
	void clearAnimation();

	void getBoundingSphere( GLVector3d &center, double &radius );

private:
	typedef std::list< Animation * > anim_container_type;

	void getAllAnimation( anim_container_type::iterator & first, anim_container_type::iterator & end )
	{
		first = mAnimationList.begin();
		end = mAnimationList.end();
	}

	Model * mPrototype;		///< \brief model prototype pointer for quick prototype access from its instances
	HierarchyNode * mRoot;	///< \brief root of hierarchic tree

	std::list< Animation * > mAnimationList;

	struct BoundingSphere 
	{
		GLVector3d center;
		double radius;
		bool valid;
	} mBoundingSphere;
		
	// QuickNodeAccess mNodeAccess;
	// Animation mAnimations;
};

#endif //_MODEL_H_
