#include "model_loader.h"
#include "tinyxml/tinyxml.h"
#include "conversions.h"
#include "gobj_all.h"
#include "hierarchy_node.h"
#include "hna_transform.h"
#include "GLMatrix.h"
#include <memory>

// todo - error handling and memory dealocations on error

void ModelLoader::resetState() {
	mData.gobjId = -1;
	mData.currentNode = 0;	
	
	for (int i = 0; (unsigned int)i < sizeof(mData.matrix) / sizeof(*mData.matrix); ++i)
		mData.matrix[i] = 0;

	mIDtoMaterialMap.clear();
	mIdGobjMap.clear();
	mModelName.clear();
}

Model * ModelLoader::loadModel(const std::string &fileName)
{	
	// initialize context of the loader
	resetState();
	mModelName = fileName;

	TiXmlDocument doc( fileName );
	
	if ( !doc.LoadFile() )
	{
		LOGE << "Unable to load " << fileName 
		     <<" Probably file not exist or isn't a valid XML file" << endl;
		return 0;
	}
	
	// create a new, empty model
	mModel = std::auto_ptr<Model>( new Model() );
	
	TiXmlHandle handle( doc.FirstChildElement( "model" ) );

	onHeaderElement( handle.FirstChildElement( "header" ) );
	onGeometryElement( handle.FirstChildElement( "geometry" ) );
	onMaterialsElement( handle.FirstChildElement( "materials" ) );
	onHierarchyNodeElement( handle.FirstChildElement( "hierarchy" ).FirstChildElement( "node" ) );
	
	mModel->setRoot(mData.currentNode); // currentNode is now root node or null	
	mData.currentNode = 0;

	// create graphical representation and return resulting model on success
	return createGraphicalRepresentation() ? mModel.release() : 0;
}

void ModelLoader::onHeaderElement(const TiXmlHandle &handle)
{
	// none header yet
}

void ModelLoader::onGeometryElement(const TiXmlHandle &handle)
{
	// TODO: move this initialization into some init-method
	NodeHandlerByName c;
	c["box"] = &ModelLoader::onBoxElement;
	c["sphere"] = &ModelLoader::onSphereElement;
	c["trimesh"] = &ModelLoader::onTrimeshElement;

	int errorCode;

	// loop through gobj elements
	TiXmlElement * child = handle.FirstChild("gobj").Element(); 
	for ( ; child != 0; child = child->NextSiblingElement() )
	{
		int gobjId;

		errorCode = child->QueryIntAttribute("id", &gobjId);
		if ( errorCode != TIXML_SUCCESS )
		{
			// TODO: handle missing attribute error
			// errorCode - TIXML_WRONG_TYPE or TIXML_NO_ATTRIBUTE
			continue;
		}
		
		TiXmlElement * gobjType = child->FirstChildElement();
		if ( !gobjType )
		{
			// TODO: handle missing specific geometry object
			continue;
		}

		// LOGI << "gobjId = " << gobjId << std::endl;
		mData.gobjId = gobjId;

		// invoke handler for specific geometry object
		NodeHandlerByName::iterator it = c.find(gobjType->ValueStr());
		if ( it != c.end() )
		{
			(this->*(it->second))(TiXmlHandle( gobjType ));
		} 
		else 
		{
			// TODO: handle unknown geometry object element
			continue;
		}
	}
}

void ModelLoader::onBoxElement(const class TiXmlHandle &handle)
{
	double length = 1, width = 1, height = 1;
	
	struct { 
		std::string name; 
		double *value; 
	} a[] = { 
		{"length", &length}, 
		{"width",  &width}, 
		{"height", &height},
	};
	
	for (int i = 0; i < 3; ++i)
	{
		TiXmlText * tmp = handle.FirstChildElement(a[i].name).FirstChild().Text();
	
		if ( tmp == 0 )
		{
			// TODO: handle non-existing element
			continue;
		}
		if ( !try_get_double(tmp->ValueStr(), *a[i].value) )
		{
			// TODO: handle bad value
			continue;
		}
	}
	
	// process length, width, height
	insertGeometryObject( mData.gobjId, new gobj::Box( width, length, height ) );	
}

void ModelLoader::onSphereElement(const class TiXmlHandle &handle)
{
	double radius = 1;
	
	struct { 
		std::string name; 
		double *value; 
	} a[] = { 
		{"radius", &radius}, 
		//{"segments",  &segments}, 
	};
	
	for (int i = 0; i < 1; ++i)
	{
		TiXmlText * tmp = handle.FirstChildElement(a[i].name).FirstChild().Text();
	
		if ( tmp == 0 )
		{
			// TODO: handle non-existing element
			continue;
		}
		if ( !try_get_double(tmp->ValueStr(), *a[i].value) )
		{
			// TODO: handle bad value
			continue;
		}
	}
	
	// process radius
	insertGeometryObject( mData.gobjId, new gobj::Sphere( radius ) );
}

void ModelLoader::onHierarchyNodeElement(const TiXmlHandle &handle)
{
	TiXmlHandle node = handle;
	TiXmlElement * el = node.Element();

	bool groupHead, groupMember, renderable;
	{
		const char * groupHeadStr = el->Attribute("group_head");
		const char * groupMemberStr = el->Attribute("group_member");
		const char * renderableStr = el->Attribute("renderable");
		groupHead = ( ( groupHeadStr != 0 && std::string( groupHeadStr ) == "true" ) ? true : false );
		groupMember = ( ( groupMemberStr != 0 && std::string( groupMemberStr ) == "true" ) ? true : false ); 
		renderable = ( ( renderableStr != 0 && std::string( renderableStr ) == "false" ) ? false : true ); 
	}
	
	int gobjId, matID;
	bool hasGobjId( el->QueryIntAttribute( "gobj", &gobjId ) == TIXML_SUCCESS );
	bool hasMatID( el->QueryIntAttribute( "material", &matID ) == TIXML_SUCCESS );
		
	TiXmlText *t = node.FirstChildElement("name").FirstChild().Text();	
	std::string name((t != 0) ? t->ValueStr() : "");

	if (mData.currentNode == 0)
		mData.currentNode = new HierarchyNode();

	HierarchyNode * hn = mData.currentNode;

	// process hasGobjId, gobjID, name
	if ( hasGobjId )
	{
		gobj::GeometryObjectRef gobjRef = getGeometryObject( gobjId );
		if ( gobjRef.isValid() )
			hn->setGeometryObject( gobjRef );
	}

	hn->setName(name);

	if ( hasMatID )
	// resolve material reference
	{
		counted_ptr<hna::Attribute> mat = getMaterialFromMap(matID);
		if ( mat.get() != 0 )
		// found in mat, assign this material to a node
		{
			if ( !mat->isShared() )
			// this material is not shared, clone it
			{
				counted_ptr<hna::Attribute> tmp(mat->clone());
				mat = tmp;
			}

			hn->setMaterial( mat );
		}
	}

	// ----------------------------
	// process renderable attribute

	hn->mRenderableNode = renderable;
	
	// ------------------------------
	// process groupHead, groupMember

	// no problem with group head
	hn->mGroupHead = groupHead;

	HierarchyNode * p = hn->getParent();
	if ( p )
	// there exists a parent node
	{
		if ( groupMember )
		// and "I should be" a group member
		{
			if (p->isGroupHead() )
			// then when parent is a group head, "I am" a group member and belong to parent
			{
				hn->mGroupMember = groupMember;
				hn->mGroup = p;
			}else if ( p->isGroupMember() )
			// when parent is a member of some group, "I will be" in the same group
			{
				hn->mGroupMember = groupMember;
				hn->mGroup = p->getGroup();
			}else
			// otherwise "I can't be" any group member
			{
				// A mistake! wrong group member flag
				LOGW << "Wrong model (" << mModelName  << ") - bad node membership! (flag ignored)" << std::endl;
			}
		}
		
	}
	else if (groupMember)
	// there don't exist any parent node, can't be group member
	{
		LOGW << "Wrong model (" << mModelName  << ") - bad node membership! (flag ignored)" << std::endl;
	}

	// ------------------
	// process attributes

	el = node.FirstChildElement("attributes").FirstChildElement().Element();
	for ( ; el != 0; el = el->NextSiblingElement() )
	{
		std::string s( el->ValueStr() );
		if ( s == "transform" )
		{
			// process transform attribute
			onTransformElement( TiXmlHandle( el ) );

		}else{
			// TODO: handle unknown attribute error
			continue;
		}
	}
	
	// ----------------
	// process children

	el = node.FirstChildElement( "children" ).FirstChildElement().Element();	
	for ( ; el != 0; el = el->NextSiblingElement() )
	{		
		mData.currentNode = new HierarchyNode();
		hn->addChild(mData.currentNode);

		onHierarchyNodeElement(TiXmlHandle(el));
	}
	
	// restore current node - last call will set currentNode to root
	mData.currentNode = hn;
}

void ModelLoader::onTransformElement(const class TiXmlHandle &handle)
{
	hna::Transformation tmpTrans;

	TiXmlElement * t = handle.FirstChildElement("node_transform").Element();
	if ( t != 0)
	{				
		TiXmlElement *m = t->FirstChildElement("matrix");
		if ( m != 0 )
		{
			// process NTM
			onMatrixElement( TiXmlHandle( m ) );
			tmpTrans.setNodeTM( GLMatrix4d( mData.matrix) );
		}
	}

	t = handle.FirstChildElement("object_transform").Element();
	if ( t != 0)
	{
		TiXmlElement *m = t->FirstChildElement("matrix");
		if ( m != 0 )
		{
			// process OTM
			onMatrixElement( TiXmlHandle( m ) );
			tmpTrans.setObjectOffsetTM( GLMatrix4d( mData.matrix) );
		}
	}

	if ( mData.currentNode != 0 )
	{
		counted_ptr<hna::Attribute > tmp(new hna::Transformation( tmpTrans ));
		mData.currentNode->setTransformation( tmp );
	}
}

void ModelLoader::onMatrixElement(const class TiXmlHandle &handle)
{
	GLMatrix4d mat( GLMatrix4d::identity() );

	TiXmlElement * r = handle.FirstChildElement( "r" ).Element();
	int rowIndex = 0;
	for ( ; r != 0; r = r->NextSiblingElement( "r" ) )
	{
		if ( rowIndex > 3 )
		{
			// TODO: handle wrong row count
			break;
		}

		int colIndex = 0;
		TiXmlElement * c = r->FirstChildElement( "c" );		
		for ( ; c != 0; c = c->NextSiblingElement( "c" ) )
		{
			if ( colIndex > 3 )
			{
				// TODO: handle wrong column count
				break;
			}

			TiXmlText * t = TiXmlHandle(c->FirstChild()).Text();
			if ( t == 0 )
			{
				//TODO: handle missing data field
				break;
			}
			double d;
			std::stringstream( t->ValueStr() ) >> std::skipws >> d;
			
			mat[ colIndex * 4 + rowIndex ] = d;
			
			++colIndex;
		}
		++rowIndex;
	}

	// copy result to context mData
	for (int i = 0; i < 16; ++i)
		mData.matrix[i] = mat[i];

}

void ModelLoader::insertGeometryObject( int id, gobj::GeometryObjectRef gobj )
{
	mIdGobjMap[id] = gobj;
}

gobj::GeometryObjectRef ModelLoader::getGeometryObject( int id )
{
	STL_EXT_NS::hash_map<int, gobj::GeometryObjectRef>::iterator it;
	it = mIdGobjMap.find( id );
	return ( it != mIdGobjMap.end() ) ? it->second : gobj::GeometryObjectRef(0);
}

void ModelLoader::insertMaterialIntoMap( int id, counted_ptr<hna::Attribute> material )
{
	mIDtoMaterialMap[id] = material;
}

counted_ptr<hna::Attribute> ModelLoader::getMaterialFromMap( int id )
{
	ModelLoader::IDtoAttributeMap::iterator it;
	it = mIDtoMaterialMap.find( id );
	return ( it != mIDtoMaterialMap.end() ) ? it->second : counted_ptr<hna::Attribute>(0);
}
