/* MAINTAINER: Martin */
#ifndef _MODEL_LOADER_H_
#define _MODEL_LOADER_H_

#include "model.h"

#include <string>
#include <memory>
#include "hash_hack.h"
#include "stlextdefs.h"

#include "gobj.h"

/**
*	\brief loads BJS model
*
*	Creates a new, empty model and loads its data from a file.
*
*/
class ModelLoader
{
public:

	virtual ~ModelLoader(){}
	
	/**
	*	\brief loads a 3D model
	*	\param fileName path to model
	*	\return new model or null
	*/
	Model * loadModel(const std::string &fileName);

protected:

	void onHeaderElement(const class TiXmlHandle &handle);
	void onGeometryElement(const class TiXmlHandle &handle);
	void onHierarchyNodeElement(const class TiXmlHandle &handle);
	void onBoxElement(const class TiXmlHandle &handle);
	void onSphereElement(const class TiXmlHandle &handle);
	
	/**
	*	\brief handler for trimesh node
	*	On server, this doesn't load any trimesh.
	*/
	void onTrimeshElement(const class TiXmlHandle &handle);
	void onTransformElement(const class TiXmlHandle &handle);
	void onMatrixElement(const class TiXmlHandle &handle);
	
	/**
	*	\brief handler for materials element
	*	On server, this code doesn't load any material.
	*/
	void onMaterialsElement(const class TiXmlHandle &handle);
	
	/** \brief type for pointer to event handler */
	typedef void (ModelLoader::* OnElementEventHandler )(const class TiXmlHandle &);
	
	/** \brief type for containter of event handlers */
	typedef STL_EXT_NS::hash_map<std::string, OnElementEventHandler> NodeHandlerByName; 
	
	/** \brief map type for mapping ID to attribute */
	typedef STL_EXT_NS::hash_map<int, counted_ptr<hna::Attribute> > IDtoAttributeMap; 

private:

	void resetState();
	
	/** \brief stores geometry object under specific id */
	void insertGeometryObject(int id, gobj::GeometryObjectRef gobj);

	/** \brief gets geometry object that was stored under specific id */
	gobj::GeometryObjectRef getGeometryObject(int id);

	void insertMaterialIntoMap( int id, counted_ptr<hna::Attribute> material );
	counted_ptr<hna::Attribute> getMaterialFromMap( int id );

	/**
	*	\brief create graphical representation in OpenGL
	*
	*	Graphical representation is created only on client.
	*
	*	\return true on success, false otherwise.
	*/
	bool createGraphicalRepresentation();

	/** \brief container of geometry objects (addressed by specific id) */
	STL_EXT_NS::hash_map<int, gobj::GeometryObjectRef> mIdGobjMap;
	
	/** 
	*	\brief maps material ID to material attribute 
	*	
	*	Material set is loaded only on client. Every material that was loaded is then
	*	inserted into this map under its id (map is empty on server). When building 
	*	hierarchy nodes and material associated with this node is needed, loader will 
	*	try to find material id in this map and will provide the requested material.
	*
	*	\note
	*/
	IDtoAttributeMap mIDtoMaterialMap;
	
	/** \brief model that is being built */
	std::auto_ptr<Model> mModel;
	
	/** \brief some methods gives data to others in this structure */
	struct {
		int gobjId;
		class HierarchyNode * currentNode;
		double matrix[16];
	} mData;

	/** \brief current model name */
	std::string mModelName;
};

#endif // _MODEL_LOADER_H_
