/* MAINTAINER: Martin */
#ifndef _MODEL_MANAGER_H_
#define _MODEL_MANAGER_H_

#include "universal_manager.h"
#include "model.h"
#include "counted_ptr.h"


class ModelLoader;

/**
*	\brief It maintains model prototypes and creates their instances.
*
*	At the first request for concrete model, the prototype is created and registered.
*	This prototype is owned by ModelManager.
*	Each request for a model (the first included) creates one new instance of the model 
*	prototype and the instance is given to client. It contains one counted reference 
*	to its prototype. Client is responsible for deleting the model instance. Deleting an 
*	instance includes freeing its reference. Model prototype will be deleted when its reference 
*	count reaches zero.
*	Creating a new instance means cloning hierarchic structure and attributes of its prototype, 
*	geometry objects - such as gobj::TriMesh, gobj::Box, gobj::Sphere - are shared.
*
*/
class ModelManager: public UniversalManager<Model, ModelManager>
{
	friend class Singleton<ModelManager>;	///< due to singleton DP
	friend class ModelCache;
protected:	
	ModelManager(): mModelLoader(0)	{}
	virtual ~ModelManager();
	
	/**
	*	\brief gets the model prototype stored under name
	*	
	*	If the prototype is not registered, it will be created and registered. Otherwise, only the
	*	counted reference to stored prototype is returned.
	*
	*	\param name model name
	*	\return reference to model prototype
	*/
	TRef getModelPrototype(std::string name);

public:
	
	/**
	*	\brief creates an instance of model prototype
	*
	*	If model prototype doesn't exist, it will be created. The prototype is then
	*	cloned and the result is returned. Model hierarchic structure and attributes are cloned,
	*	geometry objects - such as TriMesh, Sphere, Box - are shared.
	*
	*	\param name model name
	*	\return model instance
	*	\note Client is responsible for deleting the instance.
	*/
	Model * createModelInstance(std::string name);

private:
	
	ModelLoader * mModelLoader;
};

/** \brief model reference that can be used instead of ModelManager::TRef */
typedef ModelManager::TUniversalManager::TRef ModelRef; 

/**
*	\brief model instance subclass
*
*	Its purpose is to keep counted reference to model prototype.
*/
class ModelInstance: public Model
{
public:		
	ModelInstance(ModelRef prototypeRef);
	virtual ~ModelInstance(){}

private:
	
	ModelRef mPrototypeRef;	///< counted reference to model prototype
};

// Used fot the explicit template instantiation in template_instantces_*.cpp file
#ifdef INSTANTIATE_TEMPLATE
	template class Singleton< ModelManager >;
#endif


#endif //_MODEL_MANAGER_H_
