/* MAINTAINER: Lukas */

#ifndef _ASYNCRESULT_H
#define _ASYNCRESULT_H

#include "synchronizer.h"
#include "log.h"

template<class T> class AsyncFunction;

/**
 * Class that hold result from some asynchronous operation
 */
template<class T>
class AsyncResult : protected Synchronizer
{
friend class AsyncFunction<T>;

public:
	
	AsyncResult()
	{
		mResult = 0;
	}

	~AsyncResult()
	{
		mMutex.lock();

		if(mResult)
			delete mResult;

		mMutex.unlock();
	}

	/**
	 * Indicates that result was set
	 *
	 * \return true if result was set, false otherwise
	 */
	bool isDone()
	{
		return isCompleted();
	}
	
	/**
	 * Waits for operation result
	 */
	void waitForResult()
	{
		waitForCompleting();
	}

	/**
	 * Returns operation result. First time call returns result another calls returns 0.
	 * Object that is returned must be released by caller. This function blocks and waits for result.
	 *
	 * \return operation result
	 */
	T * getResult()
	{
		mMutex.lock();

		T * res;

		if(! isCompleted())
		{
			mCondVar.wait(mMutex);

			mMutex.lock();
		}

		res = mResult;

		mResult = 0;

		mMutex.unlock();

		return res;
	}
	
protected:
	/**
	 * Sets operation result. Input object must not be released and is under holder's control.
	 *
	 * \param res operation result object
	 */
	void setResult(T * res)
	{
		mMutex.lock();
		
		assertL(! isCompleted());

		mCommpleted = true;

		mResult = res;

		mMutex.unlock();

		mCondVar.broadcast();
	}

private:

	T * mResult;
};

#endif // _ASYNCRESULT_H
