/* MAINTAINER: Petko */
#ifndef _TCPIPTOOLS_H_
#define _TCPIPTOOLS_H_

#ifdef _WINDOWS
#include <winsock2.h>
#else
#include <netinet/in.h>
#include <netdb.h> // because of gethostbyaddr
#endif

namespace TcpIpTools
{
	
#ifdef _WINDOWS
	typedef SOCKET BJS_SOCKET_TYPE;
	const BJS_SOCKET_TYPE BAD_BJS_SOCKET = INVALID_SOCKET;
	const int BJS_SOCKET_ERROR = SOCKET_ERROR;
#else
	typedef int BJS_SOCKET_TYPE;
	const BJS_SOCKET_TYPE BAD_BJS_SOCKET = -1;
	const int BJS_SOCKET_ERROR = -1;
#endif
	
	/**	\brief Creates BJS socket - wrapper for socket() function
	 *
	 *	\param type Can be SOCK_DGRAM or SOCK_STREAM
	 *	\return New socket or BAD_BJS_SOCKET on error
	 */
	BJS_SOCKET_TYPE socket( int type );
	
	/**	\brief Binds socket for local endpoint - for all addresses and specified ports
	 *
	 *	\param sock Socket that will be used for receiving and transmitting data
	 *	\param port Port that will be used for communication - in usual endian - DO NOT CHANGE IT
	 *	\return 0 on success, BJS_SOCKET_ERROR otherwise
	 */
	int bind( BJS_SOCKET_TYPE sock, unsigned short port );

	/**	\brief Reads data from socket - datagram - method is blocking
	 *
	 *	\param sock Socket to read from
	 *	\param data Store for data
	 *	\param datalen Length of data (in bytes)
	 *	\param from Provide preallocated structure where will be information about packet sender stored
	 *
	 *	\return Number of bytes received, or 0 if socket was closed properly, or BJS_SOCKET_ERROR in case of error.
	 */
	int recvfrom( BJS_SOCKET_TYPE sock, char * data, int datalen, struct sockaddr_in * from );

	/**	\brief Sends data to socket - datagram - method is blocking
	 *
	 *	\param sock Socket to write to
	 *	\param data Data to transfer
	 *	\param datalen Length of data (in bytes)
	 *	\param to Structure defining the addressee
	 *
	 *	\return Number of bytes sent, or BJS_SOCKET_ERROR in case of error.
	 */
	int sendto( BJS_SOCKET_TYPE sock, const char * data, int datalen, const struct sockaddr_in * to );

	/**	\brief Writes data to socket - datagram - method is blocking
 	 *
 	 *	\param sock Socket for writing
 	 *	\param data Data to write
	 *	\param datalen Length of data (in bytes)
	 *	\param addressee String with address of recipient - can be DNS name or IP address string
	 *	\param port Port where to send data
	 *	\return Number of bytes sent, or BJS_SOCKET_ERROR in case of error.
	 */
	int sendto( BJS_SOCKET_TYPE sock, const char * data, int datalen, const char * addressee, unsigned short port  );

	/**	\brief Writes data to socket - datagram - method is blocking
	 *
	 *	\param sock Socket for writing
	 *	\param data Data to write
	 *	\param datalen Length of data (in bytes)
	 *	\param addressee IP address as long
	 *	\param port Port where to send data
	 *	\return Number of bytes sent, or BJS_SOCKET_ERROR in case of error.
	 */
	int sendto( BJS_SOCKET_TYPE sock, const char * data, int datalen, unsigned long addressee, unsigned short port  );

	/**	\brief Waits for timeout milliseconds if socked is ready for writing data - so sendto won't block
	 *
	 *	\param sock Socket to check
	 *	\param timeout Time in millisecond to wait for write-readiness
	 *	\return 1 if socket is ready, 0 if timeout elapsed, BJS_SOCKET_ERROR on error
	 */
	int selectWrite( BJS_SOCKET_TYPE sock, int timeout );

	/**	\brief Waits for timeout milliseconds if socked is ready to receive some data - so recvfrom won't block
	 *
	 *	\param sock Socket to check
	 *	\param timeout Time in millisecond to wait for read-readiness
	 *	\return 1 if socket is ready, 0 if timeout elapsed, BJS_SOCKET_ERROR on error
	 */
	int selectRead( BJS_SOCKET_TYPE sock, int timeout );

	/**	\brief Sets socket option to be able to receive and transmit broadcast data
	 *
	 *	\param sock Socket to set the parameters
	 *	\return 0 on success, BJS_SOCKET_ERROR otherwise
	 */
	int setSockBroadcast( BJS_SOCKET_TYPE sock );

	/**	\brief Close socket
	 *
	 *	\param sock Socket to close
	 */
	void closesocket( BJS_SOCKET_TYPE sock );

}


#endif	//_UDPTOOLS_H_
