/* MAINTAINER: Bernard */

// #define WANT_TRACE

#include <cstdio>
#include <cstdlib>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sys/wait.h>
#include <errno.h>
#include <pwd.h>
#include <dirent.h>
#include <cstring>
#include <fstream>
#include <dirent.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "platform.h"
#include "log.h"
#include "globals.h"
#include "settings.h"

using namespace std;

/**
 * \brief copy file
 *
 * Copy one file to another.
 * \param src source file
 * \param dst destitanion file
 * \return true if OK
 */
inline bool copy(const char* src, const char* dst)
{
	std::ifstream srcstream(src);
	std::ofstream dststream(dst, std::ios::out|std::ios::trunc|std::ios::binary);

	if (!srcstream || !dststream) return false;

	dststream << srcstream.rdbuf();

	return dststream.good();
	
/* // more unix version
	switch ( fork() )
	{
	case 0:
		execlp("cp", "cp", conf_template.c_str(), userconf.c_str(), NULL);
		exit(1);
	case -1:
		LOGE << "Unable to fork!" << endl;
		break;
	default:
		int status;
		wait(&status);
	}
*/

}

/**
 * \brief Locate home directory.
 *
 * \return ~ (resolved)
 */
const char* getHomeDir()
{
	static string homeDIR = "";
	
/*
	uid_t uid = getuid();
	struct passwd* pw = getpwuid(uid);
	homeDIR = pw->pw_dir;
	free(pw);
*/
	
	if (getenv("HOME")) homeDIR = getenv("HOME");

	return homeDIR.c_str();
}

/**
 * \brief Locate home directory and adds ./bjs
 *
 * \return ~/.bjs (resolved)
 */
const char* getHomeBJS()
{
	static string hbjs = "";
	
	if (hbjs == "") 
	{
		hbjs.append(getHomeDir()).append("/.bjs");
	}

	return hbjs.c_str();
}


/**
 * \brief Paths with 
 */
const char* PATHS_UNIX[] = 
{
	 getHomeBJS(), ".", "/usr/share/bjs",  "/usr/local/share/bjs/", 
        "/usr/share/games/bjs", "/usr/local/share/games/bjs",
        "/usr/bin", "/usr/local/bin", "/etc/bjs", /* INSTALL_DIR, */  0,
};


bool checkConfigFile(const std::string &file, const std::string &file_template)
{
	TIN
	

	string userconf = getHomeBJS();
	userconf.append("/").append(file);

	if ( exist(userconf.c_str()) )
	{
		LOGT << file << "exists" << endl;
		return true;
	}

	int rv = mkdir(getHomeBJS(), 0777);
	
	if ( !(rv == 0 || rv == EEXIST) )
	{
		LOGW << "Problem with path " << getHomeBJS() << endl;
	}

	string conf_template = locateFile(file_template);

	if (!copy( conf_template.c_str(),  userconf.c_str() ) )
	{
		LOGW << "failed to copy " << conf_template << " to " << userconf << endl;
	}
	
	return true;
}

// You should also update direcotry listing if you try to domething with this
string locateFile(const string& file)
{
	TIN
	
	if (file == "") return file;
	
	const char** dir  = SUBDIRECTORIES; 
	
	string f;
	
	for (; *dir != 0; ++dir)
	{
		for (int i = 0; PATHS_UNIX[i] != 0; ++i) 
		{
			f.clear();
			f.append( PATHS_UNIX[i]).append("/").append(*dir).append(file);
			
			// LOGD << "trying " << f << endl;
			
			if ( exist(f.c_str()) ) return f;
	
		}
	}

	LOGW << "Unable to locate file " << file << endl;

	return file;
}

bool startServer()
{
	// We need to get path to our own executable	
	
	string executable = g_settings.getString(SETTING_SERVER_EXECUTABLE);
	string serverPath = locateFile(executable);
	
	if (!exist(serverPath.c_str()))
	{
		serverPath = g_settings.getString(SETTING_ARGV0);
		size_t idx = serverPath.rfind("/");
		if (idx == string::npos) 
		{
			LOGE << "Unable to locate server executable." << endl;
			return false;
		}
		
		serverPath.resize(idx+1);
		serverPath.append(g_settings.getString(SETTING_SERVER_EXECUTABLE));
	}

	if (!exist(serverPath.c_str()))
	{
		LOGE << "Unable to locate server executable." << endl;
		return false;
	}
	
	LOGI << "Located server executable: " << serverPath << endl;
	
	switch (fork())
	{
	case 0:
		execl(serverPath.c_str(), 
			  serverPath.c_str(), 
			  g_settings.settingToOption(SETTING_SERVER_MAP).c_str(),
			  g_settings.settingToOption(SETTING_SERVER_PASSWORD).c_str(),
			  g_settings.settingToOption(SETTING_SERVER_SINGLE_PLAYER).c_str(),
			  g_settings.settingToOption(SETTING_GAME_INIT_COMMANDS).c_str(),
			  (char *) NULL);
		DIE("Server launch failed");
	case -1:
		return false;
	}
	
	return true;
}

bool addOneDirecotiry(	const std::string &directoryPath, std::vector<std::string> &filesV);

bool listDirectory(const std::string &directory, std::vector<std::string> &filesV)
{
	filesV.clear();		
	if (directory == "") return true;
	
	
	const char** dir  = SUBDIRECTORIES; 
	
	string f;
	
	for (; *dir != 0; ++dir)
	{
		for (int i = 0; PATHS_UNIX[i] != 0; ++i) 
		{
			f.clear();
			f.append( PATHS_UNIX[i]).append("/").append(*dir).append(directory);
			
			LOGT << "trying " << f << endl;
			
			addOneDirecotiry(f, filesV);	
		}
	}

	return true;
}

bool addOneDirecotiry(	const std::string &directoryPath, std::vector<std::string> &filesV)
{
	/*
	DIR *dir = opendir(directoryPath.c_str());
	if (dir == NULL) 
	{
		//LOGE << "Error when listing direcotry " << path << ": " << strerror(errno) << endl;
		return false;
	}
	
	struct dirent *dirInfo;
	while ( (dirInfo = readdir(dir)) != NULL )
	{
		filesV.push_back(dirInfo->d_name);
	}
	
	closedir(dir);
	*/
    struct dirent **namelist;
    int n;

    n = scandir(directoryPath.c_str(), &namelist, 0, alphasort);
    if (n < 0)
    {
		//LOGE << "Error when listing direcotry " << directoryPath << ": " << strerror(errno) << endl;
		// this may be cased because the direcotory doesn't exist
		return false;
	}
	else 
	{
		for (int i = 0; i < n; i++)
		{
			filesV.push_back(namelist[i]->d_name);
			free(namelist[i]);
		}
		free(namelist);
	}
		
	return true;
}

std::string getLogDir()
{
	return "";
}

void initPlatform()
{
}

