/* MAINTAINER: Martin */
#ifndef QUADTREE_H_INCLUDED
#define QUADTREE_H_INCLUDED

/** Trida pro operace s QuadTree.
*	Uchovava strukturu popisujici strom a nad ni pracuje, lze pripojit data.
*	To znamena, ze jeden strom muze pracovat nad vice daty.
*	Definuje operace nad uzly stromu, tedy je to jakysi nadrazeny manipulator.
*
*/
class Quadtree
{
public:
	enum neighbor_t { n_left, n_up, n_right, n_down };
	
	/**
	*	Typ reprezentujici jeden uzel. 
	*	@note Mozna by to slo vmestnat do 1 bytu a ne do 8. Ale timto typem neni
	*	reprezentovan samotny quadtree, tato struktura slouzi pouze k prochazeni stromu.
	*/
	struct QuadtreeNode
	{
		int width;	//!< sirka uzlu
		int x, y;	//!< position in external data matrix (data controlled by this quadtree)
	};

	Quadtree():	m_matrix(0)
	{m_root.width = m_root.x = m_root.y = 0;}

	~Quadtree()
	{Free();}

	/** Alokuje pamet pro strukturu stromu.
	*	@param width Sirka korenoveho uzlu.
	*	@note Alokuje <c>width * width</c> pameti (bool) pro matici reprezentujici quadtree.
	*/
	bool Alloc(const int width);
	
	/** Uvolni alokovanou pamet. */
	void Free(void);
	
	/** Ziska korenovy uzel stromu.
	*	@param root Koren stromu.
	*/
	inline void get_root(QuadtreeNode & root)
	{ root = m_root; }
	
	/** Vraci index hodnoty ulozene v listu stromu.
	*	@param leaf list, ze ktereho chceme hodnotu ulozenou ve stromu
	*	@return Pozice dat v poli, ktere pokryva tento strom.
	*/
	inline int get_index(const QuadtreeNode & leaf)
	{ return get_pos(leaf); }

	/** Testuje, zda je dany uzel dale rozdelen na poduzly.
	*	@return True, pokud je uzel <a> node </a> rozdelen na 4 poduzly. Jinak false.
	*	@note Pokud vraci false, znamena to, ze jsme narazili na uzel stromu.
	*/
	inline bool is_subdivided(const QuadtreeNode & node)
	{ return m_matrix[get_pos(node)]; }

	/** Ziska 4 poduzly daneho uzlu, pokud existuji.
	*	@param node Uzel, ze ktereho chceme ziskat poduzly.
	*	@param nodes Predalokovane pole pro 4 poduzly uzlu <a>node</a>.
	*	@return Povedlo/Nepovedlo.
	*/
	bool get_subnodes(const QuadtreeNode & node, QuadtreeNode *nodes);

	/** Ziska odkaz na souseda.
	*	@param type Typ souseda (vlevo, vpravo, nahore, dole).
	*	@param neighbor Sousedni uzel podle <a>type</a>.
	*/
	bool get_neighbor(const neighbor_t type, const QuadtreeNode &node, QuadtreeNode &neighbor);

	inline void set_node(const QuadtreeNode & node)
	{
		m_matrix[get_pos(node)] = true;
	}
	inline void set_leaf(const QuadtreeNode & node)
	{
		m_matrix[get_pos(node)] = false;
	}
	
protected:
	inline int get_pos(const QuadtreeNode & node){ return node.y * m_root.width + node.x; }
	QuadtreeNode m_root;
	
	/** Matice pro reprezentaci quadtree. Potrebuje <c>width*width</c> (bytu) pameti. */
	bool * m_matrix;
};

#endif
