/* MAINTAINER: Martin */

#ifndef _REF_COUNTER_H
#define _REF_COUNTER_H

#include <list>
#include "log.h"

/**
*	\brief Base class for reference counted objects.
*
*	\remarks A static variable is used inside this class! Be sure you are 
*		NOT USING this class as static member or global variable before 
*		entering the main() function!
*	
*/
class RefCounter
{
public:
	virtual ~RefCounter();

	/**
	*	\brief	Customized overloaded operator new (it only allocates memory).
	*
	*	This is used when a new object derived from RefCounter is constructed dynamically
	*	using the 'new'-operator. RefCounter needs to distinguish between instancess
	*	allocated on the heap and instances allocated on the stack. This is the place where
	*	it is performed.
	*
	*	\param	size the size of constructed instance
	*	\return	pointer to allocated memory for this instance
	*/
	void * operator new(size_t size);

	inline bool isHeapAllocated(void);
	
	int getRefCount(void){ return mRefCount; }

protected:
	RefCounter();
	inline void incRef();
	inline void decRef();

	virtual void OnZeroRefCount();

private:
	int		mRefCount;
	bool	mHeapAllocated;

	typedef std::list<void*> TTempHeapPointers;
	static  TTempHeapPointers mTmpHeapPtrs;

friend class RefCounterPtrBase;
};

void RefCounter::incRef()
{
	TIN; 
	assertL((mRefCount >= 0) && "Bad reference count!");
	++mRefCount;
}

void RefCounter::decRef()
{
	TIN; assertL((mRefCount > 0) && "Bad reference count!");
	if (--mRefCount == 0)
		OnZeroRefCount();
}

bool RefCounter::isHeapAllocated(void)
{
	return mHeapAllocated;
}

/**
*	\brief Base class for RefCounterPtr classes, a friend of class RefCounter.
*
*	The RefCounterPtrBase is declared as a friend of RefCounter class and provides 
*	protected methods for calling its private methods for increasing or decreasing 
*	reference count. Therefore only classes derived from this "base" can modify the 
*	reference count. Those should be some RefCounterPtr classes that provides 
*	overloaded operators and behaving themselves like pointers.
*/
class RefCounterPtrBase
{
protected:
	RefCounterPtrBase(){}
	static void incRef(RefCounter & rc) { rc.incRef(); }
	static void decRef(RefCounter & rc) { rc.decRef(); }
};

/**
*	\brief Reference-counting object, behaves as a pointer, works on RefCounter-derived types.
*
*	The first template argument (denotet as T) is the type with it is going to treat as with RefCounter 
*	object. The second template argument (denoted as RetT) is the type we want to publish with this 
*	RefCounter object.
*	"Publish" means that the result of access operators (operator *, operator ->, operator RetT*) is of
*	type RetT. This all is due to classes (passed to T), that are derived from RefCounter class and wrap
*	or refer to some object of type RetT. In that cases we want to access RetT and not the wrapper-class.
*/
template <class T, class RetT = T>
class RefCounterPtr : public RefCounterPtrBase
{
protected:
	T* mPtr;	///< pointer, that this class guards

public:
		
	// empty constructor - simply assign pointer p to NULL
	RefCounterPtr(void):mPtr(0)
	{}
		
	// constructing with a pointer
	RefCounterPtr(T* t):mPtr(0)
	{ *this = t; }	// we use the feature of overloaded assignment operator ;)
		
	// copy constructor
	RefCounterPtr(const RefCounterPtr<T, RetT> &t):mPtr(0)
	{ *this = t; } // we use the same feature ;o)
		
	// destructor
	~RefCounterPtr(void)
	{ if (mPtr) decRef(*mPtr); } // remove reference from object we are pointing to


	// assignment operator, assigns a pointer
	inline RefCounterPtr& operator =(T* t)
	{
		T * tmp = mPtr;
		if ((mPtr=t) != 0) incRef(*mPtr);	// add reference to object that will be pointing by "mPtr"
		if (tmp) decRef(*tmp);				// remove reference from object pointed by "mPtr"
		return *this;						// return this object for next operations
	}

	// assignment operator, assigns an another RefCounterPtr object
	inline RefCounterPtr& operator =(const RefCounterPtr<T, RetT> &t)
	{
		T * tmp = mPtr;
		if ((mPtr = t.mPtr) != 0) incRef(*mPtr);	// add reference to object that will be pointing by "mPtr"
		if (tmp) decRef(*tmp);						// remove reference from object pointed by "mPtr"
		return *this;								// return this object
	}

	/// \brief reference access
	inline RetT& operator *(void) const
	{ assertL(mPtr && "Using zero pointer in RefCounterPtr class!"); return (*mPtr); }

	/// \brief pointer access
	inline RetT* operator ->(void) const
	{ assertL(mPtr && "Using zero pointer in RefCounterPtr class!"); return &(this->operator *()); }

	/// \brief conversion - allow automatic conversion to type RetT*
	inline operator RetT*(void) const
	{ return isValid() ? &(this->operator *()) : 0; }

	/// \brief negation
	inline bool operator !(void) const
	{ return !(mPtr); }

	/// \brief equivalence with another RefCounterPtr object
	inline bool operator ==(const RefCounterPtr<T, RetT> &t) const
	{ return (mPtr == t.mPtr); }
		
	/// \brief equivalence with pointer
	inline bool operator ==(T* t) const
	{ return (mPtr == t); }

	inline bool isValid(void) const
	{ return (mPtr != 0); }

	inline int getRefCount() const
	{ 
		return isValid() ? static_cast<RefCounter*>(mPtr)->getRefCount() : 0; 
	}

	inline bool isUnique() const
	{
		return getRefCount() == 1;
	}

};


#endif //_REF_COUNTER_H

