/* MAINTAINER: Bernard */

// #define WANT_TRACE

#include <stdlib.h>
#include <fstream>
#include <sstream>
#include <iostream>
#include <cstring>
#include "log.h"
#include "settings.h"
#include "conversions.h"


#include "tinyxml/tinyxml.h"

using namespace std;

#define _SETTINGS_DEFINE_SET
#include "settings_set.h"

// global singleton
Settings g_settings;

/*****************************************************************************/
Settings::Settings()
{
	// adding default values
	#define _SETTINGS_INITIALIZE_SET
	#include "settings_set.h"
 
}

/*****************************************************************************/
bool Settings::loadSettings(const char *filename)
{
	TIN;
	
	assertL(filename != 0);
	
	TiXmlDocument doc(filename);
	TiXmlHandle   handle(&doc);


	if (!doc.LoadFile()) 
	{
		LOGE << "Unable to load " << filename 
		     <<" Probably file not exist or isn't a valid XML file" << endl;
		return false;
	}	
	

	TiXmlElement* child = handle.FirstChild( "settings" ).FirstChild( "variable" ).Element();
	if (!child)
	{
		LOGE << filename << " has bad structure" << endl;
		return false;
	}
	for (; child; child=child->NextSiblingElement()) 
	{
		const char *key   = child->Attribute("name");
		const char *value = child->Attribute("value");
		if (key != NULL && value != NULL) 
		{

			set(key, value);
			
			LOGT << "Loaded pair '" << key << "' =  '" << value << "'" << endl;
		}
	}
	
	return true;
}


/*****************************************************************************/
bool Settings::saveSettings(const char *filename)
{
	TIN

	assertL(filename != 0);
	
	TiXmlDocument doc(filename);
	TiXmlHandle   handle(&doc);

	if (!doc.LoadFile()) 
	{
		LOGE << "Unable to load " << filename 
		     <<" Probably file not exist or isn't a valid XML file" << endl;
		return false;
	}	
	

	TiXmlElement* child = handle.FirstChild( "settings" ).FirstChild( "variable" ).Element();
	for (; child; child=child->NextSiblingElement()) 
	{
		const char *key   = child->Attribute("name");

		// invalid key? Maybe some log?
		if (key == 0) continue;

		// skip read only values - this will be added later
		const char *readonly   = child->Attribute("readonly");
		if (readonly)
		{
			bool ro;
			if (try_get_bool(readonly, ro) && ro == true) continue;
		}
		
		// save the value if we have it
		if ( exist(key) ) 
		{
			child->SetAttribute("value", getString(key));
			TVARIABLE2(key, getString(key));
		}
		else
			LOGW << "Unknown log key " << key << endl;
	}

	if ( !doc.SaveFile() )
	{
		LOGE << "Unable to write to " << filename << endl;
		return false;
	}

	return true;
}

/*****************************************************************************/
std::string Settings::getString(const char *entry)
{
	map<string, string>::iterator it = find(entry);

	return it->second;
}


/*****************************************************************************/
bool Settings::compare(const char *entry, const char *value)
{
	map<string, string>::iterator it = find(entry);

	return (it->second.compare(value) == 0);
}


/*****************************************************************************/
void Settings::set(const std::string &key, const std::string &value)
{
// This TIN may be dangerous since it is used in contructor and
// thus log doesn't have to exist. Depends on compiler. So
// uncoment if you want to debug settings
//	TIN
//	LOGT << "Setting in key=" << key << " value = " << value << endl;
	
	map<string, string>::iterator it = mMap.find(key);
	if (it == mMap.end()) 
	{
		mMap.insert(make_pair(key, value));
	}
	else 
	{
		it->second = value;
	}
}

/*****************************************************************************/
bool Settings::isTrue(const char *entry)
{
	map<string, string>::iterator it = find(entry);
	
	return get_bool(it->second);
}
	
/*****************************************************************************/
map<std::string, std::string>::iterator Settings::find(const char *entry)
{
	map<string, string>::iterator it = mMap.find(entry);
	if (it == mMap.end()) 
	{
		LOGE << "Wanted setting value '" << entry << "' that doesn't exist!" << endl;
		DIE();
	}


	return it;
}
	
/*****************************************************************************/
bool Settings::exist(const char *entry)
{
	map<string, string>::iterator it = mMap.find(entry);
	if (it == mMap.end()) 
	{
		return false;
	}

	return true;
}

/*****************************************************************************/
int Settings::getInt(const char *entry)
{
	map<string, string>::iterator it = find(entry);

	return get_long(it->second);
}


double Settings::getDouble(const char *entry)
{
	map<string, string>::iterator it = find(entry);
	
	LOGT << "returning entry=" << entry << " value=" << it->second << endl;
	
	return get_double(it->second);	
}

/*
template <class T>
void Settings::set(const std::string &key, const T &val)
{
// This is used in costructor => log doesn't have to exist sice
// log and this are global variables
//	TIN 
	
	stringstream s;
	s << val;
	
	set(key, s.str());
}
*/

/*****************************************************************************/
void Settings::parseCommandLine(int argc, char* argv[])
{
	set(SETTING_ARGV0, argv[0]);
	LOGD << "This execitable: " << argv[0] << endl;
	
	for (int argcID = 1; argcID < argc; argcID++)
	{
		// We are recognizing setting in form:  --key=value
		char *arg = argv[argcID];
		if (arg[0] != '-' || arg[1] != '-') continue;
	
		arg += 2;

		char *equals = strchr(arg, '=');
		if (equals == NULL) continue;
		
		*equals = '\0'; // remove = and put '\0'  instead so we may read command easily
		
		// we have key string
		string key(arg);
		
		*equals = '='; // put = back
		
		equals++; // character after =    - it means the value
		
		string value(equals);
		
		set(key, value);
	}
}

/*****************************************************************************/
std::string Settings::settingToOption(const std::string &key)
{
	string option="--";
	option.append(key);
	option.append("=");
	option.append(getString(key.c_str()));
	
	return option;
}
