/* MAINTAINER: Bernard */

#ifndef _SETTINGS_H_
#define _SETTINGS_H_

#include <map>
#include <string>
using namespace std;

// All settings constants mus go somewhere
#define _SETTINGS_EXPORT_SET
#include "settings_set.h"

/**
 * \brief Class for holding pairs key/value intendet to be used as setting.
 * 
 * If you want to add some entry to setting, write a line to setting_data.h
 * and the you may access your variable by g_settings.getXXX(). Also don't
 * forget to add description of your variable to bjs.conf.xml.
 */
class Settings 
{

public:

	/**
	 * Set default values for all settings.
	 */
	Settings();
	
	/**
	 * \brief Load setting from config XML.
	 *
	 * Not yet known keys are added and old values are rewriten.
	 * You may call this function more times on one object.
	 * \param  filename path to file with settings
	 * \return true if loading was sucessfull
	 */
	bool loadSettings(const char *filename);
	
	/**
	 * \brief Save setttings to file.
	 *
	 * Opens and scans file. Only rewrite values in file to new values.
	 * Doesn't add any new variables. Note that indentation may change but
	 * all data including comments are kept.
	 * \param filename filename
	 * \return true if sucessfull
	 */
	bool saveSettings(const char *filename);

	/**
	 * \brief Get double asociated with key.
	 * 
	 * If the entry doesn't exist or can't be converted to double
	 * the method kills the program
	 * \param entry key to with double
	 * \return asociated double
	 */
	double getDouble(const char *entry);
	
	/**
	 * \brief Return variable from setting converted to int
	 *
	 * Die if entry not found.
	 * Die if not a number.
	 */
	int  getInt(const char *entry);
	
	/**
	 * Search for entry in settings. If entry doesn't
	 * exist, the program die. If you need store
	 * any entry in Settings, add it's default value to constructor.
	 * Else the program may fall.
	 */
	string getString(const char *entry);

	// return true if entry has value value
	bool compare(const char *entry, const char *value);

	/**
	 * \brief Search for entry and return key converted to bool.
	 * 
	 * DIE if any error encountered.
	 * \param entry Key to search for.
	 * \return Value associated with key converted to bool
	 */
	bool isTrue(const char *entry);

	/**
	 * \brief Same as isTrue.
	 * 
	 * It is here to provide more consistent interface with other getXXX
	 */
	bool getBool(const char *entry)
	{
		return 	isTrue(entry);
	}
	
	/**
	 * \brief Associate string value to key.
	 * 
	 * If the key allready exist, the old value is forgotten. If the
	 * value isn't present, new entry is created.
	 * \param key The Key.
	 * \param value String value for the Key.
	 */
	void set(const std::string &key, const std::string &value);
	
	/**
	 * \brief Associate bool value to the key.
	 * 
	 * \param key The Key.
	 * \param valye Bool value converter to string "true" or "false".
	 */
	void set(const std::string &key, bool value)
	{
		if (value) set(key, "true");
		else       set(key, "false");
	}

	void set(const std::string &key, const char *value)
	{
		set(key, string(value));
	}
	
//	template <class T>
//		void set(const std::string &key, const T &val);

	/**
	 * \brief Check if entry is known.
	 *
	 * \param entry key to check
	 * \return true if key exists
	 */
	bool exist(const char *entry);

	/**
	 * \brief Parse command line options.
	 * 
	 * It is possible to add/modify setting from command line.
	 * The recognized pattern is --key=value. All such pairs are
	 * added to settig. Plus as an extra bonus argv[0] is associated with
	 * SETTING_ARGV0
	 * The command line isn't modified.
	 * \param argc The argc from main.
	 * \param argv The argv from main.
	 */
	void parseCommandLine(int argc, char* argv[]);


	/**
	 * \brief Put key and value together such that it will be pared by parseCommandLine.
	 * 
	 * \param Key you want to know
	 * \return Formated string containing key and it's value.
	 */
	std::string settingToOption(const std::string &key);

private:

	
	// die if iteratr would be invalid
	map<string, string>::iterator find(const char *entry);
	
	map<string, string> mMap; ///\brief Here is all the setting stored.
};

extern Settings g_settings; ///\brief Global variable for accessing the setting.

#endif
