/* MAINTAINER: Martin */

#ifndef _SINGLETON_H
#define _SINGLETON_H

#include <cassert>

/**
*	\brief macro that inserts the needed singleton design pattern stuffs
*
*	Usage:
*	\code
*	class OnlyOne: public Singleton<OnlyOne>
*	{
*		// insert protected ctor, dtor and friendship with Singleton<OnlyOne>
*		_singletonDP(OnlyOne)
*
*	public:
*		// class definition
*	};
*	\endcode
*
*	\remarks sets access rights to protected
*/
#ifndef _singletonDP
#define _singletonDP(ClassName) \
	friend class Singleton<ClassName>; \
	protected: ClassName(){} \
	protected: ~ClassName(){}
#endif

/**
*	\brief resolves ambiguity with CEGUI::Singleton<T>
*/
namespace SingletonBJS
{
/**
*	\brief singleton design pattern implementation
*
*	\code
*	// Singleton<T> usage example:
*
*	class OneClass: public ClassA, ClassB, Singleton<OneClass>
*	{
*
*	// I. Singleton stuffs
*	// only the Singleton<OneClass> can call OneClass::OneClass()
*	friend class Singleton<OneClass>;
*	protected:
*		OneClass();  // protected ctor (singleton)
*		~OneClass(); // protected dtor (singleton)
*
*	// II. OneClass definition - methods, attributes
*	public:
*		some_OneClass_method()
*	};
*	
*	...
*
*	// Equivalent usage of the OneClass singleton:
*	OneClass::getInstance()->some_OneClass_method();
*	OneClass::getPtr()->some_OneClass_method();
*	OneClass::getRef().some_OneClass_method();
*	\endcode
*
*	\remarks Singleton variables are created at the first instance request. But it is STRONGLY
*		recommended DO NOT using any Singleton class at static-variables initialization time
*		(before entering the main()).
*
*	\remarks Singleton classes are destroyed at the program exit (at static variables destruction).
*		Therefore, using other classes and global or static variables (other singletons as well) 
*		inside the destructor-body of some singleton class could be unsafe (the variables that we 
*		are attempting to use might be already destroyed)!
*
*	\remarks Sometimes, it can be possible to destroy the singleton directly by using
*		<code>delete SomeSingleton::getPtr()</code>. It isn't recommended, however it 
*		shouldn't cause any program's crash...
*/
template <class T>
class Singleton
{
public:
	/**
	*	\note sets internal instance pointer to NULL due to safety reasons.
	*/
	virtual ~Singleton(void)
	{ 
		// this avoids crashes when deleting singleton directly by "delete SomeSingleton::getPtr();"
		m_InstancePtr = 0;
	}

	/**
	*	\brief returns instance pointer
	*
	*	The instance of T is created at the first getInstance() call.
	*
	*	\return pointer to instace of type T
	*	\note getInstance, getPtr and getRef are equivalent
	*/
	static inline T * getInstance(void);

	/**
	*	\brief returns instance pointer
	*
	*	The instance of T is created at the first getInstance() call.
	*
	*	\return pointer to instace of type T
	*	\note getInstance, getPtr and getRef are equivalent
	*/
	static inline T * getPtr(void)
	{ return getInstance(); }
	
	/**
	*	\brief returns instance reference
	*
	*	The instance of T is created at the first getInstance() call.
	*
	*	\return reference to instace of type T
	*	\note getInstance, getPtr and getRef are equivalent
	*/	
	static inline T & getRef(void)
	{ return *getInstance(); }

protected:
	/** \brief protected ctor due to singleton DP */
	Singleton(void){}
	
	/** \brief protected ctor due to singleton DP */
	Singleton(const Singleton&){ assert("Singleton Error: Copy constructor is not allowed!"); }

	/** \brief protected operator= due to singleton DP */
	Singleton& operator= (const Singleton&){ assert("Singleton Error: Assignment operator is not allowed!"); return *this; }

private:
	static T * m_InstancePtr;	///< \brief static pointer to the instance of the singleton class T

	/** \brief creates a new instance of singleton class T */
	static void createInstance(void)
	{ 
		//assert(&m_InstancePtr && "Using singleton before initialization of its static variables!");
		assert(!m_InstancePtr && "Attempt to create instance that is already created!");
		m_InstancePtr = new T; 
	}
	
	/** \brief destroys the instance of singleton class T */
	static void destroyInstance(void)
	{ delete m_InstancePtr; m_InstancePtr = 0; }

	/**
	*	\brief class that will destroy the singleton instance at the program exit.
	*/
	class InstanceDestruction{
	public:
		/** \brief just destroys the singleton instance of this template implementation */
		~InstanceDestruction(void)
		{
			Singleton<T>::destroyInstance();
		}
	};

	/**
	*	\brief static class that destroys singleton instance at the program exit.
	*
	*	\note it doesn't fail if the instance was destroyed before somehow and set to null
	*/
	static typename Singleton<T>::InstanceDestruction m_InstanceDestruction;
};

template <class T>
T* Singleton<T>::getInstance(void)
{
	// At the first time the Instance() is called, this will create the instance.
	static const bool volatile instance_created((createInstance(), true));
	
	assert(instance_created);
	assert(m_InstancePtr);

	return m_InstancePtr;
}

template <class T>
T * Singleton<T>::m_InstancePtr = 0;

template <class T>
typename Singleton<T>::InstanceDestruction Singleton<T>::m_InstanceDestruction;

}; // end of namespace

// explicitly use this namespace
using namespace SingletonBJS;


#endif //_SINGLETON_H

