/* MAINTAINER: Lukas */

#ifndef _SYNCHRONIZEDQUEUE_H
#define _SYNCHRONIZEDQUEUE_H

#include "log.h"
#include "condvar.h"
#include <queue>
#include "mutex.h"

/**
 * Queue that is threadsafe
 *
 * Templated copy of eventqueue
 */
template<class TObject>
class SynchronizedQueue
{

public:
	
	SynchronizedQueue() { };
	~SynchronizedQueue() { };

	/**
	 * Pushes new object to the back of the queue, threadsafe version with signaling if
	 * queue is empty
	 * 
	 * \param new object to add
	 */
	void push( TObject object )
	{
		pushWithSignal( object );
	}
	
	/**
	 * Return first object in queue, waiting for signal if queue is empty,
	 * threadsafe version
	 *
	 * \return first object in queue
	 */
	TObject pop()
	{
		return  popWithSignal();
	}

	/**
	 * Return true if queue is empty
	 *
	 * \return true if queue is empty
	 */
	bool isEmpty()
	{
		bool empty;

		mMutex.lock();
		empty = mQueue.empty();
		mMutex.unlock();

		return empty;
	}

protected:
	/**
	 * Pushes new object to the back of the queue, threadsafe version with signaling if
	 * queue is empty
	 *
	 * \param new object to add
	 */
	void pushWithSignal( TObject object )
	{
		mMutex.lock();
		mQueue.push( object );
		mMutex.unlock();
		
		mCondVar.signal();
	}

	/**
	 * Return first object in queue, waiting for signal if queue is empty,
	 * threadsafe version
	 *
	 *	\return first object in queue
	 */
	TObject popWithSignal()
	{
		mMutex.lock();
		if ( mQueue.empty() )
		{
			mCondVar.wait(mMutex);

			mMutex.lock();
		}
		
		TObject retval = mQueue.front();
		mQueue.pop();
		mMutex.unlock();
		return retval;
	}

private:
	
	std::queue< TObject > mQueue;

	/**	
	 * Private copy c-tor to avoid copying of queue
	 */
	SynchronizedQueue( const SynchronizedQueue & q )
	{
		DIE( "Attempt to copy event queue!" );
	}

	Mutex mMutex;
	CondVar mCondVar;
};

#endif // _SYNCHRONIZEDQUEUE_H
