/*
 * SynCountedPtr - sychronized reference counted pointer.
 *
 * The is a non-intrusive implementation that allocates an additional
 * int and pointer for every counted object.
 */

#ifndef SYNCOUNTEDPTR_H
#define SYNCOUNTEDPTR_H

/* For ANSI-challenged compilers, you may want to #define
 * NO_MEMBER_TEMPLATES or explicit */

#define NO_MEMBER_TEMPLATES

#include "mutex.h"

template <class X> class SynCountedPtr
{
public:
    typedef X element_type;

    explicit SynCountedPtr(X* p = 0) // allocate a new counter
        : itsCounter(0) {if (p) itsCounter = new counter(p);}
    ~SynCountedPtr()
        {release();}
    SynCountedPtr(const SynCountedPtr& r) throw()
        {acquire(r.itsCounter);}
    SynCountedPtr& operator=(const SynCountedPtr& r)
    {
        if (this != &r) {
            release();
            acquire(r.itsCounter);
        }
        return *this;
    }

#ifndef NO_MEMBER_TEMPLATES
    template <class Y> friend class SynCountedPtr<Y>;
    template <class Y> SynCountedPtr(const SynCountedPtr<Y>& r) throw()
        {acquire(r.itsCounter);}
    template <class Y> SynCountedPtr& operator=(const SynCountedPtr<Y>& r)
    {
        if (this != &r) {
            release();
            acquire(r.itsCounter);
        }
        return *this;
    }
#endif // NO_MEMBER_TEMPLATES

    X& operator*()  const throw()   {return *itsCounter->ptr;}
    X* operator->() const throw()   {return itsCounter->ptr;}
    X* get()        const throw()   {return itsCounter ? itsCounter->ptr : 0;}
    bool unique()   const throw()
        {return (itsCounter ? itsCounter->count == 1 : true);}

private:

    class counter {
	public:
        counter(X* p = 0, unsigned c = 1) : ptr(p), count(c) {}
		int inc()
		{
			countLock.lock();
			int tmp = ++count;
			countLock.unlock();

			return tmp;
		}

		int dec()
		{
			countLock.lock();
			int tmp = --count;
			countLock.unlock();

			return tmp;
		}

		X*          ptr;

	private:
		Mutex countLock;
        unsigned    count;

    }* itsCounter;

    void acquire(counter* c) throw()
    { // increment the count
        itsCounter = c;
        if (c) c->inc();
    }

    void release()
    { // decrement the count, delete if it is 0
        if (itsCounter) {
            if (itsCounter->dec() == 0) {
                delete itsCounter->ptr;
                delete itsCounter;
            }
            itsCounter = 0;
        }
    }
};

#endif // SYNCOUNTEDPTR_H
