/* MAINTAINER: Bernard */

#ifndef _TASK_H
#define _TASK_H


class WorldEntity;

/**
 * \brief Basic class for all tasks. 
 *
 */
class Task
{
public:	

	Task();

	virtual ~Task() = 0;

	/**
	 * \brief Init resources for task.
	 *
	 * This method is called before update methods.
	 * Call this method if you reimplement it.
	 * It is important you fou want to "reinit" teinit task.
	 */
	virtual void init()
	{
		mLastTime = -1;
	}
	

	/**
	 * \brief Clean resources for task. 
	 *
	 * Update shouldn't be called after resources celanup.
	 * Children doesn't have to call this method when reimplementing.
	 */
	virtual void clean();
	
	/**
	 * \brief Update call with difference from last call.
	 *
	 * Update the task. Note that atime will overflow
	 * in 20 days. dtime will be allways OK.
	 * Children doesn't have to call this method when reimplementing.
	 * This method by default compute time for ticking. If you don't need
	 * tick you may reimplement this.
	 * \param dtime time in ms sice last call. May be == 0. 
	 *              Allways is nonnegative. In first call is == 0.
	 * \param atime absolute time in ms
	 */
	virtual void update(int dtime, int atime);

	/**
	 * \brief Direct update from TaskManager
	 *
	 * Update the task. This method is called directly TaskManager
	 * Note that atime will overflow in 20 days and will start from
	 * zero.
	 * Children doesn't have to call this method when reimplementing
	 * but update(int dtime, int atime) won't work any more.
	 * \param atime absolute time in ms
	 */
	virtual void update(int atime);


	/**
	 * \brief Regullar tick function.
	 *
	 * Tick is called every mTickDelta miliseconds. If the skip between
	 * update calls was greater then mTickDelta, more calls may be performed
	 * in one update call.
	 */
	virtual void tick();	
	

	/**
	 * \brief Notification for task that new object has been added.
	 *
	 * If the task need some fast structire to hold world entity - this is the
	 * notification that new entity is added.
	 * The entity is allreardy added in world when this is called.
	 * Children doesn't have to call this method when reimplementing.
	 * \param we pointer to new entity
	 */
	virtual void addedWorldObject(WorldEntity* we);

	/**
	 * \brief Notification for task that object will be deleted.
	 *
	 * The object is still in World's list of all objects but
	 * after this call the object will be deleted so the pointer
	 * will be invalid. If task holds world entities in some super
	 * structure, this is a good time to remove it.
	 * Children doesn't have to call this method when reimplementing.
	 */
	virtual void deletedWorldObject(WorldEntity* we);

	/**
	 * \brief Notification fot task, that current map will be unloaded
	 */
	// virtual void unloadMap();
	
	/**
	 * \brief Notification for task, that new map was loaded.
	 *
	 * Task may load some aditional info from map or whatever :-)
	 */
	// virtual void loadMap(std::string mapName);
	
protected:
	int    mLastTime; ///< remembers last time of update call. Invalid if == -1
	double mTimeOverflow;	///< Time from last call of update that wasn't used.
	double mTickDelta;		///< Time for one tick. Must be > 0

};

#endif

