#include "terrain_surface.h"
#include "log.h"

#include "SDL.h"
#include "SDL_image.h"
#include "SDL_rotozoom.h"


TerrainSurface::SurfaceLayer::Tile::Tile()
	: mSurface(0)
{}

TerrainSurface::SurfaceLayer::Tile::~Tile()
{
	if (mSurface)
	{
		SDL_FreeSurface(mSurface);
		mSurface = 0;
	}
}

// if newInternalWidth or newInternalHeight is < 1 then use extents from the image itself
counted_ptr<TerrainSurface::SurfaceLayer::Tile> TerrainSurface::SurfaceLayer::Tile::createTile(const std::string & imagePath, int newInternalWidth, int newInternalHeight, double periodX, double periodY)
{
	counted_ptr<TerrainSurface::SurfaceLayer::Tile> p(0);
	
	SDL_Surface * s = IMG_Load(imagePath.c_str());
	if (!s)
	{
		LOGD << "Can't load tile image '" << imagePath << "'!" << std::endl;
		return p;
	}

	assertL(s->w > 0 && s->h > 0 && "Surface width and height should be > 0.");
	
	if (newInternalWidth < 1) 
		newInternalWidth = s->w;

	if (newInternalHeight < 1) 
		newInternalHeight = s->h;
	
	const double 
		zoomx(newInternalWidth == s->w ? 1.0 : (double)newInternalWidth / s->w),
		zoomy(newInternalHeight == s->h ? 1.0 : (double)newInternalHeight / s->h);
	const int smooth = 0;

	if (zoomx != 1.0 || zoomy != 1.0)
	{
		SDL_Surface * zoomedSurf = zoomSurface(s, zoomx, zoomy, smooth);
		SDL_FreeSurface(s);
		if ( !zoomedSurf )
		{
			LOGD << "Can't scale tile surface!" << std::endl;
			return p;
		}
		s = zoomedSurf;
	}

	p = counted_ptr<TerrainSurface::SurfaceLayer::Tile>(new TerrainSurface::SurfaceLayer::Tile());
	p->mSurface = s;
	p->mPeriodX = periodX;
	p->mPeriodY = periodY;

	return p;
}

struct SDL_Surface * TerrainSurface::SurfaceLayer::Tile::getSurface()
{
	return mSurface;
}

double TerrainSurface::SurfaceLayer::Tile::getPeriodX()
{
	return mPeriodX;
}

double TerrainSurface::SurfaceLayer::Tile::getPeriodY()
{
	return mPeriodY;
}

int TerrainSurface::SurfaceLayer::Tile::getAbsoluteWidth()
{
	return mSurface->w;
}

int TerrainSurface::SurfaceLayer::Tile::getAbsoluteHeight()
{
	return mSurface->h;
}


TerrainSurface::SurfaceLayer::SurfaceLayer():
mWeight(0), mColor(0)
{
}

void TerrainSurface::SurfaceLayer::setWeightFunction( TerrainSurface::SurfaceLayer::WeightFunctionPtr fn ) 
{ 
	mWeight = fn; 
}

void TerrainSurface::SurfaceLayer::setColorFunction( TerrainSurface::SurfaceLayer::ColorFunctionPtr fn )
{ 
	mColor = fn; 
}

TerrainSurface::SurfaceLayer::WeightFunctionPtr
TerrainSurface::SurfaceLayer::getWeightFunction()
{
	return mWeight;
}

TerrainSurface::SurfaceLayer::ColorFunctionPtr
TerrainSurface::SurfaceLayer::getColorFunction()
{
	return mColor;
}

TerrainSurface::TerrainSurface()
{
	mLayers.reserve( INITIAL_LAYERS_COLLECTION_CAPACITY );
}

TerrainSurface::LayersContainer::value_type TerrainSurface::createLayer()
{
	TerrainSurface::LayersContainer::value_type tmp( new TerrainSurface::SurfaceLayer() );
	mLayers.push_back( tmp );
	return tmp;
}

TerrainSurface::LayersContainer::size_type TerrainSurface::getLayersCount()
{
	return mLayers.size();
}

TerrainSurface::LayersContainer::value_type TerrainSurface::getLayer(TerrainSurface::LayersContainer::size_type index)
{
	return ( index >= 0 && index < getLayersCount() ) ? mLayers[ index ] : LayersContainer::value_type( 0 );
}

/*
void TerrainSurface::clearLayers()
{
	mLayers.clear();
}

void TerrainSurface::getLayers( LayersContainer::iterator & begin, LayersContainer::iterator & end )
{
	begin = mLayers.begin();
	end = mLayers.end();
}
*/


#include "terrain_base.h"
TerrainSurface::SurfaceLayer::TerrainFunction::TerrainFunction(class TerrainBase * tb):
	mTerrainBase(tb)
{}

void TerrainSurface::SurfaceLayer::TerrainFunction::sampleValues(double x1, double x2, double y, double sampleX, double sampleY, std::vector<value_type>::iterator first)
{
	assertL(mTerrainBase != 0 && "TerrainBase pointer has to be set when sampling height values!");

	// ----------------------------------------------------
	// A. get heightmap extents and calculate scale factors
	// ----------------------------------------------------
	int iHeightMapExtentX(0), iHeightMapExtentY(0);
	mTerrainBase->getExtents(iHeightMapExtentX, iHeightMapExtentY);
	if ( iHeightMapExtentX < 1 || iHeightMapExtentY < 1 ) 
	{
		// TODO: warning message? (small extents)
		return;
	}
	double heightMapExtentX(iHeightMapExtentX), heightMapExtentY(iHeightMapExtentY);
	double 
		dhx = (heightMapExtentX-1) * sampleX, // one x-step in [0,heightMapExtentX-1]x[0,heightMapExtentY-1] heightmap
		dhy = (heightMapExtentY-1) * sampleY; // one y-step

	if (dhx <= 0 || dhy <= 0) 
	{
		// TODO: handle error
		return;
	}

	// -----------
	// B. big loop 
	// -----------
	double hx, hy;	// heightmap x position, heightmap y position ([0,heightMapExtentX-1]x[0,heightMapExtentY-1] heightmap)
	
	hy = y * (heightMapExtentY - 1);
	hx = x1 * (heightMapExtentX - 1);
	for (double x = x1; x <= x2; x += sampleX ) // go through samples
	{
		using namespace std;
		// sample [x, y] from (0,1)x(0,1) corresponds with HeightMap point [hx, hy].
		// Now, we have to compute weighted average over rectangular surround of 
		// point [hx, hy]. The length of x-side is dhx, y-side dhy and position of 
		// the center is [hx, hy].
		
		// rectangle boundary (floating) points
		double 
			hx1(max(hx - dhx/2, 0.0)), 
			hy1(max(hy - dhy/2, 0.0)), 
			hx2(min(hx + dhx/2, heightMapExtentX - 1)), 
			hy2(min(hy + dhy/2, heightMapExtentY - 1));
		
		// rectangle boundary (discrete) points
		int int_hx1((int)(hx1 + 0.5)), int_hy1((int)(hy1 + 0.5)), int_hx2((int)(hx2 + 0.5)), int_hy2((int)(hy2 + 0.5));

		// initialize result to 0
		double avg_value(0);
		
		// loop through all points intersecting the rectangle
		for (int ihy = int_hy1; ihy <= int_hy2; ++ihy)
		{
			for (int ihx = int_hx1; ihx <= int_hx2; ++ihx)
			{
				// For the 1-length square centered to the point [ihx, ihy],
				// calculate the amount of intersection. This will be the weight and will 
				// lay in (0, 1].

				// The amount of intersection:
				double area = 
					( min(ihx + 0.5, hx2) - max(ihx - 0.5, hx1) ) * 
					( min(ihy + 0.5, hy2) - max(ihy - 0.5, hy1) );

				// Normalize to (0, 1] if the point [ihx, ihy] was on the border:
				if (ihx == 0 || ihx == iHeightMapExtentX - 1)
					area *= 2;
				if (ihy == 0 || ihy == iHeightMapExtentY - 1)
					area *= 2;

				// add "area * value" to the result
				avg_value += area * getValueAt(ihx, ihy);
				
			}
		}

		// calculate average values
			
		// dhx * dhy is the area of one 1-length square in heightmap extents?
		double d = 1.0 / (dhx * dhy);
			
		// clamp to [0, 1]
		avg_value = max( min( avg_value * d, 1.0 ), 0.0 );

		// store the result value
		*(first++) = value_type( avg_value );

		hx += dhx;
	}
}

TerrainSurface::SurfaceLayer::HeightFunction::HeightFunction(TerrainBase * tb)
	:TerrainFunction(tb)
{}

double TerrainSurface::SurfaceLayer::HeightFunction::getValueAt(int heightmapX, int heightmapY)
{
	assertL(mTerrainBase != 0 && "TerrainBase pointer has to be set when requesting terrain values!");
	return mTerrainBase->getTrueHeight(heightmapX, heightmapY) * (1.0 / 255.0);
}

TerrainSurface::SurfaceLayer::SlopeFunction::SlopeFunction(TerrainBase * tb)
	:TerrainFunction(tb)
{}

double TerrainSurface::SurfaceLayer::SlopeFunction::getValueAt(int heightmapX, int heightmapY)
{
	assertL(mTerrainBase != 0 && "TerrainBase pointer has to be set when requesting terrain values!");
	return mTerrainBase->slope(heightmapX, heightmapY);
}

TerrainSurface::SurfaceLayer::TerrainBasicWeightFunction::TerrainBasicWeightFunction( TerrainBase * tb )
: heightFunction( tb ), slopeFunction( tb )
{
}

/*
void TerrainSurface::SurfaceLayer::TerrainBasicWeightFunction::addSubFunction( counted_ptr< Function2D< double > > fn )
{
}
*/

void TerrainSurface::SurfaceLayer::TerrainBasicWeightFunction::sampleValues(double x1, double x2, double y, double sampleX, double sampleY, std::vector<value_type>::iterator first)
{
	int nsamples = (sampleX) != 0 ? static_cast<int>(((x2 - x1) / sampleX) + 1.5) : 0;
	if (nsamples <= 0)
		return;

	std::vector<TerrainFunction::value_type> sampledSlope(nsamples), sampledHeight(nsamples);

	heightFunction.sampleValues(x1, x2, y, sampleX, sampleY, sampledHeight.begin());
	slopeFunction.sampleValues(x1, x2, y, sampleX, sampleY, sampledSlope.begin());
	
	assertL(sampledSlope.size() == sampledHeight.size() && "Size of sampled slope values and sampled height values should be equal!");

	for (std::vector<value_type>::size_type i = 0, end = sampledSlope.size(); i < end; ++i)
	{
		// height * slope weight calculation
		*(first++) = min(1.0, value_type((1.0 - sampledHeight[i]) * (1.0 - sampledSlope[i]) ) + 0.4 * (rand() % 2));
	}
}
