#ifndef _TERRAIN_SURFACE_H_
#define _TERRAIN_SURFACE_H_

#include <vector>
#include <string>
#include <list>
#include "counted_ptr.h"
#include "function2D.h"

/**
*	\brief Maintains surface layers.
*
*	Each layer covers entire terrain and consists of weight function, color function and
*	detail function.
*/
class TerrainSurface
{
public:

	/**
	*	\brief One layer of terrain surface.
	*/
	class SurfaceLayer
	{
	public:

		class Tile
		{
		public:
			/** \brief frees internal SDL surface. */
			Tile();
			~Tile();
		
			/**
			*	\brief creates a tile
			*	\param imagePath path to the image
			*	\param newInternalWidth absolute width of internal surface, new surface is scaled from original
			*	\param newInternalHeight absolute height of internal surface, new surface is scaled from original
			*	\param periodX logical horizontal size
			*	\param periodY logical vertical size
			*	\return counted pointer to new tile
			*	\note Consider the surface in normalized coordinates (0.0, 1.0)x(0.0, 1.0). Then
			*	periodX and periodY tells how to scale it horizontally and vertically. This information
			*	is useful when aplying tile pattern as repetitive texture.
			*	
			*/
			static counted_ptr<Tile> createTile(const std::string & imagePath, int newInternalWidth, int newInternalHeight, double periodX, double periodY);
			
			
			struct SDL_Surface * getSurface();
			double getPeriodX();
			double getPeriodY();
			int getAbsoluteWidth();
			int getAbsoluteHeight();

		private:
			SDL_Surface * mSurface;
			double mPeriodX, mPeriodY;
		};

		/**
		*	\brief Helper class for colors.
		*	Consists of r, g, b channels, each channel is from interval [0,1].
		*	\note This is here because 
		*/
		class Rgb
		{
		public:
			Rgb():r(0), g(0), b(0) {}
			Rgb(float _r, float _g, float _b)
				: r(_r), g(_g), b(_b)
			{}

			float r;	///< \brief red channel value in [0, 1]
			float g;	///< \brief green channel value in [0, 1]
			float b;	///< \brief blue channel value in [0, 1]
		};

		/**
		*	\brief 2D function terrain sampler.
		*	Given a TerrainBase pointer, it is able to sample some terrain function defined at
		*	terrain heightmap support. For example height or slope on heightmap discrete points.
		*/
		class TerrainFunction: public Function2D<double>
		{
		public:
			/** \brief ctor */
			TerrainFunction(class TerrainBase * tb);
			
			/**
			*	\brief get value at heightmap discrete point
			*	\param heightmap discrete x coordinate
			*	\param heightmap discrete y coordinate
			*	\return calculated value, should be in [0, 1]
			*/
			virtual double getValueAt(int heightmapX, int heightmapY) = 0;

			/**
			*	\brief sample values at given row and stores them into some preallocated vector
			*	\param x1 start position on row y, in [0,1]
			*	\param x2 end position on row y, in [0,1]
			*	\param y row, in [0,1]
			*	\param sampleX one step in x axis (row)
			*	\param sampleY one step in y axis (column)
			*	\param first where to output first result
			*	\note The preallocated vector (the one the first iterator is taken from) must have
			*		enough entries to store all samples.
			*/
			void sampleValues(double x1, double x2, double y, double sampleX, double sampleY, std::vector<value_type>::iterator first);

		protected:
			class TerrainBase * mTerrainBase;
		};

		/**
		*	\brief terrain height function - provides values of heightmap in [0,1]
		*/
		class HeightFunction: public TerrainFunction
		{
		public:
			HeightFunction(class TerrainBase * tb);
			double getValueAt(int heightmapX, int heightmapY);
		};

		/**
		*	\brief terrain slope function - provides slope values of heightmap in [0,1]
		*/
		class SlopeFunction: public TerrainFunction
		{
		public:
			SlopeFunction(class TerrainBase * tb);
			double getValueAt(int heightmapX, int heightmapY);
		};

		typedef Function2D<double>			WeightFunction;
		typedef Function2D<Rgb>				ColorFunction;

		/**
		*	\brief terrain basic weight function - samples slope * height
		*	If you intend to write your own weight function, good idea is to extend this one
		*	and override the sampleValues method.
		*/
		class TerrainBasicWeightFunction: public WeightFunction
		{
		public:

			TerrainBasicWeightFunction( TerrainBase * tb );
			
			void sampleValues(double x1, double x2, double y, double sampleX, double sampleY, std::vector<value_type>::iterator first);
			
		protected:
			
			HeightFunction	heightFunction;
			SlopeFunction	slopeFunction;
		};

		/**
		*	\brief constant color function
		*	This can be useful when you create some number of layers. You can assign constant color
		*	function and weight function to each one of them and resulting global texture will be
		*	weighted from those layers. This should be sufficient in most cases. When you want more
		*	variability of the terrain, you can implement your own color function.
		*/
		class TerrainConstantColorFunction: public ColorFunction
		{
		public:

			/**
			*	\brief ctor
			*	\param _r red component in [0,1]
			*	\param _g green component in [0,1]
			*	\param _b blue component in [0,1]
			*/
			TerrainConstantColorFunction(double _r, double _g, double _b)
				: r(_r), g(_g), b(_b)
			{}

			/**
			*	\brief samples row with constant color value
			*/
			void sampleValues(double x1, double x2, double y, double sampleX, double sampleY, std::vector<value_type>::iterator first)
			{
				for ( double x = x1; x <= x2; x += sampleX )
					*(first++) = Rgb((float)r, (float)g, (float)b);
			}

		protected:
			double r, g, b;
		};
		
		typedef counted_ptr<WeightFunction>			WeightFunctionPtr;
		typedef counted_ptr<ColorFunction>			ColorFunctionPtr;

		SurfaceLayer();

		/** \brief set the weight function */
		void setWeightFunction( WeightFunctionPtr	fn );
		
		/** \brief set the color function */
		void setColorFunction( ColorFunctionPtr fn );

		void setDetailTile(counted_ptr<Tile> tile) { mTile = tile; }
		counted_ptr<Tile> getDetailTile() { return mTile; }

		/** \brief get the weight function */
		WeightFunctionPtr getWeightFunction();

		/** \brief get the color function */
		ColorFunctionPtr getColorFunction();

	private:		
		WeightFunctionPtr			mWeight;	///< \brief global weight function
		ColorFunctionPtr			mColor;		///< \brief global color function
		counted_ptr<Tile>			mTile;
	};

	typedef std::vector< counted_ptr<SurfaceLayer> > LayersContainer;

	enum { INITIAL_LAYERS_COLLECTION_CAPACITY = 8 };

	/**
	*	\brief ctor - reserves initial size of collection for INITIAL_LAYERS_COLLECTION_SIZE element
	*/
	TerrainSurface();

	/**
	*	\brief Add a new layer to the end of collection.
	*
	*	A new layer is created and pointer to it is returned. Then you can set its content.
	*	By default, all layer functions are set to null.
	*
	*	\return layer pointer (counted_ptr)
	*/
	LayersContainer::value_type createLayer();
	
	/**
	*	\brief get the layer count
	*	\return layers count
	*/
	LayersContainer::size_type getLayersCount();

	/**
	*	\brief get the layer at specified position in collection
	*	\param index layer index (from 0 to getLayersCount() - 1)
	*	\return layer pointer on success, null LayersContainer::value_type(0) if index is out of bounds
	*/
	LayersContainer::value_type getLayer(LayersContainer::size_type index);

	// TODO: useful?
	//void clearLayers();
	//void getLayers( LayersContainer::iterator & begin, LayersContainer::iterator & end );

private:

	/** \brief Container for all surface layers. */
	LayersContainer mLayers;
};

#endif // _TERRAIN_SURFACE_H_

