/* MAINTAINER: Bernard */

#ifndef _THREAD_H
#define _THREAD_H

#include "mutex.h"

struct SDL_Thread;

/**
 * \brief Basic class for all threads in game.
 *
 * The program is killed if any failures with threads.
 * Usage: If you want your own thread just create class
 * class MyThread : public Thread
 * {
 *   ...
 *   virtual int run();  // And add implemetation :-)
 *   ...
 * }
 *
 * NOTE that you should test mStop in run() and stop if
 * is true;
 *
 * In program:
 *
 * MyThread t1;
 * t1.start();   // the thread is running
 * ...
 * t1.wait();  // waiting until thread end.
 * 
 * 
 *  \note One instance may be running only once at time.
 */
class Thread
{
public:
	Thread();
	virtual ~Thread();

	/**
	 * \brief Start execution of thread.
	 *
	 * Create new thread and new thread calls void run().
	 * Also set stop flag to false.
	 * The program is killed if called on running thread or
	 * not joined thread even if allready ended.
	 */
	void start();

	/**
	 * \brief Code of thread.
	 *
	 * This function must be implemented by childrens.
	 * \return whatever you want to return :-)
	 */
	virtual int run() = 0;

	/**
	 * \brief Wait until end of thread execution.
	 * 
	 * This requires that the thread end when mStop is set to true.
	 * This is a blocking call. The thread is not running after
	 * return of this call. setStop(true) is called inside.
	 * The program is killed if called on not running thread.
	 * \return value returned by run();
	 */
	int wait();

	/**
	 * \brief Set stop flag.
	 * 
	 * The thread should then end. But the thread may still be running
	 * after return of this call. The call doesn't wait until thread end.
	 * \param stop set true if you want to tell the thread to end.
	 */
	void setStop(bool stop = true);

	/**
	 * \brief test if the thread is running
	 *
	 * Thread safe access to mRunning.
	 * \return true if the thread is running
	 */
	bool isRunning();

	/**
	 * \brief Return the stop flag.
	 *
	 * Thread safe access to mStop.
	 * \return mStop
	 */
	bool getStop();


	/**
	 * \brief Test if the thread started and didn't joint yet.
	 *
	 * The call return true even if the thread is not running but
	 * wasn't joined.
	 * \return true is wait (may) be called - without waiting.
	 */
	bool needWait();


	/**
	 * \brief Give CPU back to OS so the task may reschedule.
	 *
	 * This is good to be called at end of loop of thread.
	 * The running of server may be smoother.
	 *
	 * while (my_thread)
	 * {
	 *    ... // One loop
	 *    Thread::reshedule();
	 * }
	 *
	 */
	static void reshedule();

protected:

	/**
	 * \brief Set mRunning
	 *
	 * Thread safe version.
	 * \param run new value.
	 */
	void setRunning(bool run);

	bool   mStop;       ///< \brief true if someone wanted the thread to stop.
	Mutex  mStopMutex;  ///< \brief mutex for mStop.
	
	bool   mRunning;       ///< \brief true if the thread is running.
	Mutex  mRunningMutex;  ///< \brief mutex for mRunning.

	/**
	 * \brief Static function that is used for passing
	 *        to thread creating functions.
	 *
	 * \param  threadClass Pointer to Thread that created new thread.
	 * \return zero
	 */
	static int run(void *threadClass);

	SDL_Thread*  mThread; ///< \brief pointer to SDL thread. != 0 if the thread is running.
	Mutex        mThreadMutex; ///< \brief mutex for mThread.
};


#endif // _THREAD_H
