/* MAINTAINER: Bernard */

#ifndef _TIME_MANAGER_H
#define _TIME_MANAGER_H

#include "mutex.h"

/**
 * \brief Managing global time.
 * 
 * More less olny working with SDL_GetTicks and providing thread safe
 * access to it for whole application. Also measure "game time" where
 * game is running and may have different time from real time - like
 * twice faster...
 */
class TimeManager : public Mutex
{
public:
	TimeManager();
	~TimeManager();

    /**
	 * \brief Return ABSOLUTE time.
	 * 
	 * Note that it may overflow. This is thread save routine.
	 * \return ticks - some system value. Change every ms.
	 */
	int  getTicks();


	/**
	 * \brief return GAME time.
	 *
	 * Note that it may overflow - but not very probable This is thread save routine.
	 * It is modified by game speed.
	 * \return game time. Change every ms.
	 */
	int getGameTicks();

	/**
	 * \brief Set current game ticks
	 * 
	 * This is used only when some serious performance
	 * problems are encountered. Don't use if you are not sure what
	 * you are doing.
	 */
	 void setGameTicks(int new_game_ticks);

	/**
	 * \brief suspend current thread for some time
	 *
	 * \param ms sleep time in ms
	 */
	static void sleep(int ms);

	/*   -- NOT WORKING GOOD WITH NEW mPauseEnabled
	 * \brief Suspend current thread for some time.
	 *
	 * The time is given in game speed time. The sleep
	 * time is computed depending on current game speed.
	 * If game speed is changed when something is sleeping,
	 * there will be no efect to sleep time of allready sleeping.
	 */
//	void sleepGame(int game_ticks);

	/**
	 * \brief Set game speed.
	 * 
	 * 0.5 means half game speed and 2.0 means double
	 * game speed.
	 * Thread safe.
	 * \param speed new game speed
	 */
	void  setGameSpeed(double speed);
	
	/**
	 * \brief Get current game speed.
	 * 
	 * Return zero if game is paused so dont use in division :-)
	 * Thread safe.
	 * \return current game speed
	 */
	double getGameSpeed();

	/**
	 * \brief Get pause status
	 * 
	 * Thread safe.
	 * \return true if game is currently paused.
	 */
	bool isPaused();
	
	/**
	 * \brief Set pause to some state
	 * 
	 * Thread safe.
	 * \param state new pause state
	 */
	void pause(bool state);
	
	/**
	 * \brief Change pause state.
	 * 
	 * Same as pause(!isPaused()) but thread safe.
	 */
	void pauseSwitch();

	/**
	 * \brief Enable pause.
	 * 
	 * In same stages of game - like map reload pasue
	 * in engine causes some serious problems. So when reloading
	 * map or quitting you may disable pause.
	 */
	void enablePause(bool enabled); 

	/**
	 * \brief Set current simlulated time.
	 * 
	 * Don't call this function unelss you know what you are doing :)
	 * See getSimulatedTime() for explanation.
	 * \param newTime new time stamp
	 */
	void setSimulatedTime(int newTime) { mSimulatedTime = newTime; }
	
	/**
	 * \brief So far simulated time.
	 * 
	 * The simulated time is valid only on server part of game.
	 * The game works in ticks and every tick has some "timestamp"
	 * and this is the last (current) time stamp. The timestamps
	 * are created in tasksimluation.
	 * \return Current timestamp
	 */
	int  getSimulatedTime() { return mSimulatedTime; }

private:

     /**
	  * \brief Modification of game speed.
	  *
	  * The speed is double number where 1.0 means normal speed,
	  * 0.5 means half speed and 2.0 means double.
	  * Inicialized in GameServer::GameServer()
	  */
	double mGameSpeed;

	double mGameTime; ///< \brief Current game time. Updated in getGameTicks call.

	int    mLastTick; ///< \brief las time of mGameTime update.

	bool   mPause; ///< \brief True if game is paused.

	Mutex  mPauseMutex; ///< \brief Mutex used for locking the object.

	int    mSimulatedTime; ///< \brief So far simulated time. It isn't modified internally.
	
	bool   mPauseEnabled;  ///< \brief true if it is possible to pause game
};

extern TimeManager gTime; ///< \brief Global variable providing access to time for whole app.

#endif
