/* MAINTAINER: Bernard */

// #define WANT_TRACE

#include <cstdlib>
#include "translator.h"

#ifndef __SEPARATED__TRANSLATOR__
	#include "tinyxml/tinyxml.h"
	#include "log.h"
	#include "conversions.h"
#else
	#define TIN             
	#define TVARIABLE(v1)  
	#define TVARIABLE2(v1, v2)  
	#define TVARIABLE3(v1, v2, v3)  
	#define TPOINT          
	#define TPOINTS(s)      
	#define LOGT            LOGN
	#define assertL(x)   assert(x)
#endif


using namespace STL_EXT_NS;
using namespace std;


// Global variable with translator
Translator g_translator;

Translator::Translator()
{
	
}

Translator::~Translator()
{

}


const char* Translator::translate(const char *s)
{
	assertL(s != 0);

	hash_map<string, st_translation>::iterator it = mTTable.find(s);
	if (it == mTTable.end()) return s;
	
	return it->second.utf8.c_str();
}
	
wstring Translator::translateW(const char *s)
{
	assertL(s != 0);

	hash_map<string, st_translation>::iterator it = mTTable.find(s);
	if (it == mTTable.end()) 
	{
		wstring ws;

		utf8tows(s, ws);
		
		return ws;
	}
	
	return it->second.ws;
}

bool Translator::loadFile(const char *filePath)
{
	TIN
	
	assertL(filePath != 0);
	
	TiXmlDocument doc(filePath);
	TiXmlHandle   handle(&doc);


	if (!doc.LoadFile()) 
	{
		LOGE << "Unable to load " << filePath 
		     <<" Probably file not exist or isn't a valid XML file" << endl;
		return false;
	}	
	

	// Forget previous translation table
	mTTable.clear();

	// Search whole documents for translation pairs and add them to 
	// translation table
	TiXmlElement* child = handle.FirstChild( "translation" ).FirstChild( "string" ).Element();
	for (; child; child=child->NextSiblingElement()) 
	{
		const char *key   = child->Attribute("original");
		const char *value = child->Attribute("translated");
		if (key != 0 && value != 0) 
		{
			st_translation st;
			st.utf8 = value;
			utf8tows(value, st.ws); // make also wstring version of translated string
			
			mTTable.insert(make_pair(key, st));

			LOGT << "Loaded translation '" << key << "' =  '" << value << "'" << endl;
		}
	}
	
	// Making ALL translations in cache invalid	
	for (int i = 0; (unsigned int)i < mCache.size(); ++i)
	{
		mCache[i].translated_valid = false;
	}
	
	return true;
}


int  Translator::cache(const char *s)
{
	assertL(s != 0);
	
	// find free place or create a new one
	int id = -1;
	for (int i = 0; i < (int)mCache.size(); ++i)
	{
		if (mCache[i].valid == false) 
		{
			id = i;
			break;
		}
	}
	if (id == -1) // create new if none found
	{
		st_cache c;
		id = (int)mCache.size();
		mCache.push_back(c);
	}
	
	// update at index id
	mCache[id].set(s);

	return id;
}
	

const char* Translator::get(int ID)
{
	if ( ID < 0 || ID > (int)mCache.size() || mCache[ID].valid == false) 
	{
		LOGE << "Invalid ID" << endl;
		return "INVALID ID!!!";
	}

	// translating if not translated yet
	if ( mCache[ID].translated_valid == false )
	{
		mCache[ID].translated_valid = true;
		mCache[ID].translated.utf8 = translate(mCache[ID].original.c_str());
		mCache[ID].translated.ws  = translateW(mCache[ID].original.c_str());
	}
	
	return mCache[ID].translated.utf8.c_str();
}
		

const wchar_t* Translator::getW(int ID)
{
	if ( ID < 0 || ID > (int)mCache.size() || mCache[ID].valid == false) 
	{
		LOGE << "Invalid ID" << endl;
		return L"INVALID ID!!!";
	}

	// translating if not translated yet
	if ( mCache[ID].translated_valid == false )
	{
		mCache[ID].translated_valid = true;
		mCache[ID].translated.utf8 = translate(mCache[ID].original.c_str());
		mCache[ID].translated.ws  = translateW(mCache[ID].original.c_str());
	}
	
	return mCache[ID].translated.ws.c_str();
}


void Translator::uncache(int ID)
{
	if ( ID < 0 || ID > (int)mCache.size() || mCache[ID].valid == false) 
	{
		LOGE << "Invalid ID" << endl;
		return;
	}

	mCache[ID].valid = false;
	
	// remove unused places from cache - to free some memory
	while (ID == (int)mCache.size()-1 && ID >= 0 && mCache[ID].valid==false) 
	{
		mCache.pop_back();
		--ID;
	}
	
}
	

void Translator::debugPrintChache()
{
	cout << "--- Printing translation cache ---" << endl;
	
	for (unsigned int i = 0; i < mCache.size(); ++i)
	{
		cout << i << " ";
		if (!mCache[i].valid) 
		{
			cout << "--free--";
		}
		else 
		{
			cout << mCache[i].original << " *** ";
			if (mCache[i].translated_valid) cout << mCache[i].translated.utf8;
			else cout << "NOT TRANSLATED";
		}
		cout << endl;
	}

	cout << "------" << endl;
}

void Translator::loadTranslationsList(const char *filePath)
{	
	assertL(filePath != 0);
	
	TiXmlDocument doc(filePath);
	TiXmlHandle   handle(&doc);


	if (!doc.LoadFile()) 
	{
		LOGE << "Unable to load " << filePath 
		     <<" Probably file not exist or isn't a valid XML file" << endl;
		return;
	}	
	
	// Forget previous translatins list
	mTranslations.clear();

	// Search whole documents for translation pairs and add them to 
	// translation table
	TiXmlElement* child = handle.FirstChild( "languages" ).FirstChild( "lang" ).Element();
	for (; child; child=child->NextSiblingElement()) 
	{
		const char *langfile   = child->Attribute("file");
		if (langfile != 0) 
		{
			mTranslations.push_back(langfile);
		}
	}
}


