/* MAINTAINER: Bernard */

#ifndef _TRANSLATOR_H
#define _TRANSLATOR_H

#include <string>
#include "stlextdefs.h"
#include "hash_hack.h"

#include <vector>

/**
 * \brief Class for translating strings.
 *
 * Also provide cache with lazy translation.
 * Don't use directly this class (if you don't have 
 * a really good reason) Use macros defined below. They are also
 * used to locate in code strings that need translation.
 */
class Translator
{
public:
	Translator();
	~Translator();

	/**
	 * \brief Translate one string - UTF8
	 *
	 * \param s original string that you want to translate
	 * \return translated string or s if no translation 
	 *         is available
	 */
	const char* translate(const char *s);
	
	/**
	 * \brief Translate one string - WS
	 * 
	 * SORRY! Can't return wchar_t :-( Because you
	 * may ask to translate string that isn't in translation 
	 * table and then where should be the string const wchar_t* allocated?
	 * Who will free it?
	 * Or it may be done if demanded - it will be included in cache.
	 * \param s original string
	 * \return translated string or s converted to wstring
	 *         if no translation available
	 */
	std::wstring translateW(const char *s);

	/**
	 * \brief Load language file.
	 *
	 * Doesn't clear cache - only make translations in chace
	 * invalid. All previsously returned const char* are also
	 * invalid now. (be carefull not to get segfault)
	 */
	bool loadFile(const char *filePath);

	/**
	 * \brief Translate and add to cache. 
	 *
	 * Note that this only create a new space in cache table.
	 * Real translation will be perofred in first get call
	 * - lazy evaluation :-)
	 * \param s original (probably english) string
	 * \return ID to cache. It may be used in calls get, getW
	 *            and uncache
	 */
	int cache(const char *s);
	
	/**
	 * \brief Return translated string from cache
	 *
	 * Note that first call to get and firts call to get
	 * after language change will invoke searching in hash
	 * table and "real" translating".
	 * \param ID number in cache
	 * \return translated string
	 */
	const char *get(int ID);
	
	/**
	 * \brief Return translated string from cache
	 *
	 * Note that first call to get and firts call to get
	 * after language change will invoke searching in hash
	 * table and "real" translating".
	 * \param ID number in cache
	 * \return translated string - something like utf2wchar_t(get())
	 */	
	const wchar_t *getW(int ID);
	
	/**
	 * \brief clear ID from cache.
	 *
	 * The ID will not be valid after this call. 
	 * \param ID removed chached piece
	 */
	void uncache(int ID);
	

	/**
	 * \brief Print all cahce to stdout.
	 *
	 * This is only for cache tuning purposes.
	 */
	void debugPrintChache();

	/**
	 * \breif Get list of found translations.
	 * 
	 * The listconsists of filenames of translations.
	 */
	std::vector<std::string>& getTranslations() { return mTranslations; }
	
	/**
	 * \brief Load mTranslations
	 * 
	 * You don't need to init translator if you don't need to use
	 * getTranslations()
	 */
	void loadTranslationsList(const char *filePath);
		
private:


	std::vector<std::string> mTranslations; // List of all translations (filenames)
	
	/**
	 * \brief Structure with translated strings.
	 */
	typedef struct st_translation
	{
		std::string  utf8; ///< Original version - loaded from file
		std::wstring ws;   ///< wchar_t version converted from utf8
	} st_translation;
	
	/**
	 * \brief Structure for caching translated strings.
	 */
	typedef struct st_cache
	{
	public:
		/**
		 * \brief Default invalidating constructor
		 */
		st_cache()
		{
			valid    = false;
			translated_valid = false;
		}
		
		/**
		 * \brief Set original and make valid
		 */
		void set(const char* orig)
		{
			valid = true;
			translated_valid = false;
			original = orig;
		}
		
		std::string    original;    ///< Original english string
		st_translation translated;  ///< Translated version of original
		bool           valid;            ///< Is this structure valid?
		bool           translated_valid; ///< Is translated valid?
	} st_cache;
	
	
	STL_EXT_NS::hash_map<std::string, st_translation> mTTable; ///< Tranlsation table

	std::vector<st_cache> mCache; ///< Cache with translated strings
	
};

extern Translator g_translator;

// Macros for easier usage

#define _(string_to_translate)       g_translator.translate(string_to_translate)
#define LNG_CACHE_(string_to_cache)  g_translator.cache(string_to_cache)
#define LNG_GET(id_from_cache)       g_translator.get(id_from_cache)
#define LNG_UNCACHE(id_from_cache)   g_translator.uncache(id_from_cache)
#define W_(x)                        g_translator.translateW(x)
#define LNG_GETW(id_from_cahce)      g_translator.getW(id_from_cahce)

#endif
