#ifndef _UNIVERSAL_MANAGER_H
#define _UNIVERSAL_MANAGER_H

#include "singleton.h"
#include "refcounter.h"

#include "hash_hack.h"
#include "stlextdefs.h"

#include <string>


/**
*	\brief base class for various manager implementation.
*
*	usage:
*	\code
*
*	class MyManager: public UniversalManager<MyData, MyManager>
*	{
*	// A. Singleton DP stuffs:
*	protected:
*			MyManager(){}
*			~MyManager(){}
*			friend class Singleton<MyManager>;
*
*	// B. User's add(...) methods
*	public:
*		TRef add(std::string key, int param1)
*		{
*			// I. FIND
*			ManagedItem * p;
*			if ((p = find(key)) != 0)
*				return p;
*
*			// II. INIT
*			MyData * pc = new MyData(param1);
*
*			// III. REGISTER
*			return createAndRegister(key, pc);
*		}
*
*	// C. User's create(), destroy() and destroyContent() callback-methods overriding
*	// ...
*
*	}
*	\endcode
*	
*		Derive your own manager from UniversalManager. The first template argument of UniversalManager is the
*	desired managed type, the second *MUST* be the class name of the top-level derived manager due to 
*	Singleton stuffs (this "throws polymorphysm out of window", but extensibility is what we need).
*		Define protected (or private) ctor and dtor as well. This is due to Singleton DP and ensures that
*	the creation and destruction of the manager will be handled by the Singleton-class.
*		Define your own add() methods. There you can declare parameters that are needed
*	for adding new items (i.e. the file-name string and the font size int if you'd like to manage fonts).
*	Each add-method should implement three parts - FIND, INIT and REGISTER.
*	FIND-part must test if the item you are trying to add is already registered or not and return the item, 
*	if it is already registered. This can be easily done with the find(key_string) method. If the item is not 
*	registered, continue to the INIT-part, where you create the item you want to register and initialize it 
*	from arguments of add-method. Last thing to do is the registration of the item. Use the 
*	createAndRegister() method.
*		You can redefine create(), destroy() and destroyContent() callback methods if you need to. 
*	But, in common case, this shouldn't be needed.
*
*	\see remarks on RefCounter (static variables issues)
*/
template <class T, class ParentClass > // Kurna, tady nekde by mely byt traits. Ale co to vubec je zac?
class UniversalManager: public Singleton< ParentClass >
{
public:
	/**
	*	\brief a class wrapping some managed-information for UniversalManager
	*/
	class ManagedItem: public RefCounter
	{
	public:
		/**
		*	\brief initializes this wrapper of T with key (unique name used for registration) and ptr to data
		*
		*	Sets managed-state mManaged to true (manager is the owner of this instance).
		*/
		ManagedItem(std::string key, T * itemPtr):
		  mRegKey(key), mItemPtr(itemPtr), mManaged(true)
		{}
		
		virtual ~ManagedItem() {}

		/**
		*	\brief reference access
		*/
		operator T*(void) const
		{ return mItemPtr; }

		/** 
		*	\brief type-cast operator to type T
		*/
		operator T&(void) const
		{ return *mItemPtr; }

		/**
		*	\brief key used by manager
		*	\retval unique identifier used by manager for registering this item
		*/
		std::string getKey(void)
		{ return mRegKey; }

		/**
		*	\brief data pointer
		*	\retval value of data pointer that this item wraps
		*/
		T* getPtr(void)
		{ return mItemPtr; }

		/**
		*	\brief managed-status of this item
		*	\retval true, if the item is managed by manager, false otherwise.
		*/
		bool isManaged(void)
		{ return mManaged; }

		/**
		*	\brief disables managing of this item and becomes a "zombie"
		*
		*	If the item is managed, managing is disabled, content (internal data pointer mItemPtr ) is
		*	destroyed by the manager and set to null.
		*/
		void disableManaging() 
		{ 
			if (isManaged()){
				mManaged = false;
				UniversalManager<T, ParentClass>::getInstance()->destroyContent(getPtr());
				mItemPtr = 0;
			}
		}

	protected:
	
		/**
		*	\brief sends the "ZeroRefCount-event" to the appropriate specialized manager
		*
		*	It redefines the behavior of RefCounter::OnZeroRefCount (parent's ZeroRefCount-event handler).
		*	This redefined handler fires a new event for the appropriate UniversalManager (if it is managed) and 
		*	gives a reference to itself to the manager. The concrete specialized manager is the one who knows 
		*	what to do, when ZeroRefCount-event has occured. The specialized manager should do all the 
		*	"destroy-actions" related to freeing the resources allocated with this managed item.
		*
		*	\note	Manager is notyfied only when state mManaged is true (default true), but manager 
		*		can set it to false. This is the case when the manager is destroyed, but there still exist 
		*		some references to managed items. The manager can disable those items this way,
		*		remove them from its registry (but doesn't destroy them) and log some warning message. 		
		*
		*/
		virtual void OnZeroRefCount()
		{
			if (isManaged()){
				UniversalManager<T, ParentClass>::getPtr()->OnZeroRefCount(this);
			}else{
				if (isHeapAllocated()){
					delete this;
					return;
				}
			}
		}

	private:
		std::string mRegKey;	///< \brief key used in registration (key)
		T *			mItemPtr;	///< \brief pointer to the managed item
		bool		mManaged;	///< \brief true = owned by manager, false = owned by itself
	};

	/**
	*	\brief returns the number of items in registry
	*	\return number of registered items
	*/
	int size(void)
	{ return mRegister.size(); }

public:
	typedef RefCounterPtr<ManagedItem, T> TRef; ///< \brief reference to managed type T
	typedef STL_EXT_NS::hash_map<std::string, ManagedItem *> TContainer; ///< \brief containter of managed items
	typedef std::pair <std::string, ManagedItem *> TPair;  ///< (key, data)-pair used for inserting into container of managed items
	typedef UniversalManager<T, ParentClass> TUniversalManager; ///< self-type of this manager

	/**
	*	\brief destroys all unreferenced registered elements and clears the registry
	*
	*	However, if there exists some managed items with non-zero reference count, they are NOT deleted,
	*	but some warning message is logged for each item and this item becomes a zombie.
	*	Zombie means, that data wrapped by this item is destroyed by manager and manager removes this item from 
	*	registry and gives up ownership (zombie owns itself). Reference counting is still active and when it 
	*	counts zero, zombie deletes itself.
	*	
	*	This is used when the manager is being destroyed ( ~UniversalManager() ) or when there is the need
	*	of releasing registered items (for example, the graphical subsystem shutdown is the good moment for
	*	releasing textures - after that it wouldn't been possible).
	*
	*	\note Strongly thread-unsafe! ;-)
	*/		
	void erase(void);

protected:
	UniversalManager(){}	///< \brief protected ctor (singleton)
	
	/*
	*	\brief protected dtor (singleton), destroys all registered items
	*
	*	The erase() is called.
	*	Objects that contains references to managed items (obtained from manager) should be destroyed before
	*	the manager. If they are not, it would be possible to access already destroyed managed items. So, when 
	*	the manager is being destroyed and there are some referenced managed items, they are disabled (they will 
	*	lost their owner and don't notify about zero-ref-count), unregistered and some warning is logged.
	*
	*	\see ManagedItem::disableManaging()
	*/
	~UniversalManager(){ erase(); }

	/**
	*	\brief callback for creating new managed item from its key 
	*	
	*	This is used by manager when registering the item by createAndRegister()
	*
	*	\param key	unique identifier, under its name the item will be registered
	*	\param data	pointer to data that the new instance of ManagedItem will contain
	*/
	virtual ManagedItem* create(std::string key, T *data)
	{ assertL(data && "Null data pointer!"); return new ManagedItem(key, data); }
	
	
	/**
	*	\brief callback for destroying the data that ManagedItem wraps
	*	
	*	It is called by ManagedData::disableManaging() - data is filled by this instance	
	*	\param data pointer to data
	*	\note Use partial specialization, if you need your own destroy-routine.
	*/
	void destroyContent(T* data)
	{ assertL(data && "Null data pointer!"); delete data; }
	
	/**
	*	\brief destroys given managed item 
	*
	*	This is NOT the unregistering itself, but only the destruction routine of managed item and its data.
	*	Unregistering is done automatically by the OnZeroRefCount() callback. However, by overriding this method
	*	you are able to write your own destruction code.
	*
	*	\param item pointer to managed item to destroy	
	*/	
	void destroy(ManagedItem* item)
	{ assertL(item && "Null item pointer!"); item->disableManaging(); delete item; }

	/**
	*	\brief searches for managed item that is registered with the given key and returns null if not found
	*	
	*	This is a supportive method for writing user's own 
	*	<code>add(key [, params])</code> methods.
	*
	*	\param key	unique identifier, under its name the item will be registered
	*	\return item pointer on success, null if not found
	*/
	ManagedItem * find(std::string key);

	/**
	*	\brief creates and registers new item
	*	
	*	This assumes that the item isn't registered yet. This is supportive method for writing user's
	*	own <code>TRef add(key [, params])</code> methods.
	*
	*	\param key	unique identifier, under its name the item will be registered
	*	\param data	pointer to data that the new instance of ManagedItem will contain
	*/
	ManagedItem * createAndRegister(std::string key, T* data);

	TContainer mRegister;	///< \brief register of managed items

private:

	/*
	*	\brief callback called by the friend-class.
	*	
	*	Removes texture from the store and frees memory associated with the texture.
	*
	*	\param pt pointer to the texture that is being released
	*/
	void OnZeroRefCount(ManagedItem * p);
};

template <class T, class ParentClass>
void UniversalManager<T, ParentClass>::OnZeroRefCount(ManagedItem * p)
{
	TIN;
	assertL(p && "Null pointer!");

	// Try to find the key in register.
	typename TContainer::iterator
		it = mRegister.find(p->getKey());

	// If the item exists, unregister and delete it.
	if ((it != mRegister.end()) && (it->second == p))
	{
		mRegister.erase(it);
		destroy(p);
		p = 0;
	}else{
		LOGW << "Trying to release item that isn't registered." << std::endl;
	}
}

template <class T, class ParentClass>
typename UniversalManager<T, ParentClass>::ManagedItem * UniversalManager<T, ParentClass>::createAndRegister(std::string key, T* data)
{
	assert(data && "Null data pointer!");
	ManagedItem * p = create(key, data);
	mRegister.insert(TPair(key, p));
	return p;
}

template <class T, class ParentClass>
typename UniversalManager<T, ParentClass>::ManagedItem * UniversalManager<T, ParentClass>::find(std::string key)
{
	// Try to find the key of the item in the register mRegister.
	typename TContainer::iterator
		it = mRegister.find(key);

	// If the item already exists in the register, return its ManagedItem-pointer, otherwise return null.
	return (it != mRegister.end()) ? it->second : 0;
}

template <class T, class ParentClass>
void UniversalManager<T, ParentClass>::erase(void)
{
	TIN;
	for (typename TContainer::iterator it = mRegister.begin(), ite = mRegister.end(); it != ite; ++it)
	{
		ManagedItem * pmi = it->second;
		// If there doesn't exist any references to this managed item, destroy it. 
		// Otherwise log warning and disable the item.
		if (!(pmi->getRefCount())){
			destroy(pmi);
		}else{
			LOGW << "There still exist " << pmi->getRefCount() << " references to the item '" << pmi->getKey() << "'!" << std::endl;
			pmi->disableManaging();
			// Since now, pmi owns itself. It is a zombie and it will destroy 
			// itself when its reference count reaches zero.
		}
	}
	mRegister.clear();
}

#endif // _UNIVERSAL_MANAGER_H

