#ifndef _AI2_H
#define _AI2_H

#include "GLVector.h"
#include "ai.h"
#include <list>
#include <vector>

#include "log.h"
#include "application_server.h"
#include "game_server.h"
#include "commander.h"
#include "tank.h"
#include "world.h"
#include "bonusbox.h"
#include "aimanager.h"
#include "gameplaysettings.h"
#include "terrain.h"
#include "landscape.h"
#include <algorithm>

#define AI_INVALID_NODE_PTR 0

// defines how dense wil be pathfinding nodes placet on map
#define AI_NODE_DENSITY 10

// defines how far from obstacle will node be still inpassable (in units of AI_NODE_DENSITY) (should be 1.1 +-)
// avoiding scenerio that obstacle is somewhere between nodes and thus tank may hit it passing between
#define AI_TREE_AVOIDANCE_CONST 2

// if if Z distance between pathfinding nodes is AI_INPASSABLE_PITCH or more then XY distance it is considered inpassable
#define AI_INPASSABLE_PITCH 0.5

#define AI_INVALID_PATH 1

// if less then 1/AI_HEAP_RESIZE_MIN_COEF of allocated items occupied in heap and 
// AI_heap::resize() is called, resize() shrinks heap
#define AI_HEAP_RESIZE_MIN_COEF 4

// if closer then this distance from target it is considered as if it has been reached
#define AI_UNINPORTANT_DISTANCE 15

// in degrees (0 to 180) max angle at which AI dont turn left or right but continues onward
#define AI_UNINPORTANT_ANGLE 1.5

// time counted in number of updates of AI which tank sleeps after reaching destination (debugging)
// MUST BE >= 1 { new path generated at 1 ( will be generated on purpose later )}
#define AI_SLEEP_CONSTANT 1500

// some magic
#define AI_MAGIC 2

// do not change
#define AI_MAGIC_2 3

// time in seconds when to slow down before reaching breakpoint (pathfinding node)
#define AI_SLOW_APROACH_CONST 0.20

// angle at which ( or highier ) tank slows/stops to turn
#define AI_SLOW_TURNING_ANGLE 20



class World;
class Landscape;
class Terrain;
class BonusBox;
class AIManager;
class Tank;
class AI;

/**
 * \brief Node for pathfinfing purposes
 */

struct PathNode
{
	GLVector3d mPosition;
	int mX, mY;	// position inside pathfindingNodes 2d array
	bool mPassable;
	// PathNode *neighbours[8]; // squares -> 8 neighbours


	// dijkstra needed vars
	
	// todistance -> from begining to this node, fromdistance -> from this to end node
	double mToDistance, mFromDistance, mFullDistance;
	
	bool mClosed;
	PathNode *mPathPredecessor;

	// heap needed items
	int mHeapIndex;	// index to position in heap (only correct if not closed)
};

/**
 * \brief Heap for pathfinding modified dijkstra algorithm
 */

class AI_heap
{
public:
	AI_heap() { };
	~AI_heap() { };

	// heap insert
	int insert(PathNode * newitem, PathNode * new_pathPredecessor);

	// heap decrease_key
	// does not update key value!!! only updates heap after key change
	int decrease_key(PathNode *node, PathNode * new_pathPredecessor);

	// returns poniter to min node and deletes it from heap
	PathNode * extract_min();

	bool notempty();
private:

	std::vector < PathNode * > mHeapArray;

};


/**
 * \brief Basic AI class with added pathfinding methods
 */


class AI2 : public AI
{
public:
	AI2() {};
	AI2(int playerID, AIManager* manager);
	virtual ~AI2() { };
	
	/**
	 * \brief Here be the magic. Somehow compute commands for tank.
	 * 
	 * This ensure that controled player has tank and if so calls updateTank.
	 * Else it add command for creating new tank.
	 * It is SURE that mPlayerID is valid when this is called
	 */
	virtual void update();
	
	virtual void updateTank(Tank *tank);


	/**
	 * \brief Get nearest enemy.
	 *
	 * \return entiryID or INVALID_ENTITY_ID on error (like no enemy)
	 */
	int getNearestEnemy(double max_distance = 1000000000);

	int getNearestBonus( int type, double max_distance = 1000000000);

	int getNearestLandEntity(const GLVector3d &position, double max_distance = 1000000000);	



	virtual int update_neighbour(PathNode * ofwhich, PathNode * end, AI_heap &heap, const int addx, const int addy);


	// takes GLVector but uses only x and y axis, z is rewriten by terrain height
	virtual int moveclosetopoint(GLVector3d &whereto);
	virtual int movetopoint(GLVector3d &whereto);

	// in adition to movetopoint it ads to existing path so AI moves first to original position
	// and only then to selected point
	virtual int andmovetopoint(GLVector3d &whereto);

	virtual int aim(GLVector3d &whereto)
	{
		mTank->setYawPitchFrom3DTarget(whereto, 1);
		return 0;
	};

	virtual int fire()
	{
		mTank->getCmds().fire = true;
		return 0;
	};
	virtual int stopfire()
	{
		mTank->getCmds().fire = false;
		return 0;
	};
	virtual int setweapon(int weptype)
	{
		getAppServer().getCommander().addCommand("/nextweapon", mPlayerID);
		// mTank->setWeaponServer(weptype);
		return 0;
	};

protected:

	virtual int findPathadd(GLVector3d &whereto, std::list< PathNode * > &mPath );

	virtual int findPath(GLVector3d &wherefrom, GLVector3d &whereto, std::list< PathNode * > &mPath );

	virtual int pathfindingInit();

	// list
	std::list< PathNode * > mPath;	// Path being followed
	
	bool mPathCreated;
	GLVector3d mPathEnd;	// exact point of path end
	GLVector3d mTarget;	// where tank should be aiming at
	int mSleepTime;		// counted in number of updates of AI (stopping movement for debugging purposes)
	bool mMoving;
	// int mDebug;

	virtual int simpleMove(GLVector3d whereto);		// move shortest possible path to destination (only for short distance move (like between pathfinding nodes) )

	int mPathNodesMaxx;
	int mPathNodesMaxy;

	// beware! after int pathfindingInit(); call do not change this vector 
	// (pointers to PathNode used and are invalidated with vector resize)
	std::vector< std::vector< PathNode > > mPathfindingNodes;

};

double getXYangle(GLVector3d first, GLVector3d second);

#endif

