/* MAINTAINER: Bernard */

#include "log.h"
#include "application_server.h"
#include "settings.h"

#include "game_server.h"
#include "commander_server.h"
#include "networkmanager_server.h"
#include "taskmanager.h"
#include "iortransfer_server.h"
#include "timemanager.h"
#include "mapcycle.h"
#include "console_text.h"

#include <ode/ode.h>

ApplicationServer::ApplicationServer(int argc, char* argv[])
:Application(argc, argv)
{
	// init ODE
	//dInitODE();

	TIN
	mGameServer           = new GameServer;
	mCommanderServer      = new CommanderServer;
	mNetworkManagerServer = new NetworkManagerServer;
	mMapCycle             = new MapCycle;
	mIORTransferServer	  = 0;

	if (!mGameServer || !mCommanderServer || !mNetworkManagerServer || !mMapCycle)
	{
		DIE("new failed");
	}
}

ApplicationServer::~ApplicationServer()
{
	delete mGameServer;
	delete mCommanderServer;
	delete mNetworkManagerServer;
	delete mMapCycle;
	if ( mIORTransferServer ) delete mIORTransferServer;

	// close ODE
	dCloseODE();
}

ApplicationServer::ApplicationServer(ApplicationServer &copy)
:Application(0, 0)
{
	TIN

	LOGE << "It is forbidden to make copy of AplicationServer." << endl;
	DIE();
}

void ApplicationServer::init()
{
	TIN
		
	loadUserConfigFile(SETTING_SERVER_CONFIG_FILE, SETTING_SERVER_CONFIG_FILE_TEMPLATE);
	
	g_settings.parseCommandLine(mArgc, mArgv); // reparse the command line
	
	
	// Log - open log file and set log level
	g_log.init(g_settings.getString(SETTING_LOGFILE_SERVER).c_str(),
			   g_settings.isTrue(SETTING_TOSTDERR_SERVER));
	g_log.set_loglevel(g_settings.getInt(SETTING_LOGLEVEL));
	
	LOGI << "Server setting and logging initialized." << endl;
	logBasicInfo();

	// prepare server global tasks
	mTaskManager->add(getCommander());

	mNetworkManagerServer->init();

	// prepare thread for providing this server's IOR
	unsigned short IORport = (unsigned short)g_settings.getInt( SETTING_NETWORK_IOR_TRANFER_PORT );
	std::string helloString = g_settings.getString(SETTING_NETWORK_WELCOME_MESSAGE );
	std::string sIOR;

	// waiting for orb start
	mNetworkManagerServer->getOrbManager().waitForOrbStart();

	// is orb running?
	if(mNetworkManagerServer->getOrbManager().orbRunning())
		sIOR = mNetworkManagerServer->getOrbManager().getServerIOR();
	else
		DIE("CORBA start failed. Please start server again and cross your fingers.");


	LOGI << "server IOR:" << sIOR << endl;

	mIORTransferServer = new IORTransferServer( IORport , sIOR, helloString );
	if ( !mIORTransferServer ) DIE("new failed");

	mMapCycle->init();
}

void ApplicationServer::run()
{
	TIN


	LOGI << "The BJS server salutes you. Welcome to the infinite loop." << endl;
	
	// Take first file in queue
	mMapCycle->getNextMapFile();
	updateBroadcastData();
	
	mReloadMap = true;
	mGameServer->init();

	// Run thread with IOR transfer server.
	LOGI << "Starting IOR transfer server" << endl;
	mIORTransferServer->start();

	// console
	ConsoleText console;
	console.start();

	while (mReloadMap)
	{
		mReloadMap = false;
	
		// Game is going to load the map	
		mGameServer->loadMap(mMapCycle->getActualMapFile());

		updateBroadcastData();

		// start the server itself
		mGameServer->run();

		// clean the game and try it again	
		//mGameServer->unloadMap(); // it is allready done
		//if (g_settings.isTrue(SETTING_SERVER_SINGLE_PLAYER)) mReloadMap = false;
	}
	
	mGameServer->clean();
	
}

void ApplicationServer::clean()
{
	TIN

	mIORTransferServer->setStop();
	mIORTransferServer->wait();

	mNetworkManagerServer->clean();

	saveUserConfigFile(SETTING_SERVER_CONFIG_FILE, SETTING_SERVER_CONFIG_FILE_TEMPLATE);
}


GameServer& ApplicationServer::getGame()
{
	assertL(mGameServer);
	return *mGameServer;
}

Game& ApplicationServer::getGameCommon()
{
	assertL(mGameServer);
	return *mGameServer;
}

Commander& ApplicationServer::getCommander()
{
	assertL(mCommanderServer);
	return *mCommanderServer;
}

NetworkManagerServer& ApplicationServer::getNetworkManager()
{
	assertL(mNetworkManagerServer);
	return *mNetworkManagerServer;
}


void ApplicationServer::quit()
{
	getGame().endFun();	
}


void ApplicationServer::updateBroadcastData()
{
	assertL(mIORTransferServer);
	if (mIORTransferServer == 0) return;

	// Some mix of this could be transfered
	// All players:   getGame().getPlayerCount()
	// Human players: getGame().getHumanPlayerCount()
	// AI players:    getGame().getAIManager().getAICount()
	// Max humans:    g_settings.getInt(SETTING_MAX_PLAYERS)

	mIORTransferServer->setPlayerCount(getGame().getPlayerCount());
	
	// Get the map description
	getMapCycle().getActualMapInfo().name;
	mIORTransferServer->setMapDescription( getMapCycle().getActualMapInfo().name );
}

void ApplicationServer::reloadToMap(const std::string &map)
{
	mMapCycle->setNextMapToLoadNext(map);
	loadNextMap();
}

MapCycle& ApplicationServer::getMapCycle()
{
	assertL(mMapCycle);
	return *mMapCycle;
}

void ApplicationServer::loadNextMap()
{
	gTime.enablePause(false);
	
	
	mMapCycle->getNextMapFile();
	mReloadMap = true;
	
	// we need to update info about the game as soon as possible
	updateBroadcastData();
	
	// Send all players command to load new map	
	getNetworkManager().sendCommand("reconnect");
	
	getGame().endFun();
	
	// unpause if reloading...
	gTime.pause(false);
}
