/* MAINTAINER: Bernard */

#ifndef _APPLICATION_SERVER_H
#define _APPLICATION_SERVER_H

#include <string>
#include <list>
#include "application.h"
#include "globals_server.h"

class GameServer;
class Commander;
class CommanderServer;
class NetworkManagerServer;
class IORTransferServer;
class MapCycle;

class ApplicationServer : public Application
{
public:
	/**
	 * \brief Allocate all member pointer variables except mIORTransferServer.
	 * 
	 * \param argc Used for inicialization of Application.
	 * \param argv Used for inicialization of Application.
	 */
	ApplicationServer(int argc, char* argv[]);
	
	/**
	 * \brief Free all pointer variables.
	 */
	~ApplicationServer();

	/**
	 * \brief Init subsystems and allocate mIORTransferServer.
	 * 
	 * mIORTransferServer is inicialized later since it needs some data not yet available
	 * in contructor.
	 */
	virtual void init();

	/**
	 * \brief Start the server loop.
	 */
	virtual void run();

	/**
	 * \biref Clean tasks and save settings.
	 */
	virtual void clean();

	/**
	 * \brief Quit the appliaction.
	 * 
	 * Note that the call returns and the app will quit later.
	 */
	virtual void quit();

	/**
	 * \brief Application will change map as soon as possible to this new map.
	 * 
	 * \param map File with new map.
	 */
	void reloadToMap(const std::string &map);

	/**
	 * \brief Load next map from map list.
	 */
	void loadNextMap();

	/**
	 * \brief Update data sended by broadcast to clients.
	 * 
	 * You should call this if some important happnes. Like
	 * new player in game or player was kicked from game etc...
	 */
	void updateBroadcastData();
	
	/**
	 * \brief  Provides access to GameServer.
	 * 
	 * \return mGameServer
	 */
	GameServer& getGame();
	
	/**
	 * \brief Provides access to basics of game.
	 * 
	 * Sources on server should use getGame(). This function is here becaue
	 * of virtual interface of Application and access to the Game from
	 * common sources.
	 * \return mGameClient
	 */
	virtual Game& getGameCommon();
	
	/**
	 * \brief Provides access to commander on server.
	 * 
	 * \return mCommanderServer
	 */
	Commander& getCommander();
	
	/**
	 * \brief Provides access to map cycle.
	 * 
	 * \return mReloadMap
	 */
	MapCycle& getMapCycle();
	
	/**
	 * \brief Provides access to network subsystem.
	 * 
	 * \return mNetworkManagerServer
	 */
	NetworkManagerServer& getNetworkManager();

private:
	
	/**
	 * \brief Program killer - trap
	 *
	 * From security reasons we don't allow anyone to create
	 * copy of Application since it is probably a mistake.
	 * Anyway calling this constructor will kill program.
	 * \param copy not used
	 */
	ApplicationServer(ApplicationServer &copy);

	GameServer           *mGameServer;
	CommanderServer      *mCommanderServer;
	NetworkManagerServer *mNetworkManagerServer;
	IORTransferServer	 *mIORTransferServer;
	MapCycle             *mMapCycle;
	
	bool                  mReloadMap; /// \brief Distiguis between reloading map and quit.
};

#endif // _APPLICATION_SERVER_H
