/* MAINTAINER: Bernard */

#ifndef _COMMANDER_SERVER_H
#define _COMMANDER_SERVER_H

#include "commander.h"
#include <string>
#include <sstream>


class Tank;
class Player;

/**
 * \brief Processor of server commands.
 * 
 * Accepts users commands from console and from network.
 */
class CommanderServer : public Commander
{
public:
	CommanderServer();
	virtual ~CommanderServer(){ };


protected:

	/**
	 * \brief Process one command.
	 * 
	 * This function will be called in main task.
	 * The function test if the player has enough privileges!
	 * \param command string with command
	 * \param playerID ID of player who created the command
	 */
	virtual void processCommand(const std::string &command, int playerID);

	/**
	 * \brief Shortcut for getting player's tank.
	 * 
	 * \param playerID ID of player for whom we want to know the tank.
	 * \return Pointer to tank or 0 if no tank.
	 */
	static Tank*   getTank(int playerID);
	
	/**
	 * \brief Convert playerID to pointer to Player.
	 * 
	 * \param playerID ID of player we want to have pointer for.
	 * \return Valid pointer to Player or 0 on error.
	 */
	static Player* getPlayer(int playerID);

	/**
	 * \brief Hash table with admin commands
	 */
	commads_hash mAdminCommandsHash;

// ***** COMMANDS

	// Return true if playerID is admin player
	static bool isAdmin(int playerID);

	/**
	 * \brief Sends chat message to everyone.
	 * \param args message
	 */
	static void command_chat(std::stringstream &args, int playerID);
	
	/**
	 * \brief Quit the server.
	 * \param args not used
	 */
	static void command_quit(std::stringstream &args, int playerID);

	/**
	 * \brief Change player's tank weapon.
	 * \param args weapon_id
	 */
	static void command_weapon(std::stringstream &args, int playerID);

	/**
	 * \brief Change player's tank weapon to next available.
	 * \param args not used
	 */
	static void command_next_weapon(std::stringstream &args, int playerID);

	/**
	 * \brief Change player's tank weapon to previous available.
	 * \param args not used
	 */
	static void command_previous_weapon(std::stringstream &args, int playerID);

	/**
	 * \brief Change game speed.
	 * \param args game_speed
	 */
	static void command_game_speed(std::stringstream &args, int playerID);

	/**
	 * \brief Change paused state
	 * \param args not used
	 */
	static void command_pause(std::stringstream &args, int playerID);

	/**
	 * \brief create new tank for player
	 * \param args tank_id
	 */
	static void command_new_tank(std::stringstream &args, int playerID);

	/**
	 * \brief change name of current player
	 * \param args new name
	 */
	static void command_change_name(std::stringstream &args, int playerID);

	/**
	 * \brief change name of current player
	 * \param args new name
	 */
	static void command_suicide(std::stringstream &args, int playerID);
	
	/**
	 * \brief Clean old map and load new map
	 * \param args new_map
	 */
	static void command_newmap(std::stringstream &args, int playerID);
	
	/**
	 * \brief Add one AI player
	 * \params args number of AI's
	 */
	static void command_addai(std::stringstream &args, int playerID);
	 
	/**
	 * \brief kick player from game
	 * \param args playerID_for_kick
	 */ 
	static void command_kick(std::stringstream &args, int playerID);
	
	/**
	 * \brief Player confirm that he is ready for the game
	 * \param args not used
	 */ 
	static void command_ready(std::stringstream &args, int playerID);	
	
	/**
	 * \brief Player confirm that he is ready for the game
	 * \param args password
	 */ 
	static void command_wannabeadmin(std::stringstream &args, int playerID);

	/**
	 * \brief Sets (new) fraglimit for current map
	 * \param args Frag limit or 0 for no limit
	 */ 
	static void command_fraglimit(std::stringstream &args, int playerID);

	/**
	 * \brief Sets (new) time for current map
	 * \param args Time limit or 0 for no limit
	 */ 
	static void command_timelimit(std::stringstream &args, int playerID);

	/**
	 * \brief Restarts current map
	 * \param args no args
	 */ 
	static void command_restart(std::stringstream &args, int playerID);

	/**
	 * \brief Restarts current map
	 * \param args no args
	 */ 
	static void command_nextmap(std::stringstream &args, int playerID);	
	
	/**
	 * \brief Set limit for maximum players in game
	 * \param args number of players
	 */ 
	static void command_maxplayers(std::stringstream &args, int playerID);
	
	/**
	 * \brief Set limit for maximum time of play without human player
	 * \param args second_before_shutdown
	 */ 
	static void command_maxtimenohuman(std::stringstream &args, int playerID);
	
};

#endif
