/* MAINTAINER: Martin */

#ifndef _PHYSICAL_TERRAIN_H_
#define _PHYSICAL_TERRAIN_H_

#include "terrain_base.h"

/**
*	\brief specific terrain behavior on server
*	
*/
class PhysicalTerrain : public TerrainBase
{
public:
	typedef double TVertex[3];

	PhysicalTerrain(void):mBlocks(0), mBlocksW(0), mBlocksH(0), mBlockShift(0) {}
	virtual ~PhysicalTerrain(void)
	{ delete[] mBlocks; mBlocks = 0; }

	virtual void init(void);

	int getBlockShift(void) { return mBlockShift; }
	
	/**
	*	\return terrain block size (number of vertices on block's side)
	*/
	int getBlockSize(void) { return (1 << mBlockShift) + 1; }
	
	/**
	*	\return horizontal minimal number of blocks covering the heightmap
	*/
	int getBlocksCountX(void) { return mBlocksW; }

	/**
	*	\return vertical minimal number of blocks covering the heightmap
	*/
	int getBlocksCountY(void) { return mBlocksH; }		

protected:
	/**
	*	\brief small block of terrain 
	*/
	class TerrainBlock
	{
	public:			
		struct BlockDescriptor
		{
			int 
				x, ///< \brief starting x-coord of block - in heightmap units
				y, ///< \brief starting y-coord of block - in heightmap units
				w, ///< \brief block width - in heightmap units
				h; ///< \brief block width - in heightmap units
			TerrainBase *tb; ///< \brief reference to terrain base.
		};

		TerrainBlock(): mVertexBuffer(0), mIndexBuffer(0), mVertexBufferSize(0), mIndexBufferSize(0) {}
		~TerrainBlock()
		{
			delete[] mVertexBuffer; mVertexBuffer = 0;
			delete[] mIndexBuffer; mIndexBuffer = 0;
		}

		void createTriMeshData(const BlockDescriptor &bd);
		
		void getTriMeshData(TVertex * &vertices, int & verticesCount, int * &indices, int & indicesCount)		{
			vertices = mVertexBuffer; verticesCount = mVertexBufferSize;
			indices = mIndexBuffer; indicesCount = mIndexBufferSize;
		}

	private:		
		TVertex * mVertexBuffer;
		int* mIndexBuffer;
		int mVertexBufferSize;
		int mIndexBufferSize;
	};

public:	
	
	/**
	*	\brief ODE support
	*	\param bx x coordinate of block (zero based), in block-units.
	*	\param by y coordinate of block (zero based), in block-units
	*	\param vertices returned vertex buffer
	*	\param verticesCount size of the vertex buffer
	*	\param indices returned index buffer
	*	\param indicesCount size of the indexbuffer
	*	\note \c bx lies in [0, vertical block count - 1], \c by in [0, horizontal block count]
	*/
	void getTriMeshData(int bx, int by, TVertex * &vertices, int & verticesCount, int * &indices, int & indicesCount)
	{
		TerrainBlock::BlockDescriptor bd;
		getDescriptor(bx, by, bd);
		mBlocks[by * getBlocksCountX() + bx].getTriMeshData(vertices, verticesCount, indices, indicesCount);
	}

protected:

	void createBlocks();
	
	
	void render(void);
	void render(double vertices[][3], int vertCount, int indices[], int indCount);
	

	/**
	*	\brief creates block descriptor from given block's coordinates \c bx, \c by (in block matrix).
	*	\param bx x-coordinate of block in block-matrix, zero-based (i.e. second block will have bx == 1)
	*	\param by y-coordinate of block in block-matrix, zero-based
	*	\param d  descriptor to fill
	*	\return reference to \c d
	*/
	TerrainBlock::BlockDescriptor& getDescriptor(int bx, int by, TerrainBlock::BlockDescriptor &d);
	TerrainBlock::BlockDescriptor getDescriptor(int bx, int by)
	{ 
		TerrainBlock::BlockDescriptor tmp;
		return getDescriptor(bx, by, tmp);
	}

	/**
	*	\brief Array of blocks covering the whole terrain.
	*/
	TerrainBlock* mBlocks;
	int 
		mBlocksW, ///< \brief the number of blocks in x-axis
		mBlocksH; ///< \brief the number of blocks in y-axis

	/**
	*	\brief block's binary shift value
	*
	*	Determines the block size. Block size will be ((1 << mBlockShift) + 1) - this means that 
	*	all blocks are squares with common power-of-two-plus-one size. 
	*
	*/
	int mBlockShift;
	
};

#endif //_PHYSICAL_TERRAIN_H_
