#include "serverphysicsdata.h"
#include "game.h"
#include "application.h"
#include "mdesc_builder.h"
#include "conversions.h"
#include "globals.h"
#include "world.h"
#include "tank.h"

			///< Identification of model node with collision model.
const char * RIGID_GEOM = "rigid_geom";

			///< Identification of model nodes with definition of left wheels.
const char * LEFT_WHEELS[] = {	
							"wheel_front_left", 
							"wheel_back_left"
						};

			///< Identification of model nodes with definition of right wheels.
const char * RIGHT_WHEELS[] = {	
							"wheel_front_right",
							"wheel_back_right"
						};

const int SUSPENSION_AXIS = 2;				///< \brief Index of axis of suspension in array of pivot axes
const int ROTATION_AXIS = 1;				///< \brief Index of axis around that "wheels" may turn in array of pivot axes.

const double ODE_TRACK_LENGTH_MULTIPLICATOR = 1.3;		///< \brief How much longer are tracks than origina wheel distance.

/**	\brief Helper class, that creates geom and mass structure from geometry object. It works for box and sphere only.
 */
class ODEGeomCreator: public gobj::Visitor
{
public:
	ODEGeomCreator( GLVector3d & scale_vec, dSpaceID geom_space )
	{
		mGeom = 0;
		mScale = scale_vec;
		mSpace = geom_space;
	}

	/**	\brief Creates box geom and sets its inertia matrix
	 */
	virtual void visitBox(const gobj::Box * box)
	{
		double ex, ey, ez; 
		box->getExtents(ex, ey, ez);

		ex*= mScale.x;
		ey*= mScale.y;
		ez*= mScale.z;

		mGeom = dCreateBox( mSpace, ex, ey, ez );
		dMassSetBox( &mMass, 1, ex, ey, ez );
	}
  
	/**	\brief Creates sphere geom and sets its inertia matrix
	 */
	virtual void visitSphere(const gobj::Sphere * sphere)
	{
		double radius = sphere->getRadius();
		radius*= mScale.x;
		
		mGeom = dCreateSphere( mSpace, radius );
		dMassSetSphere( &mMass, 1, radius );
	}

	dGeomID getGeom()
	{
		return mGeom;
	}

	dMass getMass()
	{
		return mMass;
	}
private:
	dGeomID mGeom;
	dMass mMass;
	GLVector3d mScale;
	dSpaceID mSpace;
};

/**	\brief Helper class that gets radius of wheel (only can be used for spheres)
 */
class TrackRadiusGetter: public gobj::Visitor
{
public:
	TrackRadiusGetter()
	{
		mRadius = 0;
	}

	virtual void visitBox(const gobj::Box * box)	// je to box
	{
		DIE( "Cannog get track radius from box" );
	}
  
	/**	\brief gets sphere's radius
	 */
	virtual void visitSphere(const gobj::Sphere * sphere)	// je to Sphere
	{
		mRadius = sphere->getRadius();
	}

	double getRadius()
	{
		return mRadius;
	}
private:
	double mRadius;
};

bool ServerPhysicsData::createODERigidAndColModel( int entity_id, GLVector3d position, GLQuaterniond orientation, TaskSimulationServer * sim )
{
	WorldEntity * we = gApp->getGameCommon().getWorld().getWorldEntity( entity_id );
	
	//prepare body for ODE
	dBodyID bID= dBodyCreate( sim->mODEWorld );
	mMainBody = bID;
	mBodies.push_back( bID );

	GLVector3d pos = we->getPosition();
	dBodySetPosition( bID, pos.x, pos.y, pos.z );
	GLQuaterniond rot = we->getOrientation();
	dQuaternion quat;
	quat[0] = rot.getW();
	quat[1] = rot.getV().x;
	quat[2] = rot.getV().y;
	quat[3] = rot.getV().z;
	dBodySetQuaternion( bID, quat );

	TaskSimulationServer::ODEBodyInfo * obi = new TaskSimulationServer::ODEBodyInfo;
#ifdef _DEBUG
	if ( we->getClassType() == cTank ) 
	{
		DIE( "Fatal error! Virtual method call failed?" );
		obi->complete = false;
	}
	else
#endif
		obi->complete = true;

	obi->entID = we->getID();
	obi->type  = bGeneral;

	double body_weight = (double)( we->getWeight() ) / (double)( sim->mBasicTnkWeight );

	MDescBuilder mdb;

	mdb.init();

	Model * model = we->getModel();
	HierarchyNode * hn = model->getRoot();
	hn->accept( &mdb );

	MDescBuilder::ModelNodeDescriptor md;

	// TODO: rigid_body? left_wheel?
	int nodeNameId = mdb.getNameId( RIGID_GEOM );

	if ( nodeNameId == MDescBuilder::INVALID_ID )
	{
		DIE( "FATAL: Cannot load model node 'rigid_geom'" );
	}
	
	// I have got node with object
	if ( mdb.isComposite(nodeNameId) )
	{
		MDescBuilder::CompositeIterator it;

		dMass compositeMass;
		dMassSetZero( &compositeMass );

		mdb.getNodeDescriptor( nodeNameId, md );

		it = mdb.getCompositeIterator( nodeNameId );
		while ( !it.isDone() )
		{
			if ( ! mdb.getNodeDescriptor( it.get(), md ) )
			{
				DIE( "ERROR in physical model! Node descriptor initialisation failed!" );
			}
			const gobj::GeometryObject * obj = md.getGeometryObject();
			if ( !obj ) 
			{
				DIE( "ERROR in model! No physical representation found" );
			}

			ODEGeomCreator ogc( md.scale, 0 );
			obj->accept( &ogc );

			dGeomID subgeom = ogc.getGeom();
			dGeomID	geomtra = dCreateGeomTransform( sim->mODEOtherSpace );
			dGeomTransformSetCleanup( geomtra, 1 );

			dGeomTransformSetGeom( geomtra, subgeom );

			dQuaternion subquat;
			subquat[0] = md.rotation.getW();
			subquat[1] = md.rotation.getV().x;
			subquat[2] = md.rotation.getV().y;
			subquat[3] = md.rotation.getV().z;
			dMatrix3 rotmat;
			dRfromQ( rotmat, subquat );

			dGeomSetRotation( subgeom, rotmat );
			dGeomSetPosition( subgeom, md.position.x, md.position.y,  md.position.z );

			dMass m = ogc.getMass();
			dMassRotate( &m, rotmat );
			dMassTranslate( &m, md.position.x, md.position.y,  md.position.z );

			dMassAdd( &compositeMass, &m );

			obi->geoms.push_back( geomtra );
			dGeomSetBody( geomtra, bID );

			it.next();
		}

		//NOTE: The ODE cannot handle objects that have center of mass diferent from (0,0,0)
		// So I move the mass so that it center corresnponds to (0,0,0)
		// Try to make models like that - try to make the point of reference of object its center of mass
		//TODO: check if this was solved in latest ODE version
		if ( !vecEqual( GLVector3d( 0, 0, 0 ), GLVector3d( compositeMass.c ) ) )
		{
			LOGD << "Model for object hasn't its center of mass equal to point of reference!" << endl;
			LOGD << "It is translated by " << GLVector3d( compositeMass.c ) << endl;
			dMassTranslate( &compositeMass, -compositeMass.c[0], -compositeMass.c[1], -compositeMass.c[2]);
		}

		dMassAdjust( &compositeMass, body_weight );
		dBodySetMass( bID, &compositeMass );
		dBodySetData( bID, obi );
	}
	else	//single object
	{	
		// get descriptor (md) of this object
		if ( ! mdb.getNodeDescriptor(nodeNameId, md) )
		{
			DIE( "ERROR in physical model! Node descriptor initialisation failed!" );
		}
		
		const gobj::GeometryObject * obj = md.getGeometryObject();
		if ( ! obj ) 
		{
			DIE( "ERROR in model! No physical representation found" );
		}
			
		bool need_transform;
		if ( vecEqual( md.position, GLVector3d(0, 0, 0) ) && dblEqual( 1, md.rotation.getW() ) )
		{
			need_transform = false;
		}
		else
		{
			need_transform = true;
		}

		ODEGeomCreator ogc( md.scale, need_transform ? 0 : sim->mODEOtherSpace );
		obj->accept( &ogc );

		dMass m = ogc.getMass();
		dMassAdjust( &m , body_weight );		//nastavenie hmotnosti
				
		dGeomID gID = ogc.getGeom();

		if ( need_transform )
		{
			dMass compositeMass;
			dMassSetZero( &compositeMass );
			dGeomID subgeom = gID;
			gID = dCreateGeomTransform( sim->mODEOtherSpace );
			dGeomTransformSetCleanup( gID, 1 );

			dGeomTransformSetGeom( gID, subgeom );

			dQuaternion subquat;
			subquat[0] = md.rotation.getW();
			subquat[1] = md.rotation.getV().x;
			subquat[2] = md.rotation.getV().y;
			subquat[3] = md.rotation.getV().z;
			dMatrix3 rotmat;
			dRfromQ( rotmat, subquat );

			dGeomSetRotation( subgeom, rotmat );
			dGeomSetPosition( subgeom, md.position.x, md.position.y,  md.position.z );
			dMassRotate( &m, rotmat );
			dMassTranslate( &m, md.position.x, md.position.y,  md.position.z );

			dMassAdd( &compositeMass, &m );

			//NOTE: The ODE cannot handle objects that have center of mass diferent from (0,0,0)
			// So I move the mass so that it center corresnponds to (0,0,0)
			// Try to make models like that - try to make the point of reference of object its center of mass
			//TODO: check if this was solved in latest ODE version
			if ( !vecEqual( GLVector3d( 0, 0, 0 ), GLVector3d( compositeMass.c ) ) )
			{
				LOGD << "Model for object hasn't its center of massequal to point of reference!" << endl;
				LOGD << "It is translated by " << GLVector3d( compositeMass.c ) << endl;
				dMassTranslate( &compositeMass, -compositeMass.c[0], -compositeMass.c[1], -compositeMass.c[2]);
			}
			m = compositeMass;
		}
					
		dBodySetMass( bID, &m );
		dGeomSetBody( gID, bID );

		obi->geoms.push_back( gID );
				
		dBodySetData( bID, obi );

		return true;			
	}

	return false;
}


bool TankPhysicsData::createODERigidAndColModel( int entity_id, GLVector3d position, GLQuaterniond orientation, TaskSimulationServer * sim )
{
	mTotSpeed = 0;
	mLeftSpeed = 0;
	mRightSpeed= 0;

	Tank * we = dynamic_cast<Tank*> ( gApp->getGameCommon().getWorld().getWorldEntity( entity_id ) );
	//build the ODE bodies

	//prepare body for ODE
	dBodyID bID= dBodyCreate( sim->mODEWorld );
	mMainBody = bID;
	mBodies.push_back( bID );

	GLVector3d pos = we->getPosition();
	dBodySetPosition( bID, pos.x, pos.y, pos.z );
	GLQuaterniond rot = we->getOrientation();
	dQuaternion quat;
	quat[0] = rot.getW();
	quat[1] = rot.getV().x;
	quat[2] = rot.getV().y;
	quat[3] = rot.getV().z;
	dBodySetQuaternion( bID, quat );

	TaskSimulationServer::ODEBodyInfo * obi = new TaskSimulationServer::ODEBodyInfo;
	if ( we->getClassType() == cTank ) 
	{
		obi->complete = false;
	}
	else
	{
		DIE( "Fatal error! Virtual method call failed?" );
		obi->complete = true;
	}

	obi->entID = we->getID();
	obi->type  = bGeneral;

	double body_weight = (double)(we->getWeight()) / (double)(sim->mBasicTnkWeight);
	double track_weight= body_weight / 2;

	MDescBuilder mdb;

	mdb.init();

	Model * model = we->getModel();
	HierarchyNode * hn = model->getRoot();
	hn->accept( &mdb );

	MDescBuilder::ModelNodeDescriptor md;

	// TODO: rigid_body? left_wheel?
	int nodeNameId = mdb.getNameId( RIGID_GEOM );

	if ( nodeNameId == MDescBuilder::INVALID_ID )
	{
		DIE( "FATAL: Cannot load model node 'rigid_geom'" );
	}
	
	// I have got node with object
	if ( mdb.isComposite(nodeNameId) )
	{
		MDescBuilder::CompositeIterator it;

		dMass compositeMass;
		dMassSetZero( &compositeMass );

		it = mdb.getCompositeIterator( nodeNameId );
		while ( !it.isDone() )
		{
			if ( ! mdb.getNodeDescriptor( it.get(), md ) )
			{
				DIE( "ERROR in physical model! Node descriptor initialisation failed!" );
			}
			const gobj::GeometryObject * obj = md.getGeometryObject();
			if ( !obj ) 
			{
				DIE( "ERROR in model! No physical representation found" );
			}

			ODEGeomCreator ogc( md.scale, 0 );
			obj->accept( &ogc );

			dGeomID subgeom = ogc.getGeom();
			dGeomID	geomtra = dCreateGeomTransform( sim->mODEOtherSpace );
			dGeomTransformSetCleanup( geomtra, 1 );

			dGeomTransformSetGeom( geomtra, subgeom );

			dQuaternion subquat;
			subquat[0] = md.rotation.getW();
			subquat[1] = md.rotation.getV().x;
			subquat[2] = md.rotation.getV().y;
			subquat[3] = md.rotation.getV().z;
			dMatrix3 rotmat;
			dRfromQ( rotmat, subquat );

			dGeomSetRotation( subgeom, rotmat );
			dGeomSetPosition( subgeom, md.position.x, md.position.y,  md.position.z );

			dMass m = ogc.getMass();
			dMassRotate( &m, rotmat );
			dMassTranslate( &m, md.position.x, md.position.y,  md.position.z );

			dMassAdd( &compositeMass, &m );

			obi->geoms.push_back( geomtra );
			dGeomSetBody( geomtra, bID );

			it.next();
		}

		//NOTE: The ODE cannot handle objects that have center of mass diferent from (0,0,0)
		// So I move the mass so that it center corresnponds to (0,0,0)
		// Try to make models like that - try to make the point of reference of object its center of mass
		//TODO: check if this was solved in latest ODE version
		if ( !vecEqual( GLVector3d( 0, 0, 0 ), GLVector3d( compositeMass.c ) ) )
		{
			LOGD << "Model for object hasn't its center of massequal to point of reference!" << endl;
			LOGD << "It is translated by " << GLVector3d( compositeMass.c ) << endl;
			dMassTranslate( &compositeMass, -compositeMass.c[0], -compositeMass.c[1], -compositeMass.c[2]);
		}

		dMassAdjust( &compositeMass, body_weight );
		dBodySetMass( bID, &compositeMass );
		dBodySetData( bID, obi );
	}
	else	//single object
	{	
		// get descriptor (md) of this object
		if ( ! mdb.getNodeDescriptor(nodeNameId, md) )
		{
			DIE( "ERROR in physical model! Node descriptor initialisation failed!" );
		}
		
		const gobj::GeometryObject * obj = md.getGeometryObject();
		if ( ! obj ) 
		{
			DIE( "ERROR in model! No physical representation found" );
		}
			
		bool need_transform;
		if ( vecEqual( md.position, GLVector3d(0, 0, 0) ) && dblEqual( 1, md.rotation.getW() ) )
		{
			need_transform = false;
		}
		else
		{
			need_transform = true;
		}

		ODEGeomCreator ogc( md.scale, need_transform ? 0 : sim->mODEOtherSpace );
		obj->accept( &ogc );

		dMass m = ogc.getMass();
		dMassAdjust( &m , body_weight );		//nastavenie hmotnosti
				
		dGeomID gID = ogc.getGeom();

		if ( need_transform )
		{
			dMass compositeMass;
			dMassSetZero( &compositeMass );
			dGeomID subgeom = gID;
			gID = dCreateGeomTransform( sim->mODEOtherSpace );
			dGeomTransformSetCleanup( gID, 1 );

			dGeomTransformSetGeom( gID, subgeom );

			dQuaternion subquat;
			subquat[0] = md.rotation.getW();
			subquat[1] = md.rotation.getV().x;
			subquat[2] = md.rotation.getV().y;
			subquat[3] = md.rotation.getV().z;
			dMatrix3 rotmat;
			dRfromQ( rotmat, subquat );

			dGeomSetRotation( subgeom, rotmat );
			dGeomSetPosition( subgeom, md.position.x, md.position.y,  md.position.z );
			dMassRotate( &m, rotmat );
			dMassTranslate( &m, md.position.x, md.position.y,  md.position.z );

			dMassAdd( &compositeMass, &m );

			//NOTE: The ODE cannot handle objects that have center of mass diferent from (0,0,0)
			// So I move the mass so that it center corresnponds to (0,0,0)
			// Try to make models like that - try to make the point of reference of object its center of mass
			//TODO: check if this was solved in latest ODE version
			if ( !vecEqual( GLVector3d( 0, 0, 0 ), GLVector3d( compositeMass.c ) ) )
			{
				LOGD << "Model for object hasn't its center of massequal to point of reference!" << endl;
				LOGD << "It is translated by " << GLVector3d( compositeMass.c ) << endl;
				dMassTranslate( &compositeMass, -compositeMass.c[0], -compositeMass.c[1], -compositeMass.c[2]);
			}
			m = compositeMass;
		}
					
		dBodySetMass( bID, &m );
		dGeomSetBody( gID, bID );

		obi->geoms.push_back( gID );
				
		dBodySetData( bID, obi );
	}

	dBodyID track[2];
	dGeomID tgeom[2];
	dJointID joints[2][2];
	GLVector3d points[2][2];
	GLVector3d susp_axes[2][2];
	GLVector3d rot_axes[2][2];
	double radiuses[2][2];

	assertL( ( sizeof( RIGHT_WHEELS ) / sizeof( char* )) == 2 );	//2 joints per track
	assertL( ( sizeof( LEFT_WHEELS ) / sizeof( char* )) == 2 );

		//create ODE bodies for tracks
	for ( int i=0; i<2; i++ ) 
	{
		track[i] = dBodyCreate( sim->mODEWorld );
		dBodySetAutoDisableFlag( track[i], 0 );

			//orientation of wheel  should be same as orientation of whole tank
		dBodySetQuaternion (track[i],quat);

		for ( int j = 0; j < 2; j++ )
		{
			int tracknodeID = mdb.getNameId( i == 0 ? RIGHT_WHEELS[j] : LEFT_WHEELS[j] );
			if ( tracknodeID == MDescBuilder::INVALID_ID )
			{
				LOGE << "Unable to locate node " << ( i == 0 ? RIGHT_WHEELS[j] : LEFT_WHEELS[j] ) << endl;
				DIE( "WHEEL NODE NOT FOUND IN PHYSICAL MODEL" );
			}

			assertL( mdb.isComposite( tracknodeID ) == false );
			
			if ( !mdb.getNodeDescriptor( tracknodeID, md ) )
			{
				DIE( "Cannot get node descriptor for wheel" );
			}

			const gobj::GeometryObject * obj = md.getGeometryObject();
			if ( ! obj ) 
			{
				DIE( "ERROR in model! No physical representation of wheel found" );
			}

			TrackRadiusGetter trg;
			obj->accept( &trg );
			radiuses[i][j] = trg.getRadius()*md.scale.x;
					
			points[i][j] = pos + rot.rotateVector( md.pivotPosition );

			susp_axes[i][j] = rot.rotateVector( md.pivotAxes[SUSPENSION_AXIS] );
			rot_axes[i][j]  = rot.rotateVector( md.pivotAxes[ROTATION_AXIS] );
		}

		assertL( dblEqual( radiuses[i][0], radiuses[i][1] ) );

		dMass mass;

		dMassSetSphere( &mass, 1, radiuses[i][0] );
		dMassAdjust( &mass, track_weight );
		dBodySetMass( track[i], &mass );

		double len = (points[i][0] - points[i][1]).length();

		dGeomID tmpgeom = dCreateCapsule( 0, radiuses[i][0], len * ODE_TRACK_LENGTH_MULTIPLICATOR );
		tgeom[i] = dCreateGeomTransform( sim->mODEOtherSpace );

		dGeomTransformSetGeom(  tgeom[i], tmpgeom );
		dGeomTransformSetCleanup( tgeom[i], 1 );
		dQuaternion q1;
		dQFromAxisAndAngle( q1, 0, 1, 0, M_PI/2 );	//standard capsule is standing - we need it 'laying'
		dGeomSetQuaternion( tmpgeom, q1 );

		dGeomSetBody( tgeom[i], track[i] );

		GLVector3d tposv = (points[i][0] + points[i][1])/2;
		dBodySetPosition( track[i], tposv.x, tposv.y, tposv.z );

		TaskSimulationServer::ODEBodyInfo * tobi = new TaskSimulationServer::ODEBodyInfo;
		tobi->complete = false;
		tobi->entID = we->getID();
		tobi->geoms.push_back( tgeom[i] );
		tobi->type = bTrack;
		tobi->spec.trackTerrContact = false;

		dBodySetData( track[i], tobi );

		mBodies.push_back( track[i] );
	}

	GLVector3d lvec = (points[0][1] - points[0][0]).unit();
	GLVector3d rvec = (points[1][1] - points[1][0]).unit();

		//tracks must be parrallel
	assertL( vecEqual( lvec, rvec ) || vecEqual( lvec, -rvec ) );

		//connect  tracks with body with joints
	for ( int i = 0; i < 2; i++ )
	{
		for ( int j = 0; j < 2; j++ )
		{
			joints[i][j] = dJointCreateHinge2( sim->mODEWorld, 0 );
			dJointAttach (joints[i][j], bID, track[i] );
			
			dJointSetHinge2Anchor(joints[i][j], points[i][j].x, points[i][j].y, points[i][j].z );
			
			dJointSetHinge2Axis1( joints[i][j], susp_axes[i][j].x, susp_axes[i][j].y, susp_axes[i][j].z );
			dJointSetHinge2Axis2( joints[i][j], rot_axes[i][j].x, rot_axes[i][j].y, rot_axes[i][j].z );

			dJointSetHinge2Param( joints[i][j], dParamLoStop, 0 );
			dJointSetHinge2Param( joints[i][j], dParamHiStop, 0 );

			//tlmice/pruzenie AKO TO NASTAVIT, DPC???
			/*
			double dump_const = 1;
			double spring_const = 0;
			double step_size = 0.02;

			double s_erp = step_size * spring_const / ( step_size * spring_const + dump_const );
			double s_cfm = 1 / (step_size * spring_const + dump_const);

			dJointSetHinge2Param( joints[i][j], dParamSuspensionCFM, s_cfm);
			dJointSetHinge2Param( joints[i][j], dParamSuspensionERP, s_erp);
			*/

			mJoints.push_back( joints[i][j] );
		}
	}

	mRightTrack = track[0];
	mLeftTrack = track[1];

	mRightJoints[0]= joints[0][0];
	mRightJoints[1]= joints[0][1];
	mLeftJoints[0] = joints[1][0];
	mLeftJoints[1] = joints[1][1];
	
	return true;
}



bool LandscapeEntityPhysicsData::createODERigidAndColModel(int entity_id, GLVector3d position, GLQuaterniond orientation, TaskSimulationServer *sim)
{
	LandscapeEntity * le = gApp->getGameCommon().getWorld().getLandscape().getLandEntity( entity_id );
	
	//static - no bodies
	mMainBody = 0;
	
	GLVector3d pos = le->getPosition();
	GLQuaterniond rot = le->getOrientation();
	
	MDescBuilder mdb;

	mdb.init();

	Model * model = le->getModel();
	HierarchyNode * hn = model->getRoot();
	hn->accept( &mdb );

	MDescBuilder::ModelNodeDescriptor md;

	// TODO: rigid_body? left_wheel?
	int nodeNameId = mdb.getNameId( RIGID_GEOM );

	if ( nodeNameId == MDescBuilder::INVALID_ID )
	{
		//TODO: THIS IS A TEMPORARY HACK FOR GRASS ONLY!!!
		// THIS COULD BE REMOVED LATER. EVERY LANDSCAPE ENTITY
		// SHOULD HAVE SOME PHYSICS MODEL
		//DIE( "FATAL: Cannot load model node 'rigid_geom'" );
		return true; // WHAT TO RETURN? true or false?
		// We should really clear this hack before all is done
		// - in other words make is such it isn't needed.
	}
	
	// I have got node with object
	if ( mdb.isComposite(nodeNameId) )
	{
		MDescBuilder::CompositeIterator it;

		mdb.getNodeDescriptor( nodeNameId, md );

		it = mdb.getCompositeIterator( nodeNameId );
		while ( !it.isDone() )
		{
			if ( ! mdb.getNodeDescriptor( it.get(), md ) )
			{
				DIE( "ERROR in physical model! Node descriptor initialisation failed!" );
			}
			const gobj::GeometryObject * obj = md.getGeometryObject();
			if ( !obj ) 
			{
				DIE( "ERROR in model! No physical representation found" );
			}

			ODEGeomCreator ogc( md.scale, sim->mODETerrainSpace );
			obj->accept( &ogc );

			dGeomID subgeom = ogc.getGeom();
			
			dQuaternion subquat;
			subquat[0] = (rot * md.rotation).getW();
			subquat[1] = (rot * md.rotation).getV().x;
			subquat[2] = (rot * md.rotation).getV().y;
			subquat[3] = (rot * md.rotation).getV().z;

			dGeomSetQuaternion( subgeom, subquat );

			GLVector3d totpos = pos + md.position;
			dGeomSetPosition( subgeom, totpos.x, totpos.y,  totpos.z );

			it.next();
		}
	}
	else	//single object
	{	
		// get descriptor (md) of this object
		if ( ! mdb.getNodeDescriptor(nodeNameId, md) )
		{
			DIE( "ERROR in physical model! Node descriptor initialisation failed!" );
		}
		
		const gobj::GeometryObject * obj = md.getGeometryObject();
		if ( ! obj ) 
		{
			DIE( "ERROR in model! No physical representation found" );
		}

		ODEGeomCreator ogc( md.scale, sim->mODETerrainSpace );
		obj->accept( &ogc );
		
		dGeomID gID = ogc.getGeom();

		dQuaternion quat;
		quat[0] = ( rot * md.rotation ).getW();
		quat[1] = ( rot * md.rotation ).getV().x;
		quat[2] = ( rot * md.rotation ).getV().y;
		quat[3] = ( rot * md.rotation ).getV().z;
		
		dGeomSetQuaternion( gID, quat );

		GLVector3d totpos = pos + md.position;
		dGeomSetPosition( gID, totpos.x, totpos.y,  totpos.z );

		return true;			
	}

	return false;
}
