/* MAINTAINER: Bernard */

#include "log.h"
#include "settings.h"

#include "game_server.h"

#include "taskmanager.h"
#include "taskapplyclientsinput.h"
#include "tasksimulationserver.h"
#include "tasksendchanges.h"

#include "console_text.h"
#include "globals_server.h"
#include "application_server.h"

#include "thread.h"
#include "timemanager.h"

#include "world.h"

#include "conversions.h"

#include "tinyxml/tinyxml.h"

#include "playernetdata_server.h"
#include "aimanager.h"

#include "gameplaysettings.h"
#include "deathmatch.h"
#include "endconditions.h"

#include "commander.h"

#include "taskserverkiller.h"

GameServer::GameServer()
{
	TIN

	mAIManager = new AIManager();
	if (mAIManager == 0) DIE("new failed");

	gTime.setGameSpeed(g_settings.getDouble(SETTING_GAME_SPEED_MULTIPLICATOR));

	mPlaySettings = 0;
}

GameServer::~GameServer()
{
	TIN
   
	assertL(mAIManager);
	delete mAIManager;
	mAIManager = 0;

	if ( mPlaySettings )
	{
		delete mPlaySettings;
		mPlaySettings = 0;
	}
 
	removeAllPlayers();
}

void GameServer::removeAllPlayers()
{
  	// Prepare deletion of all players.
	for (int i = 0; (i < (int)mPlayers.size()); ++i)
	{
		if (mPlayers[i])
			preparePlayerDeletion(mPlayers[i]->getID());
	} 

	Game::removeAllPlayers();
}

GameServer::GameServer(GameServer &copy)
{
   TIN;

   LOGE << "It is forbidden to make copy of AplicationServer." << endl;
   DIE();
}


void GameServer::loadMap(const std::string mapName)
{
	TIN;
	
	
	LOGI << "Loading map: " << mapName << endl;
	
	Game::loadMap(mapName);
	
	getWorld().getLandscape().getTerrain().initServer();
	loadLandscapeEntities();
	
	getAppServer().updateBroadcastData();

	// make players not ready 
	setReadyForAllPlayers(false);
	
		//initialise game type and default game conditions
	string gtype = mMapParams.game_type;

	if (mPlaySettings) delete mPlaySettings;

	mPlaySettings = new GamePlaySettings;
	
	if ( gtype == "deathmatch" )
	{
		initDeathmatch();
		return;
	}

	// if ( gtype == "rabbit_hunt" )
	// {
	//		initRabbitHunt();
	//		return;
	// }

		//no game type specified -> error
	DIE( "Unknown type of the map." );
}

void GameServer::unloadMap()
{
	TIN;

	LOGI << "Unloading map." << endl;
	
	Game::unloadMap();
}

void GameServer::run()
{
	TIN;
	assertL(mAIManager);
   
	mEndFun = false;


	// create tasks	
	TaskApplyClientsInput applyClients;
	TaskSimulationServer  moveGameState;
	TaskSendChanges       sendChanges;
	TaskServerKiller      serverKiller;

	// and add to the manager
	mTaskManager.add(applyClients);
	mTaskManager.add(moveGameState);
	mTaskManager.add(sendChanges);
	mTaskManager.add(serverKiller);
	mTaskManager.initTasks();
	
	mTaskManager.add(getAppServer().getTaskManager());
	mTaskManager.add(*mAIManager);
	mTaskManager.add( *mPlaySettings );

	
	// START CLIENT RECEIVE THREAD
   
	setRunning(true);
	gTime.enablePause(true);
	while(!mEndFun)
	{

		mTaskManager.update();

		// Give chance to other threads
		Thread::reshedule();
	}
  	gTime.enablePause(false);
   
	unloadMap();
	
	mWorld->cleanUp();

	// kick_all_players();
	
	mTaskManager.remove( *mPlaySettings );
	mTaskManager.remove(*mAIManager);	
	mTaskManager.remove(getAppServer().getTaskManager());
	mTaskManager.cleanTasks();
}


void GameServer::init()
{
	Game::init();
}


void GameServer::clean()
{
	// Prepare deletion of all players.
	for (int i = 0; (i < (int)mPlayers.size()); ++i)
	{
		if (mPlayers[i])
			preparePlayerDeletion(mPlayers[i]->getID());
	}
	
	if ( mPlaySettings )
	{
		delete mPlaySettings;
		mPlaySettings = 0;
	}

	Game::clean();
}



#define LOAD_COORDINATE(x) \
	tmp = child->Attribute(#x); \
	if (tmp == NULL ||  !try_get_double(tmp, x)) \
	{ \
		LOGE << "Wrong Spawn point attribute " << #x << endl; \
		return false; \
	}

#define LOAD_INT(x) \
	tmp = child->Attribute(#x); \
	if (tmp == NULL ||  !try_get_long(tmp, x)) \
	{ \
		LOGE << "Wrong bonus attribute " << #x << endl; \
		return false; \
	}

bool GameServer::loadMapParts(TiXmlHandle &docHandle)
{
	//Loading spawn points from map
	mSpawnPoints.clear();

	TiXmlElement* child;
	child = docHandle.FirstChild("map").FirstChild("spawn_points").FirstChild("point").Element();
	if (!child)
	{
		DIE("Missing spawn points in map");
	}
	for (; child; child=child->NextSiblingElement()) 
	{
		double x,y,z, angle;
		const char *tmp;

		LOAD_COORDINATE(x);
		LOAD_COORDINATE(y);
		LOAD_COORDINATE(z);		
		LOAD_COORDINATE(angle);


		SpawnPoint sp;
		sp.position.x = x;
		sp.position.y = y;
		sp.position.z = z;
		sp.angle = angle;

		mSpawnPoints.push_back(sp);
	}
	if (mSpawnPoints.size() == 0)
	{
		LOGE << "No spawn points defined!" << endl;
		return false;
	}

	//Loading frequency of bonus boxes
	mBonusFrequency.clear();
	mBonusFrequency.assign(gParams.mBonus.size(), 0);
	child = docHandle.FirstChild("map").FirstChild("bonus_occurences").FirstChild("bonus").Element();
	// no need to have bonuses in map
	for (; child; child=child->NextSiblingElement()) 
	{
		long id,in_10_minutes;
		const char *tmp;

		LOAD_INT(id);
		LOAD_INT(in_10_minutes);

		if (id < 0 || (unsigned int)id >= mBonusFrequency.size())
		{
			LOGE << "Invalid bonus id " << id << endl;
			return false;
		}
		mBonusFrequency[id] = in_10_minutes;
	}


	return Game::loadMapParts(docHandle);
}

void GameServer::getSpawnPoint(GLVector3d &point, double &angle)
{
	assertL(mSpawnPoints.size() != 0);

	SpawnPoint sp = mSpawnPoints[rand()%mSpawnPoints.size()];
	point = sp.position;

	// Compute Z coordinate
	Terrain &t = getAppServer().getGame().getWorld().getLandscape().getTerrain();
	t.cropToValidXY(point);
	point.z += t.getHeight(point.x,point.y);

	angle = sp.angle;
}


int  GameServer::getBonusesIn10Minutes(int bonusID)
{
	assertL(bonusID >= 0 && (unsigned int)bonusID < mBonusFrequency.size());
	return mBonusFrequency[bonusID];
}

void GameServer::preparePlayerDeletion(int ID)
{
	assertL( isPlayerIDValid( ID ) );

	// check if we are in single player and if so shut down the server
	if (getPlayer(ID).isHumanPlayer() && g_settings.isTrue(SETTING_SERVER_SINGLE_PLAYER))
	{
		LOGI << "Quit because of ending singple player server." << endl;
		getAppServer().getCommander().addCommand("/quit");
	}

	if(getPlayer( ID ).canUpdateClient())
	{
		// disable network handler
		(dynamic_cast<PlayerNetDataServer &>(getPlayer( ID ).netData())).disableHandler();
	}

	getPlayer( ID ).setDeleteMe();

	if (mWorld)
	{
		mWorld->deletePlayersEntities( ID );
	}
	
	mAIManager->removedPlayer(ID);
}

int GameServer::newPlayer(Player &player, int ID)
{
	// TODO: how to return errors if some other reason?
	// let AI enter game even if player limit is full
	if (player.isHumanPlayer() && getHumanPlayerCount() >= g_settings.getInt(SETTING_MAX_PLAYERS) )
	{
		return INVALID_PLAYER_ID;
	}
	
	int rv = Game::newPlayer(player, ID);
	getAppServer().updateBroadcastData();
	return rv;
}

void GameServer::removePlayer(int ID)
{
	Game::removePlayer(ID);
	getAppServer().updateBroadcastData();
}

AIManager& GameServer::getAIManager()
{
	assertL(mAIManager);
	return *mAIManager;
}

void GameServer::initDeathmatch()
{
	assertL( mPlaySettings );

	Deathmatch * dm = new Deathmatch;

	mPlaySettings->setType( dm );

	loadEndConds();
}

void GameServer::loadEndConds()
{
	assertL( mPlaySettings );

	stringstream ecs( mMapParams.victory_settings );

	while ( ecs.good() )
	{
		string type;

		ecs >> type;

		if ( type != "" )
		{
			if ( type == "fraglimit" )
			{
				int limit;
				ecs >> limit;
				if ( !mPlaySettings->addEndCondition( new FragLimit( limit ) ) )
				{
					DIE( "Error in game type and game end conditions initialization" );
				}
			}

			if ( type == "timelimit" )
			{
				double time;
				ecs >> time;
				if ( !mPlaySettings->addEndCondition( new TimeLimit( time ) ) )
				{
					DIE( "Error in game type and game end conditions initialization" );
				}
			}

		}
		else
		{
			break;
		}
	}
	
}

int GameServer::getHumanPlayerCount()
{
	return 	getPlayerCount() - getAIManager().getAICount();
}
