/* MAINTAINER: Bernard */

#ifndef _GAME_SERVER_H
#define _GAME_SERVER_H

#include "mutex.h"
#include "game.h"
#include "globals.h"

class AIManager;
class GamePlaySettings;

class GameServer : public Game 
{

public:
	GameServer();
	virtual ~GameServer();

	/**
	 * \brief Start the game.
	 */
	virtual void run();

	
	/**
	 * \brief Init data structures for game.
	 *
	 */
	virtual void init();
	
    
	/**
	 * \brief Clean data structures for game. 
	 *
	 */
	virtual void clean();
	
	virtual void removeAllPlayers();
	

	/**
	 * \brief Load new map
	 * 
	 */
	virtual void loadMap(const std::string mapName);

	
	
	/**
	 * \brief Clear all static objects from map
	 */
	virtual void unloadMap();
	
	/**
	 * \brief Pick one randomly chosen spawn point and return it.
	 */	
	void getSpawnPoint(GLVector3d &point, double &angle);

	int  getBonusesIn10Minutes(int bonusID);

	/**	\brief Prepares player for deletion - sets him flag to delete and notifies world to remove its entities
	 *
	 *	\param ID of player that will be removed
	 */
	void preparePlayerDeletion( int ID );
	
	/**
	 * \brief Shut down the server if in singleplayer and network player
	 *        disconnected.
	 */
	virtual void removePlayer(int ID);
	
	/**
	 * \brief Provide access to AI manager
	 */
	AIManager& getAIManager();

	/**	\brief Returns game rules pointer. 
	 *	\note can return zero
	 *
	 *	\return Pointer to GamePlaySettings instance or 0 if map is not loaded yet or is beeing reloaded
	 */
	GamePlaySettings * getPlaySettings()
	{
		return mPlaySettings;
	}
	
	/**
	 * \brief Return number of players controlled by humans in game.
	 * 
	 * Player is considered as human if it is a network player.
	 */
	int getHumanPlayerCount();
	
	/*
	int tryAddPlayer(Player &player, int &playerID);
	
	enum
	{
		NO_ADD_PLAYER_MAX_PLAYERS,	
		NO_ADD_PLAYER_BAN,
	}
	*/
	/**
	 * \brief adding player witch check for MAX_PLAYERS limit.
	 * 
	 * This is bad solution since it should deflect why the player
	 * wasn't added if wasn't. Like BAN or such a thing.
	 */
	virtual int newPlayer(Player &player, int ID = INVALID_PLAYER_ID);
	
	
private:

	virtual bool loadMapParts(TiXmlHandle &docHandle);

	
	/**
	 * \brief Program killer - trap
	 *
	 * From security reasons we don't allow anyone to create
	 * copy of GameServer since it is probably a mistake.
	 * Anyway calling this constructor will kill program.
	 * \param copy not used
	 */
	GameServer(GameServer &copy);
	
	struct SpawnPoint
	{
		GLVector3d position;
		double     angle;
	};

	typedef std::vector<SpawnPoint> SpawnPointsVector;

	SpawnPointsVector       mSpawnPoints;
	std::vector<int>        mBonusFrequency;

	// Holder of all AI
	AIManager *mAIManager;

	GamePlaySettings * mPlaySettings;

	void initDeathmatch();
	void loadEndConds();
};

#endif // _GAME_CLIENT_H
