/* MAINTAINER: Petko */
#include "gameplaysettings.h"
#include "timemanager.h"
#include "log.h"
#include "application.h"
#include "player.h"
#include "tank.h"
#include "settings.h"
#include "application_server.h"
#include "globals.h"
#include "commander.h"
#include "game_server.h"
#include "aimanager.h"
#include "world.h"

GamePlaySettings::GamePlaySettings()
{
	mType = 0;
		//at start of map wait for players to get ready
	mState= WaitForPlayers;
	init();	
	mTickDelta = 100;
}

void GamePlaySettings::init()
{
	mPauseStart = -1;
	mEOGTime	= -1;

	const std::vector<int> & plids = getApp().getGameCommon().getValidPlayersID();
	for ( unsigned int i = 0; i < plids.size(); i++ )
	{
			//reset players' stats and ready status
		getApp().getGameCommon().getPlayer( plids[i] ).initStats();
	}
}

GamePlaySettings::GamePlaySettings(const GamePlaySettings &gps)
{
	DIE( "Attempt to copy GamePlaySettings" );
}

GamePlaySettings::~GamePlaySettings()
{
	if ( mType )
		delete mType;
}

void GamePlaySettings::killed( Player * killer, Player * dead )
{
	assertL( mType && "Game type must be initialised" );
	if ( mState != Playing ) return;

	mType->killed( killer, dead );
}

void GamePlaySettings::suicide( Player * who )
{
	assertL( mType && "Game type must be initialised" );
	if ( mState != Playing ) return;
	
	mType->suicide( who );
}

void GamePlaySettings::fall( Player * who )
{
	assertL( mType && "Game type must be initialised" );
	if ( mState != Playing ) return;
	
	mType->fall( who );
}

void GamePlaySettings::collide( Player * who, WorldEntity * with )
{
	assertL( mType && "Game type must be initialised" );
	if ( mState != Playing ) return;
	
	mType->collide( who, with );
}

void GamePlaySettings::tick()
{
	assertL( mType );

	mCondListMutex.lock();
	mType->update( mState );
	mCondListMutex.unlock();

		//handle pauses
	if ( mPauseStart < 0 && gTime.isPaused() && mState == Playing )
	{
		mPauseStart = gTime.getTicks();
		return;
	}
	if ( mPauseStart > 0 )
	{
		if ( !gTime.isPaused() )
		{
			assertL( mType );

			mCondListMutex.lock();
			mType->paused( mPauseStart, gTime.getTicks() );
			mCondListMutex.unlock();

			mPauseStart = -1;
		}
		else
			return;
	}

		//are we ready for live?
	if ( mState == WaitForPlayers )
	{
		const std::vector<int> & plids = getApp().getGameCommon().getValidPlayersID();

		if ( plids.size() == 0 )	//no players -> no play
			return;

			//AIs are always ready... so wait until any player is ready
		GameServer &game = getAppServer().getGame();
		if ( game.getHumanPlayerCount() == 0 )
			return;

		for ( unsigned int i = 0; i < plids.size(); i++ )
		{
			if ( !(getApp().getGameCommon().getPlayer( plids[i] ).isReady() ) )
			{
				return;		//players are not ready yet -> no rules
			}
		}
			//all players are ready now
		mState = Playing;

			//delete players' tanks and zero players' stats
		mCondListMutex.lock();

		start();

		mType->start();

		mCondListMutex.unlock();

		//TODO: Client notification about score counting should be here

			//let them play at least one frame :)
		return;
	}

		//if game in progress, just check if we should end
	if ( mState == Playing )
	{
		mCondListMutex.lock();
		bool ending = mType->checkTheEnd();
		mCondListMutex.unlock();

		if ( ending )
		{
				//pause game 
				//TODO: urobit nejake ine zapauzovanie, alebo zakazat odpauzovanie
			gTime.pause( true );

			mState = EndOfGame;
			mEOGTime = gTime.getTicks() + (int)(g_settings.getDouble( SETTING_GAMEPLAY_CHATTIME ) * 1000); //from milliseconds to seconds
				//TODO: send clients information about end of map -> display results table

			return;
		}
				
			//Let them play
		return;
	}

	if ( mState == EndOfGame )
	{
		if ( ( mEOGTime >= 0 ) && (mEOGTime <= gTime.getTicks())  )
		{
			/*
				//determine which map follows
			std::string nmap = getAppServer().mMapCycle.getNextMapFile();
			if ( nmap == ApplicationServer::MapCycle::NO_MAP )
			{
					//no next map - reload this
				nmap = getAppServer().mMapCycle.getLastMapFile();
				if ( nmap == "" )
				{
					DIE( "Last map is set to empty. This is some stupid error" );
				}

				restart();

					//TODO: Tu pouzit tu "systemovu" pauzu, alebo zrusit zakaz pauzovat
				gTime.pause( false );

				return;
			}

				//reload map
			getAppServer().getCommander().addCommand( string("/newmap ") + nmap );
			*/
			
			getAppServer().loadNextMap();
			
			mEOGTime = -1;
				//TODO: Tu pouzit tu "systemovu" pauzu, alebo zrusit zakaz pauzovat
			gTime.pause( false ); // O tu by se take mela starat aplikace primo
			return;
		}
	}
}

void GamePlaySettings::start()
{
	mCondListMutex.lock();

	init();

		//remove everything on map
	World & wrld = getApp().getGameCommon().getWorld();

	worldent_iterator it = wrld.getWorldEntitiesBegin();
	worldent_iterator ite= wrld.getWorldEntitiesEnd();
	for ( ; it != ite; ++it )
	{
		(*it)->setDeleteFlag();
	}

	mCondListMutex.unlock();
}

bool GamePlaySettings::restart()
{
	mCondListMutex.lock();
	mState= WaitForPlayers;
	init();
	start();

		//TODO: unpause is vital here, but this is not good - should be only "system" unpause here (players will still see "game paused")
	gTime.pause( false );
	mCondListMutex.unlock();

	return true;
}

bool GamePlaySettings::addEndCondition(MapEndCondition *cond)
{
	if ( !mType )
	{
		return false;
	}

	mCondListMutex.lock();
	mType->addEndCondition( cond );

	if ( cond->getTargetValue() == 0 )
	{
		delete cond;
	}
	else
	{
		if ( mState == Playing )
		{
			cond->start();
		}
	}
	mCondListMutex.unlock();

	return true;
}

void GamePlaySettings::setType( GameType * gt )
{
	assertL( gt != 0 );
	assertL( (mType == 0) && "Cannot change game type!" );
	mCondListMutex.lock();
	mType = gt;
	mCondListMutex.unlock();
}

const std::list< ConditionStatus > GamePlaySettings::getCondStatuses()
{
	assertL( mType );

	mCondListMutex.lock();
	const std::list< ConditionStatus > retval = mType->getCondStatuses();
	mCondListMutex.unlock();

	return retval;
}

GameType::~GameType()
{
	while ( !mEndConds.empty() )
	{
		MapEndCondition * mec = mEndConds.front();
		mEndConds.pop_front();

		assertL( mec );
		delete mec;
	}
}

bool GameType::checkTheEnd()
{
	std::list< MapEndCondition* >::iterator it  = mEndConds.begin();
	std::list< MapEndCondition* >::iterator ite = mEndConds.end();

	for ( ; it != ite; ++it )
	{
		if ( (*it)->satisfied() )
		{
			return true;
		}
	}

	return false;
}

void GameType::addEndCondition( MapEndCondition *cond )
{
	assertL( cond );
	
		//there cannot be two conditions with same name in condition list - overwrite the old one
	std::list< MapEndCondition* >::iterator it  = mEndConds.begin();
	std::list< MapEndCondition* >::iterator ite = mEndConds.end();
	for ( ; it != ite; ++it )
	{
		if ( cond->getName() == (*it)->getName() )
		{
			mEndConds.erase( it );
			break;
		}
	}

	if ( cond->getTargetValue() != 0 )
	{
		mEndConds.push_front( cond );
	}
}

void GameType::paused(int from, int to)
{
	std::list< MapEndCondition* >::iterator it  = mEndConds.begin();
	std::list< MapEndCondition* >::iterator ite = mEndConds.end();
	for ( ; it != ite; ++it )
	{
		(*it)->paused( from, to );
	}
}

void GameType::start()
{
	std::list< MapEndCondition* >::iterator it  = mEndConds.begin();
	std::list< MapEndCondition* >::iterator ite = mEndConds.end();
	for ( ; it != ite; ++it )
	{
		(*it)->start();
	}
}

const std::list< ConditionStatus >& GameType::getCondStatuses()
{
	mCondStats.clear();

	ConditionStatus cs;

	std::list< MapEndCondition* >::iterator it  = mEndConds.begin();
	std::list< MapEndCondition* >::iterator ite = mEndConds.end();
	for ( ; it != ite; ++it )
	{
		MapEndCondition * mec = *it;
		cs.name = mec->getName();
		cs.currentvalue = mec->getCurrentValue();
		cs.targetvalue  = mec->getTargetValue();

		mCondStats.push_back( cs );
	}

	return mCondStats;
}

MapEndCondition::~MapEndCondition()
{
}
