/* MAINTAINER: Petko */
#ifndef _GAMEPLAYSETTINGS_H_
#define _GAMEPLAYSETTINGS_H_

#include "task.h"
#include "mutex.h"
#include "game.h"
#include <list>
#include <string>

class Player;
class WorldEntity;

/**	\brief Abstract interface for game end conditions
 */
class MapEndCondition
{
public:
	/**	\brief Initialzes name of condition
	 *
	 *	\param name Name of the condition
	 */
	MapEndCondition( const std::string & name ) : mName( name ){};
	virtual ~MapEndCondition() = 0;

	/**	\brief Gets name of condition
	 *
	 *	\return Name of the condition
	 */
	const std::string & getName()
	{
		return mName;
	}

	/**	\brief Checks if the condition implemented is satisfied. If true, game should end.
	 *
	 *	\return True, if game should end on this condition.
	 */
	virtual bool satisfied() = 0;

	/**	\brief Returns some value that the condition tries to reach. It is converted to integer value.
	 *	\note If this value is 0, the addcondition only deletes another condition with same name
	 *
	 *	\return Some target value as integer
	 */
	virtual int getTargetValue() = 0;

	/**	\brief Returns some value that the condition has now. It is converted to integer value.
	 *
	 *	\return Some current value as integer
	 */
	virtual int getCurrentValue()= 0;

	/**	\brief If game is paused, some condition may want to do something (time limit not to count time). So this lets them know when was game paused
	 *
	 *	\param from Starting tick of pause
	 *	\param to First tick after pause
	 */
	virtual void paused( int from, int to ){};

	/**	\brief Here should be some initializations that apply on map start or restart
	 */
	virtual void start(){};

private:
	std::string mName;		///< \brief name of condition
};

/**	\brief Abstract interface for implementing actual game types - how to maintain kills, suicides, ...
 */
class GameType
{
public:
	/**	\brief Initialzes name of condition
	 *
	 *	\param name Name of the condition
	 */
	GameType( const std::string & name ) : mName( name ){};
	virtual ~GameType() = 0;

	/**	\brief Let game type handle kill
	 *
	 *	\param killer Player that killed somebody
	 *	\param dead Player that was killed
	 */
	virtual void killed( Player * killer, Player * dead ){};

	/**	\brief Let game type handle suicide
	 *
	 *	\param who Player that commited suicide
	 */
	virtual void suicide( Player * who ){};

	/**	\brief Let game type handle players' fall from map
	 *
	 *	\param who Player that fell
	 */
	virtual void fall( Player * who ){};

	/**	\brief Let game type handle collision between players' tank and some other entity
	 *
	 *	\param who Player that collided
	 *	\param with Entity with which player collided
	 */
	virtual void collide( Player * who, WorldEntity * with ){};

	/**	\brief Gets name of the game type
	 *
	 *	\return Name of the game type
	 */
	const std::string & getName()
	{
		return mName;
	}

	/**	\brief Checks all map end condition if anyone is satisfied
	 *
	 *	\return True, if map should end
	 */
	virtual bool checkTheEnd();

	/**	\brief Adds end-of-game condition (in disjunction). The value is deleted by this class, so don't do anything with that pointer
	 *	\note If there already exists condition with same name, it replaces it. If target value of condition is 0, it removes condition with same name only.
	 *
	 *	\param cond Consition to add
	 */
	void addEndCondition( MapEndCondition * cond );

	/**	\brief If game is paused, some conditions or game types may want to do something (time limit not to count time). So this lets them know when was game paused
	 *
	 *	\param from Starting tick of pause
	 *	\param to First tick after pause
	 */
	void paused( int from, int to );

	/**	\brief Here should be some initializations that apply on map start or restart
	 */
	virtual void start();

	/**	\brief Regularly called function - allows to simulate something
	 *
	 *	\param state Current state(phase) of game.
	 */
	virtual void update( eStateOfGame state ){};

	/**	\brief Prepares list with statuses of map end conditions and returns reference to it
	 *	\note This method is NOT thread safe!
	 *
	 *	\return Reference to list with condition statuses
	 */
	const std::list< ConditionStatus >& getCondStatuses();

protected:
	std::string mName;		///< \brief Name of game type
	std::list< MapEndCondition* >	mEndConds;	///< \brief Conditions to end the game - used in dusjunction
	std::list< ConditionStatus >	mCondStats;	///< \brief All states of conditions
};

/**	\brief Class that manages game type and scoring. Also manages how maps are started, ended and load next one
 */
class GamePlaySettings : public Task
{
public:
	GamePlaySettings();
	~GamePlaySettings();

	/**	\brief Let game handle kill
	 *
	 *	\param killer Player that killed somebody
	 *	\param dead Player that was killed
	 */
	void killed( Player * killer, Player * dead );

	/**	\brief Let game handle suicide
	 *
	 *	\param who Player that commited suicide
	 */
	void suicide( Player * who );
	
	/**	\brief Let game handle players' fall from map
	 *
	 *	\param who Player that fell
	 */
	virtual void fall( Player * who );

	/**	\brief Let game handle collision between players' tank and some other entity
	 *
	 *	\param who Player that collided
	 *	\param with Entity with which player collided
	 */
	void collide( Player * who, WorldEntity * with );

	/**	\brief Main tick update - decidies what do to whith thi map
	 */
	virtual void tick();

	/**	\brief Handles things that need to be done at start of live play
	 */
	void start();

	/**	\brief Handles things that need to be done at restarting map
	 *
	 *	\return True, if map is going to restart, false otherwise
	 */
	bool restart();

	/**	\brief Adds end-of-game condition (in disjunction). The value is deleted by this class or passed somewhere, so don't do anything with the value
	 *	\note If there already exists condition with same name, it replaces it. If target value of condition is 0, it removes condition with same name only.
	 *
	 *	\param cond Consition to add
	 *	\return True if addition was sucessfull. false otherwise (probably game type not initialized)
	 */
	bool addEndCondition( MapEndCondition * cond );

	/**	\brief Gets state (phase) of game
	 *
	 *	\return State of game
	 */
	eStateOfGame getStateOfGame()
	{
		return mState;
	}

	/**	\brief Sets the game type
	 *
	 *	\param gt Gametype instance - don't delete or change that value - managed by this class
	 */
	void setType( GameType * gt );

	/**	\brief Gets list of map end conditions
	 *	\note This method IS thread safe!
	 *
	 *	\return List with condition statuses
	 */
	const std::list< ConditionStatus > getCondStatuses();

private:
	GamePlaySettings( const GamePlaySettings & gps );

	/**	\brief Initialization of some basic values
	 */
	void init();

	GameType*		mType;			///< \brief Type of game
	eStateOfGame	mState;			///< \brief Current state of game
	int				mPauseStart;	///< \brief Start of current pause or -1
	int				mEOGTime;		///< \brief Time when the game ended or -1
	Mutex			mCondListMutex;	///< \brief Mutex used to lock condition status list
};

#endif
