#include "mdesc_builder.h"
#include "matrix_to_quaternion.h"

void MDescBuilder::init()
{
	mNextId = 0;
	mNameToIdMap.clear();
	mIdToDescriptorMap.clear();
	mIdToCompositeMap.clear();
	mParentToWorld = mParentToWorld.identity();
	setCurrentGroupID( INVALID_ID );
}

void MDescBuilder::transformCoordSysOrigin(GLMatrix4d & mat, GLVector3d &pos, GLVector3d axes[3])
{
	// Nitramus: I know that this is stupid... But works ;-)

	GLVector4d tmp = mat * GLVector4d(0, 0, 0, 1);
	pos.set(tmp.x, tmp.y, tmp.z);
	pos /= tmp.w;
	
	// this will be bad if scale transformation is used
	static const GLVector4d basicAxes[3] = { 
		GLVector4d(1, 0, 0, 0),
		GLVector4d(0, 1, 0, 0),
		GLVector4d(0, 0, 1, 0) };

	for ( int i = 0; i < 3; ++i )
	{
		tmp = mat * basicAxes[i];
		axes[i].set(tmp.x, tmp.y, tmp.z);
	}
}

// toto se vykona na kazdem uzlu stromu modelu
void MDescBuilder::visitHierarchyNode(HierarchyNode * node)
{
	// zpracujeme uzel:

	// --------------------------------------
	// 1. obtain transformation of this node
	node->acceptAttributeVisitor( &mTO );
	hna::Transformation * t = mTO.get();

	// --------------------------------------
	// 2. transformation into world-space 
	GLMatrix4d toWorld( GLMatrix4d::identity() ), backupParent;

	if ( t )
	{
		// preserve parent node transformation matrix
		backupParent = mParentToWorld;
	
		// move to coordinates of this node
		mParentToWorld *= t->getNodeTM();

		// multiply with offset transformation of this node
		toWorld = mParentToWorld * t->getObjectOffsetTM();
		
		// toWorld - timto kdyz se ted vynasobi vrcholy, tak uz budou ve svetovych souradnicich.

		// mParentToWorld - timto kdyz se ted vynasobi pocatek soustavy souradnic a jednotlive
		//  osy, tak dostaneme ve svetovych souradnicich bod a vektory odpovidajici
		//  lokalnimu souradnemu systemu zpracovavaneho uzlu.
		//  Ten bod bude "kloub" a osy budou osy otaceni, pripadne mozneho posunu.
	}

	// ------------------------------------
	// 3. access to geometry object

	// GeometryObject - baseclass of box, sphere, trimesh, ...
	const gobj::GeometryObject * go = node->getGeometryObject();
		
	static const GLVector3d vzero(0, 0, 0), vx(1, 0, 0), vy(0, 1, 0), vz(0, 0, 1);

	GLVector3d 
		objectPos(vzero), 
		objectAxes[3] = { vx, vy, vz }, 
		pivotPos(vzero), 
		pivotAxes[3] = { vx, vy, vz }, 
		objectScale(1, 1, 1);
	
	GLQuaterniond objectRot; // initialized as unit quaternion
	GLMatrix4d objectRotMatrix( GLMatrix4d::identity() );

	if ( go )
	{
		// Transform box reference point to box center (hack for physics subsystem).
		{
			// Every Box object have the center of reference in the center of its lower base.
			// For physics it is convenient to have it in the middle of the box.
			// This code will adjust toWorld matrix.
			class AdjustBoxPositionToItsCenter: public gobj::Visitor {
			public:
				AdjustBoxPositionToItsCenter( GLMatrix4d & toWorld )
					: mToWorld(toWorld) 
				{}

				virtual void visitBox( const gobj::Box * box ){
					double tmp, ez;
					box->getExtents(tmp, tmp, ez);
					GLMatrix4d t;
					mToWorld *= t.loadTranslate(0, 0, ez / 2);
				}
			private:

				GLMatrix4d &mToWorld;
			} adjustBoxPositionToItsCenter( toWorld );

			go->accept(&adjustBoxPositionToItsCenter);
			
			// When you now recieve the box position, it will be position of its center!
		}

		// get pivot position and axes
		transformCoordSysOrigin(mParentToWorld, pivotPos, pivotAxes);

		// get object position and axes
		transformCoordSysOrigin(toWorld, objectPos, objectAxes);
		
		// obtain object scale
		for ( int i = 0; i < 3; ++i )
			objectScale[i] = objectAxes[i].length();

		// orthogonalize object axes and normalize them
		{
			GLVector3d tmp3d = objectAxes[0];
			tmp3d.cross(objectAxes[1]);
			objectAxes[2] = ( tmp3d.dot( objectAxes[2] ) > 0 ) ? tmp3d : -tmp3d;
			
			tmp3d = objectAxes[2];
			tmp3d.cross(objectAxes[0]);
			objectAxes[1] = ( tmp3d.dot( objectAxes[1] ) > 0 ) ? tmp3d : -tmp3d;
			
			for ( int i = 0; i < 3; ++i )
				objectAxes[i].normalize();
		}
		
		// get object rotation as a quaternion 
		// ( this is not necessary, ODE can work with matrices... )		
		{
			// rotation matrix - set object axes as matrix columns
			GLMatrix4d &rm = objectRotMatrix;
			for ( int col = 0; col < 3; ++col )
				for ( int row = 0; row < 3; ++row )
					rm[ col * 4 + row ] = objectAxes[col][row];

			// create quaternion from object rotation matrix
			objectRot = matrixToQuaternion(rm);
		}
	}

	// create node descriptor
	ModelNodeDescriptor mnd(pivotPos, pivotAxes, objectPos, objectRot, objectRotMatrix, objectScale, go);

	// associate node name with node description in the map
	int nodeID = mapInsert(node->getName(), mnd);
	
	// ---------------------------------------
	// composite support and children visiting
	{
		// Note: 
		// For every group member node, there has to exist a path from group head node of this
		// group to this member with all nodes being members of this group. 
		
		if ( node->isGroupMember() )
		// this node is a member of current group
		{
			// add nodeID to current group
			int currentGroupID( getCurrentGroupID() );
			if ( isValidID( currentGroupID ) )
			{
				insertIntoComposite( currentGroupID, nodeID ); // group is in fact a composite
			}
		}
	
		const bool groupHead( node->isGroupHead() );
		int oldGroupID( INVALID_ID );

		if ( groupHead )
		{
			// preserve current group ID and set it to id of this node
			oldGroupID = setCurrentGroupID( nodeID );
		}
		
		// ----------------------------
		// recursive visit all children
		visitHierarchyChildren(node);

		if ( groupHead )
		{
			// restore current group ID
			setCurrentGroupID( oldGroupID );
		}
	}

	// restore transform matrix of parent node
	if ( t )
		mParentToWorld = backupParent;
}

void MDescBuilder::insertIntoComposite( int compositeID, int nodeID )
{
	mIdToCompositeMap[compositeID].push_front( nodeID );
}
