// MAINTAINER: Martin

#ifndef _MDESC_BUILDER_H_
#define _MDESC_BUILDER_H_

#include "hash_hack.h"
#include "stlextdefs.h"

#include <string>
#include <stack>
#include <list>

#include "model.h"
#include "hierarchynodevisitor.h"
#include "hna_visitor.h"
#include "hna_transform.h"
#include "gobj_visitor.h"
#include "gobj_all.h"
#include "transform_obtainer.h"

#include "GLQuaternion.h"
#include "GLVector.h"
#include "GLMatrix.h"


/**
*	\brief Model nodes description builder
*
*	Builds a map with node name as a key that contains model node information -
*	geometry object, position, rotation, scale, pivot point.
*
*	\code
*	// ----------------------------------
*	// An example how to use MDescBuilder
*	// ----------------------------------
*
*	MDescBuilder mdb;	// we need a model descriptors builder instance
*	Model * pmodel;		// pointer to model - every world entity has one
*	// ...
*
*	// initialize the builder
*	mdb.init();
*
*	// - build descriptors of the entire model
*	pmodel->getRoot()->accept( &mdb );
*	
*	// model node descriptor - it will be filled with data that can be used when building ODE 
*	// geometry objects and bodies
*	MDescBuilder::ModelNodeDescriptor md;
*		
*	// node name of your interest (inside a model), for example left wheel
*	int nodeNameId = mdb.getNameId("left_wheel");
*	
*	if ( nodeNameId > 0 )
*	// name found
*	{
*		if ( mdb.isComposite( nodeNameId ) )
*		// it is a composite (group of objects)!
*		{
*			MDescBuilder::CompositeIterator it;
*
*			it = mdb.getCompositeIterator( nodeNameId );
*			while ( !it.isDone() )
*			{
*				// get composite member - a single object (guaranteed)
*				int singleObjectID = it.get();
*				
*				mdb.getNodeDescriptor( singleObjectID, md );
*				// - do some work with md (create ODE geometry and/or body, 
*				// insert it into ODE composite object)
*				
*				// move on next member
*				it.next();
*			}
*		}
*		else
*		// it is a single object! - COMMON CASE
*		{
*			// get descriptor (md) of this object
*			mdb.getNodeDescriptor( nodeNameId, md );
*			
*			// - do some work with md (create ODE geometry and/or body)
*						
*			// md.position - position in world space
*			// md.rotation - orientation in world space
*			// md.scale - extents in local space (x, y, z)
*			//          - you should scale geometry object extents by this value
*			// md.getGeometryObject() - geometry object base class, see Note
*			// md.pivotPoint - can be used as joint position
*			// md.pivotAxes - can be used as joint axes
*		}
*	}
*	// Note:
*	// You can implement gobj::Visitor interface to access derived classes of md.getGeometryObject(),
*	// or you can compare the type of base class with type of your interest and then use
*	// dynamic_cast<derived_type>(base_type_ptr).
*	//
*	\endcode
*/
class MDescBuilder: public HierarchyNodeVisitor
{
public:

	/** \brief ctor - only calls MDescBuilder::init() */
	MDescBuilder()
	{ 
		init(); 
	}

	/**
	*	\brief model node descriptor
	*
	*	Describes model node using pointer to geometry object (can be null), pivot point
	*	position and axes (not normalized nor diagonalized) and object position, rotation and scale.
	*/
	class ModelNodeDescriptor
	{
	public:
		/**
		*	\brief empty ctor without initialization
		*/
		ModelNodeDescriptor():
			mGeometryObject(0)
		{}

		/**
		*	\brief copy ctor
		*/
		ModelNodeDescriptor( const ModelNodeDescriptor &mnd ):
			pivotPosition( mnd.pivotPosition ),
			position( mnd.position ),
			scale( mnd.scale ),
			rotation( mnd.rotation ),
			rotationMatrix( mnd.rotationMatrix ),
			mGeometryObject(mnd.mGeometryObject )
		{
			pivotAxes[0] = mnd.pivotAxes[0];
			pivotAxes[1] = mnd.pivotAxes[1];
			pivotAxes[2] = mnd.pivotAxes[2];
		}
		
		/**
		*	\brief ctor that will set all class data
		*/
		ModelNodeDescriptor(const GLVector3d & _pivotPos, const GLVector3d _pivotAxes[3], 
			const GLVector3d &  _pos, const GLQuaterniond & _rot, const GLMatrix4d & _rotMatrix, const GLVector3d & _scl, 
			const gobj::GeometryObject * _gobj ):
				pivotPosition(_pivotPos),
				position(_pos),
				scale(_scl),
				rotation(_rot),
				rotationMatrix(_rotMatrix),
				mGeometryObject(_gobj)
			{
				pivotAxes[0] = _pivotAxes[0];
				pivotAxes[1] = _pivotAxes[1];
				pivotAxes[2] = _pivotAxes[2];
			}

		
		GLVector3d 
			pivotPosition,		///< \brief pivot point position
			pivotAxes[3],		///< \brief three pivot axes
			position,			///< \brief object position
			scale;				///< \brief object scale
		
		GLQuaterniond rotation;	///< \brief object rotation
		GLMatrix4d rotationMatrix; ///< \brief object rotation (in matrix form)
		
		/** \brief gets geometry object */
		const gobj::GeometryObject * getGeometryObject() const
		{ 
			return mGeometryObject; 
		}
		
	private:
		const gobj::GeometryObject * mGeometryObject; ///< \brief geometry object
	};

private:
	
	/** \brief name-to-id map type*/
	typedef STL_EXT_NS::hash_map< std::string, int > TNameToIdMap;
	
	/** \brief name-to-descriptors map type */
	typedef STL_EXT_NS::hash_map< int, ModelNodeDescriptor > TIdToDescriptorMap; 
	
	/** \brief name-to-composite map type */
	typedef STL_EXT_NS::hash_map< int, std::list< int > > TIdToCompositeMap; 

public:

	/** \brief composite const iterator type used by MDescBuilder::CompositeIterator */
	typedef TIdToCompositeMap::value_type::second_type::const_iterator TCompositeConstIterator;

	/**
	*	\brief gets ID of given name
	*
	*	Node descriptors and composites are associated with IDs and not with name strings.
	*
	*	\param name node name
	*	\return ID of given name, MDescBuilder::INVALID_ID on failure
	*/
	int getNameId( const std::string & name ) const
	{
		TNameToIdMap::const_iterator it( mNameToIdMap.find(name) );
		return (it != mNameToIdMap.end()) ? it->second : INVALID_ID;
	}
	
	/**
	*	\brief gets the node descriptor
	*	\param id node identificator ( from getNameId() )
	*	\param mnd where to store result ( node descriptor )
	*	\return true means success, false means not found
	*/
	bool getNodeDescriptor( int id, ModelNodeDescriptor & mnd ) const
	{
		TIdToDescriptorMap::const_iterator it( mIdToDescriptorMap.find(id) );
		if ( it != mIdToDescriptorMap.end() )
		{
			mnd = it->second;
			return true;
		}
		return false;
	}

	/**
	*	\brief test whether the node (given by ID) is a composite or not
	*	\param id node identificator ( from getNameId() )
	*	\return true means node is composite, false means single node
	*/
	bool isComposite( int id ) const
	{
		return mIdToCompositeMap.find( id ) != mIdToCompositeMap.end();
	}

	/**
	*	\brief implements visitor interface ( builds model descriptors )
	*	
	*	On each node, it creates node descriptor, fills it with data and stores it into maps. 
	*	For a gobj::Box geometry object, its position will be translated to the center of the box.
	*	It recursively visits children.
	*/
	virtual void visitHierarchyNode(HierarchyNode * node);

	/**
	*	\brief initializes builder internal states
	*	You should call this method every time before you start building a new model.
	*/
	void init();

	/**
	*	\brief iterator for composite objects
	*	
	*	Iterates trough all composite members. Makes nested composites transparent 
	*	(automatically iterates trough them as well). Value returned by CompositeIterator::get()
	*	in each valid step is always the simple, non-composite object.
	*/
	class CompositeIterator
	{
	friend class MDescBuilder;
	public:
		/** \brief empty ctor */
		CompositeIterator(){}

		/** \brief copy ctor */
		CompositeIterator(const CompositeIterator & srcIt )
			: mStack(srcIt.mStack), mSource(srcIt.mSource)
		{}

	private:
		/** \brief creates composite iterator based on composite id and source model descriptor builder 
		*	
		*	This initializes the iterator - it tries to find first simple element in the composite.
		*	This can lead to longer search including entering nested composites until the first single
		*	element is found or collection is empty.
		*
		*	\param id composite id (if it is not a composite, the iterator will behave as if the collection was empty )
		*	\param source model composite builder
		*/
		CompositeIterator( int id, const MDescBuilder * source )
			: mSource(source)
		{
			if ( !mSource || !mSource->isComposite( id ) )
				return;

			CompositeIDListBounds b;
			if ( !mSource->getCompositeMembers( id, b.it, b.it_end ) || b.it == b.it_end )
				return;

			// push a composite on the stack - data to process
			mStack.push(b);
			if (!tryGetFirst())
				mStack.pop();
		}
	public:

		/** \brief test of composite end */
		bool isDone() const
		{
			return mStack.empty();
		}
		
		/** 
		*	\brief gets the id of a simple object
		*	
		*	While isDone() returns true, returned value is valid.
		*
		*	\return identificator of a simple object that this iterator points to
		*/
		int get() const
		{
			Stack::const_reference top = mStack.top();
			return *top.it;
		}

		/** \brief one step of the iteration */
		void next()
		{
			if ( isDone() )
				return;

			// get current composite sequence
			Stack::reference top = mStack.top();
						
			if ( ++top.it == top.it_end )
			// end of this part
			{
				mStack.pop();
				next();
			}else{
				// try to get first element
				if ( !tryGetFirst() )
					next();
			}
		}
	
	private:

		bool tryGetFirst()
		{
			int id = get();
			if ( !mSource->isComposite( id ) )
				return true;

			CompositeIDListBounds b;
			if ( !mSource->getCompositeMembers( id, b.it, b.it_end ) || b.it == b.it_end )
				return false;

			mStack.push(b);
			if (!tryGetFirst())
			{
				mStack.pop();
				return false;
			}
			return true;
		}

		typedef ::MDescBuilder::TCompositeConstIterator CompositeIDListIterator;
		
		struct CompositeIDListBounds
		{
			CompositeIDListIterator it, it_end;
		};
		typedef std::stack< CompositeIDListBounds > Stack;

		Stack mStack;
		const MDescBuilder * mSource;
	};

	/**
	*	\brief creates a iterator for given composite
	*	\param compositeID composite identificator ( from MDescBuilder )
	*	\return iterator for collection of composite members
	*	\note if compositeID represents single object, empty collection or doesn't exist, 
	*	the returned CompositeIterator will be an iterator over empty collection.
	*/
	CompositeIterator getCompositeIterator( int compositeID )
	{
		return CompositeIterator(compositeID, this);
	}

private:

	/**
	*	\brief inserts descriptor into map, uses node name as a key
	*
	*	Creates a new ID and maps node name to this id, then maps this id to
	*	node descriptor.
	*
	*	\param nodeName name of node that is described by mnd
	*	\param mnd node descriptor
	*	\param id of nodeName
	*/
	int mapInsert(std::string nodeName, const ModelNodeDescriptor & mnd)
	{
		TNameToIdMap::const_iterator it( mNameToIdMap.find(nodeName) );
		if ( it != mNameToIdMap.end() )
			return it->second;

		mNameToIdMap[nodeName] = mNextId;
		mIdToDescriptorMap[mNextId] = mnd;
		return mNextId++;
	}

	/**
	*	\brief try to get composite members (including other composites)
	*	\param id node identificator ( from getNameId() )
	*	\param begin iterator, first member
	*	\param end iterator, end of the collection, one element behind the last one
	*	\return true means thaat id represented a composite (begin with end are valid), false means invalid
	*/
	bool getCompositeMembers( int id, TCompositeConstIterator &begin, TCompositeConstIterator &end ) const
	{
		TIdToCompositeMap::const_iterator it( mIdToCompositeMap.find(id) );
		if ( it != mIdToCompositeMap.end() )
		{
			begin = it->second.begin();
			end = it->second.end();
			return true;
		}
		return false;
	}

	/** \brief sets current group identificator (internal state) */
	int setCurrentGroupID( int groupID )
	{ return (mCurrentGroupID = groupID); }
	
	/** \brief gets current group identificator (internal state) */
	int getCurrentGroupID() const
	{ return mCurrentGroupID; }

	/**
	*	\brief inserts a node into a composite
	*	\param compositeID composite identificator
	*	\param nodeID node identificator
	*/
	void insertIntoComposite( int compositeID, int nodeID );

	/**
	*	\brief transforms coordinate system origin and three axes
	*	\param mat transformation matrix
	*	\param pos position of coordinate system origin [0, 0, 0] after transformation
	*	\param axes axes of coordinate system after transformation
	*/
	static void transformCoordSysOrigin( GLMatrix4d & mat, GLVector3d &pos, GLVector3d axes[3]);
	
	TNameToIdMap		mNameToIdMap;		///< \brief maps name string to node ID
	TIdToDescriptorMap	mIdToDescriptorMap;	///< \brief maps node ID to node descriptor
	TIdToCompositeMap	mIdToCompositeMap;	///< \brief maps node ID to composite members ( a list of node ID )

	TransformObtainer	mTO;  ///< \brief transform obtainer - gets transformation from model hierarchy node
	
	/** \brief transform matrix (to world space) of parent node for current node */
	GLMatrix4d mParentToWorld;
	
	int mNextId;	///< \brief node ID generator ( starts at 0 and grows +1 )
	
	/** \brief ID of last visited group node ( a value in mNameToIdMap )*/
	int mCurrentGroupID;
public:
	
	/** \brief the value of invalid ID */
	enum { INVALID_ID = -1 };
	
	/**
	*	\brief identificator validity test
	*/
	bool isValidID( int id ) const 
	{ return id != INVALID_ID; }
};

#endif //_MDESC_BUILDER_H_
