/* Maintainer: Petko */

#include "iortransfer_server.h"
#include "log.h"
#include "globals.h"
#include "tinyxml/tinyxml.h"
#include <cstring>

IORTransferServer::IORTransferServer(unsigned short Port, const std::string & IOR, const std::string & HelloString)
{
	mPort = Port;
	mIOR = IOR;
	mVersion = BJS_VERSION_STRING;
	mHelloString = HelloString;
	mMapDescription = "";
	mPlCount = -1;
}

IORTransferServer::IORTransferServer(const IORTransferServer &its)
{
	DIE( "Attempt to copy IORTransferServer instance!" );
}

int IORTransferServer::run()
{
	LOGI << "Starting IOR transmission thread" << endl;
	LOGI << "On port " << mPort << " with IOR " << mIOR << endl;

	mSocket = TcpIpTools::socket( SOCK_DGRAM );

	int retv = TcpIpTools::bind( mSocket, mPort );
	if ( retv != 0 )
	{
		DIE("Cannot bind socket for listening to reply IOR");
	}

	if ( TcpIpTools::setSockBroadcast( mSocket ) != 0 )
	{
		LOGW << "Cannot set socket to listen broadcast! LAN discovery will not find this server!" << endl;
	}

	char * recv_buf = new char[IOR_TRANSFER_HEADER_SIZE + IOR_TRANSFER_DATA_MAX_SIZE + 1];
	
	struct sockaddr_in sa;

	while ( !getStop() )
	{
		int selval = TcpIpTools::selectRead( mSocket, 1000 );
		switch ( selval )
		{
		case 0:			//timeout elapsed - try again
			break;
		case 1:			//we have data
			{
				//receive header first
				memset( recv_buf, 0, IOR_TRANSFER_HEADER_SIZE + IOR_TRANSFER_DATA_MAX_SIZE + 1 );
				
				int really_read = TcpIpTools::recvfrom( mSocket, recv_buf, IOR_TRANSFER_HEADER_SIZE + IOR_TRANSFER_DATA_MAX_SIZE, &sa );
				if ( really_read <= IOR_TRANSFER_HEADER_SIZE )
				{
					LOGW << "Bad data size - received only " << really_read << " bytes" << endl;
					break;
				}

				recv_buf[ IOR_TRANSFER_HEADER_SIZE - 1 ] = 0;

				int data_size = atoi( recv_buf );
				if ( data_size <= 0 || data_size > IOR_TRANSFER_DATA_MAX_SIZE )
				{
					LOGW << "Bad header size - received " << recv_buf 
						 << " value (interpreted as: " << data_size << " )" << endl;
					break;
				}

				if ( really_read - IOR_TRANSFER_HEADER_SIZE != data_size )
				{
					LOGW << "Invalid request - bad header data size. Header reports " 
						 << data_size << " bytes, but I have received " 
						 << (really_read - IOR_TRANSFER_HEADER_SIZE) << " bytes" << endl;
					break;
				}

				char * data = recv_buf + IOR_TRANSFER_HEADER_SIZE;

				std::string response = handleRequest( data );

				if ( TcpIpTools::selectWrite( mSocket, 100 ) == 1 )
				{
					if ( TcpIpTools::sendto( mSocket, response.c_str(), (int)response.length(), &sa ) != (int)response.length() )
					{
						LOGW << "Sending reply did not succeeded - probably too much data" << endl;
						break;
					}
				}
				else
				{
					LOGW << "Socket was not ready for writing for too long" << endl;
				}

				break;
			}
		default:		//error
			{
				delete[] recv_buf;
				TcpIpTools::closesocket( mSocket );
				DIE( "Error while waiting for data for IOR transfer request" );
			}
		}
	}

	TcpIpTools::closesocket( mSocket );

	delete[] recv_buf;

	return 0;
}

std::string IORTransferServer::handleRequest( const char *data )
{
	TiXmlDocument request;
	TiXmlHandle   handle(&request);

	request.Parse( data );

	if ( request.Error() )
	{
		return addHeader( "<?xml version=\"1.0\" ?><ERROR>" + std::string( request.ErrorDesc() ) + "</ERROR>" );
	}

	TiXmlElement* child = handle.FirstChild( "BJS_IOR_REQUEST" ).FirstChild( "VERSION" ).Element();
	if ( child == 0 )
	{
		return addHeader( "<?xml version=\"1.0\" ?><ERROR>Version tag not found</ERROR>" );
	}

	TiXmlText * text = 0;

	for ( TiXmlNode * node = child->FirstChild(); node != 0; node = node->NextSibling()) 
	{
		if ( node->Type() == TiXmlNode::TEXT )
		{
			text = node->ToText();
			break;
		}
	}

	if ( text == 0 )
	{
		return addHeader( "<?xml version=\"1.0\" ?><ERROR>Version tag empty</ERROR>" );
	}

	if ( mVersion != text->ValueStr() )
	{
		return addHeader( "<?xml version=\"1.0\" ?><ERROR>Bad version, server has" + mVersion + "</ERROR>" );
	}

	std::string response = std::string( "<?xml version=\"1.0\" ?><BJS_IOR_RESPONSE>" ) + 
			"<IOR>" + mIOR + "</IOR>" + 
			"<WELCOME>" + mHelloString + "</WELCOME>" +
			"<MAPDESCRIPTION>" + mMapDescription + "</MAPDESCRIPTION>" +
			"<PLAYERCOUNT>" + itos( mPlCount ) + "</PLAYERCOUNT>" +
			std::string( "</BJS_IOR_RESPONSE>" );

	return addHeader( response );
}

std::string IORTransferServer::addHeader(const std::string &data)
{
	char * dsize = new char[IOR_TRANSFER_HEADER_SIZE + 1];
	memset( dsize, ' ', IOR_TRANSFER_HEADER_SIZE );
	dsize[IOR_TRANSFER_HEADER_SIZE] = 0;

#ifdef _MSC_VER
#define snprintf _snprintf
#endif

	snprintf( dsize, IOR_TRANSFER_HEADER_SIZE, "%d", (int)data.length() );
	dsize[strlen(dsize)] = ' ';

	std::string response = std::string(dsize) + data;
	delete dsize;

	return response;
}

void IORTransferServer::setPlayerCount(int PlCount)
{
	mSetMutex.lock();

	mPlCount = PlCount;

	mSetMutex.unlock();
}

void IORTransferServer::setMapDescription(const std::string &MapDesc)
{
	mSetMutex.lock();

	mMapDescription = MapDesc;

	mSetMutex.unlock();
}

