/* MAINTAINER: Lukas */

#include "networkmanager_server.h"
#include "log.h"
#include "application_server.h"
#include "playernetdata_server.h"
#include "corbaexceptionmacros.h"
#include "game_server.h"
#include "netobjects_server.h"
#include <vector>
#include "world.h"

using namespace std;

void NetworkManagerServer::init()
{
	TIN

	mPackerManager.init();

	// !!! ORB starts last - all initialized !!!

	LOGD << "Starting orb..." << endl;

	int port = g_settings.getInt(SETTING_NETWORK_CORBA_LISTEN_PORT);
	
	mOrbManager.startOrb(port);
}

void NetworkManagerServer::clean()
{
	TIN

	// !!! ORB stops first - no more requests !!!

	LOGD << "Stopping orb..." << endl;

	mOrbManager.stopOrb();

	LOGD << "Orb stopped..." << endl;

	mPackerManager.clean();
}

void NetworkManagerServer::addPlayerCallback(Client_var client, int playerID, string errorMsg)
{
	TIN

	bool connectionProblem = false;

	try
	{
		if(playerID != INVALID_PLAYER_ID)
		{
			assertL(! CORBA::is_nil(client));
			assertL(getAppServer().getGame().getPlayer(playerID).canUpdateClient());

			PlayerNetDataServer & nd = dynamic_cast<PlayerNetDataServer &>(
				getAppServer().getGame().getPlayer(playerID).netData());			

			// pass handler into corba
			Server_var ser = nd.getHandler()->_this();
			client->connectCallback(ser, playerID, errorMsg.c_str());
		}
		else
		{
			client->connectCallback(Server::_nil(), playerID, errorMsg.c_str());
		}
	}
	catch(CORBA::SystemException& e)
	{
		SYSEXCEPTION_ERROR("Can't send connect callback to player " << playerID << "(player id)")

		connectionProblem = true;
	}
	catch(CORBA::Exception& e)
	{
		EXCEPTION_ERROR("Can't send connect callback to player " << playerID << "(player id)")

		connectionProblem = true;
	}
	catch(omniORB::fatalException& e)
	{
		OMNIORB_FATAL_EXCEPTION("Can't send connect callback to player " << playerID << "(player id)")
	}
	catch(...)
	{
		UNKNOWN_ERROR("Can't send connect callback to player " << playerID << "(player id)")
	}

	if(connectionProblem && playerID != INVALID_PLAYER_ID)
	{
		LOGD << "Connection lost with player " << playerID << "(player id)." <<
			std::endl;

		getAppServer().getGame().preparePlayerDeletion( playerID );
	}
}

void NetworkManagerServer::gameInfoCallback(const GameInfo & gi, int playerID)
{
	TIN

	bool connectionProblem = false;

	try
	{
		assertL(playerID != INVALID_PLAYER_ID);
		assertL(getAppServer().getGame().getPlayer(playerID).canUpdateClient());

		PlayerNetDataServer & nd = dynamic_cast<PlayerNetDataServer &>(
			getAppServer().getGame().getPlayer(playerID).netData());			

		// data copy
		NETGameInfo netGI;
		netGI.mapName = gi.mapName.c_str();

		// send
		nd.CorbaObject()->gameInfoCallback(netGI);
	}
	catch(CORBA::SystemException& e)
	{
		SYSEXCEPTION_ERROR("Can't send game info callback to player " << playerID << "(player id)")

		connectionProblem = true;
	}
	catch(CORBA::Exception& e)
	{
		EXCEPTION_ERROR("Can't send game info callback to player " << playerID << "(player id)")

		connectionProblem = true;
	}
	catch(omniORB::fatalException& e)
	{
		OMNIORB_FATAL_EXCEPTION("Can't game info connect callback to player " << playerID << "(player id)")
	}
	catch(...)
	{
		UNKNOWN_ERROR("Can't send game info callback to player " << playerID << "(player id)")
	}

	if(connectionProblem)
	{
		LOGD << "Connection lost with player " << playerID << "(player id)." <<
			std::endl;

		getAppServer().getGame().preparePlayerDeletion( playerID );
	}
}

void NetworkManagerServer::makeSyncPoint(int playerID, NETGameState & data)
{
	assertL(playerID != INVALID_PLAYER_ID);

	// pack all data using PackerManager
	mPackerManager.pack(playerID, data);

	// TODO LUKAS debug smazat ----

	dd::iterator it = mDebugData.begin();

	while(it != mDebugData.end())
	{
		data.debug.length(
			data.debug.length() + 1);

		int i = data.debug.length() - 1;

		data.debug[i].length = it->length;
		data.debug[i].param2 = it->id;
		
		// data copy - for sure
		for(int j = 0; j < 1024; ++j)
			data.debug[i].data[j] = it->data[j];

		delete it->data;

		mDebugData.pop_front();

		it = mDebugData.begin();
	}

	// TODO LUKAS debug smazat ----
}

void NetworkManagerServer::sendDataToClients()
{
	TIN

	vector<int> ids = getAppServer().getGame().getValidPlayersID();

	vector<int>::iterator it;

	for(it = ids.begin(); it != ids.end(); ++it)
	{
		Player & player = getAppServer().getGame().getPlayer(*it);

		// skipp clients that cant be updated
		if(! player.canUpdateClient())
			continue;

		//game state that will be send
		NETGameState netData;

		makeSyncPoint(*it, netData);

		bool connectionProblem = false;

		try
		{
			assertL(player.canUpdateClient());
			assertL(
				! CORBA::is_nil(
					(dynamic_cast<PlayerNetDataServer &>(player.netData())).
					CorbaObject()
					)
				);

			(dynamic_cast<PlayerNetDataServer &>(player.netData())).
				CorbaObject()->sendGameState(netData);
		}
		catch(CORBA::SystemException& e)
		{
			SYSEXCEPTION_ERROR("Can't send data to player " << *it << "(player id)")

			connectionProblem = true;
		}
		catch(CORBA::Exception& e)
		{
			EXCEPTION_ERROR("Can't send data to player " << *it << "(player id)")

			connectionProblem = true;
		}
		catch(omniORB::fatalException& e)
		{
			OMNIORB_FATAL_EXCEPTION("Can't send data to player " << *it << "(player id)")
		}
		catch(...)
		{
			UNKNOWN_ERROR("Can't send data to player " << *it << "(player id)")
		}

		if(connectionProblem)
		{
			LOGD << "Connection lost with player " << *it << "(player id)." <<
				std::endl;

			getAppServer().getGame().preparePlayerDeletion(*it);
		}
	}
}

void NetworkManagerServer::sendCommand(string command)
{
	vector<int> ids = getAppServer().getGame().getValidPlayersID();

	vector<int>::iterator it;

	for(it = ids.begin(); it != ids.end(); ++it)
	{
		if(getAppServer().getGame().getPlayer(*it).canUpdateClient())
			(dynamic_cast<PlayerNetDataServer &>(getAppServer().getGame().getPlayer(*it).netData()))
				.addCommand(command);
	}
}

void NetworkManagerServer::sendCommand(string command, int playerID)
{
	assertL(playerID != INVALID_PLAYER_ID);

	if(getAppServer().getGame().getPlayer(playerID).canUpdateClient())
		(dynamic_cast<PlayerNetDataServer &>(getAppServer().getGame().getPlayer(playerID).netData()))
			.addCommand(command);
}

void NetworkManagerServer::sendCollision(const Collision & sc)
{
	vector<int> ids = getAppServer().getGame().getValidPlayersID();

	vector<int>::iterator it;

	for(it = ids.begin(); it != ids.end(); ++it)
	{
		if(getAppServer().getGame().getPlayer(*it).canUpdateClient())
			(dynamic_cast<PlayerNetDataServer &>(getAppServer().getGame().getPlayer(*it).netData()))
				.addCollision(sc);
	}
}

int NetworkManagerServer::numberOfClients()
{
	return getAppServer().getGame().getValidPlayersID().size();
}

void NetworkManagerServer::determinePings()
{
	// TODO LUKAS tady zavolat ping na vsechny hrace(vytvori se joby ktere pingnou)
	// TODO LUKAS udelat aby se tato funkce volala z tasksendchanges asi jednou za 1s
	// potom updaty sbirat v odpovedi a davat je do eventqueue
}

void NetworkManagerServer::collectDataForClients()
{
	World & w = getAppServer().getGame().getWorld();

	worldent_iterator itEnd = w.getWorldEntitiesEnd();

	vector<int> validPIDs = getAppServer().getGame().getValidPlayersID();

	// get and proccess world entities
	for(worldent_iterator wit = w.getWorldEntitiesBegin();
			wit != itEnd; ++wit)
	{
		WorldEntity * ent = *wit;

		assertL(ent);

		// find unpredictable
		if(ent->getUnpredictableFlag())
		{
			vector<int>::iterator pit;

			// set unpredictable in each we(for every player)
			for(pit = validPIDs.begin(); pit != validPIDs.end(); ++pit)
			{
				if(getAppServer().getGame().getPlayer(*pit).canUpdateClient())
				{
					KnownWE * kwe =
						dynamic_cast<PlayerNetDataServer &>
						(getAppServer().getGame().getPlayer(*pit).netData())
						.getKnownWorldEntities().getKnownObject(ent->getID());
					
					if(kwe)
						kwe->setPredictable(false);
				}
			}
		}
	}
}
