/* MAINTAINER: Lukas */

#ifndef _PLAYERNETDATA_SERVER_H
#define _PLAYERNETDATA_SERVER_H

#include "history.h"
#include "player.h"
#include "playernetdata.h"
#include "netobjects_server.h"
#include <string>
#include <list>
#include "log.h"
#include "sendthread.h"
#include "collision_bjs.h"

//enum eStateOfGame;
#include "game.h"

/**
 * Container for known objects
 */
template<class TKnownObject>
class KnownObjects
{
public:
	~KnownObjects()
	{
		while(! mKnownObjects.empty())
		{
			TKnownObject * ko = mKnownObjects.front();
			
			mKnownObjects.pop_front();

			delete ko;
		}
	}

	/**
	 * Remove object that client knows
	 *
	 * \param id corresponding worldentity id
	 */
	void remove(int id)
	{
		TKnownObject * ko = findKnownObject(id);

		if(ko != 0)
		{
			mKnownObjects.remove(ko);

			delete ko;
		}
	}

	/**
	 * Determines whether client knows object
	 *
	 * \return true if client knows object false otherwise
	 */
	bool isObjectKnown(int id)
	{
		return findKnownObject(id) != 0;
	}

	/**
	 * Known objects iterator
	 */
	typedef typename std::list<TKnownObject *>::iterator KnownObjectsIt;

	/**
	 * Return iterator that represents begin of known objects
	 */
	KnownObjectsIt ItBegin()
	{
		return mKnownObjects.begin();
	}

	/**
	 * Return iterator that represents end of known objects
	 */
	KnownObjectsIt ItEnd()
	{
		return mKnownObjects.end();
	}

	/**
	 * Return player's data
	 *
	 * \param id player id
	 */
	TKnownObject * getKnownObject(int id)
	{
		return findKnownObject(id);
	}

protected:
	std::list<TKnownObject *> mKnownObjects; ///< list of known objets

	/**
	 * Helper function. Returns known object by id
	 *
	 * \param id id of KnownObject
	 * \ return known object or 0
	 */
	TKnownObject * findKnownObject(int id)
	{
		KnownObjectsIt it = mKnownObjects.begin();

		while(it != mKnownObjects.end() && (*it)->getID() != id)
			++it;

		// if "it" is at the end that means we didn't find id
		// if "it == mKnownObjects.end()" we can't do "*it"
		if(it == mKnownObjects.end())
			return 0;
		else
			return *it;
	}
};

/**
 * Represents world entity that client knows
 */
class KnownWE
{
public:
	KnownWE(int id, HistoryPoint * netPoint)
	{
		mID = id;
		assertL(netPoint);
		mNetPoint = netPoint;
	}

	~KnownWE()
	{
		delete mNetPoint;
	}

	/**
	 * Returns known object's id
	 *
	 * \return known object's id
	 */
	int getID()
	{
		return mID;
	}

	/**
	 * Returns known object's netPoint(history point that was send last time)
	 *
	 * \return known object's netPoint(history point that was send last time)
	 */
	HistoryPoint * getNetPoint()
	{
		return mNetPoint;
	}

	/**
	 * Sets net point for known object's (netpoint - history point that was send last time)
	 *
	 * \param newNetPoint netpoint that will be set
	 */
	void setNetPoint(HistoryPoint * newNetPoint)
	{
		if(mNetPoint)
			delete mNetPoint;

		assertL(newNetPoint);
		mNetPoint = newNetPoint;
	}

	/**
	 * Returns whether world entity is predictable on client
	 *
	 * \param id worldentity id
	 */
	bool isPredictable()
	{
		return mPredictable;
	}

	/**
	 * Sets predictable state
	 *
	 * \param pState predictable state
	 */
	void setPredictable(bool pState)
	{
		mPredictable = pState;
	}

private:
	// no copy
	KnownWE(KnownWE &) { };
	
	int mID;					///< Object's id that client knows
	bool mPredictable;          ///< Object is predictible and can't be send
	HistoryPoint * mNetPoint;	///< Last point that was send(object with id "id")
};

/**
 * Container for known world entities
 */
class KnownWorldEntities : public KnownObjects<KnownWE>
{
public:
	/**
	 * Add entity that client knows
	 *
	 * \param id corresponding worldentity id
	 * \netPoint history point that will be send
	 */
	void add(int id, HistoryPoint * netPoint);
};

/**
 * Represents player that client knows
 */
class KnownPlayer
{
public:
	KnownPlayer(int id, std::string name)
	{
		mID = id;
		mName = name;
		
		// some values
		// if other then in player constructor then will be one update
		mKills = 0;
		mDeaths = 0;
		mScore = 0;
		mReady = false;

		// TODO LUKAS prepsat na konstantu
		mTeam = -1;
	}

	~KnownPlayer() { };

	/**
	 * Returns known player's id
	 *
	 * \return known player's id
	 */
	int getID()
	{
		return mID;
	}

	/**
	 * Returns last send player's name
	 *
	 * \return last send player's name
	 */
	std::string getName()
	{
		return mName;
	}

	/**
	 * Returns kills count that was last send
	 *
	 * \return kills count that was last send
	 */
	int getKills()
	{
		return mKills;
	}

	/**
	 * Returns deaths count that was last send
	 *
	 * \return deaths count that was last send
	 */
	int getDeaths()
	{
		return mDeaths;
	}

	/**
	 * Returns score that was last send
	 *
	 * \return score that was last send
	 */
	int getScore()
	{
		return mScore;
	}

	/**
	 * Returns team that was last send
	 *
	 * \return team that was last send
	 */
	int getTeam()
	{
		return mTeam;
	}

	/**
	 * Sets number of kills and deaths
	 *
	 * \param kills number of kills
	 * \param deaths number of deaths
	 */
	void update(int kills, int deaths, int score, int team, std::string name, bool ready)
	{
		mKills = kills;
		mDeaths = deaths;
		mScore = score;
		mTeam = team;
		mName = name;
		mReady = ready;
	}

	/**
	 * Whether data that player contains is same as are stored here
	 *
	 * \param p player reference
	 * \return true whew data is same false otherwise
	 */
	bool isSame(Player & p)
	{
		return (p.getKills() == mKills) && (p.getDeaths() == mDeaths) && (p.getScore() == mScore)
			&& (p.getTeam() == mTeam) && (p.getName() == mName) && (p.isReady() == mReady);
	}

private:
	// no copy
	KnownPlayer(KnownPlayer &) { };
	
	int mID; ///< Object's id that client knows

	std::string mName; ///< Last name that was send
	int mKills; ///< Last count that was send
	int mDeaths; ///< Last count that was send
	int mScore; ///< Last count that was send
	int mTeam; ///< Last number that was send
	bool mReady; ///< Last state that was send
};

/**
 * Container for known world entities
 */
class KnownPlayers : public KnownObjects<KnownPlayer>
{
public:
	/**
	 * Add player that client knows
	 *
	 * \param id corresponding worldentity id
	 * \netPoint history point that will be send
	 */
	void add(int id, std::string name);
};

/**
 * Data that are used by network
 */
class PlayerNetDataServer : public PlayerNetData
{
public:
	struct UpdateIntervals
	{
		// ConditionStatus will be send in every X gamestate
		static const int ConditionStatusUpdateInterval = 20;
	};

	PlayerNetDataServer(Client_var client);
	virtual ~PlayerNetDataServer();

	Client_var CorbaObject()
	{
		return mClient;
	};

	/**
	 * Add command that will be send to client
	 *
	 * \param command command that will be send
	 */
	void addCommand(std::string command)
	{
		mCommandList.push_back(command);
	};

	/**
	 * Return all commands in list
	 *
	 * \return all commands in list
	 */
	const std::list<std::string> & getCommands()
	{
		return mCommandList;
	};

	/**
	 * Clear all commands
	 */
	void clearCommands()
	{
		mCommandList.clear();
	}

	/**
	 * Return game state number that will be send next
	 *
	 * \return game state number that will be send next
	 */
	int getGameStateNumber()
	{
		return mGSNumber;
	}

	/**
	 * Increment game state number
	 */
	void incGameStateNumber()
	{
		++mGSNumber;
	}

	/**
	 * Returns KnownWorldEntities object
	 *
	 * \return KnownWorldEntities object
	 */
	KnownWorldEntities & getKnownWorldEntities()
	{
		return mKnownWEs;
	}

	/**
	 * Returns KnownPlayers object
	 *
	 * \return KnownPlayers object
	 */
	KnownPlayers & getKnownPlayers()
	{
		return mKnownPlayers;
	}

	/**
	 * Disable client's handler
	 */
	void disableHandler()
	{
		if(mServer)
		{
			mServer->deactivate();
			mServer->_remove_ref();
			mServer = 0;
		}
	}

	/**
	 * Returns pointer to client calls handler
	 *
	 * \return pointer to client calls handler
	 */
	Server_i * getHandler()
	{
		return mServer;
	}

	/**
	 * Sets client calls handler
	 *
	 * \param server pointer to client calls handler
	 */
	void setHandler(Server_i * server)
	{
		mServer = server;
	}

	/**
	 * Add new job that will be send to client
	 *
	 * param sj job to send
	 */ 
	void sendJob(SendJob * sj)
	{
		assertL(sj);
		mSendQueue->push(sj);
	}

	/**
	 * Add new collision
	 */
	void addCollision(Collision newCollision)
	{
		mCollisionList.push_back(newCollision);
	}

	/**
	 * Return all collisions in list
	 *
	 * \return all collisions in list
	 */
	const std::list<Collision> & getCollisions()
	{
		return mCollisionList;
	};	

	/**
	 * Clear all collisions
	 */
	void clearCollisions()
	{
		mCollisionList.clear();
	}

	/**
	 * Returns last send state of game
	 */
	eStateOfGame getStateOfGame()
	{
		return mStateOfGame;
	}

	/**
	 * Updates state of game
	 */
	void setStateOfGame(eStateOfGame sog)
	{
		mStateOfGame = sog;
	}

private:
	Client_var  mClient;    ///< Corba object that represents client
	Server_i * mServer;     ///< client calls handler
	int mGSNumber;			///< game state number that will be send next send(unique for player)
	std::list<std::string> mCommandList; ///< list of commands send by server
	
	std::list<Collision> mCollisionList; ///< list of all collisions send by server
	
	KnownWorldEntities mKnownWEs;
	KnownPlayers mKnownPlayers;

	SendQueue * mSendQueue;
	SendThread * mSendThread;
	eStateOfGame mStateOfGame;
};

#endif // _PLAYERNETDATA_SERVER_H
