/* MAINTAINER: Petko */
#ifndef _TASK_MOVE_GAME_STATE_H
#define _TASK_MOVE_GAME_STATE_H

#include "task.h"
#include "GLVector.h"
#include "GLQuaternion.h"
#include <ode/ode.h>
#include <list>
#include <vector>

enum BodyType
{
	bGeneral,
	bTrack
};

extern const double MOVEMENT_OPPTRACK_MULTIPLE;	///< \brief Ratio of speed of inner and outer track differences

/**	\brief Implementation of task, that makes objects in game to move
 *
 */
class TaskSimulationServer : public Task
{
public:
	TaskSimulationServer();
	virtual ~TaskSimulationServer();

	/**
	 *	\brief Init internal variables of task.
	 *
	 *	Init parent and copy tick delay from setting to local variable.
	 */
	virtual void init();
	
	/**
	 *	\brief Calls tick() every X milliseconds.
	 *
	 *	\param dtime number of milliseconds since last call - it is used
	 *	\param atime not used. Is here just because of interface.
	 */
	virtual void update(int dtime, int atime);

	/**
	 *	\brief This Method is called every X ms.
	 *
	 *	It is called from update() and simulation of all world should
	 *	be here. The number of milliseconds between calls may be
	 *	modified to change game speed.
	 *
	 *	\param time_change How much time has the server to simulate in seconds
	 *	\param time_stamp The timestamp of the simulated moment
	 */
	virtual void tick( double time_change, int time_stamp );

	/**	\brief Overloades method from Task, it's preparing and adding ODE structures here.
	 *
	 *	\param we WorldEntity that was added
	 */
	void addedWorldObject(WorldEntity* we);

	/**	\brief Overloades method from Task, it's destroying ODE structures here.
	 *
	 *	\param we WorldEntity that is going to be deleted
	 */
	virtual void deletedWorldObject(WorldEntity* we);

	/**	\brief Structure with information about one ODE body
	 */
	typedef struct ODEBodyInfo
	{
		int	 entID;					///< \brief Which entity (or its part) represents this structure
		bool complete;				///< \brief true, if this body represents complete entity
		std::list< dGeomID > geoms;	///< \brief geom that represents this body
		BodyType type;				///< \brief Type of body - needed to set correct contact point parameters in collision

		/**	\brief Union that can store one information about body
		 */
		union
		{
			bool trackTerrContact;	///< \brief True, if wheel touches the terrain
		} spec;
	} ODEBodyInfo;

	friend class ServerPhysicsData;		///< \brief For creating bodies I need access to world and space
	friend class TankPhysicsData;		///< \brief For creating bodies I need access to world and space
	friend class LandscapeEntityPhysicsData;	///< \brief For creating geoms I need access to world and space

private:

		/* ODE stuff */
	dWorldID mODEWorld;			///< \brief ID of our world for ODE
	dSpaceID mODESpace;			///< \brief ID of collision space for ODE
	dSpaceID mODETerrainSpace;	///< \brief ID of terrain collision space
	dSpaceID mODEOtherSpace;	///< \brief ID of other objects

	std::vector< dGeomID > * mTerrainGeoms;			///< \brief stores geom IDs for terrain blocks
	std::vector< dTriMeshDataID > * mTerrainMeshes;	///< \brief stores trimesh IDs for terrain blocks

	dJointGroupID mODEContacts;					///< \brief ODE contact group where contact points will be stored

	static void nearCallback (void *data, dGeomID o1, dGeomID o2);
	
	static void projectileNearCallback (void *data, dGeomID o1, dGeomID o2);

	/**	\brief Initialization of ODE
	 *
	 *	\return True if successful
	 */
	bool initODE();

	/**	\brief Destruction of ODE
	 */
	void destroyODE();

	/**	\brief Adds terrain to collision system of ODE
	 */
	void ODEAddTerrain();

	/**	\brief Adds landscape entities to collision system of ODE
	 */
	void ODEAddLandscapeEntities();

	/** \brief Initializes data for players - now it cancels all command for tanks of players, that are going to be deleted
	 */
	void initPlayers();

	/**	\brief Initializes data for simulation in entities
	 */
	void initSimData();

	/**	\brief Updates speeds that were updated outside ODE
	 */
	void updateSpeeds();

	/**	\brief Changes made during simulation updates to entities
	 */
	void updateSimChanges();

	/**	\brief Check tanks and applies appropriate changes according to the clients input
	 */
	void tankControll();

	/**	\brief Check tanks and shoots if player wants to shoot...
	 *
	 *	\param time_stamp Current simulated time
	 */
	void fireControll( const int time_stamp);

	/**	\brief Clean objects that are to far
	 */
	void cleanupObjects();

	/**	\brief Destroys players, that are marked for deletion
	 */
	void cleanupPlayers();

	/**	\brief Calls updateServer() methods of all objects
	 *
	 *	\param time_stamp Time stamp of "now"
	 */
	void updateObjects( const int time_stamp );

	/**	\brief Updates history - creates new present points and then shifts the history
	 *
	 *	\param time_stamp Time stamp of "now"
	 */
	void updateHistory( int time_stamp );

	/**	\brief Checks if projectiles hit something and applies damage to other entities (tanks)
	 */
	void applyDamage();

	/**	\brief Calculates projectile collision as ray-vs-everything
	 */
	void projectileAdvCollisionDet();

	/**	\brief Drops bonusboxes randomly to the map, according to their frequencies
	 *
	 *	\param time_elapsed Time that elapsed since last simulation tick - usually mTimeDelta
	 */
	void dropBonusBoxes( double time_elapsed );

		/* Some geometrical routines */

	/**	\brief Checks if the point lies in halfspace defined by equation 'a*x+b*y+c*z+d > 0'
	 *
	 *	\param a A parameter 'a' in halfspace equation 'a*x+b*y+c*z+d > 0'
	 *	\param b A parameter 'b' in halfspace equation 'a*x+b*y+c*z+d > 0'
	 *	\param c A parameter 'c' in halfspace equation 'a*x+b*y+c*z+d > 0'
	 *	\param d A parameter 'd' in halfspace equation 'a*x+b*y+c*z+d > 0'
	 *	\param point Point to check for inclusion
	 *	\param sharp Whether include dividing plane into halfspace - true if not - sharp inequality in halfspace equation
	 */
	bool isInHalfSpace( const double a, const double b, const double c, const double d, const GLVector3d& point, const bool sharp = true) const;

	/**	\brief Checks if the point lies in halfspace defined by point (A) and normal
	 *
	 *	\param One point, that lies in the dividing plane
	 *	\param normal Normal vector to the dividing plane
	 *	\param point Point to check for inclusion
	 *	\param sharp Whether include dividing plane into halfspace - true if not - sharp inequality in halfspace equation
	 */
	bool isInHalfSpace( const GLVector3d A, const GLVector3d normal, const GLVector3d point, const bool sharp = true) const
	{
		double d = - ( normal.dot( A ) );
		return isInHalfSpace( normal.x, normal.y, normal.z, d, point, sharp );
	}

	/**	\brief Structure that helps with deleting projectiles
	 */
	struct proj_info
	{
		GLVector3d pos;
		GLQuaterniond ori;
		int type_id;
		int player_id;
	};

	/**	\brief Structure that helps replacing projectiles to the point of collision
	 */
	struct ODE_proj_replace_info
	{
		dBodyID bid;
		GLVector3d pos;
	};
		/* Task management data */

	double	mStepSize;		///< \brief Actual (present) length of one step (dWorld{Quick}Step())
	double	mStepSizeOrig;	///< \brief Original (or basic) length of one step (dWorld{Quick}Step())

	int		mBasicTnkWeight;///< \brief Weight of tank in kgs, that will be treated as 1n ODE
	
	int 	mTimeDelta;		///< \brief Actual (present) time for one tick
	int		mTimeDeltaOrig;	///< \brief Original (or basic) time for one tick

	int		mMaxStepMultiple;	///< \brief The stepsize and timdelta can be maximim this multiple o original values
	int		mCurrentMultiple;	///< \brief current multiple of original stepsize and timedelta

	int		mCPUMaxUtilization;	///< \brief CPU utilization in percents, when simulation tries to switch to next step multiplicator

	int		mSimulatedTime;		///< \brief Total simulated time - used as timestamp

	int		mPerfMeasureTime;	///< \brief Time used for CPU utilization measurement. Utilization is computed as average for 100 steps
	int		mPerfMeasureRuns;	///< \brief Number of ticks included in mPerfMeasureTime

	std::list< int > mBonuses;		///< \brief list of living bonusboxes (their IDs), first is the first that was created
	std::list< int > mProjectiles;	///< \brief list of living projectiles (their IDs), first is the first that was created

	std::list< int > mDeletePlayers;///< \brief list of IDs of players, that I have to delete
};

#endif // _TASK_MOVE_GAME_STATE_H
